/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/


#include "constants.h"
#include "hand.h"

#include "trick.h"
#include "../player/player.h"
#include "../game/game.h"
#include "../party/rule.h"
#include "../misc/setting.h"
#include "../utils/string.h"

/** constructor
 **
 ** @param    player   player this hand belongs to
 **/
Hand::Hand(Player const& player):
  player_(&player)
{ }

/** constructor
 **
 ** @param    player   player this hand belongs to
 ** @param    hand   hand to copy
 **/
Hand::Hand(Player const& player, Hand const& hand):
  requested_position_(hand.requested_position_),
  player_(&player)
{
  this->add(hand);
} // Hand::Hand(Player player, Hand hand)

/** constructor
 **
 ** @param    cards   cards of the hand
 **/
Hand::Hand(vector<Card> const& cards)
{
  this->add(cards);
} // Hand::Hand(vector<Card> cards) :

/** constructor
 **
 ** @param    player   the player who owns the hand
 ** @param    cards   cards of the hand
 **/
Hand::Hand(Player const& player, vector<Card> const& cards) :
  player_(&player)
{
  this->add(cards);
} // Hand::Hand(Player player, vector<Card> cards) :

/** constructor
 **
 ** @param    player           the player who owns the hand
 ** @param    unplayed_cards   cards of the hand (not played)
 ** @param    played_cards   cards of the hand (played)
 **
 ** @todo   make faster
 **/
Hand::Hand(Player const& player,
           vector<Card> const& unplayed_cards,
           vector<Card> const& played_cards) :
  player_(&player)
{
  this->add_played(played_cards);
  this->add(unplayed_cards);
} // Hand::Hand(Player player, vector<Card> unplayed_cards, vector<Card> played_cards) :

/** constructor:
 ** read the hand from the string (one line, cards separated by comma)
 **
 ** @param    text   string (one line) 
 **/
Hand::Hand(string const& text)
{
  auto p = text.begin();
  while (p != text.end()) {
    auto const p2 = find(p, text.end(), ',');
    this->add(Card(string(p, p2)));
    p = p2;
    while (   (p != text.end())
           && (   (*p == ',')
               || (*p == ' ')))
      ++p;
  } // while (p != text.end())
} // Hand::Hand(string const& text)

/** constructor:
 ** read the hand from the stream
 **
 ** @param    istr   Stream with the info
 **/
Hand::Hand(istream& istr)
{
  string line;
  while (istr.good()) {
    // read the lines until there is an empty line or eof
    getline(istr, line);
    if (!line.empty() && (*line.rbegin() == '\r'))
      line.erase(line.end() - 1);
    if (line.empty())
      break;

    // the line is as followed:
    // 0  diamond king		(played 2)
    // (number of the card, card)
    String::word_first_remove(line);

    string card_name;
    card_name = String::word_first(line);
    String::word_first_remove(line);
    card_name += " ";
    card_name += String::word_first(line);
    String::word_first_remove(line);

    HandCard card(*this, card_name);
    if (card.is_empty())
      break;
    this->cards_.push_back(card);
    this->cards_all_.push_back(card);

    String::remove_blanks(line);
    if (line.empty()) {
      // the card has not been played, yet
      this->played_.push_back(UINT_MAX);
    } else { // if (line.empty())
      // the card has been played
      // the line is now like: (played 1)
      if (String::word_first(line) != "(played")
        break;
      String::word_first_remove(line);
      String::remove_blanks(line);

      this->played_.push_back(static_cast<unsigned>(atoi(line.c_str())));
    } // if !(line.empty())

  } // while (istr.good())
} // Hand::Hand(istream& istr)

/** copy constructor
 **
 ** @param    hand   hand to copy
 **/
Hand::Hand(Hand const& hand) :
  Hand(*hand.player_, hand)
{ }

/** copy operator
 **
 ** @param    hand   hand to copy
 **
 ** @return   copied hand
 **/
Hand&
Hand::operator=(Hand const& hand)
{
  if (this == &hand)
    return *this;

  this->cards_.clear();
  this->cards_all_.clear();
  this->played_.clear();
  this->add(hand);
  this->requested_position_ = hand.requested_position();
  this->player_ = hand.player_;

  return *this;
} // Hand& Hand::operator=(Hand hand)

/** converts into HandCards
 **
 ** @result    HandCards
 **/
Hand::operator HandCards() const
{
  return this->cards_;
} // Hand::operator HandCards() const

/** converts into a vector of Cards
 **
 ** @result    HandCards
 **/
Hand::operator vector<Card>() const
{
  return this->cards_;
} // Hand::operator vector<Card>() const

/** @return   iterator to the first card
 **/
Hand::const_iterator
Hand::begin() const
{ return this->cards_.begin(); }

/** @return   iterator to the end of the cards
 **/
Hand::const_iterator
Hand::end() const
{ return this->cards_.end(); }

/** @return   the cards (unplayed)
 **/
HandCards const&
Hand::cards() const
{ return this->cards_; }

/** self check
 ** when an error is found, an ASSERTION is created
 **
 ** @return   wether the self-check was successful (no error)
 **/
bool
Hand::self_check() const
{
  { // check the cards with cards_all
    unsigned c = 0;
    for (unsigned ca = 0; ca < this->cardsnumber_all(); ++ca) {
      if (!this->played(ca)) {
        DEBUG_ASSERTION((this->card(c) == this->card_all(ca)),
                        "Hand::self_check():\n"
                        "  card '" << this->card(c) << "' on position " << c
                        << " is not the same as card all '"
                        << this->card_all(ca) << "' at position " << ca << "\n"
                        << *this);
        c += 1;
      } // if (!this->played(ca))
    }
  } // check the cards with cards_all

  { // check the hand of the cards
    for (auto const& c : this->cards_)
      DEBUG_ASSERTION((&c.hand() == this),
                      "Hand::self_check():\n"
                      "  card '" << c << "' does not belong to this hand.\n"
                      << *this);
    for (auto const& c : this->cards_all_)
      DEBUG_ASSERTION((&c.hand() == this),
                      "Hand::self_check():\n"
                      "  card_all '" << c
                      << "' does not belong to this hand.\n"
                      << *this);
  } // check the hand of the cards

  return true;
} // bool Hand::self_check() const

/** @return   whether the hand is empty (cardsnumber == 0)
 **/
bool
Hand::empty() const
{
  return (this->cardsnumber() == 0);
} // bool Hand::empty() const

/** @return   number of cards still in the hand
 **/
size_t
Hand::cardsnumber() const
{
  return this->cards_.size();
} // size_t Hand::cardsnumber() const

/** @return   number of cards still in the hand
 **/
size_t
Hand::cardsnumber_all() const
{
  return this->cards_all_.size();
} // size_t Hand::cardsnumber_all() const

/** @return   points on the hand
 **/
unsigned
Hand::points() const
{
  unsigned points = 0;

  for (auto const& c : this->cards_)
    points += c.points();

  return points;
} // unsigned Hand::points() const

/** set an unknown card to a known one
 **
 ** @param    card   card to set
 ** @param    n      how many cards to set (default: 1)
 **/
void
Hand::set_known(Card const& card, unsigned n)
{
  if (n == 0)
    return;

  DEBUG_ASSERTION(this->count(Card::unknown) >= n,
                  "Hand::set_unknown(" << card << ", " << n << ")\n"
                  "  not enough unknown cards:\n"
                  << *this);

  for (; n > 0; --n) {
    auto const pos = this->getpos(Card::unknown);
    this->cards_[pos] = card;
    this->cards_all_[this->pos_to_pos_all(pos)] = card;
  }
} // void Hand::set_unknown(Card card, unsigned n = 1)

/** @param    i	number of card
 **
 ** @return   card i of the hand (with remaining cards)
 **/
HandCard const&
Hand::card(unsigned const i) const
{
#ifdef WORKAROUND
  if ( (this->cardsnumber() < 3)
      && (i <= 3) ){
    DEBUG_ASSERTION((i < this->cardsnumber()),
                    "Hand::card(i):\n"
                    "  invalid value 'i' = " << i
                    << " (max = " << this->cardsnumber() << ")\n"
                    "Hand:\n"
                    << *this
                    << "\n"
                    << "You seem to encounter this bug in a poverty while changing the cards.\n"
                    << "Please write us how the cards are lying (or send us a screenshot from the table) and tell us what your last action was (where you have clicked with which button with your mouse)."
                    << "Thank you\n"
                    << "--\n"
                    << "Sie scheinen diesen Fehler waehrend einer Armut bei dem Auswechseln der Karten zu erhalten.\n"
                    << "Bitte schreiben Sie uns zusaetzlich zum Fehlerbericht, wie die Karten liegen (oder machen Sie einen Bildschirmabzug (screenshot) ) und schreiben Sie uns, welche Aktionen Sie zuletzt durchgefuehrt haben (wohin Sie mit welcher Maustaste geklickt haben).\n"
                    << "Vielen Dank");
  } // if (in poverty)
#endif

  DEBUG_ASSERTION((i < this->cardsnumber()),
                  "Hand::card(i):\n"
                  "  invalid value 'i' = " << i
                  << " (max = " << this->cardsnumber() << ")\n"
                  "Hand:\n"
                  << *this);

  DEBUG_ASSERTION(!this->cards_[i].is_empty(),
                  "Hand::card(i):\n"
                  "  empty card " << i << ":\n"
                  << *this);

  return this->cards_[i];
} // HandCard Hand::card(unsigned i) const

/** @param    i	cardnumber
 **
 ** @return   the card number 'i'
 **/
HandCard const&
Hand::card_all(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
                  "Error: file " << __FILE__ << ", line " << __LINE__ << ", "
                  << "function Hand::card_all()\n"
                  << "  illegal value " << i
                  << " (maximal value is: " << this->cardsnumber_all() - 1 << ")");


  DEBUG_ASSERTION(!this->cards_all_[i].is_empty(),
                  "Hand::card_all(i):\n"
                  "  empty card " << i << ":\n"
                  << *this);

  return this->cards_all_[i];
} // HandCard Hand::card_all(unsigned i) const

/** @param    i	cardnumber
 **
 ** @return   whether the card 'i' has been played
 **/
bool
Hand::played(unsigned const i) const
{
  return (this->played_trick(i) != UINT_MAX);
} // bool Hand::played(unsigned const i) const

/** @param    i	number of card
 **
 ** @return   in which trick the card has been played
 **   (UINT_MAX, if not played, yet)
 **/
unsigned
Hand::played_trick(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
                  "Hand::played()\n"
                  "  illegal value i = " << i
                  << " (>= " << this->cardsnumber_all() << ")");

  return this->played_[i];
} // unsigned Hand::played_trick(i)

/** @return   all (not played) cards, but only one of multiple same
 **/
HandCards
Hand::cards_single() const
{
  HandCards cards_single;

  for (auto const& c : this->cards_) {
    // search the card in the til now selected cards
    if (!std::contains(cards_single, c))
      cards_single.push_back(c);
  } // for (c : this->cards_)

  return cards_single;
} // HandCards Hand::cards_single() const

/** @return   the cards and its number
 **/
map<Card, unsigned>
Hand::counted_cards() const
{
  map<Card, unsigned> counter;

  for (auto c : this->cards_)
    counter[c] += 1;

  return counter;
} // map<Card, unsigned> Hand::counted_cards() const

/** @return   the highest value of the hand
 **/
Card::Value
Hand::highest_value() const
{
  return this->cards_.highest_value();
} // Card::Value Hand::highest_value() const

/** @param    tcolor   tcolor to check
 **
 ** @return   the highest value of the given tcolor
 **/
Card::Value
Hand::highest_value(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand:highest_value(tcolor):\n"
                  "  hand does not belongs to a player");

  return this->cards_.highest_value(tcolor);
} // Card::Value Hand::highest_value(Card::TColor tcolor) const

/** @return   the lowest value of the hand
 **/
Card::Value
Hand::lowest_value() const
{
  return this->cards_.lowest_value();
} // Card::Value Hand::lowest_value() const

/** @param    tcolor   tcolor to check
 **
 ** @return   the lowest value of the given tcolor
 **/
Card::Value
Hand::lowest_value(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand:lowest_value(tcolor):\n"
                  "  hand does not belongs to a player");

  return this->cards_.lowest_value(tcolor);
} // Card::Value Hand::lowest_value(Card::TColor tcolor) const

/** @return   the highest card of the hand
 **/
HandCard const&
Hand::highest_card() const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand::highest_card():\n"
                  "  hand does not belongs to a player");

  return this->cards_.highest_card();
} // HandCard Hand::highest_card() const

/** @return   the highest trump of the hand
 **/
HandCard const&
Hand::highest_trump() const
{
  return this->cards_.highest_trump();
} // HandCard Hand::highest_trump() const

/** @param    card_limit   card limit
 **
 ** @return   the highest trump till the card limit
 **/
HandCard const&
Hand::highest_trump_till(Card const& card_limit) const
{
  return this->cards_.highest_trump_till(card_limit);
} // HandCard Hand::highest_trump_till(Card card_limit) const

/** @param    tcolor   tcolor to check the cards of
 **
 ** @return   the highest card of the given tcolor
 **   (Card() if no card of the color is present)
 **/
HandCard const&
Hand::highest_card(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand::highest_card():\n"
                  "  hand does not belongs to a player");

  return this->cards_.highest_card(tcolor);
} // HandCard Hand::highest_card(Card::TColor tcolor) const

/** @return   the lowest trump of the hand
 **/
HandCard const&
Hand::lowest_trump() const
{
  return this->cards_.lowest_trump();
} // HandCard Hand::lowest_trump() const

/** @param    tcolor   tcolor to check the cards of
 **
 ** @return   the lowest card of the given tcolor
 **   (Card() if no card of the tcolor is present)
 **/
HandCard const&
Hand::lowest_card(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand::lowest_card():\n"
                  "  hand does not belongs to a player");

  return this->cards_.lowest_card(tcolor);
} // HandCard Hand::lowest_card(Card::TColor tcolor) const

/** @param    card   card to compare with
 **
 ** @return   whether there exists a higher card than 'card' on the hand
 **/
bool
Hand::higher_card_exists(Card const& card) const
{
  return this->cards_.higher_card_exists(card);
} // bool Hand::higher_card_exists(Card card) const

/** @param    card   card to compare with
 **
 ** @return   whether there exists a lower card than 'card' on the hand
 **/
bool
Hand::lower_card_exists(Card const& card) const
{
  return this->cards_.lower_card_exists(card);
} // bool Hand::lower_card_exists(Card card) const

/** @param    card   card to compare with
 **
 ** @return   number of higher cards then 'card'
 **/
unsigned
Hand::higher_cards_no(Card const& card) const
{
  return this->cards_.higher_cards_no(card);
} // unsigned Hand::higher_cards_no(Card card) const

/** @param    card   card to compare with
 **
 ** @return   the next jabbing card of 'card', empty if no such exists
 **/
HandCard const&
Hand::next_jabbing_card(Card const& card) const
{
  return this->cards_.next_jabbing_card(card);
} // HandCard Hand::next_jabbing_card(Card card) const

/** @param    card   card to compare with
 **
 ** @return   the next higher card then 'card', empty if no such exists
 **/
HandCard const&
Hand::next_higher_card(Card const& card) const
{
  return this->cards_.next_higher_card(card);
} // HandCard Hand::next_higher_card(Card card) const

/** @param    card   card to compare with
 **
 ** @return   'card' or the next higher card then 'card', empty if no such exists
 **/
HandCard const&
Hand::same_or_higher_card(Card const& card) const
{
  return this->cards_.same_or_higher_card(card);
} // HandCard Hand::same_or_higher_card(Card card) const

/** @param    card   card to compare with
 **
 ** @return   the next lower card then 'card', empty if no such exists
 **/
HandCard const&
Hand::next_lower_card(Card const& card) const
{
  return this->cards_.next_lower_card(card);
} // HandCard Hand::next_lower_card(Card card) const

/** @param    card   card to compare with
 **
 ** @return   'card' or the next lower card then 'card', empty if no such exists
 **/
HandCard const&
Hand::same_or_lower_card(Card const& card) const
{
  return this->cards_.same_or_lower_card(card);
} // HandCard Hand::same_or_lower_card(Card card) const

/** @param    trick   trick to check
 **
 ** @return   whether the trick can be jabbed
 **
 ** @todo       improve performance: do not use validcards
 **/
bool
Hand::can_jab(Trick const& trick) const
{
  return this->validcards(trick).higher_card_exists(trick.winnercard());
} // bool Hand::can_jab(Trick trick) const

/** -> result
 ** The pos of the argument is the pos in the remaining cards.
 ** It is searched, what position in all cards 'pos' stands for.
 **
 ** @param    pos   position in the unplayed card
 **
 ** @return   position the played cards
 **/
unsigned
Hand::pos_to_pos_all(unsigned pos) const
{
  if (pos == UINT_MAX)
    return UINT_MAX;

  if ( (pos >= this->cardsnumber())
      || (this->played_.empty()) ) {
    DEBUG_ASSERTION(false,
                    "Error: file " << __FILE__ << ", line " << __LINE__ << ", "
                    << "function Hand::pos_to_pos_all(" << pos << ")\n"
                    << "  illegal value " << pos
                    << " (max value is: " << cardsnumber() - 1 << ")");
    return UINT_MAX;
  }

  unsigned n;
  // translate the position
  for (n = 0, pos += 1; pos > 0; n++)
    if (!this->played(n))
      pos--;

  return (n - 1);
} // unsigned Hand::pos_to_pos_all(pos)


/** -> result
 ** The pos of the argument is the pos in all cards.
 ** It is searched, what position in the remaining cards 'pos' stands for.
 **
 ** @param    pos   position in all card
 **
 ** @return   position in the unplayed cards
 **/
unsigned
Hand::pos_all_to_pos(unsigned pos) const
{
  if (pos == UINT_MAX)
    return UINT_MAX;

  if ( (pos >= cardsnumber_all())
      || (played_.empty()) ) {
    DEBUG_ASSERTION(false,
                    "Error: file " << __FILE__ << ", line " << __LINE__ << ", "
                    << "function Hand::pos_all_to_pos(" << pos << ")\n"
                    << "  illegal value " << pos
                    << " (max value is: " << cardsnumber_all() - 1 << ")");
  }

  unsigned n;
  // translate the position
  for (n = 0, pos += 1; pos > 0; pos--)
    if (!played(pos - 1))
      n++;

  return (n - 1);
} // unsigned Hand::pos_all_to_pos(pos)

/** @return   the player this hand belongs to
 **/
Player const&
Hand::player() const
{
  DEBUG_ASSERTION(this->player_,
                  "Hand::player():\n"
                  "  this->player_ == nullptr");

  return *this->player_;
} // Player Hand::player() const

/** sets the player
 **
 ** @param    player   player of this hand
 **/
void
Hand::set_player(Player const& player)
{
  this->player_ = &player;
} // void Hand::set_player(Player player)

/** @return   the corresponding game
 **/
Game const&
Hand::game() const
{
  return this->player().game();
} // Game Hand::game() const

/** @param    c	card to search
 **
 ** @return   position in the unplayed cards of 'c'
 **   if 'c' is the requested card, its position is returned
 **/
unsigned
Hand::getpos(Card const& c) const
{
  // if the card is in the requested position, take that
  if (   (this->requested_position() < this->cardsnumber())
      && (this->card(requested_position()) == c)
     )
    return this->requested_position();

  // search the card
  for (unsigned i = 0; i < this->cardsnumber(); i++)
    if (c == this->card(i))
      // the card has been found 
      return i;

  if (   (c != Card::unknown)
      && this->contains(Card::unknown))
    return this->getpos(Card::unknown);

  DEBUG_ASSERTION(false,
                  "Hand::getpos(" << c << ")\n"
                  "  card is not in the hand: " << c << '\n'
                  << *this);

  // the card is not in the hand
  return UINT_MAX;
} // unsigned Hand::getpos(Card)

/** @param    c	card to search
 **
 ** @return   position in all cards of 'c'
 **   if 'c' is the requested card, its position is returned
 **/
unsigned
Hand::getpos_all(Card const& c) const
{
  // if the card is in the requested position, take that
  if (   (requested_position() < cardsnumber())
      && (card_all(requested_position()) == c)
     )
    return requested_position();

  // search the card (in played form)
  for (unsigned i = 0; i < cardsnumber_all(); i++)
    if (this->played(i))
      if (c == card_all(i))
        // the card has been found
        return i;

  // search the card
  for (unsigned i = 0; i < cardsnumber_all(); i++)
    if (c == card_all(i))
      // the card has been found
      return i;

  // the card is not in the hand
  return UINT_MAX;
} // unsigned Hand::getpos_all(Card)

/** marks card 'c' as played
 **
 ** @param    c   card to play
 **/
void
Hand::playcard(HandCard const& c)
{
  DEBUG_ASSERTION((this == &c.hand()),
                  "Hand::playcard(" + to_string(c) + ")\n"
                  "  The card does not belong to the hand:\n"
                  << *this);

  auto const pos = this->getpos(c);
  if (this->card(pos) == Card::unknown) {
    this->cards_[pos] = c;
    this->cards_all_[this->pos_to_pos_all(pos)] = c;
  }
  this->playcard(pos);
} // void Hand::playcard(HandCard c)

/** marks card at 'pos' as played
 **
 ** @param    pos   position of the card to play
 **/
void
Hand::playcard(unsigned const pos)
{
  DEBUG_ASSERTION((pos != UINT_MAX),
                  "function Hand::playcard(pos)\n"
                  "  illegal position 'UINT_MAX'"
                  << *this);


  DEBUG_ASSERTION((pos < this->cardsnumber()),
                  "function Hand::playcard(" << pos << ")\n"
                  << "  illegal position"
                  << " (maximal value is " << cardsnumber() << ")");

  DEBUG_ASSERTION(!this->card(pos).is_empty(),
                  "Hand::playcard(" << pos << ")\n"
                  << "  empty card: " << card(pos));

  // mark the card as being played
  if (::game_status == GameStatus::game_poverty_shift) {
    this->played_[this->pos_to_pos_all(pos)] = pos;
  } else { //
    this->played_[this->pos_to_pos_all(pos)]
      = this->cardsnumber_all() - this->cardsnumber();
  }

  this->cards_.erase(this->cards_.begin() + pos);
  this->requested_position_ = UINT_MAX;
} // void Hand::playcard(unsigned pos)

/** marks a played 'card' as not played
 **
 ** @param    c card to unmark
 **/
void
Hand::unplaycard(HandCard const& c)
{
  DEBUG_ASSERTION((this == &c.hand()),
                  "Hand::playcard(c):\n"
                  "  the card does not belong to the hand");

  for (unsigned i = 0; i < this->cardsnumber_all(); i++) {
    if ((this->card_all(i) == c)
        && (this->played(i))) {
      // card found:
      // unmark the card and add it to the hand
      this->played_[i] = UINT_MAX;
      unsigned j = 0; // position in 'cards_'
      for (; i > 0; i--)
        if (!this->played(i - 1))
          j += 1;

      this->cards_.insert(this->cards_.begin() + j, HandCard(*this, c));

      return ;
    } // card found
  } // for (i)

  // the card has not been found
  DEBUG_ASSERTION(false,
                  "Hand::unplaycard(" << c << "):\n"
                  "  card not found:\n"
                  << *this);
} // void Hand::unplaycard(HandCard c)

/** adds 'card' to the hand
 **
 ** @param    card   card to add
 ** @param    n      how many times to add (default: 1)
 **/
void
Hand::add(Card const& card, unsigned const n)
{
  this->cards_.insert(this->cards_.end(), n, HandCard(*this, card));
  this->cards_all_.insert(this->cards_all_.end(), n,HandCard(*this, card));
  this->played_.insert(this->played_.end(), n, UINT_MAX);
} // void Hand::add(Card card, unsigned n = 1)

/** adds the cards 'cards' to the hand
 **
 ** @param    cards   cards to be added
 **/
void
Hand::add(vector<Card> const& cards)
{
  for (auto const& c : cards)
    this->add(c);
} // void Hand::add(vector<Card> cards)

/** adds the cards from 'hand' to the hand
 **
 ** @param    hand   hand to be added
 **/
void
Hand::add(Hand const& hand)
{
  for (unsigned c = 0; c < hand.cardsnumber_all(); c++) {
    if (!hand.played(c))
      this->cards_.push_back(HandCard(*this, hand.card_all(c)));
    this->cards_all_.push_back(HandCard(*this, hand.card_all(c)));
    this->played_.push_back(hand.played_trick(c));
  } // for (c < hand.cardsnumber_all())
} // void Hand::add(Hand const& hand)

/** adds 'card' to the hand
 **
 ** @param    card   card to add
 ** @param    n      how many times to add (default: 1)
 **/
void
Hand::add_played(Card const& card, unsigned const n)
{
  for (unsigned i = 0; i < n; ++i) {
    this->played_.push_back(this->cards_all_.size() - this->cards_.size());
    this->cards_all_.push_back(HandCard(*this, card));
  } // for (unsigned i = 0; i < n; ++i)
} // void Hand::add_played(Card card, unsigned n = 1)

/** adds the played cards 'cards' to the hand
 **
 ** @param    cards   played cards to be added
 **/
void
Hand::add_played(vector<Card> const& cards)
{
  for (auto const& c : cards)
    this->add_played(c);
} // void Hand::add_played(vector<Card> cards)

/** removes 'card' from the hand (one card)
 **
 ** @param    card   card to remove
 **
 ** @return   position of the removed card
 **/
unsigned
Hand::remove(HandCard const& card)
{
  DEBUG_ASSERTION((this->player_ == card.hand().player_),
                  "Hand::remove(card):\n"
                  "  the card does not belong to the same player: " << card << '\n'
                  << *this);

  auto const pos_all = this->getpos_all(card);
  if (   (pos_all == UINT_MAX)
      && card.is_empty())
    return UINT_MAX;

  DEBUG_ASSERTION((pos_all != UINT_MAX),
                  "Hand::remove(card):"
                  "  card '" << card << "' not in the hand: \n"
                  << *this);

  if (!this->played(pos_all))
    this->cards_.erase(this->cards_.begin() + this->pos_all_to_pos(pos_all));
  this->cards_all_.erase(this->cards_all_.begin() + pos_all);
  this->played_.erase(this->played_.begin() + pos_all);

  return pos_all;
} // unsigned Hand::remove(HandCard card)

/** removes the card 'card' from the hand
 **
 ** @param    card   card to remove
 **
 ** @return   position of the removed card
 **/
unsigned
Hand::remove(Card const card)
{
  return this->remove(HandCard(*this, card));
} // unsigned Hand::remove(Card card)

/** removes the cards 'cards' from the hand
 **
 ** @param    cards   cards to remove
 **/
void
Hand::remove(HandCards const& cards)
{
  for (auto const& c : cards)
    this->remove(c);
} // void Hand::remove(HandCards cards)

/** @return   the requested position
 **/
unsigned
Hand::requested_position() const
{ return this->requested_position_; }

/** @return   the requested card (to be played next)
 **/
HandCard
Hand::requested_card() const
{
  if (this->requested_position() == UINT_MAX)
    return HandCard(*this, Card());

  return this->card(this->requested_position());
} // HandCard Hand::requested_card() const

/** @return   the requested position
 **/
unsigned
Hand::requested_position_all() const
{
  return this->pos_to_pos_all(this->requested_position());
} // unsigned Hand::requested_position_all() const

/** request the card in 'position' to be played next
 **
 ** @param    position   position of the requested card
 **/
void
Hand::request_position(unsigned const position)
{
  this->requested_position_ = position;
} // void Hand::request_position(unsigned position)

/** request the card in 'position_all' of all cards to be played next
 **
 ** @param    position_all   position (of all cards) of the requested card
 **/
void
Hand::request_position_all(unsigned const position_all)
{
  if (!this->played(position_all))
    this->request_position(this->pos_all_to_pos(position_all));
} // void Hand::request_position_all(unsigned position_all)

/** request the card 'card' to be played next
 **
 ** @param    card   card to be requested
 **/
void
Hand::request_card(Card const& card)
{
  DEBUG_ASSERTION((this->count(card) > 0),
                  "Hand::request_card(card):\n"
                  "  'card' " << card << " is not in the hand:\n"
                  << *this);

  this->request_position(this->getpos(card));
} // void Hand::request_card(Card card)

/** forget the request
 **/
void
Hand::forget_request()
{
  this->requested_position_ = UINT_MAX;
} // void Hand::forget_request()

/** -> result:
 **
 ** @param    trick   trick
 **
 ** @return   the cards from the hand, that are valid to play in the trick
 **    (of two same cards only one is taken)
 **/
HandCards
Hand::validcards(Trick const& trick) const
{
  HandCards cards;

  for (auto const& c : this->cards_)
    if (c.isvalid(trick))
      if (!std::contains(cards, c))
        cards.push_back(c);

  return cards;
} // HandCards Hand::validcards(trick)

/** -> result:
 **
 ** @param    tcolor   tcolor
 **
 ** @return   the cards from the hand of 'tcolor'
 **    (of two same cards only one is taken)
 **/
HandCards
Hand::cards(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION((this->cardsnumber() > 0),
                  "Hand::cards(tcolor):\n"
                  "  The hand has no cards");

  HandCards cards;

  for (auto const& c : this->cards_)
    if (c.tcolor() == tcolor)
      if (!std::contains(cards, c))
        cards.push_back(c);

  return cards;
} // HandCards Hand::cards(Card::TColor tcolor)

/** @return   the trumps of the hand
 **    (of two same cards only one is taken)
 **/
HandCards
Hand::trumps() const
{
  return this->cards(Card::trump);
} // HandCards Hand::trumps() const

/** @return   whether there is a trump still on the hand
 **/
bool
Hand::hastrump() const
{
  return this->cards_.hastrump();
} // bool Hand::hastrump() const

/** @return   whether there is a color card still on the hand
 **/
bool
Hand::hascolor() const
{
  return this->cards_.hascolor();
} // bool Hand::hascolor() const

/** @return   whether there is a genscher on the hand
 **/
bool
Hand::has_possible_genscher() const
{
  return (   this->player().game().rule()(Rule::Type::genscher)
          && (this->player().game().type() == GameType::normal)
          && (this->count_trump_kings()
              == this->player().game().rule()(Rule::Type::number_of_same_cards)));
} // bool Hand::has_possible_genscher() const

/** @return   whether the hand does contain swines (or can announce them)
 **/
bool
Hand::has_swines() const
{
  auto const& swines = this->game().swines();

  if (swines.swines_announced())
    return (   (swines.swines_owner() == this->player())
            && (this->count_trump_aces() > 0));

  if (   (::game_status == GameStatus::game_reservation)
      && (   !::setting(Setting::Type::show_all_hands)
          || (this->player().type() == Player::Type::human) )
      && !this->player().reservation().swines)
    return false;

  return swines.swines_announcement_valid(this->player());
} // bool Hand::has_swines() const

/** @return   whether the hand does contain hyperswines (or can announce them)
 **
 ** @todo       when all cards are shown: show also hyperswines if another player has the swines
 **/
bool
Hand::has_hyperswines() const
{
  auto const& swines = this->game().swines();

  if (swines.hyperswines_announced())
    return (   (swines.hyperswines_owner() == this->player())
            && (   this->game().rule()(Rule::Type::with_nines)
                 ? this->count_trump_nines()
                 : this->count_trump_kings()
                > 0));

  if (   (::game_status == GameStatus::game_reservation)
      && (   !::setting(Setting::Type::show_all_hands)
          || (this->player().type() == Player::Type::human) )
      && (this->player() == this->game().players().current_player())
      && (   !this->player().reservation().swines
          || !this->player().reservation().hyperswines) )
    return false;

  return swines.hyperswines_announcement_valid(this->player());
} // bool Hand::has_hyperswines() const

/** @param    value	value to count
 **
 ** @return   number of cards with the value 'value'
 **/
unsigned
Hand::count(Card::Value const value) const
{
  return this->cards_.count(value);
} // unsigned Hand::count(Card::Value value)


/** @return   number of cards with value 10 or mor
 **/
unsigned
Hand::count_rich_cards() const
{
  return this->cards_.count_rich_cards();
} // unsigned Hand::count_rich_cards() const

/** @param    tcolor	trump-color to count
 **
 ** @return   number of cards with the trump-color 'tcolor'
 **/
unsigned
Hand::count(Card::TColor const tcolor) const
{
  return this->cards_.count(tcolor);
} // unsigned Hand::count(Card::TColor tcolor)

/** @param    card	card to count
 **
 ** @return   number of cards 'card'
 **/
unsigned
Hand::count(Card const& card) const
{
  return this->cards_.count(card);
} // unsigned Hand::count(Card const& card) const

/** -> result
 ** Note: 'Hand::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    tcolor   trump-color of the card
 ** @param    value   value of the card
 **
 ** @return   number of cards with trump-color 'tcolor' and value 'value'
 **/
unsigned
Hand::count(Card::TColor const tcolor, Card::Value const value) const
{
  return this->cards_.count(tcolor, value);
} // unsigned Hand::count(Card::TColor, Card::Value)

/** -> result
 ** Note: 'Hand::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    color   color to count
 ** @param    game   game
 **
 ** @return   number of cards with color 'color' given the gametype 'gt'
 **/
unsigned
Hand::count(Card::Color const color, Game const& game) const
{
  return this->count(color, game.type(),
                     game.rule()(Rule::Type::dollen));
} // unsigned Hand::count(Card::Color color, Game game) const

/** -> result
 ** Note: 'Hand::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    color   color to count
 ** @param    gametype   the gametype
 **
 ** @return   number of cards with color 'color' given the gametype
 **/
unsigned
Hand::count(Card::Color const color,
            GameType const gametype) const
{
  return this->cards_.count(color, gametype,
                            this->game().rule()(Rule::Type::dollen));
} // unsigned Hand::count(Card::Color color, GameType gametype) const

/** -> result
 ** Note: 'Hand::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    color   color to count
 ** @param    gametype   the gametype
 ** @param    dollen   whether dollen are allowed
 **
 ** @return   number of cards with color 'color' given the gametype and the dollen rule
 **/
unsigned
Hand::count(Card::Color const color,
            GameType const gametype, bool const dollen) const
{
  return this->cards_.count(color, gametype, dollen);
} // unsigned Hand::count(Card::Color color, GameType gametype, bool dollen) const

/** @param    card   card to check
 **
 ** @return   how many times 'card' has been played
 **/
unsigned
Hand::played(Card const& card) const
{
  return (this->count_all(card) - this->count(card));
} // unsigned Hand::played(Card const& card) const

/** @return   number of trumps on the hand
 **/
unsigned
Hand::count_trumps() const
{
  return this->cards_.count_trumps();
} // unsigned Hand::count_trumps() const

/** @param    game_type   the game type
 **
 ** @return   number of trumps on the hand for the game type 'game_type'
 **/
unsigned
Hand::count_trumps(GameType const game_type) const
{
  return this->cards_.count_trumps(game_type);
} // unsigned Hand::count_trumps(GameType game_type) const

/** @param    game_type   the game type
 **
 ** @return   number of fehl on the hand for the game type 'game_type'
 **/
unsigned
Hand::count_fehl() const
{
  return (this->cardsnumber() - this->count_trumps());
} // unsigned Hand::count_fehl() const

/** @param    game_type   game type
 **
 ** @return   number of fehl colors for the game
 **/
unsigned
Hand::count_fehl_colors(GameType const game_type) const
{
  unsigned n = 0;
  for (auto const c : this->game().rule().card_colors()) {
    if (   (c != Game::Cards::trumpcolor(game_type))
        && !this->contains(c))
      n += 1;
  }
  return n;
} // unsigned Hand::count_fehl_colors(GameType game_type) const

/** @param    gametype    game type
 **
 ** @result   number of single tens
 **/
unsigned
Hand::count_single_tens(GameType const gametype) const
{

  auto const dollen = this->game().rule()(Rule::Type::dollen);

  unsigned result = 0;
  for (auto const c : this->game().rule().card_colors()) {
    if (   this->count(c, gametype) <= 2
        && this->count(c, Card::ace) == 0 
        && !Card(c, Card::ace).istrump(gametype, dollen)
        && !Card(c, Card::ten).istrump(gametype, dollen))
      result += this->count(c, Card::ten);
  }

  return result;
} // unsigned Hand::count_single_tens(GameType gametype) const

/** @param    tcolor	card tcolor
 ** @param    value   card value
 **
 ** @return   number of 'Card(tcolor, value)' on the hand with all cards
 **/
unsigned
Hand::count_all(Card::TColor const tcolor, Card::Value const value) const
{
  return count_if(this->cards_all_,
                  [tcolor, value](auto const& c) {
                  return (   (c.tcolor() == tcolor)
                          && (c.value() == value));
                  });
} // unsigned Hand::count_all(Card::TColor tcolor, Card::Value value) const

/** @param    card	card to search
 **
 ** @return   number of 'card' on the hand with all cards
 **/
unsigned
Hand::count_all(Card const& card) const
{
  return count_if(this->cards_all_,
                  [&card](auto const& c) {
                  return (c == card);
                  });
} // unsigned Hand::count_all(Card card) const

/** @return   number of trump aces on the hand with all cards
 **/
unsigned
Hand::count_all_trump_aces() const
{
  return this->count_all(Card::trump, Card::ace);
} // unsigned Hand::count_all_trump_aces() const

/** @return   number of trump kings on the hand
 **/
unsigned
Hand::count_trump_kings() const
{
  return this->cards_.count_trump_kings();
} // unsigned Hand::count_trump_kings() const

/** @return   number of nines on the hand
 **/
unsigned
Hand::count_nines() const
{
  return this->cards_.count_nines();
} // unsigned Hand::numbernines()

/** @return   number of kings on the hand
 **/
unsigned
Hand::count_kings() const
{
  return this->cards_.count_kings();
} // unsigned Hand::numberkings()

/** @return   number of club queens on the hand
 **/
unsigned
Hand::count_club_queens() const
{
  return this->cards_.count_club_queens();
} // unsigned Hand::count_club_queens()

/** @return   number of dolle on the hand
 **/
unsigned
Hand::count_dolle() const
{
  if (!this->game().rule()(Rule::Type::dollen))
    return 0;
  return this->count(Card::dolle);
} // unsigned Hand::count_dolle()

/** @return   number of swines on the hand
 **/
unsigned
Hand::count_swines() const
{
  if (!this->has_swines())
    return 0;
  return this->count_trump_aces();
} // unsigned Hand::count_swines()

/** @return   number of hyperswines on the hand
 **/
unsigned
Hand::count_hyperswines() const
{
  if (!this->has_hyperswines())
    return 0;
  return (this->game().rule()(Rule::Type::with_nines)
          ? this->count_trump_aces()
          : this->count_trump_kings());
} // unsigned Hand::count_hyperswines()

/** @return   number of trump aces on the hand
 **/
unsigned
Hand::count_trump_aces() const
{
  return this->cards_.count_trump_aces();
} // unsigned Hand::count_trump_aces() const

/** @param    game_type   game type
 **
 ** @return   number of trump aces on the hand
 **/
unsigned
Hand::count_trump_aces(GameType const game_type) const
{
  return this->cards_.count_trump_aces(game_type);
} // unsigned Hand::count_trump_aces(GameType game_type) const

/** @return   number of trump nines on the hand
 **/
unsigned
Hand::count_trump_nines() const
{
  return this->cards_.count_trump_nines();
} // unsigned Hand::count_trump_nines() const

/** @return   number of cards to shift in the poverty
 **             UINT_MAX when no poverty is valid
 **/
unsigned
Hand::count_poverty_cards() const
{
  if (!this->has_poverty())
    return UINT_MAX;

  auto const& rule = this->game().rule();

  if (rule(Rule::Type::poverty_shift_only_trump))
    return this->count_trumps();
  else if (!rule(Rule::Type::poverty_fox_do_not_count))
    return rule(Rule::Type::max_number_of_poverty_trumps);
  else if (!rule(Rule::Type::poverty_fox_shift_extra))
    return max(rule(Rule::Type::max_number_of_poverty_trumps),
               this->count_trumps(GameType::poverty));
  else
    // Cannot use 'this->has_swines()' in the case that swines can be
    // announced during the game.
    return max(rule(Rule::Type::max_number_of_poverty_trumps)
               + (this->game().swines().swines_announced()
                  ? 0
                  : this->count_trump_aces(GameType::poverty)
                 ),
               this->count_trumps(GameType::poverty));
} // unsigned Hand::count_poverty_cards() const

/** @return   whether the hand has a poverty
 **/
bool
Hand::has_poverty() const
{
  auto const& rule = this->game().rule();
  if (!rule(Rule::Type::poverty))
    return false;

  return (this->cards_.count_trumps(GameType::poverty)
          - ( (   rule(Rule::Type::poverty_fox_do_not_count)
               && !this->game().swines().swines_announced())
             ? this->count_trump_aces(GameType::poverty)
             : 0)
          <= rule(Rule::Type::max_number_of_poverty_trumps));
} // bool Hand::has_poverty() const

/** @param    card	card to compare with
 **
 ** @return   number of cards greater or equal to 'card'
 **/
unsigned
Hand::count_ge(Card const& card) const
{
  return this->cards_.count_ge(card);
} // unsigned Hand::count_ge(Card card)

/** @param    card	card to compare with
 **
 ** @return   number of cards less or equal to 'card'
 **/
unsigned
Hand::count_le(Card const& card) const
{
  return count_if(this->cards_,
                  [&card](auto const& c) {
                  return (c.less(card) || (c == card));
                  });
} // unsigned Hand::count_le(Card card)

/** @param    card	card to compare with
 **
 ** @return   number of trump cards greater or equal to 'card'
 **/
unsigned
Hand::count_ge_trump(Card const& card) const
{
  return count_if(this->cards_,
                  [&card](auto const& c) {
                  return (c.istrump() && !c.less(card));
                  });
} // unsigned Hand::count_ge_trump(Card card)

/** @param    card	card to compare with
 **
 ** @return   number of trump cards less or equal to 'card'
 **/
unsigned
Hand::count_le_trump(Card const& card) const
{
  return count_if(this->cards_,
                  [&card](auto const& c) {
                  return (   c.istrump() 
                          && (c.less(card) || (c == card)));
                  });
} // unsigned Hand::count_le_trump(Card card)

/** @param    card	trump card to compare with
 **
 ** @return   the relative position of the card (between 0 and 1)
 **/
float
Hand::rel_pos_trump(Card const& card) const
{
  if (this->count_trumps() == 0)
    return 0;
  if (!card.istrump(this->game()))
    return 1;

  unsigned const le = this->count_le_trump(card);
  unsigned const ge = this->count_ge_trump(card);

  return (static_cast<float>(le) / (le + ge));
} // float Hand::rel_pos_trump(Card card)

/** @param    hand	hand to compare with
 **
 ** @return   whether the cards of 'hand' and 'this' are the same
 **/
bool
Hand::has_same_cards(Hand const& hand) const
{
  return hand.cards().equal(this->cards_);
} // bool Hand::has_same_cards(Hand const& hand) const

/** @param    tcolor   tcolor to search
 **
 ** @return   whether 'tcolor' is contained in 'this'
 **/
bool
Hand::contains(Card::TColor const tcolor) const
{
  return this->cards_.contains(tcolor);
} // bool Hand::contains(Card::TColor tcolor) const

/** @param    card   card to search
 **
 ** @return   whether 'card' is contained in 'this'
 **/
bool
Hand::contains(Card const& card) const
{
  return this->cards_.contains(card);
} // bool Hand::contains(Card card) const

/** @param    color   color of the card to search
 ** @param    value   value of the card to search
 **
 ** @return   whether Card(color, value) is contained in 'this'
 **/
bool
Hand::contains(Card::Color const color, Card::Value const value) const
{
  if (   (color == Card::nocardcolor)
      || (color == Card::trump)
      || (value == Card::nocardvalue))
    return false;
  return this->contains({color, value});
} // bool Hand::contains(Card::Color color, Card::Value value) const

/** @param    hand	hand to compare with
 **
 ** @return   whether all cards of 'hand' are contained in 'this'
 **/
bool
Hand::contains(Hand const& hand) const
{
  for (auto const& c : hand.cards_)
    if (this->count(c) < hand.count(c))
      return false;

  return true;
} // bool Hand::contains(Hand hand) const

/** @param    hand	hand to compare with
 **
 ** @return   the cards which are in 'hand' but not in 'this'
 **/
HandCards
Hand::missing_cards(Hand const& hand) const
{
  HandCards cards;
  for (auto const& c : hand.cards_)
    if (this->count(c) < hand.count(c))
      cards.push_back(c);

  return cards;
} // HandCards Hand::missing_cards(Hand hand) const

/** @return   whether there is a trump ten or ace on the hand
 **/
bool
Hand::contains_ten_points_in_trump() const
{
  for (auto const& c : this->cards_) {
    if (   (c.points() >= 10)
        && c.istrump()) {
      return true;
    }
  }
  return false;
}

/** -> result:
 ** compares the hands.
 ** The order of the cards is ignored.
 **
 ** @param    lhs   first hand
 ** @param    rhs   second hand
 **
 ** @return   whether the two hands are equal
 **
 ** @todo   compare the players
 **/
bool
operator==(Hand const& lhs, Hand const& rhs)
{
  if ((lhs.cardsnumber_all() != rhs.cardsnumber_all())
      || (lhs.cardsnumber() != rhs.cardsnumber()))
    return false;

  vector<bool>found(lhs.cardsnumber_all(), false);

  for (unsigned c = 0; c < lhs.cardsnumber_all(); c++) {
    unsigned d;
    for (d = 0; d < rhs.cardsnumber_all(); d++) {
      if (!found[d]
          && (lhs.card_all(c) == rhs.card_all(d))
          && (lhs.played_trick(c) == rhs.played_trick(d)))
        break;
    } // for (d < rhs.cardsnumber_all())
    if (d < rhs.cardsnumber_all())
      found[d] = true;
    else
      return false;
  } // for (c < lhs.cardsnumber_all())

  return true;
} // bool operator==(Hand lhs; Hand rhs)

/** -> result:
 ** compares the hands.
 ** The order of the cards is ignored.
 **
 ** @param    lhs   first hand
 ** @param    rhs   second hand
 **
 ** @return   whether the two hands are different
 **/
bool
operator!=(Hand const& lhs, Hand const& rhs)
{
  return !(lhs == rhs);
} // bool operator!=(Hand lhs; Hand rhs)

/** write the hand on 'ostr'
 **
 ** @param    ostr   stream to write the hand to
 ** @param    hand   hand to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Hand const& hand)
{
  for(unsigned i = 0; i < hand.cardsnumber_all(); i++) {
    ostr << std::hex << i << std::dec << "  " << hand.card_all(i);
    if (hand.played(i))
      ostr << "\t(played " << setw(2) << hand.played_trick(i) << ")";
    ostr << "\n";
  } // for (i < hand.cardsnumber_all())

  return ostr;
} // ostream operator<<(ostream, Hand)
