/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"
#include "hand_cards.h"

#include "../player/player.h"
#include "../game/game.h"
#include "../party/rule.h"

/** constructor
 **
 ** @param    cards   cards of the hand
 **/
HandCards::HandCards(vector<HandCard> const& cards) :
  vector<HandCard>(cards)
{
#ifndef DEBUG_NO_ASSERTS
  // test whether all cards are from the same player
  Player const* const player = &cards[0].player();
  for (HandCards::const_iterator c = this->begin();
       c != this->end();
       ++c)
    DEBUG_ASSERTION((player == &c->player()),
		    "HandCards::HandCards(cards):\n"
		    "  the cards belong to different players: "
		    << player->no() << " and " << c->player().no());
#endif
} // HandCards::HandCards(vector<HandCard> const& cards)

/** constructor
 **
 ** @param    cards   cards of the hand
 **/
HandCards::HandCards(vector<Card> const& cards)
{
  for (auto c : cards)
    this->emplace_back(HandCard(c));
} // HandCards::HandCards(vector<Card> cards)

/** constructor
 **
 ** @param    cards   cards of the hand
 ** @param    hand    corresponding hand
 **/
HandCards::HandCards(Hand const& hand,
		     vector<Card> const& cards)
{
  for (auto c : cards)
    this->emplace_back(HandCard(hand, c));
} // HandCards::HandCards(vector<HandCard> cards)

/** constructor
 **
 ** @param    n   number of cards
 **/
HandCards::HandCards(unsigned const n) :
  vector<HandCard>(n)
{ }

/** reads the cards from the input stream
 **
 ** @param    istr   input stream
 **/
HandCards::HandCards(istream& istr)
{
  while (istr.good()) {
#ifndef OUTDATED
    // DK: 0.6.8
    while (std::isdigit(istr.peek()))
      istr.get();
#endif
    string line;
    std::getline(istr, line);
    if (!line.empty() && (*line.rbegin() == '\r'))
      line.erase(line.end() - 1);
    if (line.empty())
      break;

    this->push_back(HandCard(Card(line)));
  } // while (this->istr->good())
} // HandCards::HandCards(istream& istr)

/** converts the vector in a vector of simple cards
 **
 ** @return   this cards
 **/
HandCards::operator vector<Card>() const
{
  vector<Card> cards;
  for (auto const& c : *this)
    cards.emplace_back(c);
  return cards;
} // HandCards::vector<Card>() const

/** @return   number of cards
 **/
unsigned
HandCards::cardsnumber() const
{
  return this->size();
} // unsigned HandCards::cardsnumber() const

/** -> result
 **
 ** @param    i   number of card
 **
 ** @return   card i of the hand (with remaining cards)
 **/
HandCard const&
HandCards::card(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->size()),
		  "HandCards::card(i):\n"
		  "  invalid value 'i' = " << i
		  << " (max = " << this->size() << ")\n"
		  "HandCards: "
		  << *this);

  return (*this)[i];
} // HandCard HandCards::card(unsigned i) const

/** @return   the player this hand belongs to
 **/
Player const&
HandCards::player() const
{
  DEBUG_ASSERTION(!this->empty(),
                  "HandCards::player()\n"
                  "  no card");
  return (*this)[0].player();
} // Player HandCards::player() const

/** @return   the corresponding game
 **/
Game const&
HandCards::game() const
{
  return this->player().game();
} // Game HandCards::game() const

/** add the card
 **
 ** @param    card   card to add
 **/
void
HandCards::add(HandCard const& card)
{
  this->emplace_back(card);
} // void HandCards::add(HandCard card)

/** add the cards
 **
 ** @param    cards   cards to add
 **/
void
HandCards::add(HandCards const& cards)
{
  this->insert(this->end(), cards.begin(), cards.end());
} // void HandCards::add(HandCards cards)

/** removes 'card' from the vector
 **
 ** @param    card   card to remove
 **/
void
HandCards::remove(Card const& card)
{
  auto d = std::find_if(*this, [card](auto const& d) { return (d == card); });
  DEBUG_ASSERTION((d != this->end()),
                  "HandCards::remove(card):\n"
                  "  card '" << card << "' not found in the cards.\n"
                  << *this);
  this->erase(d);
} // void HandCards::remove(Card card)

/** removes 'cards' from the vector
 **
 ** @param    cards   cards to remove
 **/
void
HandCards::remove(HandCards const& cards)
{
  for (auto const& c : cards) {
    auto d = std::find_if(*this, [&c](auto const& d) { return (d == c); });
    DEBUG_ASSERTION((d != this->end()),
                    "HandCards::remove(cards):\n"
                    "  card '" << c << "' not found in the cards.\n"
                    << *this);
    this->erase(d);
  } // for (c \in cards)
} // void HandCards::remove(HandCards cards)

/** @return   the highest value of the cards
 **/
Card::Value
HandCards::highest_value() const
{
  if (this->cardsnumber() == 0)
    return Card::nocardvalue;

  Card::Value value = this->begin()->value();
  for (auto c : *this) {
    if (c.value() > value)
      value = c.value();
  } // for (c)

  return value;
} // Card::Value HandCards::highest_value() const

/** -> result
 **
 ** @param    tcolor   tcolor to check
 **
 ** @return   the highest value of the given tcolor
 **/
Card::Value
HandCards::highest_value(Card::TColor const tcolor) const
{
  if (this->cardsnumber() == 0)
    return Card::nocardvalue;

  Card::Value value = Card::nine;
  for (auto c : *this) {
    if (   (c.value() > value)
        && (c.tcolor() == tcolor) )
      value = c.value();
  } // for (c)

  return value;
} // Card::Value HandCards::highest_value(Card::TColor tcolor) const

/** @return   the lowest value of the cards
 **/
Card::Value
HandCards::lowest_value() const
{
  if (this->empty())
    return Card::nocardvalue;

  return std::min_element(*this, [](auto a, auto b) {
                          return (a.value() < b.value());
                          })->value();
} // Card::Value HandCards::lowest_value() const

/** -> result
 **
 ** @param    tcolor   tcolor to check
 **
 ** @return   the lowest value of the given tcolor
 **/
Card::Value
HandCards::lowest_value(Card::TColor const tcolor) const
{
  if (this->empty())
    return Card::nocardvalue;

  auto max = std::min_element(*this, [tcolor](auto a, auto b) {
                          return (   (a.tcolor() == tcolor)
                                  && (a.value() < b.value()));
                          });
  if (max == this->end())
    return Card::nocardvalue;
  return max->value();
} // Card::Value HandCards::lowest_value(Card::TColor tcolor) const

/** @return   the highest card of the cards
 **/
HandCard const&
HandCards::highest_card() const
{
  if (this->empty())
    return HandCard::empty;

  auto highest_card = this->begin();

  for (auto c = this->begin() + 1;
       c != this->end();
       ++c) {
    if (highest_card->less(*c))
      highest_card = c;
  } // for (c)

  return *highest_card;
} // HandCard HandCards::highest_card() const

/** @return   the highest trump of the cards
 **/
HandCard const&
HandCards::highest_trump() const
{
  return this->highest_card(Card::trump);
} // HandCard HandCards::highest_trump() const

/** -> result
 **
 ** @param    card_limit   card limit
 **
 ** @return   the highest trump till the card limit
 **/
HandCard const&
HandCards::highest_trump_till(Card const& card_limit) const
{
  return this->highest_card_till(Card::trump, card_limit);
} // HandCard HandCards::highest_trump_till(Card card_limit) const

/** -> result
 **
 ** @param    tcolor   tcolor to check the cards of
 **
 ** @return   the highest card of the given tcolor
 **   (Card() if no card of the color is present)
 **/
HandCard const&
HandCards::highest_card(Card::TColor const tcolor) const
{
  if (this->count(tcolor) == 0)
    return HandCard::empty;

  auto highest_card = this->begin();
  for (;
       highest_card != this->end();
       ++highest_card)
    if (highest_card->tcolor() == tcolor)
      break;

  if (highest_card == this->end())
    return HandCard::empty;

  for (auto c = highest_card + 1;
       c != this->end();
       ++c) {
    if (   (c->tcolor() == tcolor)
        && (highest_card->less(*c)) )
      highest_card = c;
  } // for (c)

  return *highest_card;
} // HandCard HandCards::highest_card(Card::TColor tcolor) const

/** -> result
 **
 ** @param    tcolor       tcolor
 ** @param    card_limit   card limit
 **
 ** @return   the highest card of the given tcolor equal/less the card limit
 **   (Card() if no card is found)
 **/
HandCard const&
HandCards::highest_card_till(Card::TColor const tcolor,
                             Card const& card_limit) const
{
  if (this->count(tcolor) == 0)
    return HandCard::empty;

  auto highest_card = this->begin();
  for (;
       highest_card != this->end();
       ++highest_card)
    if (   (highest_card->tcolor() == tcolor)
        && (   (*highest_card == card_limit)
            || highest_card->less(card_limit)) )
      break;

  if (highest_card == this->end())
    return HandCard::empty;

  for (auto c = highest_card + 1;
       c != this->end();
       ++c) {
    if (   (c->tcolor() == tcolor)
        && (highest_card->less(*c))
        && (   (*c == card_limit)
            || c->less(card_limit)) )
      highest_card = c;
  } // for (c)

  return *highest_card;
} // HandCard HandCards::highest_card_till(Card::TColor tcolor, Card card_limit) const

/** @return   the lowest trump of the cards
 **/
HandCard const&
HandCards::lowest_trump() const
{
  return this->lowest_card(Card::trump);
} // HandCard HandCards::lowest_trump() const

/** -> result
 **
 ** @param    tcolor   tcolor to check the cards of
 **
 ** @return   the lowest card of the given tcolor
 **   (Card() if no card of the tcolor is present)
 **/
HandCard const&
HandCards::lowest_card(Card::TColor const tcolor) const
{
  if (this->count(tcolor) == 0)
    return HandCard::empty;

  auto lowest_card = this->begin();
  for (;
       lowest_card != this->end();
       ++lowest_card)
    if (lowest_card->tcolor() == tcolor)
      break;

  if (lowest_card == this->end())
    return HandCard::empty;

  for (auto c = lowest_card + 1;
       c != this->end();
       ++c) {
    if (   (c->tcolor() == tcolor)
        && (c->less(*lowest_card)) )
      lowest_card = c;
  } // for (c)

  return *lowest_card;
} // HandCard HandCards::lowest_card(Card::TColor tcolor) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   whether there exists a higher card than 'card' on the hand
 **/
bool
HandCards::higher_card_exists(Card const& card) const
{
  for (auto& c : *this)
    if (card.less(c))
      return true;

  return false;
} // bool HandCards::higher_card_exists(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   whether there exists a lower card than 'card' on the hand of the same color
 **/
bool
HandCards::lower_card_exists(Card const& card) const
{
  for (auto& c : *this)
    if (   (c.tcolor() == card.tcolor(this->game()))
        && c.less(card))
      return true;

  return false;
} // bool HandCards::lower_card_exists(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   the number of higher cards then 'card'
 **/
unsigned
HandCards::higher_cards_no(Card const& card) const
{
  return std::count_if(*this, [card](auto const& c) {
                       return card.less(c);
                       });
} // unsigned HandCards::higher_cards_no(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   the next jabbing card of 'card', empty if no such exists
 **/
HandCard const&
HandCards::next_jabbing_card(Card const& card) const
{
  if (this->game().cards().less(card, card))
    return this->same_or_higher_card(card);
  else
    return this->next_higher_card(card);
} // HandCard HandCards::next_jabbing_card(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   the next higher card then 'card', empty if no such exists
 **/
HandCard const&
HandCards::next_higher_card(Card const& card) const
{
  HandCard const* higher_card = nullptr;
  for (auto const& c : *this) {
    if (   card.less(c)
        && (c != card)
        && (   (higher_card == nullptr)
            || c.less(*higher_card)) )
      higher_card = &c;
  } // for (c)

  if (higher_card == nullptr)
    return HandCard::empty;
  return *higher_card;
} // HandCard HandCards::next_higher_card(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   'card' or the next higher card then 'card', empty if no such exists
 **/
HandCard const&
HandCards::same_or_higher_card(Card const& card) const
{
  for (auto const& c : *this) {
    if (c == card)
      return c;
  }

  return this->next_higher_card(card);
} // HandCard HandCards::same_or_higher_card(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   the next lower card then 'card' (in the same color), empty if no such exists
 **/
HandCard const&
HandCards::next_lower_card(Card const& card) const
{
  HandCard const* lower_card = nullptr;
  auto const tcolor = card.tcolor(this->game());
  for (auto const& c : *this) {
    if (   c.less(card)
        && (c != card)
        && (c.tcolor() == tcolor)
        && (   (lower_card == nullptr)
            || lower_card->less(c)) )
      lower_card = &c;
  } // for (c)

  if (lower_card == nullptr)
    return HandCard::empty;
  return *lower_card;
} // HandCard HandCards::next_lower_card(Card card) const

/** -> result
 **
 ** @param    card   card to compare with
 **
 ** @return   'card' or the next lower card then 'card', empty if no such exists
 **/
HandCard const&
HandCards::same_or_lower_card(Card const& card) const
{
  for (auto const& c : *this)
    if (c == card)
      return c;

  return this->next_lower_card(card);
} // HandCard HandCards::same_or_lower_card(Card card) const

/** @return   whether there is a trump still on the hand
 **/
bool
HandCards::hastrump() const
{
  for (auto const& c : *this)
    if (c.istrump())
      return true;

  return false;
} // bool HandCards::hastrump()

/** @return   whether there is a color card still on the hand
 **/
bool
HandCards::hascolor() const
{
  for (auto const& c : *this)
    if (!c.istrump())
      return true;

  return false;
} // bool HandCards::hascolor()

/** -> result
 **
 ** @param    value   value to count
 **
 ** @return   number of cards with the value 'value'
 **/
unsigned
HandCards::count(Card::Value const value) const
{
  return std::count_if(*this, [value](auto const& c) {
                       return (c.value() == value);
                       });
} // unsigned HandCards::count(Card::Value value)

/** @return   number of cards with value 10 or 11
 **/
unsigned
HandCards::count_rich_cards() const
{
  return std::count_if(*this, [](auto const& c) {
                       return (c.value() >= 10);
                       });
} // unsigned HandCards::count_rich_cards() const

/** -> result
 **
 ** @param    tcolor   trump-color to count
 **
 ** @return   number of cards with the trump-color 'tcolor'
 **/
unsigned
HandCards::count(Card::TColor const tcolor) const
{
  return std::count_if(*this, [tcolor](auto const& c) {
                       return (c.tcolor() == tcolor);
                       });
} // unsigned HandCards::count(Card::TColor tcolor)

/** -> result
 **
 ** @param    card   card to count
 **
 ** @return   number of cards 'card'
 **/
unsigned
HandCards::count(Card const& card) const
{
  return std::count_if(*this, [&card](auto const& c) {
                       return (c == card);
                       });
} // unsigned HandCards::count(Card const& card) const

/** -> result
 ** Note: 'HandCards::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    tcolor   trump-color of the card
 ** @param    value   value of the card
 **
 ** @return   number of cards with trump-color 'tcolor' and value 'value'
 **/
unsigned
HandCards::count(Card::TColor const tcolor,
                 Card::Value const value) const
{
  if (this->empty())
    return 0;

  if (tcolor == this->game().cards().trumpcolor())
    return this->count(Card(tcolor, value));

  return std::count_if(*this, [tcolor, value](auto const& c) {
                       return (   (c.tcolor() == tcolor)
                               && (c.value() == value));
                       });
} // unsigned HandCards::count(Card::TColor, Card::Value)

/** -> result
 ** Note: 'HandCards::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    color   color to count
 ** @param    game   game
 **
 ** @return   number of cards with color 'color' given the gametype 'gt'
 **/
unsigned
HandCards::count(Card::Color const color, Game const& game) const
{
  return this->count(color, game.type(), game.rule()(Rule::Type::dollen));
} // unsigned HandCards::count(Card::Color color, Game game) const


/** -> result
 **
 ** @param    card   card to search
 **
 ** @return   whether 'card' is contained in 'this'
 **/
bool
HandCards::contains(Card const& card) const
{
  return (this->count(card) > 0);
} // bool HandCards::contains(Card card) const

/** -> result
 **
 ** @param    color   card color
 ** @param    value   card vallue
 **
 ** @return   whether 'card' is contained in 'this'
 **/
bool
HandCards::contains(Card::Color const color,
                    Card::Value const value) const
{
  return this->contains(Card(color, value));
} // bool HandCards::contains(Card::Color color, Card::Value value) const

/** -> result
 **
 ** @param    tcolor   trump-color to search for
 **
 ** @return   whether there is a card with trump-color 'tcolor' on the hand
 **/
bool
HandCards::contains(Card::TColor const tcolor) const
{
  for (auto const& c : *this) {
    if (c.tcolor() == tcolor)
      return true;
  }

  return false;
} // bool HandCards::contains(Card::TColor tcolor)

/** -> result
 ** Note: 'HandCards::count(Card)' does not cover the case (trump, ace)
 **
 ** @param    color   color to count
 ** @param    gametype   the gametype
 ** @param    dollen   whether dollen are allowed
 **
 ** @return   number of cards with color 'color' given the gametype and the dollen rule
 **/
unsigned
HandCards::count(Card::Color const color,
                 GameType const gametype, bool const dollen) const
{
  return std::count_if(*this, [color, gametype, dollen](auto const& c) {
                       return (   (c.color() == color)
                               && !c.Card::istrump(gametype, dollen));
                       });
} // unsigned HandCards::count(Card::Color color, GameType gametype, bool dollen) const

/** @return   number of trumps on the hand
 **/
unsigned
HandCards::count_trumps() const
{
  return this->count(Card::trump);
} // unsigned HandCards::count_trumps() const

/** -> result
 **
 ** @param    gametype   the game type
 **
 ** @return   number of trumps on the hand for the game type 'gametype'
 **/
unsigned
HandCards::count_trumps(GameType const gametype) const
{
  if (this->empty())
    return 0;
  auto const dollen = this->game().rule()(Rule::Type::dollen);
  return std::count_if(*this, [gametype, dollen](auto const& c) {
                       return (c.Card::istrump(gametype, dollen));
                       });
} // unsigned HandCards::count_trumps(GameType gametype) const

/** @return   number of trump aces on the hand
 **/
unsigned
HandCards::count_trump_aces() const
{
  return this->count(Card::trump, Card::ace);
} // unsigned HandCards::count_trump_aces() const

/** -> result
 **
 ** @param    gametype   game type
 **
 ** @return   number of trump aces on the hand
 **/
unsigned
HandCards::count_trump_aces(GameType const gametype) const
{
  if (this->empty())
    return 0;
  auto const dollen = this->game().rule()(Rule::Type::dollen);
  return std::count_if(*this, [gametype, dollen](auto const& c) {
                       return (c.Card::istrump(gametype, dollen)
                               && (c.value() == Card::ace));
                       });
} // unsigned HandCards::count_trump_aces(GameType gametype) const

/** @return   number of trump kings on the hand
 **/
unsigned
HandCards::count_trump_kings() const
{
  return this->count(Card::trump, Card::king);
} // unsigned HandCards::count_trump_kings() const

/** @return   number of nines on the hand
 **/
unsigned
HandCards::count_nines() const
{
  return this->count(Card::nine);
} // unsigned HandCards::numbernines()

/** @return   number of kings on the hand
 **/
unsigned
HandCards::count_kings() const
{
  return this->count(Card::king);
} // unsigned HandCards::numberkings()

/** @return   number of club queens on the hand
 **/
unsigned
HandCards::count_club_queens() const
{
  return this->count(Card::club_queen);
} // unsigned HandCards::count_club_queens()

/** @return   number of trump nines on the hand
 **/
unsigned
HandCards::count_trump_nines() const
{
  return this->count(Card::trump, Card::nine);
} // unsigned HandCards::count_trump_nines() const

/** -> result
 **
 ** @param    cards   cards to compare with
 **
 ** @return   whether 'cards' is equal to 'this' (without order)
 **/
bool
HandCards::equal(HandCards const& cards) const
{
  for (auto const& c : cards)
    if (this->count(c) != cards.count(c))
      return false;

  return true;
} // bool HandCards::equal(HandCards cards) const

/** @return  number of cards greater or equal to 'card'
 **/
unsigned
HandCards::count_ge(Card const card) const
{
  return count_if(*this,
                  [&card](auto const& c) {
                  return !c.less(card);
                  });
}

/** @return   cards, that do not jab 'card'
 **/
HandCards
HandCards::cards_not_ge(Card const card) const
{
  HandCards cards;
  for (auto const c : *this) {
    if (c.less(card)) {
      cards.add(c);
    }
  }
  return cards;
}

/** write the hand on 'ostr'
 **
 ** @param    ostr   stream to write the hand to
 ** @param    cards   cards to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, HandCards const& cards)
{
  for (auto const& c : cards)
    ostr << c << '\n';

  return ostr;
} // ostream& operator<<(ostream&, HandCards)
