/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "sorted_hand.h"

#include "hand.h"

#include "../misc/setting.h"
#include "../utils/random.h"

/** constructor
 **
 ** @param    hand   the corresponding hand
 **/
SortedHand::SortedHand()
{ }

/** constructor
 **
 ** @param    hand   the corresponding hand
 **/
SortedHand::SortedHand(Hand& hand) :
  hand_(&hand),
  to_sorted_pos_(hand.cardsnumber_all()),
  from_sorted_pos_(hand.cardsnumber_all()),
  mixed(::setting(Setting::Type::cards_order).mixed())
{
  for (unsigned i = 0; i < this->cardsnumber_all(); ++i)
    this->to_sorted_pos_.push_back(i);

  this->set_sorting();
} // SortedHand::SortedHand(Hand& hand)

/** constructor
 **
 ** @param    sorted_hand   the sorted hand to copy
 ** @param    hand   the corresponding hand
 **/
SortedHand::SortedHand(SortedHand const& sorted_hand, Hand& hand) :
  hand_(&hand),
  to_sorted_pos_(sorted_hand.to_sorted_pos_),
  from_sorted_pos_(sorted_hand.from_sorted_pos_),
  mixed(sorted_hand.mixed)
{ }

/** @return   the hand
 **/
Hand const&
SortedHand::hand() const
{ return *this->hand_; }

/** @return   the hand
 **/
SortedHand::operator Hand() const
{ return *this->hand_; }

/** @return   the number of (unplayed) cards
 **/
unsigned
SortedHand::cardsnumber() const
{
  return this->hand_->cardsnumber();
} // unsigned SortedHand::cardsnumber() const

/** @return   the number of (all) cards
 **/
unsigned
SortedHand::cardsnumber_all() const
{
  return this->hand_->cardsnumber_all();
} // unsigned SortedHand::cardsnumber_all() const

/** @param    i	number of card
 **
 ** @return   card i of the hand (with remaining cards)
 **/
HandCard const&
SortedHand::card(unsigned const i) const
{
  return this->hand_->card(this->from_sorted_pos(i));
} // HandCard const& SortedHand::card(unsigned i) const

/** @param    i	cardnumber
 **
 ** @return   the card number 'i'
 **/
HandCard const&
SortedHand::card_all(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
		  "SortedHand::card_all(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber_all() - 1 << ")");

  return this->hand_->card_all(this->from_sorted_pos_all(i));
} // HandCard const& SortedHand::card(unsigned i) const

/** @param    i	cardnumber
 **
 ** @return   whether the card 'i' has been played
 **/
bool
SortedHand::played(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
		  "SortedHand::played(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber_all() - 1 << ")");

  return this->hand_->played(this->from_sorted_pos_all(i));
} // bool SortedHand::played(unsigned i) const

/** @param    i	number of card
 **
 ** @return   in which trick the card has been played
 **   (UINT_MAX, if not played, yet)
 **/
unsigned
SortedHand::played_trick(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
		  "SortedHand::played_trick(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber_all() - 1 << ")");

  return this->hand_->played_trick(this->from_sorted_pos_all(i));
} // unsigned SortedHand::played_trick(unsigned i) const

/** @param    i	number of card
 **
 ** @return   real position of the card (in the played cards)
 **/
unsigned
SortedHand::from_sorted_pos(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber()),
		  "SortedHand::from_sorted_pos(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber() - 1 << ")");

  return this->hand_->pos_all_to_pos(this->from_sorted_pos_all(this->pos_to_pos_all(i)));
} // unsigned SortedHand::from_sorted_pos(unsigned i) const

/** @param    i	number of card
 **
 ** @return   real position of the card (in all cards)
 **/
unsigned
SortedHand::from_sorted_pos_all(unsigned const i) const
{
  DEBUG_ASSERTION((i < this->cardsnumber_all()),
		  "SortedHand::from_sorted_pos_all(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber_all() - 1 << ")");

  return this->from_sorted_pos_[i];
} // unsigned SortedHand::from_sorted_pos_all(unsigned i) const

/** @param    i	number of card
 **
 ** @return   translated position of the card (real pos is 'i')
 **/
unsigned
SortedHand::to_sorted_pos(unsigned const i) const
{
  if (i == UINT_MAX)
    return UINT_MAX;

  DEBUG_ASSERTION((i < this->cardsnumber()),
		  "SortedHand::to_sorted_pos(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber() - 1 << ")");

  return this->hand_->pos_all_to_pos(this->to_sorted_pos_all(this->pos_to_pos_all(i)));
} // unsigned SortedHand::to_sorted_pos(unsigned i) const

/** @param    i	number of card
 **
 ** @return   translated position of the card (real pos is 'i')
 **/
unsigned
SortedHand::to_sorted_pos_all(unsigned const i) const
{
  if (i == UINT_MAX)
    return UINT_MAX;

  DEBUG_ASSERTION((i < this->cardsnumber_all()),
		  "SortedHand::to_sorted_pos_all(i):\n"
		  "  illegal value " << i
		  << " (maximal value is: "
		  << this->cardsnumber_all() - 1 << ")");

  return this->to_sorted_pos_[i];
} // unsigned SortedHand::to_sorted_pos_all(unsigned i) const

/** -> result
 ** The pos of the argument is the pos in the unplayed cards.
 ** It is searched, what position in all cards 'pos' stands for.
 **
 ** @param    pos   position in the unplayed card
 **
 ** @return   position the played cards
 **/
unsigned
SortedHand::pos_to_pos_all(unsigned pos) const
{
  if (pos == UINT_MAX)
    return UINT_MAX;

  unsigned n;
  // translate the position
  for (n = 0, pos += 1; pos > 0; ++n)
    if (!this->hand_->played(this->from_sorted_pos_all(n)))
      pos -= 1;

  return (n - 1);
} // unsigned SortedHand::pos_to_pos_all(unsigned pos)

/** -> result
 ** The pos of the argument is the pos in all cards.
 ** It is searched, what position in the remaining cards 'pos' stands for.
 **
 ** @param    pos   position in all card
 **
 ** @return   position in the unplayed cards
 **/
unsigned
SortedHand::pos_all_to_pos(unsigned pos) const
{
  if (pos == UINT_MAX)
    return UINT_MAX;

  unsigned n;
  // translate the position
  for (n = 0, pos += 1; pos > 0; pos--)
    if (!this->hand_->played(this->from_sorted_pos_all(pos - 1)))
      n++;

  return (n - 1);
} // unsigned SortedHand::pos_all_to_pos(unsigned pos)

/** @return   the requested card (to be played next)
 **/
HandCard
SortedHand::requested_card() const
{
  return this->hand_->requested_card();
} // HandCard Hand::requested_card() const

/** @return   the position of the requested card (to be played next)
 **/
unsigned
SortedHand::requested_position() const
{
  return this->to_sorted_pos(this->hand_->requested_position());
} // unsigned Hand::requested_position() const

/** @return   the position of the requested card (to be played next)
 **/
unsigned
SortedHand::requested_position_all() const
{
  return this->to_sorted_pos_all(this->hand_->requested_position_all());
} // unsigned Hand::requested_position_all() const

/** The pos of the argument is the pos in all cards.
 ** It is searched, what position in the remaining cards 'pos' stands for.
 **
 ** @param    pos_all   position of the card to request (in all cards)
 **/
void
SortedHand::request_position_all(unsigned const pos_all)
{
  this->hand_->request_position_all(this->from_sorted_pos_all(pos_all));
} // void SortedHand::request_position_all(unsigned const pos_all)

/** marks card 'c' as played
 **
 ** @param    c   card to play
 **/
void
SortedHand::playcard(HandCard const& c)
{
  this->hand_->playcard(c);
} // void SortedHand::playcard(HandCard c)

/** marks card at 'pos' as played
 **
 ** @param    pos   position of the card to play
 **/
void
SortedHand::playcard(unsigned const pos)
{
  this->hand_->playcard(this->from_sorted_pos(pos));
} // void SortedHand::playcard(unsigned pos)

/** marks a played 'card' as not played
 **
 ** @param    c card to unmark
 **/
void
SortedHand::unplaycard(HandCard const& c)
{
  this->hand_->unplaycard(c);
} // void SortedHand::unplaycard(HandCard c)

/** adds the cards to the hand
 **
 ** @param    cards   cards to be added
 **/
void
SortedHand::add(vector<Card> const& cards)
{
  // here we know, that the cards are added at the end of the hand
  for (unsigned c = 0; c < cards.size(); c++) {
    this->to_sorted_pos_.push_back(c + this->cardsnumber());
    this->from_sorted_pos_.push_back(c + this->cardsnumber());
  } // for (c < hand.cardsnumber_all())
  this->hand_->add(cards);

  if (!this->mixed)
    this->sort();
} // void SortedHand::add(vector<Card> cards)

/** adds the cards to the hand
 **
 ** @param    cards   cards to be added
 **/
void
SortedHand::add(HandCards const& cards)
{
  this->add(static_cast<vector<Card> >(cards));
} // void SortedHand::add(HandCards cards)

/** removes 'card' from the hand
 **
 ** @param    card   card to remove
 **
 ** @return   position of the removed card
 **/
unsigned
SortedHand::remove(HandCard const& card)
{
  // the position in the hand of the removed card
  auto const pos_all = this->hand_->getpos_all(card);
  // the position in the sorted hand of the removed card
  auto const sorted_pos_all = this->to_sorted_pos_all(pos_all);

  this->hand_->remove(card);

  // update the sorting vectors
  //vector<unsigned> to_sorted_pos_;
  //vector<unsigned> from_sorted_pos_;

  this->from_sorted_pos_.erase(this->from_sorted_pos_.begin()
			       + sorted_pos_all);
  for_each(this->from_sorted_pos_,
           [pos_all](auto& p) {
           if (p > pos_all)
           p -= 1;
           });

  this->to_sorted_pos_.erase(this->to_sorted_pos_.begin() + pos_all);
  for_each(this->to_sorted_pos_,
           [sorted_pos_all](auto& p) {
           if (p > sorted_pos_all)
           p -= 1;
           });

  return sorted_pos_all;
} // unsigned SortedHand::remove(HandCard card)

/** removes the cards 'cards' from the hand
 **
 ** @param    cards   cards to be removeed
 **/
void
SortedHand::remove(HandCards const& cards)
{
  for (auto const& c : cards)
    this->remove(c);
} // void SortedHand::remove(HandCards cards)

/** the hand has been changed - update the sorting
 **/
void
SortedHand::set_sorting()
{
  if (this->mixed)
    this->mix();
  else
    this->sort();
} // void SortedHand::set_sorting()

/** sorts the cards (as set in the settings)
 **
 ** @return   whether the order has changed
 **
 ** @todo       overview, especially the return value
 ** @todo       use std::sort
 **/
bool
SortedHand::sort()
{
  this->from_sorted_pos_.clear();
  for (unsigned i = 0; i < this->cardsnumber_all(); ++i)
    this->from_sorted_pos_.push_back(i);

  bool changed = false;
  if (::setting(Setting::Type::cards_order).sorted()) {
    // a bad bubble sort, but we have at max 12 cards

    for (unsigned c1 = 0; c1 < this->cardsnumber_all(); ++c1)
      for (unsigned c2 = c1 + 1; c2 < this->cardsnumber_all(); ++c2)
	if (HandCard::relative_position(this->hand_->card_all(this->from_sorted_pos_[c1]),
					this->hand_->card_all(this->from_sorted_pos_[c2]))
	    > 0) {
	  std::swap(this->from_sorted_pos_[c1], this->from_sorted_pos_[c2]);
	  changed = true;
	}

    this->set_positions();
  } // bubble sort over all cards

  this->mixed = false;

  return changed;
} // bool SortedHand::sort()

/** unsort the cards (make them the original order)
 **/
void
SortedHand::unsort()
{
  this->from_sorted_pos_.clear();
  for (unsigned i = 0; i < this->cardsnumber_all(); ++i)
    this->from_sorted_pos_.push_back(i);

  this->set_positions();

  this->mixed = true;
} // void SortedHand::unsort()

/** mix the (order of the) cards
 **
 ** @todo    use std::shuffle
 **/
void
SortedHand::mix()
{
  if (!this->mixed) {
    this->unsort();
  } else { // if !(!this->mixed)
    this->from_sorted_pos_.clear();
    for (unsigned i = 0; i < this->cardsnumber_all(); ++i)
      this->from_sorted_pos_.push_back(i);
    if (::setting(Setting::Type::cards_order).mixed()) {
      for (unsigned c = 0; c < this->cardsnumber_all(); c++)
	std::swap(this->from_sorted_pos_[c], this->from_sorted_pos_[::random_value(c + 1)]);
    }

    this->set_positions();
  } // if !(!this->mixed)
} // void SortedHand::mix()

/** calculate the position of the cards from the hand
 **/
void
SortedHand::set_positions()
{
  this->to_sorted_pos_.resize(this->from_sorted_pos_.size());

  for (unsigned c = 0; c < this->cardsnumber_all(); c++)
    this->to_sorted_pos_[this->from_sorted_pos_[c]] = c;
} // void SortedHand::set_positions()

/** compare the hands
 **
 ** @param    lhs   left operand
 ** @param    rhs   right operand
 **
 ** @return   whether the two sorted hands are equal
 **/
bool
operator==(SortedHand const& lhs, SortedHand const& rhs)
{
    return (   (lhs.hand_ == rhs.hand_)
            && (lhs.mixed == rhs.mixed)
            && (lhs.to_sorted_pos_ == rhs.to_sorted_pos_));
} // bool operator==(SortedHand lhs, SortedHand rhs)

/** compare the hands
 **
 ** @param    lhs   left operand
 ** @param    rhs   right operand
 **
 ** @return   whether the two sorted hands differ
 **/
bool
operator!=(SortedHand const& lhs, SortedHand const& rhs)
{
  return !(lhs == rhs);
} 

/** writes the (sorted) hand in 'ostr'
 **
 ** @param    ostr   stream to write to
 ** @param    sorted_hand   hand to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, SortedHand const& sorted_hand)
{
  for (unsigned i = 0; i < sorted_hand.cardsnumber_all(); i++) {
    ostr << std::hex << i << std::dec << "  " << sorted_hand.card_all(i);
    if (sorted_hand.played(i))
      ostr << "\t(played " << setw(2) << sorted_hand.played_trick(i) << ")";
    ostr << "\n";
  } // for (i < hand.cardsnumber_all())

  return ostr;
} // ostream& operator<<(ostream& ostr, SortedHand sorted_hand)
