/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/


#include "constants.h"
#include "trick.h"

#include "hand.h"
#include "../game/game.h"
#include "../player/player.h"
#include "../party/party.h"
#include "../party/rule.h"
#include "../ui/ui.h"
#include "../utils/string.h"

Trick const Trick::empty = Trick();

#define GETLINE(istr, line) \
  std::getline(istr, line);  \
  if (*(line.end() - 1) == '\r') \
  line.erase(line.end() - 1)

/** constructor
 **
 ** @param    startplayer   the player, who has to play the first card
 **/
Trick::Trick(Player const& startplayer) :
  game_(&startplayer.game()),
  no_(startplayer.game().tricks().current_no()),
  startplayer_(startplayer.no())
{ }

/** constructor
 **
 ** @param    startplayer    the number of the startplayer
 ** @param    cards          the cards in the trick (in order of playing)
 ** @param    winnerplayer   the number of the winnerplayer
 **                            default: UINT_MAX
 **/
Trick::Trick(unsigned const startplayer,
             vector<Card> const& cards,
             unsigned const winnerplayer) :
  cards_(cards),
  startplayer_(startplayer),
  winnerplayer_(winnerplayer)
{ }

/** constructor
 **
 ** @param    istr   the stream to read the trick from
 **/
Trick::Trick(istream& istr)
{
  string line;
  GETLINE(istr, line);
  if (String::word_first(line) != "startplayer:")
    return ;

  String::word_first_remove(line);
  this->startplayer_ = static_cast<unsigned>(stoi(line));

  while (istr.good()) {
    GETLINE(istr, line);
    if (line.empty())
      break;
    if (String::word_first(line) == "winner:")
      break;

    // the line is like: 1: club ace
    String::word_first_remove(line);
    String::remove_blanks(line);
    this->cards_.push_back(HandCard(Card(line)));
  } // while (istr.good())
  if (line.empty()
      || !istr.good())
    return ;

  if (String::word_first(line) == "winner:") {
    String::word_first_remove(line);
    this->winnerplayer_ = static_cast<unsigned>(stoi(line));
  } // if (String::word_first(line) == "winner:")
} // Trick::Trick(istream& istr)

/** default constructor
 **/
Trick::Trick() = default;

/** copy constructor
 **/
Trick::Trick(Trick const& trick) = default;

/** copy operator
 **/
Trick&
Trick::operator=(Trick const& trick) = default;

/** @return   iterator to the first card
 **/
Trick::const_iterator
Trick::begin() const
{ return this->cards_.begin(); }

/** @return   iterator to the end of the cards
 **/
Trick::const_iterator
Trick::end() const
{ return this->cards_.end(); }

/** set the game
 **
 ** @param    game   new game of the trick
 **/
void
Trick::set_game(Game const& game)
{
  this->game_ = &game;
  for (auto& c : this->cards_)
    c.set_hand(game.player(c.player().no()).hand());
} // void Trick::set_game(Game game)

/** @return   the game
 **/
Game const&
Trick::game() const
{
  DEBUG_ASSERTION(this->game_,
                  "Trick::game()\n"
                  "  game is not set");
  return *this->game_;
} // Game Trick::game() const

/** @return   the cards
 **/
HandCards const&
Trick::cards() const
{ return this->cards_; }


/** @return   the number of the trick
 **/
unsigned
Trick::no() const
{ return this->no_; }

/** @return   whether the trick is in a trickpile
 **/
bool
Trick::intrickpile() const
{ return this->intrickpile_; }

/** self check
 ** when an error is found, an ASSERTION is created
 **
 ** @return   wether the self-check was successful (no error)
 **/
bool
Trick::self_check() const
{
  DEBUG_ASSERTION(this->game_,
                  "Trick::self_check()\n"
                  "  no game");

  DEBUG_ASSERTION((this->no_ <= this->game_->tricks().current_no()),
                  "Trick::self_check()\n"
                  "  number not valid: " << this->no_);

  DEBUG_ASSERTION((this == &this->game_->tricks().trick(this->no_)),
                  "Trick::self_check()\n"
                  "  Not the trick the number says to be");

  DEBUG_ASSERTION(this->startplayerno() != UINT_MAX,
                  "Trick::self_check()\n"
                  "  no startplayer");

  DEBUG_ASSERTION((!this->intrickpile_
                   || (this->winnerplayerno() != UINT_MAX)),
                  "Trick::self_check()\n"
                  "  in trickpile but no winnerplayer");

  return true;
} // bool Trick::self_check() const

/** output of the trick
 **
 ** @param    ostr   output stream
 ** @param    trick   Trick to be put in the output stream
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Trick const& trick)
{
  ostr << "startplayer: " << trick.startplayerno() << "\n";
  for(unsigned i = 0; i < trick.actcardno(); i++) {
    ostr
      << (trick.game_
          ? trick.playerno_of_card(i)
          : (trick.startplayerno() + i))
      << ": "
      << trick.card(i) << "\n";
  }

  if (trick.isfull())
    ostr << "winner: " << trick.winnerplayerno() << "\n";

  return ostr;
} // ostream& operator<<(ostream& ostr, Trick trick)

/** adds the card 'c' to the trick
 **
 ** @param    card   card, which should be added to this trick
 **
 ** @return   trick with c as last card
 **/
void
Trick::add(HandCard const& card)
{
  DEBUG_ASSERTION(!card.is_empty(),
                  "Trick::operator+=(HandCard):\n"
                  "  shall add empty card to the trick");

  DEBUG_ASSERTION(   !this->game_
                  || !this->isfull(),
                  "Trick::operator+=(HandCard):\n"
                  "  trick is already full");

  DEBUG_ASSERTION((   !this->game_
                   || (&card.player() == &this->actplayer())),
                  "Trick::operator+=(HandCard):\n"
                  "  Card is from the wrong player "
                  << card.player().no()
                  << " (should be " << this->actplayer().no() << ")\n"
                  "  Pointer: " << &card.player() << " != "
                  << &this->actplayer());

  if (this->game_) {
    // update the winnerplayer
    if (this->isstartcard()) {
      this->winnerplayer_ = this->startplayer_;
    } else {
      if (this->winnercard().less(card)) {
        this->winnerplayer_ = this->actplayer().no();
      }
    } // if !(this->isstartcard())
  } // if (this->game_)

  // add the card to the trick
  this->cards_.push_back(card);
} // void Trick::add(Card c)

/** -> result
 **
 ** @param    c   number of card
 **
 ** @return   the trick till the given card
 **/
Trick
Trick::till_card(unsigned const c) const
{
  DEBUG_ASSERTION((c <= this->actcardno()),
                  "Trick::till_card(" << c << ")\n"
                  "  the trick contains only " << this->actcardno() << " cards");

  Trick t(this->startplayer());
  for (unsigned i = 0; i <= c; ++i)
    t.add(this->card(i));

  return t;
} // Trick Trick::till_card(unsigned c) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the trick till the given player
 **/
Trick
Trick::till_player(Player const& player) const
{
  DEBUG_ASSERTION(this->has_played(player),
                  "Trick::till_player(" << player.no() << ")\n"
                  "  the player has not played so far");

  Trick t(this->startplayer());
  auto const n = this->cardno_of_player(player);
  for (unsigned i = 0; i <= n; ++i)
    t.add(this->card(i));

  return t;
} // Trick Trick::till_player(Player player) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the trick till before the given player
 **/
Trick
Trick::before_player(Player const& player) const
{
  DEBUG_ASSERTION(this->has_played(player),
                  "Trick::before_player(" << player.no() << ")\n"
                  "  the player has not played so far");

  Trick t(this->startplayer());
  auto const n = this->cardno_of_player(player);
  for (unsigned i = 0; i < n; ++i)
    t.add(this->card(i));

  return t;
} // Trick Trick::before_player(Player player) const

/** @return   the trick without the last played card
 **/
Trick
Trick::before_last_played_card() const
{
  DEBUG_ASSERTION(!this->isempty(),
                  "Trick::before_before_last_played_card()\n"
                  "  the trick is empty");

  Trick t(this->startplayer());
  for (unsigned i = 0; i + 1 < this->actcardno(); ++i)
    t.add(this->card(i));

  return t;
} // Trick Trick::before_last_played_card() const

/** -> result
 **
 ** @param    c   number of card for the result
 **
 ** @return   card c of actual trick
 **/
HandCard const&
Trick::card(unsigned const c) const
{
  DEBUG_ASSERTION((c < this->actcardno()),
                  "Trick::card(c):\n"
                  "  'c' is to great (" << c << ">=" << this->actcardno()
                  << ")");

  return this->cards_[c];
} // HandCard const& Trick::card(unsigned c) const

/** -> result
 **
 ** @param    player   player whose card is returned
 **
 ** @return   card of 'player' of actual trick
 **/
HandCard const&
Trick::card_of_player(Player const& player) const
{
  return ((player.no() >= this->startplayerno())
          ? this->card(player.no()
                       - this->startplayerno())
          : this->card(player.no()
                       + this->game().playerno()
                       - this->startplayerno())
         );
} // HandCard Trick::card_of_player(Player player) const

/** -> result
 **
 ** @param    c   number of card for the result
 **
 ** @return   the player, who has played the 'c'th card
 **/
Player const&
Trick::player_of_card(unsigned const c) const
{
  return this->game().player(this->playerno_of_card(c));
} // Player const& Trick::player_of_card(unsigned c) const

/** -> result
 **
 ** @param    c   number of card for the result
 **
 ** @return   the nubmer of the player, who has played the 'c'th card
 **/
unsigned
Trick::playerno_of_card(unsigned const c) const
{
  return ((this->startplayerno() + c
           < this->game().playerno())
          ? (c + this->startplayerno())
          : (c + this->startplayerno()
             - this->game().playerno())
         );
} // unsigned Trick::playerno_of_card(unsigned c) const

/** -> result
 **
 ** @param    player   player, whose cardno is returnes
 **
 ** @return   the number of the card, the player has played
 **/
unsigned
Trick::cardno_of_player(Player const& player) const
{
  return ((this->startplayerno()
           <= player.no())
          ? (player.no() - this->startplayerno())
          : (player.no()
             + this->game().playerno()
             - this->startplayerno())
         );
} // unsigned Trick::cardno_of_player(Player player) const

/** -> result
 **
 ** @param    player   player to check
 **
 ** @return   whether the player has already played a card
 **/
bool
Trick::has_played(Player const& player) const
{
  return (this->cardno_of_player(player) < this->actcardno());
} // bool Trick::has_played(Player player) const

/** -> result
 **
 ** @param    card   card to check
 **
 ** @return   whether the card is in the trick
 **/
bool
Trick::contains(Card const& card) const
{
  return (find(this->cards_, card) != this->cards_.end());
} // bool Trick::contains(Card card) const

/** @return   whether the trick contains a trump
 **/
bool
Trick::contains_trump() const
{
  return (find_if(this->cards_,
                  [](auto const& card) 
                  { return card.istrump();
                  })
          != this->cards_.end());
} // bool Trick::contains_trump() const

/** @return   whether the trick contains a fox
 **/
bool
Trick::contains_fox() const
{
  return (find_if(this->cards_,
                  [](auto const& card) 
                  { return card.isfox();
                  })
          != this->cards_.end());
} // bool Trick::contains_fox() const

/** -> result
 **
 ** @param    game    corresponding game
 **
 ** @return   whether the trick contains a possible extrapoint
 **/
bool
Trick::contains_possible_extrapoint(Game const& game) const
{
  if (game.is_solo())
    return false;

  auto const& rule = game.rule();
  auto const is_last_trick = game.tricks().is_last_trick();

  if (   rule(Rule::Type::extrapoint_catch_fox)
      && this->contains(Card::fox)
      && !this->game().swines().swines_announced())
    return true;
  if (   rule(Rule::Type::extrapoint_fox_last_trick)
      && is_last_trick
      && this->winnercard().isfox())
    return true;
  if (   rule(Rule::Type::extrapoint_catch_fox_last_trick)
      && is_last_trick
      && this->contains(Card::fox) )
    return true;
  if (   rule(Rule::Type::extrapoint_charlie)
      && is_last_trick
      && (this->winnercard() == Card::charlie) )
    return true;
  if (   rule(Rule::Type::extrapoint_catch_charlie)
      && is_last_trick
      && this->contains(Card::charlie) )
    return true;
  if (   rule(Rule::Type::extrapoint_dolle_jabs_dolle)
      && (this->winnercard() == Card::dolle) )
    return true;
  if (rule(Rule::Type::extrapoint_heart_trick)) {
    unsigned c = 0;
    for (c = 0; c < this->actcardno(); c++)
      if (this->card(c).tcolor() != Card::heart)
        break;
    if (c == this->actcardno())
      return true;
  }

  return false;
} // bool Trick::contains_possible_extrapoint(Game game) const

/** @return   the number of the startplayer
 **/
unsigned
Trick::startplayerno() const
{
  return this->startplayer_;
} // unsigned Trick::startplayerno() const

/** @return   the startplayer
 **/
Player const&
Trick::startplayer() const
{
  DEBUG_ASSERTION(this->startplayerno() != UINT_MAX,
                  "Trick::startplayer():\n"
                  "  startplayerno() == NULL != UINT_MAX");

  return this->game().player(this->startplayerno());
} // Player const& Trick::startplayer() const

/** @return   the last player
 **/
Player const&
Trick::lastplayer() const
{
  DEBUG_ASSERTION(this->startplayerno() != UINT_MAX,
                  "Trick::lastplayer():\n"
                  "  startplayerno() == N != UINT_MAX");

  return this->game().player((this->startplayerno()
                              + this->game().playerno() - 1)
                             % this->game().playerno());
} // Player const& Trick::lastplayer() const

/** @return   the last player
 **/
Player const&
Trick::secondlastplayer() const
{
  DEBUG_ASSERTION(this->startplayerno() != UINT_MAX,
                  "Trick::secondlastplayer():\n"
                  "  startplayerno() == N != UINT_MAX");

  return this->game().player((this->startplayerno()
                              + this->game().playerno() - 2)
                             % this->game().playerno());
} // Player const& Trick::secondlastplayer() const

/** set the startplayer
 **
 ** @param    startplayer   new startplayer
 **
 ** @return   the startplayer
 **/
void
Trick::set_startplayer(Player const& startplayer)
{
  DEBUG_ASSERTION(this->startplayerno() == UINT_MAX,
                  "Trick::startplayer_set(startplayer):\n"
                  "  startplayer already set");

  this->startplayer_ = startplayer.no();
} // void Trick::set_startplayer(Player startplayer)

/** set the startplayer
 **
 ** @param    startplayerno   number of new startplayer
 **
 ** @return   the startplayer
 **/
void
Trick::set_startplayer(unsigned const startplayerno)
{
  DEBUG_ASSERTION(this->startplayerno() == UINT_MAX,
                  "Trick::startplayer_set(startplayer):\n"
                  "  startplayer already set");

  this->startplayer_ = startplayerno;
} // void Trick::set_startplayer(unsigned startplayerno)

/** @return   the startcard
 **/
HandCard const&
Trick::startcard() const
{
  DEBUG_ASSERTION(!this->isstartcard(),
                  "Trick::startcard():\n"
                  "  trick is empty");

  return this->cards_[0];
} // HandCard Trick::startcard() const

/** @return   whether the trick is empty
 **/
bool
Trick::isstartcard() const
{
  return this->cards_.empty();
} // bool Trick::isstartcard() const

/** @return   whether the trick is empty
 **/
bool
Trick::isempty() const
{
  return this->cards_.empty();
} // bool Trick::isempty() const

/** @return   whether the trick is full
 **/
bool
Trick::isfull() const
{
  if (!this->game_)
    return true;
  return (this->remainingcardno() == 0);
} // bool Trick::isfull() const

/** @return   whether the trick lacks only one card
 **/
bool
Trick::islastcard() const
{
  return (this->remainingcardno() == 1);
} // bool Trick::islastcard() const

/** moves the trick in the trickpile
 **
 ** @bug   const cast
 **/
void
Trick::move_in_trickpile()
{
  // Bug: const cast
  const_cast<Player&>(this->winnerplayer()).move_in_trickpile(*this);

  this->intrickpile_ = true;

  if (!this->game().isvirtual())
    ::ui->trick_move_in_pile();
} // void Trick::move_in_trickpile()

/** @return   the number of the acting card
 **   (that is, how many cards there are in the tricks)
 **/
unsigned
Trick::actcardno() const
{
  return this->cards_.size();
} // unsigned Trick::actcardno() const

/** @return   the number of remaining cards to play in the trick
 **/
unsigned
Trick::remainingcardno() const
{
  return (this->game().playerno() - this->actcardno());
} // unsigned Trick::remainingcardno() const

/** @return   the acting player
 **/
Player const&
Trick::actplayer() const
{
  return this->game().player(this->actplayerno());
} // Player const& Trick::actplayer() const

/** @return   the acting player number
 **/
unsigned
Trick::actplayerno() const
{
  return ((this->startplayerno() + this->actcardno())
          % this->game().playerno());
} // unsigned Trick::actplayerno() const

/** @return   the points in the trick
 **/
unsigned
Trick::points() const
{
  unsigned points = 0;
  for (auto const& c : this->cards_)
    points += (c.is_unknown() ? 0 : c.points());
  return points;
} // unsigned Trick::points() const

/** -> result
 **
 ** @param    card   card which is checked
 **
 ** @return   whether it is valid to play 'card' with its hand
 **/
bool
Trick::isvalid(HandCard const& card) const
{
  return card.isvalid(*this);
} // bool Trick::isvalid(HandCard card) const

/** -> result
 **
 ** @param    card   card which is checked
 ** @param    hand   hand of the card
 **
 ** @return   whether it is valid to play 'card' with the hand 'hand'
 **/
bool
Trick::isvalid(Card const& card, Hand const& hand) const
{
  return this->isvalid(HandCard(hand, card));
} // bool Trick::isvalid(Card card, Hand hand) const

/** @return   the number of the winnerplayer
 **/
unsigned
Trick::winnerplayerno() const
{
  return this->winnerplayer_;
} // unsigned Trick::winnerplayerno() const

/** @return   the player which wins this trick
 **/
Player const&
Trick::winnerplayer() const
{
  // for Game::numberoftricks_of_player()
  if (this->winnerplayer_ == UINT_MAX)
    return this->startplayer();

  DEBUG_ASSERTION(this->winnerplayer_ != UINT_MAX,
                  "Trick::winnerplayer():\n"
                  "  winnerplayerno() == UINT_MAX");

  return this->game().player(this->winnerplayer_);
} // Player const& Trick::winnerplayer() const

/** @return   the team which wins the trick
 **/
Team
Trick::winnerteam() const
{
  return this->winnerplayer().team();
} // Team Trick::winnerteam() const

/** @return   card that `makes` the trick
 **/
HandCard const&
Trick::winnercard() const
{
  if (this->isempty())
    return HandCard::empty;

  return this->card_of_player(this->winnerplayer());
} // HandCard Trick::winnercard() const

/** @return   cardno that `makes` the trick
 **/
unsigned
Trick::winnercardno() const
{
  return this->cardno_of_player(this->winnerplayer());
} // unsigned Trick::winnercardno() const

/** @return   card which wins the trick without taking the last card played into account
 **/
HandCard const&
Trick::winnercard_before() const
{
  if (this->actcardno() <= 1)
    return HandCard::empty;

  auto const* winnercard = &this->card(0);
  for (unsigned c = 1; c < this->actcardno() - 1; ++c)
    if (winnercard->less(this->card(c)))
      winnercard = &this->card(c);

  return *winnercard;
} // HandCard Trick::winnercard_before() const

/** -> result
 **
 ** @param    cardno   card to analyse
 **
 ** @return   whether the card jabs the cards before
 **/
bool
Trick::jabs_cards_before(unsigned const cardno) const
{
  DEBUG_ASSERTION(cardno < this->actcardno(),
                  "Trick::jabs_cards_before(cardno = " << cardno << ")\n"
                  "  Card is not played, yet.\n"
                  "  Trick:\n" << *this);

  auto const& card = this->card(cardno);
  for (unsigned c = 0; c < cardno; ++c)
    if (!this->card(c).less(card))
      return false;

  return true;
} //  bool Trick::jabs_cards_before(unsigned cardno) const

/** -> result
 **
 ** @param    player   player to analyse
 **
 ** @return   whether the player jabs the cards before
 **/
bool
Trick::jabs_cards_before(Player const& player) const
{
  DEBUG_ASSERTION(this->has_played(player),
                  "Trick::jabs_cards_before(player = " << player.no() << ")\n"
                  "  Player has not played, yet.\n"
                  "  Trick:\n" << *this);

  return -this->jabs_cards_before(this->cardno_of_player(player));
} // bool Trick::jabs_cards_before(Player player) const

/** @return   whether this is the last trick in the game
 **/
bool
Trick::islast() const
{
  return (this->no() == this->game().tricks().max_size() - 1);
} // bool Trick::islast() const

/** -> result
 **
 ** @param    card   card to test
 **
 ** @return   whether the trick will be jabbed by this card
 **/
bool
Trick::isjabbed(Card const& card) const
{
  if (this->isempty())
    return true;
  else
    return this->winnercard().less(card);
} // bool Trick::isjabbed(Card card) const

/** -> result
 ** take possible special cards into account (p.e. swines, that are not announced)
 **
 ** @param    card   card to test
 **
 ** @return   whether the trick will be jabbed by this card
 **/
bool
Trick::isjabbed(HandCard const& card) const
{
  if (this->isempty())
    return true;
  else {
    if (card.possible_hyperswine())
      return true;
    else if (card.possible_swine())
      return !this->winnercard().ishyperswine();
    else
      return this->winnercard().less(card);
  }
} // bool Trick::isjabbed(HandCard card) const

/** -> result
 ** take possible special cards into account (p.e. swines, that are not announced)
 **
 ** @param    card   card to test
 ** @param    hand   corresponding hand
 **
 ** @return   whether the trick will be jabbed by this card
 **/
bool
Trick::isjabbed(Card const& card, Hand const& hand) const
{
  return this->isjabbed(HandCard(hand, card));
} // bool Trick::isjabbed(Card card, Hand hand) const

/** -> result
 **
 ** @param    cardno_a   number of the first card
 ** @param    cardno_b   number of the second card
 **
 ** @return   whether the first card is less than the second card in this trick
 **/
bool
Trick::less(unsigned cardno_a, unsigned cardno_b) const
{
  DEBUG_ASSERTION(   (cardno_a < this->actcardno())
                  && (cardno_b < this->actcardno()),
                  "Trick::less(" << cardno_a << ", " << cardno_b << ")\n"
                  "  one cardno is less than the actcardno " << this->actcardno()
                 );
  if (cardno_a == cardno_b)
    return false;
  else if (cardno_a < cardno_b)
    return this->card(cardno_a).less(this->card(cardno_b));
  else // cardno_a > cardno_b
    return !this->card(cardno_b).less(this->card(cardno_a));
} // bool Trick::less(unsigned cardno_a, unsigned cardno_b) const

/** comparison of two tricks
 **
 ** @param    trick1   first trick to compare
 ** @param    trick2   second trick to compare
 **
 ** @return   whether the tricks are equal
 **/
bool
operator==(Trick const& trick1, Trick const& trick2)
{
  if ((trick1.startplayerno() != trick2.startplayerno())
      || (trick1.actcardno() != trick2.actcardno()))
    return false;

  for (unsigned c = 0; c < trick1.actcardno(); c++)
    if (trick1.card(c) != trick2.card(c))
      return false;

  return (trick1.winnerplayerno() == trick2.winnerplayerno());
} // bool operator==(Trick trick1, Trick trick2)

/** comparison of two tricks
 **
 ** @param    trick1   first trick to compare
 ** @param    trick2   second trick to compare
 **
 ** @return   whether the tricks are different
 **/
bool
operator!=(Trick const& trick1, Trick const& trick2)
{
  return !(trick1 == trick2);
} // bool operator!=(Trick const& trick1, Trick const& trick2)

/** counts all points witch are turned on in rules and determines
 ** team which has won this point.
 ** takes all information from the game the trick belongs to.
 **
 ** @return   a vector with names of specialpoints and type of them
 **   for solo games return value is an empty vector
 **/
Specialpoints
Trick::specialpoints() const
{
  vector<Team> teams;

  for (auto const& p : this->game().players())
    teams.push_back(this->game().teaminfo().get(p));

  return this->specialpoints(teams);
} // Specialpoints Trick::specialpoints() const

/** counts all points witch are turned on in rules and determines
 ** team which has won this point
 **
 ** @param    teams   the teams of the players
 **
 ** @return   a vector with names of specialpoints and type of them
 **   for solo games return value is an empty vector
 **/
Specialpoints
Trick::specialpoints(vector<Team> const& teams) const
{
  Specialpoints sp;

  Team const local_winnerteam
    = ( (  (teams[this->winnerplayerno()] != Team::unknown)
         ? teams[this->winnerplayerno()]
         : Team::noteam) );

  auto const& rule = this->game().rule();

  // if this game is a solo there are no specialpoints
  if (   is_solo(this->game().type())
      && !(   (this->game().type() == GameType::marriage_solo)
           && (this->no() < this->game().marriage().determination_trickno()) ) )
    return sp;

  bool dolle_caught = false;

  // doppelkopf
  if (this->points() >= rule(Rule::Type::points_for_doppelkopf)) {
    Specialpoint s(Specialpoint::Type::doppelkopf, local_winnerteam);
    s.player_get_no = this->winnerplayerno();
    sp.push_back(s);
  } // if (this->points() >= rule(Rule::Type::points_for_doppelkopf))


  bool fl_caught_charlie=false;
  for (unsigned i = 0; i < this->actcardno(); i++) {
    // the card belongs to the winner team
    if (   ::is_real(local_winnerteam)
        && (local_winnerteam == teams[this->playerno_of_card(i)]) )
      continue;

    // fox caught
    if (   this->card(i).isfox()
        && rule(Rule::Type::extrapoint_catch_fox)
        && (this->winnerplayerno() != this->playerno_of_card(i))) {
      Specialpoint s(Specialpoint::Type::caught_fox, local_winnerteam);
      s.player_get_no = this->winnerplayerno();
      s.player_of_no = this->playerno_of_card(i);

      sp.push_back(s);

      if (this->islast()
          && rule(Rule::Type::extrapoint_catch_fox_last_trick)) {
        Specialpoint s(Specialpoint::Type::caught_fox_last_trick, local_winnerteam);
        s.player_get_no = this->winnerplayerno();
        s.player_of_no = this->playerno_of_card(i);

        sp.push_back(s);
      }
    }

    // Charlie caught
    if (this->islast()
        && (card(i) == Card::charlie)
        && !fl_caught_charlie
        && rule(Rule::Type::extrapoint_catch_charlie)
        && (!rule(Rule::Type::extrapoint_catch_charlie_only_with_diamond_queen)
            || (this->winnercard() == Card::diamond_queen) )
        && (local_winnerteam != teams[this->playerno_of_card(i)])) {

      fl_caught_charlie=true;

      Specialpoint s(Specialpoint::Type::caught_charlie, local_winnerteam);
      s.player_get_no = this->winnerplayerno();
      s.player_of_no = this->playerno_of_card(i);
      sp.push_back(s);
      // check for double charlies caught
      if (rule(Rule::Type::extrapoint_catch_double_charlie))
        for (unsigned n = i + 1; n < actcardno(); ++n) {
          if (   (card(n) == Card::charlie)
              && (teams[this->playerno_of_card(n)] != s.team) ) {
            Specialpoint s(Specialpoint::Type::caught_charlie,
                           local_winnerteam);
            s.player_get_no = this->winnerplayerno();
            s.player_of_no = this->playerno_of_card(n);
            sp.push_back(s);
          }
        }
    } // if (charlie caught)

    // dolle caught
    dolle_caught |= this->card(i).isdolle();

  } // for (i < this->actcardno())

  if (rule(Rule::Type::extrapoint_dolle_jabs_dolle)
      && dolle_caught
      && (winnercard() == Card::dolle)) {
    for (unsigned c = 0; c < this->actcardno(); c++) {
      if (   (c != this->winnercardno())
          && (this->card(c) == Card::dolle)
          && (local_winnerteam != teams[this->playerno_of_card(c)]) ) {
        Specialpoint s(Specialpoint::Type::dolle_caught_dolle,
                       local_winnerteam);
        s.player_get_no = this->winnerplayerno();
        s.player_of_no = this->playerno_of_card(c);

        sp.push_back(s);
      } // if (dolle caught of opposite team)
    } // for (c)
  } // if (dolle caught)

  if (rule(Rule::Type::extrapoint_heart_trick)
      && (is_normal(this->game().type()))
      && (this->winnercard().tcolor() == Card::heart)) {
    unsigned c;
    for (c = 0; c < this->actcardno(); ++c)
      if (this->card(c).tcolor() != Card::heart)
        break;
    if (c == this->actcardno()) {
      Specialpoint s(Specialpoint::Type::heart_trick,
                     local_winnerteam);
      s.player_get_no = this->winnerplayerno();

      sp.push_back(s);
    }
  } // if (heart trick)

  // check for foxes who are winning the last trick
  if (   this->islast()
      && (winnercard().isfox())
      && rule(Rule::Type::extrapoint_fox_last_trick)) {

    Specialpoint s(Specialpoint::Type::fox_last_trick,
                   local_winnerteam);
    s.player_get_no = this->winnerplayerno();
    //s.player_of_no = this->winnerplayerno();
    sp.push_back(s);

    // check for double fox winning
    if (rule(Rule::Type::extrapoint_double_fox_last_trick))
      for (unsigned n=winnercardno()+1; n<actcardno(); n++)
      {
        if(card(n).isfox() &&
           teams[this->playerno_of_card(n)]==s.team) {
          Specialpoint s(Specialpoint::Type::fox_last_trick,
                         local_winnerteam);
          s.player_get_no = this->winnerplayerno();
          //s.player_of_no = this->playerno_of_card(n);
          sp.push_back(s);
        }
      }
  }

  // check for charlies in lasttrick
  if (this->islast()) {
    if ((winnercard()==Card::charlie)
        && rule(Rule::Type::extrapoint_charlie))
    {
      Specialpoint s(Specialpoint::Type::charlie,
                     local_winnerteam);

      s.player_get_no = this->winnerplayerno();
      //s.player_of_no = this->winnerplayerno();
      sp.push_back(s);

      // check for double charlies
      if (rule(Rule::Type::extrapoint_double_charlie))
        for (unsigned n=winnercardno()+1; n<actcardno(); n++)
        {
          if(card(n)==Card::charlie &&
             teams[this->playerno_of_card(n)]==s.team)
          {
            Specialpoint s(Specialpoint::Type::charlie,
                           local_winnerteam);
            s.player_get_no = this->winnerplayerno();
            //s.player_of_no = this->playerno_of_card(n);

            sp.push_back(s);
          }
        }
    } // if (charlie)
  }

  return sp;
} // Specialpoints Trick::specialpoints(vector<Team> teams) const

/** @return   a vector with the cards that made a specialpoint
 **/
HandCards
Trick::specialpoints_cards() const
{
  auto const spv = this->specialpoints();

  HandCards cards;

  for (auto sp : this->specialpoints()) {
    switch (sp.type) {
    case Specialpoint::Type::nospecialpoint:
      break;
    case Specialpoint::Type::caught_fox:
    case Specialpoint::Type::caught_fox_last_trick:
      for (auto const& c : this->cards_) {
        if (c.istrumpace()) {
          cards.push_back(c);
          break;
        }
      }
      break;
    case Specialpoint::Type::fox_last_trick:
      cards.push_back(this->winnercard());
      break;
    case Specialpoint::Type::charlie:
      cards.push_back(this->winnercard());
      break;
    case Specialpoint::Type::caught_charlie:
      for (auto const& c : this->cards_) {
        if (c == Card::charlie) {
          cards.push_back(c);
          break;
        }
      }
      break;
    case Specialpoint::Type::dolle_caught_dolle:
      for (auto c = this->cards_.rbegin();
           c != this->cards_.rend();
           ++c)
        if (*c == Card::dolle) {
          cards.push_back(*c);
          break;
        }
      break;
    case Specialpoint::Type::heart_trick:
      cards.push_back(this->winnercard());
      break;
    case Specialpoint::Type::doppelkopf:
      // check for confusing winnercard
      if (   !this->winnercard().istrumpace()
          && (this->winnercard() != Card::dolle)) {
        cards.push_back(this->winnercard());
        break;
      }
      { // search better card
        HandCards::const_iterator c;
        for (c = this->cards_.begin();
             c != this->cards_.end();
             ++c) {
          if (   !c->istrumpace()
              && (*c != Card::dolle)) {
            cards.push_back(*c);
            break;
          }
        } // for (c \in this->cards_)
        if (c != this->cards_.end())
          break;
      } // search better card

      // found no better card -- taking winnercard
      cards.push_back(this->winnercard());
      break;
    case Specialpoint::Type::won:
    case Specialpoint::Type::no90:
    case Specialpoint::Type::no60:
    case Specialpoint::Type::no30:
    case Specialpoint::Type::no0:
    case Specialpoint::Type::no120_said:
    case Specialpoint::Type::no90_said:
    case Specialpoint::Type::no60_said:
    case Specialpoint::Type::no30_said:
    case Specialpoint::Type::no0_said:
    case Specialpoint::Type::no90_said_120_got:
    case Specialpoint::Type::no60_said_90_got:
    case Specialpoint::Type::no30_said_60_got:
    case Specialpoint::Type::no0_said_30_got:
    case Specialpoint::Type::no120_reply:
    case Specialpoint::Type::no90_reply:
    case Specialpoint::Type::no60_reply:
    case Specialpoint::Type::no30_reply:
    case Specialpoint::Type::no0_reply:
    case Specialpoint::Type::contra_won:
    case Specialpoint::Type::solo:
    case Specialpoint::Type::bock:
      break;
    } // switch (sp.type)
  } // for (sp : spv)

  return cards;
} // HandCards Trick::specialpoints_cards() const

/** @return   c   cardnumber of actual trick
 **
 ** @return   whether this is a specialpoints card
 **/
bool
Trick::is_specialpoints_card(unsigned const c) const
{
  DEBUG_ASSERTION((c < this->actcardno()),
                  "Trick::is_specialpoints_card(c):\n"
                  "  'c' is to great (" << c << ">=" << this->actcardno()
                  << ")");

  auto const playerno = this->playerno_of_card(c);
  for (auto const& sp_card : this->specialpoints_cards()) {
    if (sp_card.player().no() == playerno)
      return true;
  } // for (sp_card : sp_cards)

  return false;
} // bool Trick::is_specialpoints_card(unsigned c) const
