/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"

#include "weighted_card_list.h"

/** standard constructor
 **
 ** @param    game   corresponding game
 **/
WeightedCardList::WeightedCardList(Game const& game) :
  List(),
  game_(&game),
  tcolor_ptr(Card::number_of_tcolors, WeightedCardList::end()),
  tcolor_pos(Card::number_of_tcolors, -1)
{ }

/** copy constructor
 **
 ** @param    card_list   card list to copy
 **/
WeightedCardList::WeightedCardList(WeightedCardList const& card_list) :
  List(card_list),
  game_(card_list.game_),
  tcolor_ptr(Card::number_of_tcolors, WeightedCardList::end()),
  tcolor_pos(Card::number_of_tcolors, -1)
{ this->recreate_tcolor_ptrs(); }

/** copy operator
 **
 ** @param    card_list   card list to copy
 **
 ** @return   the card list
 **/
WeightedCardList&
WeightedCardList::operator=(WeightedCardList const& card_list)
{
  static_cast<List&>(*this) = card_list;
  this->recreate_tcolor_ptrs();

  return *this;
} // WeightedCardList& WeightedCardList::operator=(WeightedCardList card_list)

/** @return   the corresponding game
 **/
Game const&
WeightedCardList::game() const
{ return *this->game_; }


/** clear the list
 **/
void
WeightedCardList::clear()
{
  this->List::clear();
  this->tcolor_ptr
    = vector<WeightedCardList::iterator>(Card::number_of_tcolors,
                                         this->end());
  this->tcolor_pos = vector<int>(Card::number_of_tcolors, -1);
  return ;
} // void WeightedCardList::clear()

/** write the list in the stream
 **
 ** @param    ostr   output stream
 **/
void
WeightedCardList::write(ostream& ostr) const
{
  for (auto wc = this->begin(); wc != this->end(); ++wc) {
    ostr << std::setw(5) << wc->first << ' ';
    char c = ' ';
    for (unsigned tcolor = 0; tcolor < Card::number_of_tcolors; ++tcolor) {
      if (wc == this->tcolor_ptr[tcolor]) {
        c = to_string(static_cast<Card::TColor>(tcolor))[0];
        break;
      }
    }
    ostr << c << ' ' << wc->second << '\n';
  }
  ostr << '\n';
  for (unsigned tcolor = 0; tcolor < Card::number_of_tcolors; ++tcolor) {
    ostr << setw(7) << static_cast<Card::TColor>(tcolor) << "  "
      << setw(2) << this->tcolor_pos[tcolor] << ' ';

    auto const& highest = this->highest(static_cast<Card::TColor>(tcolor));
    if (highest.is_empty())
      ostr << "--\n";
    else
      ostr << highest << '\n';
  }
} // void WeightedCardList::write(ostream& ostr) const

/** @return   the number of cards
 **/
unsigned
WeightedCardList::cards_no() const
{
  return this->size();
} // unsigned WeightedCardList::cards_no() const

/** add the card
 **
 ** @param    card        card to add
 ** @param    weighting   the weighting of the card
 **/
void
WeightedCardList::add(Card const card, int const weighting)
{
  // pointer to the position to insert the card
  auto wc = this->begin();
  // position of the card
  int pos = 0;
  // search the position to insert the card
  for (; wc != this->end(); ++wc, ++pos)
    if (weighting >= wc->first)
      break;
  wc = this->insert(wc, pair<int, Card>(weighting, card));

  { // update the pos of the highest tcolors
    for (unsigned tcolor = 0; tcolor < Card::number_of_tcolors; ++tcolor)
      if (this->tcolor_pos[tcolor] >= pos)
        this->tcolor_pos[tcolor]  += 1;
  } // update the pos of the highest tcolors

  { // update the highest pointer of the tcolor of the card
    auto const tcolor = card.tcolor(this->game());
    if (   (this->tcolor_pos[tcolor] == -1)
        || (this->tcolor_pos[tcolor] >= pos) ) {
      this->tcolor_pos[tcolor] = pos;
      this->tcolor_ptr[tcolor] = wc;
    } // if (this->tcolor_pos[tcolor] >= pos)
  } // update the highest pointer of the tcolor of the card
} // void WeightedCardList::add(Card card, int weighting)

/** @return   the card with the highest weighting
 **/
Card
WeightedCardList::highest() const
{
  if (this->empty())
    return Card::empty;

  return this->front().second;
} // Card WeightedCardList::highest() const

/** -> result
 **
 ** @param    tcolor   tcolor to check
 **
 ** @return   the card with the highest weighting of 'tcolor'
 **/
Card
WeightedCardList::highest(Card::TColor const tcolor) const
{
  if (this->tcolor_ptr[tcolor] == this->end())
    return Card::empty;

  return this->tcolor_ptr[tcolor]->second;
} // Card WeightedCardList::highest(Card::TColor tcolor) const

/** -> result
 **
 ** @param    tcolor   tcolor not to check
 **
 ** @return   the card with the highest weighting not of 'tcolor'
 **/
Card
WeightedCardList::highest_not_of(Card::TColor const tcolor) const
{
  auto const& ptr = this->highest_ptr_not_of(tcolor);
  if (ptr == this->end())
    return Card::empty;

  return ptr->second;
} // Card WeightedCardList::highest_not_of(Card::TColor tcolor) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 **
 ** @return   the card with the highest weighting of 'tcolor_a' or 'tcolor_b'
 **/
Card
WeightedCardList::highest(Card::TColor const tcolor_a,
                          Card::TColor const tcolor_b) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b);
  if (ptr == this->end())
    return Card::empty;

  return ptr->second;
} // Card WeightedCardList::highest(Card::TColor tcolor_a, Card::TColor tcolor_b) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 **
 ** @return   the card with the highest weighting of 'tcolor_'.
 **/
Card
WeightedCardList::highest(Card::TColor const tcolor_a,
                          Card::TColor const tcolor_b,
                          Card::TColor const tcolor_c) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b, tcolor_c);
  if (ptr == this->end())
    return Card::empty;

  return ptr->second;
} // Card WeightedCardList::highest(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 ** @param    tcolor_d   fourth tcolor to check
 **
 ** @return   the card with the highest weighting of 'tcolor_'.
 **/
Card
WeightedCardList::highest(Card::TColor const tcolor_a,
                          Card::TColor const tcolor_b,
                          Card::TColor const tcolor_c,
                          Card::TColor const tcolor_d) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b, tcolor_c, tcolor_d);
  if (ptr == this->end())
    return Card::empty;

  return ptr->second;
} // Card WeightedCardList::highest(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c, Card::TColor tcolor_d) const

/** @return   the highest weighting
 **            INT_MIN, if none exists
 **/
int
WeightedCardList::highest_weighting() const
{
  if (this->empty())
    return INT_MIN;

  return this->front().first;
} // int WeightedCardList::highest_weighting() const

/** -> result
 **
 ** @param    tcolor   tcolor
 **
 ** @return   the highest weighting of tcolor
 **            INT_MIN, if none exists
 **/
int
WeightedCardList::highest_weighting(Card::TColor const tcolor) const
{
  if (this->tcolor_ptr[tcolor] == this->end())
    return INT_MIN;

  return this->tcolor_ptr[tcolor]->first;
} // int WeightedCardList::highest_weighting(Card::TColor tcolor) const

/** -> result
 **
 ** @param    tcolor   tcolor
 **
 ** @return   the highest weighting not of tcolor
 **            INT_MIN, if none exists
 **/
int
WeightedCardList::highest_weighting_not_of(Card::TColor const tcolor) const
{
  auto const& ptr = this->highest_ptr_not_of(tcolor);
  if (ptr == this->end())
    return INT_MIN;

  return ptr->first;
} // int WeightedCardList::highest_weighting_not_of(Card::TColor tcolor) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 **
 ** @return   the highest weighting of 'tcolor_a' or 'tcolor_b'
 **/
int
WeightedCardList::highest_weighting(Card::TColor const tcolor_a,
                                    Card::TColor const tcolor_b) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b);
  if (ptr == this->end())
    return INT_MIN;

  return ptr->first;
} // int WeightedCardList::highest_weighting(Card::TColor tcolor_a, Card::TColor tcolor_b) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 **
 ** @return   the highest weighting of 'tcolor_'.
 **/
int
WeightedCardList::highest_weighting(Card::TColor const tcolor_a,
                                    Card::TColor const tcolor_b,
                                    Card::TColor const tcolor_c) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b, tcolor_c);
  if (ptr == this->end())
    return INT_MIN;

  return ptr->first;
} // int WeightedCardList::highest_weighting(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 ** @param    tcolor_d   fourth tcolor to check
 **
 ** @return   the card with the highest weighting of 'tcolor_'.
 **/
int
WeightedCardList::highest_weighting(Card::TColor const tcolor_a,
                                    Card::TColor const tcolor_b,
                                    Card::TColor const tcolor_c,
                                    Card::TColor const tcolor_d) const
{
  auto const& ptr = this->highest_ptr(tcolor_a, tcolor_b, tcolor_c, tcolor_d);
  if (ptr == this->end()) {
    return INT_MIN;
  }

  return ptr->first;
} // int WeightedCardList::highest_weighting(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c, Card::TColor tcolor_d) const

/** @return   the pointer to the highest weighting
 **/
WeightedCardList::const_iterator
WeightedCardList::highest_ptr() const
{
  return this->begin();
} // WeightedCardList::const_iterator WeightedCardList::highest_ptr() const

/** -> result
 **
 ** @param    tcolor   tcolor to check
 **
 ** @return   the pointer to the highest weighting of 'tcolor'
 **/
WeightedCardList::iterator const&
WeightedCardList::highest_ptr(Card::TColor const tcolor) const
{
  return this->tcolor_ptr[tcolor];
} // WeightedCardList::iterator WeightedCardList::highest_ptr(Card::TColor const tcolor) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 **
 ** @return   the ptr with the highest weighting of 'tcolor_a' or 'tcolor_b'
 **/
WeightedCardList::iterator const&
WeightedCardList::highest_ptr(Card::TColor const tcolor_a,
                              Card::TColor const tcolor_b) const
{
  if (this->tcolor_ptr[tcolor_a] == this->end())
    return this->tcolor_ptr[tcolor_b];

  if (this->tcolor_ptr[tcolor_b] == this->end())
    return this->tcolor_ptr[tcolor_a];

  if (this->tcolor_pos[tcolor_a] < this->tcolor_pos[tcolor_b])
    return this->tcolor_ptr[tcolor_a];
  else
    return this->tcolor_ptr[tcolor_b];
} // WeightedCardList::iterator WeightedCardList::highest_ptr(Card::TColor tcolor_a, Card::TColor tcolor_b) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 **
 ** @return   the ptr with the highest weighting of 'tcolor_'.
 **/
WeightedCardList::iterator const&
WeightedCardList::highest_ptr(Card::TColor const tcolor_a,
                              Card::TColor const tcolor_b,
                              Card::TColor const tcolor_c) const
{
  if (this->tcolor_ptr[tcolor_a] == this->end())
    return this->highest_ptr(tcolor_b, tcolor_c);

  if (this->tcolor_ptr[tcolor_b] == this->end())
    return this->highest_ptr(tcolor_a, tcolor_c);

  if (this->tcolor_pos[tcolor_a] < this->tcolor_pos[tcolor_b])
    return this->highest_ptr(tcolor_a, tcolor_c);
  else
    return this->highest_ptr(tcolor_b, tcolor_c);
} // WeightedCardList::iterator WeightedCardList::highest_ptr(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c) const

/** -> result
 **
 ** @param    tcolor_a   first tcolor to check
 ** @param    tcolor_b   second tcolor to check
 ** @param    tcolor_c   third tcolor to check
 ** @param    tcolor_d   fourth tcolor to check
 **
 ** @return   the ptr with the highest weighting of 'tcolor_'.
 **/
WeightedCardList::iterator const&
WeightedCardList::highest_ptr(Card::TColor const tcolor_a,
                              Card::TColor const tcolor_b,
                              Card::TColor const tcolor_c,
                              Card::TColor const tcolor_d) const
{
  if (this->tcolor_ptr[tcolor_a] == this->end())
    return this->highest_ptr(tcolor_b, tcolor_c, tcolor_d);

  if (this->tcolor_ptr[tcolor_b] == this->end())
    return this->highest_ptr(tcolor_a, tcolor_c, tcolor_d);

  if (this->tcolor_pos[tcolor_a] < this->tcolor_pos[tcolor_b])
    return this->highest_ptr(tcolor_a, tcolor_c, tcolor_d);
  else
    return this->highest_ptr(tcolor_b, tcolor_c, tcolor_d);
} // WeightedCardList::iterator WeightedCardList::highest_ptr(Card::TColor tcolor_a, Card::TColor tcolor_b, Card::TColor tcolor_c, Card::TColor tcolor_d) const

/** -> result
 **
 ** @param    tcolor   tcolor not to check
 **
 ** @return   the pointer to the highest weighting not of 'tcolor'
 **/
WeightedCardList::iterator const&
WeightedCardList::highest_ptr_not_of(Card::TColor const tcolor) const
{
  Card::TColor const tc = this->highest().tcolor(this->game());
  if (tc != tcolor)
    return this->highest_ptr(tc);

  return this->highest_ptr(( (tcolor == Card::diamond)
                            ? Card::heart
                            : Card::diamond),
                           (  (tcolor <= Card::heart)
                            ? Card::spade
                            : Card::heart),
                           (  (tcolor <= Card::spade)
                            ? Card::club
                            : Card::spade),
                           (  (tcolor <= Card::club)
                            ? Card::trump
                            : Card::club)
                          );
} // WeightedCardList::iterator WeightedCardList::highest_ptr_not_of(Card::TColor const tcolor) const

/** -> result
 **
 ** @param    card   card to return the (heighest) weighting of
 **
 ** @return   the (highest) weighting weighting of 'card'
 **            INT_MIN, if none exists
 **/
int
WeightedCardList::weighting(Card const card) const
{
  // search the card in the pointers
  for (auto ptr
       = this->tcolor_ptr[card.tcolor(this->game())];
       ptr != this->end();
       ++ptr)
    if (ptr->second == card)
      return ptr->first;

  return INT_MIN;
} // int WeightedCardList::weighting(Card card) const

/** remove the highest card
 **/
void
WeightedCardList::pop_highest()
{
  if (this->empty())
    return ;

  // the tcolor of the card to remove
  auto const tcolor = this->front().second.tcolor(this->game());

  // remove the card
  this->pop_front();

  // adjust the positions
  for (unsigned tc = 0; tc < Card::number_of_tcolors; ++tc)
    if (this->tcolor_pos[tc] >= 0)
      this->tcolor_pos[tc] -= 1;

  // search the first of 'tcolor' in the list
  this->recreate_tcolor_ptr(tcolor);
} // void WeightedCardList::pop_highest()

/** remove the highest card of the tcolor
 **
 ** @param    tcolor   tcolor to remove the highest card from
 **/
void
WeightedCardList::pop_highest(Card::TColor const tcolor)
{
  // search the card in the pointers
  auto& ptr = this->tcolor_ptr[tcolor];
  if (ptr == this->end())
    return ;

  int& pos = this->tcolor_pos[tcolor];
  // update the pos of the further tcolors
  for (unsigned tc = 0; tc < Card::number_of_tcolors; ++tc)
    if (this->tcolor_pos[tc] > pos)
      this->tcolor_pos[tc]  += 1;
  // remove the pointer
  ptr = this->erase(ptr);
  for (; ptr != this->end(); ++ptr, ++pos)
    if (ptr->second.tcolor(this->game()) == tcolor)
      break;
  if (ptr == this->end())
    pos = -1;
} // void WeightedCardList::pop_highest(Card::TColor tcolor)

/** remove the highest card not of the tcolor
 **
 ** @param    tcolor   tcolor to remove the highest card from
 **
 ** @todo      own code (for better performance)
 **/
void
WeightedCardList::pop_highest_not_of(Card::TColor const tcolor)
{
  this->pop(this->highest_not_of(tcolor));
} // void WeightedCardList::pop_highest_not_of(Card::TColor tcolor)

/** remove the given card
 **
 ** @param    card   card to remove
 **/
void
WeightedCardList::pop(Card const card)
{
  // search the card in the pointers
  auto const tcolor = card.tcolor(this->game());
  auto& ptr = this->tcolor_ptr[tcolor];
  if (  (ptr != this->end())
      && (ptr->second == card) ) {
    int& pos = this->tcolor_pos[tcolor];
    // update the pos of the further tcolors
    for (unsigned tc = 0; tc < Card::number_of_tcolors; ++tc)
      if (this->tcolor_pos[tc] > pos)
        this->tcolor_pos[tc]  += 1;
    // remove the pointer
    ptr = this->erase(ptr);
    for (; ptr != this->end(); ++ptr, ++pos)
      if (ptr->second.tcolor(this->game()) == tcolor)
        break;
    if (ptr == this->end())
      pos = -1;
  } else { // if !(this->tcolor_ptr[tcolor].second == card)
    int pos = this->tcolor_pos[tcolor];
    for (; ptr != this->end(); ++ptr, ++pos)
      if (ptr->second == card)
        break;
    DEBUG_ASSERTION(ptr != this->end(),
                    "WeightedCardList::pop(card)\n"
                    "  card '" << card << "' not found in the list:\n"
                    << *this);
    ptr = this->erase(ptr);
    // update the pos of the further tcolors
    for (unsigned tc = 0; tc < Card::number_of_tcolors; ++tc)
      if (this->tcolor_pos[tc] > pos)
        this->tcolor_pos[tc] -= 1;
  } // if !(this->tcolor_ptr[tcolor].second == card)
} // void WeightedCardList::pop(Card card)

/** recreate the pointers to the highest card of each tcolor
 **/
void
WeightedCardList::recreate_tcolor_ptrs()
{
  // search the first of 'tcolor' in the list
  for (unsigned tcolor = 0; tcolor < Card::number_of_tcolors; ++tcolor)
    this->recreate_tcolor_ptr(static_cast<Card::TColor>(tcolor));
} // void WeightedCardList::recreate_tcolor_ptrs()

/** recreate the pointer to the highest card of the tcolor
 **
 ** @param    tcolor   tcolor to recreate the pointer of
 **/
void
WeightedCardList::recreate_tcolor_ptr(Card::TColor const tcolor)
{
  // search the first of 'tcolor' in the list
  int& pos = this->tcolor_pos[tcolor];
  auto& ptr = this->tcolor_ptr[tcolor];
  for (pos = 0, ptr = this->begin(); ptr != this->end(); ++ptr, ++pos)
    if (ptr->second.tcolor(this->game()) == tcolor)
      break;
  if (ptr == this->end())
    pos = -1;
} // void WeightedCardList::recreate_tcolor_ptr(Card::TColor tcolor)

/** write 'card_list' in 'ostr'
 **
 ** @param    ostr        output stream
 ** @param    card_list   card list to write
 **/
ostream&
operator<<(ostream& ostr, WeightedCardList const& card_list)
{
  card_list.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, WeightedCardList card_list)
