/**********************************************************************
 *
 *   Some routines for effective Debugging
 *
 *   Copyright (C) 2001 – 2018 by Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *
 *********************************************************************/

#pragma once

#include <iostream>
#include <fstream>
using std::cout;
using std::cerr;
using std::endl;
using std::ostream;
using std::ofstream;
#include <sstream>
using std::ostringstream;
#include <string>
using std::string;
#include <set>
using std::set;
#include <memory>
using std::unique_ptr;
using std::make_unique;

#include <cstdlib>
using std::set_terminate;

#include "utils/time.h"

#define DEBUG_ASSERTION(predicate, text)   \
  if (!(predicate)) {            \
    ostringstream ostr;            \
    ostr << "ASSERT(file " << __FILE__      \
         << ", line " << __LINE__      \
         << "): " << text << endl;      \
    cerr << ostr.str();      \
    SEGFAULT;               \
  } else

#define DEBUG_ASSERT(predicate) \
  DEBUG_ASSERTION(predicate, "failed")

typedef void (*DebugFunction)(string const&);
extern DebugFunction debug_function;

#define DEBUG_THROW(Type, constructor)       \
do {                                                                 \
  Type exception = constructor;  \
    ExceptionTerminateHandler::set( __FILE__, __LINE__, exception.what()); \
    DEBUG("throw") << "exception: " __FILE__ << '#' << __LINE__ << '\n'        \
      << "  " << exception.what() << endl;               \
  throw (exception);                                     \
} while(false)

#define DEBUG_CAUGHT()           \
  ExceptionTerminateHandler::unset()

// Ausgabe der History einer ungefangenen Exception auf fs::error
struct UncaughtException
{
  UncaughtException(string const file,
                    int const line,
                    string const what) :
    file_(file), line_(line), what_(what)
  { }
  string const file_;
  int const line_;
  string const what_;
}; // struct UncaughtException

namespace ExceptionTerminateHandler {

extern unique_ptr<UncaughtException> uncaughtException_;

// neue terminate Funktion, die what, CPOS und Bugreport ausgibt
// fuer eine nicht gefangene Exception
void terminate();
void set(string const& file, int line, string const& what);
void unset();

}; // namespace ExceptionTerminateHandler


// class for writing debug informations
// Usage:
//   debug.add("filter name");
//   DEBUG("filter name") << "debug text\n";
// special filters:
//   "codeline": prepend the file and code line in front of the debug text (default: on)
//   "time": prepend the time (of the day) in front of the debug text (default: off)
//   "clock": prepend the processor time in front of the debug text (default: off)
class Debug {
  public:
    Debug();
    ~Debug();

    ostream& ostr();
    void set_ostr(ostream& ostr);
    void set_file(string file);
    void add(string filter);
    void remove(string filter);
    bool filter(string filter);

    bool operator()(string filter) const;
    ostream& ostr(string filter);
    ostream& ostr(string filter, string file, long line, string function_name);

  private:
    set<string> filter_;
    ostream* ostr_;
}; // class Debug

extern Debug debug;
#define DEBUG(f) if (!::debug(f)) ; else ::debug.ostr(f, __FILE__, __LINE__, __func__)
