/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "cards.h"
#include "game.h"
#include "../party/rule.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GameCards::GameCards(Game& game) :
  game_(&game),
  cards_(Card::number_of_tcolors)
{ }

/** constructor
 **
 ** @param    cards   cards to copy
 ** @param    game    corresponding game
 **/
GameCards::GameCards(GameCards const& cards, Game& game) :
  game_(&game),
  cards_(cards.cards_)
{ }

/** update the cards
 **/
void
GameCards::update()
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();

  this->cards_ = vector<vector<Card>>(Card::number_of_tcolors);

  // first all color cards

  for (auto c : rule.card_colors())
    for (auto v : rule.card_values())
      if (!Card(c, v).istrump(game))
        this->cards_[c].insert(this->cards_[c].begin(), Card(c, v));


  // second all trumps (manual fill)

  vector<Card>& trumps = this->cards_[Card::trump];
  auto const trumpcolor = this->trumpcolor();
  switch (game.type()) {
  case GameType::normal:
  case GameType::fox_highest_trump:
  case GameType::poverty:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_diamond:
  case GameType::redistribute:
    if (rule(Rule::Type::with_nines))
      trumps.push_back(Card(trumpcolor, Card::nine));
    trumps.push_back(Card(trumpcolor, Card::king));
    if (!(   (trumpcolor == Card::heart)
          && rule(Rule::Type::dollen)))
      trumps.push_back(Card(trumpcolor,  Card::ten));
    trumps.push_back(Card(trumpcolor,    Card::ace));
    trumps.push_back(Card::diamond_jack);
    trumps.push_back(Card::heart_jack);
    trumps.push_back(Card::spade_jack);
    trumps.push_back(Card::club_jack);
    trumps.push_back(Card::diamond_queen);
    trumps.push_back(Card::heart_queen);
    trumps.push_back(Card::spade_queen);
    trumps.push_back(Card::club_queen);
    if (rule(Rule::Type::dollen))
      trumps.push_back(Card::dolle);
    break;
  case GameType::thrown_nines:
    trumps = {
    Card::diamond_nine,
    Card::heart_nine,
    Card::spade_nine,
    Card::club_nine,
    };
    break;
  case GameType::thrown_kings:
    trumps = {
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  case GameType::thrown_nines_and_kings:
    trumps = {
      Card::diamond_nine,
      Card::heart_nine,
      Card::spade_nine,
      Card::club_nine,
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  case GameType::thrown_richness:
    trumps = {
      Card::club_ace,
      Card::spade_ace,
      Card::heart_ace,
      Card::diamond_ace,
      Card::club_ten,
      Card::spade_ten,
      Card::heart_ten,
      Card::diamond_ten,
      Card::club_king,
      Card::spade_king,
      Card::heart_king,
      Card::diamond_king,
      Card::club_queen,
      Card::spade_queen,
      Card::heart_queen,
      Card::diamond_queen,
      Card::club_jack,
      Card::spade_jack,
      Card::heart_jack,
      Card::diamond_jack,
    };
    if (rule(Rule::Type::with_nines)) {
      trumps.push_back(Card::club_nine);
      trumps.push_back(Card::spade_nine);
      trumps.push_back(Card::heart_nine);
      trumps.push_back(Card::diamond_nine);
    }
    break;
  case GameType::solo_meatless:
    break;
  case GameType::solo_jack:
    trumps = {
      Card::diamond_jack,
      Card::heart_jack,
      Card::spade_jack,
      Card::club_jack,
    };
    break;
  case GameType::solo_queen:
    trumps = {
      Card::diamond_queen,
      Card::heart_queen,
      Card::spade_queen,
      Card::club_queen,
    };
    break;
  case GameType::solo_king:
    trumps = {
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  case GameType::solo_queen_jack:
    trumps = {
      Card::diamond_jack,
      Card::heart_jack,
      Card::spade_jack,
      Card::club_jack,
      Card::diamond_queen,
      Card::heart_queen,
      Card::spade_queen,
      Card::club_queen,
    };
    break;
  case GameType::solo_king_jack:
    trumps = {
      Card::diamond_jack,
      Card::heart_jack,
      Card::spade_jack,
      Card::club_jack,
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  case GameType::solo_king_queen:
    trumps = {
      Card::diamond_queen,
      Card::heart_queen,
      Card::spade_queen,
      Card::club_queen,
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  case GameType::solo_koehler:
    trumps = {
      Card::diamond_jack,
      Card::heart_jack,
      Card::spade_jack,
      Card::club_jack,
      Card::diamond_queen,
      Card::heart_queen,
      Card::spade_queen,
      Card::club_queen,
      Card::diamond_king,
      Card::heart_king,
      Card::spade_king,
      Card::club_king,
    };
    break;
  }; // switch (this->type())

  DEBUG_ASSERTION((this->trumps().size()
                   == (this->count_trumps()
                       / rule(Rule::Type::number_of_same_cards))),
                  "GameCards::update()\n"
                  "  trumps size " << this->trumps().size()
                  << " does not match trumps number "
                  << (this->count_trumps()
                      / rule(Rule::Type::number_of_same_cards)));
} // void GameCards::update()

/** update the sorting of the swines
 **/
void
GameCards::update_swines()
{
  for (auto t = this->cards_[Card::trump].begin();
       t != this->cards_[Card::trump].end();
       ++t) {
    if (*t == this->swine()) {
      this->cards_[Card::trump].erase(t);
      break;
    }
  } // for (t : this->trump_)
  this->cards_[Card::trump].push_back(this->swine());
}

/** update the sorting of the hyperswines
 **/
void
GameCards::update_hyperswines()
{
  for (auto t = this->cards_[Card::trump].begin();
       t != this->cards_[Card::trump].end();
       ++t) {
    if (*t == this->hyperswine()) {
      this->cards_[Card::trump].erase(t);
      break;
    }
  } // for (t \in this->trump_)
  this->cards_[Card::trump].push_back(this->hyperswine());
}

/** @return   number of cards
 **/
unsigned
GameCards::count() const
{
  auto const& rule = this->game_->rule();
  return (this->count_unique()
          * rule(Rule::Type::number_of_same_cards));
}

/** @return   number of different cards
 **/
unsigned
GameCards::count_unique() const
{
  auto const& rule = this->game_->rule();
  return rule(Rule::Type::number_of_cards);
}

/** -> result
 **
 ** @param    tcolor   tcolor
 **
 ** @return   number of cards of the tcolor
 **/
unsigned
GameCards::count(Card::TColor const tcolor) const
{
  auto const& rule = this->game_->rule();
  if (tcolor == Card::unknowncardcolor) {
    return this->count();
  } else {
    return (this->count_unique(tcolor)
            * rule(Rule::Type::number_of_same_cards));
  }
}

/** -> result
 **
 ** @param    tcolor   tcolor
 **
 ** @return   number of different cards of the tcolor
 **/
unsigned
GameCards::count_unique(Card::TColor const tcolor) const
{
  auto const& rule = this->game_->rule();
  if (tcolor == Card::unknowncardcolor)
    return this->count();
  if (tcolor == Card::trump)
    return (this->count_trumps() / rule(Rule::Type::number_of_same_cards));
  if (tcolor == this->trumpcolor())
    return 0;

  switch (this->game_->type()) {
  case GameType::normal:
  case GameType::fox_highest_trump:
  case GameType::poverty:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_diamond:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_club:
  case GameType::redistribute:
    {
      unsigned number = rule(Rule::Type::number_of_card_values) - 2;
      if (   (tcolor == Card::heart)
          && rule(Rule::Type::dollen)) {
        number -= 1;
      }
      return number;
    }
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
    return (rule(Rule::Type::number_of_card_values) - 1);

  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
    return (rule(Rule::Type::number_of_card_values) - 2);

  case GameType::solo_koehler:
    return (rule(Rule::Type::number_of_card_values) - 3);

  case GameType::solo_meatless:
    return rule(Rule::Type::number_of_card_values);

  case GameType::thrown_nines:
  case GameType::thrown_kings:
    return rule(Rule::Type::number_of_card_colors);

  case GameType::thrown_nines_and_kings:
    return (2 * rule(Rule::Type::number_of_card_colors));

  case GameType::thrown_richness:
    return 0;
  } // this->type()

  return 0;
} // unsigned GameCards::count_unique(Card::TColor tcolor) const

/** @return   number of trumps in the game
 **/
unsigned
GameCards::count_trumps() const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();

  unsigned trumps = 0;

  switch (game.type()) {
  case GameType::normal:
  case GameType::fox_highest_trump:
  case GameType::poverty:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_diamond:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_club:
  case GameType::redistribute:
    trumps = 2 * rule(Rule::Type::number_of_card_colors); // jack and queen
    trumps += (rule(Rule::Type::number_of_card_values)
               - 2); // rest trump color
    if (   rule(Rule::Type::dollen)
        && (game.type() != GameType::solo_heart))
      trumps += 1;
    break;
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
    trumps = rule(Rule::Type::number_of_card_colors);
    break;
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
    trumps = 2 * rule(Rule::Type::number_of_card_colors);
    break;
  case GameType::solo_koehler:
    trumps = 3 * rule(Rule::Type::number_of_card_colors);
    break;
  case GameType::solo_meatless:
    trumps = 0;
    break;
  case GameType::thrown_nines:
  case GameType::thrown_kings:
    trumps = rule(Rule::Type::number_of_card_colors);
    break;

  case GameType::thrown_nines_and_kings:
    trumps = 2 * rule(Rule::Type::number_of_card_colors);
    break;

  case GameType::thrown_richness:
    trumps = (rule(Rule::Type::number_of_cards)
              / rule(Rule::Type::number_of_same_cards));
    break;
  } // this->type()
  trumps *= rule(Rule::Type::number_of_same_cards);

  return trumps;
} // unsigned GameCards::count_trumps() const

/** @return   number of played trumps
 **
 ** @todo        perhaps use a variable,
 **              so that this function does not take much time
 **/
unsigned
GameCards::count_played_trumps() const
{
  auto const& game = *this->game_;
  auto const& tricks = game.tricks();
  // number of trumps
  unsigned trumps = 0;
  for (auto const& trick : tricks) {
    for (unsigned c = 0; c < trick.actcardno(); ++c) {
      if (trick.card(c).istrump()) {
        trumps += 1;
      }
    }
  }

  return trumps;
}

/** @return   number of played cards
 **/
unsigned
GameCards::count_played() const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  auto const& tricks = game.tricks();

  if (tricks.empty())
    return 0;

  if (::game_status == GameStatus::game_finished) {
    return (rule(Rule::Type::number_of_tricks_in_game)
            * rule(Rule::Type::number_of_players_in_game));
  }

  return (tricks.real_current_no()
          * rule(Rule::Type::number_of_players_in_game)
          + tricks.current().actcardno());
}

/** @return   number of cards remaining to be played
 **/
unsigned
GameCards::count_remaining() const
{
  auto const& rule = this->game_->rule();
  return (rule(Rule::Type::number_of_tricks_in_game)
          * rule(Rule::Type::number_of_players_in_game)
          - this->count_played());
}

/** -> result
 **
 ** @param    card   card
 **
 ** @return   number of played cards 'card'
 **
 ** @todo        perhaps use a variable,
 **              so that this function does not take much time
 **/
unsigned
GameCards::count_played(Card const card) const
{
  auto const& tricks = this->game_->tricks();
  // number of trumps
  unsigned no = 0;
  for (auto const& trick : tricks) {
    for (auto c : trick.cards()) {
      if (c == card) {
        no += 1;
      }
    }
  }

  return no;
}

/** @param    game_type   the game type
 **
 ** @return   the trumpcolor
 **/
Card::Color
GameCards::trumpcolor(GameType game_type)
{
  switch (game_type) {
  case GameType::normal:
  case GameType::poverty:
  case GameType::fox_highest_trump:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_diamond:
  case GameType::redistribute:
    return Card::diamond;
  case GameType::solo_heart:
    return Card::heart;
  case GameType::solo_spade:
    return Card::spade;
  case GameType::solo_club:
    return Card::club;
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
  case GameType::solo_koehler:
  case GameType::solo_meatless:
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
    return Card::nocardcolor;
  } // switch(game_type)

  DEBUG_ASSERTION(false,
                  "GameCards::trumpcolor(game_type)\n"
                  "  after return");
  return Card::nocardcolor;
} // static Card::Color GameCards::trumpcolor(GameType game_type)

/** @return   the trumpcolor
 **/
Card::Color
GameCards::trumpcolor() const
{
  return GameCards::trumpcolor(this->game_->type());
}

/** @return   the colors without trump
 **/
vector<Card::Color>
GameCards::colors() const
{
  switch (this->game_->type()) {
  case GameType::normal:
  case GameType::poverty:
  case GameType::fox_highest_trump:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_diamond:
  case GameType::redistribute:
    return {Card::club, Card::spade, Card::heart};
  case GameType::solo_heart:
    return {Card::club, Card::spade, Card::diamond};
  case GameType::solo_spade:
    return {Card::club, Card::heart, Card::diamond};
  case GameType::solo_club:
    return {Card::spade, Card::heart, Card::diamond};
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
  case GameType::solo_koehler:
  case GameType::solo_meatless:
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
    return {Card::club, Card::spade, Card::heart, Card::diamond};
  } // switch(game_type)

  DEBUG_ASSERTION(false,
                  "GameCards::colors(game_type)\n"
                  "  after return");
  return {};
}

/** -> result
 **
 ** @param    tcolor   tcolor
 **
 ** @return   the cards of the tcolor
 **/
vector<Card> const&
GameCards::cards(Card::TColor const tcolor) const
{
  DEBUG_ASSERTION(tcolor <= Card::trump,
                  "GameCards::cards(tcolor)\n"
                  "  tcolor = " << tcolor << " is not valid.");

  return this->cards_[tcolor];
} // vector<Card> GameCards::cards(Card::TColor tcolor) const

/** @return   the trumps
 **/
vector<Card> const&
GameCards::trumps() const
{
  return this->cards(Card::trump);
} // vector<Card> GameCards::trumps() const

/** -> result
 **
 ** @note   if both cards are equal, the first (this) is not less than
 **   the second (card)
 **   (but the dollen -- see rules)
 **
 ** @param    card_a   first card
 ** @param    card_b   second card
 **
 ** @return   whether 'card_a' is less than 'card_b'
 **/
bool
GameCards::less(Card const lhs, Card const rhs) const
{
  if (rhs.is_empty())
    return false;
  if (lhs.is_empty())
    return true;

  Game const& game = *this->game_;
  Rule const& rule = game.rule();

  // both cards are not trump
  if (!lhs.istrump(game) && !rhs.istrump(game)) {
    if (lhs.color() == rhs.color()) // same color
      return (lhs.value() < rhs.value());
    else // different color
      return false;
  }

  if (lhs.istrump(game) && !rhs.istrump(game))
    return false;

  if (!lhs.istrump(game) && rhs.istrump(game))
    return true;


  // The difficult case: both cards are trump

  switch(game.type()) {
  case GameType::normal:
  case GameType::poverty:
  case GameType::fox_highest_trump:
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_diamond:
  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::redistribute:

    // Hyperswines
    if (rule(Rule::Type::hyperswines)) {
      if (game.swines().hyperswines_announced()) {
        if (lhs == this->hyperswine())
          return false;
        if (rhs == this->hyperswine())
          return true;
      }
    }

    // Swines
    if (rule(Rule::Type::swines)) {
      if (game.swines().swines_announced()) {
        if (lhs == this->swine())
          return false;
        if (rhs == this->swine())
          return true;
      }
    }

    // Dollen
    if (rule(Rule::Type::dollen)) {
      if (lhs == Card::dolle) {
        if (rhs == Card::dolle) { // both cards are dollen
          if (    (rule(Rule::Type::dollen_first_over_second_with_swines)
               && game.swines().swines_announced()) )
            return false;
          if (rule(Rule::Type::dollen_contrary_in_last_trick)
              && game.tricks().is_last_trick()) {
            // we are in the last trick
            // and the rule 'dollen_contrary_in_last_trick' is valid
            if (rule(Rule::Type::dollen_second_over_first)) {
              return false;
            } else { // if !(rule(Rule::Type::dollen_second_over_first)
              return true;
            } // if !(rule(Rule::Type::dollen_second_over_first)
          } else { // if !(game.tricks().is_last_trick())
            if (rule(Rule::Type::dollen_second_over_first)) {
              return true;
            } else { // if !(rule(Rule::Type::dollen_second_over_first)
              return false;
            } // if !(rule(Rule::Type::dollen_second_over_first)
          } // if !(game.tricks().is_last_trick())
        } else { // if !(b == Card::dolle)
          return false;
        } // if !(b == Card::dolle)
      } else { // if !(a == Card::dolle)
        if (rhs == Card::dolle)
          return true;
      } // if !(a == Card::dolle)
    } // if (rule.dollen)

    if (lhs.value() == Card::queen) {
      if (rhs.value() == Card::queen) {
        return (lhs.color() < rhs.color());
      } else {
        return false;
      }
    }

    if (lhs.value() == Card::jack) {
      if (rhs.value() == Card::queen) {
        return true;
      } else if (rhs.value() == Card::jack) {
        return (lhs.color() < rhs.color());
      } else {
        return false;
      }
    }

    // 'a' is a simple ace/ten/king/nine
    if ( (rhs.value() == Card::queen) ||
        (rhs.value() == Card::jack) ) // 'b' is higher than an ace/ten/king/nine
      return true;

    // both cards are simple ace/ten/king/nine
    return (lhs.value() < rhs.value());
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
    return (lhs.color() < rhs.color());
  case GameType::solo_queen_jack:
    if (lhs.value() != rhs.value()) // one card is queen, one is jack
      return (lhs.value() == Card::jack);
    // both card are jack/queen
    return (lhs.color() < rhs.color());
  case GameType::solo_king_jack:
    if (lhs.value() != rhs.value()) // one card is king, one is jack
      return (lhs.value() == Card::jack);
    // both card are jack/king
    return (lhs.color() < rhs.color());
  case GameType::solo_king_queen:
    if (lhs.value() != rhs.value()) // one card is king, one is queen
      return (lhs.value() == Card::queen);
    // both card are queen/king
    return (lhs.color() < rhs.color());
  case GameType::solo_koehler:
    if (lhs.value() == rhs.value()) // both cards are the same value
      return (lhs.color() < rhs.color());
    // both cards have different values
    if (lhs.value() == Card::jack)
      return true;
    if (rhs.value() == Card::jack)
      return false;
    // one card is queen, one is king
    // -- hey, that is like in 'solo_king_queen', take that code
    return (lhs.value() == Card::queen);
  case GameType::solo_meatless:
    // cannot be, because no card is trump
    DEBUG_ASSERTION(false,
                    "  solo_meatless with trump ");

    break;
  case GameType::thrown_nines:
    return (lhs.color() < rhs.color());
  case GameType::thrown_kings:
    return (lhs.color() < rhs.color());
  case GameType::thrown_nines_and_kings:
    if (   (lhs.value() == Card::nine)
        && (rhs.value() == Card::king) )
      return true;
    if (   (lhs.value() == Card::king)
        && (rhs.value() == Card::nine) )
      return false;
    return (lhs.color() < rhs.color());

  case GameType::thrown_richness:
    return (   (lhs.value() < rhs.value())
            || (   (lhs.value() == rhs.value())
                && (lhs.color() < rhs.color())));
  } // switch(game.type())

  DEBUG_ASSERTION(false,
                  "Card::less(lhs, rhs):\n"
                  "  after 'switch(game.type())'");

  return false;
} // bool GameCards::less(Card lhs, Card rhs) const

/** -> result
 **
 ** @param    card   card
 **
 ** @return   the next lower card than 'card'
 **
 ** @todo      use code that is more performant
 **/
Card
GameCards::next_lower_card(Card const card) const
{
  vector<Card> lower_cards = this->lower_cards(card);
  if (lower_cards.empty())
    return Card::empty;

  return lower_cards.front();
} // Card GameCards::next_lower_card(Card card) const

/** -> result
 **
 ** @param    card   card
 **
 ** @return   the next higher card than 'card'
 **
 ** @todo      use code that is more performant
 **/
Card
GameCards::next_higher_card(Card const card) const
{
  vector<Card> higher_cards = this->higher_cards(card);
  if (higher_cards.empty())
    return Card::empty;

  return higher_cards.front();
} // Card GameCards::next_higher_card(Card card) const

/** -> result
 **
 ** @param    card   card
 ** @param    n      number of cards to get higher
 **
 ** @return   the next higher card in n steps than 'card'
 **
 ** @todo      use code that is more performant
 **/
Card
GameCards::next_higher_card(Card const card, unsigned const n) const
{
  if (n == 0)
    return card;
  else if (n == 1)
    return this->next_higher_card(card);
  else // if (n > 1)
    return this->next_higher_card(card, n-1);
}

/** -> result
 **
 ** @param    card   card
 **
 ** @return   all cards lower than 'card' (in ascending order)
 **/
vector<Card>
GameCards::lower_cards(Card const card) const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  if (card.istrump(game)) {
    auto c = this->trumps().rbegin();
    for (; c != this->trumps().rend(); ++c)
      if (*c == card)
        break;
    return vector<Card>(++c, this->trumps().rend());
  } else { // if !(card.istrump(*this))
    vector<Card> cards;
    for (auto const& v : rule.card_values()) {
      Card const c(card.color(), v);
      if (   !c.istrump(game)
          && this->less(c, card))
        cards.insert(cards.begin(), c);
    } // for (v \in card values)

    return cards;
  } // if !(card.istrump(*this))
}

/** -> result
 **
 ** @param    card   card
 **
 ** @return   all cards higher than 'card' (in ascending order)
 **/
vector<Card>
GameCards::higher_cards(Card const card) const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  if (card.istrump(game)) {
    auto c = this->trumps().begin();
    for (; c != this->trumps().end(); ++c)
      if (this->less(card, *c))
        break;
    return vector<Card>(c, this->trumps().end());

  } else { // if !(card.istrump(*this))
    vector<Card> cards;
    for (auto const& v : rule.card_values()) {
      Card const c(card.color(), v);
      if (   !c.istrump(game)
          && this->less(card, c))
        cards.push_back(c);
    } // for (v \in card values)

    cards.insert(cards.end(), this->trumps().begin(), this->trumps().end());
    return cards;
  } // if !(card.istrump(*this))
}

/** -> result
 **
 ** @param    lhs   left card
 ** @param    rhs   right card
 **
 ** @return   the cards between 'lhs' and 'rhs' (in ascending order)
 **/
vector<Card>
GameCards::cards_between(Card const lhs, Card const rhs) const
{
  // the cards between
  auto cards = this->higher_cards(lhs);

  // search the first card not less than 'rhs'
  auto c = cards.begin();
  for (c = cards.begin(); c != cards.end(); ++c)
    if (   (*c == rhs)
        || (!this->less(*c, rhs)) )
      break;

  return vector<Card>(cards.begin(), c);
}

/** -> result
 **
 ** @param    card   card
 **
 ** @return   all trumps lower than 'card' (in ascending order)
 **/
vector<Card>
GameCards::lower_trumps(Card const card) const
{
  auto const& game = *this->game_;

  if (!card.istrump(game))
    return {};

  auto c = this->trumps().begin();
  for (; c != this->trumps().end(); ++c)
    if (!this->less(*c, card))
      break;
  return vector<Card>(this->trumps().begin(), c);
}

/** -> result
 **
 ** @param    card   card to check
 **
 ** @return   whether the card is a special card
 **             (according to the rules and the hand)
 **/
bool
GameCards::is_special(HandCard const& card) const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  if (card.is_empty())
    return false;

  // dolle
  if (   rule(Rule::Type::dollen)
      && (card == Card::dolle))
    return true;

  // swines
  if (   rule(Rule::Type::swines)
      && (card == this->swine())
      && (   game.swines().swines_announced()
          || game.swines().swines_announcement_valid(card.player())) )
    return true;

  // hyperswines
  if (   rule(Rule::Type::hyperswines)
      && (card == this->hyperswine())
      && (   game.swines().hyperswines_announced()
          || game.swines().hyperswines_announcement_valid(card.player())) )
    return true;

  return false;
} // bool GameCards::is_special(HandCard card) const

/** @return   a fox card
 **/
Card
GameCards::fox() const
{
  auto const game_type = this->game_->type();
  auto const& rule = this->game_->rule();
  if (   rule(Rule::Type::dollen)
      && (   ::is_normal(game_type)
          || ::is_poverty(game_type)
          || ::is_marriage(game_type))) {
    return {this->trumpcolor(), Card::ace};
  } else {
    return {};
  }
} // Card GameCards::fox() const

/** @return   a dolle card
 **/
Card
GameCards::dolle() const
{
  auto const game_type = this->game_->type();
  auto const& rule = this->game_->rule();
  if (!rule(Rule::Type::dollen))
    return {};

  if (   ::is_normal(game_type)
      || ::is_poverty(game_type)
      || ::is_marriage(game_type)
      || ::is_color_solo(game_type)) {
    return Card::dolle;
  } else {
    return {};
  }
} // Card GameCards::dolle() const

/** @return   a swine card
 **/
Card
GameCards::swine() const
{
  auto const game_type = this->game_->type();
  auto const& rule = this->game_->rule();
  if (!rule(Rule::Type::swines))
    return {};

  if (   ::is_normal(game_type)
      || ::is_poverty(game_type)
      || ::is_marriage(game_type)
      || (   rule(Rule::Type::swines_in_solo)
          && ::is_color_solo(game_type)) ) {
    return {this->trumpcolor(), Card::ace};
  } else {
    return {};
  }
} // Card GameCards::swine()

/** @return   a hyperswine card
 **/
Card
GameCards::hyperswine() const
{
  auto const game_type = this->game_->type();
  auto const& rule = this->game_->rule();
  if (!rule(Rule::Type::hyperswines))
    return {};

  if (   ::is_normal(game_type)
      || ::is_poverty(game_type)
      || ::is_marriage(game_type)
      || (   rule(Rule::Type::hyperswines_in_solo)
          && ::is_color_solo(game_type)) ) {
    return {this->trumpcolor(), rule(Rule::Type::with_nines) ? Card::nine : Card::king};
  } else {
    return {};
  }
} // Card GameCards::hyperswine()

