/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "game.h"

#include "gameplay.h"
#include "../party/party.h"
#include "../party/rule.h"
#include "../player/player.h"
#include "../card/trick.h"

Signal<void(Game&)> Game::signal_open_;

/** constructor
 **
 ** @param    party   party, the game belongs to
 ** @param    startplayer   the startplayer of the game
 **/
Game::Game(Party& party, Player& startplayer) :
  party_(&party),
  startplayer_(&startplayer),
  gameplay_(make_unique<Gameplay>(*this)),
  seed_(party.seed()),
  cards_(*this),
  poverty_(*this),
  marriage_(*this),
  swines_(*this),
  tricks_(*this),
  announcements_(*this),
  teaminfo_(*this)
{
  players_ = make_unique<Players>(*this);
}

/** constructor
 **
 ** @param    game   game to copy
 ** @param    players   the players for the new game
 **/
Game::Game(Game const& game, vector<unique_ptr<Player>>& players) :
  party_(game.party_),
  players_(make_unique<Players>(*this, game.players(), players)),
  gameplay_(game.gameplay_->clone(*this)),
  seed_(game.seed_),
  type_(game.type_),
  cards_(game.cards_, *this),
  poverty_(game.poverty_, *this),
  marriage_(game.marriage_, *this),
  swines_(game.swines_, *this),
  tricks_(game.tricks_, *this),
  announcements_(game.announcements_, *this),
  teaminfo_(game.teaminfo_, *this),
  finished_(game.finished_)
{
  // set the players
  if (game.startplayer_)
    this->startplayer_ = &this->player(game.startplayer_->no());

  DEBUG_ASSERT(this->self_check());
} // Game::Game(Game game, vector<unique_ptr<Players>> players)

/** destructor
 **/
Game::~Game() = default;

/** self check
 ** when an error is found, an ASSERTION is created
 **
 ** @return   wether the self-check was successful (no error)
 **/
bool
Game::self_check() const
{
  if (!this->isvirtual())
    DEBUG_ASSERTION((this == &::party->game()),
		    "Game::self_check()\n"
		    "  this game is not virutal and not the game of the party");

  DEBUG_ASSERTION(this->party_,
		  "Game::self_check()\n"
		  "  no pointer to a party\n"
		  "  this = " << this);

  DEBUG_ASSERTION((this->party_ == ::party.get()),
		  "Game::self_check()\n"
		  "  different party: "
		  "this->party  = " << this->party_
		  << " != " << ::party.get() << " = ::party");

  if (this->startplayer_)
    DEBUG_ASSERTION((this == &this->startplayer_->game()),
                    "Game::self_check()\n"
                    "  the startplayer is not of the same game");

  this->players().self_check();
  this->swines().self_check();
  this->tricks().self_check();

  return true;
} // bool Game::self_check() const

/** output of the game
 **
 ** @param    ostr   output stream
 **/
void
Game::write(ostream& ostr) const
{
  ostr << "Game:\n"
    << "-----\n"
    << "\n"
    << "seed" << ": " << this->seed() << "\n"
    << "current trick number: " << this->tricks().current_no() << "\n"
    << "type: " << this->type() << "\n";

#ifdef OUTDATED
  // DK 2005-03-31
  // this information is saved in the party

  ostr << "\n"
    << "players" << ":\n";
  for (auto const& player : this->players())
    ostr << *player << "\n";
#endif

  ostr << "\n"
    << "teams" << ": \n";
  for (auto const& player : this->players())
    ostr << player.no() << ": " << player.team() << "\n";

  ostr << "\n"
    << "teaminfo" << ": \n";
  this->teaminfo().write(ostr);

  this->tricks().write(ostr);
} // void Game::write(ostream& ostr) const

/** output of the hands
 **
 ** @param    ostr   output stream
 **/
void
Game::write_hands(ostream& ostr) const
{
  ostr << "\n"
    << "hands:\n"
    << "\n";
  for (auto const& p : this->players()) {
    ostr << p.no() << ' ' << p.name() << '\n'
      << p.hand()
      << '\n';
  }
}

/** output of the game
 **
 ** @param    ostr   output stream
 ** @param    game   game to be printed
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr,  Game const& game)
{
  game.write(ostr);
  return ostr;
}

/** @return   whether this is a virtual game
 **/
bool
Game::isvirtual() const
{ return (this != &this->party().game()); }

/** @return   whether the game is finished
 **/
bool
Game::is_finished() const
{ return this->finished_; }

/** @return   the rules of the game
 **/
Party const&
Game::party() const
{
  DEBUG_ASSERTION(this->party_ == ::party.get(),
                  "Game::party()\n"
                  "  different party: "
                  "this->party  = " << this->party_
                  << " != " << ::party.get() << " = ::party");
  return *this->party_;
}

/** @return   the rules of the game
 **/
Rule const&
Game::rule() const noexcept
{
  return this->party().rule();
}

/** @return   whether the second dolle goes over the first in the current trick
 **/
bool
Game::second_dolle_over_first() const
{
  auto const& rule = this->rule();
  return (   rule(Rule::Type::dollen)
          && !(   rule(Rule::Type::dollen_first_over_second_with_swines)
               && this->swines().swines_announced())
          && (  !this->tricks().is_last_trick()
              ? rule(Rule::Type::dollen_second_over_first)
              : (rule(Rule::Type::dollen_second_over_first)
                 != rule(Rule::Type::dollen_contrary_in_last_trick))
             )
         );
} // bool Game::second_dolle_over_first() const

/** @return   the bock multiplier
 **/
int
Game::bock_multiplier() const
{
  return this->party().current_bock_multiplier();
}

/** @return   whether this is a duty solo
 **/
bool
Game::is_duty_solo() const noexcept
{
  return this->is_duty_solo_;
}

/** @return   the current seed
 **/
Seed
Game::seed() const noexcept
{ return this->seed_; }

/** @param   seed   new seed
 **/
void
Game::set_seed(Seed const seed) noexcept
{ this->seed_ = seed; }

/** @return   the gameplay
 **/
Gameplay const&
Game::gameplay() const noexcept
{ return *this->gameplay_; }

/** @return   current game-type
 **
 ** @todo   show 'marriage_silent' if both club queens have been played
 **/
GameType
Game::type() const noexcept
{
  if (   !this->is_finished()
      && (this->type_ == GameType::marriage_silent)
      && (this->marriage().selector() == MarriageSelector::silent) )
    return GameType::normal;

  return this->type_;
}

/** sets the gametype
 **
 ** @param    type   new gametype
 **/
void
Game::set_type(GameType const type) noexcept
{
  this->type_ = type;
  this->cards().update();
}

/** @return   whether this is a solo
 **/
bool
Game::is_solo() const noexcept
{
  return ::is_solo(this->type());
}

/** @return   whether this is a real solo (solo, but no silent marriage)
 **/
bool
Game::is_real_solo() const noexcept
{
  return ::is_real_solo(this->type());
}

/** sets whether this is a duty solo
 **
 ** @return   whether this is a duty solo
 **/
bool
Game::set_is_duty_solo()
{
  this->is_duty_solo_ = false;
  auto const& party = this->party();

  if (   party.is_duty_soli_round()
      && (::game_status <= GameStatus::game_redistribute)) {
    this->is_duty_solo_ = true;
    return true;
  }

  if (this->is_real_solo()) {
    auto const& soloplayer = (  (::game_status <= GameStatus::game_reservation)
                              ? this->players().current_player()
                              : this->players().soloplayer());
    if (   is_color_solo(this->type())
        && party.remaining_duty_color_soli(soloplayer))
      this->is_duty_solo_ = true;

    if (   is_picture_solo(this->type())
        && party.remaining_duty_picture_soli(soloplayer))
      this->is_duty_solo_ = true;

    if (party.remaining_duty_free_soli(soloplayer))
      this->is_duty_solo_ = true;
  }

  return this->is_duty_solo();
} // bool Game::set_is_duty_solo()

/** set the startplayer
 **
 ** @param    startplayer   the startplayer
 **/
void
Game::set_startplayer(unsigned const startplayer)
{
  DEBUG_ASSERTION(::game_status <= GameStatus::game_init,
                  "Game::set_startplayer(" << startplayer << ")\n"
                  "  game_status = " << ::game_status << " > game_init");

  this->startplayer_ = &this->player(startplayer);
}

/** @return   whether the startplayer is not incremented (t.i. duty solo but not in the duty solo round or lust solo and lustsolo player leads)
 **/
bool
Game::startplayer_stays() const
{
  if (!::is_real_solo(this->type()))
    return false;
  // some solo
  if (this->party().is_duty_soli_round())
    return (   (::game_status == GameStatus::game_reservation)
            || (this->players().soloplayer() != this->party().players()[this->party().startplayer()]));
  if (this->is_duty_solo())
    return true;
  // some lust solo
  return this->rule()(Rule::Type::lustsolo_player_leads);
} // bool Game::startplayer_stays() const

/** @return   the cards
 **/
GameCards const&
Game::cards() const noexcept
{
  return this->cards_;
}

/** @return   the cards
 **/
GameCards&
Game::cards() noexcept
{
  return this->cards_;
}

/** @return   the poverty
 **/
GamePoverty const&
Game::poverty() const noexcept
{
  return this->poverty_;
}

/** @return   the poverty
 **/
GamePoverty&
Game::poverty() noexcept
{
  return this->poverty_;
}

/** @return   the marriage
 **/
GameMarriage const&
Game::marriage() const noexcept
{
  return this->marriage_;
}

/** @return   the marriage
 **/
GameMarriage&
Game::marriage() noexcept
{
  return this->marriage_;
}

/** @return   the swines
 **/
GameSwines const&
Game::swines() const noexcept
{
  return this->swines_;
}

/** @return   the swines
 **/
GameSwines&
Game::swines() noexcept
{
  return this->swines_;
}

/** @return   the tricks
 **/
GameTricks const&
Game::tricks() const noexcept
{
  return this->tricks_;
}

/** @return   the tricks
 **/
GameTricks&
Game::tricks() noexcept
{
  return this->tricks_;
}

/** @return   the announcements
 **/
Announcements const&
Game::announcements() const noexcept
{
  return this->announcements_;
}

/** @return   the announcements
 **/
Announcements&
Game::announcements() noexcept
{
  return this->announcements_;
}

/** @return   the teaminfo
 **/
Game::TeamInfo const&
Game::teaminfo() const noexcept
{
  return this->teaminfo_;
}

/** @return   the teaminfo
 **/
Game::TeamInfo&
Game::teaminfo() noexcept
{
  return this->teaminfo_;
}

/** -> result
 **
 ** @param    game1   game 1 to compare
 ** @param    game2   game 2 to compare
 **
 ** @return   whether the games are equal (they lie in the same memory)
 **/
bool
operator==(Game const& game1, Game const& game2) noexcept
{
  return (&game1 == &game2);
}

/** -> result
 **
 ** @param    game1   game 1 to compare
 ** @param    game2   game 2 to compare
 **
 ** @return   whether the games are different
 **/
bool
operator!=(Game const& game1, Game const& game2) noexcept
{
  return !(game1 == game2);
}

/** @return   signal when the game is opened
 **/
Signal<void(Game&)>&
Game::signal_open() noexcept
{ return Game::signal_open_; }

/** @return   signal, when the cards are distributed
 **/
Signal<void()>&
Game::signal_cards_distributed() noexcept
{ return this->signal_cards_distributed_; }

/** @return   signal, when the cards are redistributed
 **/
Signal<void()>&
Game::signal_redistribute() noexcept
{ return this->signal_redistribute_; }

/** @return   signal, when a reservation is got
 **/
Signal<void(Reservation const&, Player const&)>&
Game::signal_reservation_got() noexcept
{ return this->signal_reservation_got_; }

/** @return   signal, when the game is started
 **/
Signal<void()>&
Game::signal_start() noexcept
{ return this->signal_start_; }

/** @return   signal, when the game is finished
 **/
Signal<void()>&
Game::signal_finish() noexcept
{ return this->signal_finish_; }

/** @return   signal, when the game is closed
 **/
Signal<void()>&
Game::signal_close() noexcept
{ return this->signal_close_; }

/** @return   signal for a gameplay action
 **/
Signal<void(GameplayAction const&)>&
Game::signal_gameplay_action() noexcept
{ return this->signal_gameplay_action_; }

/** @return   signal, when a card shall be get
 **/
Signal<void(Player const&)>&
Game::signal_get_card() noexcept
{ return this->signal_get_card_; }

/** @return   signal, when a card is played
 **/
Signal<void(HandCard const&)>&
Game::signal_card_played() noexcept
{ return this->signal_card_played_; }

/** @return   signal, when the ai tests a card
 **/
Signal<void(Card const&, unsigned)>&
Game::signal_ai_test_card() noexcept
{ return this->signal_ai_test_card_; }

/** @return   signal, for a card weighting calculated from the ai
 **/
Signal<void(unsigned)>&
Game::signal_ai_card_weighting() noexcept
{ return this->signal_ai_card_weighting_; }

/** @return   signal, when a card has been played in a virtual game
 **/
Signal<void(HandCard const&)>&
Game::signal_virtual_card_played() noexcept
{ return this->signal_virtual_card_played_; }

/** @return   signal, when a virtual trick is full
 **/
Signal<void()>&
Game::signal_virtual_trick_full() noexcept
{ return this->signal_virtual_trick_full_; }

/** @return   signal, when the gametype has changed
 **/
Signal<void()>&
Game::signal_gametype_changed() noexcept
{ return this->signal_gametype_changed_; }
