/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "game.h"

#include "gameplay_actions.h"

#include "../party/party.h"
#include "../party/rule.h"
#include "../player/playersDb.h"
#include "../player/ai/ai.h"
#include "../player/ai/heuristics.h"
#include "../card/trick.h"
#include "../ui/ui.h"
#include "../misc/setting.h"
#include "../os/bug_report_replay.h"
#include "../utils/random.h"

#include <ctime>


#ifdef DKNOF
#include "../card/weighted_card_list.h"
#endif

/** plays a game (get reservation, make the tricks, give the result)
 **/
void
Game::play()
{
  if (::game_status != GameStatus::game_new)
    return ;

redistribute: // if the cards have to be redistributed
  for (auto& p : *this->players_)
    p.game_open(*this);

  this->set_is_duty_solo();
  if (::game_status == GameStatus::game_redistribute) {
    ::game_status = GameStatus::game_new;
    this->teaminfo().reset();
    this->set_type(GameType::normal);
  } else { // if !(::game_status == GameStatus::game_redistribute)
    if (!this->isvirtual()) {
      Game::signal_open_(*this);
      ::ui->game_open(*this);
    }
  } // if !(::game_status == GameStatus::game_redistribute)

  if (::game_status != GameStatus::game_new) {
    if (!this->isvirtual()) {
      this->signal_close_();
      ::ui->game_close();
    }
    return ;
  } // if (::game_status != GameStatus::game_new)

  ::game_status = GameStatus::game_init;
  this->init();
  if (::fast_play & FastPlay::seed_info)
    return ;

  if (   (::game_status != GameStatus::game_reservation)
      && (::game_status != GameStatus::game_redistribute))
    return ;

  if (::game_status == GameStatus::game_reservation)
    ::game_status = GameStatus::game_start;

  this->players().sort_hands();

  // update the team info in 'game' and in the players
  this->teaminfo().set_at_gamestart();
  this->set_is_duty_solo();

  // set the startplayer
  if (   this->is_duty_solo()
      || (   this->rule()(Rule::Type::lustsolo_player_leads)
          && ::is_solo(this->type())) )
    this->players().set_current_player(this->players().soloplayer());
  else
    this->players().set_current_player(this->startplayer());

  if (!this->isvirtual()) {
    this->signal_start_();
    ::ui->game_start();
  }
  for (auto& p : *this->players_) {
    p.game_start();
  }

  if (::game_status == GameStatus::game_redistribute) {
    this->poverty().reset();

    this->set_seed(this->party_->next_seed());

    if (!this->isvirtual()) {
      this->signal_redistribute_();
      ::ui->game_redistribute();
    }

    goto redistribute;
  } // if (::game_status == GameStatus::game_redistribute)

  // testing swines
  this->swines().test_swines_from_reservations();

  if (::game_status != GameStatus::game_start)
    return ;

  // resort for not announced swines/hyperswines
  this->players().sort_hands();

  ::game_status = GameStatus::game_play;

  // the reserve is needed because the players keep pointers to the tricks
  // in their trickpiles
  for (unsigned t = 0; t < this->rule()(Rule::Type::number_of_tricks_in_game); ++t) {
    try {
      this->tricks_.add(Trick(this->players().current_player()));
      for (auto& p : *this->players_)
        p.trick_open(this->tricks().current());
      this->gameplay_->gameplay_action(GameplayActions::TrickOpen());
      if (!this->isvirtual()) {
        this->tricks().signal_trick_open_(this->tricks().current());
        ::ui->trick_open(this->tricks().current());
        this->signal_gameplay_action_(GameplayActions::TrickOpen());
        ::ui->gameplay_action(GameplayActions::TrickOpen());
      }
      this->announcements().request_from_players();

      while(!this->tricks().current().isfull()) {
        this->announcements().request_from_players();
        this->nextplayer();
        if (::game_status != GameStatus::game_play)
          throw(::game_status);
      } // while(!this->tricks().current().isfull())

      ::game_status = GameStatus::game_full_trick;
      this->evaluatetrick();
      ::game_status = GameStatus::game_play;
      this->announcements().request_from_players();
    } catch (...) {
      if (!this->isvirtual()) {
        this->tricks().signal_trick_close_(this->tricks().current());
        ::ui->trick_close();
      }
      throw;
    } // try

    if (!this->isvirtual()) {
      this->tricks().signal_trick_close_(this->tricks().current());
      ::ui->trick_close();
    }
  } // for (t < this->trickno())

  return ;
} // void Game::play()

/** plays a game according to the gameplay
 ** This is a virtual game, the gameplay is from another game
 **
 ** @param    gameplay    the gameplay
 **/
void
Game::play(Gameplay const& gameplay)
{
  for (auto const& action : gameplay.actions()) {
    switch (action->type()) {
    case GameplayActions::Type::reservation:
    case GameplayActions::Type::poverty_shift:
    case GameplayActions::Type::poverty_accepted:
    case GameplayActions::Type::poverty_returned:
    case GameplayActions::Type::poverty_denied:
    case GameplayActions::Type::poverty_denied_by_all:
      // already done
      break;
    case GameplayActions::Type::card_played:
      this->play_card(dynamic_cast<GameplayActions::CardPlayed*>(action.get())->card);
      break;
    case GameplayActions::Type::announcement:
      this->announcements().make_announcement(dynamic_cast<GameplayActions::Announcement*>(action.get())->announcement,
                                              this->player(dynamic_cast<GameplayActions::Announcement*>(action.get())->player));
      break;
    case GameplayActions::Type::swines:
      this->swines().swines_announce(this->player(dynamic_cast<GameplayActions::Swines*>(action.get())->player));
      break;
    case GameplayActions::Type::hyperswines:
      this->swines().hyperswines_announce(this->player(dynamic_cast<GameplayActions::Hyperswines*>(action.get())->player));
      break;
    case GameplayActions::Type::marriage:
      // todo
      break;
    case GameplayActions::Type::genscher:
      // todo
      break;
    case GameplayActions::Type::trick_open:
      this->tricks_.add(Trick(this->players().current_player()));
      for (auto& p : *this->players_)
        p.trick_open(this->tricks().current());
      break;
    case GameplayActions::Type::trick_full:
      this->evaluatetrick();
      break;
    case GameplayActions::Type::trick_taken:
    case GameplayActions::Type::check:
    case GameplayActions::Type::print:
      // nothing to do
      break ;
    } // switch (action->type())
  } // for (auto const& action : gameplay.actions())
} // void Game::play(Gameplay gameplay)

/** initializes a new game -- get reservations and set the game type
 **/
void
Game::init()
{
  this->players_->reset();
  this->poverty_.reset();
  this->marriage_.reset();
  this->swines_.reset();
  this->announcements_.init();
  this->teaminfo_.reset();
  this->finished_ = false;

  this->players().set_current_player(this->startplayer());

  this->set_type(GameType::normal);

  // initialise the players
  auto& players = *this->players_;
  for (auto& p : players)
    p.new_game(*this);

  this->distributecards();

  if (game_status != GameStatus::game_init)
    return ;

  ::game_status = GameStatus::game_reservation;
  this->signal_cards_distributed_();
  ::ui->game_cards_distributed();

  this->players().set_current_player(this->startplayer());
  { // get the reservation
    this->signal_reservation_ask_(this->players().current_player());
    ::ui->reservation_ask(this->players().current_player());
    this->set_type(GameType::normal);
    this->players().current_player().reservation_get(this->party().is_duty_soli_round());
    this->gameplay_->gameplay_action(GameplayActions::Reservation(this->players().current_player().no(),
                                                                  this->players().current_player().reservation()));
    this->signal_gameplay_action_(GameplayActions::Reservation(this->players().current_player().no(),
                                                               this->players().current_player().reservation()));
    ::ui->gameplay_action(GameplayActions::Reservation(this->players().current_player().no(),
                                                       this->players().current_player().reservation()));
    this->signal_reservation_got_(this->players().current_player().reservation(),
                                  this->players().current_player());
    ::ui->reservation_got(this->players().current_player().reservation(),
                          this->players().current_player());
    if (game_status != GameStatus::game_reservation)
      return ;
    if (this->players().current_player().reservation().game_type == GameType::redistribute) {
      this->set_type(GameType::redistribute);
      ::game_status = GameStatus::game_redistribute;
      players.set_soloplayer(this->players().current_player());
      return ;
    }

    if (this->party().is_duty_soli_round()) {
      DEBUG_ASSERTION(this->startplayer().is_remaining_duty_solo(this->startplayer().reservation().game_type),
                      "Game::init()\n"
                      "  Duty solo round and player still has to player a solo:\n"
                      << "  free duty solo:    " << this->startplayer().remaining_duty_free_soli() << '\n'
                      << "  free picture solo: " << this->startplayer().remaining_duty_picture_soli() << '\n'
                      << "  free color solo:   " << this->startplayer().remaining_duty_color_soli() << '\n'
                      << "  But he has selected no duty solo:\n"
                      << this->startplayer().reservation());
    }


    if (!(   this->party().is_duty_soli_round()
          && !this->reservation(this->startplayer()).offer_duty_solo) ) {
      for (this->players().next_current_player();
           this->players().current_player() != this->startplayer();
           this->players().next_current_player()) {
        this->set_type(GameType::normal);
        this->signal_reservation_ask_(this->players().current_player());
        ::ui->reservation_ask(this->players().current_player());
        this->set_type(GameType::normal);
        this->players().current_player().reservation_get();
        this->gameplay_->gameplay_action(GameplayActions::Reservation(this->players().current_player().no(),
                                                                      this->players().current_player().reservation()));
        this->signal_gameplay_action_(GameplayActions::Reservation(this->players().current_player().no(),
                                                                   this->players().current_player().reservation()));
        ::ui->gameplay_action(GameplayActions::Reservation(this->players().current_player().no(),
                                                           this->players().current_player().reservation()));
        this->signal_reservation_got_(this->players().current_player().reservation(),
                                      this->players().current_player());
        ::ui->reservation_got(this->players().current_player().reservation(),
                              this->players().current_player());
        if (game_status != GameStatus::game_reservation)
          return ;
      }
    } // if (duty solo selected)
    this->set_type(GameType::normal);
  } // get the reservation

  if (::fast_play & FastPlay::seed_info) {
    this->print_seed_statistics();
    return ;
  }

  // testing a solo game
  for (unsigned i = 0;
       i < this->playerno();
       i++,
       this->players().next_current_player()) {
    if (::is_solo(this->reservation(this->players().current_player()).game_type)
        && this->rule()(this->reservation(this->players().current_player()).game_type)) {
      // the player 'p' is playing a solo
      this->set_type(this->reservation(this->players().current_player()).game_type);
      players.set_soloplayer(this->players().current_player());

      // no other player can play a solo
      if (!(   this->party().is_duty_soli_round()
            && (this->players().current_player() == this->startplayer())
            && this->reservation(this->players().current_player()).offer_duty_solo) )
        break;
    }
  } // for (i < this->playerno())

  if (   (this->type() == GameType::normal)
      || this->rule()(Rule::Type::throwing_before_solo) ) {
    if (this->rule()(Rule::Type::force_redistribute)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( (this->reservation(this->players().current_player()).game_type
              == GameType::redistribute)) {
          this->set_type(GameType::redistribute);
          ::game_status = GameStatus::game_redistribute;
          players.set_soloplayer(this->players().current_player());
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (this->rule()(Rule::Type::force_redistribute))
    // testing fox highest trump
    if (this->rule()(Rule::Type::throw_when_fox_highest_trump)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( (this->reservation(this->players().current_player()).game_type
              == GameType::fox_highest_trump)) {
          // look, whether there is a card that is higher than the fox
          unsigned c;
          for (c = 0; c < this->players().current_player().hand().cardsnumber(); c++)
            // check with diamond ten because of swines!
            if (   Card::diamond_ten.less(this->players().current_player().hand().card(c))
                && !this->players().current_player().hand().card(c).istrumpace())
              break;
          if (c == this->players().current_player().hand().cardsnumber()) {
            this->set_type(GameType::fox_highest_trump);
            ::game_status = GameStatus::game_redistribute;
            players.set_soloplayer(this->players().current_player());
            break;
          } // if (c < this->player(p).hand().cardsnumber())
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::throw_when_fox_highest_trump))

    // testing thrown nines
    if (this->rule()(Rule::Type::throw_with_nines)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( ((this->reservation(this->players().current_player()).game_type
               == GameType::thrown_nines)
             )
            && (this->players().current_player().hand().count_nines()
                >= this->rule()(Rule::Type::min_number_of_throwing_nines))) {
          this->set_type(GameType::thrown_nines);
          ::game_status = GameStatus::game_redistribute;
          players.set_soloplayer(this->players().current_player());
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::throw_with_nines))

    // testing thrown kings
    if (this->rule()(Rule::Type::throw_with_kings)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( ((this->reservation(this->players().current_player()).game_type
               == GameType::thrown_kings)
             )
            && (this->players().current_player().hand().count_kings()
                >= this->rule()(Rule::Type::min_number_of_throwing_kings))) {
          this->set_type(GameType::thrown_kings);
          ::game_status = GameStatus::game_redistribute;
          players.set_soloplayer(this->players().current_player());
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::throw_with_kings))

    // testing thrown nines and kings
    if (this->rule()(Rule::Type::throw_with_nines_and_kings)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( ((this->reservation(this->players().current_player()).game_type
               == GameType::thrown_nines_and_kings)
             )
            && (this->players().current_player().hand().count_nines()
                + this->players().current_player().hand().count_kings()
                >= this->rule()(Rule::Type::min_number_of_throwing_nines_and_kings))) {
          this->set_type(GameType::thrown_nines_and_kings);
          ::game_status = GameStatus::game_redistribute;
          players.set_soloplayer(this->players().current_player());
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::throw_with_nines_and_kings))

    // testing thrown richness
    if (this->rule()(Rule::Type::throw_with_richness)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if ( ((this->reservation(this->players().current_player()).game_type
               == GameType::thrown_richness)
             )
            && (this->players().current_player().hand().points()
                >= this->rule()(Rule::Type::min_richness_for_throwing))) {
          this->set_type(GameType::thrown_richness);
          ::game_status = GameStatus::game_redistribute;
          players.set_soloplayer(this->players().current_player());
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::throw_with_richness))

  } // if (this->type() == GameType::normal)

  if (this->type() == GameType::normal) {
    // testing a marriage
    for (unsigned i = 0;
         i < this->playerno();
         i++,
         this->players().next_current_player()) {
      if ( ((this->reservation(this->players().current_player()).game_type
             == GameType::marriage)
           )
          && (this->players().current_player().hand().count_club_queens() == 2)) {
        this->set_type(GameType::marriage);
        players.set_soloplayer(this->players().current_player());
        this->marriage().selector_
          = this->reservation(this->players().current_player()).marriage_selector;
        DEBUG_ASSERTION(this->rule()(this->marriage().selector()),
                        "Game::init():\n"
                        "  marriage selector '" << this->marriage().selector()
                        << "' is not valid.\n");
      }
    } // for (p < playerno())
  } // if (type() == GameType::normal)

  if (   (this->type() == GameType::normal)
      || (   this->is_solo()
          && this->rule()(Rule::Type::throwing_before_solo) )
      || (   (this->type() == GameType::marriage)
          && !this->rule()(Rule::Type::marriage_before_poverty) )
     ) {
    // testing a poverty
    if (this->rule()(Rule::Type::poverty)) {
      for (unsigned i = 0;
           i < this->playerno();
           i++,
           this->players().next_current_player()) {
        if (   (   (this->reservation(this->players().current_player()).game_type
                    == GameType::poverty)
               )
            && this->players().current_player().hand().has_poverty() ) {
          if (   this->rule()(Rule::Type::throwing_before_solo)
              && (this->type() != GameType::normal)
              && !(   (this->type() == GameType::marriage)
                   && !this->rule()(Rule::Type::marriage_before_poverty) )
              && this->rule()(Rule::Type::poverty_shift)
              && !(   (this->players().current_player().hand().count_trumps() > 1)
                   && this->rule()(Rule::Type::throw_with_one_trump) )
             )
            continue;
          players.set_soloplayer(this->players().current_player());
          this->set_type(GameType::poverty);
          this->swines().test_swines_from_reservations();
          this->poverty().shift();
          if (::game_status != GameStatus::game_reservation)
            return ;
          break;
        } // if (reservation)
      } // for (p < playerno())
    } // if (rule()(Rule::Type::poverty))
  } // if (type() == GameType::normal)

  if (this->type() != GameType::marriage)
    this->marriage().selector_ = MarriageSelector::team_set;

  this->marriage().check_for_silent_marriage();
} // void Game::init()

/** resets the game to the first trick
 ** Does not reset the hands of the players.
 ** This function is to be used to create a virtual copy of game that can be used to replay the game from the beginning (that is after shifting of the poverty)
 **/
void
Game::reset_to_first_trick()
{

#ifdef POSTPONED
  this->player_current_ = &this->player(this->startplayer().no());

  this->marriage_selector_ = MarriageSelector::team_set;
  this->marriage_determination_trickno_ = UINT_MAX;
  this->tricks_.clear();
  this->teaminfo_set_at_gamestart();
  // ToDo: add human to info
  this->announcement_.clear();
  this->swines_owner_ = nullptr;
  this->first_fox_catcher_ = nullptr;
  this->swines_announced_ = false;
  this->hyperswines_owner_ = nullptr;
  this->hyperswines_announced_ = false;
  this->finished_ = false;
  for (auto& p : *this->players_)
    p.set_hand({p->hand().cards_all()});

  if (this->gameplay_) {
    // gameplay 
    Gameplay gameplay;
    gameplay.game_open(*this);
    gameplay.game_cards_distributed();
    gameplay.game_start();
    auto a = this->gameplay_->actions_.begin();
    for (;
         a != this->gameplay_->actions_.end();
         ++a) {
      gameplay.gameplay_action(**a);
    } // for (a : this->gameplay().actions())
    for (auto a2 = a;
         a2 != this->gameplay().actions().end();
         ++a2)
      delete *a2;
    this->gameplay_->actions_.erase(a, this->gameplay_->actions_.end());

    for (unsigned p = 0; p < this->playerno(); ++p)
      this->gameplay_->hands_[p] = Hand(this->player(p),
                                        this->player(p).hand().cards_all());
  } // gameplay 
#endif

  return ;
} // void Game::reset_to_first_trick()

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   whether the player has made a reservation
 **/
bool
Game::has_made_reservation(Player const& player) const
{
  if (::game_status < GameStatus::game_reservation)
    return false;
  if (::game_status > GameStatus::game_reservation)
    return true;

  return ( ( player.no()
            + this->playerno()
            - this->startplayer().no()
           ) % this->playerno()
          <  (this->players().current_player().no()
              + this->playerno()
              - this->startplayer().no()
             ) % this->playerno()
         );
} // bool Game::has_made_reservation(Player player) const

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   the reservation of 'player'
 **/
Reservation const&
Game::reservation(Player const& player) const
{
  return player.reservation();
} // Reservation const& Game::reservation(Player const& player) const


/** distribute the cards randomly at the players
 **/
void
Game::distributecards()
{
  unique_ptr<RandomBase> rand;
  if (::bug_report_replay) {
    if (::bug_report_replay->version() <= Version(0, 7, 16))
      rand = make_unique<RandomPre_0_7_17>(this->seed_);
    else
      rand = make_unique<Random>(this->seed_);
  } else {
    rand = make_unique<Random>(this->seed_);
  }

  vector<Card> all_cards = this->rule().cards();
  for (unsigned i = 1; i < this->rule()(Rule::Type::number_of_same_cards); ++i)
    all_cards.insert(all_cards.end(),
                     this->rule().cards().begin(),
                     this->rule().cards().end());

#if 0
  // Take the random generator from the standard library
  // Bug reports are no longer replayable
  std::random_shuffle(all_cards.begin(), all_cards.end());
#endif

  vector<vector<Card> > hands(this->playerno());

  while (!all_cards.empty()) {
    for (auto& h : hands) {
      DEBUG_ASSERTION(!all_cards.empty(),
                      "Game::distributecards():\n"
                      "  valid cards cannot be distributed between the players");
      auto c = (all_cards.begin() + (*rand)(all_cards.size()));
      h.push_back(*c);
      all_cards.erase(c);
    } // for (h : hands)
  } // while (!cards.empty())

  // create the hands for the players
  for (unsigned p = 0;
       p < this->playerno();
       p++) {
    this->player((p + this->startplayer().no()) % this->playerno()).set_hand(Hand{hands[p]});
  }

  return ;
} // void Game::distributecards()

/** the next player has to play a card
 **
 ** @bug         uses processor time, not real time
 */
void
Game::nextplayer()
{
  DEBUG_ASSERT(this->self_check());

#ifdef POSTPONED
  DEBUG_ASSERTION((this->players().current_player().hand().cardsnumber()
                   == this->trickno() - this->tricks().current_no()),
                  "Game::nextplayer():\n"
                  "  current player has not the right number of cards on the hand\n"
                  "trick " << this->tricks().current_no() << "\n"
                  << this->players().current_player());
#endif


  // ***ToDo DK this uses the processor time
  const clock_t time_before_playing_card = ::clock();

  HandCard const played_card = this->players().current_player().card_get();

  if (!(::game_status & GameStatus::game))
    return ;

  DEBUG_ASSERTION(!played_card.is_empty(),
                  "Game::nextplayer():\n"
                  "  result of 'Player::card_get()' is an empty card.\n"
                  "  Player number: " << this->players().current_player().no() << "\n"
                  "  Player type:   " << this->players().current_player().type());
  // here always update,
  // so that the user can do something during the calculation of the ai
  ::ui->update();

  // testing, whether the card is valid.
  // if it is not, take the first valid card of the hand.
  if (!this->tricks().current().isvalid(played_card)
      && !(   this->isvirtual()
           && (this->players().current_player().hand().cardsnumber()
               > this->players().current_player().cards_to_play()))
     ) {
#ifndef RELEASE
    cerr << "ai type = " << played_card.player().type() << endl;
    cerr << "heuristic = " << dynamic_cast<Ai const&>(this->players().current_player()
                                                     ).last_heuristic()
      << endl;
    cerr << *this;
#endif
#ifndef POSTPONED
    cerr << "hand:   " << this->players().current_player().hand() << endl;
    cerr << "current trick:\n" << this->tricks().current() << endl;
    cerr << "played card: " << played_card << endl;
    cerr << "heuristic:   " << dynamic_cast<Ai const&>(this->players().current_player()).last_heuristic() << endl;
    DEBUG_ASSERTION(false,
                    "Game::nextplayer()\n"
                    "  card \"" << played_card << "\" is not valid"
                    << endl);
#else
    // *** the opposite team has won
#endif
  }

  if (   !(::fast_play & FastPlay::pause)
      && !(   ::bug_report_replay
           && ::bug_report_replay->auto_action())) {

    // wait a bit, before the card is played
    // if it is the human who has to play
    // Bug: DK
    // When the time wraps around, it is assumed, that the calculation
    // has taken no time

    if (this->players().current_player().type() != Player::Type::human) {
      if (!this->isvirtual()) {

#ifdef POSTPONED
        // ToDo: cannot use this code when using 'clock()'
        while (time_before_playing_card
               + (::setting(Setting::Type::card_play_delay)
                  * static_cast<double>(CLOCKS_PER_SEC)
                  / 1000)
               > ::clock() ) {
#endif

          ::ui->sleep(max(long(0),
                          long(::setting(Setting::Type::card_play_delay)
                               - ((::clock()
                                   - time_before_playing_card)
                                  / (static_cast<double>(CLOCKS_PER_SEC)
                                     / 1000) ))));

#ifdef POSTPONED
          // ToDo: cannot use this code when using 'clock()'
        } // while (still have to wait)
#endif
      } // if (!this->isvirtual())
    } // if (this->players().current_player().type() & Player::Type::ai)
  } // if (!(::fast_play & FastPlay::pause))

  this->play_card(this->players().current_player(), played_card);

  // request announcement of the players
  this->announcements().request_from_players();
} // void Game::nextplayer()

/** the player plays the card
 **
 ** @param    player   current player
 ** @param    card     card that is played
 **/
void
Game::play_card(Player& player, Card const& card)
{
  // check that it is the right player
  DEBUG_ASSERTION(player == this->players().current_player(),
                  "Game::play_card(" + player.name() + ", " + to_string(card) + ")\n"
                  "The player is not the current one (" + this->players().current_player().name() + ")");
  this->play_card(card);
} // void Game::play_card(Player& player, Card card)

/** the player plays the card
 **
 ** @param    card     card that is played
 **/
void
Game::play_card(Card const& card_)
{
  Player& player = this->players().current_player();
  // check that the hand contains the card
  DEBUG_ASSERTION(player.hand().contains(card_),
                  "Game::play_card(" + player.name() + ", " + to_string(card_) + ")\n"
                  "The card is not on the hand:\n"
                  << player.hand());

  auto const card = HandCard(player.hand(), card_);

  // testing, whether the card is valid.
  // if it is not, take the first valid card of the hand.
  if (!this->tricks().current().isvalid(card)
      && !(   this->isvirtual()
           && (this->players().current_player().hand().cardsnumber()
               > this->players().current_player().cards_to_play()))
     ) {
    DEBUG_ASSERTION(false,
                    "Game::play_card(" + player.name() + ", " + to_string(card) + ")\n"
                    "Card is not valid.\n");
  }
  if (!this->isvirtual())
    this->marriage().check_for_genscher(card);

  // remove the card from the hand, add it to the trick
  // and give the message to the ui
  this->players().current_player().hand().playcard(card);
  this->tricks().current().add(card);
  // set the next player as current player
  if (this->tricks().current().isfull())
    this->players().set_current_player(this->tricks().current().winnerplayer());
  else
    this->players().next_current_player();

  this->teaminfo().update();
  for (auto& p : this->players())
    p.card_played(card);

  this->marriage().determine_silent_marriage();
  if (this->isvirtual()) {
    this->signal_virtual_card_played_(card);
  } else {
    this->signal_card_played_(card);
    ::ui->card_played(card);
    auto heuristic = Aiconfig::Heuristic::no_heuristic;
    switch (player.type()) {
    case Player::Type::human:
    case Player::Type::ai:
      heuristic = dynamic_cast<Ai const&>(player).last_heuristic();
      break;
    case Player::Type::unset:
    case Player::Type::ai_dummy:
    case Player::Type::ai_random:
      heuristic = Aiconfig::Heuristic::no_heuristic;
      break;
    } // switch (player.type())

    this->gameplay_->gameplay_action(GameplayActions::CardPlayed(player.no(),
                                                                 card,
                                                                 heuristic));
    this->signal_gameplay_action_(GameplayActions::CardPlayed(player.no(),
                                                              card,
                                                              heuristic));
    ::ui->gameplay_action(GameplayActions::CardPlayed(player.no(),
                                                      card,
                                                      heuristic));
  }
} // void Game::play_card(Player& player, Card card)


/** evaluates who is startplayer for nextcard and updates player
 **/
void
Game::evaluatetrick()
{
  auto const& trick = this->tricks().current();
  auto const& winnerplayer = trick.winnerplayer();

  // test for marriage
  this->marriage().determine_marriage();
  if (!(::game_status & GameStatus::game))
    return ;

  this->swines().evaluate_trick(trick);
  this->announcements().evaluate_trick(trick);

  // tell the players and the ui, that the trick is full
  for (auto& p : *this->players_)
    p.trick_full(this->tricks().current());
  if (this->isvirtual()) {
    this->signal_virtual_trick_full_();

    // request announcement of the players
    this->announcements().request_from_players();

  } else { // if !(this->isvirtual())
    this->gameplay_->gameplay_action(GameplayActions::TrickFull(this->tricks().current()));
    this->signal_gameplay_action_(GameplayActions::TrickFull(this->tricks().current()));
    ::ui->gameplay_action(GameplayActions::TrickFull(this->tricks().current()));
    this->tricks().signal_trick_full_(this->tricks().current());
    ::ui->trick_full();

    this->announcements().request_from_players();

    for (auto& p : *this->players_)
      p.trick_taken();

    this->announcements().request_from_players();

    ::game_status = GameStatus::game_trick_taken;
    this->gameplay_->gameplay_action(GameplayActions::TrickTaken());
    this->signal_gameplay_action_(GameplayActions::TrickTaken());
    ::ui->gameplay_action(GameplayActions::TrickTaken());
  } // if !(this->isvirtual())
  if (!(::game_status & GameStatus::game))
    return ;

  // move the trick in the trickpile
  this->tricks().current().move_in_trickpile();

  // Startplayer for next round 
  this->players().set_current_player(winnerplayer);
} // void Game::evaluatetrick();

/** the game is finished
 **/
void
Game::finish()
{
  if (this->marriage().is_undetermined()) {
    auto& players = this->players();
    auto const& soloplayer = players.soloplayer();
    this->marriage().determination_trickno_ = this->tricks().current_no();
    this->set_type(GameType::marriage_solo);
    for (auto& p : players) {
      auto const team = ((p == soloplayer)
                         ? Team::re
                         : Team::contra);
      this->teaminfo_.set(p, team);
    }
    this->marriage().selector_ = MarriageSelector::team_set;
    // update the teaminfo of the players
    for (auto& p : players)
      p.marriage(soloplayer, soloplayer);
  } // if (undetermined marriage)

  this->finished_ = true;
} // void Game::finish()

/** the game is finished
 **/
void
Game::close()
{
  for (auto& p : *this->players_)
    p.game_close(*this);
  this->signal_close()();
  ::ui->game_close();
} // void Game::close()

/** print seed information on stdout
 **/
void
Game::print_seed_statistics() const
{
  // format (for each player)
  // seed
  // playerno
  // playername
  // number trumps
  // number club queens
  // number dollen
  // number queens
  // number jacks
  // number swines
  // number hyperswines
  // number fehl colors
  // number color aces
  // reservation (gametype)

  // next line must be called in 'freedoko.cpp'
  //cout << "seed;no;name;trumpno;club queens;spade queens;dollen;queens;jacks;diamond aces;hyperswines;fehl colors;color aces;reservation;hand value\n";
  for (unsigned p = 0; p < this->playerno(); ++p) {
    Player const& player = this->player(p);
    Hand const& hand = player.hand();
    cout << this->seed() << ';';
    cout << p << ';' << player.name() << ';';
    cout << hand.count_trumps() << ';';
    cout << hand.count_club_queens() << ';';
    cout << hand.count(Card::spade_queen) << ';';
    cout << hand.count(Card::dolle) << ';';
    cout << hand.count(Card::queen) << ';';
    cout << hand.count(Card::jack) << ';';
    cout << hand.count(Card::diamond_ace) << ';';
    cout << hand.count(Card::diamond_nine) << ';';
    // cout << (hand.has_hyperswines() ? "hyperswines" : "") << ';';
    cout << 3 - (  (hand.contains(Card::club) ? 1 : 0)
                 + (hand.contains(Card::spade) ? 1 : 0)
                 + (hand.contains(Card::heart) ? 1 : 0)) << ';';
    cout << hand.count(Card::ace) - hand.count_trump_aces() << ';';
    cout << player.reservation().game_type << ';';
    cout << Heuristics::CalcHandValue(dynamic_cast<Ai const&>(player)) << ';';
    cout << '\n';
  } // for (p)
} // void Game::print_seed_statistics() const
