/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "reservation.h"
#include "players.h"
#include "poverty.h"
#include "marriage.h"
#include "swines.h"
#include "cards.h"
#include "tricks.h"
#include "announcements.h"
#include "teaminfo.h"
#include "../card/card.h"
#include "../card/hand_cards.h"
class Party;
class Rule;
class Player;
class Human;
class Trick;
class Hand;

class Gameplay;
#include "gameplay_action.h"

class Game {
  friend class Party;
  friend class GamePlayers;
  friend class GamePoverty;
  friend class GameMarriage;
  friend class GameSwines;
  friend class GameSummary;
  friend class Announcements;
  friend class GameTeamInfo;
  friend class Ai;
  friend class WVirtualGames;
  friend class WMonteCarlo;
  friend class Gametree;
  public:
  using Players = GamePlayers;
  using Player = GamePlayer;
  using Poverty = GamePoverty;
  using Marriage = GameMarriage;
  using Swines = GameSwines;
  using Cards = GameCards;
  using Tricks = GameTricks;
  using TeamInfo = GameTeamInfo;

  public:
  Game(Party& party, Player& startplayer);
  // constructor (for virtual games)
  //Game(Game const& game, vector<Player*>& players);
  Game(Game const& game, vector<unique_ptr<Player>>& players);
  Game() = delete;
  Game(Game const& game) = delete;
  Game& operator=(Game const& g) = delete;
  ~Game();


  bool self_check() const;
  // resets the game to the beginning of the first trick
  // (that is after the cards distribution and reservation)
  void reset_to_first_trick();

  void write(ostream& ostr) const;
  void write_hands(ostream& ostr = cout) const;

  bool isvirtual() const;
  bool is_finished() const;

  Party const& party() const;
  Rule const& rule() const noexcept;
  bool second_dolle_over_first() const;

  int bock_multiplier() const;
  bool is_duty_solo() const noexcept;

  Players const& players() const;
  Players& players();
  Player const& player(unsigned p) const;
  Player& player(unsigned p);
  unsigned playerno() const;

  Seed seed() const noexcept;
  void set_seed(Seed seed) noexcept;

  Gameplay const& gameplay() const noexcept;

  GameType type() const noexcept;
  void set_type(GameType type) noexcept;
  bool is_solo() const noexcept;
  bool is_real_solo() const noexcept;
  bool set_is_duty_solo();

  void set_startplayer(unsigned startplayer);
  bool startplayer_stays() const;

  // update the cards (according to the game type)
  void update_cards();


  // players

  Player const& startplayer() const;

  bool has_made_reservation(Player const& player) const;
  Reservation const& reservation(Player const& player) const;

  // infos of the gameplay

  // the number of points the player has got
  unsigned points_of_player(Player const& player) const;
  // the number of points the player has got
  unsigned points_of_player(unsigned playerno) const;
  // the number of points the player till the trick 'trickno' has got
  unsigned points_of_player(Player const& player,
                            unsigned trickno) const;
  // the number of points the player till the trick 'trickno' has got
  unsigned points_of_player(unsigned playerno,
                            unsigned trickno) const;
  // the points, the team has got
  unsigned points_of_team(Team const& team) const;
  // the played points
  unsigned played_points() const;
  // the points still in the game
  unsigned remaining_points() const;
  // the points needed to win the game
  unsigned needed_points_to_win(Team const& team) const;
  // the winnerteam
  Team winnerteam() const;


  Cards const& cards() const noexcept;
  Cards& cards() noexcept;

  Poverty const& poverty() const noexcept;
  Poverty& poverty() noexcept;

  Marriage const& marriage() const noexcept;
  Marriage& marriage() noexcept;

  Swines const& swines() const noexcept;
  Swines& swines() noexcept;

  Tricks const& tricks() const noexcept;
  Tricks& tricks() noexcept;

  Announcements const& announcements() const noexcept;
  Announcements& announcements() noexcept;

  TeamInfo const& teaminfo() const noexcept;
  TeamInfo& teaminfo() noexcept;


  // team (real team)

  // the team of the player
  Team team(Player const& player) const;


  // gameplay

  // play the game
  void play();
  // play the game according to the gameplay
  void play(Gameplay const& gameplay);
  private:

  // init the game
  void init();
  // distribute the cards
  void distributecards();
  protected:
  // the next player has to play a card
  void nextplayer();
  public:
  // the player plays a card
  void play_card(Player& player, Card const& card);
  void play_card(Card const& card);
  // evaluate the full trick
  void evaluatetrick();
  private:
  // finish the game
  void finish();

  public:
  void close();

  private:
  // print statistics of the seeds
  void print_seed_statistics() const;

  public:
  // signals
  static Signal<void(Game&)>& signal_open() noexcept;
  Signal<void()>& signal_cards_distributed() noexcept;
  Signal<void()>& signal_redistribute() noexcept;
  Signal<void(Reservation const&, Player const&)>& signal_reservation_got() noexcept;
  Signal<void()>& signal_start() noexcept;
  Signal<void()>& signal_finish() noexcept;
  Signal<void()>& signal_close() noexcept;

  Signal<void(GameplayAction const&)>& signal_gameplay_action() noexcept;

  Signal<void(Player const&)>& signal_get_card() noexcept;
  Signal<void(HandCard const&)>& signal_card_played() noexcept;

  Signal<void(Card const&, unsigned)>& signal_ai_test_card() noexcept;
  Signal<void(unsigned)>& signal_ai_card_weighting() noexcept;
  Signal<void(HandCard const&)>& signal_virtual_card_played() noexcept;
  Signal<void()>& signal_virtual_trick_full() noexcept;

  Signal<void()>& signal_gametype_changed() noexcept;

  private:
  // party, the game belongs to
  Party* party_ = {};

  unique_ptr<Players> players_;
  Player const* startplayer_ = {};

  unique_ptr<Gameplay> gameplay_;

  Seed seed_ = 0;
  GameType type_ = GameType::normal;
  bool is_duty_solo_ = false;

  Cards cards_;

  Poverty poverty_;
  Marriage marriage_;
  Swines swines_;

  Tricks tricks_;
  Announcements announcements_;

  TeamInfo teaminfo_;

  bool finished_ = false;

  // signals
 
  static Signal<void(Game&)> signal_open_;
  Signal<void()> signal_cards_distributed_;
  Signal<void()> signal_redistribute_;
  Signal<void(Player const&)> signal_reservation_ask_;
  Signal<void(Reservation const&, Player const&)> signal_reservation_got_;
  Signal<void()> signal_start_;
  Signal<void()> signal_finish_;
  Signal<void()> signal_close_;

  Signal<void(GameplayAction const&)> signal_gameplay_action_;

  Signal<void(Player const&)> signal_get_card_;
  Signal<void(HandCard const&)> signal_card_played_;

  Signal<void(Card const&, unsigned)> signal_ai_test_card_;
  Signal<void(unsigned)> signal_ai_card_weighting_;
  Signal<void(HandCard const&)> signal_virtual_card_played_;
  Signal<void()> signal_virtual_trick_full_;

  Signal<void()> signal_gametype_changed_;
}; // class Game

// output of the game
ostream& operator<<(ostream& ostr, Game const& game);

// comparison
bool operator==(Game const& game1, Game const& game2) noexcept;
bool operator!=(Game const& game1, Game const& game2) noexcept;
