/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"

#include "game_summary.h"

#include "game.h"
#include "../party/party.h"
#include "../party/rule.h"
#include "../card/trick.h"
#include "../utils/string.h"
#include "../class/readconfig/readconfig.h"

/** Constructor
 **
 ** @param    game   game the summary is made of
 **
 ** @todo   remove const cast
 **/
GameSummary::GameSummary(Game const& game) :
  rule_(&game.rule()),
  seed_(game.seed()),
  players_number_(game.players().size()),
  party_players_(game.players().size()),
  startplayer_(game.startplayer().no()),
soloplayer_(  game.players().exists_soloplayer()
            ? game.players().soloplayer().no()
            : UINT_MAX),
  game_type_(game.type()),
  duty_solo_(game.is_duty_solo()),
swines_player_(   game.swines().swines_announced()
               ? game.swines().swines_owner().no()
               : UINT_MAX),
hyperswines_player_(  game.swines().hyperswines_announced()
                    ? game.swines().hyperswines_owner().no()
                    : UINT_MAX)
{
  for (auto const& p : game.players())
    this->party_players_[p.no()] = p.no();
  this->evaluate(game);
} // GameSummary::GameSummary(Game game) :

/** Constructor
 **
 ** @param    rule   the corresponding rule
 ** @param    istr   input stream to read the game from
 **/
GameSummary::GameSummary(Rule const& rule, istream& istr) :
  rule_(&rule)
{
  this->read(istr);
} // GameSummary::GameSummary(Rule rule, istream& istr) :

/** writes the game summary into 'ostr'
 **
 ** @param    ostr   output stream
 **/
void
GameSummary::write(ostream& ostr) const
{
  ostr << "seed = " << Unsigned(this->seed()) << '\n';
  ostr << "players number = " << Unsigned(this->players_number()) << '\n';
  ostr << "party players =";
  for (auto const& p : this->party_players_)
    ostr << " " << p;
  ostr << '\n';
  ostr << "startplayer number = " << Unsigned(this->startplayer()) << '\n';
  ostr << "soloplayer number = " << Unsigned(this->soloplayer()) << '\n';
  ostr << "gametype = " << this->game_type() << '\n';
  ostr << "duty solo = " << this->is_duty_solo() << '\n';
  ostr << "swines player number = " << Unsigned(this->swines_player()) << '\n';
  ostr << "hyperswines player number = " << Unsigned(this->hyperswines_player()) << '\n';
  ostr << "teams =";
  for (auto t = begin(this->teams_); t != end(this->teams_); ++t) {
    if (t != begin(this->teams_))
      ostr << ',';
    ostr << ' ' << *t;
  }
  ostr << '\n';
  ostr << "highest announcement re = " << this->highest_announcement_re() << '\n';
  ostr << "highest announcement contra = " << this->highest_announcement_contra() << '\n';
  ostr << "trick points =";
  for (auto p = begin(this->trick_points_);
       p != end(this->trick_points_);
       ++p) {
    if (p != begin(this->trick_points_))
      ostr << ',';
    ostr << ' ' << *p;
  }
  ostr << '\n';
  ostr << "winnerteam = " << this->winnerteam() << '\n';
  ostr << "points = " << this->points() << '\n';
  ostr << "special points\n"
    << "{\n"
    << this->specialpoints_
    << "}\n";
  if (!this->bock_triggers_.empty()) {
    ostr << "bock triggers\n"
      << "{\n";
    for (auto const& c : this->bock_triggers_)
      ostr << c << '\n';
    ostr << "}\n";
  } // if (!this->bock_triggers().empty())
} // void GameSummary::write(ostream& ostr) const

/** reads the game summary from 'ostr'
 **
 ** @param    istr   input stream
 **/
void
GameSummary::read(istream& istr)
{

  Unsigned seed = UINT_MAX;
  Unsigned players_number = UINT_MAX;
  vector<unsigned> party_players;
  Unsigned startplayer = UINT_MAX;
  Unsigned soloplayer = UINT_MAX;

  GameType game_type = GameType::normal;
  bool duty_solo = false;
  Unsigned swines_player = UINT_MAX;
  Unsigned hyperswines_player = UINT_MAX;
  vector<Team> teams;
  Announcement highest_announcement_re = Announcement::noannouncement;
  Announcement highest_announcement_contra = Announcement::noannouncement;
  vector<unsigned> trick_points;
  Team winnerteam = Team::noteam;
  int points = INT_MAX;
  Specialpoints specialpoints;
  vector<BockTrigger> bock_triggers;

  { // read the data
    Config config; // the config read
    int depth = 0; // the depth in the blocks
    do { // while (istr)
      istr >> config;

      if (istr.eof())
        break;

      if (istr.fail())
        break;

      if ((config.name == "") && (config.value == ""))
        break;

      if (config.value.empty()) {
        if (config.name == "{") {
          depth += 1;
        } else if (config.name == "}") {
          depth -= 1;
          // finished with the game summary
          if (depth == -1)
            istr.putback('}');
          if (depth <= 0)
            break;

        } else if (config.name == "special points") {
          string line;
          line = String::remove_blanks(::read_line(istr));
          if (line != "{")
            throw ReadException("Game summary read: "
                                "special points: expected '{', "
                                " found '" + line + "'");

          while (istr.good()
                 && (istr.peek() != '}')) {
            specialpoints.emplace_back(istr);
            while (isspace(istr.peek())
                   && istr.good())
              istr.get();
          }

          line = String::remove_blanks(::read_line(istr));
          if (line != "}")
            throw ReadException("Game summary read: "
                                "special points: expected '}', "
                                " found '" + line + "'");
        } else if (config.name == "bock triggers") {
          string line;
          line = String::remove_blanks(::read_line(istr));
          if (line != "{")
            throw ReadException("Game summary read: "
                                "bock triggers: expected '{', "
                                " found '" + line + "'");

          while (istr.good()
                 && (istr.peek() != '}')) {
            line = String::remove_blanks(::read_line(istr));
            bock_triggers.push_back(bock_trigger_from_string(line));
            while (isspace(istr.peek())
                   && istr.good())
              istr.get();
          }

          line = String::remove_blanks(::read_line(istr));
          if (line != "}")
            throw ReadException("Game summary read: "
                                "bock triggers: expected '}', "
                                " found '" + line + "'");
        } else { // if (config.name == ...)
          throw ReadException("Game summary read: unknown line "
                              "'" + config.name + "'");
        } // if (config.name == ...)
      } else { // if !(config.value.empty())
        istringstream istr(config.value);
        if (config.name == "seed") {
          istr >> seed;
        } else if (config.name == "players number") {
          istr >> players_number;
          if (party_players.empty()) {
            for (unsigned p = 0; p < players_number; ++p)
              party_players.push_back(p);
          }
        } else if (config.name == "party players") {
          for (auto& p : party_players)
            istr >> p;
        } else if (config.name == "startplayer number") {
          istr >> startplayer;
        } else if (config.name == "soloplayer number") {
          istr >> soloplayer;
        } else if (config.name == "gametype") {
          game_type = game_type_from_string(config.value);
        } else if (config.name == "duty solo") {
          duty_solo = (   (config.value == "true")
                       || (config.value == "yes")
                       || (config.value == "1"));
        } else if (config.name == "swines player number") {
          istr >> swines_player;
        } else if (config.name == "hyperswines player number") {
          istr >> hyperswines_player;
        } else if (config.name == "teams") {
          string::size_type pos_begin = 0;
          while ((pos_begin < config.value.size())
                 && isspace(config.value[pos_begin]))
            ++pos_begin;
          while (pos_begin < config.value.size()) {
            string::size_type const pos_end
              = config.value.find(',', pos_begin);
            auto const name = string(config.value,
                                     pos_begin,
                                     ((pos_end == string::npos)
                                      ? string::npos
                                      : (pos_end - pos_begin)));

            teams.push_back(team_from_string(name));
            if (pos_end == string::npos)
              break;
            pos_begin = pos_end + 1;
            while ((pos_begin < config.value.size())
                   && isspace(config.value[pos_begin]))
              ++pos_begin;
          } // while (pos_begin < config.value.size())
        } else if (config.name == "highest announcement re") {
          highest_announcement_re = announcement_from_string(config.value);
        } else if (config.name == "highest announcement contra") {
          highest_announcement_contra = announcement_from_string(config.value);
        } else if (config.name == "trick points") {
          string::size_type pos_begin = 0;
          while ((pos_begin < config.value.size())
                 && isspace(config.value[pos_begin]))
            ++pos_begin;
          while (pos_begin < config.value.size()) {
            auto const pos_end = config.value.find(',', pos_begin);
            auto const name = string(config.value,
                                     pos_begin,
                                     ((pos_end == string::npos)
                                      ? string::npos
                                      : (pos_end - pos_begin)));

            trick_points.push_back(Unsigned(name));
            if (trick_points.back() == UINT_MAX)
              throw ReadException("Game summary read: invalid trickpoints "
                                  "'" + string(config.value,
                                               pos_begin,
                                               pos_end - pos_begin)
                                  + "'");
            if (pos_end == string::npos)
              break;
            pos_begin = pos_end + 1;
            while ((pos_begin < config.value.size())
                   && isspace(config.value[pos_begin]))
              ++pos_begin;
          } // while (pos_begin < config.value.size())
        } else if (config.name == "winnerteam") {
          winnerteam = team_from_string(config.value);
        } else if (config.name == "points") {
          istr >> points;
        } else { // if (config.name == ...)
          throw ReadException("Game summary read: unknown line "
                              "'" + config.name + " = " + config.value + "'");
        } // if (config.name == ...)
        if (!istr.good() && !istr.eof())
          throw ReadException("Game summary read: error loading "
                              "'" + config.name + "'");
      } // if !(config.value.empty())

    } while (istr.good());

    if (!istr.good())
      throw ReadException("Game summary read: unknown error");
  } // read the data

  { // test whether all data was read
    if (seed == UINT_MAX)
      throw ReadException("Game summary read: seed not loaded");
    if (players_number == UINT_MAX)
      throw ReadException("Game summary read: players number not loaded");
    if (startplayer == UINT_MAX)
      throw ReadException("Game summary read: startplayer number not loaded");
    if (teams.size() != players_number)
      throw ReadException("Game summary read: teams not loaded");
    if (trick_points.size() != players_number)
      throw ReadException("Game summary read: trick points not loaded");
    if (points == INT_MAX)
      throw ReadException("Game summary read: points number not loaded");

  } // test whether all data was read

  { // set the data
    // explicit casts of Unsigned because of problems with MinGW g++-3.2
    this->seed_ = seed;
    this->players_number_ = players_number;
    this->party_players_ = party_players;
    this->startplayer_ = startplayer;
    this->soloplayer_ = soloplayer;
    this->game_type_ = game_type;
    this->duty_solo_ = duty_solo;
    this->swines_player_ = swines_player;
    this->hyperswines_player_ = hyperswines_player;
    this->teams_ = teams;
    this->highest_announcement_re_ = highest_announcement_re;
    this->highest_announcement_contra_ = highest_announcement_contra;
    this->trick_points_ = trick_points;
    this->winnerteam_ = winnerteam;
    this->points_ = points;
    this->specialpoints_ = specialpoints;
    this->bock_triggers_ = bock_triggers;
  } // set the data
} // void GameSummary::read(istream& istr)

/** @return   the rule
 **/
Rule const&
GameSummary::rule() const
{ return *this->rule_; }

/** @return   the seed
 **/
Seed
GameSummary::seed() const
{ return this->seed_; }

/** @return   the number of players
 **/
unsigned
GameSummary::players_number() const
{ return this->players_number_; }

/** -> result
 **
 ** @param    p   number of the player in the game
 **
 ** @return   the number of the player in the party
 **/
unsigned
GameSummary::party_player(unsigned const p) const
{
  DEBUG_ASSERTION(p < this->party_players_.size(),
                  "GameSummary::party_player(" << p << ")\n"
                  "  There are only " << this->party_players_.size() << " players");
  return this->party_players_[p];
} // unsigned GameSummary::party_player(unsigned p) const

/** -> result
 **
 ** @param    player_no   number of the player in the party
 **
 ** @return   whether the player has played
 **/
unsigned
GameSummary::game_player(Player const& player) const
{
  for (unsigned p = 0; p < this->party_players_.size(); ++p)
    if (this->party_players_[p] == ::party->players().no(player))
      return p;
  return UINT_MAX;
} // unsigned GameSummary::game_player(Player const& player) const

/** -> result
 **
 ** @param    player    player of the party
 **
 ** @return   whether the player has played
 **/
bool
GameSummary::has_played(Player const& player) const
{
  for (auto const& p : this->party_players_) {
    if (p == ::party->players().no(player))
      return true;
  }
  return false;
} // bool GameSummary::has_played(Player player) const

/** @return   the startplayer
 **/
unsigned
GameSummary::startplayer() const
{ return this->startplayer_; }

/** @return   the soloplayer
 **/
unsigned
GameSummary::soloplayer() const
{ return this->soloplayer_; }

/** @return   the game type
 **/
GameType
GameSummary::game_type() const
{ return this->game_type_; }

/** @return   the swines player
 **/
unsigned
GameSummary::swines_player() const
{ return this->swines_player_; }

/** @return   the hyperswines player
 **/
unsigned
GameSummary::hyperswines_player() const
{ return this->hyperswines_player_; }

/** @return   the highest announcement of re
 **/
Announcement
GameSummary::highest_announcement_re() const
{ return this->highest_announcement_re_; }

/** @return   the highest announcement of contra
 **/
Announcement
GameSummary::highest_announcement_contra() const
{ return this->highest_announcement_contra_; }

/** @return   the winner team
 **/
Team
GameSummary::winnerteam() const
{ return this->winnerteam_; }

/** @return   whether the game was a duty solo
 **/
bool
GameSummary::is_solo() const
{
  return ::is_solo(this->game_type());
} // bool GameSummary::is_solo() const

/** @return   whether the game was a duty solo
 **/
bool
GameSummary::is_duty_solo() const
{
  return this->duty_solo_;
} // bool GameSummary::is_duty_solo() const

/** @return   whether the game was a lust solo
 **/
bool
GameSummary::is_lust_solo() const
{
  return (   this->is_solo()
          && !this->is_duty_solo());
} // bool GameSummary::is_lust_solo() const

/** -> result
 ** the check of duty solo round must be made by party
 **
 ** @return   whether the startplayer is not incremented (t.i. duty solo or lust solo and lustsolo player leads)
 **/
bool
GameSummary::startplayer_stays() const
{
  if (!is_real_solo(this->game_type()))
    return false;
  // some solo
  if (this->is_duty_solo())
    return true;
  // some lust solo
  return this->rule()(Rule::Type::lustsolo_player_leads);
} // bool GameSummary::startplayer_stays() const

/** @return   the number of the next startplayer
 **/
unsigned
GameSummary::next_startplayer() const
{
  if (this->startplayer_stays())
    return this->startplayer();
  else
    return (this->startplayer() + 1) % this->players_number();
} // unsigned GameSummary::next_startplayer() const

/** @return   the team which gets the points
 **/
Team
GameSummary::team_of_points() const
{
  if (this->winnerteam() == Team::noteam) {
    switch(this->rule()(Rule::Type::counting)) {
    case Counting::plus:
      if (this->points() > 0)
        return Team::re;
      else
        return Team::contra;

    case Counting::minus:
      if (this->points() < 0)
        return Team::re;
      else
        return Team::contra;

    case Counting::plusminus:
      return Team::re;

    } // switch(this->rule()(Rule::Type::counting))
  } else { // if !(this->winnerteam() == Team::noteam)
    switch(this->rule()(Rule::Type::counting)) {
    case Counting::plus:
      return this->winnerteam();

    case Counting::minus:
      return opposite(this->winnerteam());

    case Counting::plusminus:
      return this->winnerteam();
    } // switch(this->rule()(Rule::Type::counting))
  } // if !(this->winnerteam() == Team::noteam)

  return Team::noteam;
} // Team GameSummary::team_of_points() const

/** -> result
 **
 ** @param    team   the team
 **
 ** @return   whether the team gets points
 **/
bool
GameSummary::get_points(Team const team) const
{
  if (team == Team::noteam)
    return false;

  if (this->winnerteam() == Team::noteam) {
    switch(this->rule()(Rule::Type::counting)) {
    case Counting::plus:
      return ( ( (this->points() > 0)
                && (team == Team::re))
              || ( (this->points() < 0)
                  && (team == Team::contra)) );
    case Counting::minus:
      return ( ( (this->points() < 0)
                && (team == Team::re))
              || ( (this->points() > 0)
                  && (team == Team::contra)) );

    case Counting::plusminus:
      return (this->points_ != 0);

    } // switch(this->rule()(Rule::Type::counting))

  } else { // if !(p.winnerteam() == Team::noteam)
    switch(this->rule()(Rule::Type::counting)) {
    case Counting::plus:
    case Counting::minus:
      return (team == this->team_of_points());

    case Counting::plusminus:
      return true;
    } // switch(this->rule()(Rule::Type::counting))
  } // if !(p.winnerteam() == Team::noteam)

  DEBUG_ASSERT(false);
  return false;
} // bool GameSummary::get_points(Team team) const

/** @return   the points
 **/
int
GameSummary::points() const
{ return this->points_; }

/** -> result
 **
 ** @param    team   the team
 **
 ** @return   the points the team gets
 **/
int
GameSummary::points(Team const team) const
{
  if (team == Team::noteam)
    return 0;
  if (!this->get_points(team))
    return 0;

  // in a solo game 'points()' returns three times the value of 'points_'
  switch(this->rule().counting()) {
  case Counting::plus:
    if (   this->is_solo()
        && (   this->rule()(Rule::Type::solo_always_counts_triple)
            || (this->winnerteam() == Team::re)) )
      return 3 * this->points();

    if ( (this->winnerteam() == Team::noteam)
        && (team == Team::contra) )
      return -this->points();

    return this->points();

  case Counting::minus:
    if (   this->is_solo()
        && (   this->rule()(Rule::Type::solo_always_counts_triple)
            || (this->winnerteam() == Team::contra)) )
      return 3 * (-this->points());

    if ( (this->winnerteam() == Team::noteam)
        && (team == Team::re) )
      return this->points();

    return -this->points();

  case Counting::plusminus:
    if (this->is_solo()) {
      if (this->winnerteam() == Team::re) {
        if (team == this->winnerteam())
          return 3 * this->points();
        else
          return -this->points();
      } else { // if !(this->winnerteam() == Team::re)
        if (team == this->winnerteam())
          return this->points();
        else
          return -3 * this->points();
      } // if !(this->winnerteam() == Team::re)
    } // if (this->is_solo())

    if (team == this->team_of_points())
      return  this->points();
    else
      return -this->points();
  } // switch(this->rule().counting())

  DEBUG_ASSERTION(false,
                  "GameSummary::points(team):\n"
                  "  counting type not known: " << this->rule().counting());
  return 0;
} // int GameSummary::points(Team team) const

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   whether the player gets points
 **/
bool
GameSummary::get_points(Player const& player) const
{
  return get_points(this->team(player));
} // bool GameSummary::get_points(Player player) const

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   the points the player gets, 0 if he has not played in the game
 **/
int
GameSummary::points(Player const& player) const
{
  return this->points(this->team(player));
} // int GameSummary::points(Player player) const

/** -> result
 **
 ** @param    playerno   the player
 **
 ** @return   the number of points player has made in his tricks
 **/
unsigned
GameSummary::trick_points(Player const& player) const
{
  return (  this->has_played(player)
          ? this->trick_points_[this->game_player(player)]
          : 0);
} // unsigned GameSummary::trick_points(Player player) const

/** -> result
 **
 ** @param    team   the team
 **
 ** @return   the points made by the team
 **/
unsigned
GameSummary::trick_points(Team const team) const
{
  unsigned points = 0;
  for (unsigned p = 0; p < this->trick_points_.size(); ++p)
    if (this->teams_[p] == team)
      points += this->trick_points_[p];

  return points;
} // unsigned GameSummary::trick_points(Team team) const

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   team of the player, noteam if the player has not played
 **/
Team
GameSummary::team(Player const& player) const
{
  return (  this->has_played(player)
          ? this->teams_[this->game_player(player)]
          : Team::noteam);
} // Team GameSummary::team(Player player) const

/** @return   the specialpoints
 **/
Specialpoints const&
GameSummary::specialpoints() const
{ return this->specialpoints_; }

/** -> result
 **
 ** @param    i   number of special point
 **
 ** @return   the i'th special point
 **/
Specialpoint const&
GameSummary::specialpoint(unsigned const i) const
{
  DEBUG_ASSERTION(i < this->specialpoints_.size(),
                  "GameSummary::specialpoints(i = " << i << ")\n"
                  "i is too great, must be less then " << this->specialpoints_.size());

  return this->specialpoints_[i];
} // Specialpoint GameSummary::specialpoint(unsigned i) const

/** @return   the bock multiplier
 **/
int
GameSummary::bock_multiplier() const
{
  int multiplier = 1;
  for (auto const& s : this->specialpoints_)
    if (s.type == Specialpoint::Type::bock)
      multiplier *= s.value;

  return multiplier;
} // int GameSummary::bock_multiplier() const

/** @return   the bock triggers
 **/
vector<BockTrigger> const&
GameSummary::bock_triggers() const
{ return this->bock_triggers_; }


/** evaluate the game
 **
 ** @param    game   game to evaluate
 **/
void
GameSummary::evaluate(Game const& game)
{
  auto const& rule = game.rule();

  // save the team and the trick points
  for (unsigned p = 0; p < game.players().size(); ++p) {
    auto const& player = game.player(p);
    this->teams_.push_back(player.team());
    this->trick_points_.push_back(game.points_of_player(player));
  } // for (p < game.playerno())

  // save the winnerteam
  this->winnerteam_ = game.winnerteam();

  // calculate highest announcements
  map<Announcement, unsigned> re_announcement_player;
  map<Announcement, unsigned> contra_announcement_player;
  Announcement reply_announcement = Announcement::noannouncement;
  unsigned reply_player = UINT_MAX;
  for (unsigned p = 0; p < game.players().size(); ++p) {
    auto const& player = game.player(p);
    auto const& announcements = game.announcements().all(player);

    if (   !announcements.empty()
        && is_reply(announcements.back())) {
      reply_announcement = announcements.back();
      reply_player = p;
    }

    if (player.team() == Team::re) {
      if (player.announcement() > this->highest_announcement_re_)
        this->highest_announcement_re_ = player.announcement();

      for (auto const& a : announcements)
        re_announcement_player[a.announcement] = p;
    } // if (player.team() == Team::re)

    if (player.team() == Team::contra) {
      if( player.announcement() > highest_announcement_contra_)
        this->highest_announcement_contra_ = player.announcement();

      for (auto const& a : announcements)
        contra_announcement_player[a.announcement] = p;
    } // if (player.team() == Team::re)
  } // for (p < game.players().size())

  { // special points
    // Order of the points
    // * won
    // * re no 120 said
    // * contra no 120 said
    // * re no 90
    // * contra no 90
    // * no 90 reply
    // * re no 90 said
    // * contra no 90 said
    // * re no 90 said && contra > 120 points
    // * contra no 90 said && re > 120 points
    // * re no 60
    // * contra no 60
    // * no 60 reply
    // * re no 60 said
    // * contra no 60 said
    // * re no 60 said && contra > 90 points
    // * contra no 60 said && re > 90 points
    // * re no 30
    // * contra no 30
    // * no 30 reply
    // * re no 30 said
    // * contra no 30 said
    // * re no 30 said && contra > 60 points
    // * contra no 30 said && re > 60 points
    // * re no 0
    // * contra no 0
    // * no 0 reply
    // * re no 0 said
    // * contra no 0 said
    // * no 120 reply
    // * re no 0 said && contra > 30 points
    // * contra no 0 said && re > 30 points
    // * against club queen
    // * multiplicator 3
    // * specialpoints

    { // announcements

      // shortcuts
      // got no 90
#define TRICKpoints_NO(A) \
      if ((game.points_of_team(Team::re) < A) \
          && (game.winnerteam() != Team::re)) \
      this->specialpoints_.emplace_back(Specialpoint::Type::no ## A, \
                                        Team::contra); \
      if ((game.points_of_team(Team::contra) < A) \
          && (game.winnerteam() != Team::contra)) \
      this->specialpoints_.emplace_back(Specialpoint::Type::no ## A, \
                                        Team::re)
      // announced no 60
#define TRICKpoints_NO_ANNOUNCED(A) \
      if (rule(Rule::Type::knocking)) { \
        for (unsigned p = 0; p < this->players_number_; ++p) { \
          if (   (game.player(p).announcement() >= Announcement::no ## A) \
              && (game.winnerteam() != Team::noteam)) { \
            Specialpoint sp(Specialpoint::Type::no ## A ## _said, \
                            game.player(p).team(), \
                            game.winnerteam()); \
            sp.player_of_no = p; \
            this->specialpoints_.push_back(sp); \
          } \
        } \
      } else { \
        if (is_real(this->highest_announcement_re()) \
            && (this->highest_announcement_re() >= Announcement::no ## A ) \
            && (game.winnerteam() != Team::noteam)) { \
          Specialpoint sp(Specialpoint::Type::no ## A ## _said, \
                          Team::re, \
                          game.winnerteam()); \
          sp.player_of_no = re_announcement_player[Announcement::no ## A]; \
          this->specialpoints_.push_back(sp); \
        } \
        if (is_real(this->highest_announcement_contra()) \
            && (this->highest_announcement_contra() >= Announcement::no ## A ) \
            && (game.winnerteam() != Team::noteam)) { \
          Specialpoint sp(Specialpoint::Type::no ## A ## _said, \
                          Team::contra, \
                          game.winnerteam()); \
          sp.player_of_no = contra_announcement_player[Announcement::no ## A]; \
          this->specialpoints_.push_back(sp); \
        } \
      }
      // got 120 (against 'no 90')
#define TRICKpoints_GOT_AGAINST(A, B) \
      if ((this->highest_announcement_contra() == Announcement::no ## A) \
          && (game.points_of_team(Team::re) >= A + 30)) \
      this->specialpoints_.emplace_back(Specialpoint::Type::no ## A ## _said_ ## B ## _got, \
                                        Team::re); \
      if ((this->highest_announcement_re() == Announcement::no ## A) \
          && (game.points_of_team(Team::contra) >= A + 30)) \
      this->specialpoints_.emplace_back(Specialpoint::Type::no ## A ## _said_ ## B ## _got, \
                                        Team::contra)
      // replied on announcement
#define ANNOUNCEMENT_REPLY(A) \
      if ( (reply_announcement == Announcement::no ## A ## _reply) \
          && (game.winnerteam() != Team::noteam)) { \
        Specialpoint sp( Specialpoint::Type::no ## A ## _reply, \
                        game.player(reply_player).team(), \
                        game.winnerteam()); \
        sp.player_of_no = reply_player; \
        this->specialpoints_.push_back(sp); \
      } else
      // re announced no 60

#define points(A, B) \
      TRICKpoints_NO(A); \
      TRICKpoints_NO_ANNOUNCED(A); \
      ANNOUNCEMENT_REPLY(A); \
      TRICKpoints_GOT_AGAINST(A, B);

      { // won
        if (game.winnerteam() != Team::noteam)
          this->specialpoints_.emplace_back(Specialpoint::Type::won,
                                            game.winnerteam());
      } // won
      if (rule(Rule::Type::announcement_contra_doubles_against_re)) {
        // against club queens
        if (   !::is_solo(this->game_type())
            && (game.winnerteam() == Team::contra) )
          this->specialpoints_.emplace_back(Specialpoint::Type::contra_won,
                                            Team::contra);
      } // if (Rule::Type::announcement_contra_doubles_against_re)

      if (rule(Rule::Type::announcement_re_doubles)) {
        points(90, 120);
        points(60, 90);
        points(30, 60);

        { // 0
          // re no 0
          if (( !game.tricks().has_trick(Team::re))
              && (game.winnerteam() != Team::re))
            this->specialpoints_.emplace_back(Specialpoint::Type::no0,
                                              Team::contra);
          // contra no 0
          if (( !game.tricks().has_trick(Team::contra) )
              && (game.winnerteam() != Team::contra))
            this->specialpoints_.emplace_back(Specialpoint::Type::no0,
                                              Team::re);
          TRICKpoints_NO_ANNOUNCED(0);
          ANNOUNCEMENT_REPLY(0);
          TRICKpoints_GOT_AGAINST(0, 30);
        } // 0

        ANNOUNCEMENT_REPLY(120);
      } // if (rule(Rule::Type::announcement_re_doubles))

      if (game.winnerteam() != Team::noteam) {
        if (rule(Rule::Type::knocking)) {
          // no 120 announced
          for (unsigned p = 0; p < game.players().size(); ++p) {
            if (   (game.player(p).announcement() >= Announcement::no120)
                && (game.winnerteam() != Team::noteam)) {
              Specialpoint sp(Specialpoint::Type::no120_said,
                              game.player(p).team(),
                              game.winnerteam());
              sp.player_of_no = p;
              this->specialpoints_.push_back(sp);
            }
          } // for (p)
        } else { // if !(rule(Rule::Type::knocking))
          // re announced
          if (   (  (this->highest_announcement_re() >= Announcement::no120)
                  && is_real(this->highest_announcement_re()) )
              || (this->highest_announcement_re() == Announcement::reply) ) {
            Specialpoint sp(Specialpoint::Type::no120_said,
                            Team::re,
                            game.winnerteam());
            sp.player_of_no
              = re_announcement_player[( (this->highest_announcement_re()
                                          == Announcement::reply)
                                        ? Announcement::reply
                                        : Announcement::no120) ];
            this->specialpoints_.push_back(sp);
          }
          // contra announced
          if (   (   (this->highest_announcement_contra() >= Announcement::no120)
                  && is_real(this->highest_announcement_contra()) )
              || (this->highest_announcement_contra() == Announcement::reply) ) {
            Specialpoint sp(Specialpoint::Type::no120_said,
                            Team::contra,
                            game.winnerteam());
            sp.player_of_no
              = contra_announcement_player[( (this->highest_announcement_contra()
                                              == Announcement::reply)
                                            ? Announcement::reply
                                            : Announcement::no120) ];
            this->specialpoints_.push_back(sp);
          }
        } // if !(rule(Rule::Type::knocking))
      } // if (game.winnerteam() != Team::noteam)

      if (!rule(Rule::Type::announcement_re_doubles)) {
        points(90, 120);
        points(60, 90);
        points(30, 60);

        { // 0
          // re no 0
          if (( !game.tricks().has_trick(Team::re))
              && (game.winnerteam() != Team::re))
            this->specialpoints_.emplace_back(Specialpoint::Type::no0,
                                              Team::contra);
          // contra no 0
          if (( !game.tricks().has_trick(Team::contra) )
              && (game.winnerteam() != Team::contra))
            this->specialpoints_.emplace_back(Specialpoint::Type::no0,
                                              Team::re);
          TRICKpoints_NO_ANNOUNCED(0);
          ANNOUNCEMENT_REPLY(0);
          TRICKpoints_GOT_AGAINST(0, 30);
        } // 0

        ANNOUNCEMENT_REPLY(120);
      } // if (!rule(Rule::Type::announcement_re_doubles))

      if (!rule(Rule::Type::announcement_contra_doubles_against_re)) {
        // against club queens
        if (   !::is_solo(this->game_type())
            && (game.winnerteam() == Team::contra) )
          this->specialpoints_.emplace_back(Specialpoint::Type::contra_won,
                                            Team::contra);
      } // if (Rule::Type::announcement_contra_doubles_against_re)

#undef TRICKpoints_RE_NO
#undef TRICKpoints_CONTRA_NO
#undef TRICKpoints_RE_GOT
#undef TRICKpoints_CONTRA_GOT
#undef ANNOUNCEMENT_REPLY
#undef TRICKpoints_RE_NO_ANNOUNCED
#undef TRICKpoints_CONTRA_NO_ANNOUNCED
#undef points

    } // announcements

    // multiplicator 3 for solo games
    if (   ::is_solo(this->game_type())
        && (   rule(Rule::Type::solo_always_counts_triple)
            || (this->winnerteam() == Team::re) ) ) {
      Specialpoint sp(Specialpoint::Type::solo, this->winnerteam());
      sp.player_of_no = game.players().soloplayer().no();
      this->specialpoints_.push_back(sp);
    }

    // last collect all specialpoints of each trick
    for (auto const& t : game.tricks()) {
      for (auto const& s : t.specialpoints()) {
        if (s.is_valid(this->teams_)) {
          this->specialpoints_.push_back(s);
        }
      }
    }

    { // bock triggers
      if (rule(Rule::Type::bock_120)) {
        // bock after 120-120 points
        if (game.points_of_team(Team::re) == game.points_of_team(Team::contra))
          this->bock_triggers_.push_back(BockTrigger::equal_points);
      } // if (rule(Rule::Type::bock_120))

      if (rule(Rule::Type::bock_solo_lost)) {
        // bock after a lost solo game
        if (   ::is_solo(game.type())
            && (game.winnerteam() == Team::contra))
          this->bock_triggers_.push_back(BockTrigger::solo_lost);
      } // if (rule(Rule::Type::bock_solo_lost))

      if (rule(Rule::Type::bock_re_lost)) {
        // bock after a lost re/contra game (only re/contra is announced)
        if (   (   game.announcements().last(Team::re)
                != Announcement::noannouncement)
            && (game.winnerteam() != Team::re))
          this->bock_triggers_.push_back(BockTrigger::re_lost);
        if (   (   game.announcements().last(Team::contra)
                != Announcement::noannouncement)
            && (game.winnerteam() != Team::contra))
          this->bock_triggers_.push_back(BockTrigger::contra_lost);
      } // if (rule(Rule::Type::bock_re_lost))

      if (rule(Rule::Type::bock_heart_trick)) {
        if (!::is_solo(game.type())) {
          // bock after a real heart trick
          for (auto const& t : game.tricks()) {
            auto c = t.cards().begin();
            for (; c != t.cards().end(); ++c) {
              if (c->tcolor() != Card::heart) {
                break;
              }
            }

            if (c == t.cards().end())
              this->bock_triggers_.push_back(BockTrigger::heart_trick);
          } // for (t : game.tricks())
        } // if (!::is_solo(game.type()))
      } // if (rule(Rule::Type::bock_heart_trick))

      if (rule(Rule::Type::bock_black)) {
        // bock after a black game (a team got no trick)
        if (   !game.tricks().has_trick(Team::re)
            || !game.tricks().has_trick(Team::contra) )
          this->bock_triggers_.push_back(BockTrigger::black);
      } // if (rule(Rule::Type::bock_black))

      if (game.bock_multiplier() != 1) {
        this->specialpoints_.emplace_back(Specialpoint::Type::bock,
                                          game.bock_multiplier());
      } // if (game.bock_multiplier())
    } // bock triggers

    // calculating the main game points
    this->points_ = sum(this->specialpoints_, this->winnerteam(), game);

    DEBUG_ASSERTION((   (this->winnerteam() == Team::re)
                     || (this->winnerteam() == Team::contra)
                     || (this->winnerteam() == Team::noteam)),
                    "GameSummary::evaluate(Game):" << '\n'
                    << "  wrong team at end: " << this->winnerteam());

  } // special points
} // void GameSummary::evaluate(Game game)

/** writes the game summary into the stream
 **
 ** @param    ostr      output stream
 ** @param    game_summary   game summary
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, GameSummary const& game_summary)
{
  game_summary.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, GameSummary game_summary)
