/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "gameplay.h"
#include "gameplay_actions.h"

#include "game.h"
#include "../card/trick.h"
#include "../card/hand.h"
#include "../player/player.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
Gameplay::Gameplay(Game const& game) :
  game_(&game)
{ }

/** constructor
 **
 ** @param    game       corresponding game
 ** @param    gameplay   gameplay to copy
 **/
Gameplay::Gameplay(Game const& game, Gameplay const& gameplay) :
  game_(&game),
  seed_(gameplay.seed_),
  gametype_(gameplay.gametype_),
  soloplayer_(gameplay.soloplayer_),
  hands_(gameplay.hands_)
{
  for (auto const& a : gameplay.actions_)
    this->actions_.push_back(a->clone());
  for (auto& h : this->hands_)
    h.set_player(game.player(h.player().no()));

} // Gameplay::Gameplay(Game game, Gameplay gameplay)

/** destructor
 **/
Gameplay::~Gameplay() = default;

/** @param    game new corresponding game
 **
 ** @return   clone of the gameplay
 **/
unique_ptr<Gameplay>
Gameplay::clone(Game const& game) const
{
  return make_unique<Gameplay>(game, *this);
} // unique_ptr<Gameplay> Gameplay::clone() const


/** writes the gameplay in the output stream
 **
 ** @param    ostr   output stream
 **
 ** @return   output stream
 **/
ostream&
Gameplay::write(ostream& ostr) const
{
  ostr
    << "seed = " << this->seed_ << '\n'
    << "gametype = " << this->gametype_ << '\n'
    << "soloplayer = " << this->soloplayer_ << '\n';

  ostr
    << "hands\n"
    << "{\n";
  for (auto const& h : this->hands_) {
    ostr
      << "{\n"
      << h
      << "}\n";
  }
  ostr << "}\n";

  ostr
    << "actions\n"
    << "{\n";
  for (auto const& a : this->actions_)
    ostr << "  " << *a << '\n';
  ostr << "}\n";

  return ostr;
} // ostream& Gameplay::write(ostream& ostr) const

/** the cards are distributed
 **/
void
Gameplay::cards_distributed()
{
  this->seed_ = this->game_->seed();

  this->hands_.clear();
  for (auto const& p : this->game_->players())
    this->hands_.emplace_back(p.hand());
} // void Gameplay::cards_distributed()

/** the game is redistributed
 **/
void
Gameplay::redistribute()
{
  this->actions_.clear();
} // void Gameplay::redistribute()

/** the game is started
 **/
void
Gameplay::game_start()
{
  this->soloplayer_ = (  this->game_->is_solo()
                       ? this->game_->players().soloplayer().no()
                       : UINT_MAX);
} // void Gameplay::game_start()

/** @return   the actions
 **/
vector<unique_ptr<GameplayAction>> const&
Gameplay::actions() const
{ return this->actions_; }

/** a gameplay action
 **
 ** @param    action   the gameplay action
 **/
void
Gameplay::gameplay_action(GameplayAction const& action)
{
  this->actions_.push_back(action.clone());
} // void Gameplay::gameplay_action(GameplayAction action)

/** writes the gameplay in the output stream
 **
 ** @param    ostr   output stream
 ** @param    gameplay   gameplay
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Gameplay const& gameplay)
{
  gameplay.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, Gameplay gameplay)
