/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "gameplay_action.h"
#include "gameplay_actions.h"

#include "../card/trick.h"
#include "../card/hand.h"
#include "../player/player.h"

#include "../utils/string.h"

namespace GameplayActions {

/** returns an action type determined by the line
 **
 ** @param    line   line with the action data
 **
 ** @return   action corresponding to the line
 **   (to be freed)
 **/
unique_ptr<GameplayAction>
GameplayAction::create(string const& line)
{
  if (line.empty())
    return {};

  auto const keyword
    = String::remove_blanks(  (std::find(line.begin(), line.end(), '=')
					  != line.end())
				       ? string(line.begin(),
						std::find(line.begin(), line.end(), '=') - 1)
				       : line
				      );

  if (keyword == to_string(Type::reservation))
    return make_unique<Reservation>(line);
  else if (keyword == to_string(Type::marriage))
    return make_unique<Marriage>(line);
  else if (keyword == to_string(Type::genscher))
    return make_unique<Genscher>(line);
  else if (keyword == to_string(Type::poverty_shift))
    return make_unique<PovertyShift>(line);
  else if (keyword == to_string(Type::poverty_accepted))
    return make_unique<PovertyAccepted>(line);
  else if (keyword == to_string(Type::poverty_returned))
    return make_unique<PovertyReturned>(line);
  else if (keyword == to_string(Type::poverty_denied))
    return make_unique<PovertyDenied>(line);
  else if (keyword == to_string(Type::poverty_denied_by_all))
    return make_unique<PovertyDeniedByAll>(line);
  else if (keyword == to_string(Type::announcement))
    return make_unique<Announcement>(line);
  else if (keyword == to_string(Type::swines))
    return make_unique<Swines>(line);
  else if (keyword == to_string(Type::hyperswines))
    return make_unique<Hyperswines>(line);
  else if (keyword == to_string(Type::card_played))
    return make_unique<CardPlayed>(line);
  else if (keyword == to_string(Type::trick_open))
    return make_unique<TrickOpen>(line);
  else if (keyword == to_string(Type::trick_full))
    return make_unique<TrickFull>(line);
  else if (keyword == to_string(Type::trick_taken))
    return make_unique<TrickTaken>(line);
  else if (   (keyword == to_string(Type::check))
           || (keyword == "CHECK"))
    return make_unique<Check>(line);
  else if (   (keyword == to_string(Type::print))
           || (keyword == "PRINT"))
    return make_unique<Print>(line);

  return {};
} // static unique_ptr<GameplayAction> GameplayAction::create(string line)

/** returns an action type determined by the line
 **
 ** @param    line   line with the action data
 ** @param    istr   input stream (behind the line)
 **
 ** @return   action corresponding to the line
 **   (to be freed)
 **
 ** @todo   remove (take the function without istr)
 **/
unique_ptr<GameplayAction>
GameplayAction::create(string const& line, istream& istr)
{
  if (line.empty())
    return {};

#ifndef OUTDATED
  // old bug report format (< 0.7.3)
  auto const keyword
    = String::remove_blanks
    ( (std::find(line.begin(), line.end(), ':')
       < std::find(line.begin(), line.end(), '='))
     ? string(line.begin(),
	      std::find(line.begin(), line.end(), ':'))
     : (  (std::find(line.begin(), line.end(), '=')
	   != line.end())
	? string(line.begin(),
		 std::find(line.begin(), line.end(), '='))
	: line
       )
    );
#else
  auto const keyword
    = String::remove_blanks
    (  (std::find(line.begin(), line.end(), '=')
	!= line.end())
     ? string(line.begin(),
	      std::find(line.begin(), line.end(), '=') - 1)
     : line
    );
#endif

#ifndef OUTDATED
  // old bug report format (< 0.7.3)
  if (keyword == "Marriage")
    return make_unique<Marriage>(line);
  else if (keyword == "Genscher")
    return make_unique<Genscher>(line);
  else if (keyword == "Poverty Shift")
    return make_unique<PovertyShift>(line);
  else if (keyword == "Poverty Returned")
    return make_unique<PovertyReturned>(line);
  else if (keyword == "Poverty Denied By All")
    return make_unique<PovertyDeniedByAll>(line);
  else if (keyword == "Announcement")
    return make_unique<Announcement>(line);
  else if (keyword == "Swines")
    return make_unique<Swines>(line);
  else if (keyword == "Hyperswines")
    return make_unique<Hyperswines>(line);
  else if (keyword == "Played")
    return make_unique<CardPlayed>(line);
  else if (keyword == "Trick full")
    return make_unique<TrickFull>(line, istr);
#endif

  if (keyword == to_string(Type::reservation))
    return make_unique<Reservation>(line);
  else if (keyword == to_string(Type::marriage))
    return make_unique<Marriage>(line);
  else if (keyword == to_string(Type::genscher))
    return make_unique<Genscher>(line);
  else if (keyword == to_string(Type::poverty_shift))
    return make_unique<PovertyShift>(line);
  else if (keyword == to_string(Type::poverty_accepted))
    return make_unique<PovertyAccepted>(line);
  else if (keyword == to_string(Type::poverty_returned))
    return make_unique<PovertyReturned>(line);
  else if (keyword == to_string(Type::poverty_denied))
    return make_unique<PovertyDenied>(line);
  else if (keyword == to_string(Type::poverty_denied_by_all))
    return make_unique<PovertyDeniedByAll>(line);
  else if (keyword == to_string(Type::announcement))
    return make_unique<Announcement>(line);
  else if (keyword == to_string(Type::swines))
    return make_unique<Swines>(line);
  else if (keyword == to_string(Type::hyperswines))
    return make_unique<Hyperswines>(line);
  else if (keyword == to_string(Type::card_played))
    return make_unique<CardPlayed>(line);
  else if (keyword == to_string(Type::trick_open))
    return make_unique<TrickOpen>(line);
  else if (keyword == to_string(Type::trick_full))
    return make_unique<TrickFull>(line, istr);
  else if (keyword == to_string(Type::trick_taken))
    return make_unique<TrickTaken>(line);
  else if (keyword == to_string(Type::check))
    return make_unique<Check>(line);
  else if (keyword == to_string(Type::print))
    return make_unique<Print>(line);
  else if (string(line, 0, to_string(Type::print).size()) == to_string(Type::print))
    return make_unique<Print>(line);

  return {};
} // static unique_ptr<GameplayAction> GameplayAction::create(string line, istream& istr)

/** @return   clone of this gameplay action
 **/
unique_ptr<GameplayAction>
GameplayAction::clone() const
{
  switch (this->type()) {
  case Type::reservation:
    return make_unique<Reservation>(dynamic_cast<Reservation const&>(*this));
  case Type::marriage:
    return make_unique<Marriage>(dynamic_cast<Marriage const&>(*this));
  case Type::genscher:
    return make_unique<Genscher>(dynamic_cast<Genscher const&>(*this));
  case Type::poverty_shift:
    return make_unique<PovertyShift>(dynamic_cast<PovertyShift const&>(*this));
  case Type::poverty_accepted:
    return make_unique<PovertyAccepted>(dynamic_cast<PovertyAccepted const&>(*this));
  case Type::poverty_returned:
    return make_unique<PovertyReturned>(dynamic_cast<PovertyReturned const&>(*this));
  case Type::poverty_denied:
    return make_unique<PovertyDenied>(dynamic_cast<PovertyDenied const&>(*this));
  case Type::poverty_denied_by_all:
    return make_unique<PovertyDeniedByAll>(dynamic_cast<PovertyDeniedByAll const&>(*this));
  case Type::announcement:
    return make_unique<Announcement>(dynamic_cast<Announcement const&>(*this));
  case Type::swines:
    return make_unique<Swines>(dynamic_cast<Swines const&>(*this));
  case Type::hyperswines:
    return make_unique<Hyperswines>(dynamic_cast<Hyperswines const&>(*this));
  case Type::card_played:
    return make_unique<CardPlayed>(dynamic_cast<CardPlayed const&>(*this));
  case Type::trick_open:
    return make_unique<TrickOpen>(dynamic_cast<TrickOpen const&>(*this));
  case Type::trick_full:
    return make_unique<TrickFull>(dynamic_cast<TrickFull const&>(*this));
  case Type::trick_taken:
    return make_unique<TrickTaken>(dynamic_cast<TrickTaken const&>(*this));
  case Type::check:
    return make_unique<Check>(dynamic_cast<Check const&>(*this));
  case Type::print:
    return make_unique<Print>(dynamic_cast<Print const&>(*this));
  } // switch (this->type())

  return {};
} // unique_ptr<GameplayAction> GameplayAction::clone() const

/** -> result
 **
 ** @param    action   action to compare with
 **
 ** @return   whether the action is equal to this
 **/
bool
GameplayAction::equal(GameplayAction const& action) const
{
  return (this->type() == action.type());
} // bool GameplayAction::equal(GameplayAction action) const

} // namespace GameplayActions

namespace GameplayActions {
/** @param    gameplay action
 **
 ** @return   player number of the action (UINT_MAX, if the actions contains no player)
 **/
unsigned
  player(GameplayActions::GameplayAction const& action)
  {
    using namespace GameplayActions;
    switch (action.type()) {
    case Type::reservation:
      return dynamic_cast<GameplayActions::Reservation const&>(action).player;
    case Type::marriage:
      return dynamic_cast<GameplayActions::Marriage const&>(action).player;
    case Type::genscher:
      return dynamic_cast<GameplayActions::Genscher const&>(action).player;
    case Type::poverty_shift:
      return dynamic_cast<GameplayActions::PovertyShift const&>(action).player;
    case Type::poverty_accepted:
      return dynamic_cast<GameplayActions::PovertyAccepted const&>(action).player;
    case Type::poverty_returned:
      return dynamic_cast<GameplayActions::PovertyReturned const&>(action).player;
    case Type::poverty_denied:
      return dynamic_cast<GameplayActions::PovertyDenied const&>(action).player;
    case Type::announcement:
      return dynamic_cast<GameplayActions::Announcement const&>(action).player;
    case Type::swines:
      return dynamic_cast<GameplayActions::Swines const&>(action).player;
    case Type::hyperswines:
      return dynamic_cast<GameplayActions::Hyperswines const&>(action).player;
    case Type::card_played:
      return dynamic_cast<GameplayActions::CardPlayed const&>(action).player;
    case Type::poverty_denied_by_all:
    case Type::trick_open:
    case Type::trick_full:
    case Type::trick_taken:
    case Type::check:
    case Type::print:
      return UINT_MAX;
    } // switch (action.type())
    return UINT_MAX;
  } // unsigned player(GameplayAction action)
} // namespace GameplayActions

/** -> result
 **
 ** @param    type   action tye
 **
 ** @return   name of the action type
 **/
string
to_string(GameplayActions::Type const type)
{
  using namespace GameplayActions;
  switch(type) {
  case Type::reservation:
    return "reservation";
    (void)_("GameplayAction::reservation");
  case Type::marriage:
    return "marriage";
    (void)_("GameplayAction::marriage");
  case Type::genscher:
    return "genscher";
    (void)_("GameplayAction::genscher");
  case Type::poverty_shift:
    return "poverty shift";
    (void)_("GameplayAction::poverty shift");
  case Type::poverty_accepted:
    return "poverty accepted";
    (void)_("GameplayAction::poverty accepted");
  case Type::poverty_returned:
    return "poverty returned";
    (void)_("GameplayAction::poverty returned");
  case Type::poverty_denied:
    return "poverty denied";
    (void)_("GameplayAction::poverty denied");
  case Type::poverty_denied_by_all:
    return "poverty denied by all";
    (void)_("GameplayAction::poverty denied by all");
  case Type::announcement:
    return "announcement";
    (void)_("GameplayAction::announcement");
  case Type::swines:
    return "swines";
    (void)_("GameplayAction::swines");
  case Type::hyperswines:
    return "hyperswines";
    (void)_("GameplayAction::hyperswines");
  case Type::card_played:
    return "card played";
    (void)_("GameplayAction::card played");
  case Type::trick_open:
    return "trick open";
    (void)_("GameplayAction::trick open");
  case Type::trick_full:
    return "trick full";
    (void)_("GameplayAction::trick full");
  case Type::trick_taken:
    return "trick taken";
    (void)_("GameplayAction::trick taken");
  case Type::check:
    return "check";
    (void)_("GameplayAction::check");
  case Type::print:
    return "print";
    (void)_("GameplayAction::print");
  } // switch(type)

  DEBUG_ASSERTION(false,
                  "to_string(GameplayActions::Type type):\n"
                  "  unknown type '" << static_cast<int>(type) << "'");

  return "";
} // string to_string(GameplayActions::Type type)

/** -> result
 **
 ** @param    type    action type
 **
 ** @return   translation for the action type
 **/
string
gettext(GameplayActions::Type const type)
{
  return gettext("GameplayAction::" + to_string(type));
} // string gettext(GameplayActions::Type type)

/** writes the type in the output stream
 **
 ** @param    ostr   output stream
 ** @param    type   action type
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, GameplayActions::Type const type)
{
  return (ostr << to_string(type));
}

/** -> result
 **
 ** @param    discrepancy   discrepancy
 **
 ** @return   name of the discrepancy
 **/
string
to_string(GameplayActions::Discrepancy const discrepancy)
{
  using namespace GameplayActions;
  switch(discrepancy) {
  case Discrepancy::future:
    return "future";
    (void)_("Discrepancy::future");
  case Discrepancy::none:
    return "none";
    (void)_("Discrepancy::none");
  case Discrepancy::skipped:
    return "skipped";
    (void)_("Discrepancy::skipped");
  case Discrepancy::player:
    return "player";
    (void)_("Discrepancy::player");
  case Discrepancy::card:
    return "card";
    (void)_("Discrepancy::card");
  case Discrepancy::other:
    return "other";
    (void)_("Discrepancy::other");
  } // switch(discrepancy)

  DEBUG_ASSERTION(false,
                  "to_string(GameplayActions::Discrepancy discrepancy):\n"
                  "  unknown discrepancy "
                  << "'" << static_cast<int>(discrepancy) << "'");

  return "";
} // string to_string(GameplayActions::Discrepancy discrepancy)

/** writes the type in the output stream
 **
 ** @param    ostr          output stream
 ** @param    discrepancy   action discrepancy
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, GameplayActions::Discrepancy const discrepancy)
{
  return (ostr << to_string(discrepancy));
}

/** @param    discrepancy
 **
 ** @return   whether the discrepancy is not none
 **/
bool
operator!(GameplayActions::Discrepancy const discrepancy)
{ return (discrepancy == GameplayActions::Discrepancy::none); }

/** @param    lhs   left hand operand
 ** @param    rhs   left hand operand
 **
 ** @return   lhs || rhs
 **/
bool
operator||(GameplayActions::Discrepancy const lhs, bool const rhs)
{ return ((lhs != GameplayActions::Discrepancy::none) || rhs); }

/** writes the action in the output stream
 **
 ** @param    ostr   output stream
 ** @param    action   action
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, GameplayAction const& action)
{
  action.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, GameplayAction action)

/** @return   a string representing the action
 **/
string
to_string(GameplayAction const& action)
{
  ostringstream ostr;
  action.write(ostr);
  return ostr.str();
} // string to_string(GameplayAction action)

/** -> result
 **
 ** @param    lhs   first action
 ** @param    rhs   second action
 **
 ** @return   whether the actions are equal
 **/
bool
operator==(GameplayAction const& lhs, GameplayAction const& rhs)
{
  return rhs.equal(lhs);
} // bool operator==(GameplayAction lhs, GameplayAction rhs)

/** -> result
 **
 ** @param    lhs   first action
 ** @param    rhs   second action
 **
 ** @return   whether the actions are different
 **/
bool
operator!=(GameplayAction const& lhs, GameplayAction const& rhs)
{
  return !(lhs == rhs);
} // bool operator!=(GameplayAction lhs, GameplayAction rhs)
