/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "card_played.h"

#include "../../utils/string.h"

namespace GameplayActions {

/** constructor
 **
 ** @param    playerno   player who has played the card
 ** @param    card   played card
 **/
CardPlayed::CardPlayed(unsigned const playerno,
                       Card const card) :
  BasePlayer(playerno),
  card(card)
{ }

/** constructor
 **
 ** @param    playerno    player who has played the card
 ** @param    card        played card
 ** @param    heuristic   heuristic of the card
 **/
CardPlayed::CardPlayed(unsigned const playerno,
                       Card const card,
                       Aiconfig::Heuristic const heuristic) :
  BasePlayer(playerno),
  card(card),
  heuristic(heuristic)
{ }

/** constructor
 **
 ** @param    line   line with the data
 **
 ** @todo      read heuristic
 **/
CardPlayed::CardPlayed(string line) :
  BasePlayer()
{
#ifndef OUTDATED
  // old format in the bug report:
  //   Played: Player 2: club ace
  if (string(line, 0, strlen("Played: ")) == "Played: ") {
    // the first word is 'Played'
    String::word_first_remove_with_blanks(line);
    // the next word is 'player', remove it also
    String::word_first_remove_with_blanks(line);
    // next comes the player number, followed by a ':'
    this->player = static_cast<unsigned>(atoi(line.c_str()));
    // remove the player number
    String::word_first_remove_with_blanks(line);

    // the last part is the card
    this->card = Card(line);

    return ;
  }
#endif

  // ex: card played = 0, club ace
  auto const pre = to_string(this->type()) + " =";
  if (string(line, 0, pre.length()) != pre)
    return ;

  istringstream istr(string(line, pre.length()));
  istr >> this->player;
  if (istr.get() != ',')
    return ;
  istr >> this->card;


  // ToDo: read heuristic
  while (   istr.good()
         && ::isspace(istr.peek()))
    istr.get();

  if (!istr.good())
    return ;

  if (istr.peek() != '(')
    return ;
  istr.get();

  string heuristic_name;
  while (   istr.good()
         && (istr.peek() != ')'))
    heuristic_name.push_back(istr.get());

  if (istr.peek() != ')')
    return ;

  this->heuristic = Aiconfig::Heuristic_from_name(heuristic_name);
} // CardPlayed(string line)

/** -> result
 **
 ** @param    action   action to compare with
 **
 ** @return   whether the action is equal to this (do not compare the heuristic)
 **/
bool
CardPlayed::equal(GameplayAction const& action) const
{
  return (   (this->BasePlayer::equal(action))
          && (this->card == dynamic_cast<CardPlayed const&>(action).card)
          //&& (this->heuristic == dynamic_cast<CardPlayed const&>(action).heuristic)
          );
} // bool CardPlayed::equal(GameplayAction action) const

/** writes the action in the output stream
 **
 ** @param    ostr   output stream
 **/
void
CardPlayed::write(ostream& ostr) const
{
  ostr << this->type();
  ostr << " = " << this->player;
  ostr << ", " << this->card;

  if (this->heuristic != Aiconfig::Heuristic::no_heuristic)
    ostr << " (" << this->heuristic << ")";
} // void write(ostream& ostr) const

/** @return   a translation representing the data
 **/
string
CardPlayed::data_translation() const
{
  if (this->heuristic == Aiconfig::Heuristic::no_heuristic)
    return _(this->card);
  else
    return (_(this->card)
            + " (" + _(this->heuristic) + ")");
} // virtual string CardPlayed::data_translation() const

} // namespace GameplayActions
