/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "players.h"
#include "game.h"

#include "../party/party.h"
#include "../party/rule.h"
#include "../player/player.h"
#include "../player/human/human.h"
#include "../ui/ui.h"
#include "../misc/setting.h"
#include "../utils/string.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GamePlayers::GamePlayers(Game& game) :
  game_(&game)
{
  (void)game_;
  for (auto& p : game.party_->players()) {
    this->players_.push_back(&p);
  }
  for (auto& player : *this) {
    player.new_game(game);
  }
}

/** constructor
 **
 ** @param    game          corresponding game
 ** @param    old_players   players from the game to copy
 ** @param    new_players   players to use
 **/
GamePlayers::GamePlayers(Game& game,
                         GamePlayers const& old_players,
                         vector<Player*>& new_players) :
  game_(&game)
{
  DEBUG_ASSERTION(new_players.size() == game.rule()(Rule::Type::number_of_players_in_game),
                  "GamePlayers::GamePlayers(game, players)\n"
                  "  Number of players does not match: " << new_players.size() << " != " << game.rule()(Rule::Type::number_of_players_in_game) << '\n');
  this->players_ = new_players;
  for (auto& player : *this) {
    player.set_game(game);
  }
  this->current_player_ = &this->player(old_players.current_player().no());
  if (old_players.exists_soloplayer()) {
    this->soloplayer_ = &this->player(old_players.soloplayer().no());
  }
}

/** constructor
 **
 ** @param    game      corresponding game
 ** @param    old_players   players from the game to copy
 ** @param    new_players   players to use
 **/
GamePlayers::GamePlayers(Game& game,
                         GamePlayers const& old_players,
                         vector<unique_ptr<Player>>& new_players) :
  game_(&game)
{
  DEBUG_ASSERTION(new_players.size() == game.rule()(Rule::Type::number_of_players_in_game),
                  "GamePlayers::GamePlayers(game, players)\n"
                  "  Number of players does not match: " << new_players.size() << " != " << game.rule()(Rule::Type::number_of_players_in_game) << '\n');
  for (auto& p : new_players)
    this->players_.push_back(p.get());
  for (auto& player : *this) {
    player.set_game(game);
  }
  this->current_player_ = &this->player(old_players.current_player().no());
  if (old_players.exists_soloplayer()) {
    this->soloplayer_ = &this->player(old_players.soloplayer().no());
  }
}

/** self check
 **/
void
GamePlayers::self_check() const
{
  for (auto const& player : *this) {
    DEBUG_ASSERTION(&player.game() == this->game_,
                    "GamePlayers::self_check()\n"
                    "  a player is not of this game");
    DEBUG_ASSERTION(player.self_check(),
                    "GamePlayers::self_check()\n"
                    "  self check of a player returned 'false'");
  }

  if (this->current_player_)
    DEBUG_ASSERTION((this->game_ == &this->current_player_->game()),
                    "GamePlayers::self_check()\n"
                    "  the current player is not of the same game");

  if (this->soloplayer_)
    DEBUG_ASSERTION((this->game_ == &this->soloplayer_->game()),
                    "GamePlayers::self_check()\n"
                    "  the soloplayer is not of the same game");
}

/** reset the values
 **/
void
GamePlayers::reset() noexcept
{
  this->soloplayer_ = {};
}

/** @return   whether there is no player
 **/
bool
GamePlayers::empty() const noexcept
{
  return this->players_.empty();
}

/** @return   the number of players
 **/
unsigned
GamePlayers::size() const noexcept
{
  return this->players_.size();
}

/** @return   iterator to first player
 **/
GamePlayers::iterator
GamePlayers::begin() noexcept
{
  return GamePlayers::iterator(*this, 0);
}

/** @return   iterator behind the last player
 **/
GamePlayers::iterator
GamePlayers::end() noexcept
{
  return GamePlayers::iterator(*this, this->size());
}

/** @return   iterator to first player
 **/
GamePlayers::const_iterator
GamePlayers::begin() const noexcept
{
  return GamePlayers::const_iterator(*this, 0);
}

/** @return   iterator behind the last player
 **/
GamePlayers::const_iterator
GamePlayers::end() const noexcept
{
  return GamePlayers::const_iterator(*this, this->size());
}

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player&
GamePlayers::player(unsigned const p)
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "GamePlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
}

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player const&
GamePlayers::player(unsigned const p) const
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "GamePlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
}

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   the position of the player, UINT_MAX if not found
 **/
unsigned
GamePlayers::no(Player const& player) const noexcept
{
  for (size_t p = 0; p < this->size(); ++p)
    if (&this->player(p) == &player)
      return p;
  return UINT_MAX;
}

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   whether the player is in the game
 **/
bool
GamePlayers::contains(Player const& player) const noexcept
{
  return std::contains(this->players_, &player);
}

/** @return   current player
 **/
Player const&
GamePlayers::current_player() const
{
  DEBUG_ASSERTION(this->current_player_,
                  "GamePlayers::current_player:\n"
                  "  this->current_player_ == NULL");

  return *this->current_player_;
}

/** @return   current player
 **/
Player&
GamePlayers::current_player()
{
  DEBUG_ASSERTION(this->current_player_,
                  "GamePlayers::current_player:\n"
                  "  this->current_player_ == NULL");

  return *this->current_player_;
}

/** set the current player
 **/
void
GamePlayers::set_current_player(Player const& player)
{
  DEBUG_ASSERTION(&player.game() == this->game_,
                  "GamePlayers::set_current_player(player)\n"
                 "  Game does not match");
  this->current_player_ = this->players_[player.no()];
}

/** set the current player to the next one
 **/
void
GamePlayers::next_current_player() noexcept
{
  this->current_player_ = this->players_[(this->current_player_->no() + 1) % this->size()];
}

/** @return   whether a soloplayer exists
 **/
bool
GamePlayers::exists_soloplayer() const noexcept
{
  return this->soloplayer_;
}

/** -> result
 **
 ** @param    player   player to check
 **
 ** @return   whether 'player' is the soloplayer (or poverty / marriage player)
 **/
bool
GamePlayers::is_soloplayer(Player const& player) const noexcept
{
  return (   this->exists_soloplayer()
          && (this->soloplayer() == player));
}

/** @return   soloplayer
 **/
Player const&
GamePlayers::soloplayer() const
{
  DEBUG_ASSERTION(this->soloplayer_,
                  "GamePlayers::soloplayer:\n"
                  "  this->soloplayer_ == NULL");

  return *this->soloplayer_;
}

/** @return   soloplayer
 **/
Player&
GamePlayers::soloplayer()
{
  DEBUG_ASSERTION(this->soloplayer_,
                  "GamePlayers::soloplayer:\n"
                  "  this->soloplayer_ == NULL");

  return *this->soloplayer_;
}

/** set the soloplayer
 **/
void
GamePlayers::set_soloplayer(Player& player)
{
  DEBUG_ASSERTION(&player.game() == this->game_,
                  "GamePlayers::set_soloplayer(player)\n"
                  "  Game does not match");
  this->soloplayer_ = &player;
}

/** unset the soloplayer
 **/
void
GamePlayers::unset_soloplayer() noexcept
{
  this->soloplayer_ = {};
}

/** @return   the player who takes the poverty
 **/
Player const&
GamePlayers::poverty_partner() const
{
  DEBUG_ASSERTION((this->game_->type() == GameType::poverty),
                  "GamePlayers::poverty_partner()\n"
                  "  gametype is not poverty: " << this->game_->type());
  for (auto const& p : *this) {
    if (   (p.team() == Team::re)
        && (&p != this->soloplayer_ )) {
      return p;
    }
  }

  return this->player(0);
}

/** @return   number of humans in the game
 **/
unsigned
GamePlayers::count_humans() const noexcept
{
  return std::count_if(this->players_,
                       [](auto const p) {
                       return p->type() == Player::Type::human;
                       });
}

/** @return   human player if there exists exactly one,
 **   else 'NULL'
 **/
Human&
GamePlayers::human()
{
  DEBUG_ASSERTION(this->count_humans() == 1,
                  "GamePlayers::humans()\n"
                  "  Number of humans is not 1");

  auto const p = find_if(this->players_,
                         [](auto const p) {
                         return (p->type() == Player::Type::human);
                         });
  return dynamic_cast<Human&>(**p);
}

/** @return   human player if there exists exactly one,
 **   else 'NULL'
 **/
Human const&
GamePlayers::human() const
{
  DEBUG_ASSERTION(this->count_humans() == 1,
                  "GamePlayers::humans()\n"
                  "  Number of humans is not 1");

  auto const p = find_if(this->players_,
                         [](auto const p) {
                         return (p->type() == Player::Type::human);
                         });
  return dynamic_cast<Human const&>(**p);
}

/** -> result
 **
 ** @param    player   player, whose following player is returned
 **
 ** @return   the next player after 'player'
 **/
Player const&
GamePlayers::following(Player const& player) const
{
  DEBUG_ASSERTION(player == this->player(player.no()),
                  "GamePlayers::following(player):\n"
                  "  player not found in the game-players");

  return this->player((player.no() + 1)
                      % this->size());
}

/** -> result
 **
 ** @param    player   player, whose following player is returned
 **
 ** @return   the next player after 'player'
 **/
Player&
GamePlayers::following(Player& player)
{
  DEBUG_ASSERTION(player == this->player(player.no()),
                  "GamePlayers::following(player):\n"
                  "  player not found in the game-players");

  return this->player((player.no() + 1)
                      % this->size());
}

/** -> result
 **
 ** @param    player   player, whose previous player is returned
 **
 ** @return   the player before 'player'
 **/
Player const&
GamePlayers::previous(Player const& player) const
{
  DEBUG_ASSERTION(player == this->player(player.no()),
                  "GamePlayers::previous(player):\n"
                  "  player not found in the game-players");

  return this->player((player.no() + this->size() - 1)
                      % this->size());
}

/** sort the hands of the players
 **/
void
GamePlayers::sort_hands()
{
  if (!::setting(Setting::Type::cards_order).sorted())
    return;

  for (auto& player : *this) {
    player.hand_sort();
  }
}

/** mix the hands of the players
 **/
void
GamePlayers::mix_hands()
{
  if (::setting(Setting::Type::cards_order).sorted())
    return;

  for (auto& player : *this) {
    player.hand_mix();
  }
}

/** write the game into the output stream
 ** The current game is not written.
 **
 ** @param    ostr   output stream
 **
 ** @todo   output of the game?
 **/
void
GamePlayers::write(ostream& ostr) const
{
  ostr << "players\n"
    << "{\n";
  for (auto const& player : this->players_)
    ostr << "{\n" << *player << "}\n\n";
  ostr << "}\n";
}

/** write the players in short form (name, type) in the output stream
 **
 ** @param    ostr   output stream
 **/
void
GamePlayers::write_short(ostream& ostr) const
{
  ostr << "game players:\n";
  for (unsigned p = 0; p < this->players_.size(); ++p) {
    ostr << p << ": ";
    this->player(p).write_short(ostr);
    ostr << '\n';
  }
}
