/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "game.h"
#include "gameplay.h"
#include "gameplay_actions/poverty.h"
#include "../os/bug_report_replay.h"
#include "../misc/setting.h"
#include "../party/rule.h"
#include "../ui/ui.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GamePoverty::GamePoverty(Game& game) noexcept :
  game_(&game)
{ }

/** constructor
 **
 ** @param    poverty   poverty to copy
 ** @param    game   corresponding game
 **/
GamePoverty::GamePoverty(GamePoverty const& poverty, Game& game) :
  game_(&game),
  shifted_cards_(poverty.shifted_cards_),
  returned_cards_(poverty.returned_cards_),
  shifted_(poverty.shifted_)
{ }

/** @return   whether cards were shifted
 **/
bool
GamePoverty::shifted() const noexcept
{
  return this->shifted_;
}

/** @return   the number of cards, that are shifted by the poverty player
 **/
unsigned
GamePoverty::shifted_cardno() const noexcept
{
  return this->shifted_cards().size();
}

/** @return   the number of cards, that are returned to the poverty player
 **/
unsigned
GamePoverty::returned_cardno() const noexcept
{
  return this->returned_cards().size();
}

/** @return   the number of trumps, that are returned to the poverty player
 **/
unsigned
GamePoverty::returned_trumpno() const noexcept
{
  return this->returned_cards().count_trumps();
}

/** @return   the shifted poverty cards
 **/
HandCards const&
GamePoverty::shifted_cards() const
{
  DEBUG_ASSERTION((this->game_->type() == GameType::poverty),
                  "GamePoverty::shifted_cards():\n"
                  "  invalid gametype");

  return this->shifted_cards_;
}

/** @return   the returned poverty cards
 **/
HandCards const&
GamePoverty::returned_cards() const
{
  DEBUG_ASSERTION((this->game_->type() == GameType::poverty),
                  "GamePoverty::returned_cards():\n"
                  "  invalid gametype '" << this->game_->type() << "'");

  return this->returned_cards_;
}

/** reset all data
 **/
void
GamePoverty::reset() noexcept
{
  this->shifted_cards_.clear();
  this->returned_cards_.clear();
  this->shifted_ = false;
}

/** make the shifting of the poverty cards
 **/
void
GamePoverty::shift()
{
  auto& game = *this->game_;
  auto& players = game.players();
  auto const& rule = game.rule();
  auto& soloplayer = players.soloplayer();
  auto& gameplay = *game.gameplay_;

  if (   (game.type() != GameType::poverty)
      || !rule(Rule::Type::poverty_shift)
      || (   (soloplayer.hand().count_trumps() <= 1)
          && rule(Rule::Type::throw_with_one_trump)
         )
     ) {
    players.set_current_player(soloplayer);
    ::game_status = GameStatus::game_redistribute;
    return ;
  } // if (not poverty shift)

  ::game_status = GameStatus::game_poverty_shift;

  game.teaminfo().set(soloplayer, Team::re);
  soloplayer.set_team(Team::re);

  game.swines().test_swines_from_reservations();
  // resort for not announced swines/hyperswines
  players.sort_hands();

  if (!game.isvirtual()) {
    ::ui->poverty_shifting(soloplayer);
    this->signal_shifting_(soloplayer);
  }

  this->shifted_cards_ = soloplayer.poverty_shift();

  DEBUG_ASSERTION((this->shifted_cards().player()
                   == soloplayer),
                  "Game::poverty_shift():\n"
                  "  the soloplayer '" << soloplayer.no()
                  << "' shifts cards from player '"
                  << this->shifted_cards().player().no() << "'");

  DEBUG_ASSERTION((soloplayer.hand().count_trumps()
                   == 0),
                  "Game::poverty_shift():\n"
                  "  Poverty: the soloplayer still has trump on the hand.\n"
                  "remaining trumps: "
                  << soloplayer.hand().count_trumps() << '\n'
                  << "Hand:\n"
                  << soloplayer.hand()
                  << "shifted cards:\n"
                  << this->shifted_cards());

  DEBUG_ASSERTION( (this->shifted_cardno()
                    + soloplayer.hand().cardsnumber()
                    == rule(Rule::Type::number_of_tricks_in_game)),
                  "Game::poverty_shift():\n"
                  "  wrong number of cards shifted: "
                  << this->shifted_cardno());

  if (rule(Rule::Type::poverty_shift_only_trump))
    DEBUG_ASSERTION(this->shifted_cards().count_trumps()
                      == this->shifted_cardno(),
                    "Game::poverty_shift():\n"
                    "  wrong number of cards shifted: "
                    << "trumps = " << this->shifted_cards().count_trumps()
                    << " != " << this->shifted_cardno() << " = cards shifted");
  else if (!rule(Rule::Type::poverty_fox_do_not_count)
           || game.swines().swines_announced() )
    // normal case: shift at max 3 cards
    DEBUG_ASSERTION(this->shifted_cardno()
                      == rule(Rule::Type::max_number_of_poverty_trumps),
                    "Game::poverty_shift():\n"
                    "  wrong number of cards shifted: "
                    << "poverty cards = " << rule(Rule::Type::max_number_of_poverty_trumps)
                    << " != " << this->shifted_cardno() << " = cards shifted");
  else if (!rule(Rule::Type::poverty_fox_shift_extra))
    // at max 3 cards or number of trumps
    DEBUG_ASSERTION( (   (this->shifted_cards().count_trumps()
                          == this->shifted_cardno())
                      || (this->shifted_cardno()
                          == rule(Rule::Type::max_number_of_poverty_trumps)) ),
                    "Game::poverty_shift():\n"
                    "  wrong number of cards shifted: "
                    << "trumps or "
                    << "poverty cards = " << rule(Rule::Type::max_number_of_poverty_trumps)
                    << " != " << this->shifted_cardno() << " = cards shifted");
  else
    // fox are shifted extra
    DEBUG_ASSERTION( (this->shifted_cardno()
                      - this->shifted_cards().count_trump_aces()
                      == rule(Rule::Type::max_number_of_poverty_trumps)),
                    "Game::poverty_shift():\n"
                    "  wrong number of cards shifted: "
                    << "poverty cards = " << rule(Rule::Type::max_number_of_poverty_trumps)
                    << " != " << this->shifted_cardno() << " - "
                    << this->shifted_cards().count_trump_aces()
                    << " = cards shifted - fox");

  for (auto& p : players)
    p.poverty_shift(soloplayer, this->shifted_cardno());

  gameplay.gameplay_action(GameplayActions::PovertyShift(soloplayer.no(),
                                                                this->shifted_cards()));
  if (!game.isvirtual()) {
    ::ui->poverty_shift(soloplayer, this->shifted_cardno());
    this->signal_shift_(soloplayer, this->shifted_cardno());
    ::ui->gameplay_action(GameplayActions::PovertyShift(soloplayer.no(),
                                                        this->shifted_cards()));
    game.signal_gameplay_action_(GameplayActions::PovertyShift(soloplayer.no(),
                                                                this->shifted_cards()));
  }

  // ask all players, whether they accept the poverty or not
  bool accept = false;
  players.set_current_player(soloplayer);
  for (players.next_current_player();
       players.current_player() != soloplayer;
       players.next_current_player()) {
    auto& current_player = players.current_player();
    if (!game.isvirtual()) {
      ::ui->poverty_ask(current_player, this->shifted_cardno());
      this->signal_ask_(current_player, this->shifted_cardno());
      if (   !(::fast_play & FastPlay::pause)
          && !(   ::bug_report_replay
               && ::bug_report_replay->auto_action())) {
        if (current_player.type() != Player::Type::human)
          ::ui->sleep(::setting(Setting::Type::card_play_delay));
      }
      if (::game_status != GameStatus::game_poverty_shift)
        return ;
    } // if (!this->isvirtual())


    accept = current_player.poverty_take_accept(this->shifted_cardno());
    if (::game_status != GameStatus::game_poverty_shift)
      return ;

    if (accept) {
      { // test for swines and hyperswines before taking the cards
        if (rule(Rule::Type::swines)
            && rule(Rule::Type::swines_announcement_begin)
            && game.reservation(current_player).swines)
          game.swines().swines_announce(current_player);
        if (rule(Rule::Type::hyperswines)
            && rule(Rule::Type::hyperswines_announcement_begin)
            && game.reservation(current_player).hyperswines)
          game.swines().hyperswines_announce(current_player);
      } // test for swines and hyperswines before taking the cards
      gameplay.gameplay_action(GameplayActions::PovertyAccepted(current_player.no()));
      ::ui->gameplay_action(GameplayActions::PovertyAccepted(current_player.no()));
      game.signal_gameplay_action_(GameplayActions::PovertyAccepted(current_player.no()));
      break;
    } else { // if !(accept)
      for (auto& p : players) {
        p.poverty_take_denied(current_player);
      }
      gameplay.gameplay_action(GameplayActions::PovertyDenied(current_player.no()));
      if (!game.isvirtual()) {
        ::ui->poverty_take_denied(current_player);
        this->signal_take_denied_(current_player);
        ::ui->gameplay_action(GameplayActions::PovertyDenied(current_player.no()));
        game.signal_gameplay_action_(GameplayActions::PovertyDenied(current_player.no()));
      }
    } // if !(accept)
  } // for (current_player)

  if (!accept) {
    players.set_current_player(soloplayer);

    for (auto& p : players)
      p.poverty_take_denied_by_all();
    gameplay.gameplay_action(GameplayActions::PovertyDeniedByAll());
    if (!game.isvirtual()) {
      ::ui->poverty_take_denied_by_all();
      this->signal_take_denied_by_all_();
      ::ui->gameplay_action(GameplayActions::PovertyDeniedByAll());
      game.signal_gameplay_action_(GameplayActions::PovertyDeniedByAll());
    }

    soloplayer.sorted_hand().add(this->shifted_cards());
    soloplayer.hand_sort();

    ::game_status = GameStatus::game_redistribute;
  } else { // if !(!accept)
    auto& current_player = players.current_player();
    // set the teaminfo
    auto& teaminfo = game.teaminfo();
    teaminfo.set(current_player, Team::re);
    current_player.set_team(Team::re);

    for (auto& p : players) {
      if (teaminfo.get(p) == Team::unknown) {
        teaminfo.set(p, Team::contra);
        p.set_team(Team::contra);
      }
    } // for (p)
    teaminfo.set_at_gamestart();

    this->returned_cards_
      = current_player.poverty_cards_change(this->shifted_cards());
    if (::game_status != GameStatus::game_poverty_shift)
      return ;

    // check for shifted swines
    if (rule(Rule::Type::swines)
        && game.swines().swines_announced()
        && (game.swines().swines_owner() == soloplayer)) {
      game.swines().swines_announce(current_player);
    }
    if (rule(Rule::Type::hyperswines)
        && game.swines().hyperswines_announced()
        && (game.swines().hyperswines_owner() == soloplayer)) {
      if (!rule(Rule::Type::swines_and_hyperswines_joint)) {
        game.swines().hyperswines_owner_ = nullptr;
        game.swines().hyperswines_announced_ = false;
      }
      game.swines().hyperswines_announce(current_player);
    }

    DEBUG_ASSERTION((this->returned_cards().player()
                     == current_player),
                    "Game::poverty_shift():\n"
                    "  the current player '" << current_player.no()
                    << "' shifts back cards from player '"
                    << this->returned_cards().player().no() << "'");

    DEBUG_ASSERTION((this->returned_cards().size()
                     == this->shifted_cardno()),
                    "Game::poverty_shift():\n"
                    "  number of cards returned ("
                    << this->returned_cards().size()
                    << ") is not the same as are shifted ("
                    << this->shifted_cardno() << ").");

    for (auto& p : players) {
      p.poverty_take_accepted(current_player,
                              this->returned_cardno(),
                              this->returned_trumpno());
    }
    gameplay.gameplay_action(GameplayActions::PovertyReturned(current_player.no(),
                                                              this->returned_cards()));
    if (!game.isvirtual()) {
      ::ui->poverty_take_accepted(current_player,
                                  this->returned_cardno(),
                                  this->returned_trumpno());
      this->signal_take_accepted_(current_player,
                                  this->returned_cardno(),
                                  this->returned_trumpno());
      ::ui->gameplay_action(GameplayActions::PovertyReturned(current_player.no(),
                                                             this->returned_cards()));
      game.signal_gameplay_action_(GameplayActions::PovertyReturned(current_player.no(),
                                                                    this->returned_cards()));
    }


    // add the cards to the player
    players.set_current_player(soloplayer);
    soloplayer.poverty_cards_get_back(this->returned_cards());
    if (!game.isvirtual()) {
      ::ui->gametype_changed();
      game.signal_gametype_changed_();
    }

    if (::game_status != GameStatus::game_poverty_shift)
      return ;

  } // if !(!accept)

  this->shifted_ = true;

  if (accept) {
    if (game.swines().swines_announced()) {
      // check, that the swines announcement is correct
      auto const& swines_owner = game.swines().swines_owner();
      if (swines_owner.team() == Team::re) {
        auto p = std::find_if(players,
                              [this](auto player) {
                              return true;
                              });

        if (p != players.end()) {
          game.swines().swines_announce(*p);
        } else {
          game.swines().swines_owner_ = nullptr;
          game.swines().swines_announced_ = false;
        }
      } // if (this->swines_owner().team() == Team::re)
    } // if (this->swines_owner())
  } // if (accept)

  if (::game_status == GameStatus::game_poverty_shift)
    ::game_status = GameStatus::game_reservation;
} // void GamePoverty::shift()


/** signals
 **/

Signal<void(Player const&)>&
GamePoverty::signal_shifting() noexcept
{ return this->signal_shifting_; }

Signal<void(Player const&, unsigned)>&
GamePoverty::signal_shift() noexcept
{ return this->signal_shift_; }

Signal<void(Player const&, unsigned)>&
GamePoverty::signal_ask() noexcept
{ return this->signal_ask_; }

Signal<void(Player const&)>&
GamePoverty::signal_take_denied() noexcept
{ return this->signal_take_denied_; }

Signal<void()>&
GamePoverty::signal_take_denied_by_all() noexcept
{ return this->signal_take_denied_by_all_; }

Signal<void(Player const&, unsigned, unsigned)>&
GamePoverty::signal_take_accepted() noexcept
{ return this->signal_take_accepted_; }

Signal<void(Player const&)>&
GamePoverty::signal_get_cards_to_shift() noexcept
{ return this->signal_get_cards_to_shift_; }

Signal<void(Player const&, unsigned)>&
GamePoverty::signal_get_take_accept() noexcept
{ return this->signal_get_take_accept_; }

Signal<void(Player const&, HandCards)>&
GamePoverty::signal_get_exchanged_cards() noexcept
{ return this->signal_get_exchanged_cards_; }

Signal<void(Player const&, HandCards)>&
GamePoverty::signal_cards_get_back() noexcept
{ return this->signal_cards_get_back_; }
