/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "tricks.h"
#include "game.h"
#include "../party/rule.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GameTricks::GameTricks(Game& game) noexcept :
  game_(&game)
{ }

/** constructor
 **
 ** @param    tricks   tricks to copy
 ** @param    game     corresponding game
 **/
GameTricks::GameTricks(GameTricks const& tricks, Game& game) :
  game_(&game),
  tricks_(tricks.tricks_)
{
  for (auto& t : this->tricks_)
    t.set_game(game);
}

/** self check
 **/
void
GameTricks::self_check() const
{
  for (auto const& trick : *this) {
    DEBUG_ASSERTION(&trick.game() == this->game_,
		    "GameTricks::self_check()\n"
		    "  a trick is not of this game");

    DEBUG_ASSERTION(trick.self_check(),
		    "GameTricks::self_check()\n"
		    "  self check of a trick returned 'false'");
  }
}

/** @return   whether there is no trick
 **/
bool
GameTricks::empty() const noexcept
{
  return this->tricks_.empty();
}

/** @return   the number of tricks
 **/
unsigned
GameTricks::size() const noexcept
{
  return this->tricks_.size();
}

/** @return   the number of tricks
 **/
unsigned
GameTricks::max_size() const
{
  return this->game_->rule()(Rule::Type::number_of_tricks_in_game);
}

/** @return   iterator to first trick
 **/
GameTricks::iterator
GameTricks::begin() noexcept
{
  return GameTricks::iterator(*this, 0);
}

/** @return   iterator behind the last trick
 **/
GameTricks::iterator
GameTricks::end() noexcept
{
  return GameTricks::iterator(*this, this->size());
}

/** @return   iterator to first trick
 **/
GameTricks::const_iterator
GameTricks::begin() const noexcept
{
  return GameTricks::const_iterator(*this, 0);
}

/** @return   iterator behind the last trick
 **/
GameTricks::const_iterator
GameTricks::end() const noexcept
{
  return GameTricks::const_iterator(*this, this->size());
}

/** -> result
 **
 ** @param    t   number of the trick
 **
 ** @return   the trick with the number 't'
 **/
Trick&
GameTricks::trick(unsigned const t)
{
  DEBUG_ASSERTION(t < this->tricks_.size(),
                  "GameTricks::trick(t):\n"
                  "'t' is not valid (" << t << ")");

  return this->tricks_[t];
}

/** -> result
 **
 ** @param    t   number of the trick
 **
 ** @return   the trick with the number 't'
 **/
Trick const&
GameTricks::trick(unsigned const t) const
{
  DEBUG_ASSERTION(t < this->tricks_.size(),
                  "GameTricks::trick(t):\n"
                  "'t' is not valid (" << t << ")");

  return this->tricks_[t];
}

/** add a trick
 **
 ** @param   trick to add
 **/
void
GameTricks::add(Trick trick)
{
  this->tricks_.push_back(trick);
  this->tricks_.back().set_game(*this->game_);
}

/** @return  the current trick
 **/
Trick const&
GameTricks::current() const
{ return this->tricks_.back(); }

/** @return  the current trick
 **/
Trick&
GameTricks::current()
{ return this->tricks_.back(); }

/** -> result
 ** a full trick in trickpile is one trick further
 **
 ** @return   the number of tricks in the trickpiles
 **/
unsigned
GameTricks::tricks_in_trickpiles() const
{
  if (this->tricks_.empty())
    if (::game_status < GameStatus::game_play)
      return UINT_MAX;
    else
      return 0;

  if (   (::game_status == GameStatus::game_finished)
      || (::game_status == GameStatus::game_trick_taken))
    return this->tricks_.size();

  return (this->tricks_.size() - 1);
}

/** -> result
 ** a full trick is one trick further
 **
 ** @return   the number of the current trick
 **
 ** @todo   replace with 'real_trick_current_no'
 **/
unsigned
GameTricks::current_no() const
{
  if (this->tricks_.empty())
    if (::game_status < GameStatus::game_play)
      return UINT_MAX;
    else
      return 0;

  if (   (::game_status == GameStatus::game_finished)
      || (this->current().isfull()))
    return this->tricks_.size();

  return (this->tricks_.size() - 1);
}

/** @return   the number of the current trick
 **/
unsigned
GameTricks::real_current_no() const noexcept
{
  if (this->tricks_.empty())
    if (::game_status < GameStatus::game_play)
      return UINT_MAX;
    else
      return 0;

  if (::game_status == GameStatus::game_finished)
    return this->tricks_.size();

  return (this->tricks_.size() - 1);
}

/** @return   the number of remaining tricks
 **/
unsigned
GameTricks::remaining_no() const
{
  if (   (::game_status == GameStatus::game_reservation)
      || (::game_status == GameStatus::game_poverty_shift) )
    return this->game_->rule()(Rule::Type::number_of_tricks_in_game);

  DEBUG_ASSERTION((this->current_no() != UINT_MAX),
		  "Game::tricks().remaining_no():\n"
		  "  this->tricks().current_no() == -1");

  return (this->game_->rule()(Rule::Type::number_of_tricks_in_game)
          - this->current_no());
}

/** @return   the number of remaining tricks
 **/
unsigned
GameTricks::real_remaining_no() const
{
  DEBUG_ASSERTION((this->real_current_no() != UINT_MAX),
                  "Game::tricks().real_remaining_no():\n"
                  "  this->real_tricks().current_no() == -1");

  return (this->game_->rule()(Rule::Type::number_of_tricks_in_game)
          - this->real_current_no());
}

/** @return   whether this is the last trick
 **/
bool
GameTricks::is_last_trick() const
{
  if (   (::game_status == GameStatus::game_reservation)
      || (::game_status == GameStatus::game_poverty_shift) )
    return false;
  return (this->remaining_no() == 1);
}

/** @return   number of trump tricks (with the current trick)
 **/
unsigned
GameTricks::trump_tricks_no() const
{
  return count_if(this->tricks_,
                  [](Trick const& trick) {
                  return (   !trick.isempty()
                          && trick.startcard().istrump());
                  });
}

/** -> result
 **
 ** @param    team    Team
 **
 ** @return   whether the team has got a trick
 **/
bool
GameTricks::has_trick(Team const team) const
{
  return contains_if(this->tricks_,
                     [team](auto const& trick) {
                     return (   (trick.winnerteam() == team)
                             && trick.intrickpile());
                     });
}

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   whether the player has got a trick
 **/
bool
GameTricks::has_trick(Player const& player) const
{
  return contains_if(this->tricks_,
                     [&player](auto const& trick) {
                     return (   (trick.winnerplayer() == player)
                             && trick.intrickpile());
                     });
}

/** -> result
 **
 ** @param    player  player
 **
 ** @return   number of tricks, the player has won so far
 **              (with the current trick)
 **
 */
unsigned
GameTricks::number_of_tricks(Player const& player) const
{
  return std::count_if(this->tricks_, 
                       [&player](auto const& trick) {
                       return (   (trick.winnerplayer() == player)
                               && trick.intrickpile());
                       });
}

/** -> result
 **
 ** @param    player  player
 ** @param    trickno last trick to view
 **
 ** @return   number of tricks, the player has won til the trick 'trickno'
 **
 */
unsigned
GameTricks::number_of_tricks_till_trick(Player const& player,
                                        unsigned trickno) const
{
  unsigned number = 0;

  if (trickno != UINT_MAX)
    trickno += 1;
  for (auto trick = this->tricks_.begin();
       (   (trickno > 0)
        && (trick != this->tricks_.end()));
       ++trick, --trickno) {
    if (   (trick->winnerplayer() == player)
        && trick->intrickpile()) {
      number += 1;
    }
  }

  return number;
}

/** -> result
 **
 ** @param    team    team
 **
 ** @return   number of tricks, the team has won so far
 **              (without the current trick)
 **/
unsigned
GameTricks::number_of_tricks(Team const team) const
{
  return count_if(this->tricks_,
                  [team](auto const& trick) {
                  return (   (trick.winnerteam() == team)
                          && trick.intrickpile());
                  });
}

/** write the game into the output stream
 ** The current game is not written.
 **
 ** @param    ostr   output stream
 **
 ** @todo   output of the game?
 **/
void
GameTricks::write(ostream& ostr) const
{
  ostr << '\n'
    << "tricks:\n"
    << '\n';
  unsigned i = 0;
  for (auto const& trick : *this) {
    ostr << i << ". trick:\n"
      << trick << '\n';
    i += 1;
  }

  ostr << '\n';
}

/** @return   signal, when a trick is opened
 **/
Signal<void(Trick const&)>&
GameTricks::signal_trick_open() noexcept
{ return this->signal_trick_open_; }

/** @return   signal, when a trick is full
 **/
Signal<void(Trick const&)>&
GameTricks::signal_trick_full() noexcept
{ return this->signal_trick_full_; }

/** @return   signal, when a trick is moved into the trickpile
 **/
Signal<void(Trick const&)>&
GameTricks::signal_trick_move_in_pile() noexcept
{ return this->signal_trick_move_in_pile_; }

/** @return   signal, when a trick is closed
 **/
Signal<void(Trick const&)>&
GameTricks::signal_trick_close() noexcept
{ return this->signal_trick_close_; }
