/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 * 
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "setting.h"
#include "setting.cardsorder.h"
#include "setting.theme.h"

#ifdef LINUX
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#endif
#include <sys/stat.h>

#include "../party/party.h"
#include "../game/game.h"
#include "../player/human/human.h"

#include "../ui/ui.h"

#include "../utils/file.h"
#include "../utils/string.h"
#include "../utils/windows.h"
#include "../class/readconfig/readconfig.h"


/** constructor
 **/
Setting::Setting() :
  cards_order_(make_unique<CardsOrder>(*this))
{ }

/** copy constructor
 **
 ** @param    setting   setting to copy
 **/
Setting::Setting(Setting const& setting):
  bool_(setting.bool_),
  unsigned_(setting.unsigned_),
  string_(setting.string_),
  cardset_license_(setting.cardset_license_),
  cards_order_(make_unique<CardsOrder>(*setting.cards_order_)),
  data_directories_(setting.data_directories_),
  string_path_(setting.string_path_),
  themes_()
{
  for (auto const& t : setting.themes_)
    this->themes_.push_back(make_unique<Theme>(*t));

  this->update_all();
} // Setting::Setting(Setting setting)

/** copy operator
 **
 ** @param    setting   setting to copy
 **
 ** @return   this
 **/
Setting&
Setting::operator=(Setting const& setting)
{
  this->cardset_license_ = setting.cardset_license_;
  this->data_directories_ = setting.data_directories_;
  this->string_path_ = setting.string_path_;
  for (auto t : Type::bool_list)
    this->set(t, setting.value(t));

  for (auto t : Type::unsigned_list)
    this->set(t, setting.value(t));

  for (auto t : Type::string_list)
    this->set(t, setting.value(t));

  this->set(Type::cards_order,
            setting.value(Type::cards_order));

  this->themes_.clear();
  for (auto const& t : setting.themes_)
    this->themes_.push_back(make_unique<Theme>(*t));

  return *this;
} // bool Setting::operator=(Setting setting) const

/** initializes the settings
 **/
void
Setting::init()
{
  this->set_data_directories();
  this->set_to_hardcoded();
  this->create_themes_list();
}

/** @return   signal: a rule is changed
 **/
Signal<void(int, void const*)>& 
Setting::signal_changed()
{ return this->signal_changed_; }

/** set the values to the hardcoded
 **/
void
Setting::set_to_hardcoded()
{
#if 0
  this->set(Type::public_data_directory,    "");
  this->set(Type::private_data_directory,   "");
  this->set(Type::settings_file,            "FreeDokorc");
  this->set(Type::logo_file,                "logo.png");
  this->set(Type::icon_file,                "icon.png");
  this->set(Type::graphic_extension,        "png");
  this->set(Type::sound_extension,          "wav");
  this->set(Type::cardsets_directory,       "cardsets");
  this->set(Type::cards_directory,          "cards");
  this->set(Type::cards_back_directory,     "back");
  this->set(Type::icons_directory,          "icons");
  this->set(Type::backgrounds_directory,    "backgrounds");
  this->set(Type::sounds_directory,         "sound");
  this->set(Type::ai_directory,             "ai");
  this->set(Type::parties_directory,        "rules");
  this->set(Type::rules_directory,          "rules");
  this->set(Type::settings_directory,       "settings");
  this->set(Type::manual_directory,         "/usr/share/doc/manual");
#endif

  this->bool_[Type::automatic_savings]          = true;
  this->bool_[Type::save_party_changes]         = true;
  this->bool_[Type::additional_party_settings]  = false;

  this->string_[Type::language]                 = "";

  this->string_[Type::name]                     = "";

  this->bool_[Type::show_splash_screen]         = true;
  this->bool_[Type::splash_screen_transparent]  = false;

  this->bool_[Type::show_bug_report_button_in_game_finished_window] = true;
  this->bool_[Type::save_bug_reports_on_desktop] = true;

  this->bool_[Type::sound]                      = true;

  this->bool_[Type::automatic_card_suggestion]  = false;
  this->bool_[Type::announce_swines_automatically] = true;
  this->bool_[Type::show_if_valid]              = true;
  this->bool_[Type::emphasize_valid_cards]      = false;
  this->bool_[Type::announce_in_table]          = true;
  this->bool_[Type::show_all_hands]             = false;
  this->bool_[Type::show_ai_information_hands]  = false;
  this->bool_[Type::show_ai_information_teams]  = false;
  this->bool_[Type::show_trickpiles_points]     = true;
  this->bool_[Type::show_known_teams_in_game]   = true;
  this->bool_[Type::show_soloplayer_in_game]    = true;

  this->bool_[Type::show_full_trick_window]     = true;
  this->bool_[Type::show_full_trick_window_if_special_points]= true;
  this->bool_[Type::close_full_trick_automatically] = false;

  this->bool_[Type::show_gametype_window]                = true;
  this->bool_[Type::close_gametype_window_automatically] = false;
  this->bool_[Type::show_marriage_window]                = true;
  this->bool_[Type::close_marriage_window_automatically] = false;
  this->bool_[Type::show_genscher_window]                = true;
  this->bool_[Type::close_genscher_window_automatically] = false;
  this->bool_[Type::show_announcement_window]            = true;
  this->bool_[Type::close_announcement_window_automatically] = false;
  this->bool_[Type::show_swines_window]                  = true;
  this->bool_[Type::close_swines_window_automatically]   = false;

  this->bool_[Type::rotate_trick_cards]        = false;
  this->set(Type::cards_order, "trump up, club down, heart down, spade down, diamond down");


#ifdef USE_THREADS
  this->unsigned_[Type::threads_max]                     = 1;
#endif
  this->unsigned_[Type::card_play_delay]                 = 500;
  this->unsigned_[Type::full_trick_close_delay]          = 2000;

  this->unsigned_[Type::gametype_window_close_delay]     = 2000;
  this->unsigned_[Type::marriage_window_close_delay]     = 20;
  this->unsigned_[Type::genscher_window_close_delay]     = 2000;
  this->unsigned_[Type::announcement_window_close_delay] = 2000;
  this->unsigned_[Type::swines_window_close_delay]       = 2000;

  this->bool_[Type::original_cards_size]                 = true;
  this->unsigned_[Type::cards_height]                    = 140;

  this->bool_[Type::own_hand_on_table_bottom]            = true;
  this->unsigned_[Type::table_rotation]                  = 0;

  this->bool_[Type::ui_main_window_fullscreen]           = false;
  this->unsigned_[Type::ui_main_window_pos_x]            = 0;
  this->unsigned_[Type::ui_main_window_pos_y]            = 0;
  this->unsigned_[Type::ui_main_window_width]            = 0;
  this->unsigned_[Type::ui_main_window_height]           = 0;

  //this->string_[Type::cardset]        = "xskat/french";
  //this->string_[Type::cards_back]            = "xskat";
  { // set the cardset
    this->string_[Type::cardset] = "";
    this->string_path_[Type::cardset] = "";
    this->string_[Type::cards_back] = "";
    this->string_path_[Type::cards_back] = "";

    vector<string> cardsets;

    if ((*this)(Type::language) == "de") {
      if (Date() <= Setting::innocard_expiration_date)
        cardsets.push_back("InnoCard/french/185");
      cardsets.push_back("xskat/french");
      cardsets.push_back("xskat/german");
    } else if ((*this)(Type::language) == "de-alt") {
      cardsets.push_back("xskat/german");
      cardsets.push_back("freedoko-pingucards");
    } else { // english and so
      if (Date() <= Setting::innocard_expiration_date)
        cardsets.push_back("InnoCard/english/185");
      cardsets.push_back("dondorf");
    } // if (language == )
    cardsets.push_back("xskat/french");
    cardsets.push_back("dondorf");
    cardsets.push_back("kdecarddecks/spaced");
    cardsets.push_back("pysol/gpl");
    cardsets.push_back("pysol/a-m/gpl");
    cardsets.push_back("openclipart");

    auto cardset = cardsets.begin();
    for (; cardset != cardsets.end(); ++cardset) {
      this->set(Type::cardset, *cardset);
      if (!this->path(Type::cardset).empty())
        break;
    } // for (cardset \in cardsets)

    if (cardset == cardsets.end()) {
      this->string_[Type::cardset] = "";
      if (::ui)
        ::ui->error(_("Error::loading default cardset %s",
                      cardsets.front()
                     ));
    } else {
      this->string_[Type::cardset] = *cardset;
    }
  } // set the cardset
  { // set the cards back
    vector<string> backs = {"penguin", "back"};

    auto back = backs.begin();
    for (; back != backs.end(); ++back) {
      this->string_[Type::cards_back] = *back;
      if (!this->path(Type::cards_back).empty())
        break;
    } // for (back \in backs)

    if (back == backs.end()) {
      this->string_[Type::cards_back] = "";
      if (::ui)
        // gettext: %s = back, %s = cardset
        ::ui->error(_("Error::loading default cards back %s (%s)",
                      backs.front(),
                      this->value(Type::cardset)
                     ));
    } else {
      this->string_[Type::cards_back] = *back;
    }
  } // set the cards back
  { // set the icons
    this->string_[Type::iconset] = "default";
    this->string_path_[Type::iconset] = "";
  } // set the icons
  { // set the background
    this->string_[Type::background] = "";
    this->string_path_[Type::background] = "";

    vector<string> backgrounds = {"table"};

    auto background = backgrounds.begin();
    for (; background != backgrounds.end(); ++background) {
      this->string_[Type::background] = *background;
      this->update_path(Type::background);
      if (!this->path(Type::background).empty())
        break;
    } // for (background \in backgrounds)

    if (background == backgrounds.end()) {
      this->string_[Type::background] = "";
      if (::ui)
        ::ui->error(_("Error::loading default background %s",
                      backgrounds.front()
                     ));
    }
  } // set the background


#ifdef WINDOWS
  this->string_[Type::name_font]            = "Arial Bold Italic 24";
#else
  this->string_[Type::name_font]            = "Sans Serif Bold Italic 24";
#endif
  this->string_[Type::name_font_color]        = "black";
  this->string_[Type::name_active_font_color]    = "red";
  this->string_[Type::name_reservation_font_color] = "blue";
#ifdef WINDOWS
  this->string_[Type::trickpile_points_font]    = "Arial Bold 24";
#else
  this->string_[Type::trickpile_points_font]    = "Sans Serif Bold 24";
#endif
  this->string_[Type::trickpile_points_font_color] = "black";

  this->string_[Type::poverty_shift_arrow_color] = "black";

#ifdef USE_SOUND_COMMAND
  this->string_[Type::play_sound_command]        = "";
#endif
  this->string_[Type::browser_command]        = "";
} // void Setting::set_to_hardcoded()

/** @return   the data directories
 **/
vector<string> const&
Setting::data_directories() const
{ return this->data_directories_; }

/** sets the data diretories
 **/
void
Setting::set_data_directories()
{
  // search the directory
  this->data_directories_.clear();
  if (getenv("FREEDOKO_DATA_DIRECTORY"))
    this->data_directories_.push_back(getenv("FREEDOKO_DATA_DIRECTORY"));
  if (getenv("FREEDOKO_MORE_DATA_DIRECTORY"))
    this->data_directories_.push_back(getenv("FREEDOKO_MORE_DATA_DIRECTORY"));
  this->data_directories_.push_back(this->value(Type::private_data_directory));


#ifdef PUBLIC_DATA_DIRECTORY_VALUE
  this->data_directories_.push_back(PUBLIC_DATA_DIRECTORY_VALUE);
#else // !#ifdef PUBLIC_DATA_DIRECTORY_VALUE

#ifdef RELEASE

#ifdef WINDOWS
  this->data_directories_.push_back(Reg_read(HKEY_CURRENT_USER,
                                             "Software\\FreeDoko",
                                             "Directory"));
  if (this->data_directories_.back().empty())
    this->data_directories_.pop_back();
  this->data_directories_.push_back(Reg_read(HKEY_LOCAL_MACHINE,
                                             "Software\\FreeDoko",
                                             "Directory"));
  if (this->data_directories_.back().empty())
    this->data_directories_.pop_back();
#endif // #ifdef WINDOWS

#else // !#ifdef RELEASE
  this->data_directories_.push_back("../data");
  this->data_directories_.push_back("../../more_data");
#endif // !#ifdef RELEASE

#endif // !#ifdef PUBLIC_DATA_DIRECTORY_VALUE

  this->data_directories_.push_back(".");

  for (auto& d : this->data_directories_) {
    d = File::absolute_path_from_executable_directory(d);
    if (!File::isdirectory(d))
      d = File::absolute_path_from_executable_directory(".");
  }

  // make each name unique
  for (auto& d : this->data_directories_)
    if (   (d.size() >= 2)
        && (d[0] == '.')
        && (d[1] == '/'))
      d.erase(0, 2);
  unique(this->data_directories_);
} // Setting::set_data_directories()

/** adds the given directory to the data directories
 **
 ** @param    path   directory to add
 **/
void
Setting::add_data_directory(string const& path)
{
  auto d = (   (path.length() >= 2)
              && (path[0] == '.')
              && (path[1] == '/')
              ? string(path, 2)
              : path);
  d = File::absolute_path_from_executable_directory(d);
  if (!contains(this->data_directories_, d)) {
    this->data_directories_.push_back(d);
  }
} // void Setting::add_data_directory(string path)

/** write the data directories in the stream, separated by newlines
 **
 ** @param    ostr     stream to write into
 ** @param    indent   indention of each line, default: ""
 **
 ** @return   stream
 **/
ostream&
Setting::write_data_directories(ostream& ostr, string const& indent) const
{
  for (auto const& d : this->data_directories_)
    ostr << indent << d << '\n';
  return ostr;
} // ostream& Setting::write_data_directories(ostream& ostr, string indent)

/** -> result
 **
 ** @param    indent   indention of each line, default: ""
 **
 ** @return   the public directories as string, separated by newlines
 **/
string
Setting::data_directories_string(string const& indent) const
{
  string l;
  for (auto const& d : this->data_directories_)
    l += indent + d + "\n";
  return l;
} // string Setting::data_directories_string(string indent)

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the setting 'type'
 **/
bool
Setting::operator()(Type::Bool const type) const
{
  if (   (::fast_play & FastPlay::show_all_hands)
      && (type == Type::show_all_hands))
    return true;

  return (this->dependencies(type)
          && this->value(type));
} // bool Setting::operator()(Type::Bool type) const

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the setting 'type'
 **/
unsigned
Setting::operator()(Type::Unsigned const type) const
{
  return this->value(type);
} // unsigned Setting::operator()(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the setting 'type'
 **/
string
Setting::operator()(Type::String const type) const
{
  string value = this->value(type);

  if (value.empty()) {
    switch(type) {
    case Type::name:
#ifdef LINUX
      value = getpwuid(geteuid())->pw_gecos;
      if (value.empty())
        // no real name - take the login
        value = getpwuid(geteuid())->pw_name;
#endif
#ifdef WINDOWS
#ifdef USE_REGISTRY
      {
        TCHAR  infoBuf[1001];
        infoBuf[1000] = 0;
        DWORD  bufCharCount = 1000;
        if (!GetUserName(infoBuf, &bufCharCount))
          value = "Player";
        else
          value = infoBuf;
      }
#else
      value = "Player";
#endif
#endif
      break;
    case Type::language:
      value = "en";
#ifdef LINUX
      if (getenv("LANG"))
        if (string(getenv("LANG"), 0, 2) == string("de"))
          value = "de";
#endif
#ifdef WINDOWS
      value = Reg_read(HKEY_CURRENT_USER,
                       "Software\\FreeDoko",
                       "Language");
      if (value.empty())
        value = Reg_read(HKEY_LOCAL_MACHINE,
                         "Software\\FreeDoko",
                         "Language");

      if (value.empty()) {
        // Ein bisschen kompliziert, aus der Registry die Sprache auszulesen:
        // zuerst wird der Laenderkode gelesen (Deutsch: 407),
        // anschliessend wird geschaut, welche Sprache zu dem Laendercode
        // gehört.
        string const language_code
          = Reg_read(HKEY_CURRENT_USER,
                     "Control Panel\\International",
                     "Locale");

        if (!language_code.empty()) {
          if (string(language_code, language_code.size() - 1)
              == "7") { // deutsch
            value = "de";
          } else if (string(language_code, language_code.size() - 1)
                     == "9") { // english
            value = "en";
          } else if (string(language_code, language_code.size() - 1)
                     == "A") { // spanish
            value = "en";
          } else if (string(language_code, language_code.size() - 1)
                     == "C") { // francais
            value = "en";
          } else { // unknown language code
            value = "en";
          } // unknown langugage code
        } // if (!language_code.empty())

      } // if (value.empty())

      if (value.empty()) {
        // *** DK
        // I didn't find the language in the registry, so I use this workaround
        if ((Reg_read(HKEY_LOCAL_MACHINE,
                      "Software\\Microsoft\\Windows\\CurrentVersion",
                      "ProgramFilesDir") == string("C:\\Programme"))
            && (Reg_read(HKEY_LOCAL_MACHINE,
                         "Software\\Microsoft\\Windows\\CurrentVersion",
                         "ProgramFilesPath") == string("C:\\Programme")
               )
           )
          value = "de";
      } // if (value.empty())
      // We have mostly german users
      if (value.empty())
        value = "de";
#endif
      break;
    case Type::cardset:
      value = "xskat/french";
      break;
    case Type::cards_back:
      value = "default";
      break;
    case Type::iconset:
      value = "default";
      break;
    case Type::background:
      value = "default";
      break;
    case Type::name_font:
      value = "Sans Serif Bold Italic 16";
      break;
    case Type::name_font_color:
      value = "black";
      break;
    case Type::name_active_font_color:
      value = "red";
      break;
    case Type::name_reservation_font_color:
      value = "blue";
      break;
    case Type::trickpile_points_font:
      value = "Sans Serif Bold 20";
      break;
    case Type::trickpile_points_font_color:
      value = "black";
      break;
    case Type::poverty_shift_arrow_color:
      value = "black";
      break;
#ifdef USE_SOUND_COMMAND
    case Type::play_sound_command:
#ifdef WINDOWS
#ifdef USE_REGISTRY
      { // Look in the registry for the default program for wav files
        string class_str
          = Reg_read(HKEY_CURRENT_USER,
                     "Software\\CLASSES\\.wav",
                     "");
        if (class_str.empty())
          class_str = Reg_read(HKEY_LOCAL_MACHINE,
                               "Software\\CLASSES\\.wav",
                               "");
        if (!class_str.empty()) {
          value = Reg_read(HKEY_CURRENT_USER,
                           "Software\\CLASSES\\" + class_str
                           + "\\shell\\open\\command",
                           "");
          if (value.empty())
            value = Reg_read(HKEY_LOCAL_MACHINE,
                             "Software\\CLASSES\\" + class_str
                             + "\\shell\\open\\command",
                             "");
          if (   (value.size() >= 5)  
              && string(value, value.size() - 4, std::string::npos) == "\"\%1\"")
            value = string(value, 0, value.size() - 4);
        } // if (!class_str.empty())
      } // Look in the registry for the default program for wav files
#else
#endif
      if (value.empty())
        value = "explorer";
#else // #ifdef WINDOWS
      value = "aplay";
#endif // #ifdef WINDOWS
      break;
#endif // #ifdef USE_SOUND_COMMAND

    case Type::browser_command:
#ifdef LINUX
      value = "echo could not open a browser for"; // if nothing else works
      if (getenv("BROWSER") != nullptr) {
        value = string(getenv("BROWSER"));
        break;
      }  // if (getenv("BROWSER") != nullptr)

      { // test some browsers and take the first found
        static auto constexpr browsers = {
          "firefox",
          "chromium",
          "google-chrome",
          "epiphany",
          "galeon",
          "mozilla",
          "opera",
          "konqueror",
          "rekonq",
          "midori",
          "arora",
          "web",
          "xterm -e w3m",
          "xterm -e links",
          "xterm -e lynx",
        };

        for (auto const& browser : browsers) {
          if (::system(("test -x /usr/bin/"s + browser
                        + " -o -x /usr/bin/X11/"s + browser
                        + " -o -x /usr/local/bin/"s + browser).c_str())
              == 0) {
            value = browser;
            break;
          }
        }
      } // test some browsers and take the first found
#endif
#ifdef WINDOWS
#ifdef USE_REGISTRY
      { // Look in the registry for the default browser for html
        string class_str
          = Reg_read(HKEY_CURRENT_USER,
                     "Software\\CLASSES\\.html",
                     "");
        if (class_str.empty())
          class_str = Reg_read(HKEY_LOCAL_MACHINE,
                               "Software\\CLASSES\\.html",
                               "");
        if (!class_str.empty()) {
          value = Reg_read(HKEY_CURRENT_USER,
                           "Software\\CLASSES\\" + class_str
                           + "\\shell\\open\\command",
                           "");
          if (value.empty())
            value = Reg_read(HKEY_LOCAL_MACHINE,
                             "Software\\CLASSES\\" + class_str
                             + "\\shell\\open\\command",
                             "");
          if (   (value.size() >= 5)  
              && string(value, value.size() - 4, std::string::npos) == "\"\%1\"")
            value = string(value, 0, value.size() - 4);
        } // if (!class_str.empty())
      } // Look in the registry for the default browser for html
#else
#endif
      if (value.empty())
        value = "explorer";
#endif
      break;
    } // switch(type)
  } // if (value.empty())

  switch(type) {
  case Type::cardset:
    value = ((*this)(Type::cardsets_directory) + "/"
             + value);
    break;
  case Type::cards_back:
    value = ((*this)(Type::cards_back_directory) + "/"
             + value);
    break;
  case Type::iconset:
    value = ((*this)(Type::iconsets_directory) + "/"
             + value);
    break;
  case Type::background:
    value = ((*this)(Type::backgrounds_directory) + "/"
             + value);
    break;
  case Type::name:
  case Type::language:
  case Type::name_font:
  case Type::name_font_color:
  case Type::name_active_font_color:
  case Type::name_reservation_font_color:
  case Type::trickpile_points_font:
  case Type::trickpile_points_font_color:
  case Type::poverty_shift_arrow_color:
#ifdef USE_SOUND_COMMAND
  case Type::play_sound_command:
#endif
  case Type::browser_command:
    break;
  } // switch(type)

  return value;
} // string Setting::operator()(Type::String type) const

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   value of setting 'type'
 **/
string
Setting::operator()(Type::StringConst const type) const
{
  switch(type) {
  case Type::cardset_license:
    return this->value(Type::cardset_license);
  case Type::iconset_license:
    return this->value(Type::iconset_license);
  case Type::public_data_directory:
    return File::filename_expand(this->value(Type::public_data_directory));
  case Type::private_data_directory:
    return File::filename_expand(this->value(Type::private_data_directory));
  case Type::settings_file:
    return ((*this)(Type::private_data_directory)
            + "/" + this->value(type));
  case Type::logo_file:
    return ((*this)(Type::public_data_directory)
            + this->value(type));
  case Type::icon_file:
    return ((*this)(Type::public_data_directory)
            + this->value(type));
  case Type::graphic_extension:
    return this->value(type);
  case Type::sound_extension:
    return this->value(type);
  case Type::cardsets_directory:
    return this->value(type);
  case Type::cards_directory:
    return ((*this)(Type::cardset) + "/"
            + this->value(type));
  case Type::cards_back_directory:
    return ((*this)(Type::cardset) + "/"
            + this->value(type));
  case Type::icons_directory:
    return (this->value(Type::iconsets_directory)
            + "/" + this->value(Type::iconset));
  case Type::iconsets_directory:
    return this->value(type);
  case Type::icons_from_cardset_directory:
    return ((*this)(Type::cardset) + "/"
            + this->value(type));
  case Type::backgrounds_directory:
    return this->value(type);
  case Type::sounds_directory:
    return (this->value(type) + "/" + (*this)(Type::language));
  case Type::ai_directory:
    return this->value(type);
  case Type::parties_directory:
    return this->value(type);
  case Type::rules_directory:
    return this->value(type);
  case Type::settings_directory:
    return this->value(type);
  case Type::manual_directory:
    return this->value(type);
  } // switch(type)

  return "";
} // string Setting::operator()(Type::StringConst type) const

/** -> result
 **
 ** @param    type   cards order
 **
 ** @return   the cards order
 **/
Setting::CardsOrder const&
Setting::operator()(Type::CardsOrder type) const
{
  return this->value(type);
} // CardsOrder Setting::operator()(Type::CardsOrder type) const

/** -> result
 **
 ** @param    type   cards order
 **
 ** @return   the cards order
 **/
Setting::CardsOrder&
Setting::operator()(Type::CardsOrder const type)
{
  return this->value(type);
} // CardsOrder& Setting::operator()(Type::CardsOrder type)

/** -> result
 **
 ** @param    name   name of the type
 **
 ** @return   the type with the name 'name';
 **   -1, if the name is not recognized
 **/
int
Setting::type(string const& name) const
{
  for (auto t : Type::bool_list)
    if (name == to_string(t))
      return t;

  for (auto t : Type::unsigned_list)
    if (name == to_string(t))
      return t;

  for (auto t : Type::string_list)
    if (name == to_string(t))
      return t;

  if (name == to_string(Type::cards_order))
    return Type::cards_order;

  return -1;
} // int Setting::type(string name) const

/** -> result
 **
 ** @param    name   name of the type
 **
 ** @return   the type with the translated name 'name';
 **   -1, if the name is not recognized
 **/
int
Setting::type_translated(string const& name) const
{
  for (auto t : Type::bool_list)
    if (name == _(t))
      return t;

  for (auto t : Type::unsigned_list)
    if (name == _(t))
      return t;

  for (auto t : Type::string_list)
    if (name == _(t))
      return t;

  if (name == _(Type::cards_order))
    return Type::cards_order;

  return -1;
} // int Setting::type_translated(string name) const

/** -> result
 **
 ** @param    name   name of the type
 **
 ** @return   the type with the translated or normal name 'name';
 **   -1, if the name is not recognized
 **/
int
Setting::type_translated_or_normal(string const& name) const
{
  int type = this->type_translated(name);
  if (type == -1)
    type = this->type(name);

  return type;
} // int Setting::type_translated_or_normal(string name) const

/** -> result
 **
 ** @param    type   the type (bool)
 **
 ** @return   whether the dependencies of the setting 'type' are met
 **/
bool
Setting::dependencies(Type::Bool const type) const
{
  switch(type) {
  case Type::automatic_savings:
    return true;
  case Type::save_party_changes:
    return (*this)(Type::automatic_savings);
  case Type::additional_party_settings:
    return true;
  case Type::show_splash_screen:
    return true;
  case Type::splash_screen_transparent:
    return (*this)(Type::show_splash_screen);
  case Type::show_bug_report_button_in_game_finished_window:
  case Type::save_bug_reports_on_desktop:
  case Type::sound:
  case Type::automatic_card_suggestion:
  case Type::announce_swines_automatically:
  case Type::show_if_valid:
  case Type::emphasize_valid_cards:
  case Type::announce_in_table:
  case Type::show_all_hands:
  case Type::show_ai_information_hands:
  case Type::show_ai_information_teams:
  case Type::show_trickpiles_points:
  case Type::show_known_teams_in_game:
  case Type::show_soloplayer_in_game:
    // the showing is also dependent of the rule
    return true;
  case Type::show_full_trick_window_if_special_points:
    return !(*this)(Type::show_full_trick_window);
  case Type::close_gametype_window_automatically:
    return (*this)(Type::show_gametype_window);
  case Type::close_marriage_window_automatically:
    return (*this)(Type::show_marriage_window);
  case Type::close_genscher_window_automatically:
    return (*this)(Type::show_genscher_window);
  case Type::close_announcement_window_automatically:
    return (*this)(Type::show_announcement_window);
  case Type::close_swines_window_automatically:
    return (*this)(Type::show_swines_window);
  case Type::show_full_trick_window:
  case Type::close_full_trick_automatically:
  case Type::show_gametype_window:
  case Type::show_marriage_window:
  case Type::show_genscher_window:
  case Type::show_announcement_window:
  case Type::show_swines_window:
  case Type::rotate_trick_cards:
  case Type::own_hand_on_table_bottom:
  case Type::original_cards_size:
  case Type::ui_main_window_fullscreen:
    return true;
  } // switch(type)

  return true;
} // bool Setting::dependencies(Type::Bool type) const

/** -> result
 **
 ** @param    type   the type (unsigned)
 **
 ** @return   whether the dependencies of the setting 'type' are met
 **/
bool
Setting::dependencies(Type::Unsigned const type) const
{
  switch(type) {
  case Type::full_trick_close_delay:
    return (*this)(Type::close_full_trick_automatically);
  case Type::gametype_window_close_delay:
    return (*this)(Type::close_gametype_window_automatically);
  case Type::marriage_window_close_delay:
    return (*this)(Type::close_marriage_window_automatically);
  case Type::genscher_window_close_delay:
    return (*this)(Type::close_genscher_window_automatically);
  case Type::announcement_window_close_delay:
    return (*this)(Type::close_announcement_window_automatically);
  case Type::swines_window_close_delay:
    return (*this)(Type::close_swines_window_automatically);
  case Type::cards_height:
    return !(*this)(Type::original_cards_size);
  default:
    return true;
  } // switch(type)

  return true;
} // bool Setting::dependencies(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   the type (string)
 **
 ** @return   whether the dependencies of the setting 'type' are met
 **/
bool
Setting::dependencies(Type::String const type) const
{
  (void)type;
  return true;
} // bool Setting::dependencies(Type::String type) const

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the value of the setting 'type'
 **/
bool
Setting::value(Type::Bool const type) const
{
  return this->bool_.at(type);
} // bool Setting::value(Type::Bool type) const

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the value of the (bool) setting 'type'
 **/
unsigned
Setting::value(Type::Unsigned const type) const
{
  return this->unsigned_.at(type);
} // unsigned Setting::value(Type::Unsigned type)

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the value of the (string) setting 'type'
 **/
string const&
Setting::value(Type::String const type) const
{
  return this->string_.at(type);
} // string Setting::value(Type::String type)

/** searches the file for 'type' in the directories
 **
 ** @param    type   setting type
 **
 ** @return   the file/directory with 'type'
 **            empty string if not found
 **/
string const&
Setting::path(Type::String const type) const
{
  return this->string_path_.at(type);
} // string Setting::path(Type::String type) const

/** searches the file for 'type' in the directories
 **
 ** @param    type   setting type
 **
 ** @return   the file/directory with 'type'
 **            empty string if not found
 **/
string
Setting::path(Type::StringConst const type) const
{
  switch (type) {
  case Type::cardset_license:
    return "";
  case Type::iconset_license:
    return "";

  case Type::public_data_directory:
  case Type::private_data_directory:
    return this->value(type);

  case Type::settings_file:
    return (this->value(Type::private_data_directory) + "/"
            + this->value(type));

  case Type::logo_file:
  case Type::icon_file: {
    // search the directory

    for (auto const& dd : this->data_directories()) {
      string const file = (dd + "/" + this->value(type));
      // check for the file
      ifstream istr(file.c_str());
      if (istr.good()) {
        return file;
      }
    } // for (dd : this->data_directories())

    return "";
  } // case Type::logo_file, icon_file

    // the following directories are subdirectories of the
    // public/private data directory
  case Type::cardsets_directory:
  case Type::cards_directory:
  case Type::cards_back_directory:
  case Type::icons_from_cardset_directory:
  case Type::iconsets_directory:
  case Type::icons_directory:
  case Type::backgrounds_directory:
  case Type::sounds_directory:
  case Type::ai_directory:
  case Type::parties_directory:
  case Type::rules_directory:
  case Type::settings_directory:
  case Type::manual_directory:
    return "";

  case Type::graphic_extension:
  case Type::sound_extension:
    return "";
  } // switch(type)

  return "";
} // string Setting::path(Type::StringConst type) const

/** search a cards back
 **/
void
Setting::search_cards_back()
{
  if (!this->path(Type::cards_back).empty())
    return ;

  string new_back;

  { // search the file from the settings
    auto const file = (this->path(Type::cardset) + "/"
                       + this->value(Type::cards_back_directory) + "/"
                       + this->value(Type::cards_back)
                       + "." + (*this)(Type::graphic_extension));
    // check for the file
    ifstream istr(file.c_str());
    if (istr.good()) {
      new_back = file;
    }

    if (new_back.empty()) {
      for (auto const& dd : this->data_directories()) {
        auto const file = (dd + "/" + (*this)(Type::cardset) + "/"
                           + this->value(Type::cards_back_directory) + "/"
                           + this->value(Type::cards_back)
                           + "." + (*this)(Type::graphic_extension));
        // check for the file
        ifstream istr(file.c_str());
        if (istr.good()) {
          new_back = file;
          break;
        }
      } // for (dd : this->data_directories)
    } // if (new_back.empty())
  } // search the file from the settings


  // search any back
  for (auto const& dd : this->data_directories()) {
    auto const directory = (dd + "/" + (*this)(Type::cards_back_directory));

    DIR *dir = opendir(directory.c_str());

    if (dir == nullptr)
      continue;

    if (dir != nullptr) {
      struct dirent *entry;

      // search all entries in the directory
      while ((entry = readdir(dir)) != nullptr) {
        if ((strcmp(entry->d_name, ".") == 0) // NOLINT
            || (strcmp(entry->d_name, "..") == 0)) // NOLINT
          continue;
        if (File::isfile(directory + "/" + entry->d_name)) { // NOLINT
          string const basename = File::basename(entry->d_name); // NOLINT
          if ((basename.size()
               > 1 + (*this)(Type::graphic_extension).size())
              && (string(basename,
                         basename.size()
                         - 1
                         - (*this)(Type::graphic_extension).size(),
                         1 + (*this)(Type::graphic_extension).size())
                  == "." + (*this)(Type::graphic_extension))
             ) {
            new_back = string(basename,
                              0,
                              basename.size()
                              - 1 - (*this)(Type::graphic_extension).size());
            break;
          } // if (extension is right)
        } // if (File::isfile(entry->d_name)) // NOLINT
      } // while (entry != nullptr) ;
      closedir(dir);
    } // if (dir != nullptr)
  } // for (dd : this->data_directories())

  if (new_back.empty()) {
    if (::ui)
      ::ui->error(_("Error::could not find any back for cardset %s", this->value(Type::cardset)));
  } else {
    this->set(Type::cards_back, new_back);
  }
} // void Setting::search_cards_back()

/** search a background
 **/
void
Setting::search_background()
{
  if (!this->path(Type::background).empty())
    return ;

  string new_background;

  // search any background
  for (auto const& dd : this->data_directories()) {
    auto const directory = (dd + "/" + (*this)(Type::backgrounds_directory));

    DIR *dir = opendir(directory.c_str());

    if (dir == nullptr)
      continue;

    if (dir != nullptr) {
      struct dirent *entry;

      // search all entries in the directory
      while ((entry = readdir(dir))) {
        if (   (strcmp(entry->d_name, ".") == 0) // NOLINT
            || (strcmp(entry->d_name, "..") == 0)) // NOLINT
          continue;
        if (File::isfile(directory + "/" + entry->d_name)) { // NOLINT
          string const basename = File::basename(entry->d_name); // NOLINT
          if ((basename.size()
               > 1 + (*this)(Type::graphic_extension).size())
              && (string(basename,
                         basename.size()
                         - 1
                         - (*this)(Type::graphic_extension).size(),
                         1 + (*this)(Type::graphic_extension).size())
                  == "." + (*this)(Type::graphic_extension))
             ) {
            new_background = string(basename,
                                    0,
                                    basename.size()
                                    - 1 - (*this)(Type::graphic_extension).size());
            break;
          } // if (extension is right)
        } // if (File::isfile(entry->d_name)) // NOLINT
      } // while (entry)
      closedir(dir);
    } // if (dir != nullptr)
  } // for (dd : datadir)

  if (new_background.empty()) {
    if (::ui)
      ::ui->error(_("Error::could not find any background"));
  } else {
    this->set(Type::background, new_background);
  }
} // void Setting::search_background()

/** -> result
 **
 ** @param    type   setting type
 **
 ** @return   the value of the (string) setting 'type'
 **/
string
Setting::value(Type::StringConst const type) const
{
  switch(type) {
  case Type::cardset_license:
    return this->cardset_license_;
  case Type::iconset_license:
    return this->iconset_license_;
  case Type::public_data_directory:
    {
      string value;

#ifdef PUBLIC_DATA_DIRECTORY_VALUE
      value = PUBLIC_DATA_DIRECTORY_VALUE;
#else // !#ifdef PUBLIC_DATA_DIRECTORY_VALUE

#ifdef RELEASE

#ifndef WINDOWS
      value = "/usr/share/games/FreeDoko";
#else // !#ifndef WINDOWS
      value = Reg_read(HKEY_LOCAL_MACHINE,
                       "Software\\FreeDoko",
                       "Directory");
#endif // #ifndef WINDOWS

#else // !#ifdef RELEASE
      value = "../data";
#endif // !#ifdef RELEASE

      { // test that the directory exists
        struct stat status;
        stat(value.c_str(), &status);
        if (!S_ISDIR(status.st_mode))
          value = ".";
      } // test that the directory exists

#endif // !#ifdef PUBLIC_DATA_DIRECTORY_VALUE
      if (value.empty())
        value = ".";
      return value;
    }
    break;
  case Type::private_data_directory:
    {
      string value;
#if defined(LINUX) || defined(HPUX)
      value = "~/.FreeDoko";
#endif
#ifdef WINDOWS
      value = "~/FreeDoko";
#endif
      return File::filename_expand(value);
    }
  case Type::settings_file:
    return "FreeDokorc";
  case Type::logo_file:
    return "logo.png";
  case Type::icon_file:
    return "icon.png";
  case Type::graphic_extension:
    return "png";
  case Type::sound_extension:
    return "wav";
  case Type::cardsets_directory:
    return "cardsets";
  case Type::cards_directory:
    return "cards";
  case Type::cards_back_directory:
    return "backs";
  case Type::icons_from_cardset_directory:
    return "icons";
  case Type::iconsets_directory:
    return "iconsets";
  case Type::icons_directory:
    return "default";
  case Type::backgrounds_directory:
    return "backgrounds";
  case Type::sounds_directory:
    return "sounds";
  case Type::ai_directory:
    return "ai";
  case Type::parties_directory:
    return "parties";
  case Type::rules_directory:
    return "rules";
  case Type::settings_directory:
    return "settings";
  case Type::manual_directory:
#ifdef MANUAL_DIRECTORY_VALUE
    return MANUAL_DIRECTORY_VALUE;
#endif // !#ifdef MANUAL_DIRECTORY_VALUE
    break;
  } // switch(type)

  if (   File::isdirectory("../src")
      && File::isdirectory("../manual"))
    return "../manual";

  return "";
} // string Setting::value(Type::StringConst type)

/** -> result
 **
 ** @param    type   setting type (only 'cards_order')
 **
 ** @return   the value of the cards order
 **/
Setting::CardsOrder const&
Setting::value(Type::CardsOrder const type) const
{
  return *this->cards_order_;
} // Setting::CardsOrder Setting::value(Type::CardsOrder type) const

/** -> result
 **
 ** @param    type   setting type (only 'cards_order')
 **
 ** @return   the value of the cards order
 **/
Setting::CardsOrder&
Setting::value(Type::CardsOrder const type)
{
  return *this->cards_order_;
} // Setting::CardsOrder& Setting::value(Type::CardsOrder type)

/** -> result
 **
 ** @param    type   setting type (Type Unsigned)
 **
 ** @return   the minimal value of the setting 'type'
 **/
unsigned
Setting::min(Type::Unsigned const type) const
{
  switch(type) {
#ifdef USE_THREADS
  case Type::threads_max:
    return 1;
#endif
  case Type::table_rotation:
    return 0;
  case Type::cards_height:
    return 2;
  default:
    return 0;
  } // switch(type)
} // unsigned Setting::min(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   setting type (Type Unsigned)
 **
 ** @return   the maximal value of the setting 'type'
 **/
unsigned
Setting::max(Type::Unsigned const type) const
{
  switch(type) {
#ifdef USE_THREADS
  case Type::threads_max:
    return 12;
#endif
  case Type::card_play_delay:
    return 60000; // 60 seconds
  case Type::full_trick_close_delay:
  case Type::gametype_window_close_delay:
  case Type::marriage_window_close_delay:
  case Type::genscher_window_close_delay:
  case Type::announcement_window_close_delay:
  case Type::swines_window_close_delay:
    return 60000; // 60 seconds
  case Type::table_rotation:
    return 3;
  case Type::cards_height:
    return 1000;
  case Type::ui_main_window_pos_x:
  case Type::ui_main_window_pos_y:
  case Type::ui_main_window_width:
  case Type::ui_main_window_height:
    return UINT_MAX;
  } // switch(type)

  return 0;
} // unsigned Setting::max(Type::Unsigned type) const

/** @return   the themes list
 **/
vector<unique_ptr<Setting::Theme>> const&
Setting::themes() const
{
  return this->themes_;
} // vector<unique_ptr<Setting::Theme>> Setting::themes() const

/** -> result
 **
 ** @param    name   name of the theme
 **
 ** @return   the theme with the name 'name'
 **            'nullptr' if none is found
 **/
Setting::Theme const*
Setting::theme(string const& name) const
{
  for (auto const& t : this->themes_)
    if (t->name() == name)
      return t.get();

  return nullptr;
} // Setting::Theme const* Setting::theme(string const& name) const

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type Bool)
 ** @param    value   new value
 **/
void
Setting::set(Type::Bool const type, bool const value)
{
  if (this->value(type) == value)
    return ;

  auto const old_value = this->bool_[type];
  this->bool_[type] = value;

  if (   ::ui
      && (this == &::setting))
    this->signal_changed_(type, &old_value);
} // void Setting::set(Type::Bool type, bool value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type Bool)
 ** @param    value   new value (as string)
 **/
void
Setting::set(Type::Bool const type, string const& value)
{
  if (   (value != "true")
      && (value != "false")
      && (value != "yes")
      && (value != "no")
      && (value != "0")
      && (value != "1")) {
    cerr << "illegal value '" << value << "' for '" << type << "', "
      << "must be a boolean ('true' or 'false' or 'yes' or 'no' or '1' or '0'. "
      << "Taking 'false'."
      << endl;
  }

  this->set(type, ((   (value == "false")
                    || (value == "no")
                    || (value == "0"))
                   ? false : true)); // NOLINT
} // void Setting::set(Type::Bool type, string value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type Unsigned)
 ** @param    value   new value
 **/
void
Setting::set(Type::Unsigned const type, unsigned const value)
{
  if (this->value(type) == value)
    return ;

  auto const old_value = this->unsigned_[type];
  this->unsigned_[type] = value;

  this->signal_changed_(type, &old_value);
} // void Setting::set(Type::Unsigned type, unsigned value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type Unsigned)
 ** @param    value   new value (as string)
 **/
void
Setting::set(Type::Unsigned const type, string const& value)
{
  try {
    auto const number = stol(value);
    this->set(type, number);
  } catch (...) {
    cerr << "illegal value '" << value << "' for '" << type << "', "
      << "must be a digit. "
      << endl;
  }
} // void Setting::set(Type::Unsigned type, string value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type String)
 ** @param    value   new value
 **/
void
Setting::set(Type::String const type, string const& value)
{
  if (this->value(type) == value)
    return ;

  auto const old_value = this->string_[type];

  // First test, whether this is the default value.
  // If it is, then leave the value empty.
#ifdef POSTPONED
  this->string_[type] = "";
#endif // #ifdef POSTPONED 

  this->string_[type] = value;

  this->update_path(type);

  { // check for path values
    switch(type) {
    case Type::background:
    case Type::cards_back:
    case Type::cardset:
    case Type::iconset:
      if (this->path(type).empty()) {
        this->string_[type] = old_value;
        return ;
      }
      break;
    default:
      break;
    } // switch (type)
  } // check for path values

  if (this == &::setting) {
    switch(type) {
    case Type::name:
      break;
    case Type::language:
      set_language((*this)(Type::language));
      break;
    case Type::cardset:
    case Type::iconset:
      { // read the license
        vector<string> const license_files
          = {"copyright",
            "Copyright",
            "COPYright",
            "license",
            "License",
            "LICENSE",
            "copyright.txt",
            "Copyright.txt",
            "COPYright.txt",
            "license.txt",
            "License.txt",
            "LICENSE.txt"};

        string directory = this->path(type);

        string license_text;

        for (auto const& license : license_files) {
          if (File::isfile(directory + "/"
                           + license)) {
            license_text = String::getfile(directory + "/"
                                           + license);
          } else if (File::isfile(directory + "/../"
                                  + license)) {
            license_text = String::getfile(directory + "/../"
                                           + license);
          } // if (license found)
          if (!license_text.empty())
            break;
        } // for (license : license_files)

        if (license_text.empty())
          // no license found
          license_text = "license not found :-(";

        if (type == Type::cardset) {
          this->cardset_license_ = license_text;
        } else {
          this->iconset_license_ = license_text;
        }
      } // read the license
      if (type == Type::cardset) {
        this->search_cards_back();
      }
      break;

    default:
      break;
    } // switch(type)
  } // if (this == &::setting)

  this->signal_changed_(type, &old_value);
} // void Setting::set(Type::String type, string value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type CardsOrder)
 ** @param    value   new value
 **/
void
Setting::set(Type::CardsOrder const type, CardsOrder const& value)
{
  if (this->value(type) == value) {
    return ;
  }

  auto const old_value = *this->cards_order_;

  // the following code keeps the sorting order when the cards are mixed,
  if (!value.sorted()) {
    this->cards_order_->sorted_set(value.sorted());
  } else { // if !(!value.sorted())
    *this->cards_order_ = value;

    if (this == &::setting) {
      if (game_status & GameStatus::game)
        ::party->game().players().sort_hands();

      this->signal_changed_(type, &old_value);
    } // if (this == &::setting)
  } // if !(!value.sorted())
} // void Setting::set(Type::CardsOrder type, CardsOrder value)

/** sets the 'type' to 'value'
 **
 ** @param    type   setting type (Type CardsOrder)
 ** @param    value   new value (as string)
 **/
void
Setting::set(Type::CardsOrder const type, string const& value)
{
  this->set(type, CardsOrder(*this, value));
} // void Setting::set(Type::CardsOrder type, string value)

/** sets the 'type' to 'value'
 ** the parameters are strings
 **
 ** @param    type   name of the setting type
 ** @param    value   new value
 **
 ** @return   whether the setting was successful
 **/
bool
Setting::set(string const& type, string const& value)
{
  for (auto t : Type::bool_list) {
    if (   (type == to_string(t))
        || (type == _(t))) {
      this->set(t, value);
      return true;
    }
  }
  for (auto t : Type::unsigned_list) {
    if (   (type == to_string(t))
        || (type == _(t))) {
      this->set(t, value);
      return true;
    }
  }
  for (auto t : Type::string_list) {
    if (   (type == to_string(t))
        || (type == _(t))) {
      this->set(t, value);
      return true;
    }
  }

  if ((type == to_string(Type::CardsOrder::cards_order))
      || (type == _(Type::cards_order))) {
    this->set(Type::cards_order, value);
    return true;
  }

  return false;
} // bool Setting::set(string type, string value)

/** set the values from the theme
 **
 ** @param    theme   theme to set
 **/
void
Setting::set(Theme const& theme)
{
  for (auto const& t : theme.bool_)
    this->set(t.first, t.second);
  for (auto const& t : theme.unsigned_)
    this->set(t.first, t.second);
  for (auto const& t : theme.string_)
    this->set(t.first, t.second);
} // void Setting::set(Theme theme)

/** set the values from the theme 'name'
 **
 ** @param    name   name of the theme
 **
 ** @return   whether the theme was found
 **/
bool
Setting::set_theme(string const& name)
{
  auto const& theme = this->theme(name);
  if (!theme)
    return false;

  this->set(*theme);
  return true;
} // bool Setting::set_theme(string const& name)

/** -> result
 **
 ** @param    playerno   the number of the player
 **
 ** @return   the position of the player playerno
 **/
Position
Setting::position(unsigned const playerno) const
{
  auto const& game = ::party->game();
  auto rotation = ::setting(Type::table_rotation);
  if (   ::setting(Type::own_hand_on_table_bottom)
      && ::party->players().count_humans() == 1
      && ::game_status & GameStatus::game) {
    auto const& human = *::party->players().human_player();

    if (game.players().contains(human))
      rotation = game.players().no(human);
  }

  switch ( (playerno + 4 - rotation) % 4) {
  case 0:
    return Position::south;
  case 1:
    return Position::west;
  case 2:
    return Position::north;
  case 3:
    return Position::east;
  default:
    DEBUG_ASSERTION(false,
                    "Table::position(playerno):\n"
                    "  player number " << playerno
                    << " invalid.");
    break;
  } // switch(playerno)

  return Position::north;
} // Position Table::position(unsigned playerno) const

/** load the settings from the default file
 **/
void
Setting::load()
{
  this->load((*this)(Type::public_data_directory) + "/"
             + this->value(Type::settings_file),
             false);
  this->load((*this)(Type::settings_file), false);
} // void Setting::load()

/** load the settings from the given file
 **
 ** @param    filename   file to load the settings from
 ** @param    filename_not_exists_output   whether to show an information
 **   if the file could not be opened
 **/
void
Setting::load(string const& filename, bool filename_not_exists_output)
{
  ifstream istr(filename.c_str());

  if (istr.fail()) {
    if (filename_not_exists_output)
      cerr << "Setting::load(" << filename << "):\n"
        << "  Error while opening the file"
        << endl;
    return ;
  }

  // load the settings
  while (istr.good()) {
    Config config;
    istr >> config;

    // finished with the config file
    if ((config.name == "") && (config.value == ""))
      break;

    if (config.separator) {
      // a setting
      if (!set(config.name, config.value)) {
        cerr << "setting file:    "
          << "ignoring unknown setting '" << config.name << "'."
          << endl;
      }
    } else { // if (config.separator)
      // a setting
      // if the value is in parentencies, remove both
      if (config.name == "!input") {
        // include the given file
        this->load((File::dirname(filename) + "/"
                    + config.value));
      } else if (config.name == "!end") {
        // ignore the rest of the file
        break;
      } else if (config.name == "!stdout") {
        // output of the data to 'stdout'
        cout << config.value << endl;
      } else if (config.name == "!stderr") {
        // output of the data to 'stderr'
        cerr << config.value << endl;
      } else if (config.name == "") {
        cerr << "Setting file \'" << filename << "\':    "
          << "Ignoring line \'" << config.value << "\'.\n";
      } else {
        cerr << "Setting file \'" << filename << "\':    "
          << "Setting '" << config.name << "' unknown.\n"
          << "Ignoring it.\n";
      } // if (config.name == .)
    } // config.separator

  } // while (istr.good())
} // void Setting::load(string filename, bool filename_not_exists_output)

/** check whether a cardset is outdated and replaces it
 **/
void
Setting::check_for_outdated_cardset()
{
  // since version 0.7.12 (from 2014)
  // replace Altenburg cardset with InnoCard
  auto const cardset_old = this->value(Type::cardset);
  bool changed = false;
  if (this->value(Type::cardset) == "InnoCard/french") {
    this->set(Type::cardset, "InnoCard/french/185");
    changed = true;
  }
  //if (string(this->value(Type::cardset), 0, 8) == "InnoCard") {
  //this->set(Type::cardset, "xskat/french");
  //changed = true;
  //}
  if (this->value(Type::cardset) == "Altenburg/english") {
    this->set(Type::cardset, "InnoCard/english/185");
    //this->set(cardset, "dondorf");
    changed = true;
  }
  if (this->value(Type::cardset) == "Altenburg/german") {
    this->set(Type::cardset, "xskat/german");
    changed = true;
  }
  if (string(this->value(Type::cardset), 0, 10) == "Altenburg") {
    this->set(Type::cardset, "InnoCard/french/185");
    //this->set(Type::cardset, "xskat/french");
    changed = true;
  }
  if (changed && (this->value(Type::cardset) == cardset_old)) {
    this->set(Type::cardset, "xskat/french");
  }
  if (changed) {
    // gettext: %s = old cardset, %s = replacement
    ::ui->information(_("cardset replaced: %s, %s", cardset_old, this->value(Type::cardset)),
                      InformationType::problem, true);
  }
} // void Setting::check_for_outdated_cardset()

/** save the settings in the default file
 **
 ** @return   whether the saving was successful
 **/
bool
Setting::save() const
{
  return this->save((*this)(Type::settings_file));
} // bool Setting::save() const

/** save the settings in the given file
 **
 ** @param    filename   file into which to save the settings
 **
 ** @return   whether the saving was successful
 **/
bool
Setting::save(string const& filename) const
{
  auto filename_tmp = filename + ".tmp";
  ofstream ostr(filename_tmp.c_str());
  if (!ostr.good()) {
    ::ui->information(_("Error::Setting::save: Error opening temporary file %s. Aborting saving.", filename_tmp), InformationType::problem);
    return false;
  }

  ostr << "#!FreeDoko -s\n"
    << "# FreeDoko settings file (version " << *::version << ")\n"
    << '\n';
  this->write(ostr);

  if (!ostr.good()) {
    ::ui->information(_("Error::Setting::save: Error saving in temporary file %s. Keeping temporary file (for bug tracking).", filename_tmp), InformationType::problem);
    return false;
  }
  ostr.close();

#ifdef WINDOWS
  unlink(filename.c_str());
#endif
  if (rename(filename_tmp.c_str(), filename.c_str())) {
    ::ui->information(_("Error::Setting::save: Could not rename temporary file %s to requested file %s. Keeping temporary file.", filename_tmp, filename), InformationType::problem);
    return false;
  }

  return true;
} // bool Setting::save(string filename)

/** updates all settings in the ui
 **/
void
Setting::update_all()
{
  if (this != &::setting)
    return ;

  for (auto t : Type::bool_list) {
    auto const v = this->value(t);
    this->signal_changed_(t, &v);
  }
  for (auto t : Type::unsigned_list) {
    auto const v = this->value(t);
    this->signal_changed_(t, &v);
  }
  for (auto t : Type::string_list) {
    this->signal_changed_(t, &(this->value(t)));
  }
  this->signal_changed_(Type::cards_order, &(this->value(Type::cards_order)));
} // void Setting::update_all()

/** create the themes list
 **/
void
Setting::create_themes_list()
{
  this->themes_.clear();

  { // InnoCard french
    auto theme = make_unique<Theme>("french (InnoCard)");
    (void)_("Setting::Theme::french (InnoCard)");
    theme->set(Type::cardset, "InnoCard/french/185");
    theme->set(Type::cards_back, "penguin");
    theme->set(Type::language, "de");
    theme->set(Type::background, "table");
    this->themes_.push_back(move(theme));
  } // InnoCard french
  { // xskat german
    auto theme = make_unique<Theme>("german (xskat)");
    (void)_("Setting::Theme::german (xskat)");
    theme->set(Type::cardset, "xskat/german");
    theme->set(Type::cards_back, "penguin");
    //theme->set(Type::language, "de-alt");
    theme->set(Type::language, "de");
    theme->set(Type::background, "table");
    this->themes_.push_back(move(theme));
  } // xskat german
  { // InnoCard english
    auto theme = make_unique<Theme>("english (InnoCard)");
    (void)_("Setting::Theme::english (InnoCard)");
    theme->set(Type::cardset, "InnoCard/english/185");
    theme->set(Type::cards_back, "penguin");
    theme->set(Type::language, "en");
    theme->set(Type::background, "table");
    this->themes_.push_back(move(theme));
  } // InnoCard english
#if 0
  { // dondorf english
    auto theme = make_unique<Theme>("english (dondorf)");
    (void)_("Setting::Theme::english (dondorf)");
    theme->set(Type::cardset, "dondorf");
    theme->set(Type::cards_back, "penguin");
    theme->set(Type::language, "en");
    this->themes_.push_back(move(theme));
  } // dondorf english
#endif
#if 0
  { // autumn
    auto theme = make_unique<Theme>("autumn"));
    (void)_("Setting::Theme::autumn");
    theme->set(Type::cards_back, "leaves");
    theme->set(Type::background, "grass-with-leaves");
    this->themes_.push_back(move(theme));
  } // autumn
#endif
  { // christmas
    auto theme = make_unique<Theme>("christmas");
    (void)_("Setting::Theme::christmas");
    theme->set(Type::cards_back, "christmas/candle");
    theme->set(Type::background, "christmas/sky");
    this->themes_.push_back(move(theme));
  } // christmas
} // void Setting::create_themes_list()

/** update the path of the string setting 'type'
 **
 ** @param    type   setting type
 **/
void
Setting::update_path(Type::String const type)
{
  auto const path_bak = this->string_path_[type];

  switch(type) {
  case Type::name:
  case Type::name_font:
  case Type::name_font_color:
  case Type::name_active_font_color:
  case Type::name_reservation_font_color:
  case Type::trickpile_points_font:
  case Type::trickpile_points_font_color:
  case Type::poverty_shift_arrow_color:
  case Type::language:
    break;

  case Type::cardset: {
    // search the directory
    vector<string> subdir;
    subdir.push_back((*this)(Type::cardset));
    auto const datadir = this->data_directories();

    for (auto const& sd : subdir) {
#ifdef WORKAROUND
      // check for Altenburg cardset
      if (   (altenburg_expiration_date < Date() )
          && (sd.find("Altenburg") != string::npos) ) {
        ::ui->information(_("License: Altenburg expired"),
                          InformationType::warning);
        continue;
      } // if (Altenburg license expired)
      // check for InnoCard cardset
      if (   (innocard_expiration_date < Date() )
          && (sd.find("InnoCard") != string::npos) ) {
        ::ui->information(_("License: InnoCard expired"),
                          InformationType::warning);
        continue;
      } // if (InnoCard license expired)
      // check for Elite cardset
      if (   (elite_expiration_date < Date() )
          && (sd.find("Elite") != string::npos) ) {
        ::ui->information(_("License: Elite expired"),
                          InformationType::warning);
        continue;
      } // if (Elite license expired)
#endif
      for (auto const& dd : datadir) {
        auto const directory = dd + "/" + sd;
        // check for the file
        if (File::isfile(directory + "/"
                         + this->value(Type::cards_directory) + "/"
                         + to_string(Card::club) + "/"
                         + to_string(Card::ace)
                         + "." + (*this)(Type::graphic_extension)
                        )) {
          this->string_path_[type] = directory;
          break;
        }
      } // for (dd : datadir)
      if (!this->string_path_[type].empty())
        break;
    } // for (sd : subdir)

    this->update_path(Type::cards_back);
    break;
  } // case Type::cardset: 

  case Type::cards_back: {

    vector<string> const backs // backs to search
      = {this->value(Type::cards_back),
        "penguin",
        "back",
        "back01",
        "deck1"};
    for (auto const& back : backs) {
      auto file = (this->path(Type::cardset) + "/"
                   + this->value(Type::cards_back_directory) + "/"
                   + back + "." + (*this)(Type::graphic_extension));
      // check for the file
      if (!File::isfile(file)) {
        for (auto const& dd : this->data_directories()) {
          file = (dd + "/" + (*this)(Type::cardset) + "/"
                  + this->value(Type::cards_back_directory) + "/"
                  + back + "." + (*this)(Type::graphic_extension));
          if (File::isfile(file)) {
            break;
          }
        } // for (dd : this->data_directiories())
      }

      if (File::isfile(file)) {
        if (back == this->value(Type::cards_back)) {
          this->string_path_[type] = file;
          break;
        } else {
          this->set(Type::cards_back, back);
          break;
        }
      }

    } // for (back : backs)
    break;
  } // case Type::cards_back:

  case Type::iconset: {
    // search a directory
    for (auto const& dd : this->data_directories()) {
      auto const dir = (dd + "/" + (*this)(Type::iconset));
      if (File::isdirectory(dir)) {
        this->string_path_[type] = dir;
        break;
      }
    } // for (dd : this->data_directiories())
    break;
  } // case Type::iconset:

  case Type::background: {
    // search the directory
    for (auto const& dd : this->data_directories()) {
      auto const file = (dd + "/" + (*this)(Type::backgrounds_directory) + "/"
                         + this->value(Type::background)
                         + "." + (*this)(Type::graphic_extension)
                        );
      // check for the file
      if (File::isfile(file)) {
        this->string_path_[type] = file;
        break;
      }
    } // for (dd : this->data_directiories())

    break;
  } // case background:

#ifdef USE_SOUND_COMMAND
  case Type::play_sound_command:
    this->string_path_[type] = (*this)(type);
    break ;
#endif
  case Type::browser_command:
    this->string_path_[type] = (*this)(type);
    break ;
  } // switch(type)

  if (this->string_path_[type].empty())
    this->string_path_[type] = path_bak;
} // void Setting::update_path(Type::String type)

/** prints the settings in the output stream
 **
 ** @param    ostr   ouput stream
 **
 ** @return   output stream
 **/
ostream&
Setting::write(ostream& ostr) const
{
  auto const flags = ostr.flags();
  ostr << std::boolalpha;

  ostr << to_string(Type::name) << " = "
    << this->value(Type::name) << "\n"

    << to_string(Type::language) << " = "
    << this->value(Type::language) << "\n"

    << to_string(Type::sound) << " = "
    << this->value(Type::sound) << "\n"

#ifdef USE_SOUND_COMMAND
    << to_string(Type::play_sound_command) << " = "
    << this->value(Type::play_sound_command) << "\n"
#endif

    << to_string(Type::browser_command) << " = "
    << this->value(Type::browser_command) << "\n"

    << to_string(Type::show_bug_report_button_in_game_finished_window) << " = "
    << this->value(Type::show_bug_report_button_in_game_finished_window) << "\n"

    << to_string(Type::save_bug_reports_on_desktop) << " = "
    << this->value(Type::save_bug_reports_on_desktop) << "\n"

#ifdef USE_THREADS
    << to_string(Type::threads_max) << " = "
    << this->value(Type::threads_max) << "\n"
#endif

    << "\n"
    << "# party settings\n"
    << to_string(Type::additional_party_settings) << " = "
    << this->value(Type::additional_party_settings) << "\n"


    << "\n"
    << "# information and delay\n"

    << to_string(Type::card_play_delay) << " = "
    << ((this->value(Type::card_play_delay) == UINT_MAX)
        ? "-1"
        : String::to_string(this->value(Type::card_play_delay)).c_str())
    << "\n"

    << to_string(Type::automatic_card_suggestion) << " = "
    << this->value(Type::automatic_card_suggestion) << "\n"

    << to_string(Type::announce_swines_automatically) << " = "
    << this->value(Type::announce_swines_automatically) << "\n"

    << to_string(Type::show_if_valid) << " = "
    << this->value(Type::show_if_valid) << "\n"

    << to_string(Type::emphasize_valid_cards) << " = "
    << this->value(Type::emphasize_valid_cards) << "\n"

    << to_string(Type::announce_in_table) << " = "
    << this->value(Type::announce_in_table) << "\n"

    << to_string(Type::show_full_trick_window) << " = "
    << this->value(Type::show_full_trick_window) << "\n"

    << to_string(Type::show_full_trick_window_if_special_points) << " = "
    << this->value(Type::show_full_trick_window_if_special_points) << "\n"

    << to_string(Type::close_full_trick_automatically) << " = "
    << this->value(Type::close_full_trick_automatically) << "\n"

    << to_string(Type::full_trick_close_delay) << " = "
    << this->value(Type::full_trick_close_delay) << "\n"


    << to_string(Type::show_gametype_window) << " = "
    << this->value(Type::show_gametype_window) << "\n"

    << to_string(Type::close_gametype_window_automatically) << " = "
    << this->value(Type::close_gametype_window_automatically) << "\n"

    << to_string(Type::gametype_window_close_delay) << " = "
    << this->value(Type::gametype_window_close_delay) << "\n"


    << to_string(Type::show_marriage_window) << " = "
    << this->value(Type::show_marriage_window) << "\n"

    << to_string(Type::close_marriage_window_automatically) << " = "
    << this->value(Type::close_marriage_window_automatically) << "\n"

    << to_string(Type::marriage_window_close_delay) << " = "
    << this->value(Type::marriage_window_close_delay) << "\n"


    << to_string(Type::show_genscher_window) << " = "
    << this->value(Type::show_genscher_window) << "\n"

    << to_string(Type::close_genscher_window_automatically) << " = "
    << this->value(Type::close_genscher_window_automatically) << "\n"

    << to_string(Type::genscher_window_close_delay) << " = "
    << this->value(Type::genscher_window_close_delay) << "\n"


    << to_string(Type::show_announcement_window) << " = "
    << this->value(Type::show_announcement_window) << "\n"

    << to_string(Type::close_announcement_window_automatically) << " = "
    << this->value(Type::close_announcement_window_automatically) << "\n"

    << to_string(Type::announcement_window_close_delay) << " = "
    << this->value(Type::announcement_window_close_delay) << "\n"


    << to_string(Type::show_swines_window) << " = "
    << this->value(Type::show_swines_window) << "\n"

    << to_string(Type::close_swines_window_automatically) << " = "
    << this->value(Type::close_swines_window_automatically) << "\n"

    << to_string(Type::swines_window_close_delay) << " = "
    << this->value(Type::swines_window_close_delay) << "\n"


    << to_string(Type::show_trickpiles_points) << " = "
    << this->value(Type::show_trickpiles_points) << "\n"

    << to_string(Type::show_known_teams_in_game) << " = "
    << this->value(Type::show_known_teams_in_game) << "\n"

    << to_string(Type::show_soloplayer_in_game) << " = "
    << this->value(Type::show_soloplayer_in_game) << "\n"

    << to_string(Type::show_all_hands) << " = "
    << this->value(Type::show_all_hands) << "\n"

    << to_string(Type::show_ai_information_hands) << " = "
    << this->value(Type::show_ai_information_hands) << "\n"

    << to_string(Type::show_ai_information_teams) << " = "
    << this->value(Type::show_ai_information_teams) << "\n"

    << "\n";

  // so the sorting is saved, even if the mixed modus is aktivated

  if (!this->value(Type::cards_order).sorted()) {
    CardsOrder cards_order_tmp(this->value(Type::cards_order));

    cards_order_tmp.sorted_set(true);

    ostr << to_string(Type::cards_order) << " = "
      << cards_order_tmp << "\n";
  } // if (this->value(Type::cards_order).mixed())

  ostr << to_string(Type::cards_order) << " = "
    << this->value(Type::cards_order) << "\n";

  ostr << "# appearance\n"

    << "#   table\n"

    << to_string(Type::own_hand_on_table_bottom) << " = "
    << this->value(Type::own_hand_on_table_bottom) << "\n"

    << to_string(Type::table_rotation) << " = "
    << this->value(Type::table_rotation) << "\n"

    << to_string(Type::rotate_trick_cards) << " = "
    << this->value(Type::rotate_trick_cards) << "\n"

    << "#   cards\n"

    << to_string(Type::original_cards_size) << " = "
    << this->value(Type::original_cards_size) << "\n"

    << to_string(Type::cards_height) << " = "
    << this->value(Type::cards_height) << "\n"

    << "#   theme\n"

    << to_string(Type::cardset) << " = "
    << this->value(Type::cardset) << "\n"

    << to_string(Type::cards_back) << " = "
    << this->value(Type::cards_back) << "\n"

    << to_string(Type::iconset) << " = "
    << this->value(Type::iconset) << "\n"

    << to_string(Type::background) << " = "
    << this->value(Type::background) << "\n"

    << "#   fonts and colors\n"

    << to_string(Type::name_font) << " = "
    << this->value(Type::name_font) << "\n"

    << to_string(Type::name_font_color) << " = "
    << this->value(Type::name_font_color) << "\n"

    << to_string(Type::name_active_font_color) << " = "
    << this->value(Type::name_active_font_color) << "\n"

    << to_string(Type::name_reservation_font_color) << " = "
    << this->value(Type::name_reservation_font_color) << "\n"

    << to_string(Type::trickpile_points_font) << " = "
    << this->value(Type::trickpile_points_font) << "\n"

    << to_string(Type::trickpile_points_font_color) << " = "
    << this->value(Type::trickpile_points_font_color) << "\n"

    << to_string(Type::poverty_shift_arrow_color) << " = "
    << this->value(Type::poverty_shift_arrow_color) << "\n"

    << "#   splash screen\n"

    << to_string(Type::show_splash_screen) << " = "
    << this->value(Type::show_splash_screen) << "\n"

    << to_string(Type::splash_screen_transparent) << " = "
    << this->value(Type::splash_screen_transparent) << "\n"

    << "# ui\n"

    << to_string(Type::ui_main_window_pos_x) << " = "
    << this->value(Type::ui_main_window_pos_x) << "\n"

    << to_string(Type::ui_main_window_pos_y) << " = "
    << this->value(Type::ui_main_window_pos_y) << "\n"

    << to_string(Type::ui_main_window_width) << " = "
    << this->value(Type::ui_main_window_width) << "\n"

    << to_string(Type::ui_main_window_height) << " = "
    << this->value(Type::ui_main_window_height) << "\n"

    << to_string(Type::ui_main_window_fullscreen) << " = "
    << this->value(Type::ui_main_window_fullscreen) << "\n"
    ;

  ostr << "\n";

  ostr.flags(flags);
  return ostr;
} // ostream& Setting::write(ostream& ostr) const

/** write the setting in the output stream
 **
 ** @param    ostr   output stream
 ** @param    setting   setting to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Setting const& setting)
{
  setting.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, Setting setting)

/** -> result
 **
 ** @param    setting1   first setting that is compared
 ** @param    setting2   second setting that is compared
 **
 ** @return   whether the two settings are equal (according to value)
 **/
bool
operator==(Setting const& setting1, Setting const& setting2)
{
  for (auto t : Setting::Type::bool_list) {
    if (setting1.value(t) != setting2.value(t)) {
      return false;
    }
  }
  for (auto t : Setting::Type::unsigned_list) {
    if (setting1.value(t) != setting2.value(t)) {
      return false;
    }
  }
  for (auto t : Setting::Type::string_list) {
    if (setting1.value(t) != setting2.value(t)) {
      return false;
    }
  }
  for (auto t : Setting::Type::cards_order_list) {
    if (setting1.value(t) != setting2.value(t)) {
      return false;
    }
  }

  return true;
} // bool operator==(Setting setting1, Setting setting2)

/** -> result
 **
 ** @param    setting1   first setting that is compared
 ** @param    setting2   second setting that is compared
 **
 ** @return   whether the two settings are different (according to value)
 **/
bool
operator!=(Setting const& setting1, Setting const& setting2)
{
  return !(setting1 == setting2);
} // bool operator!=(Setting setting1, Setting setting2)
