/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#pragma once

#include "../basetypes.h"

#include "../utils/date.h"

#include "setting_type.h"

class Setting;
extern Setting setting;

/**
 ** the class for all preferences
 **/
class Setting {
  public:
    // cards order
    class CardsOrder;
    // theme (some settings)
    class Theme;

  public:
    // the date of the expiration of the Altenburg cardset
    static Date const altenburg_expiration_date;
    // the date of the expiration of the InnoCard cardset
    static Date const innocard_expiration_date;
    // the date of the expiration of the Elite cardset
    static Date const elite_expiration_date;

  public:
    using Type = SettingType;

  public:
    Setting();
    Setting(Setting const& setting);
    Setting& operator=(Setting const& setting);

    void init();

    Signal<void(int, void const*)>& signal_changed();

    vector<string> const& data_directories() const;
    void set_data_directories();
    void add_data_directory(string const& path);
    ostream& write_data_directories(ostream& ostr,
                                    string const& indent = "") const;
    // a string containing the data directories
    string data_directories_string(string const& indent = "") const;


    // returns, whether the setting is activ or not (or the value)
    bool operator()(Type::Bool type) const;
    unsigned operator()(Type::Unsigned type) const;
    string operator()(Type::String type) const;
    string operator()(Type::StringConst type) const;
    CardsOrder const& operator()(Type::CardsOrder type) const;
    CardsOrder& operator()(Type::CardsOrder type);

    // returns the type with the name
    int type(string const& name) const;
    // returns the type with the translated name
    int type_translated(string const& name) const;
    // returns the type with the translated or normal name
    int type_translated_or_normal(string const& name) const;

    // returns, whether the dependencies of the settings are met
    bool dependencies(Type::Bool type) const;
    bool dependencies(Type::Unsigned type) const;
    bool dependencies(Type::String type) const;

    // returns the setting
    bool value(Type::Bool type) const;
    unsigned value(Type::Unsigned type) const;
    string const& value(Type::String type) const;
    string const& path(Type::String type) const;
    string value(Type::StringConst type) const;
    string path(Type::StringConst type) const;
    CardsOrder const& value(Type::CardsOrder type) const;
    CardsOrder& value(Type::CardsOrder type);

    // the minimum and maximum value
    unsigned min(Type::Unsigned type) const;
    unsigned max(Type::Unsigned type) const;

    // return the themes list
    vector<unique_ptr<Theme>> const& themes() const;
    // return the theme with the given name
    Theme const* theme(string const& name) const;

    // sets the setting
    bool set(string const& type, string const& value);
    void set(Type::Bool type, bool value);
    void set(Type::Bool type, string const& value);
    void set(Type::String type, string const& value);
    void set(Type::Unsigned type, unsigned value);
    void set(Type::Unsigned type, string const& value);
    void set(Type::CardsOrder type,
             CardsOrder const& value);
    void set(Type::CardsOrder type,
             string const& value);
    // set the values from the theme
    void set(Theme const& theme);
    // set the values from the theme 'name'
    bool set_theme(string const& name);

    // returns the position of 'playerno'
    Position position(unsigned playerno) const;

    void load();
    void load(string const& filename,
              bool filename_not_exists_output = true);

    bool save() const;
    bool save(string const& filename) const;

    // check whether a cardset is outdated and replaces it
    void check_for_outdated_cardset();

    // write of the settings
    ostream& write(ostream& ostr) const;

  private:
    // set to the hardcoded values
    void set_to_hardcoded();

    // search a cards back
    void search_cards_back();
    // search a background
    void search_background();

    // update all
    void update_all();
    // create the themes list
    void create_themes_list();
    // update the path
    void update_path(Type::String type);

  private:
    // the settings
    map<Type::Bool, bool> bool_;
    map<Type::Unsigned, unsigned> unsigned_;
    map<Type::String, string> string_;

    // the cardset license (string const setting)
    string cardset_license_ = "";
    // the iconset license (string const setting)
    string iconset_license_ = "";
    // cards order setting
    unique_ptr<CardsOrder> cards_order_;

    // directories for the data
    vector<string> data_directories_;
    // string settings (paths)
    map<Type::String, string> string_path_;
    // themes
    vector<unique_ptr<Theme>> themes_;

    Signal<void(int, void const*)> signal_changed_;
}; // class Setting

// output of the Settings
ostream& operator<<(ostream& ostr, Setting const& setting);

// comparison of two settings (value)
bool operator==(Setting const& setting1, Setting const& setting2);
bool operator!=(Setting const& setting1, Setting const& setting2);

//extern string GPL;
//extern string cardset_license;

#include "setting.cardsorder.h"
#include "setting.theme.h"
