/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"
#include "../utils/file.h"
#include "setting.h"

#ifndef LOCALE_DIRECTORY
#define LOCALE_DIRECTORY "../po"
#endif

/** initializes the translation
 **/
void
translation_init()
{
  setlocale(LC_ALL, "");
  textdomain("FreeDoko");
  vector<string> paths;
  if (getenv("FREEDOKO_LOCALE_DIRECTORY"))
    paths.push_back(getenv("FREEDOKO_LOCALE_DIRECTORY"));
  paths.push_back(File::absolute_path_from_executable_directory(LOCALE_DIRECTORY));
  for (auto const p : ::setting.data_directories()) {
    paths.push_back(p + "/po");
  }
  for (auto const path : paths) {
    if (File::isdirectory(path)) {
      bindtextdomain("FreeDoko", path.c_str());
      break;
    }
  }

  bind_textdomain_codeset("FreeDoko", "utf-8");
} // void translation_init()

/** @return   the locale (LC_MESSAGES)
 **/
string
get_language()
{
#ifdef WINDOWS
  string locale = Reg_read(HKEY_CURRENT_USER,
                           "Software\\FreeDoko",
                           "Language");
  if (locale.empty())
    locale = Reg_read(HKEY_LOCAL_MACHINE,
                      "Software\\FreeDoko",
                      "Language");

  if (locale.empty()) {
    // Ein bisschen kompliziert, aus der Registry die Sprache auszulesen:
    // zuerst wird der Laenderkode gelesen (Deutsch: 407),
    // anschliessend wird geschaut, welche Sprache zu dem Laendercode
    // gehört.
    string const language_code
      = Reg_read(HKEY_CURRENT_USER,
                 "Control Panel\\International",
                 "Locale");

    if (!language_code.empty()) {
      if (string(language_code, language_code.size() - 1)
          == "7") { // deutsch
        locale = "de";
      } else if (string(language_code, language_code.size() - 1)
                 == "9") { // english
        locale = "en";
      } else if (string(language_code, language_code.size() - 1)
                 == "A") { // spanish
        locale = "en";
      } else if (string(language_code, language_code.size() - 1)
                 == "C") { // francais
        locale = "en";
      } else { // unknown language code
        locale = "en";
      } // unknown langugage code
    } // if (!language_code.empty())

  } // if (locale.empty())

  if (locale.empty()) {
    // *** DK
    // I didn't find the language in the registry, so I use this workaround
    if ((Reg_read(HKEY_LOCAL_MACHINE,
                  "Software\\Microsoft\\Windows\\CurrentVersion",
                  "ProgramFilesDir") == string("C:\\Programme"))
        && (Reg_read(HKEY_LOCAL_MACHINE,
                     "Software\\Microsoft\\Windows\\CurrentVersion",
                     "ProgramFilesPath") == string("C:\\Programme")
           )
       )
      locale = "de";
  } // if (locale.empty())
#else
  // non Windows
  string locale = setlocale(LC_MESSAGES, nullptr);
#endif
  return locale;
} // string get_language()

/** set the language
 **
 ** @param    language   new language
 **/
void
set_language(string const language)
{
#ifdef WINDOWS
  SetEnvironmentVariableA("language", language.c_str());
  putenv(("language=" + language).c_str());
#endif
#ifndef WINDOWS
  setenv("language", language.c_str(), 1);
#endif
} // void set_language(string const language)

/** translate the text
 **
 ** @param    text   text to translate
 **
 ** @return   translation of 'text'
 **/
string
gettext(string const text)
{
  if (::debug("gettext")) {
    auto const translation = gettext(text.c_str());
    if (translation == text)
      ::debug.ostr("gettext") << _("Translation missing: %s", text) << '\n'; 
    return translation;
  } else {
    return gettext(text.c_str());
  }
}

/** translate the text
 **
 ** @param    text   text to translate
 **
 ** @return   translation of 'text'
 **/
string
_(string const& text)
{
  return gettext(text);
}

/** replace a substring (only %%)
 **
 ** @param    text   text
 **
 ** @return   text with replacements %% -> %
 **/
string
replace_substrings(string const& text)
{
  auto const pos = text.find('%');
  if (pos == text.npos)
    return text;
  if (pos + 1 == text.npos)
    return text;
  if (text[pos + 1] == '%')
    return (  string(text, 0, pos)
            + replace_substrings(string(text, pos + 2, text.npos)));
  return text;
}

/** replace a substring
 **
 ** @param    text           text with string to replace
 ** @param    param_string   string to insert
 **
 ** @return   text with replacements
 **/
string
replace_substring(string const& text, string const& param_string)
{
  return replace_substring(text, param_string, 's', 't');
} // string replace_substring(string text, string param_string)

/** replace a substring
 **
 ** @param    text        text with string to replace
 ** @param    param_int   integer to insert
 **
 ** @return   text with replacements
 **/
string
replace_substring(string const& text, int const param_int)
{
  return replace_substring(text, std::to_string(param_int), 'i', 'd');
} // string replace_substring(string const& text, int param_int)

/** replace a substring
 **
 ** @param    text             text with string to replace
 ** @param    param_unsigned   unsigned integer to insert
 **
 ** @return   text with replacements
 **/
string
replace_substring(string const& text, uint32_t const param_unsigned)
{
  return replace_substring(text, std::to_string(param_unsigned), 'u');
}

/** replace a substring
 **
 ** @param    text             text with string to replace
 ** @param    param_unsigned   unsigned integer to insert
 **
 ** @return   text with replacements
 **/
string
replace_substring(string const& text, uint64_t const param_unsigned)
{
  return replace_substring(text, std::to_string(param_unsigned), 'u');
}

/** replace a substring
 **
 ** @param    text          text with string to replace
 ** @param    replacement   replacement
 ** @param    c1            character for identifying the part to replace
 ** @param    c2            optional second character for identifying the part to replace
 **
 ** @return   text with replacements
 **/
string
replace_substring(string const& text, string const& replacement,
                  char c1, char c2)
{
  // start of a '%'-substring
  auto const pos = text.find('%');
  DEBUG_ASSERTION(   (pos != text.npos)
                  && (pos + 1 != text.npos),
                  "replace_substring(text, replacement, " << c1 << ", " << c2 << ")\n"
                  "  Found no %, but should replace a string '" << replacement << "'. Text is: "
                  << text);
  // end of the '%'-substring
  DEBUG_ASSERTION(   (text[pos + 1]) == c1
                  || (text[pos + 1]) == c2,
                  "replace_substring(text, replacement, " << c1 << ", " << c2 << ")\n"
                  "  wrong type at position " << pos + 1 << ":\n"
                  "  The text says '" << text[pos + 1] << "' but the code says '" << c1 << "'"
                  << (c2 ? (" or '" + string(1, c2) + "'") : ""s)
                  << ".\n"
                  "  Text: " << text);


  return (  replace_substrings(string(text, 0, pos))
          + replacement
          + replace_substrings(string(text, pos + 2, text.npos)));
} // string replace_substring(string const& text, string const& replacement, char c1, char c2 = '\0')
