/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "auto_bug_report.h"
#include "bug_report.h"

#include "../party/party.h"
#include "../game/game_summary.h"
#include "../game/game.h"
#include "../card/trick.h"

#include "../utils/string.h"

namespace OS_NS {

/** constructor
 **/
AutoBugReport::AutoBugReport()
{
  Game::signal_open().connect(*this, &AutoBugReport::game_open, this->disconnector_);
}

/** destructor
 **/
AutoBugReport::~AutoBugReport() = default;

/** the game is opened
 **
 ** @param    game   game that is opened
 **/
void
AutoBugReport::game_open(Game const& game)
{
  {
    auto& game = ::party->game();
    game.tricks().signal_trick_full().connect(*this, &AutoBugReport::trick_full, this->disconnector_);
    game.signal_finish().connect(*this, &AutoBugReport::game_finish, this->disconnector_);
  }
} // void AutoBugReport::game_open(Game game)

/** the trick is full
 ** create automatic bug reports
 **
 ** @param    trick   current trick
 **/
void
AutoBugReport::trick_full(Trick const& trick)
{
  for (unsigned c = 0; c < trick.actcardno(); ++c) {
    auto const& card = trick.card(c);
    auto const& player = trick.player_of_card(c);

    if (player != trick.winnerplayer()) {
      if (card.isdolle()) {
        this->create_report("dolle lost:\n"
                            + String::to_string(trick)
                           );
      }
      if (   card.isfox()
          && (player.team() != trick.winnerplayer().team()) ) {
        this->create_report("fox lost:\n"
                            + String::to_string(trick)
                           );
      }
      if (card.isswine()) {
        this->create_report("swine lost:\n"
                            + String::to_string(trick)
                           );
      }
    } // if (player != trick.winnerplayer())
  } // for (c)
} // void AutoBugReport::trick_full(Trick trick)

/** the game is finished
 ** create automatic bug reports
 **/
void
AutoBugReport::game_finish()
{
  auto const& game_summary = ::party->last_game_summary();
  { // lost poverty
    if (   (game_summary.game_type() == GameType::poverty)
        && (game_summary.winnerteam() != Team::re) )
      this->create_report("poverty lost by re"
                          " ("
                          + String::to_string(game_summary.trick_points(Team::re))
                          + ")");
  } // lost poverty
  { // lost solo
    if (   is_solo(game_summary.game_type())
        && (game_summary.winnerteam() != Team::re)
        && (game_summary.trick_points(Team::re) < 90) )
      this->create_report(to_string(game_summary.game_type()) + " lost by re"
                          + " ("
                          + String::to_string(game_summary.trick_points(Team::re))
                          + ")");
  } // lost solo

  { // bad announcement
    if (   is_real(game_summary.highest_announcement_re())
        && (game_summary.trick_points(Team::re)
            < (needed_points(game_summary.highest_announcement_re())
               - 2 * 30) )
       )
      this->create_report("bad annoucement for re: "
                          + to_string(game_summary.highest_announcement_re())
                          + " ("
                          + String::to_string(game_summary.trick_points(Team::contra))
                          + ")");
    if (   (game_summary.highest_announcement_re()
            == Announcement::noannouncement)
        && (game_summary.trick_points(Team::re)
            > (needed_points(Announcement::no120))
            + 2 * 30)
       )
      this->create_report("no annoucement for re: "
                          + String::to_string(game_summary.trick_points(Team::contra))
                          + " points for contra");

    if (   is_real(game_summary.highest_announcement_contra())
        && (game_summary.trick_points(Team::contra)
            < (needed_points(game_summary.highest_announcement_contra())
               - 2 * 30) )
       )
      this->create_report("bad annoucement for contra: "
                          + to_string(game_summary.highest_announcement_contra())
                          + " ("
                          + String::to_string(game_summary.trick_points(Team::re))
                          + ")");
    if (   (game_summary.highest_announcement_contra()
            == Announcement::noannouncement)
        && (game_summary.trick_points(Team::contra)
            > (needed_points(Announcement::no120))
            + 2 * 30)
       )
      this->create_report("no annoucement for contra: "
                          + String::to_string(game_summary.trick_points(Team::re))
                          + " points for re");

    if (   (game_summary.game_type() == GameType::poverty)
        && (game_summary.winnerteam() != Team::re) ) {
    } // if (bad announcement)
  } // bad announcement
} // void AutoBugReport::game_finish()

/** create the report
 **/
void
AutoBugReport::create_report(string const& text) const
{
  cout << "automatic bug report creation\n"
    << "  " << ::bug_report->filename() << '\n'
    << "  " << text << '\n'
    << endl;

  ::bug_report->report(text, "automatic");
} // void AutoBugReport::create_report(string text) const

} // namespace OS_NS
