/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "../basetypes.h"
#include "../utils/file.h"
#include "../utils/string.h"

#include "bug_report.h"

#include <fstream>

#include "../party/party.h"
#include "../party/rule.h"
#include "../game/game.h"
#include "../game/gameplay.h"
#include "../game/game_summary.h"
#include "../game/gameplay_action.h"
#include "../player/aiconfig.h"
#include "../card/trick.h"
#include "../misc/setting.h"
#include "../ui/ui.h"

#ifdef WINDOWS
// for 'mkdir'
#include <io.h>
#else
// for 'mkdir'
#include <sys/stat.h>
#endif
#include <ctime>

// the bug report
unique_ptr<OS_NS::BugReport> bug_report;

namespace OS_NS {

string const bug_reports_directory = "FreeDoko.BugReports";
string const bug_report_file = "BugReport.FreeDoko";

/** create a report
 **
 ** @param    message   message of the user
 **
 ** @return   the filename in which the report is written
 **/
string
  BugReport::report(string const& message)
  {
    return this->report(message, ".");
  } // string BugReport::report(string message = "")

/** create a report
 **
 ** @param    message   message of the user
 ** @param    subdir    sub directory to save into
 **
 ** @return   the filename in which the report is written
 **/
string
  BugReport::report(string const& message, string const& subdir)
  {
    unique_ptr<ofstream> ofstr_ptr;
    std::ostream* ofstr = nullptr;
    string filename;
    { // open a bug report file

      // Remark: could use a stack here
      vector<string> directories = {
        this->directory(),
        ::setting(Setting::Type::private_data_directory)
          + "/" + bug_reports_directory,
        ::setting(Setting::Type::private_data_directory),
        "./" + bug_reports_directory,
        "."
      };

      for (auto const& dir : directories) {
        auto const directory = ((subdir.empty() || (subdir == "."))
                                ? dir
                                : dir + "/" + subdir);
        bool created_dir = false;
        if (!File::isdirectory(directory)) {
#ifdef WINDOWS
          mkdir(directory.c_str());
#else
          mkdir(directory.c_str(), 00777);
#endif
          created_dir = true;
        } // if (!File::isdirectory(this->directory()))

        filename = directory + "/" + this->filename(directory);
        ofstr_ptr = make_unique<ofstream>(filename.c_str(), std::ios::app);
        ofstr = ofstr_ptr.get();
        if (!ofstr_ptr->fail()) {
          if (created_dir)
            ::ui->information(_("BugReport::created directory '%s'", directory),
                              InformationType::normal);
        } else { // if !(!ofstr_ptr->fail())
          ::ui->information(_("BugReport::could not save at '%s', trying '%s'.",
                                          filename.c_str(),
                                          (dir + "/" + subdir + "/" + this->filename(dir + "/" + subdir)
                                          )),
                            InformationType::normal);
        } // if !(!ofstr->fail())
        if (!ofstr->fail())
          break;
      } // for (dir \in directories)
      if (!ofstr || ofstr->fail()) {
        cerr << _("BugReport::could not save")
          << endl;
        cerr << _("Press <Return> to display the bug report.")
          << endl;
        std::cin.get();
        ofstr = &cerr;
        filename = "standard error output";
      } // if (ofstr->fail())
    } // open a bug report file

    const time_t time_now = time(nullptr);

    string const bug_report_output = this->ostr_->str();

    { // write the version, system, time and message
      *ofstr << "# FreeDoko Bugreport\n"
        << '\n'
        << "version: " << *::version
        << '\n'
        << "compiled: " << __DATE__ << ", " << __TIME__ << '\n'
#ifdef WINDOWS
        << "system: Windows\n"
#endif
#ifdef LINUX
        << "system: Linux\n"
#endif
#ifdef HPUX
        << "system: HPUX\n"
#endif
        << "time: " << asctime(localtime(&time_now)) << "\n"
        << "language: " << ::setting(Setting::Type::language) << "\n";

      if (   (::game_status & GameStatus::game)
          && (::game_status != GameStatus::game_new))
        *ofstr << "trick: " << ::party->game().tricks().current_no() << '\n';

      *ofstr << '\n'
        << "/---------\\\n"
        << "| message |\n"
        << "\\---------/\n"
        << message << "\n\n"
        << "/--------\\\n"
        << "| output |\n"
        << "\\--------/\n";
    } // write the version, system, time and message
    if (bug_report_output.empty())
      *ofstr << "finished\n";
    if (!bug_report_output.empty()) {
      *ofstr << bug_report_output;
      if (   (::game_status == GameStatus::game_reservation)
          || (::game_status == GameStatus::game_poverty_shift) ) {
        *ofstr << '\n'
          << "hands\n"
          << "{\n";
        auto const& game = ::party->game();
        for (auto const& player : game.players()) {
          *ofstr << "Player " << player.no()
            << " (" << player.name() << ")\n"
            << player.hand()
            << '\n';
        } // for (player)
        *ofstr << "}\n";
      } // if (::game_status == game_poverty_shift or game_reservation)
      { // write the gameplay actions
        *ofstr << '\n'
          << "gameplay actions\n"
          << "{\n";
        unsigned t = 0;
        for (auto const& a : this->actions_) {
          if (a->type() == GameplayActions::Type::trick_open) {
            t += 1;
            *ofstr << "  # trick " << t << '\n';
          } else {
            *ofstr << "  " << *a << '\n';
          }
        }
        *ofstr << "}\n";
      } // write the gameplay actions

      { // write the game summary
        if (this->game_summary_) {

          *ofstr << '\n'
            << "game summary\n"
            << "{\n"
            << *this->game_summary_
            << "}\n";

          { // special points
            *ofstr << "special points\n"
              << "{\n"
              << "\tteam\tvalue\tdescription:\n";
            for (auto const& special_point
                 : this->game_summary_->specialpoints()) {
              *ofstr << '\t';
              *ofstr << special_point.team;
              // extra rule
              if ((special_point.type == Specialpoint::Type::no120_said)
                  && (::party->rule()(Rule::Type::announcement_re_doubles)))
                *ofstr << "\t" << "*2";
              else
                *ofstr << "\t"
                  << value(special_point.type);
              *ofstr << "\t" << to_string(special_point.type)
                << '\n';
            } // for (special_point)
            *ofstr << "\t= " << this->game_summary_->points() << '\n';

            *ofstr << "}\n";

          } // special points
        } // if (this->game_summary_)
      } // write the game summary
    } // if (!bug_report_output.empty())

    // close the file stream
    *ofstr << flush;


    // I don't know, how to reinitialize 'ostrstream'
    this->ostr_ = make_unique<ostringstream>();

    // copy the old data
    *(this->ostr_) << bug_report_output;

    // one more bug report written
    this->written_ += 1;

    return filename;
  } // virtual string BugReport::report(string message, string subdir)

/** Constructor
 **/
BugReport::BugReport() :
  ostr_(make_unique<ostringstream>())
{
  Game::signal_open().connect(*this, &BugReport::game_open, this->disconnector_);
} // BugReport::BugReport()

/** Destructor
 **/
BugReport::~BugReport() = default;

/** the game is opened
 **
 ** @param    game   opened game
 **/
void
BugReport::game_open(Game const& game)
{
  this->new_game();
  {
    auto& game = ::party->game();
    game.signal_redistribute().connect(*this, &BugReport::new_game, this->disconnector_);
    game.signal_start().connect(*this, &BugReport::game_start, this->disconnector_);
    game.signal_finish().connect(*this, &BugReport::game_finish, this->disconnector_);
    game.signal_gameplay_action().connect(*this, &BugReport::gameplay_action, this->disconnector_);
  }
} // void BugReport::game_open(Game game)

/** a new game has been initialized
 **
 ** @param    game   game
 **/
void
BugReport::new_game()
{
  this->clear();

  auto const& game = ::party->game();
  this->seed_ = game.seed();

  *this->ostr_ << "seed: " << game.seed() << '\n'
    << "startplayer: " << game.startplayer().no() << '\n'
    << '\n';
  {
#ifdef POSTPONED
    // ToDo: include
    auto const write_database_bak = Player::write_database;
    Player::write_database = false;
    auto const write_standard_type_bak = Aiconfig::write_standard_type;
    Aiconfig::write_standard_type = false;
#endif

    *this->ostr_ << "party\n"
      << "{\n"
      << game.party()
      << "}\n";

#ifdef POSTPONED
    // ToDo: include
    Aiconfig::write_standard_type = write_standard_type_bak;
    Player::write_database = write_database_bak;
#endif
  }
} // void BugReport::new_game()

/** the game is started
 **
 ** @param    game  the game
 **
 ** @todo      poverty
 **/
void
BugReport::game_start()
{
  auto const& game = ::party->game();
  { // write the game type
    *(this->ostr_) << '\n'
      << "gametype: " << game.type()
      << '\n';

    switch (game.type()) {
    case GameType::normal:
      break;

    case GameType::marriage_solo:
    case GameType::marriage: {
      *(this->ostr_)
        << "selector: " << to_string(game.marriage().selector()) << '\n'
        << "bride: player "
        << game.players().soloplayer().no()
        << '\n';
    } break;

    case GameType::poverty: {
      // ToDo
    } break;

    case GameType::thrown_nines:
    case GameType::thrown_kings:
    case GameType::thrown_nines_and_kings:
    case GameType::thrown_richness:
    case GameType::fox_highest_trump:
    case GameType::redistribute:
    case GameType::genscher:
    case GameType::marriage_silent:
    case GameType::solo_meatless:
    case GameType::solo_jack:
    case GameType::solo_queen:
    case GameType::solo_king:
    case GameType::solo_queen_jack:
    case GameType::solo_king_jack:
    case GameType::solo_king_queen:
    case GameType::solo_koehler:
    case GameType::solo_club:
    case GameType::solo_heart:
    case GameType::solo_spade:
    case GameType::solo_diamond: {
      *(this->ostr_) << "soloplayer: player "
        << game.players().soloplayer().no()
        << '\n';
      break;
    }
    } // switch (game.type())
  } // write the game type

  { // write the hands
    *(this->ostr_) <<'\n'
      <<"hands\n"
      << "{\n";

    for (auto const& player : game.players()) {
      *(this->ostr_) << "player " << player.no()
        << " (" << player.name()
        << ")\n"
        << player.hand()
        << '\n';
    }

    *(this->ostr_) << "}\n";
  } // write the hands
} // void BugReport::game_start()

/** the game is finished
 **
 ** @param    game  the game
 **/
void
BugReport::game_finish()
{ 
  this->game_summary_ = make_unique<GameSummary>(::party->last_game_summary());
} // void BugReport::game_finish()

/** a gameplay action has happened
 **
 ** @param    action    gameplay action
 **/
void
BugReport::gameplay_action(GameplayAction const& action)
{
  // do not save all actions
  switch (action.type()) {
  case GameplayActions::Type::poverty_denied_by_all:
  case GameplayActions::Type::trick_taken:
  case GameplayActions::Type::check:
  case GameplayActions::Type::print:
    // do not save
    return ;

  case GameplayActions::Type::reservation:
  case GameplayActions::Type::poverty_shift:
  case GameplayActions::Type::poverty_accepted:
  case GameplayActions::Type::poverty_returned:
  case GameplayActions::Type::poverty_denied:
  case GameplayActions::Type::trick_open:
  case GameplayActions::Type::card_played:
  case GameplayActions::Type::announcement:
  case GameplayActions::Type::swines:
  case GameplayActions::Type::hyperswines:
  case GameplayActions::Type::marriage:
  case GameplayActions::Type::genscher:
  case GameplayActions::Type::trick_full:
    // do save
    break;
  } // switch (action.type)

  this->actions_.push_back(action.clone());

#ifdef OUTDATED
  if (dynamic_cast<GameplayAction::Player1 const*>(&action))
    if (this->game().player(static_cast<GameplayAction::Player1 const&>(action).player).type()
        == Player::Type::human)
      this->human_actions_.push_back(action.clone());
#endif
} // void BugReport::gameplay_action(GameplayAction action)

/** clear the bug report stream and make a backup before
 **/
void
BugReport::clear()
{
  this->seed_ = UINT_MAX;
  this->ostr_ = make_unique<ostringstream>();

  this->actions_.clear();
  this->human_actions_.clear();

  this->game_summary_.reset();
} // virtual void BugReport::clear()

/** @return   the directory to save the bug report in
 **/
string
BugReport::directory() const
{
  if (::setting(Setting::Type::save_bug_reports_on_desktop)) {
    return (File::desktop_directory()
            + "/" + bug_reports_directory);
  } // if (::setting(Setting::Type::save_bug_reports_on_desktop))

  return (::setting(Setting::Type::private_data_directory)
          + "/" + bug_reports_directory);
} // virtual void BugReport::directory()

/** @return   the filename (bug_report_file.`seed`)
 **/
string
BugReport::filename() const
{
  return this->filename(this->directory());
} // virtual void BugReport::filename()

/** -> result
 **
 ** @param    directory   the directory to save the bug report in
 **
 ** @return   the filename (`seed`.bug_report_file)
 **/
string
BugReport::filename(string const directory) const
{
  auto const& game = ::party->game();
  // The counter 'written_' is used so that multiple creating of
  // a bug report with moving the previous one does lead to different names.
  string filename = (((::game_status >= GameStatus::game_init)
                      ? String::to_string(game.seed(), 6, '0')
                      : string("party")
                     )
                     + ( (this->written_ > 0)
                        ? "_" + String::to_string(this->written_)
                        : "")
                     + "." + bug_report_file);

  // Check whether the file does not exists, yet.
  // If the file does already exists,
  // append a number and search the first which does not exists.
  for (unsigned n = this->written_;
       File::isfile(directory + "/" + filename);
       n += 1,
       filename = (((::game_status >= GameStatus::game_init)
                    ? String::to_string(game.seed(), 6, '0')
                    : string("party")
                   ) + "_" + String::to_string(n)
                   + "." + bug_report_file))
    ;

  return filename;
} // void BugReport::filename(string directory)

} // namespace OS_NS
