/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "bug_report_replay.h"

#include "../player/ai/ai.h"
#include "../party/party.h"
#include "../game/game.h"
#include "../game/gameplay_actions.h"
#include "../game/gameplay_actions/check.h"
#include "../card/trick.h"

#include "../ui/ui.h"

#include "../misc/setting.h"
#include "../misc/references_check.h"

#include "../utils/string.h"
#include "../utils/file.h"

#include <sstream>
#include <chrono>


// There can only be one 'bug report replay' instance.
// If a new is created, the old is removed.
unique_ptr<OS_NS::BugReportReplay> bug_report_replay;


namespace OS_NS {

Signal<void(BugReportReplay&)> BugReportReplay::signal_open_;

// returns the value from the line 'line'
static string get_value(string const& line);
// returns the keyword from the line 'line'
static string get_keyword(string const& line);


/** -> result
 **
 ** @param    line   the line
 **
 ** @return   the value from the given line
 **/
string
  get_value(string const& line)
  {
    if (line.empty())
      return "";

    string value = line;
    String::word_first_remove_with_blanks(value);

    return value;
  } // static string get_value(string line)

/** -> result
 **
 ** @param    line   the line
 **
 ** @return   the keyword from the given line
 **/
string
  get_keyword(string const& line)
  {
    if (line.empty())
      return "";

    string const word
      = (  (std::find(line.begin(), line.end(), ':')
            != line.end())
         ? string(line.begin(),
                  std::find(line.begin(), line.end(), ':'))
         : line
        );

    return word;
  } // static string get_keyword(string line)

/** Constructor
 ** (to be called from a child class)
 **
 ** @param    filename   the file with the bug report
 **/
BugReportReplay::BugReportReplay(string const& filename,
                                 unsigned verbose) :
  filename_(filename),
  verbose_(verbose)
{
  if (!File::isfile(filename)) {
    cerr << "Selected bug report '" << filename << "' is no file." << endl;
    return ;
  } // if (!File::isfile(filename))

  try {
    this->read_file();
  } catch (ReadException const& read_exception) {
    return ;
  }

  this->init();

  this->print_header();
} // BugReportReplay::BugReportReplay(string filename)

/** Destructor
 **/
BugReportReplay::~BugReportReplay()
{
  this->signal_close()();
} // BugReportReplay::~BugReportReplay()

/** @return   signal when the party is opened
 **/
Signal<void(BugReportReplay&)>&
BugReportReplay::signal_open()
{ return BugReportReplay::signal_open_; }

/** @return   signal: the party is opened
 **/
Signal<void()>&
BugReportReplay::signal_close()
{ return this->signal_close_; }


/** initializes the bug report
 **/
void
BugReportReplay::init()
{
  Party::signal_open().connect(*this, &BugReportReplay::party_open, this->disconnector_);
  Game::signal_open().connect(*this, &BugReportReplay::game_open, this->disconnector_);

  ::setting.set(Setting::Type::save_party_changes, false);
  ::setting.set(Setting::Type::additional_party_settings, true);
  this->check_action_ = false;

  if (::game_status == GameStatus::party_new) {
    this->BugReportReplay::party_open(*::party);
    this->BugReportReplay::party_get_settings();
  }

#ifdef POSTPONED
  // is called in 'party_open' so that this party is set
  // (-> UI_GTKMM::BugReportReplay::update_info)
  ::ui->bug_report_replay_open(*this);
#endif
} // void BugReportReplay::init()

/** write the data of this bug report into 'ostr'
 **
 ** @param    ostr   output stream
 **/
void
BugReportReplay::write(ostream& ostr) const
{
  ostr << "file: "      << this->filename_ << '\n';
  if (this->version_)
    ostr << "version: " << *this->version_ << '\n';
  else
    ostr << "version: -\n";
  ostr << "compiled: "  << this->compiled_ << '\n';
  ostr << "system: "    << this->system_ << '\n';
  ostr << "time: "      << this->time_ << '\n';
  ostr << "language: "  << this->language_ << '\n';
  ostr << "trickno: "   << this->trickno_ << '\n';

  ostr << '\n';

  ostr << "message\n"
    << "{\n"
    << this->message_
    << "}\n";

  ostr << '\n';

  ostr << "seed: "              << this->seed_ << '\n';
  ostr << "startplayer_no: "    << this->startplayer_no_ << '\n';
  ostr << "game_type: "         << this->game_type_ << '\n';
  ostr << "marriage_selector: " << this->marriage_selector_ << '\n';
  ostr << "soloplayer_no: "     << this->soloplayer_no_ << '\n';

  ostr << '\n';

  ostr << "rule\n"
    << "{\n"
    << this->rule_
    << "}\n";

  ostr << '\n';

  ostr << "persons\n"
    << "{\n";
  for (auto const& p : this->persons_)
    ostr << "{\n" << *p << "}\n";
  ostr << "}";

  ostr << '\n';

  if (!this->poverty_cards_shifted_.empty()) {
    ostr << "poverty cards shifted\n"
      << "{\n";
    for (auto const c : this->poverty_cards_shifted_)
      ostr << c << '\n';
    ostr << "}\n";
  } else {
    ostr << "poverty cards shifted: -\n";
  }

  if (!this->poverty_cards_returned_.empty()) {
    ostr << "poverty cards returned\n"
      << "{\n";
    for (auto const c : this->poverty_cards_returned_)
      ostr << c << '\n';
    ostr << "}\n";
  } else {
    ostr << "poverty cards returned: -\n";
  }

  ostr << '\n';

  ostr << "hands\n"
    << "{\n";
  for (auto const& h : this->hands_)
    ostr << "{\n" << h << "}\n";
  ostr << "}\n";

  ostr << '\n';

  ostr << "swines: ";
  if (this->swines_player_no_ == UINT_MAX)
    ostr << '-';
  else
    ostr << this->swines_player_no_;
  ostr << '\n';
  ostr << "hyperswines: ";
  if (this->hyperswines_player_no_ == UINT_MAX)
    ostr << '-';
  else
    ostr << this->hyperswines_player_no_;
  ostr << '\n';


  ostr << '\n';

  ostr << "game actions\n"
    << "{\n";
  for (auto const& a : this->actions_)
    ostr << *a << '\n';
  ostr << "}\n";

  ostr << '\n';

  ostr << "full tricks\n"
    << "{\n";
  for (auto const& t : this->full_tricks_)
    ostr << t << '\n';
  ostr << "}\n";

  ostr << '\n';

  ostr << "game summary\n"
    << "{\n";
  if (this->game_summary_)
    ostr << *this->game_summary_;
  ostr << "}\n";

  ostr << '\n';

  ostr << "current hands\n"
    << "{\n";
  for (auto const& h : this->current_hands_)
    ostr << "{\n" << h << "}\n";
  ostr << "}\n";

  ostr << '\n';

  ostr << "current trick\n"
    << this->current_trick_;

  ostr << '\n';

  ostr << "human actions\n"
    << "{\n";
  for (auto const& a : this->human_actions_)
    ostr << *a << '\n';
  ostr << "}\n";
} // void BugReportReplay::write(ostream& ostr) const

/** test, whether the line is the expected
 **
 ** @param    read_line       line that was read
 ** @param    expected_line   line that is read
 **
 ** @return   whether the line is the expected
 **/
bool
BugReportReplay::expect_line(string const& read_line,
                             string const& expected_line)
{
  if (read_line != expected_line) {
    cerr << "BugReportReplay:\n"
      << "  expected line '" << expected_line << "', "
      << "got: '" << read_line << "'"
      << "\n";

    this->mismatch();
    return false;
  } // if (read_line != expected_line)

  return true;
} // bool BugReportReplay::expect_line(string read_line, string expected_line)

/** test, whether the keyword is the expected
 **
 ** @param    read_keyword       keyword that was read
 ** @param    expected_keyword   keyword that is read
 **
 ** @return   whether the keyword is the expected
 **/
bool
BugReportReplay::expect_keyword(string const& read_keyword,
                                string const& expected_keyword)
{
  if (read_keyword != expected_keyword) {
    cerr << "BugReportReplay:\n"
      << "  expected keyword '" << expected_keyword << "', "
      << "got: '" << read_keyword << "'"
      << "\n";

    this->mismatch();
    return false;
  } // if (read_keyword != expected_keyword)

  return true;
} // bool BugReportReplay::expect_keyword(string read_keyword, string expected_line)

/** reads the whole file
 **
 ** @todo   poverty
 ** @todo   game summary
 **/
void
BugReportReplay::read_file()
{
  cout << "Loading bug report " << this->filename() << endl;
  ifstream istr(this->filename().c_str());
  string line;    // the line read
  string keyword; // the keyword read
  string value;   // the value read

#define GETLINE \
  do { \
    if (istr.eof()) { \
      throw EOF; \
    } \
    std::getline(istr, line);  \
    String::remove_blanks(line); \
    if ( (line.size() > 0) && (*(line.end() - 1) == '\r') ) \
    line.erase(line.end() - 1); \
  } while (line.empty() && istr.good()) ; \
  keyword = get_keyword(line); \
  value = get_value(line); \

  // reads the variable from 'value'
#define READ_FROM_VALUE(var)  \
  { \
    value = get_value(line); \
    istringstream istr(value); \
    istr >> var; \
  }

  // expect the keyword
  // if it is not valid, return from the function
#define EXPECT_LINE(expected_line) \
  if (true) { \
    if (!this->expect_line(line, expected_line)) \
    throw ReadException("expected line '" + string(line) + "'") ; \
  } else
#if 0
  ;
#endif

#define EXPECT_KEYWORD(expected_keyword) \
  if (true) { \
    if (!this->expect_keyword(keyword, expected_keyword)) \
    throw ReadException("expected keyword '" + string(keyword) + "'") ; \
  } else
#if 0
  ;
#endif

  try {
    this->human_finished_ = true;
    { // read the first lines (including message)
      // the first line
      GETLINE;
      // Byte Order Mark for UTF-8
      string bom = "...";
      bom[0] = 0xef;
      bom[1] = 0xbb;
      bom[2] = 0xbf;

      if (   (line != "# FreeDoko Bugreport")
          && (line != bom + "# FreeDoko Bugreport")
          && (line != "# FreeDoko Reference")
          && (line != bom + "# FreeDoko Reference") ) {
        cerr << "BugReportReplay::read_first_lines()\n"
          << "  The file is no bug report! First line does not match. Expecting\n"
          << "# FreeDoko Bugreport\n"
          << "  or\n"
          << "# FreeDoko Reference\n"
          << "  but got\n"
          << line
          << endl;
        this->mismatch();
        return ;
      } // if (line != "# FreeDoko Bugreport")

      // read and print the next informations
      // Example:
      // Version: 0.6.7b  (2004-12-19)
      // Compiled: Dec 19 2004, 19:36:18
      // System: Windows
      // Time: Fri Feb 25 22:55:22 2005
      //
      // Language: de
      // Trick: 9
      do { // while(!line.empty());
        GETLINE;

        if (line == "") {
        } else if (line == "/---------\\") {
          break;
        } else if (keyword == "version") {
          auto version = Version::create(value);
          if (version)
            this->version_ = make_unique<Version>(*version);
          else
            this->version_ = {};
          if (!this->version_) {
            cerr << "BugReportReplay: "
              << "could not read version: '" << value << "'\n"
              << "ignoring it." << endl;

          }
          if (   this->version_
              && (*this->version_ <= Version(0, 6, 6)) ) {
            cerr << "BugReportReplay: Too old version "
              << *this->version_
              << " < 0.6.6"
              << endl;
#ifndef DKNOF
            exit(EXIT_SUCCESS);
#endif
          }
        } else if (keyword == "compiled") {
          this->compiled_  = value;
        } else if (keyword == "system") {
          this->system_    = value;
        } else if (keyword == "time") {
          this->time_      = value;
        } else if (keyword == "language") {
          this->language_  = value;
        } else if (keyword == "trick") {
          READ_FROM_VALUE(this->trickno_);
        } else { // if !(keyword == ...)
          cerr << "BugReportReplay: "
            << "unknown keyword '" << keyword << "'\n"
            << "ignoring it." << endl;
        } // if !(keyword == ...)

      } while(line != "/---------\\");

      { // read the message

        // the header exists of three lines -- ignore them
        GETLINE;
        GETLINE;

        this->message_.clear();
        do { // while (line != "/---------\\")
          GETLINE;
          if (line != "/--------\\")
            this->message_ += "  " + line + "\n";

          if (istr.fail()) {
            cerr << "BugReportReplay::read_first_lines()\n"
              << "  (unexpected) Error reading the first lines.\n"
              << "Aborting."
              << endl;
            exit(EXIT_FAILURE);
          } // if (istr.fail())
        } while(line != "/--------\\") ;

        // the foot exists of three lines -- ignore them (one is already ignored)
        GETLINE;
        GETLINE;
      } // read the message
      GETLINE;
    } // read the first lines
    if (line == "finished") {
      throw EOF;
    } // if (line == "finished")
    { // seed, startplayer
      EXPECT_KEYWORD("seed");
      READ_FROM_VALUE(this->seed_);

      GETLINE;
      EXPECT_KEYWORD("startplayer");
      READ_FROM_VALUE(this->startplayer_no_);
    } // seed, startplayer
    { // rules, players
      GETLINE;

      if (line == "party") {
        EXPECT_LINE("party");
        // a '{'
        GETLINE;
        EXPECT_LINE("{");
        ::party->read(istr);
        this->persons_.clear();
        for (auto const& p : ::party->players())
          this->persons_.push_back(p.clone());
        GETLINE;
        this->rule_ = ::party->rule();
        EXPECT_LINE("}");
#ifdef WORKAROUND
        // ToDo: this shouldn't be necessary
        ::party->set_seed(this->seed_);
#endif
      } else { // if !(line == "party")
        EXPECT_LINE("rules");
        this->rule_.read(istr);
        // set the rules here, so that the default ai types see the setting of with_nines
        this->set_party_rule();

        GETLINE;
        EXPECT_LINE("players");

        // a '{'
        GETLINE;
        EXPECT_LINE("{");

        while (istr.peek() != '}') {
          this->persons_.push_back(unique_ptr<Player>(Player::create(istr)));
          if (   !istr.good()
              || !this->persons_.back()) {
            cerr << "BugReportReplay::read_file()\n"
              << "  could not load player.\n"
              << "Aborting."
              << endl;
            exit(EXIT_FAILURE);
          } // if (!this->players.back())
          while ((istr.peek() == '\n')
                 || (istr.peek() == '\r'))
            istr.get();
        } // while (istr.peek() != '}')

        GETLINE;
        EXPECT_LINE("}");
      } // if !(line == "party")
    } // players, players
    { // gametype
      GETLINE;
      if (keyword == "hands") {
        this->game_type_ = GameType::normal;
      } else { // if !(keyword == "hands")
        EXPECT_KEYWORD("gametype");
        String::word_first_remove_with_blanks(line);
        this->game_type_ = game_type_from_string(line);
        { // special cases
          if (   is_solo(this->game_type_)
              || (this->game_type_ == GameType::thrown_nines)
              || (this->game_type_ == GameType::thrown_kings)
              || (this->game_type_ == GameType::thrown_nines_and_kings)
              || (this->game_type_ == GameType::thrown_richness)
              || (this->game_type_ == GameType::fox_highest_trump) ) {
            GETLINE;
            EXPECT_KEYWORD("soloplayer");
            String::word_first_remove_with_blanks(line);
            String::word_first_remove_with_blanks(line);
            this->soloplayer_no_
              = static_cast<unsigned>(stoi(line));
          }
          if (this->game_type_ == GameType::marriage) {
            // read the selector and the bride
            GETLINE;
            EXPECT_KEYWORD("selector");
            String::word_first_remove_with_blanks(line);
            this->marriage_selector_ = marriage_selector_from_string(line);

            GETLINE;
            EXPECT_KEYWORD("bride");
            String::word_first_remove_with_blanks(line);
            String::word_first_remove_with_blanks(line);
            this->soloplayer_no_ = static_cast<unsigned>(stoi(line));
          } // if (this->game_type() == GameType::marriage)
          if (this->game_type_ == GameType::poverty) {
            // ToDo
          } // if (this->game_type == GameType::poverty)
        } // special cases
      } // if !(keyword == "hands")
    } // gametype
    { // hands
      if (keyword != "hands")
        GETLINE;
      EXPECT_LINE("hands");
      GETLINE;
      EXPECT_LINE("{");

      for (auto const& person : this->persons_) {
        (void)person;
        // the player number
        GETLINE;
        this->hands_.emplace_back(istr);
      } // for (player : this->game().player)
      // a '}'
      GETLINE;
      EXPECT_LINE("}");
    } // hands
    { // actions, tricks
      GETLINE;
      EXPECT_LINE("gameplay actions");
      GETLINE;
      EXPECT_LINE("{");
      do { // while (line != "}")
        /* Format for a trick:
         *
         * Trick: 0
         * Played: Player 0: spade ace
         * Played: Player 1: spade nine
         * Played: Player 2: spade ace
         * Played: Player 3: spade nine
         * 
         * Trick full: 1
         * Startplayer: 0
         * 0: spade ace
         * 1: spade nine
         * 2: spade ace
         * 3: spade nine
         * Winner: 0
         */

        GETLINE;

        if (line == "}") {
          break;
        } else if (line[0] == '#') {
          continue;
        } else if (line == "stop") {
          this->auto_action_end_ = this->actions_.size();
          this->auto_start_party_ = true;
          continue;
        }

        auto action = GameplayAction::create(line, istr);
        if (!action) {
          cerr << "BugReportReplay:\n"
            << "  Unknown action '" << line << "'\n"
            << "  Ignoring it." << endl;
#ifdef DKNOF
          SEGFAULT;
#endif
          continue;
        }

#ifndef POSTPONED
        if (action->type() == GameplayActions::Type::trick_full)
          this->full_tricks_.push_back(dynamic_cast<GameplayActions::TrickFull*>(action.get())->trick);
#endif
#ifndef WORKAROUND
        if (action->type() == GameplayActions::Type::trick_full) {
          // Replace 'TrickFull' with 'TrickClosed',
          // since the trick in 'TrickFull' needs a corresponding game.
          action = make_unique<GameplayActions::TrickTaken>();
        } // if (action->type == GameplayAction::trick_full)
#endif

        if (action->type() == GameplayActions::Type::check) {
          this->auto_action_end_ = this->actions_.size() + 1;
          this->auto_start_party_ = true;
        }

        { // special behaviour in a poverty: save the shifted cards
          if (action->type() == GameplayActions::Type::poverty_shift) {
            this->poverty_cards_shifted_
              = vector<Card>(dynamic_cast<GameplayActions::PovertyShift*>(action.get())->cards);
          } else if (action->type() == GameplayActions::Type::poverty_returned) {
            this->poverty_cards_returned_
              = vector<Card>(dynamic_cast<GameplayActions::PovertyReturned*>(action.get())->cards);
          } // if (action->type == GameplayAction::poverty_shift)
        } // special behaviour in a poverty: save the shifted cards

        this->actions_.push_back(move(action));

      } while (line != "}");
    } // actions, tricks
    { // game summary
      GETLINE;
      if (line == "game summary") {
        EXPECT_LINE("game summary");
        this->game_summary_ = make_unique<GameSummary>(this->rule_, istr);
        GETLINE;
        if (line == "special points") {
          // Just skip them -- they are contained in the game summary
          EXPECT_LINE("special points");
          GETLINE;
          EXPECT_LINE("{");

          do {
            GETLINE;
          } while (line != "}");
          GETLINE;
        } // if (line == "special points")
      } // if (line == "game summary")
    } // game summary
    if (istr.eof())
      throw EOF;
    { // current hands
      EXPECT_LINE("current hands");
      GETLINE;
      EXPECT_LINE("{");

      for (auto const& person : this->persons_) {
        (void)person;
        GETLINE;
        this->current_hands_.emplace_back(istr);
      } // for (person : this->persons_)
      // a '}'
      GETLINE;
      EXPECT_LINE("}");
      GETLINE;
    } // current hands
    if (istr.eof())
      throw EOF;
    { // current trick
      if (line == "current trick")
      {
        this->current_trick_ = Trick(istr);
        GETLINE;
      }
    } // current trick
    if (istr.eof())
      throw EOF;
    { // human actions
      EXPECT_LINE("human actions");
      GETLINE;
      EXPECT_LINE("{");
      do { // while (line != "}")

        GETLINE;
        if (line == "}")
          break;

        auto action = GameplayAction::create(line, istr);

        if (!action) {
          cerr << "BugReportReplay:\n"
            << "  Unknown human action '" << line << "'\n"
            << "  Ignoring it." << endl;
          continue;
        }

        this->human_actions_.push_back(std::move(action));

      } while (line != "}");
      this->human_finished_ = this->human_actions_.empty();
    } // human actions

  } catch (int const status) {
    if (status == EOF) {
      // finished
      // return ;
    } else {
      throw;
    }
  } catch (ReadException const& read_exception) {
    cerr << "BugReportReplay::read_file()\n"
      << "  read exception: " << read_exception.message()
      << endl;
    throw;
  } catch (...) {
    cerr << "BugReportReplay::read_file()\n"
      << "  unknown exception" << endl;
    throw;
  }

#undef GETLINE
#undef READ_FROM_VALUE
#undef EXPECT_LINE
#undef EXPECT_KEYWORD

  this->check_action_ = false;
  this->loaded_ = true;
} // void BugReportReplay::read_file()

/** @return   the current gameplay action
 **/
GameplayAction const&
BugReportReplay::current_action() const
{
  DEBUG_ASSERTION((this->current_action_no_ < this->actions_.size()),
                  "BugReportReplay::current_action()\n"
                  "  no further gameplay action");
  return *this->actions_[this->current_action_no_];
} // GameplayAction BugReportReplay::current_action() const

/** @return   the number of remaining actions
 **/
unsigned
BugReportReplay::remaining_actions() const
{
  return this->actions_.size() - this->current_action_no_;
}

/** @return   the current gameplay action
 **/
GameplayAction const&
BugReportReplay::next_action() const
{
  DEBUG_ASSERTION((this->current_action_no_ + 1 < this->actions_.size()),
                  "BugReportReplay::next_action()\n"
                  "  no further gameplay action");
  return *this->actions_[this->current_action_no_ + 1];
}

/** the current action has been processed
 **
 ** @param    action        processed action
 ** @param    discrepancy   discrepancy of the processed and the saved action
 **/
void
BugReportReplay::action_processed(GameplayAction const& action,
                                  GameplayActions::Discrepancy const
                                  discrepancy)
{
  if (this->finished_)
    return ;

  // check the action when requested
  if (   this->check_action_
      && (this->current_action().type() != GameplayActions::Type::check) ) {
    DEBUG_ASSERTION(this->actions_[this->current_action_no_ - 1]->type()
                    == GameplayActions::Type::check,
                    "BugReportReplay::action_processed(discrepancy)\n"
                    "  last action is not 'check'");
    this->reference_check_game_action(action, discrepancy);
    this->check_action_ = false;
  } // if (this->check_action_)

  // test for a human action
  if (!this->human_finished_
      && this->current_action() == this->current_human_action()) {
    this->human_actions_discrepancies_.push_back(discrepancy);
    this->current_human_action_no_ += 1;
    if (this->current_human_action_no_ == this->human_actions_.size()) {
      this->human_finished_ = true;
    }
  } // if (human action)

  // save the discrepancy and go to the next action
  this->actions_discrepancies_.push_back(discrepancy);
  this->current_action_no_ += 1;

  if (this->current_action_no_ == this->actions_.size()) {
    this->end_reached();
    return ;
  }

  // automatic actions for the human player
  if (::game_status >= GameStatus::game_play)
    this->handle_current_human_action();

  if (this->finished_)
    return ;

  // test for printing to perform
  if (this->current_action().type() == GameplayActions::Type::print) {
    if (   (::game_status >= GameStatus::game_poverty_shift)
        && (::game_status < GameStatus::game_finished) ) {
      // in a running game
      auto const& action = dynamic_cast<GameplayActions::Print const&>(this->current_action());
      auto const& player = ::party->game().players().current_player();
      auto ai = dynamic_cast<Ai const*>(&player);
      if (action.info == "help") {
        cout << "print teams: print the team information of the ai\n";
        cout << "print cards: print the cards information of the ai\n";
        cout << "print time:  print the current time\n";
      } else if (action.info == "time") {
        auto time =  std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        auto timeinfo = localtime(&time);
        char s[20];
        std::strftime(s, 19, "%H:%M:%S", timeinfo); // NOLINT
        cout << s << '\n'; // NOLINT
      } else if (ai && (action.info == "teams")) {
        cout << '\n'
          << ai->name() << ": "
          << ai->team_information() << '\n';
      } else if (ai && (action.info == "cards")) {
        if (   (::game_status >= GameStatus::game_poverty_shift)
            && (::game_status < GameStatus::game_finished) ) {
          cout << '\n'
            << ai->name() << ": "
            << ai->cards_information() << '\n';
        }
      } else {
        cout << action.info << '\n';
      }
    } // if (game_status ...)
    this->action_processed(this->current_action(), GameplayActions::Discrepancy::none);
  } // if (this->current_action().type() == GameplayAction::print)

  // test for check to perform
  if (this->current_action().type() == GameplayActions::Type::check) {
    this->check_action_ = true;
    this->action_processed(this->current_action(), GameplayActions::Discrepancy::none);
  } // if (this->current_action().type() == GameplayAction::check)
} // void BugReportReplay::action_processed(GameplayAction action, GameplayAction::Discrepancy discrepancy)

/** check the game action of a reference
 ** Print and save the result.
 ** End the game, if it is the last check
 **
 ** @param    action        processed action
 ** @param    discrepancy   discrepancy of the processed and the saved action
 **/
void
BugReportReplay::reference_check_game_action(GameplayAction const& action,
                                             GameplayActions::Discrepancy
                                             discrepancy)
{
  DEBUG_ASSERTION(this->current_action_no_ > 0,
                  "BugreportReplay::reference_check_game_action(action, discrepancy)\n"
                  "  this is the first action in the bug report replay");

  DEBUG_ASSERTION((this->actions_[this->current_action_no_ - 1]->type()
                   == GameplayActions::Type::check),
                  "BugreportReplay::reference_check_game_action(action, discrepancy)\n"
                  "  last action is no check");
  auto const& check_action
    = dynamic_cast<GameplayActions::Check const&>(*this->actions_[this->current_action_no_ - 1]);

  if (::references_check)
    ::references_check->check(::party->game(),
                              action, this->current_action(),
                              check_action.comment,
                              discrepancy,
                              (this->auto_action_end_
                               == this->current_action_no_) );
} // void BugReportReplay::reference_check_game_action(GameplayAction action, GameplayAction::Discrepancy discrepancy)

/** -> result
 **
 ** @param    action_no   number of the action
 **
 ** @return   the discrepancy for the given action
 **/
GameplayActions::Discrepancy
BugReportReplay::discrepancy(unsigned const action_no) const
{
  if (action_no >= this->actions_discrepancies_.size())
    return GameplayActions::Discrepancy::future;

  return this->actions_discrepancies_[action_no];
} // Discrepancy BugReportReplay::discrepancy(unsigned action_no) const

/** @return   the current human action
 **/
GameplayAction const&
BugReportReplay::current_human_action() const
{
  DEBUG_ASSERTION((this->current_human_action_no_ < this->human_actions_.size()),
                  "BugReportReplay::current_human_action()\n"
                  "  no further human action");
  return *this->human_actions_[this->current_human_action_no_];
} // GameplayAction BugReportReplay::current_human_action() const

/** -> result
 **
 ** @param    action_no   number of the human action
 **
 ** @return   the discrepancy for the given human action
 **/
GameplayActions::Discrepancy
BugReportReplay::human_discrepancy(unsigned const action_no) const
{
  if (action_no >= this->human_actions_discrepancies_.size())
    return GameplayActions::Discrepancy::future;

  return this->human_actions_discrepancies_[action_no];
} // Discrepancy BugReportReplay::human_discrepancy(unsigned action_no) const

/** handle the current human action
 **/
void
BugReportReplay::handle_current_human_action()
{
  if (   this->finished_
      || this->human_finished_
      || (this->current_action() != this->current_human_action()) )
    return ;

  // print the coming action
  this->print_current_human_action();

  auto& game = ::party->game();

  // partly execute the action automatically
  switch (this->current_human_action().type()) {
  case GameplayActions::Type::announcement: {
    auto const& announcement_action
      = dynamic_cast<GameplayActions::Announcement const&>(this->current_human_action());
    if (this->verbose_ & verbose_action)
      cout << "BugReportReplay:\n"
        << "  making announcement '" << announcement_action.announcement << "'"
        << " for player " << announcement_action.player
        << endl;

    // make the announcement in place of the human and tell it the user
    if (!is_reply(announcement_action.announcement))
      game.announcements().make_announcement(announcement_action.announcement,
                                             game.player(announcement_action.player));
  } break;

  case GameplayActions::Type::swines: {
    auto const& swines_action
      = dynamic_cast<GameplayActions::Swines const&>(this->current_human_action());
    if (this->verbose_ & verbose_action)
      cout << "BugReportReplay:\n"
        << "  announce 'swines' for human player "
        << swines_action.player
        << endl;
    if (!game.swines().swines_announce(game.player(swines_action.player)))
      cerr << "Error announcing 'swines' for human player"
        << swines_action.player
        << endl;
  } break;

  case GameplayActions::Type::hyperswines: {
    auto const& hyperswines_action
      = dynamic_cast<GameplayActions::Hyperswines const&>(this->current_human_action());
    if (this->verbose_ & verbose_action)
      cout << "BugReportReplay:\n"
        << "  announce 'hyperswines' for human player "
        << hyperswines_action.player
        << endl;
    if (!game.swines().hyperswines_announce(game.player(hyperswines_action.player)))
      cerr << "Error announcing 'hyperswines' for human player"
        << hyperswines_action.player
        << endl;
  } break;

  default:
    break;
  } // switch (this->current_human_action().type())
} // void BugReportReplay::handle_current_human_action()

/** print the next human action
 **/
void
BugReportReplay::print_current_human_action() const
{
  if (!(this->verbose_ & verbose_human_action))
    return ;
  switch (this->current_human_action().type()) {
  case GameplayActions::Type::card_played: {
    auto const& card_played_action
      = dynamic_cast<GameplayActions::CardPlayed const&>(this->current_human_action());
    cout << "BugReportReplay: Human\n"
      << "  play " << card_played_action.card
      << endl;
  } break;

  case GameplayActions::Type::reservation: {
    auto const& reservation_action
      = dynamic_cast<GameplayActions::Reservation const&>(this->current_human_action());
    cout << "BugReportReplay: Human\n"
      << "  gametype " << reservation_action.reservation.game_type << '\n';
    if (reservation_action.reservation.game_type == GameType::marriage)
      cout << "  marriage selector " << reservation_action.reservation.marriage_selector <<'\n'; 
    if (reservation_action.reservation.swines)
      cout << "  swines\n";
    if (reservation_action.reservation.hyperswines)
      cout << "  hyperswines\n";
  } break;

  case GameplayActions::Type::poverty_shift: {
    auto const& poverty_shift_action
      = dynamic_cast<GameplayActions::PovertyShift const&>(this->current_human_action());
    cout << "BugReportReplay: Human\n"
      << "  poverty: shift\n";
    for (auto const& c : poverty_shift_action.cards)
      cout << "    " << c << '\n';
  } break;

  case GameplayActions::Type::poverty_returned: {
    auto const& poverty_returned_action
      = dynamic_cast<GameplayActions::PovertyReturned const&>(this->current_human_action());
    cout << "BugReportReplay: Human\n"
      << "  poverty: return\n";
    for (auto const& c : poverty_returned_action.cards)
      cout << "    " << c << '\n';
  } break;

  case GameplayActions::Type::poverty_accepted: {
    cout << "BugReportReplay: Human\n"
      << "  poverty: accept\n";
  } break;

  case GameplayActions::Type::poverty_denied: {
    cout << "BugReportReplay: Human\n"
      << "  poverty: deny\n";
  } break;

  case GameplayActions::Type::poverty_denied_by_all:
  case GameplayActions::Type::trick_open:
  case GameplayActions::Type::trick_full:
  case GameplayActions::Type::trick_taken:
  case GameplayActions::Type::marriage:
  case GameplayActions::Type::check:
  case GameplayActions::Type::print:
    // nothing to do
    break;

  case GameplayActions::Type::announcement:
  case GameplayActions::Type::swines:
  case GameplayActions::Type::hyperswines:
    // -> action_processed

  case GameplayActions::Type::genscher:
    // ToDo
    break;
  } // switch (this->current_human_action().type())
} // void BugReportReplay::print_current_human_action() const

/** the bug report is inconsistent:
 **/
void
BugReportReplay::mismatch()
{
  cerr << "BugReport mismatch!" << endl;
} // void BugReportReplay::mismatch()

/** the bug report is finished.
 **/
void
BugReportReplay::end_reached()
{
  this->finished_ = true;
  this->human_finished_ = true;

  if (this->verbose_ & verbose_finish) {
    cout << "BugReport: finished" << endl;
    cout << "Message:\n"
      << "{\n"
      << this->message_
      << "}"
      <<endl;
  }
} // void BugReportReplay::end_reached()

/** prints the header and the differences
 **/
void
BugReportReplay::print_header() const
{
  if (!(this->verbose_ & verbose_header))
    return;
  // print the first lines
  cout << "BugReport\n"
    << "\n";

  if (this->version_)
    cout << "version:     " << *this->version_  << '\n';
  else
    cout << "version:     -\n";
  cout
    << "compiled:    " << this->compiled_        << '\n'
    << "system:      " << this->system_          << '\n'
    << "time:        " << this->time_            << '\n'
    << "language:    " << this->language_        << '\n'
    << "seed:        " << this->seed_            << '\n'
    << "startplayer: " << this->startplayer_no_  << '\n'
    << "trick:       " << this->trickno_         << '\n'
    << '\n'
    << "message:\n"
    << "{\n"
    <<   this->message_
    << "}\n";
} // void BugReportReplay::print_header() const

/** @return   the filename
 **/
string const&
BugReportReplay::filename() const
{ return this->filename_; }

/** @return   wether the bug report was loaded successfully
 **/
bool
BugReportReplay::loaded() const
{ return this->loaded_; }

/** @return   whether the bug report replay is finished
 **/
bool
BugReportReplay::finished() const
{ return this->finished_; }

/** @return   wether the party is automatically started
 **/
bool
BugReportReplay::auto_start_party() const
{ return this->auto_start_party_; }

/** @return   the version
 **/
Version const&
BugReportReplay::version() const
{ return *this->version_; }

/** @return   the compiled string
 **/
string const&
BugReportReplay::compiled() const
{ return this->compiled_; }

/** @return   the system string
 **/
string const&
BugReportReplay::system() const
{ return this->system_; }

/** @return   the time string
 **/
string const&
BugReportReplay::time() const
{ return this->time_; }

/** @return   the language string
 **/
string const&
BugReportReplay::language() const
{ return this->language_; }

/** @return   the trick number
 **/
unsigned
BugReportReplay::trickno() const
{ return this->trickno_; }

/** @return   the message
 **/
string const&
BugReportReplay::message() const
{ return this->message_; }

/** @return   the game type
 **/
GameType
BugReportReplay::game_type() const
{ return this->game_type_; }

/** @return   the seed
 **/
::Seed
BugReportReplay::seed() const
{ return this->seed_; }

/** @return   the number of the startplayer
 **/
unsigned
BugReportReplay::startplayer_no() const
{ return this->startplayer_no_; }

/** @return   the number of the soloplayer
 **/
unsigned
BugReportReplay::soloplayer_no() const
{ return this->soloplayer_no_; }

/** @return   the shifted cards
 **/
vector<Card> const&
BugReportReplay::poverty_cards_shifted() const
{ return this->poverty_cards_shifted_; }

/** @return   the returned cards
 **/
vector<Card> const&
BugReportReplay::poverty_cards_returned() const
{ return this->poverty_cards_returned_; }

/** @return   the marriage selector
 **/
MarriageSelector
BugReportReplay::marriage_selector() const
{ return this->marriage_selector_; }

/** @return   the number of the current action
 **/
unsigned
BugReportReplay::current_action_no() const
{ return this->current_action_no_; }

/** @return   the number of the current human action
 **/
unsigned
BugReportReplay::current_human_action_no() const
{ return this->current_human_action_no_; }

/** @return   the game summary
 **/
unique_ptr<GameSummary> const&
BugReportReplay::game_summary() const
{ return this->game_summary_; }

/** @return   the actions
 **/
vector<unique_ptr<GameplayAction>> const&
BugReportReplay::actions() const
{ return this->actions_; }

/** @return   the action discrepancies
 **/
vector<GameplayActions::Discrepancy> const&
BugReportReplay::actions_discrepancies() const
{ return this->actions_discrepancies_; }

/** @return   the human actions
 **/
vector<unique_ptr<GameplayAction>> const&
BugReportReplay::human_actions() const
{ return this->human_actions_; }

/** @return   the human action discrepancies
 **/
vector<GameplayActions::Discrepancy> const&
BugReportReplay::human_actions_discrepancies() const
{ return this->human_actions_discrepancies_; }

/** a party is opened
 **
 ** @param    party   the party that is opened
 **/
void
BugReportReplay::party_open(Party const& party)
{
  ::party->signal_get_settings().connect(*this, &BugReportReplay::party_get_settings, this->disconnector_);
  ::party->signal_loaded().connect(*this, &BugReportReplay::party_loaded, this->disconnector_);
  ::party->signal_start().connect(*this, &BugReportReplay::party_start, this->disconnector_);

  this->finished_ = false;
  this->current_action_no_ = 0;
  this->current_human_action_no_ = 0;
  this->actions_discrepancies_.clear();
  this->human_actions_discrepancies_.clear();

  this->human_finished_ = this->human_actions_.empty();

  BugReportReplay::signal_open()(*this);

  if (this->current_action_no_ == this->actions_.size()) {
    this->end_reached();
    return ;
  }

  if (this->current_action().type() == GameplayActions::Type::check) {
    this->check_action_ = true;
    this->action_processed(this->current_action(),
                           GameplayActions::Discrepancy::none);
  } // if (this->current_action().type == GameplayActions::Type::check)
} // void BugReportReplay::party_open(Party party)

/** set the settings of the party
 **/
void
BugReportReplay::party_get_settings()
{
  ::party->set_seed(this->seed_);
  ::party->set_startplayer(this->startplayer_no_);

  // set the rules and print the rule differences
  this->set_party_rule();

  { // set the players
    for (unsigned n = 0; n < this->persons_.size(); ++n) {
      ::party->players().set(n, this->persons_[n]->clone());
    }
  } // set the players

  if (this->verbose_ & verbose_message)
    // repeat the message
    cout << '\n'
      << "Message:\n"
      << "{\n"
      <<   this->message_
      << "}\n";
} // void BugReportReplay::party_get_settings()

/** the party is loaded
 **/
void
BugReportReplay::party_loaded()
{
  // (re)set the rules and print the rule differences
  this->set_party_rule();

  if (this->verbose_ & verbose_message) {
    // repeat the message
    cout << '\n'
      << "Message:\n"
      << "{\n"
      <<   this->message_
      << "}\n";
  }
} // void BugReportReplay::party_loaded()

/** set the rule of the party
 **/
void
BugReportReplay::set_party_rule()
{
  // the hardcoded ruleset (the harddcoded)
  Rule const rule_hardcoded;
  // set the party rules
  auto& rule = ::party->rule();
  rule = this->rule_;

  if (this->verbose_ & verbose_rules) {
    cout << '\n'
      << "rule differences:\n"
      << "{\n";
    cout << setw(38) << ""
      << setw(12) << "bug_report"
      << setw(12) << "hardcoded"
      << "\n";
    for (unsigned i = Rule::Type::first; i <= Rule::Type::last; i++) {
      if (   (i >= (Rule::Type::bool_first))
          && (i <= Rule::Type::bool_last)) {
        if (   (rule(Rule::Type::Bool(i))
                != rule_hardcoded(Rule::Type::Bool(i))) ) {
          cout << setw(38) << Rule::Type::Bool(i)
            << setw(12) << rule(Rule::Type::Bool(i))
            << setw(12) << rule_hardcoded(Rule::Type::Bool(i))
            << "\n";
        } // if (rule differs)
      } else if (   (i >= Rule::Type::unsigned_first)
                 && (i <= Rule::Type::unsigned_last)) {
        if (   (rule(Rule::Type::Unsigned(i))
                != rule_hardcoded(Rule::Type::Unsigned(i))) ) {
          cout << setw(38) << Rule::Type::Unsigned(i)
            << setw(12) << rule(Rule::Type::Unsigned(i))
            << setw(12) << rule_hardcoded(Rule::Type::Unsigned(i))
            << "\n";
        } // if (rule differs)
      } else if (   (i >= Rule::Type::unsigned_extra_first)
                 && (i <= Rule::Type::unsigned_extra_last)) {
        if (   (rule(Rule::Type::UnsignedExtra(i))
                != rule_hardcoded(Rule::Type::UnsignedExtra(i)))) {
          cout << setw(38) << Rule::Type::UnsignedExtra(i)
            << setw(12) << rule(Rule::Type::UnsignedExtra(i))
            << setw(12) << rule_hardcoded(Rule::Type::UnsignedExtra(i))
            << "\n";
        } // if (rule differs)
      } else if (i == Rule::Type::counting) {
        if (   (rule(Rule::Type::counting)
                != rule_hardcoded(Rule::Type::counting)) ) {
          cout << setw(38) << Rule::Type::counting
            << setw(12) << rule(Rule::Type::counting)
            << setw(12) << rule_hardcoded(Rule::Type::counting)
            << "\n";
        } // if (rule differs)
      } else { // if (i: type unknown)
        DEBUG_ASSERTION(false,
                        "BugReportReplay::party_get_settings():\n"
                        "  unknown ruletype number " << i);
      } // if (i: type unknown)
    } // for (unsigned i = Rule::Type::first; i <= Rule::Type::last; i++)
    cout << "}" << endl;
  } // if (this->verbose_ & verbose_rules)
} // void BugReportReplay::set_party_rule()

/** the party is started
 **/
void
BugReportReplay::party_start()
{
  this->auto_start_party_ = false;
} // void BugReportReplay::party_start()

/** the game is opened
 **
 ** @param    game   game that is opened
 **/
void
BugReportReplay::game_open(Game const& game)
{
  {
    auto& game = ::party->game();
    game.signal_gameplay_action().connect(*this, &BugReportReplay::gameplay_action, this->disconnector_);
    game.poverty().signal_ask().connect(*this, &BugReportReplay::poverty_ask, this->disconnector_);
    game.tricks().signal_trick_open().connect(*this, &BugReportReplay::trick_open, this->disconnector_);
  }
} // void BugReportReplay::game_open(Game game)

/** a gameplay action
 **
 ** @param    action   the action
 **/
void
BugReportReplay::gameplay_action(GameplayAction const& action)
{
  if (this->finished_)
    return ;

  auto const& game = ::party->game();
  // do handle all actions (see BugReport::gameplay_action(action) )
  switch (action.type()) {
  case GameplayActions::Type::poverty_denied_by_all:
  case GameplayActions::Type::trick_taken:
    // do not handle
    return ;
  case GameplayActions::Type::check:
  case GameplayActions::Type::print:
    // shall not happen
    DEBUG_ASSERTION(false,
                    "Gameplay action 'check'/'print' should not be processed here");
    return ;

  case GameplayActions::Type::trick_full: // replaced by TRICK_OPEN
    if (!game.tricks().is_last_trick())
      return ;
  case GameplayActions::Type::trick_open: // replaced by TRICK_FULL
    if (game.tricks().current_no() == 0)
      return ;
  case GameplayActions::Type::reservation:
  case GameplayActions::Type::poverty_shift:
  case GameplayActions::Type::poverty_returned:
  case GameplayActions::Type::poverty_accepted:
  case GameplayActions::Type::poverty_denied:
  case GameplayActions::Type::card_played:
  case GameplayActions::Type::announcement:
  case GameplayActions::Type::swines:
  case GameplayActions::Type::hyperswines:
  case GameplayActions::Type::marriage:
  case GameplayActions::Type::genscher:
    // do handle
    break;
  } // switch (action.type())

  // check that this action conforms to the gameplay
  if (   (action == this->current_action())
      || (   (action.type() == GameplayActions::Type::trick_open)
          && (this->current_action().type() == GameplayActions::Type::trick_full) )
     ) {
    this->action_processed(action, GameplayActions::Discrepancy::none);
    return ;
  }
  if (   this->remaining_actions() >= 2
      && action == this->next_action()) {
    cerr << "BugReport action skipped:\n"
      << "  bug report: " << this->current_action() << endl;
    this->action_processed(this->current_action(), GameplayActions::Discrepancy::skipped);
    this->action_processed(action, GameplayActions::Discrepancy::none);
    return ;
  }
  // the action differs from the one of the bug report
  cerr << "BugReport different actions:\n"
    << "  game:       " << action << '\n'
    << "  bug report: " << this->current_action() << endl;

  // if the type is equal there is only some other gameplay
  if (action.type() == this->current_action().type()) {
    switch (this->current_action().type()) {
    case GameplayActions::Type::card_played:
      if (dynamic_cast<GameplayActions::CardPlayed const&>(action).card
          != dynamic_cast<GameplayActions::CardPlayed const&>(this->current_action()).card) {
        this->action_processed(action, GameplayActions::Discrepancy::card);
        break;
      }
    default:
      this->action_processed(action, GameplayActions::Discrepancy::other);
      break;
    } // switch (this->current_action().type())

    return ;
  } // if (action.type() == this->current_action().type())

  // skip announcements
  switch (this->current_action().type()) {
  case GameplayActions::Type::announcement:
  case GameplayActions::Type::swines:
  case GameplayActions::Type::hyperswines:
  case GameplayActions::Type::marriage:
  case GameplayActions::Type::genscher:
#ifdef OUTDATED
    // 2017-02-02: why skipping?
    if (action.type() >= GameplayAction::card_played) {
      // just skip
      this->action_processed(action, GameplayActions::Discrepancy::skipped);
      this->gameplay_action(action);
      return ;
    }
#endif
    break;

  default:
    break;
  } // switch (this->current_action().type())

  switch (action.type()) {
  case GameplayActions::Type::card_played:
    if (this->current_action().type() == GameplayActions::Type::card_played) {
      this->action_processed(action, GameplayActions::Discrepancy::skipped);
      break;
    }
    // search the next card-played action
    while (!this->finished_) {
      this->action_processed(action, GameplayActions::Discrepancy::skipped);
      if (this->finished_)
        break ;
      if (this->current_action().type() == GameplayActions::Type::card_played)
        break ;
    } // while (!this->finished_)
    break;

  case GameplayActions::Type::reservation:
  case GameplayActions::Type::poverty_shift:
  case GameplayActions::Type::poverty_accepted:
  case GameplayActions::Type::poverty_returned:
  case GameplayActions::Type::poverty_denied:
  case GameplayActions::Type::poverty_denied_by_all:
  case GameplayActions::Type::trick_open:
  case GameplayActions::Type::trick_full:
  case GameplayActions::Type::trick_taken:
    // ToDo
    break;

  case GameplayActions::Type::announcement:
  case GameplayActions::Type::marriage:
  case GameplayActions::Type::genscher:
  case GameplayActions::Type::swines:
  case GameplayActions::Type::hyperswines:
  case GameplayActions::Type::check:
  case GameplayActions::Type::print:
    // ignore
    break;
  } // switch (action.type())

  return ;
} // void BugReportReplay::gameplay_action(GameplayAction action)

/** ask 'player' whether to accept the poverty
 **
 ** @param    player   player that is asked
 ** @param    cardno   number of shifted cards
 **/
void
BugReportReplay::poverty_ask(Player const& player,
                             unsigned const cardno)
{
  this->handle_current_human_action();
} // void BugReportReplay::poverty_ask(Player player, unsigned cardno);

/** a new trick is opened
 **
 ** @param    trick   trick that is opened
 **/
void
BugReportReplay::trick_open(Trick const& trick)
{
  // automatic actions for the human player
  // De facto this is only needed for the first trick so that the player
  // can announce before any other action.
  this->handle_current_human_action();
} // void BugReportReplay::trick_open(Trick const& trick)

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the next card played by the player
 **/
HandCard
BugReportReplay::next_card(Player const& player) const
{
  if (this->human_finished_) {
    if (this->finished_)
      return {};
    auto const& action = this->current_action();
    if (action.type() != GameplayActions::Type::card_played)
      return {};
    auto const& card_played = dynamic_cast<GameplayActions::CardPlayed const&>(action);
    if (card_played.player != player.no())
      return {};
    return HandCard(player.hand(), card_played.card);
  }

  auto const& action = this->current_human_action();

  if (action.type() == GameplayActions::Type::genscher)
    return HandCard(player.hand(), Card::diamond, Card::king);

  if (action.type() != GameplayActions::Type::card_played) {
    cerr << "Bug report: human action should be 'card played' but is:\n"
      << "  " << this->current_human_action() << endl;
    return {};
  } // if (this->current_human_action().type() != GameplayActions::Type::card_played)

  return HandCard(player.hand(), dynamic_cast<GameplayActions::CardPlayed const&>(this->current_human_action()).card);
} // HandCard BugReportReplay::next_card(Player player) const

/** set the auto action end
 **
 ** @param    auto_action_end   the end of the auto actions
 **/
void
BugReportReplay::set_auto_action_end(unsigned const auto_action_end)
{
  this->auto_action_end_ = auto_action_end;
  this->auto_start_party_ = true;
} // void BugReportReplay::set_auto_action_end(unsigned auto_action_end)

/** @return   whether the current action should be executed automatically
 **/
bool
BugReportReplay::auto_action() const
{
  if (this->finished_)
    return false;

  return (!this->check_action_
          && (this->current_action_no_ < this->auto_action_end_));
} // bool BugReportReplay::auto_action() const

} // namespace OS_NS
