/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "../game/gameplay.h"
#include "../game/gameplay_action.h"
#include "../card/hand.h"
#include "../card/trick.h"
#include "../party/party.h"
#include "../party/rule.h"
#include "../game/game.h"
#include "../game/game_summary.h"

#include "../basetypes.h"
#include "../utils/version.h"

namespace OS_NS {
class BugReportReplay;
} // namespace OS_NS 
extern unique_ptr<OS_NS::BugReportReplay> bug_report_replay;

namespace OS_NS {

/**
 ** A class to simplify the replay of the bug report.
 ** All info's from the bug report are used and differences to the
 ** gameplay are shown.
 **
 ** Parts of the bug report:
 ** - version      is printed
 ** - system       is printed
 ** - date         is printed
 ** - message      is printed
 ** - seed         is set
 ** - startplayer  is set
 ** - ruleset      is set, differences to the standard ruleset and
 **                the default are printed
 ** - players      are set, differences to the standard aiconfig
 **                are printed
 ** - hands        are checked
 **                * marriage: bridge is checked
 ** - gametype     is checked
 **                * marriage: bridge is checked
 **                * poverty: shifted cards are checked
 **
 ** @todo   poverty
 ** @todo   marriage
 ** @todo   automatic announcements for the player
 **/
class BugReportReplay {
  public:
    enum Verbose {
      verbose_none = 0,
      verbose_seed = 1 << 1,
      verbose_header = 1 << 2,
      verbose_message = 1 << 3,
      verbose_rules = 1 << 4,
      verbose_action = 1 << 5,
      verbose_human_action = 1 << 6,
      verbose_finish = 1 << 7,
      verbose_all = ((1 << 8) - 1)
    }; // enum Verbose
  public:
    BugReportReplay() = delete;
    BugReportReplay(string const& filename, unsigned verbose);
    BugReportReplay(BugReportReplay const&) = delete;
    BugReportReplay& operator=(BugReportReplay const&) = delete;
    ~BugReportReplay();

    // signals
    static Signal<void(BugReportReplay&)>& signal_open();
    Signal<void()>& signal_close();

    void write(ostream& ostr) const;

    // prints the header and the differences
    void print_header() const;

    string const& filename() const;
    bool loaded() const;
    bool finished() const;

    bool auto_start_party() const;
    Version const& version() const;
    string const& compiled() const;
    string const& system() const;
    string const& time() const;
    string const& language() const;
    unsigned trickno() const;
    string const& message() const;
    GameType game_type() const;
    ::Seed seed() const;
    unsigned startplayer_no() const;
    unsigned soloplayer_no() const;
    MarriageSelector marriage_selector() const;
    vector<Card> const& poverty_cards_shifted() const;
    vector<Card> const& poverty_cards_returned() const;
    unsigned current_action_no() const;
    unsigned current_human_action_no() const;
    unique_ptr<GameSummary> const& game_summary() const;
    vector<unique_ptr<GameplayAction>> const& actions() const;
    vector<GameplayActions::Discrepancy> const& actions_discrepancies() const;
    vector<unique_ptr<GameplayAction>> const& human_actions() const;
    vector<GameplayActions::Discrepancy> const& human_actions_discrepancies() const;

    void party_open(Party const& party);
    void party_get_settings();
    void party_loaded();
    void set_party_rule();
    void party_start();

    void game_open(Game const& game);
    void gameplay_action(GameplayAction const& action);

    void poverty_ask(Player const& player, unsigned cardno);
    void trick_open(Trick const& trick);

    // returns the next card played by the human player
    HandCard next_card(Player const& player) const;

    // the discrepancy for the given action
    GameplayActions::Discrepancy discrepancy(unsigned action_no) const;
    // the discrepancy for the given human action
    GameplayActions::Discrepancy human_discrepancy(unsigned action_no) const;

    GameplayAction const& current_action() const;
    unsigned remaining_actions() const;
    GameplayAction const& next_action() const;
    // whether the current action should be executed automatically
    void set_auto_action_end(unsigned auto_action_end);
    // whether the current action should be executed automatically
    bool auto_action() const;

  private:
    // initializes the bug report replay
    void init();

  private:
    // reads the whole file
    void read_file();

    // 'line' is expected -- test the current line
    bool expect_line(string const& read_line,
                     string const& expected_line);
    // 'value' is expected -- test the current value
    bool expect_value(string const& read_value,
                      string const& expected_value);
    // 'keyword' is expected -- test the current keyword
    bool expect_keyword(string const& read_keyword,
                        string const& expected_keyword);


    // the current action has been processed
    void action_processed(GameplayAction const& action,
                          GameplayActions::Discrepancy discrepancy);
    // check game action when requested
    void reference_check_game_action(GameplayAction const& action,
                                     GameplayActions::Discrepancy discrepancy);

    // the current human action
    GameplayAction const& current_human_action() const;
    // handle the coming human action
    void handle_current_human_action();
    // print the coming human action
    void print_current_human_action() const;

    // this bug report is inconsistent
    void mismatch();
    // the end of the bug report is reached
    void end_reached();

  private:
    AutoDisconnector disconnector_;

    // the filename of the bugreport
    string filename_;
    // the verbose level
    unsigned verbose_ = verbose_none;
    // whether this bug report was loaded successfully
    bool loaded_ = false;
    // whether this bug report is inconsistent
    bool inconsistent_ = false;
    // whether this bug report is finished
    bool finished_ = false;
    // whether this bug report is finished in the human parts
    bool human_finished_ = false;
    // whether the party should be started automatically
    bool auto_start_party_ = false;
    // till which action to process automatically
    unsigned auto_action_end_ = 0;
    // whether to check the next action
    bool check_action_ = false;

    // the version
    unique_ptr<Version> version_;
    // the date of the compilation
    string compiled_;
    // the system
    string system_;
    // the time the bug report was created
    string time_;
    // the language
    string language_;
    // the trick of the bug report
    unsigned trickno_ = UINT_MAX;
    // the mesage
    string message_;
    // the seed
    ::Seed seed_ = UINT_MAX;
    // the number of the startplayer
    unsigned startplayer_no_ = UINT_MAX;
    // the gametype
    ::GameType game_type_ = GameType::normal;
    // the marriage selector
    ::MarriageSelector marriage_selector_ = MarriageSelector::team_set;
    // the number of the soloplayer
    unsigned soloplayer_no_ = UINT_MAX;

    // the rule
    ::Rule rule_;
    // the persons
    vector<unique_ptr<Player>> persons_;
    // poverty cards
    vector<Card> poverty_cards_shifted_;
    vector<Card> poverty_cards_returned_;
    // hands
    vector<Hand> hands_;
    // the number of the player with swines
    unsigned swines_player_no_ = UINT_MAX;
    // the number of the player with hyperswines
    unsigned hyperswines_player_no_ = UINT_MAX;
    // gameplay
    vector<unique_ptr<GameplayAction>> actions_;
    // the number of the current action
    unsigned current_action_no_ = 0;
    // whether the action is conform to the gameplay
    vector<GameplayActions::Discrepancy> actions_discrepancies_;

    // full tricks
    vector<::Trick> full_tricks_;
    // game summary
    unique_ptr<::GameSummary> game_summary_;

    // current hands
    vector<::Hand> current_hands_;
    // current trick
    ::Trick current_trick_;
    // human actions
    vector<unique_ptr<GameplayAction>> human_actions_;
    // the number of the current human action
    unsigned current_human_action_no_ = 0;
    // whether the action is conform to the gameplay
    vector<GameplayActions::Discrepancy> human_actions_discrepancies_;

    // signals
    static Signal<void(BugReportReplay&)> signal_open_;
    Signal<void()> signal_close_;
}; // class BugReportReplay

} // namespace OS_NS
