/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include <iterator>

#include "constants.h"
#include "party.h"

#include "../utils/string.h"
#include "../utils/random.h"
#include "../utils/container_algorithm.h"

#include "rule.h"
#include "../game/game_summary.h"
#include "../player/player.h"
#include "../player/ai/ai.h"
#include "../player/human/human.h"
#include "../card/trick.h"

#include "../misc/setting.h"

#include "../ui/ui.h"

#ifdef CHECK_RUNTIME
#include "../runtime.h"
#endif

#include <ctime>

Signal<void(Party&)> Party::signal_open_;

/** constructor
 **/
Party::Party() = default;

/** initialize the party
 ** load the players from the private data directory
 **/
void
Party::init()
{
  rule_ = make_unique<Rule>();
  players_ = make_unique<Players>(*this);
}

/** @return   signal when the party is opened
 **/
Signal<void(Party&)>&
Party::signal_open() noexcept
{ return Party::signal_open_; }

/** @return   signal: the party is opened
 **/
Signal<void()>&
Party::signal_get_settings() noexcept
{ return this->signal_get_settings_; }

/** @return   signal: the party is loaded
 **/
Signal<void()>&
Party::signal_loaded() noexcept
{ return this->signal_loaded_; }

/** @return   signal: the party is started
 **/
Signal<void()>&
Party::signal_start() noexcept
{ return this->signal_start_; }

/** @return   signal: a round is started
 **/
Signal<void(unsigned)>&
Party::signal_start_round() noexcept
{ return this->signal_start_round_; }

/** @return   signal: the party is finished
 **/
Signal<void()>&
Party::signal_finish() noexcept
{ return this->signal_finish_; }

/** @return   signal: the party is closed
 **/
Signal<void()>&
Party::signal_close() noexcept
{ return this->signal_close_; }

/** @return   signal: the seed has changed
 **/
Signal<void()>&
Party::signal_seed_changed() noexcept
{ return this->signal_seed_changed_; }

/** @return   signal: the startplayer has changed
 **/
Signal<void()>&
Party::signal_startplayer_changed() noexcept
{ return this->signal_startplayer_changed_; }

/** @return   the rule
 **/
Rule const&
Party::rule() const noexcept
{
  return *this->rule_;
} // Rule Party::rule() const

/** @return   the rule
 **/
Rule&
Party::rule() noexcept
{
  return *this->rule_;
} // Rule Party::rule()

/** @return   the players
 **/
Party::Players const&
Party::players() const noexcept
{
  return *this->players_;
} // Party::Players const& Party::players() const

/** @return   the players
 **/
Party::Players&
Party::players() noexcept
{
  return *this->players_;
} // Party::Players& Party::players()

/** @return   whether the seed is random
 **/
bool
Party::random_seed() const noexcept
{
  return this->random_seed_;
} // bool Party::random_seed() const

/** @return   the seed
 **/
Seed
Party::seed() const noexcept
{
  if (this->in_configuration()) {
    if (this->random_seed_)
      return 0;
    return this->first_seed_;
  }

  return this->seed_;
} // Seed Party::seed() const

/** sets the seed
 **
 ** @param    seed   new seed
 **/
void
Party::set_seed(Seed const seed) noexcept
{
  if (this->in_configuration()) {
    if (this->first_seed_ != seed) {
      this->first_seed_ = seed;
      this->signal_seed_changed_();
    }
  }
  if (!::setting(Setting::Type::additional_party_settings))
    this->set_random_seed();
  else if (seed == UINT_MAX)
    this->set_random_seed();
  else {
    this->random_seed_ = false;
    this->seed_ = seed;
  }
} // void Party::set_seed(Seed seed)

/** sets the seed of the party to a random value
 ** (uses the system clock)
 **
 ** @todo      'better' random value
 **/
void
Party::set_random_seed() noexcept
{
  if (this->in_configuration()) {
    if (!this->random_seed_) {
      this->random_seed_ = true;
      this->signal_seed_changed_();
    }
  }

  this->seed_ = ::random_value.get();
} // void Party::set_random_seed()

/** calculates the next seed with the old one
 **
 ** @return   the new seed
 **/
Seed
Party::next_seed() noexcept
{
  // just increase the seed by one for a new cards distribution
  this->seed_ += 1;

  return this->seed_;
}

/** @return   whether the startplayer is random
 **/
bool
Party::random_startplayer() const noexcept
{
  return this->random_startplayer_;
} // bool Party::random_startplayer() const

/** @return   the startplayer
 **/
unsigned
Party::startplayer() const noexcept
{
  if (this->in_configuration()) {
    return this->first_startplayer_;
  }

  return this->startplayer_;
} // unsigned Party::startplayer() const

/** sets the first startplayer
 **
 ** @param    startplayer   number of the startplayer
 **/
void
Party::set_startplayer(unsigned const startplayer) noexcept
{
  if (this->in_configuration()) {
    if (this->first_startplayer_ != startplayer) {
      this->first_startplayer_ = startplayer;
      this->signal_startplayer_changed_();
    }
  }
  if (!::setting(Setting::Type::additional_party_settings))
    this->startplayer_ = 1;
  else if (startplayer == UINT_MAX)
    this->set_random_startplayer();
  else {
    this->random_startplayer_ = false;
    this->startplayer_ = startplayer;
  }
} // void Party::set_startplayer(unsigned const startplayer)

/** sets the startplayer to random
 **/
void
Party::set_random_startplayer() noexcept
{
  if (this->in_configuration()) {
    if (!this->random_startplayer_) {
      this->random_startplayer_ = true;
      this->signal_startplayer_changed_();
    }
  }

  this->startplayer_ = ::random_value(this->players_->size());
} // void Party::set_random_startplayer_first()

/** @return   points
 **/
int
Party::points() const noexcept
{
  return this->points_;
} // int Party::points() const

/** @return   the game
 **/
Game const&
Party::game() const noexcept
{
  return *this->game_;
} // Game Party::game() const

/** @return   the game
 **/
Game&
Party::game() noexcept
{
  return *this->game_;
} // Game Party::game()

/** the last game shall be replayed
 **/
void
Party::set_replay_game() noexcept
{
  this->replay_game_ = true;
} // void Party::set_replay_game()

/** @return   whether this game is a replayed game 
 **/
bool
Party::is_replayed_game() const noexcept
{
  return this->is_replayed_game_;
} // bool Party::is_replayed_game() const

/** @return   whether the party can be configured (seed, startplayer)
 **/
bool
Party::in_configuration() const noexcept
{
  switch (::game_status) {
  case GameStatus::programstart:
  case GameStatus::party_new:
    return true;
  case GameStatus::party_initial_loaded:
    return (this->gameno() == 0);
  default:
    return false;
  } // switch (::game_status)
} // bool Party::in_configuration() const

/** @return   the number of the current round
 **/
unsigned
Party::roundno() const noexcept
{
  if (this->round_startgame_.empty())
    return 0;

  if (::game_status == GameStatus::party_finished)
    return this->round_startgame_.size();

  return (this->round_startgame_.size() - 1);
} // unsigned Party::roundno() const

/** -> result
 ** Started rounds are not counted.
 ** If the last game of a round is finished, the round is counted as full.
 **
 ** @return   the number of remaining rounds, excluding a duty soli round
 **/
unsigned
Party::remaining_rounds() const noexcept
{
  if (::game_status == GameStatus::party_finished)
    return 0;

  if (!this->rule()(Rule::Type::number_of_rounds_limited))
    return 0;

  if (this->is_duty_soli_round())
    return 0;

  if (   (this->rule()(Rule::Type::number_of_rounds)
          == this->roundno())
      && this->starts_new_round(this->gameno()))
    return 0;

  return (this->rule()(Rule::Type::number_of_rounds)
          - this->roundno()
          - 1);
} // unsigned Party::remaining_rounds() const

/** -> result
 ** A started game is not counted as remaining.
 ** A finished game (gamestatus = game_finished) is not counted as remaining.
 **
 ** @return   the number of remaining games (without duty solo games)
 **/
unsigned
Party::remaining_normal_games() const noexcept
{
  if (!this->rule()(Rule::Type::number_of_rounds_limited))
    return 0;
  if (this->is_duty_soli_round())
    return 0;

  return (  (  this->rule()(Rule::Type::number_of_rounds)
             * this->rule()(Rule::Type::number_of_games_per_round))
          - this->finished_normal_games_);
} // unsigned Party::remaining_normal_games() const

/** @return   the number of remaining games (with duty solo games)
 **/
unsigned
Party::remaining_games() const noexcept
{
  if (!this->rule()(Rule::Type::number_of_rounds_limited))
    return 0;

  return (this->remaining_normal_games() + this->remaining_duty_soli());
} // unsigned Party::remaining_games() const

/** @return   the number of the current game
 **/
unsigned
Party::gameno() const noexcept
{
  if (::game_status == GameStatus::game_finished)
    return (this->game_summaries_.size() - 1);
  else
    return this->game_summaries_.size();
} // unsigned Party::gameno() const

/** @return   the number of finished games
 **/
unsigned
Party::finished_games() const noexcept
{
  return this->game_summaries_.size();
} // unsigned Party::finished_games() const

/** @return   whether this is the duty soli round
 **/
bool
Party::is_duty_soli_round() const noexcept
{
  if (   (game_status == GameStatus::party_finished)
      && !this->game_summaries_.empty()
      && !is_real_solo(this->last_game_summary().game_type()))
    return false;

  return (   this->duty_soli_round_
          && (this->roundno() >= *this->duty_soli_round_));
} // bool Party::is_duty_soli_round() const

/** @param    round
 **
 ** @return   whether 'round' is the duty soli round
 **/
bool
Party::is_duty_soli_round(unsigned const round) const noexcept
{
  return (   this->duty_soli_round_
          && (round == *this->duty_soli_round_));
} // bool Party::is_duty_soli_round(unsigned round) const

/** @return   whether this is the last game
 **/
bool
Party::is_last_game() const noexcept
{
  // must be in a running party
  if (!(::game_status >= GameStatus::party_play))
    return false;

  // if there are still duty soli to play
  if (this->remaining_duty_soli()
      > ( (   (::game_status > GameStatus::game_reservation)
           && (::game_status < GameStatus::game_finished)
           && this->game().is_duty_solo())
         ? 1 : 0))
    return false;

  if (   this->rule()(Rule::Type::number_of_rounds_limited)
      && (this->remaining_normal_games() == 0)
      && (this->remaining_duty_soli() == 0)
     ) {
    return true;
  }

  // last game in duty soli round
  if (this->is_duty_soli_round()
      && (this->remaining_duty_soli()
          == ((::game_status < GameStatus::game_finished)
              ? 1
              : 0) )
      && (   (::game_status <= GameStatus::game_reservation)
          || this->game().is_duty_solo() ) )
    return true;

  // the points are limited
  if (   this->rule()(Rule::Type::points_limited)
      && (this->remaining_points() <= 0)
      && (this->remaining_duty_soli() == 0) )
    return true;

  return false;
} // bool Party::is_last_game() const

/** @return   whether the party is finished
 **/
bool
Party::is_finished() const noexcept
{
  return (   (this->remaining_duty_soli() == 0)
          && (   (   this->rule()(Rule::Type::number_of_rounds_limited)
                  && (this->remaining_normal_games() == 0) )
              || (   this->rule()(Rule::Type::points_limited)
                  && (this->remaining_points() <= 0) ) ) );
} // bool Party::is_finished() const

/** @return   the remaining points
 **/
int
Party::remaining_points() const noexcept
{
  if (!this->rule()(Rule::Type::points_limited))
    return -1;

  return (static_cast<int>(this->rule()(Rule::Type::points)) - this->points());
} // unsigned Party::remaining_points() const

/** -> result
 **
 ** @param    gameno   number of the game
 **
 ** @return   whether game 'gameno' starts a new round
 **/
bool
Party::starts_new_round(unsigned const gameno) const noexcept
{
  if ((gameno == 0) || this->round_startgame_.empty())
    return true;
  if (gameno <= this->round_startgame_.back()) {
    for (auto const& s : this->round_startgame_) {
      if (s == gameno) {
        return true;
      }
      else if (s > gameno) {
        return false;
      }
    }
    return false;
  } else { // if !(gameno <= this->gameno())
    // the startgame of the next round
    unsigned next_round_startgame
      = this->round_startgame_.back() + this->players_->size();

    // check for played solo games in the current round
    for (unsigned g = this->round_startgame_.back(); g < this->gameno(); ++g)
      if (this->game_summary(g).startplayer_stays())
        next_round_startgame += 1;

    // check whether the current game is a solo
    if (   (::game_status & GameStatus::game)
        //&& (::game_status < GameStatus::game_finished)
        && this->game().startplayer_stays())
      next_round_startgame += 1;

    if (next_round_startgame > gameno)
      return false;
    if (this->rule()(Rule::Type::number_of_rounds_limited)
        && (  (this->roundno() 
               + (gameno - next_round_startgame) / this->players_->size())
            > this->rule()(Rule::Type::number_of_rounds)) )
      return false;
    return ((gameno - next_round_startgame) % this->players_->size()
            == 0);
  } // if !(gameno <= this->gameno())
} // bool Party::starts_new_round(unsigned const gameno) const

/** -> result
 **
 ** @param    gameno   number of the game
 **
 ** @return   the round game 'gameno' is in
 **/
unsigned
Party::round_of_game(unsigned const gameno) const noexcept
{
  if (gameno == 0)
    return 0;

  if (gameno == this->gameno()) {
    return this->roundno();
  }
  if (gameno < max(this->round_startgame_.back(), this->gameno())) {
    for (unsigned r = 1; r <= this->roundno(); ++r) {
      if (this->round_startgame_[r] > gameno) {
        return r - 1;
      }
    }
    return this->roundno();
  } // if (gameno <= this->gameno())

  // the startgame of the next round
  unsigned next_round_startgame
    = this->round_startgame_.back() + this->players_->size();

  // check for played solo games in the current round
  for (unsigned g = this->round_startgame_.back(); g < this->gameno(); ++g)
    if (this->game_summary(g).startplayer_stays())
      next_round_startgame += 1;

  // check whether the current game is a solo
  if (   (::game_status & GameStatus::game)
      //&& (::game_status < GameStatus::game_finished)
      && this->game().startplayer_stays())
    next_round_startgame += 1;

  if (this->is_duty_soli_round())
    next_round_startgame = gameno;
  if (gameno < next_round_startgame)
    return this->roundno();

  unsigned const round
    = (this->roundno() + 1
       + (gameno - next_round_startgame) / this->players_->size());

#ifdef OUTDATED
  // for future games, we have to estimate
  unsigned const round
    = (  (this->roundno() == 0)
       ? (gameno / this->players_->size())
       : (  (gameno * this->roundno())
          / this->round_startgame_[this->roundno()]) );
#endif

  // do not exceed the limited rounds (+ duty soli round)
  if (this->rule()(Rule::Type::number_of_rounds_limited))
    return min(round,
               this->rule()(Rule::Type::number_of_rounds)
               + ( (this->remaining_duty_soli() || this->is_duty_soli_round())
                  ? 1 : 0));
  else
    return round;
} // unsigned Party::round_of_game(unsigned const gameno) const

/** add a game summary
 ** add the points and manage the bock rounds
 **
 ** @param    game_summary   game summary to add
 **/
void
Party::add_game_summary(unique_ptr<GameSummary> game_summary_p)
{
  this->game_summaries_.push_back(move(game_summary_p));
  auto const game_summary = this->game_summaries_.back().get();
  if (!game_summary->startplayer_stays() ) {
    this->finished_normal_games_ += 1;
  }
  if (!this->bock_multipliers_.empty()
      && !game_summary->startplayer_stays() ) {
    this->bock_multipliers_.erase(this->bock_multipliers_.begin());
  }

  for (auto const& p : *this->players_) {
    if (this->rule()(Rule::Type::counting) == Counting::plus)
      this->points_ += game_summary->points(p);
    else
      this->points_ -= game_summary->points(p);
  } // for (auto const& p : *this->players_)

  // add the bock multipliers
  if (this->rule()(Rule::Type::bock)) {
    if (!game_summary->bock_triggers().empty()) {
      if (this->rule()(Rule::Type::bock_append)) {
        for (unsigned g = 0;
             g < this->rule()(Rule::Type::number_of_games_per_round);
             ++g) {
          this->bock_multipliers_.push_back(this->rule()(Rule::Type::bock_multiplier));
        }
      } else { // if !(this->rule()(Rule::Type::bock_append))
        for (unsigned g = 0;
             g < this->rule()(Rule::Type::number_of_games_per_round);
             ++g) {
          if (this->bock_multipliers_.size() > g)
            this->bock_multipliers_[g] *= this->rule()(Rule::Type::bock_multiplier);
          else
            this->bock_multipliers_.push_back(this->rule()(Rule::Type::bock_multiplier));
        } // for (g < this->rule(number_of_games_per_round))
      } // if !(this->rule()(Rule::Type::bock_append))
    } // if (!this->game().game_summaries().bock_triggers().empty())

    if (   this->rule()(Rule::Type::number_of_rounds_limited)
        && (this->bock_multipliers_.size()
            > (this->remaining_normal_games()
               + (this->remaining_duty_soli() ? 1 : 0) ) ) ) {
      this->bock_multipliers_.resize(this->remaining_normal_games()
                                     + (this->remaining_duty_soli() ? 1 : 0));
    }
  } // if (this->rule()(Rule::Type::bock)
} // void Party::add_game_summary(unique_ptr<GameSummary> game_summary)

/** @return   the game summary of the last game
 **/
GameSummary const&
Party::last_game_summary() const noexcept
{
  return *this->game_summaries_.back();
} // GameSummary const& Party::last_game_summary() const;

/** -> result
 **
 ** @param    gameno   number of the game
 **
 ** @return   the game summary of game 'gameno'
 **/
GameSummary const&
Party::game_summary(unsigned const gameno) const
{
  DEBUG_ASSERTION((gameno < this->game_summaries_.size()),
                  "Party::game_summary(gameno):\n"
                  "  gameno = " << gameno << " is too great "
                  "(maximal " << this->game_summaries_.size() << ")");
  return *this->game_summaries_[gameno];
} // GameSummary const& Party::game_summary(unsigned const gameno) const;

/** @return   the bock multipliers
 **/
vector<unsigned> const&
Party::bock_multipliers() const noexcept
{
  return this->bock_multipliers_;
} // vector<unsigned> Party::bock_multipliers() const

/** @return   the bock multiplier for the current game
 **/
unsigned
Party::current_bock_multiplier() const
{
  if (this->bock_multipliers_.empty())
    return 1;

  DEBUG_ASSERTION(this->bock_multipliers_[0] > 1,
                  "Party::current_bock_multiplier()\n"
                  "  bock multiplier is not > 1: " << this->bock_multipliers_[0]);
  return this->bock_multipliers_[0];
} // unsigned Party::current_bock_multiplier() const;

/** -> result
 **
 ** @param    gameno   game number
 **
 ** @return   the bock multiplier for the game gameno
 **/
unsigned
Party::bock_multiplier(unsigned const gameno) const noexcept
{
  if (gameno < this->finished_games())
    return this->game_summary(gameno).bock_multiplier();

  if (gameno >= this->finished_games() + this->bock_multipliers_.size())
    return 1;

  return this->bock_multipliers_[gameno - this->finished_games()];
} // unsigned Party::bock_multiplier(unsigned gameno) const

/** -> result
 **
 ** @param    roundno   number of round
 **
 ** @return   the number of the first game in the round
 **/
unsigned
Party::round_startgame(unsigned const roundno) const noexcept
{
  if (roundno <= this->roundno())
    return this->round_startgame_[roundno];
  if (roundno == 0)
    return 0;

  // if roundno > current round we have to estimate
  if (::game_status & GameStatus::game) {
    return (this->gameno()
            + (this->players_->size()
               - ((this->players_->size()
                   + this->startplayer() - this->first_startplayer_
                  ) % this->players_->size())
              )
            + (this->game().startplayer_stays() ? 1 : 0)
            + (roundno - this->roundno() - 1) * this->players_->size()
           );
  } else {
    return (this->gameno()
            + (this->last_game_summary().startplayer_stays() ? 1 : 0)
            + (this->players_->size()
               - ((this->startplayer() - this->first_startplayer_
                   + this->players_->size()) % this->players_->size())
              )
            + (roundno - this->roundno() - 1) * this->players_->size()
           );
  }
} // unsigned Party::round_startgame(unsigned const roundno) const

/** @return   how many points has been made so far
 **/
int
Party::pointsum() const
{
  if (this->game_summaries_.empty())
    return 0;
  return this->pointsum_till_game(this->game_summaries_.size() - 1);
} // int Party::pointsum() const

/** -> result
 **
 ** @param    gameno   game number
 **
 ** @return   how many points has been made till game 'gameno'
 **/
int
Party::pointsum_till_game(unsigned const gameno) const
{
  int points = 0;

  for (unsigned g = 0; g <= min(gameno, this->finished_games()); ++g)
    points += this->game_summary(g).points();

  return points;
} // int Party::pointsum_till_game(unsigned const gameno) const

/** -> result
 **
 ** @param    roundno      number of round
 **
 ** @return   how many points has been made till round 'roundno'
 **/
int
Party::pointsum_till_round(unsigned const roundno) const
{
  int points = 0;

  for (unsigned r = 0; r <= roundno; ++r)
    points += this->pointsum_in_round(r);

  return points;
} // int Party::pointsum_till_round(unsigned const roundno) const

/** -> result
 **
 ** @param    roundno      number of round
 **
 ** @return   how many points has been made in round 'roundno'
 **/
int
Party::pointsum_in_round(unsigned const roundno) const
{
  unsigned const gameno_end
    = ( (roundno == this->roundno())
       ? this->gameno() + ((::game_status == GameStatus::game_finished)
                           ? 1
                           : 0)
       : this->round_startgame(roundno + 1) );

  int points = 0;

  for (unsigned g = this->round_startgame(roundno);
       g < min(gameno_end, this->finished_games());
       ++g) {
    points += this->game_summary(g).points();
  }

  return points;
} // int Party::pointsum_in_round(unsigned const roundno) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   how many points 'player' has made so far
 **/
int
Party::pointsum(Player const& player) const
{
  return this->pointsum(this->players_->no(player));
} // int Party::pointsum(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   how many points player 'p' has made so far
 **/
int
Party::pointsum(unsigned const p) const
{
  if (this->game_summaries_.empty())
    return 0;
  return this->pointsum_till_game(this->game_summaries_.size() - 1, p);
} // int Party::pointsum(unsigned p) const

/** -> result
 **
 ** @param    gameno      number of the game
 ** @param    p   number of player
 **
 ** @return   how many points player 'p' has made till game 'gameno'
 **/
int
Party::pointsum_till_game(unsigned const gameno, unsigned const p) const
{
  int points = 0;
  auto const& player = this->players_->player(p);

  for (unsigned g = 0; g < gameno; ++g) {
    if (this->game_summary(g).has_played(player))
      points += this->game_summary(g).points(player);
  }
  if (   (gameno < this->gameno())
      || (::game_status == GameStatus::game_finished) ) {
    if (this->game_summary(gameno).has_played(player))
      points += this->game_summary(gameno).points(player);
  }

  return points;
} // int Party::pointsum_till_game(unsigned gameno, unsigned p) const

/** -> result
 **
 ** @param    roundno      number of round
 ** @param    p   number of player
 **
 ** @return   how many points player 'p' has made till round 'roundno'
 **/
int
Party::pointsum_till_round(unsigned const roundno,
                           unsigned const p) const
{
  int points = 0;

  for (unsigned r = 0; r <= roundno; ++r)
    points += this->pointsum_in_round(r, p);

  return points;
} // int Party::pointsum_till_round(unsigned roundno, unsigned p) const

/** -> result
 **
 ** @param    roundno      number of round
 ** @param    p   number of player
 **
 ** @return   how many points player 'p' has made in round 'roundno'
 **/
int
Party::pointsum_in_round(unsigned const roundno, unsigned const p) const
{
  auto const& player = this->players_->player(p);
  unsigned const gameno_end
    = ( (roundno == this->roundno())
       ? this->gameno() + ((::game_status == GameStatus::game_finished)
                           ? 1
                           : 0)
       : this->round_startgame(roundno + 1) );

  int points = 0;

  for (unsigned g = this->round_startgame(roundno);
       g < min(gameno_end, this->finished_games());
       ++g) {
    if (this->game_summary(g).has_played(player))
      points += this->game_summary(g).points(player);
  }

  return points;
} // int Party::pointsum_in_round(unsigned roundno, unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the rang of 'player'
 **/
unsigned
Party::rang(Player const& player) const
{
  return this->rang(this->players_->no(player));
} // unsigned Party::rang(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the rang of the player
 **/
unsigned
Party::rang(unsigned const p) const
{
  auto const points = this->pointsum(p);
  return std::count_if(*this->players_,
                       [points, this](Player const& player)
                       { return (points < this->pointsum(player)); }
                      );
} // unsigned Party::rang(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of duty solo games 'player' has played
 **/
unsigned
Party::played_duty_soli(Player const& player) const
{
  return this->played_duty_soli(this->players_->no(player));
} // unsigned Party::played_duty_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of duty solo games the player has played
 **/
unsigned
Party::played_duty_soli(unsigned const p) const
{
  unsigned n = 0;

  for (auto const& game_summary : this->game_summaries_) {
    if (   game_summary->is_duty_solo()
        && (game_summary->soloplayer() == p) )
      n += 1;
  }

  return min(n, this->rule()(Rule::Type::number_of_duty_soli));
} // unsigned Party::played_duty_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of duty color solo games 'player' has played
 **/
unsigned
Party::played_duty_color_soli(Player const& player) const
{
  return this->played_duty_color_soli(this->players_->no(player));
} // unsigned Party::played_duty_color_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of duty color solo games the player has played
 **/
unsigned
Party::played_duty_color_soli(unsigned const p) const
{
  unsigned n = 0;

  for (auto const& game_summary : this->game_summaries_)
    if (   game_summary->is_duty_solo()
        && is_color_solo(game_summary->game_type())
        && (game_summary->soloplayer() == p) )
      n += 1;

  return min(n, this->rule()(Rule::Type::number_of_duty_color_soli));
} // unsigned Party::played_duty_color_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of duty picture solo games 'player' has played
 **/
unsigned
Party::played_duty_picture_soli(Player const& player) const
{
  return this->played_duty_picture_soli(this->players_->no(player));
} // unsigned Party::played_duty_picture_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of duty picture solo games the player has played
 **/
unsigned
Party::played_duty_picture_soli(unsigned const p) const
{
  unsigned n = 0;

  for (auto const& game_summary : this->game_summaries_) {
    if (   game_summary->is_duty_solo()
        && is_picture_solo(game_summary->game_type())
        && (game_summary->soloplayer() == p) )
      n += 1;
  }

  return min(n, this->rule()(Rule::Type::number_of_duty_picture_soli));
} // unsigned Party::played_duty_picture_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of free duty solo games 'player' has played
 **/
unsigned
Party::played_duty_free_soli(Player const& player) const
{
  return this->played_duty_free_soli(this->players_->no(player));
} // unsigned Party::played_duty_free_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of free duty solo games the player has played
 **/
unsigned
Party::played_duty_free_soli(unsigned const p) const
{
  return (  this->played_duty_soli(p)
          - this->played_duty_color_soli(p)
          - this->played_duty_picture_soli(p));
} // unsigned Party::played_duty_free_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of lust solo games 'player' has played
 **
 ** @todo   not calculate duty soli
 **/
unsigned
Party::played_lust_soli(Player const& player) const
{
  return this->played_lust_soli(this->players_->no(player));
} // unsigned Party::played_lust_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of lust solo games the player has played
 **
 ** @todo   not calculate duty soli
 **/
unsigned
Party::played_lust_soli(unsigned const p) const
{
  return count_if(this->game_summaries_,
                  [p](auto const& gs) {
                  return (   gs->is_lust_solo()
                          && (gs->soloplayer() == p) );
                  });
} // unsigned Party::played_lust_soli(unsigned p) const

/** @return   the number of remaining duty picture solo games still to be played
 **/
unsigned
Party::remaining_duty_soli() const
{
  unsigned n = 0;
  for (auto const& p : *this->players_)
    n += this->remaining_duty_soli(p);

  return n;
} // unsigned Party::remaining_duty_soli() const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of remaining duty picture solo games 'player' has to play
 **/
unsigned
Party::remaining_duty_soli(Player const& player) const
{
  return this->remaining_duty_soli(this->players_->no(player));
} // unsigned Party::remaining_duty_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of remaining free duty solo games the player has to play
 **/
unsigned
Party::remaining_duty_soli(unsigned const p) const
{
  return (this->rule()(Rule::Type::number_of_duty_soli)
          - this->played_duty_soli(p));
} // unsigned Party::remaining_duty_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of remaining color duty picture solo games 'player' has to play
 **/
unsigned
Party::remaining_duty_color_soli(Player const& player) const
{
  return this->remaining_duty_color_soli(this->players_->no(player));
} // unsigned Party::remaining_duty_color_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of remaining color duty solo games the player has to play
 **/
unsigned
Party::remaining_duty_color_soli(unsigned const p) const
{
  return (this->rule()(Rule::Type::number_of_duty_color_soli)
          - this->played_duty_color_soli(p));
} // unsigned Party::remaining_duty_color_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of remaining picture duty picture solo games 'player' has to play
 **/
unsigned
Party::remaining_duty_picture_soli(Player const& player) const
{
  return this->remaining_duty_picture_soli(this->players_->no(player));
} // unsigned Party::remaining_duty_picture_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of remaining picture duty solo games the player has to play
 **/
unsigned
Party::remaining_duty_picture_soli(unsigned const p) const
{
  return (this->rule()(Rule::Type::number_of_duty_picture_soli)
          - this->played_duty_picture_soli(p));
} // unsigned Party::remaining_duty_picture_soli(unsigned p) const

/** -> result
 **
 ** @param    player   player
 **
 ** @return   the number of remaining free duty picture solo games 'player' has to play
 **/
unsigned
Party::remaining_duty_free_soli(Player const& player) const
{
  return this->remaining_duty_free_soli(this->players_->no(player));
} // unsigned Party::remaining_duty_free_soli(Player player) const

/** -> result
 **
 ** @param    p   number of player
 **
 ** @return   the number of remaining free duty solo games the player has to play
 **/
unsigned
Party::remaining_duty_free_soli(unsigned const p) const
{
  return (  this->remaining_duty_soli(p)
          - this->remaining_duty_color_soli(p)
          - this->remaining_duty_picture_soli(p));
} // unsigned Party::remaining_duty_free_soli(unsigned p) const

/** write the party into the output stream
 ** The current game is not written.
 **
 ** @param    ostr   output stream
 **
 ** @todo   output of the game?
 **/
void
Party::write(ostream& ostr) const
{
  if (::game_status < GameStatus::party_play) {
    ostr << "first seed = " << (this->random_seed_ ? "-"s : std::to_string(this->first_seed_)) << "\n"
      << "first startplayer = " << (this->random_startplayer_ ? "-"s : std::to_string(this->first_startplayer_)) << "\n"
      << "\n";
  } else {
    ostr << "# first seed (setting) = " << (this->random_seed_ ? "-"s : std::to_string(this->first_seed_)) << "\n"
      << "# first startplayer (setting) = " << (this->random_startplayer_ ? "-"s : std::to_string(this->first_startplayer_)) << "\n"
      << "first seed = " << this->first_seed_ << "\n"
      << "first startplayer = " << this->first_startplayer_ << "\n"
      << "\n";
  }

  ostr << "rules \n"
    << "{\n"
    << this->rule()
    << "}\n";

  ostr<< '\n';

  ostr << "players\n"
    << "{\n";
  for (auto const& player : *this->players_)
    ostr << "{\n" << player << "}\n\n";
  ostr << "}\n";

  ostr << '\n';

  ostr << "game summaries\n";
  ostr << "{\n";
  for (unsigned i = 0; i < this->game_summaries_.size(); ++i) {
    if (this->starts_new_round(i))
      ostr << "# round " << this->round_of_game(i) << '\n';
    ostr << "#  game " << i << "\n"
      << "{\n" << this->game_summary(i) << "}\n";
  } // for (i)
  ostr << "}\n";

  ostr << '\n';
  ostr << "points\n";
  ostr << "{\n";
  for (auto const& p : *this->players_)
    ostr << p.name() << ": " << this->pointsum(p) << endl;
  ostr << "}\n";
} // void Party::write(ostream& ostr) const

/** write the party points into the output stream
 **
 ** @param    ostr   output stream
 **/
void
Party::write_pointstable(ostream& ostr) const
{
  ostr << '\n';
  ostr << "party\n" ;
  ostr << "gameno: " << this->gameno() << '\n';
  ostr << "round startgame" << '\n';
  for (unsigned r = 0; r < this->round_startgame_.size(); ++r)
    ostr << r << ": " << this->round_startgame(r) << '\n';

  ostr << "game ";
  for (auto const& p : *this->players_)
    ostr << "| " << p.name() << ' ';
  ostr << "| bm ";
  ostr << "| gametype \n";
  ostr << "-----";
  for (auto const& p : *this->players_)
    ostr << "+-" << setw(p.name().length())
      << "-" << '-';
  ostr << "+----";
  ostr << "+-----\n";
  for (unsigned g = 0; g < this->gameno(); ++g) {
    if (this->starts_new_round(g)) {
      if (this->is_duty_soli_round(this->round_of_game(g)))
        ostr << "-duty";
      else
        ostr << "-" << setw(3) << this->round_of_game(g) << " ";
      for (auto const& p : *this->players_)
        ostr << "+-" << setw(p.name().length())
          << "-" << '-';
      ostr << "+----";
      ostr << "+-----\n";
    }
    ostr << setw(3) << g << "  ";
    for (auto const& p : *this->players_)
      ostr << "| " << setw(p.name().length())
        << this->game_summary(g).points(p)
        << ' ';
    ostr << "|" << setw(3);
    if (this->bock_multiplier(g) == 1)
      ostr << " ";
    else
      ostr << this->bock_multiplier(g);
    ostr << ' ';
    ostr << "| ";
    ostr << (this->game_summary(g).is_solo() ? '*' : ' ');
    if (!this->game_summary(g).bock_triggers().empty())
      ostr << this->game_summary(g).bock_triggers().size();
    ostr << '\n';
  } // for (g < this->gameno())
} // void Party::write_pointstable(ostream& ostr) const

/** reads the party from the input stream
 **
 ** @param    istr   input stream
 **
 ** @return   whether the reading was successful
 **/
bool
Party::read(istream& istr)
{
  // party config values
  optional<unsigned> first_seed = 0;
  optional<unsigned> first_startplayer = 0;
  unique_ptr<Rule> rule;
  unique_ptr<Players> players;
  vector<unique_ptr<GameSummary>> game_summaries;

  Config config; // the config read
  try {
    int depth = 0; // the depth in the blocks
    do { // while (istr)
      istr >> config;

      if (istr.eof())
        break;

      if (istr.fail())
        break;

      if (   (config.name == "")
          && (config.value == ""))
        break;

      if (config.value.empty()) {
        if (config.name == "{") {
          depth += 1;
        } else if (config.name == "}") {
          depth -= 1;
          // finished with the party
          if (depth == -1)
            istr.putback('}');
          if (depth <= 0)
            break;

        } else if (config.name == "rules") {
          if (rule)
            throw ReadException("Party read: rule already read");
          rule = make_unique<Rule>();
          if (!rule->read(istr))
            throw ReadException("Party read: rule failed");

        } else if (config.name == "players") {
          if (players)
            throw ReadException("Party read: multiple players");

          istr >> config;
          if (! (config.value.empty()
                 && (config.name == "{")) )
            throw ReadException("Party read: players: expected '{'");

          players = make_unique<Players>(*this, istr);

          istr >> config;
          if (! (config.value.empty()
                 && (config.name == "}")) )
            throw ReadException("Party read: players: expected '}'");

        } else if (config.name == "game summaries") {
          {
            string const line = read_line(istr);
            if (line != "{")
              throw ReadException("Party read: game summaries: expected '{', "
                                  " found '" + line + "'");
          }
          unique_ptr<GameSummary> game_summary;
          while (istr.good()
                 && istr.peek() != '}') {
            try {
              game_summary = make_unique<GameSummary>(*rule, istr);
            } catch(...) {
              throw;
            }
            game_summaries.push_back(move(game_summary));
          } // while (istr.good() && istr.peek() !=- '}')
          {
            string const line = read_line(istr);
            if (line != "}")
              throw ReadException("Party read: game summaries: expected '}', "
                                  " found '" + line + "'");
          }

        } else if (config.name == "points") {
          // ignore the block
          {
            string const line = read_line(istr);
            if (line != "{")
              throw ReadException("Party read: points: expected '{', "
                                  " found '" + line + "'");
          }
          while (istr.good()
                 && istr.peek() != '}') {
            Config points_config;
            istr >> points_config;
          } // while (istr.good() && istr.peek() != '}')
          {
            string const line = read_line(istr);
            if (line != "}")
              throw ReadException("Party read: points: expected '}', "
                                  " found '" + line + "'");
          }

        } else { // if (config.name == ...)
          throw ReadException("Party read: unknown line "
                              "'" + config.name + "'");
        } // if (config.name == ...)
      } else { // if !(config.value.empty())
        if (config.name == "first seed") {
          istringstream istr2(config.value);
          istr2 >> first_seed;
          if (!istr2.good() && !istr2.eof())
            throw ReadException("Party read: error loading first seed");
        } else if (config.name == "first startplayer") {
          istringstream istr2(config.value);
          istr2 >> first_startplayer;
          if (!istr2.good() && !istr2.eof())
            throw ReadException("Party read: error loading first seed");
        } else { // if (config.name == ...)
          throw ReadException("Party read: unknown line "
                              "'" + config.name + " = " + config.value + "'");
        } // if (config.name == ...)
      } // if !(config.value.empty())

    } while (istr.good());

    if (!istr.good() && !istr.eof())
      throw ReadException("Party read: unknown error in the input stream");

    if (!rule)
      throw ReadException("Party read: rule not loaded");

    if (players->size() != (*rule)(Rule::Type::number_of_players_in_game))
      throw ReadException("Party read: incorrect number of players");

  } catch (...) {
    throw;
  } // try

  ::game_status = GameStatus::party_loaded;
  this->rule_ = std::move(rule);

  this->players_ = move(players);

  if (first_seed) {
    this->random_seed_ = false;
    this->first_seed_ = *first_seed;
  } else {
    this->random_seed_ = true;
    this->first_seed_ = 0;
  }
  if (first_startplayer) {
    this->random_startplayer_ = false;
    this->first_startplayer_ = *first_startplayer;
  } else {
    this->random_startplayer_ = true;
    this->first_startplayer_ = 0;
  }

  this->game_summaries_.clear();
  this->bock_multipliers_.clear();

  if (game_summaries.empty()) {
    this->seed_ = this->first_seed_;
    this->startplayer_ = this->first_startplayer_;
    this->round_startgame_.clear();
    this->finished_normal_games_ = 0;
  } else { // if !(game_summaries.empty())
    this->first_seed_ = game_summaries.front()->seed();
    this->first_startplayer_ = game_summaries.front()->startplayer();
    this->seed_ = game_summaries.back()->seed();
    this->startplayer_ = game_summaries.back()->startplayer();
    this->finished_normal_games_ = 0;

    { // calculate round startgame
      unsigned gameno = 0;
      this->round_startgame_.clear();
      this->round_startgame_.push_back(0);
      for (auto& gs : game_summaries) {
        // check for a new round
        // ToDo: in the duty soli round not all players do play, so the check of the startplayer should be able to accept skipped players
        if (   (gameno > 0)
            && (this->last_game_summary().startplayer()
                != this->first_startplayer_)
            && (this->last_game_summary().next_startplayer()
                == this->first_startplayer_)
            && !this->is_duty_soli_round()) {
          this->round_startgame_.push_back(this->finished_games());
        }
        // check for duty soli round
        if (!this->is_duty_soli_round()
            && (   (   this->rule()(Rule::Type::number_of_rounds_limited)
                    && (this->roundno()
                        >= this->rule()(Rule::Type::number_of_rounds))
                    && !this->last_game_summary().startplayer_stays()
                    && (this->last_game_summary().next_startplayer()
                        == this->first_startplayer_)
                   )
                || (   this->rule()(Rule::Type::points_limited)
                    && (this->remaining_points() <= 0) )
               ) ) {
          this->duty_soli_round_ = this->roundno();
        }
        this->add_game_summary(move(gs));
        gameno += 1;
      } // for (gs : game_summaries)

      // We have to check for a new round because of the summary window after loading a party
      // check for a new round
      if (   (gameno > 0)
          && (this->last_game_summary().startplayer()
              != this->first_startplayer_)
          && (this->last_game_summary().next_startplayer()
              == this->first_startplayer_)
          && !this->is_duty_soli_round()) {
        this->round_startgame_.push_back(this->finished_games());
      }
      if (!this->is_duty_soli_round()
          && (   (   this->rule()(Rule::Type::number_of_rounds_limited)
                  && (this->roundno()
                      >= this->rule()(Rule::Type::number_of_rounds)) )
              || (   this->rule()(Rule::Type::points_limited)
                  && (this->remaining_points() <= 0) )
             ) ) {
        this->duty_soli_round_ = this->roundno();
      }

      if (this->is_duty_soli_round()) {
        if (   !this->game_summaries_.empty()
            && !this->last_game_summary().is_duty_solo()
            && !this->round_startgame_.empty()
            && (this->round_startgame_.back() != this->gameno())
           ) {
          this->round_startgame_.push_back(this->gameno());
        }
      } // if (this->is_duty_soli_round())
    } // calculate round startgame

    // we want to start the following game
    if (   !this->last_game_summary().startplayer_stays()
        || (   this->is_duty_soli_round()
            && this->last_game_summary().is_duty_solo())) {
      (this->startplayer_ += 1) %= this->players_->size();
    }
    this->next_seed();
  } // if !(this->game_summaries_.empty())

  return (istr.good() || istr.eof());
} // bool Party::read(istream& istr)

/** saves the party in the given file
 **
 ** @param    filename   file to save the party in
 **
 ** @return   whether the party could be saved
 **/
bool
Party::save(string const& filename) const
{
  string filename_tmp = filename + ".tmp";
  ofstream ostr(filename_tmp.c_str());
  if (!ostr.good()) {
    ::ui->information(_("Error::Party::save: Error opening temporary file %s. Aborting saving.", filename_tmp), InformationType::problem);
    return false;
  }

  ostr
    << "#!FreeDoko\n"
    << "# FreeDoko party (version " << *::version << ")\n"
    << '\n';
  this->write(ostr);

  if (!ostr.good()) {
    ::ui->information(_("Error::Party::save: Error saving in temporary file %s. Keeping temporary file (for bug tracking).", filename_tmp), InformationType::problem);
    return false;
  }
  ostr.close();

#ifdef WINDOWS
  unlink(filename.c_str());
#endif
  if (rename(filename_tmp.c_str(), filename.c_str())) {
    ::ui->information(_("Error::Party::save: Could not rename temporary file %s to requested file %s. Keeping temporary file.", filename_tmp, filename), InformationType::problem);
    return false;
  }

  return true;
} // bool Party::save(string const& filename) const

/** loads the party in the given file
 **
 ** @param    filename   file to load the party from
 **
 ** @return   whether the loading was successful
 **/
bool
Party::load(string const& filename)
{
  try {
    ifstream istr(filename.c_str());
    if (istr.fail())
      throw ReadException("Party load: open failed");

    this->read(istr);

    if (!istr.good() && !istr.eof())
      throw ReadException("Party load: unknown error");
  } catch (ReadException const& exception) {
    ::ui->information(_("Message::party load: load error '%s'",
                        filename)
                      + "\n"
                      + exception.message(),
                      InformationType::problem);
    return false;
  } // try

  return true;
} // bool Party::load(string const& filename)

/** load the players
 ** loads the files 'player.?' from the FreeDoko private directory
 **/
bool
Party::load_players()
{
  bool loaded_fail = false;
  for (auto const& p : *this->players_) {
    auto const n = this->players_->no(p);
    auto player = Player::create(::setting(Setting::Type::private_data_directory)
                                 + "/player."
                                 + String::to_string(n + 1));
    if (player)
      this->players_->set(n, move(player));
    else
      loaded_fail = true;
  } // for (p : *this->players_)

  return !loaded_fail;
} // bool Party::load_players()

/** reset the party:
 ** get the info of the first startplayer, the first seed,
 ** the ruleset and the players
 **/
void
Party::reset()
{
  for (auto const& player : *this->players_) {
    ::ui->name_changed(player);
  }

  if (ui->type() == UIType::DUMMY) {
    for (unsigned p = 0; p < this->players_->size(); ++p)
      if (this->players_->player(p).type() == Player::Type::human)
        this->players_->set_type(p, Player::Type::ai);
  } // if (::ui->type() == UIType::DUMMY)


  // reset the points
  this->points_ = 0;
  this->duty_soli_round_ = UINT_MAX;

  // save the starting values
  this->first_seed_ = this->seed_;
  this->first_startplayer_ = this->startplayer_;

  if (::setting(Setting::Type::save_party_changes))
    this->save_changes();
} // void Party::reset()

/** the party is opened
 **/
void
Party::open()
{
  ::ui->party_open(*this);
  Party::signal_open_(*this);
} // void Party::open()

/** play the party:
 ** play games until the user ends the party
 ** or the party is ended according to the rules
 **/
void
Party::play()
{
  if (this->seed() == UINT_MAX)
    this->set_seed(this->seed());
  if (this->startplayer() == UINT_MAX)
    this->set_startplayer(this->startplayer());

  this->signal_start_();
  if (this->gameno() == 0) {
    this->round_startgame_.push_back(0);
    this->signal_start_round_(this->roundno());
  }

  try {
    while (::game_status == GameStatus::party_play) {
      // check for a new round
      if (   (this->gameno() > 0)
          && (this->startplayer() == this->first_startplayer_)
          && !this->is_duty_soli_round()
          && !this->game_summaries_.back()->startplayer_stays()) {
        // new round
        // when a party is loaded, the round startgame is already set
        if (this->round_startgame_.back() < this->finished_games())
          this->round_startgame_.push_back(this->finished_games());
        this->signal_start_round_(this->roundno());
      } // if (new round)

      // check for duty soli round
      if (!this->is_duty_soli_round()
          && (   (   this->rule()(Rule::Type::number_of_rounds_limited)
                  && (this->roundno()
                      >= this->rule()(Rule::Type::number_of_rounds)) )
              || (   this->rule()(Rule::Type::points_limited)
                  && (this->remaining_points() <= 0) )
             ) ) {
        this->duty_soli_round_ = this->roundno();
      }

      if (   this->is_duty_soli_round()
          && (this->remaining_duty_soli() == 0)) {
        ::game_status = GameStatus::party_finished;
        break;
      }

      if (this->is_duty_soli_round()) {
        if (   !this->game_summaries_.empty()
            && !this->last_game_summary().is_duty_solo()) {
          if (!this->starts_new_round(this->gameno())) {
            this->round_startgame_.push_back(this->gameno());
          }
        }
        // search the next player to play a duty solo
        while (!remaining_duty_soli(this->startplayer()))
          (this->startplayer_ += 1) %= this->players().size();
      } // if (this->is_duty_soli_round())

      this->play_game();

      if (this->is_last_game()) {
        ::game_status = GameStatus::party_finished;
      }

      this->next_seed();

      if (::fast_play & FastPlay::seed_info)
        continue;

      if (   !this->last_game_summary().startplayer_stays()
          || (   this->is_duty_soli_round()
              && this->last_game_summary().is_duty_solo())) {
        (this->startplayer_ += 1) %= this->players().size();
      }

      if (this->rule()(Rule::Type::mutate)) {
        this->rule_->mutate();
      }

    } // while (game_status == GameStatus::party_play)

    this->round_startgame_.push_back(this->finished_games());
    ::game_status = GameStatus::party_finished;
  } catch (GameStatus const& gamestatus) {
    if (gamestatus != GameStatus::party_finished)
      throw;

    ::game_status = GameStatus::party_finished;
  } // try

  if (::setting(Setting::Type::save_party_changes))
    this->remove_auto_save();
} // void Party::play()

/** play a game
 **/
void
Party::play_game()
{
  // create a new game
  ::game_status = GameStatus::game_new;
  this->game_ = make_unique<Game>(*this, this->players_->player(this->startplayer()));
  try { // play the game
    this->game().play();
    if (::fast_play & FastPlay::seed_info) {
      this->game().close();

      this->game_.reset();
      if (this->seed_ == 100000) {
        exit(0);
      }
      ::game_status = GameStatus::party_play;
      return ;
    } // if (::fast_play & FastPlay::seed_info)

    DEBUG_ASSERTION((::game_status == GameStatus::game_play),
                    "Party::play():\n"
                    "  wrong gamestatus (needs 'game_play'): "
                    << ::game_status);

    ::game_status = GameStatus::game_finished;
    this->game().finish();

    if (!this->is_replayed_game()) {
      auto game_summary = make_unique<GameSummary>(this->game());
      auto const gs = game_summary.get();
      this->add_game_summary(move(game_summary));
      for (auto& p : *this->players_)
        p.db_->evaluate(p, *gs);
    }

    if (::setting(Setting::Type::save_party_changes))
      this->save_changes();

    this->game().signal_finish()();
    ::ui->game_finish();
    ::game_status = GameStatus::game_closed;
  } catch (...) { // play the game
    this->game().close();
    this->game_.reset();

    throw;
  } // try

  this->game().close();
#ifdef CHECK_RUNTIME
  cout << '\n' << ::runtime << endl;
#endif

  this->game_.reset();

  ::game_status = GameStatus::party_play;
  if (this->replay_game_) {
    this->replay_game_ = false;
    this->is_replayed_game_ = true;
    this->play_game();
    this->is_replayed_game_ = false;
  }
} // void Party::play_game()

/** the party is closed
 **/
void
Party::close()
{
  ::ui->party_close();
  this->signal_close_();

  this->round_startgame_.clear();
  this->finished_normal_games_ = 0;

  this->game_summaries_.clear();

  this->bock_multipliers_.clear();

  // remove all connections
  this->signal_get_settings_.disconnect_all();
  this->signal_loaded_.disconnect_all();
  this->signal_start_.disconnect_all();
  this->signal_start_round_.disconnect_all();
  this->signal_finish_.disconnect_all();
  this->signal_close_.disconnect_all();
  this->signal_seed_changed_.disconnect_all();
  this->signal_startplayer_changed_.disconnect_all();
} // void Party::close()

/** the rule has changed
 **
 ** @param    type        rule type to have changed
 ** @param    old_value   the old value of the rule
 **/
void
Party::rule_changed(int const type, void const* const old_value)
{
  for (auto& player : *this->players_)
    player.rule_changed(type, old_value);
} // void Party::rule_changed(int const type, void const* const old_value)

/** save the file
 **
 ** @return   whether the saving was successful
 **/
bool
Party::auto_save() const
{
  // save the party in the private data directory
  return this->save(::setting(Setting::Type::private_data_directory)
                    + "/"
                    + ::setting(Setting::Type::parties_directory)
                    + "/"
                    + "current");
} // bool Party::auto_save() const

/** remove the auto saved file
 **/
void
Party::remove_auto_save() const
{
  std::remove((::setting(Setting::Type::private_data_directory)
               + "/"
               + ::setting(Setting::Type::parties_directory)
               + "/"
               + "current").c_str());

  return;
} // void Party::remove_auto_save() const

/** save the changes (rules and players)
 **/
void
Party::save_changes() const
{
  ::ui->set_busy();
  try {
    this->auto_save();
    // save the rules in the private data directory
    this->rule().save(::setting(Setting::Type::private_data_directory)
                      + "/"
                      + ::setting(Setting::Type::rules_directory)
                      + "/"
                      + "last");

    this->players().save();
    ::ui->set_not_busy();
  } catch(...) {
    ::ui->set_not_busy();
    throw;
  } // try
} // void Party::save_changes() const

/** write the party into the output stream
 **
 ** @param    ostr   output stream
 ** @param    party   party
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Party const& party)
{
  party.write(ostr);

  return ostr;
} // ostream& operator<<(ostream& ostr, Party const& party)
