/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "rule_types.h"
#include "../basetypes.h"
#include "../game/specialpoint.h"
#include "../card/card.h"

/**
 ** controls the rules of the game/party
 **/
class Rule {
  public:
    enum class Ruleset {
      ddv,
      freedoko
    }; // enum Ruleset

    using Type = RuleType;

  public:
    Rule();
    Rule(Rule const& rule);
    Rule& operator=(Rule const& rule);

    Signal<void(int, void const*)>& signal_changed();

    // returns, whether the rule is activ or not (or the value)
    bool operator()(Type::Bool type) const;
    unsigned operator()(Type::Unsigned type) const;
    unsigned operator()(Type::UnsignedExtra type) const;
    Counting operator()(Type::Counting type) const;

    // returns, whether the gametype can be selected
    bool operator()(GameType game_type) const;
    // returns, whether the marriage selector can be selected
    bool operator()(MarriageSelector marriage_selector) const;
    // returns, whether the specialpoint is valid
    bool operator()(Specialpoint::Type special_point) const;
    // returns the number of remaining cards for the announcement
    unsigned remaining_cards(Announcement announcement) const;

    // returns, whether the dependencies of the rule are met
    bool dependencies(Type::Bool type) const;
    bool dependencies(Type::Unsigned type) const;

    // returns the setting of the rule
    bool value(Type::Bool type) const;
    unsigned value(Type::Unsigned type) const;
    Counting value(Type::Counting type) const;
    Counting const& counting() const;

    // the minimum and maximum value
    unsigned min(Type::Unsigned type) const;
    unsigned max(Type::Unsigned type) const;

    // Remark:
    // these are vectors and not sets because of compilation problems
    // all card colors
    vector<Card::Color> const& card_colors() const;
    // all card values (depending on rule 'with nines') in descending order 
    vector<Card::Value> const& card_values() const;
    // all cards (depending on rule 'with nines')
    vector<Card> const& cards() const;

    // sets the rule
    void set(Type::Bool type, bool value);
    void set(Type::Bool type, string const& value);
    void set(Type::Unsigned type, unsigned value);
    void set(Type::Unsigned type, string const& value);
    void set(Type::Counting type, Counting value);
    void set(Type::Counting type, string const& value);

    bool set(string const& type, string const& value);

    // set to a ruleset
    void set_to(Ruleset const& ruleset);

    // mutate the rule
    void mutate();

    // load the rules
    bool load(string const& filename);
    bool read(istream& istr);
    // save the rules in the given file
    bool save(string const& filename) const;

    // output of the rules
    void write(ostream& ostr) const;

  private:
    // set to the default values
    void reset_to_hardcoded();
  public:
    // set to the values for the seed statistics
    void reset_to_seed_statistics();

  private:
    vector<bool> bool_;
    vector<unsigned> unsigned_;
    Counting counting_ = Counting::plus;

    Signal<void(int, void const*)> signal_changed_;
}; // class Rule

// output of the Rule
ostream& operator<<(ostream& ostr, Rule const& rule);

// comparison of two rules (value)
bool operator==(Rule const& rule1, Rule const& rule2);
bool operator!=(Rule const& rule1, Rule const& rule2);
