/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "ai.h"

#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

/** constructor
 **/
AiRandom::AiRandom() :
  Player(Player::Type::ai_random)
{ }

/** construktor
 **
 ** @param    istr   stream with the infos
 **
 ** @todo   all
 **/
AiRandom::AiRandom(istream& istr) :
  Player(Player::Type::ai_random)
{
  (void)istr;
  DEBUG_ASSERTION(false,
		  "AiRandom::AiRandom(istr):\n"
		  "  Contructor not implemented");
} // AiRandom::AiRandom(istream& istr)

/** copy constructor
 **
 ** @param    player   player to copy
 **/
AiRandom::AiRandom(Player const& player) :
  Player(player),
  Aiconfig()
{
  this->set_type(Player::Type::ai_random);
} // AiRandom::AiRandom(Player player)

/** copy constructor
 **
 ** @param    player   player to copy
 ** @param    aiconfig   configuration to copy
 **/
AiRandom::AiRandom(Player const& player, Aiconfig const& aiconfig) :
  Player(player),
  Aiconfig(aiconfig)
{
  this->set_type(Player::Type::ai_random);
} // AiRandom::AiRandom(Player player, Aiconfig aiconfig)

/** clone the player
 **
 ** @return   pointer of a clone
 **/
unique_ptr<Player>
AiRandom::clone() const
{  
  return make_unique<AiRandom>(*this);
} // virtual unique_ptr<Player> AiRandom::clone() const

/** -> result
 ** (test of poverty, marriage, swines, hyperswines)
 **
 ** @param    duty_solo   whether the player has to play a solo
 **    (default: false)
 **
 ** @return   simple Reservation
 **/
Reservation const&
AiRandom::reservation_get(bool const duty_solo)
{
  this->reservation() = this->Player::reservation_get_default();

  if (   duty_solo
      || (::random_value(this->game().playerno() * this->game().playerno()) == 0) ) {
    vector<GameType> soli;
    for (auto t : game_type_list) {
      if (is_real_solo(t)
          && this->game().rule()(t) ) {
        soli.push_back(t);
      }
    } // for (t)
    if (!soli.empty()) {
      this->reservation().game_type = soli[::random_value(soli.size())];
    }
  } // if (duty_solo)

  return this->reservation();
} // Reservation const& AiRandom::reservation_get(bool duty_solo)

/** -> result:
 ** take a random card of all valid ones
 **
 ** @return   valid card
 **/
HandCard
AiRandom::card_get()
{
  HandCard card;
  do {
    card = this->hand().card(::random_value(this->hand().cardsnumber()));
  } while (!this->game().tricks().current().isvalid(card));

  return card;
} // HandCard AiRandom::card_get()

/** @return   announcement of the player
 **/
Announcement
AiRandom::announcement_request() const
{
  // points of the own team
  unsigned points = 0;

  // count the points, the own team has made 
  for(auto const& player : this->game().players()) {
    if (this->game().teaminfo().get(player) == this->team())
      points += this->game().points_of_player(player);
  }

  if (points == 240)
    return Announcement::no0;
  if (points > 240 - 30)
    return Announcement::no30;
  if (points > 240 - 60)
    return Announcement::no60;
  if (points > 240 - 90)
    return Announcement::no90;
  if (points > 240 - 120)
    return Announcement::no120;

  return Announcement::noannouncement;
} // Announcement AiRandom::announcement_request() const

/** @return   NULL (no new partner)
 **/
Player const*
AiRandom::genscher_partner()
{ 
  return {};
} // Player const* AiRandom::genscher_partner()

/** @return   shifted cards
 **/
HandCards
AiRandom::poverty_shift()
{
  HandCards hand_cards = this->hand().cards();
  HandCards cards_to_shift;

  // first add all trump
  for (auto c = hand_cards.begin(); c != hand_cards.end(); ) {
    if (c->istrump()) {
      cards_to_shift.push_back(*c);
      c = hand_cards.erase(c);
    } else {
      ++c;
    }
  } // for (c : hand_cards)

  // then fill up with cards randomly chosen
  while (cards_to_shift.size() < this->hand().count_poverty_cards()) {
    auto const i = ::random_value(hand_cards.size());
    cards_to_shift.push_back(hand_cards[i]);
    hand_cards.erase(hand_cards.begin() + i);
  }

  this->sorted_hand().remove(cards_to_shift);

  return cards_to_shift;
} // HandCards AiRandom::poverty_shift()

/** -> result
 **
 ** @param    cardno   number of shifted cards (ignored)
 **
 ** @return   false (do not accept the poverty)
 **/
bool
AiRandom::poverty_take_accept(unsigned const cardno)
{
  return false;
} // bool AiRandom::poverty_take_accept(unsigned cardno)


// the player changes the cards of the poverty
/** should not be called
 **
 ** @param    cards   the cards the player takes
 **
 ** @return   empty cards
 **/
HandCards
AiRandom::poverty_cards_change(HandCards const& cards)
{
  DEBUG_ASSERTION(false,
                  "AiRandom::poverty_cards_change(cards)\n"
                  "  function should not be called!");
  return HandCards();
} // HandCards AiRandom::poverty_cards_change(HandCards cards)

/** the poverty player gets 'cards'
 **
 ** @param    cards   cards the player gets back
 **/
void
AiRandom::poverty_cards_get_back(HandCards const& cards)
{
  this->sorted_hand().add(cards);
} // void AiRandom::poverty_cards_get_back(HandCards cards)
