/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#pragma once

#include "../aiconfig.h"
#include "heuristic_rationale.h"
class Ai;
class TeamInformation;
class CardsInformation;
class Player;
class Hand;
class Trick;
class Card;

/** Base class for heuristics
 **/
class Heuristic {
  public:
    using Rationale = HeuristicRationale;
  public:
    class CallFunction; // Übergangslösung
    class PfundInFirstColorRun;
    class PlayHighestColor;
    class PlayTrump;
    class PlayToJabLater;
    class PlayBadColor;
    class JabColorTrickWithDolle;
    class SaveDolle;
    class ServeTrumpTrick;
    class StartWithLowTrump;
    class JabColorOverFox;
    class JabForDoppelkopf;
    class TryForDoppelkopf;
    class DrawTrump;
    class MeatlessPlayHighestColor;
    class PictureSecondBestTrump;
    class ColorJabForAce;
    class BestWinningCard;
  public:
    static unique_ptr<Heuristic> create(Aiconfig::Heuristic heuristic,
                                        Ai const& ai);

    Heuristic(Ai const& ai, Aiconfig::Heuristic heuristic);
    Heuristic(Heuristic const&) = delete;
    Heuristic& operator=(Heuristic const&) = delete;

    virtual ~Heuristic();

    void set_player(Player const& player);
    Player const& player() const;

    Aiconfig::Heuristic heuristic() const;
    string const& name() const;
    Rationale const& rationale() const;
    virtual Card get_card(Trick const& trick) = 0;

  protected: // Zugriffsfunktionen
    Hand const& hand() const;
    Team team() const;

  protected: // Hilfsfunktionen
    bool only_own_team_behind(Trick const& trick) const;
    bool soloplayer_behind(Trick const& trick) const;
    bool same_team(Player const& player) const;
    bool guessed_same_team(Player const& player) const;
    unsigned color_runs(Card::TColor color) const;

    bool value(Aiconfig::Type::Bool type) const;
    int value(Aiconfig::Type::Int type) const;
    Card value(Aiconfig::Type::Card type) const;

    Card trump_card_limit() const;

    TeamInformation const& team_information() const;
    Team guessed_team(Player const& player) const;
    CardsInformation const& cards_information() const;
    Hand const& guessed_hand(Player const& player) const;

    Card lowest_best_trump_card(Trick const& trick) const;

    bool condition_solo_game(Trick const& trick);
    bool condition_solo_meatless(Trick const& trick);
    bool condition_solo_picture(Trick const& trick);
    bool condition_solo_color(Trick const& trick);
    bool condition_soloplayer_behind(Trick const& trick);

    bool condition_startcard(Trick const& trick);
    bool condition_second_card(Trick const& trick);
    bool condition_second_last_card(Trick const& trick);
    bool condition_last_card(Trick const& trick);
    bool condition_first_color_run(Trick const& trick);
    bool condition_color_trick(Trick const& trick);
    bool condition_pure_color_trick(Trick const& trick);
    bool condition_startcard_is(Trick const& trick, Card card);

    bool condition_partner_guessed(Trick const& trick);
    bool condition_last_player_guessed_same_team(Trick const& trick);

    bool condition_opposite_team_announcement();

    bool condition_has_color();
    bool condition_has_trump();
    bool condition_has(Card::Color color);
    bool condition_has_ten_points_in_trump();
    bool condition_has_dolle();

    bool condition_opposite_team_behind(Trick const& trick);
    bool condition_opposite_swine_or_hyperswine(Trick const& trick);
    bool condition_opposite_swine_or_hyperswine_behind(Trick const& trick);
    bool condition_can_be_jabbed_by_opposite_team(Trick const& trick, Card card);
    bool condition_opposite_team_has_trump();

    bool condition_trump_trick(Trick const& trick);
    bool condition_winnerplayer_same_team(Trick const& trick);
    bool condition_jabbed_with_trump(Trick const& trick);
    bool condition_own_team_jabbed_with_trump_card_limit(Trick const& trick);
    bool condition_can_jab(Trick const& trick);
    bool condition_can_jab_with_trump(Trick const& trick);
    bool condition_can_jab_with(Trick const& trick, Card card);
    bool condition_enough_trick_points_for_doppelkopf(Trick const& trick);

  protected:
    Ai const* ai_;
    Player const* player_;
    Aiconfig::Heuristic heuristic_;
    string const name_;
    Rationale rationale_;
}; // class Heuristic
