/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/ 

#include "constants.h"

#include "heuristics.h"

#define DEBUG_HEURISTIC \
  if (   (   ::debug.filter("heuristics") \
          || ::debug.filter(name)) \
      && !hi.game().isvirtual()) \
cout << __FILE__ << '#' << __LINE__ << " (" << trick.actcardno() << ')' << hi.player().name() << ": " << name << " – "


// debugging information
#ifndef RELEASE
#ifdef DKNOF
// best jabbing card (card selection)
//#define DEBUG_BJC_OSTR cout << "[#" << __LINE__ << "] "
#endif
#endif
#ifndef DEBUG_BJC_OSTR
#define DEBUG_BJC_OSTR null_ostr
#endif

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

unsigned number_of_no_trumps_per_color();
Card lowest_card(Hand const& hand, Card::Value const cv,
                 HeuristicInterface const& hi);
Card lowest_jack(Hand const& hand, HeuristicInterface const& hi);
Card lowest_queen(Hand const& hand, HeuristicInterface const& hi);
Card highest_queen(Hand const& hand, HeuristicInterface const& hi);
Card highest_card(Hand const& hand, Card::Value const cv,
                  HeuristicInterface const& hi);
Card highest_trump_card( Trick const& t, Hand h );
bool better_points_optimize(HandCard const& old_card,
                            HandCard const& test_card,
                            HeuristicInterface const& hi);

// Checks if the card will win the trick
bool cardWillWinTheTrick(Trick const &t, HeuristicInterface const& hi,
                         HandCard const& card);
unsigned calcPointsOfOwnTeam( HeuristicInterface const& hi, Game const& g, bool const with_current_trick = true );
unsigned calcPointsOfOppositeTeam( HeuristicInterface const& hi, Game const& g, bool const with_current_trick = true );

// whether all players behind are of the own team
bool checkAllMyTeam(Trick const& t, HeuristicInterface const& hi);
// whether all players behind are of the opposite team
bool checkAllOppositeTeam(Trick const& t, HeuristicInterface const& hi);
// whether the opposite team behind still has a trump
bool opposite_trump_behind(Trick const& trick, HeuristicInterface const& hi);
// the highest card of the playes of the opposite team behind
HandCard const& highest_card_behind_of_opposite_team(Trick const& trick,
                                                     HeuristicInterface const& hi);
// the valid cards of the playes of the opposite team behind
HandCards valid_cards_behind_of_opposite_team(Trick const& trick,
                                              HeuristicInterface const& hi);
// the highest card of the playes of the same team behind
HandCard const& highest_card_behind_of_same_team(Trick const& trick,
                                                 HeuristicInterface const& hi);
// the lowest of the highest trumps
HandCard lowest_highest_trump(Trick const& trick, HeuristicInterface const& hi);
// the highest trump of the playes of the opposite team
HandCard const& highest_trump_of_opposite_team(HeuristicInterface const& hi);
// the number of tricks the team will get
unsigned secure_tricks_of_team(HeuristicInterface const& hi);



namespace BestWinningCard {
  Card poverty_re(Trick const& trick,
                  HeuristicInterface const& hi);
  Card poverty_contra(Trick const& trick,
                      HeuristicInterface const& hi);
  Card color_solo_re(Trick const& trick,
                     HeuristicInterface const& hi);
  Card color_solo_contra(Trick const& trick,
                         HeuristicInterface const& hi);
  Card single_picture_solo_re(Trick const& trick,
                              HeuristicInterface const& hi);
  Card single_picture_solo_contra(Trick const& trick,
                                  HeuristicInterface const& hi);
  Card double_picture_solo_re(Trick const& trick,
                              HeuristicInterface const& hi);
  Card double_picture_solo_contra(Trick const& trick,
                                  HeuristicInterface const& hi);
  Card koehler_solo_re(Trick const& trick,
                       HeuristicInterface const& hi);
  Card koehler_solo_contra(Trick const& trick,
                           HeuristicInterface const& hi);
  Card meatless_solo_re(Trick const& trick,
                        HeuristicInterface const& hi);
  Card meatless_solo_contra(Trick const& trick,
                            HeuristicInterface const& hi);
}; // namespace BestWinningCard


/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   a card to jab the trick
 **
 ** @todo      aiconfig aggressive
 **/
Card 
Heuristics::best_winning_card(Trick const& trick,
                                  HeuristicInterface const& hi)
{
  constexpr auto name = "best winning card";
  // @heuristic::name   best winning card new
  // @heuristic::idea   take a card to win the trick

  Game const& game = hi.game();

  // @heuristic::condition   not first player
  if (trick.isempty()) {
    DEBUG_HEURISTIC << "startcard\n";
    return {};
  }

  // @heuristic::condition   can overjab the trick
  if (!hi.hand().can_jab(trick)) {
    DEBUG_HEURISTIC << "cannot overjab\n";
    return {};
  }

  // call the 
  switch (game.type()) {
  case GameType::normal:
    break;
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
  case GameType::fox_highest_trump:
  case GameType::redistribute:
    // should not happen
    break;
  case GameType::poverty:
    if (hi.team() == Team::re) {
      if (hi.same_player(game.players().soloplayer())) {
        DEBUG_HEURISTIC << "poverty player\n";
        return {};
      }
      return BestWinningCard::poverty_re(trick, hi);
    } else {
      return BestWinningCard::poverty_contra(trick, hi);
    }
  case GameType::genscher:
    break;
  case GameType::marriage:
    // do not jab a decision trick as bride
    if (   hi.same_player(game.players().soloplayer())
        && is_selector(trick.startcard().tcolor(),
                       game.marriage().selector()) ) {
        DEBUG_HEURISTIC << "bride and decision trick\n";
      return {};
    }
    break;
  case GameType::marriage_solo:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::color_solo_re(trick, hi)
            : BestWinningCard::color_solo_contra(trick, hi) );
  case GameType::marriage_silent:
    if (hi.team() == Team::re)
      return BestWinningCard::color_solo_re(trick, hi);
    break;
  case GameType::solo_meatless:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::meatless_solo_re(trick, hi)
            : BestWinningCard::meatless_solo_contra(trick, hi) );
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::single_picture_solo_re(trick, hi)
            : BestWinningCard::single_picture_solo_contra(trick, hi) );
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::double_picture_solo_re(trick, hi)
            : BestWinningCard::double_picture_solo_contra(trick, hi) );
  case GameType::solo_koehler:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::koehler_solo_re(trick, hi)
            : BestWinningCard::koehler_solo_contra(trick, hi) );
  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_diamond:
    return ( (hi.team() == Team::re)
            ? BestWinningCard::color_solo_re(trick, hi)
            : BestWinningCard::color_solo_contra(trick, hi) );
  } // switch (game.type())

  // 1) startplayer
  // 2) second player
  // 3) second last player
  // 4) last player

  // card to choose
  HandCard card = HandCard(hi.hand(), best_jabbing_card(trick, hi));
  DEBUG_HEURISTIC << "best jabbing card: " << card << '\n';
  if (card == HandCard::empty)
    return {};
  DEBUG_ASSERTION(trick.isjabbed(card),
                  "Heuristics::best_wining_card(trick, hi)\n"
                  "  card = " << card << " = jabbing_card(trick, hi) does not jab the trick:\n" << trick);

  // 2), 3), color trick, has color
  if (   !trick.startcard().istrump()
      && hi.hand().contains(trick.startcard().color())) {
    DEBUG_HEURISTIC << "pos 2/3, color trick, has color\n";
    auto const color = trick.startcard().color();
    // the opposite team behind me has the color
    for (unsigned c = trick.actcardno() + 1; c < game.playerno(); ++c) {
      auto const& p = trick.player_of_card(c);
      if (!hi.guessed_same_team(p)
          && (   !hi.handofplayer(p).contains(color)
              || card.less(hi.handofplayer(p).highest_card(color))) ) {
        DEBUG_HEURISTIC << "opposite team behind does not have the color\n";
        return {};
      }
    } // for (c)

    // There are enough remaining color cards
    if (hi.cards_information().remaining_others(color) + 1
        < trick.remainingcardno()) {
      DEBUG_HEURISTIC << "not enough color cards\n";
      return {};
    }

    DEBUG_HEURISTIC << "select " << card << "\n";
    return card;
  } // 2), 3), color trick, has color

  // 2), 3), color trick, first round, no color
  // 'choose pfund' must be selected before
  if (   !trick.startcard().istrump()
      && (hi.color_runs(trick.startcard().tcolor()) == 0)
      && !hi.hand().contains(trick.startcard().tcolor())) {
    DEBUG_HEURISTIC << "pos 2/3, color trick, first round, have not color\n";
    // if the trick already goes to my team and the opposite team does have the color, leave it
    if (   hi.guessed_same_team(trick.winnerplayer())
        && (   (trick.winnercard().value() == Card::ace)
            || trick.winnercard().istrump())
        && hi.handofplayer(trick.lastplayer()
                          ).contains(trick.startcard().tcolor())
        && (   (trick.actcardno() == 2)
            || (hi.handofplayer(trick.player_of_card(2)
                               ).contains(trick.startcard().tcolor()))) ) {
      DEBUG_HEURISTIC << "trick goes to my team " << card << "\n";
      return {};
    }

    DEBUG_HEURISTIC << "select " << card << "\n";
    return card;
  } // 2), 3), color trick, first round, no color

  // 4)
  if (trick.islastcard()) {
    DEBUG_HEURISTIC << " 4) lastplayer\n";
    // do not jab if the partner gets the trick and does not get at the last position
    if (   hi.guessed_same_team(trick.winnerplayer())
        && !hi.guessed_same_team(game.players().previous(hi.player()))) {
      DEBUG_HEURISTIC << "   partner gets the trick\n";
      return {};
    }

    // stay in the back, if the partner gets the trick
    if (   (trick.winnerplayer() == trick.startplayer())
        && hi.guessed_same_team(trick.startplayer())) {
      DEBUG_HEURISTIC << "   partner gets the trick\n";
      return {};
    }
    if (!hi.guessed_same_team(trick.winnerplayer())) {
      // jab a trick according to the aiconfig
      if (   (   (card.value() == Card::queen)
              || (   (card == Card::dolle)
                  && (   (   game.swines().swines_announced()
                          && !hi.guessed_same_team(game.swines().swines_owner()))
                      || (   game.swines().hyperswines_announced()
                          && !hi.guessed_same_team(game.swines().hyperswines_owner()))
                     )
                 )
             )
          && (static_cast<int>(trick.points())
              >= hi.value(Aiconfig::Type::limitqueen)) ) {
        DEBUG_HEURISTIC << "   points >= limitqueen: take " << card << "\n";
        return card;
      }
      if (static_cast<int>(trick.points()) >= hi.value(Aiconfig::Type::limitdolle)) {
        DEBUG_HEURISTIC << "   points >= limitdolle: take " << card << "\n";
        return card;
      }
    } // if (!hi.guessed_same_team(trick.winnerplayer()))

    // not many trumps
    if (   (hi.hand().count_trumps() * 2 <= hi.hand().cardsnumber())
        && (hi.cards_information().higher_cards_no_of_others(hi.hand().highest_trump())
            >= 2 * hi.hand().count_trumps()) ) {
      DEBUG_HEURISTIC << "   get a trick with few and low trumps: take " << card << "\n";
      return card;
    }

    // do not jab a trick with few points
    if (   (hi.hand().rel_pos_trump(card) >= 0.75)
        && (   (trick.points() <= 4)
            || (   (trick.points() <= 6)
                && hi.guessed_same_team(game.players().previous(trick.winnerplayer())))
           )
       ) {
      DEBUG_HEURISTIC << "   too few points\n";
      return {};
    }

    // jab a trick, if there are enough trumps
    if (hi.hand().rel_pos_trump(card)
        <= ((trick.points() / 20.0)
            + (hi.value(Aiconfig::Type::aggressive) ? 0.15 : 0)
            + (hi.guessed_same_team(trick.winnerplayer()) ? 0 : 0.3))) {
      DEBUG_HEURISTIC << "   enough trumps: take " << card << "\n";
      return card;
    }

    DEBUG_HEURISTIC << "   do not jab\n";
    return {};
  } // 4)

  if (trick.remainingcardno() == 2) { // 3)
    if (hi.guessed_same_team(trick.lastplayer())) { // 3) + partner behind
      DEBUG_HEURISTIC << "  3), partner behind\n";
      // show myself
      if (   (game.type() == GameType::normal)
          // ToDo: check the own gameplay for team informations
          && (game.teaminfo().get(hi.player()) == Team::unknown)
          && (card == Card::club_queen)) {
      DEBUG_HEURISTIC << "      show my team: take " << card << "\n";
        return card;
      }
      // do not jab, if the partner behind has a higher card and I do not have very many high cards
      if (   hi.handofplayer(trick.lastplayer()
                            ).higher_card_exists(trick.winnercard())
          && (hi.hand().higher_cards_no(card) + hi.hand().count(card)
              < (hi.hand().count_trumps() / 2
                 + (hi.value(Aiconfig::Type::aggressive) ? 0 : 1)))
          && (hi.hand().count_trumps()
              > 3 + hi.hand().higher_cards_no(card)) ) {
        DEBUG_HEURISTIC << "   partner behind has a higher card and I do not have very many high cards\n";
        return {};
      }
      DEBUG_HEURISTIC << "   take " << card << "\n";
      return card;
    } // 3) + partner behind
    else { // 3) + opposite behind
      if (hi.guessed_same_team(trick.winnerplayer())) {
        DEBUG_HEURISTIC << "  3), opposite player behind, own team has trick\n";
        // only jab the trick, if the partner has played a small card or there are many points in the trick
        if (trick.isjabbed(Card::diamond_jack)) {
          DEBUG_HEURISTIC << "  jab with low trump: take " << card << "\n";
          return card;
        }
        if (   (static_cast<int>(trick.points())
                >= hi.value(Aiconfig::Type::limitqueen))
            && (   trick.isjabbed(Card::diamond_queen)
                || (hi.hand().rel_pos_trump(card) <= 0.4)
                || (hi.hand().higher_cards_no(card) + hi.hand().count(card)
                    >= (hi.hand().count_trumps() / 2
                        + (hi.value(Aiconfig::Type::aggressive) ? 0 : 1))
                   ) ) ) {
          DEBUG_HEURISTIC << "  jab limitqueen: take " << card << "\n";
          return card;
        }
        // jab to get the partner in backhand
        if (   hi.guessed_same_team(game.players().previous(hi.player()))
            && (   !hi.guessed_same_team(trick.winnerplayer())
                || trick.winnercard().less(hi.trump_card_limit()) )
            && (hi.hand().higher_cards_no(card) + hi.hand().count(card)
                >= (hi.hand().count_trumps() / 2 
                    + (hi.value(Aiconfig::Type::aggressive) ? 0 : 1))) ) {
          DEBUG_HEURISTIC << "  get partner in backhand: take " << card << "\n";
          return card;
        }
        return {};
      } else { // if !(hi.guessed_same_team(trick.winnerplayer()))
        DEBUG_HEURISTIC << "  3), opposite player behind, opposite team has trick\n";
        // use a club queen
        if (   (card == Card::club_queen)
            && (hi.game().teaminfo().get(hi.player()) == Team::unknown)
            && (hi.guessedteamofplayer(trick.lastplayer()) == Team::unknown) ) {
          DEBUG_HEURISTIC << "  try club queen for partner behind: take " << card << "\n";
          return card;
        }
        // do not jab a small trick with a high card
        if (    (trick.points() >= 10)
            && (hi.hand().count_ge(card) >= 2) ) {
          DEBUG_HEURISTIC << "  jab big trick with not so high card: take " << card << "\n";
          return card;
        }
        if (   (static_cast<int>(trick.points())
                < hi.value(Aiconfig::Type::limitqueen))
            && hi.trump_card_limit().less(card)) {
          DEBUG_HEURISTIC << "  points < limitqueen, but queen selected\n";
          return {};
        }
        if (   (static_cast<int>(trick.points())
                < hi.value(Aiconfig::Type::limitdolle))
            && Card::club_queen.less(card)) {
          DEBUG_HEURISTIC << "  points < limitdolle, but dolle selected\n";
          return {};
        }
        DEBUG_HEURISTIC << "  take " << card << "\n";
        return card;
      } // if !(hi.guessed_same_team(trick.winnerplayer()))
    } // 3) + opposite behind
  } // 3)

  if (trick.actcardno() == 1) { // 2)
    if (   hi.guessed_same_team(trick.player_of_card(2))
        && !hi.guessed_same_team(trick.player_of_card(3)) ) {
      // 2), 3: same team, 4: opposite team
      DEBUG_HEURISTIC << "  2), 3: same team, 4: opposite team\n";
      // if card is not so high and the partner has a higher card, let him overjab
      if (   card.less(hi.trump_card_limit())
          && hi.handofplayer(trick.player_of_card(2)).higher_card_exists(card)) {
        DEBUG_HEURISTIC << "  let the partner have the trick\n";
        return {};
      }
      if (   !card.less(hi.trump_card_limit())
          && (hi.hand().count_ge(card) >= hi.hand().count_trumps() / 2) ) {
        DEBUG_HEURISTIC << "  have many high trumps: take " << card << "\n";
        return card;
      }
      if (   !card.less(hi.trump_card_limit())
          && (trick.points() >= 10) ) {
        DEBUG_HEURISTIC << "  many points: take " << card << "\n";
        return card;
      }
      // ToDo
#ifdef OUTDATED
      DEBUG_HEURISTIC << "  best winning card old\n";
      return best_winning_card_old(trick, hi);
#endif
    } // 2), 3: same team, 4: opposite team
    else if (   !hi.guessed_same_team(trick.player_of_card(2))
             && hi.guessed_same_team(trick.player_of_card(3)) ) {
      // 2), 3: opposite team, 4: same team
      DEBUG_HEURISTIC << "  2), 3: opposite team, 4: same team\n";
      if (   !card.less(hi.trump_card_limit())
          && (hi.hand().count_le(card) <= hi.hand().count_trumps() / 3) ) {
        DEBUG_HEURISTIC << "  have many high trumps: take " << card << "\n";
        return card;
      }
      // ToDo
#ifdef OUTDATED
      DEBUG_HEURISTIC << "  best winning card old\n";
      return best_winning_card_old(trick, hi);
#endif
    } // 2), 3: opposite team, 4: same team
    else { // 2, two times opposite team behind
      DEBUG_HEURISTIC << "  2), 3: opposite team, 4: opposite team\n";
      // always jab many points
      if (trick.points() >= 10) {
        DEBUG_HEURISTIC << "  jab many points: take " << card << "\n";
        return card;
      }
      // always jab for a high announcement
      if (game.announcements().last(hi.team()) > Announcement::no120) {
        DEBUG_HEURISTIC << "  jab for high announcement: take " << card << "\n";
        return card;
      }
      // show team with club queen
      if (   (card == Card::club_queen)
          && (game.teaminfo().get(hi.player()) == Team::unknown)) {
        DEBUG_HEURISTIC << "  show team with club queen: take " << card << "\n";
        return card;
      }
      // let the trick pass, if the player in the front has played a high card
      if (hi.trump_card_limit().less(trick.startcard()) ) {
        DEBUG_HEURISTIC << "  the startplayer has already played a high card\n";
        return {};
      }
      // jab, if the team has made an annoucement
      if (!!game.announcements().last(hi.team())) {
        DEBUG_HEURISTIC << "  jab because of announcement: take " << card << "\n";
        return card;
      }
      // as aggressive: jab, if the player has enough high trumps
      if (hi.value(Aiconfig::Type::aggressive)
          && (hi.hand().higher_cards_no(card) + hi.hand().count(card)
              >= hi.hand().count_trumps() / 2)
          && (hi.hand().count_trumps() >= 4) ) {
        DEBUG_HEURISTIC << "  jab because of aggressive: take " << card << "\n";
        return card;
      }

      // ToDo
#ifdef OUTDATED
      DEBUG_HEURISTIC << "  best winning card old\n";
      return best_winning_card_old(trick, hi);
#endif
    } // 2), two times opposite team behind
  } // 2)

  DEBUG_HEURISTIC << "  nothing matched\n";
  return {};
} // Card Heuristics::best_winning_card(Trick trick, HeuristicInterface hi)


/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a poverty
 **
 ** @todo      all
 **/
Card
BestWinningCard::poverty_re(Trick const& trick,
                            HeuristicInterface const& hi)
{
  if (trick.islastcard())
    return Heuristics::best_jabbing_card(trick, hi);
  return Card::empty;
} // Card BestWinningCard::poverty_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a poverty
 **
 ** @todo      all
 **/
Card
BestWinningCard::poverty_contra(Trick const& trick,
                                HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::poverty_best_winning_card(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::poverty_contra(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a color solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::color_solo_re(Trick const& trick,
                               HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::color_solo_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a color solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::color_solo_contra(Trick const& trick,
                                   HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::color_solo_contra(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a single picture solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::single_picture_solo_re(Trick const& trick,
                                        HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::single_picture_solo_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a single picture solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::single_picture_solo_contra(Trick const& trick,
                                            HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::single_picture_solo_contra(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a double picture solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::double_picture_solo_re(Trick const& trick,
                                        HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::double_picture_solo_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a double picture solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::double_picture_solo_contra(Trick const& trick,
                                            HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::double_picture_solo_contra(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a koeher solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::koehler_solo_re(Trick const& trick,
                                 HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::koehler_solo_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a koeher solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::koehler_solo_contra(Trick const& trick,
                                     HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::koehler_solo_contra(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the re player in a meatless solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::meatless_solo_re(Trick const& trick,
                                  HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::meatless_solo_re(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   best winning card for the contra player in a meatless solo
 **
 ** @todo      all
 **/
Card
BestWinningCard::meatless_solo_contra(Trick const& trick,
                                      HeuristicInterface const& hi)
{
#ifndef OUTDATED
  return Heuristics::best_winning_card_old(trick, hi);
#endif
  return Card::empty;
} // Card BestWinningCard::meatless_solo_contra(Trick trick, HeuristicInterface hi)


/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 ** @param    jab_always   whether to jab always -- else return empty if jabbing is senseless (default: false)
 **
 ** @return   the "best" card to jab the trick
 **
 ** @todo      aiconfig aggressive
 **/
Card
Heuristics::best_jabbing_card(Trick const& trick,
                              HeuristicInterface const& hi,
                              bool const jab_always)
{
  // notes
  // * first trick: duty announcement
  // * color trick: not below a fox 
  // * re: fox must not be saved (beware of dollen)
  // * small points: close overjab
  // * as contra: take heart instead of spade queen, since re will overjab with club queen
  // * high jabbing, also without a 10 in the trick: check remaining points
  // * overjab, so that the last player cannot pfund the partner
  // * check the position of the card according to the hand
  // * check the position of the card according to the remaining cards
  // * check few trumps (don't play the fox free)
  // * regard announcement of partner (take a high card)
  // * ensure announcement / winning
  //   -> heuristic 'get trick for announcement'
  // * senseless: overjab the partner close

  // card decision
  // a) narrow jabbing
  // b) high jabbing
  // c) medium
  // d) show team
  // e) save points (few trump)

  // @heuristic::condition   not first player
  if (trick.isempty())
    return Card::empty;

  // @heuristic::condition   can overjab the trick
  if (!hi.hand().can_jab(trick))
    return Card::empty;

  Game const& game = hi.game();
  // card to choose
  Card card; 

  DEBUG_BJC_OSTR << hi.player().name() << ": best jabbing_card()\n" << trick;
  if (game.isvirtual())
    DEBUG_BJC_OSTR << "** virtual game **\n";

  // 1) startplayer
  // 2) second player
  // 3) second last player
  // 4) last player

  // 4) / 3) + partner behind
  if (   trick.islastcard()
      || (   (trick.remainingcardno() == 2)
          && hi.guessed_same_team(trick.lastplayer()) )
     ) {
    DEBUG_BJC_OSTR << "4), 3) + partner behind\n";
    // ToDo: check for duty announcement
    // color ace / ten
    if (   !trick.startcard().istrump()
        && hi.hand().contains(trick.startcard().tcolor()))
      card = hi.hand().highest_card(trick.startcard().tcolor());
    if (!card) {
      // trump ace (fox) / trump ten / trump king
      auto const trumpcolor = game.cards().trumpcolor();
      vector<Card> cards = {
        Card(trumpcolor, Card::ace),
        Card(trumpcolor, Card::ten),
        Card(trumpcolor, Card::king)
      };
      if (game.rule()(Rule::Type::with_nines))
        cards.push_back(Card(trumpcolor, Card::nine));
      cards.push_back(Card::diamond_jack);
      for (auto const c : cards) {
        if (   hi.hand().contains(c)
            && !HandCard(hi.hand(), c).is_special()
            && trick.isjabbed(c)) {
          card = c;
          DEBUG_BJC_OSTR << "  " << card << " (manual order -- fox to nine, diamond jack)\n";
          break;
        }
      }
    } // if (trump is a color)
    // lowest winning card
    if (!card) {
      card = hi.hand().next_higher_card(trick.winnercard());
      DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
    }
  } // 4) / 3) + partner behind
  else { // 2), 3)
    DEBUG_BJC_OSTR << "2), 3)\n";
    // 2), 3), color trick, has color
    if (!trick.startcard().istrump()) { //color trick
      DEBUG_BJC_OSTR << "color trick\n";
      Card::Color const color = trick.startcard().color();
      if (hi.hand().contains(color)) {
        DEBUG_BJC_OSTR << "has color\n";
        // highest card
        // ToDo: check whether to spare an ace
        card = hi.hand().highest_card(color);
        DEBUG_BJC_OSTR << "  " << card << " (highest card)\n";
      } else if (hi.color_runs(color) == 0) {
        // 2), 3), color trick, no color
        DEBUG_BJC_OSTR << "has not color\n";

        // first round, no color
        // ToDo: check whether another player has thrown the color / has the color in the hand
        // trump ace (fox) / trump ten / trump king
        {
          if (!(   (hi.cards_information().remaining(color)
                    >= trick.remainingcardno())
                && (   hi.guessed_same_team(trick.lastplayer())
                    || hi.handofplayer(trick.lastplayer()).contains(color))
                && (   (trick.actcardno() == 2)
                    || hi.guessed_same_team(trick.player_of_card(2))
                    || hi.handofplayer(trick.player_of_card(2)).contains(color)))
             ) {
            // if the trick can not get through, take a high card
            // ToDo: check short colors (without nines, dolle)
            card = hi.hand().same_or_higher_card(hi.trump_card_limit());
            if (!card)
              card = hi.hand().same_or_higher_card(hi.lowest_trump_card_limit());
            if (!card)
              card = hi.hand().same_or_higher_card(Card::diamond_jack);
            DEBUG_BJC_OSTR << "  " << card << " (manual order -- high trump)\n";
          }

          if (!card) {
            if (!(   (   !hi.guessed_same_team(trick.lastplayer())
                      && hi.cards_information().of_player(trick.lastplayer()).played(color))
                  || (   (trick.actcardno() == 2)
                      && !hi.guessed_same_team(trick.player_of_card(2))
                      && hi.cards_information().of_player(trick.player_of_card(2)).played(color))
                 )) {
              // only play below diamond jack, if no opposite player has played the color before
              auto const trumpcolor = game.cards().trumpcolor();
              vector<Card> cards = {
                Card(trumpcolor, Card::ace),
                Card(trumpcolor, Card::ten),
                Card(trumpcolor, Card::king)
              };
              if (game.rule()(Rule::Type::with_nines))
                cards.push_back(Card(trumpcolor, Card::nine));
              for (auto c : cards) {
                if (   hi.hand().contains(c)
                    && !HandCard(hi.hand(), c).is_special()
                    && trick.isjabbed(c)) {
                  card = c;
                  DEBUG_BJC_OSTR << "  " << card << " (manual order -- fox to nine, diamond jack)\n";
                  break;
                }
              }
            }
          }

          if (   !card
              && !trick.winnercard().istrump()) {
            card = hi.hand().same_or_higher_card(Card::diamond_jack);
            DEBUG_BJC_OSTR << "  " << card << " (same or higher card(diamond jack))\n";
          }
        } // if (trump is a color)

        if (   !card
            || !trick.isjabbed(card)) {
          card = hi.hand().next_higher_card(trick.winnercard());
          DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
        }

        // show with club queen
        if (   (game.type() == GameType::normal)
            && game.cards().less(Card::charlie, card)
            && (hi.team() == Team::re)
            && (game.teaminfo().get(hi.player()) != Team::re)
            && hi.hand().contains(Card::club_queen)
            && trick.isjabbed(Card::club_queen, hi.hand())) {
          card = Card::club_queen;
          DEBUG_BJC_OSTR << "  " << card << " (show with club queen)\n";
        }

      } else { // 2), 3), color trick (other cases)
        DEBUG_BJC_OSTR << "color trick (other cases)\n";
        // ToDo: jack / queen
        // check how many points are still out
        // check whether another player has thrown the color / has the color in the hand
        if (   (trick.points() >= 10)
            || (   hi.cards_information().remaining_others(color, Card::ace)
                && !Card(color, Card::ace).istrump(game))

            || (   hi.cards_information().remaining_others(color, Card::ten)
                && !Card(color, Card::ten).istrump(game)) ) {
          // make a very high jab
          if (   (trick.points() >= 20)
              || (   (trick.points() >= 10)
                  && (trick.actcardno() == 1) ) ) {
            card = hi.hand().highest_trump();
            DEBUG_BJC_OSTR << "  " << card << " (highest trump)\n";
          }
          // make a high jab
          if (   !card
              && (   (trick.points() >= 14)
                  || checkAllOppositeTeam(trick, hi) )
             ) {
            card = hi.hand().same_or_lower_card(Card::club_queen);
            DEBUG_BJC_OSTR << "  " << card << " (same or lower card(club queen))\n";
          }
          if (!card) {
            card = hi.hand().same_or_higher_card(hi.lowest_trump_card_limit());
            DEBUG_BJC_OSTR << "  " << card << " (same or higher card(lowest trump card limit))\n";
          }
        } else if (   (trick.actcardno() == 1)
                   && hi.guessed_same_team(trick.startplayer()) ) {
          // jab high, if both opposite players are behind
          card = hi.hand().same_or_higher_card(hi.trump_card_limit());
          DEBUG_BJC_OSTR << "  " << card << " (same or higher card(trump card limit))\n";
        } 

        if (!card) {
          vector<Card> cards;
          // ToDo: check short colors (without nines, dolle)
          if (hi.cards_information().remaining(color)
              >= trick.remainingcardno()) {
            auto const trumpcolor = game.cards().trumpcolor();
            cards.push_back(Card(trumpcolor, Card::ace));
            cards.push_back(Card(trumpcolor, Card::ten));
            cards.push_back(Card(trumpcolor, Card::king));
          }
          cards.push_back(Card::diamond_jack);
          for (auto const c : cards) {
            if (   hi.hand().contains(c)
                && !HandCard(hi.hand(), c).is_special()
                && trick.isjabbed(c)) {
              card = c;
              DEBUG_BJC_OSTR << "  " << card << " (manual order -- fox to nine, diamond jack -- highest jack)\n";
              break;
            }
          }
        }
        if (!card) {
          // not many points or no high trump
          card = hi.hand().same_or_higher_card(hi.lowest_trump_card_limit());
          DEBUG_BJC_OSTR << "  " << card << " (same or higher card(lowest trump card limit))\n";
        }
        if (   !card
            || !trick.isjabbed(card) ) {
          card = hi.hand().next_higher_card(trick.winnercard());
          DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
        }

      } // cases in color trick
    } else { // if !(color trick)
      if (trick.actcardno() == game.playerno() - 1) {
        DEBUG_BJC_OSTR << "last player\n";
        // 3), opposite team behind
        if (static_cast<int>(trick.points())
            >= hi.value(Aiconfig::Type::limitqueen)) {
          card = hi.hand().same_or_higher_card(hi.trump_card_limit());
          DEBUG_BJC_OSTR << "  " << card << " (same or higher card(trump card limit))\n";
        }
      } else { // if !3 (opposite team behind)

        if (   hi.guessed_same_team(trick.player_of_card(2))
            && !hi.guessed_same_team(trick.player_of_card(3)) ) {
          // 2), 3: same team, 4: opposite team
          // take a high card (over trump card limit)
          // or the lowest winning card
          if (   (static_cast<int>(trick.points())
                  >= hi.value(Aiconfig::Type::limitdolle))
              || (   (static_cast<int>(trick.points() >= 10))
                  && (hi.hand().higher_cards_no(Card::club_queen)
                      >= 2) ) ) {
            card = hi.hand().highest_trump();
            DEBUG_BJC_OSTR << "  " << card << " (highest trump (many points) )\n";
          }
          if (!card) {
            card = hi.hand().same_or_higher_card(hi.trump_card_limit());
            DEBUG_BJC_OSTR << "  " << card << " (same or higher card(trump card limit) )\n";
            if (   !card
                || !trick.isjabbed(card)
                || game.cards().less(Card::club_queen, card)) {
              card = hi.hand().next_higher_card(Card::charlie);
              DEBUG_BJC_OSTR << "  " << card << " (next higher card(club jack))\n";
            }
            if (!trick.isjabbed(card)) {
              card = hi.hand().next_higher_card(trick.winnercard());
              DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
            }
          } // if (!card)
        } // 2), 3: same team, 4: opposite team
        else if (   !hi.guessed_same_team(trick.player_of_card(2))
                 && hi.guessed_same_team(trick.player_of_card(3)) ) {
          // 2), 3: opposite team, 4: same team
          // take lowest winning card
          card = hi.hand().next_jabbing_card(trick.winnercard());
          while (  (   (card.value() >= 10)
                    || (   (   !game.is_real_solo()
                            || is_color_solo(game.type()) )
                        && HandCard(hi.hand(), card).less(Card::diamond_jack)) )
                 && !HandCard(hi.hand(), card).is_special()
                 && card)
          {
            card = hi.hand().next_higher_card(card);
          }
          if (!card)
            card = hi.hand().next_jabbing_card(trick.winnercard());
          DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
        } // 2), 3: opposite team, 4: same team
        else { // 2, two times opposite team behind
          // take the highest card, if the partner has played ten points
          if (trick.points() >= 10) {
            card = hi.hand().highest_trump();
            // do not take a dolle when it can be jabbed
            if (   game.rule()(Rule::Type::dollen_second_over_first)
                && (card == Card::dolle)
                && hi.handofplayer(trick.player_of_card(2)).contains(Card::dolle)
                && hi.handofplayer(trick.player_of_card(3)).contains(Card::dolle)
               ) {
              card = hi.hand().same_or_lower_card(Card::club_queen);
              if (   (card.value() != Card::queen)
                  && (card.value() != Card::jack))
                card = Card::empty;
              DEBUG_BJC_OSTR << "  " << card << " (same or lower card(club queen))\n";
            } 
          } // if (trick.points() >= 10)
          // if there is an announcement higher then 'no 120', then take the highest queen
          if (game.announcements().last(hi.team()) > Announcement::no120) {
            card = hi.hand().same_or_lower_card(Card::club_queen);
            if (   (card.value() != Card::queen)
                && (card.value() != Card::jack))
              card = Card::empty;
            DEBUG_BJC_OSTR << "  " << card << " (same or lower card(club queen))\n";
          }
          if (!card) {
            // take lowest queen, the highest jack or the lowest winnercard
            card = hi.hand().same_or_higher_card(Card::diamond_queen);
            DEBUG_BJC_OSTR << "  " << card << " (same or higher card(diamond queen))\n";
            if (   !card
                || !trick.isjabbed(card)
                || game.cards().less(Card::club_queen, card)) {
              card = hi.hand().same_or_lower_card(Card::club_jack);
              if (card.value() != Card::jack)
                card = Card::empty;
              DEBUG_BJC_OSTR << "  " << card << " (same or lower card(club jack))\n";
            }
          }
          if (   !card
              || !trick.isjabbed(card)) {
            card = hi.hand().next_higher_card(trick.winnercard());
            DEBUG_BJC_OSTR << "  " << card << " (next higher card)\n";
          }
        } // 2), two times opposite team behind
      } // if !3 (opposite team behind)
    } // if !(color trick)
  } // 2), 3)

  { // special case: dolle
    if (   !card
        && (trick.winnercard() == Card::dolle)
        && game.rule()(Rule::Type::dollen_second_over_first)
        && hi.hand().contains(Card::dolle))
      card = Card::dolle;

    Card const card_bak = card;
    // do not play a dolle / swine if the opposite player can overjab
    if ( (  (card == Card::dolle)
          || HandCard(hi.hand(), card).isswine())
        && (   ( (trick.actcardno() <= 2)
                ? (   !hi.guessed_same_team(trick.lastplayer())
                   && hi.handofplayer(trick.lastplayer()).higher_card_exists(card))
                : false)
            || ( (trick.actcardno() <= 1)
                ? (   !hi.guessed_same_team(trick.player_of_card(2))
                   && hi.handofplayer(trick.player_of_card(2)).higher_card_exists(card))
                : false) )
       ) {
      card = hi.hand().next_lower_card(card);
      if (!trick.isjabbed(card))
        card = card_bak;
    }
  } // special case: dolle

  // ToDo: check for using a special card
  card = optimized_card_to_play(card, trick, hi);
  DEBUG_BJC_OSTR << "  " << card << " (optimized card to play)\n";

  // prefer a swine over a dolle with the rule 'dollen second over first'
  if (   card.isdolle(game)
      && game.rule()(Rule::Type::dollen_second_over_first)
      && (   ( (trick.actcardno() <= 2)
              ? (   !hi.guessed_same_team(trick.lastplayer())
                 && hi.handofplayer(trick.lastplayer()).contains(Card::dolle))
              : false)
          || ( (trick.actcardno() <= 1)
              ? (   !hi.guessed_same_team(trick.player_of_card(2))
                 && hi.handofplayer(trick.player_of_card(2)).contains(Card::dolle))
              : false) )
     ) {
    if (hi.hand().next_higher_card(card) != Card::empty) {
      card = hi.hand().next_higher_card(card);
      DEBUG_BJC_OSTR << "  " << card << " (swine over a dolle)\n";
      return card;
    }
  }

  // with swines try to save a dolle
  if (   hi.hand().count_dolle()
      && trick.isvalid(Card::dolle, hi.hand())
      && game.cards().less(card, game.cards().dolle())
      && (   (   game.swines().swines_announced()
              && !hi.player().has_swines()
              && !hi.guessed_same_team(game.swines().swines_owner())
              && trick.has_played(game.swines().swines_owner()))
          ||   (   game.swines().hyperswines_announced()
                && !hi.player().has_hyperswines()
                && !hi.guessed_same_team(game.swines().hyperswines_owner())
                && trick.has_played(game.swines().hyperswines_owner()))
         ) ) {
    unsigned n = 2;
    if (   game.swines().swines_announced()
        && !hi.player().has_swines()
        && !hi.guessed_same_team(game.swines().swines_owner())) {
      n += hi.cards_information().remaining(hi.game().cards().swine());
    }
    if (   game.swines().hyperswines_announced()
        && !hi.player().has_hyperswines()
        && !hi.guessed_same_team(game.swines().hyperswines_owner())) {
      n += hi.cards_information().remaining(hi.game().cards().hyperswine());
    }
    if (hi.hand().count_fehl() <= 2)
      n += 1;
    else if (trick.islastcard())
      n += 1;
    if (hi.hand().count_trumps() <= (hi.hand().count_dolle() + n) ) {
      card = Card::dolle;
      DEBUG_BJC_OSTR << "  " << card << " (save dolle)\n";
      return card;
    }
  } // if (game.swines_owner() && hi.hand().count_dolle())

  // check whether to show the team (club queen)
  if (   (game.type() == GameType::normal)
      // ToDo: check the own gameplay for team informations
      && (game.teaminfo().get(hi.player()) == Team::unknown)
      && (hi.team() == Team::re)
      && hi.hand().contains(Card::club_queen)
      && trick.isvalid(Card::club_queen, hi.hand())
      && trick.isjabbed(Card::club_queen)) {
    if (   (card == Card::spade_queen)
        || (card == Card::heart_queen)
        || (   (card == Card::diamond_queen)
            && !hi.cards_information().remaining_others(Card::heart_queen)
            && !hi.cards_information().remaining_others(Card::spade_queen)
           )
       ) {
      card = Card::club_queen;
      DEBUG_BJC_OSTR << "  " << card << " (show with club queen)\n";
      return card;
    }
    if (   !trick.startcard().istrump()
        && (hi.color_runs(trick.startcard().tcolor()) > 0)
        && !trick.islastcard()
        && hi.guessed_same_team(trick.lastplayer())) {
      card = Card::club_queen;
      DEBUG_BJC_OSTR << "  " << card << " (show with club queen)\n";
      return card;
    }
  } // if (re, not shown)

  if (!jab_always) {
    // do not take 10 points (fox, dolle, swine) if the player behind can overjab
    if (   (card.points() >= 10)
        && card.istrump(game)
        && !(   !trick.startcard().istrump()
             && hi.cards_information().color_runs(trick.startcard().color()) == 0)
        && (   ( (trick.actcardno() <= 2)
                ? (   !hi.guessed_same_team(trick.lastplayer())
                   && hi.handofplayer(trick.lastplayer()).higher_card_exists(card))
                : false)
            || ( (trick.actcardno() <= 1)
                ? (   !hi.guessed_same_team(trick.player_of_card(2))
                   && hi.handofplayer(trick.player_of_card(2)).higher_card_exists(card))
                : false) ) ) {
      card = Card::empty;
      DEBUG_BJC_OSTR << "  " << card << " (test for overjabbing)\n";
    }

    // do not overjab the partner close
    if (card
        && hi.guessed_same_team(trick.winnerplayer())
        && (   trick.islastcard()
            || (card == game.cards().next_higher_card(trick.winnercard()))
           ) ) {
      card = Card::empty;
      DEBUG_BJC_OSTR << "  " << card << " (no close overjabbing the partner)\n";
    }

  } // if (!jab_always)

  return card;
} // Card Heuristics::best_jabbing_card(Trick trick, HeuristicInterface hi, bool jab_always)

/** finds best card on Hand of hi, which wins this trick,
 ** non trump ace or lowest trump
 ** greater than jack, only in lastcard lower trumps are allowed
 ** the card will be chosen taking the value of the trick in account
 **
 ** @param    trick     trick to check
 ** @param    hi        heuristic interface
 ** @param    pointmodi point modifikator to add to points of t
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::best_winning_card_old(Trick const& t, HeuristicInterface const& hi,
                                  unsigned pointmodi )
{
  // @heuristic::name   best winning card
  // @heuristic::idea   Tries to find the best card to win a trick depending on different values like Aiconfig::Type::limitqueen

  switch (hi.game().type()) {
  case GameType::solo_jack:
    return best_winning_card_single_picture( t, Card::jack, hi);
  case GameType::solo_queen:
    return best_winning_card_single_picture( t, Card::queen, hi);
  case GameType::solo_king:
    return best_winning_card_single_picture( t, Card::king, hi);
  case GameType::solo_queen_jack:
    return best_winning_card_double_picture( t, Card::jack, Card::queen, hi);
  case GameType::solo_king_jack:
    return best_winning_card_double_picture( t, Card::jack, Card::king, hi);
  case GameType::solo_king_queen:
    return best_winning_card_double_picture( t, Card::king, Card::queen, hi);
  case GameType::solo_koehler:
    return best_winning_card_triple_picture( t, hi);
  default:
    break;
  } // switch (hi.game().type())


  if( t.isstartcard() )
    return Card::empty;

  HandCard c(hi.hand());
  auto const ha = hi.hand().validcards(t);
  unsigned i;
  auto const& game = t.game();

  bool allmyteam = checkAllMyTeam(t, hi);

  int const tpoints = t.points() + pointmodi + (t.remainingcardno()-1)*1;


  bool const trickLost = t.winnercard().less(highest_card_behind_of_opposite_team(t, hi));

  // check for enough points in the trick
  if (   t.islastcard()
      && (t.points() <= 6))
    return Card::empty;

  if( !trickLost || allmyteam ) {
    // take lowest trump which takes this trick
    for( i = 0; i < ha.cardsnumber(); i++ ) {
      HandCard const& card = ha.card(i);

      if (!t.isjabbed(card))
        continue;
      if( (   c.is_empty()
           || card.value() < c.value()
           || ( card.value() == c.value()
               && card.less( c ) )
          )
        )
      {
        if (card.is_special()) {
          if ( tpoints >= hi.value( Aiconfig::Type::limitdolle ) ) {
            c = card;
          }
        } else {
          c = card;
        }
      }

    } // for (i < ha.cardsnumber())

    if (!c.is_empty()) {
      // improve the card: take the low trumps in the order ace, ten, king, nine
      auto const trumpcolor = game.cards().trumpcolor();
      if (trumpcolor != Card::nocardcolor) {
        if (c.istrump()) {
          vector<Card> cards = {
          Card(trumpcolor, Card::ace),
          Card(trumpcolor, Card::ten),
          Card(trumpcolor, Card::king)
          };
          if (hi.game().rule()(Rule::Type::with_nines))
            cards.push_back(Card(trumpcolor, Card::nine));
          for (auto const c : cards) {
            if (!HandCard(hi.hand(), c).is_special())
              if (hi.hand().contains(c)
                  && t.isjabbed(c)) {
                return c;
              }
          } // for (c)

        } // if c.istrump())
      } // if (trumpcolor != Card::nocardcolor)

      { // take club queen over heart/spade queen to show myself
        if (   (game.type() == GameType::normal)
            && (game.announcements().last(hi.player())
                == Announcement::noannouncement)
            && (hi.team() == Team::re)
            && hi.hand().contains(Card::club_queen)
            && (   (c == Card::heart_queen)
                || (c == Card::spade_queen) )
           )
          c = Card::club_queen;
      } // take club queen over heart/spade queen

      return c;

    } // if (!c.is_empty())

  } // if( !trickLost || allmyteam )

  // find any card that wins this trick greater or equal to lowest_trump_card_limit
  for (auto const& card : ha) {
    if (!(   hi.lowest_trump_card_limit().less(card)
          || allmyteam) )
      continue;

    if(   (card.value() == Card::queen )
       && (tpoints >= hi.value(Aiconfig::Type::limitqueen))
      ) {
      c = card;
      if (!t.islastcard()) {
        break;
      }
    }
    if( card.value() == Card::queen )
      continue;

    if(  (   card.isdolle()
          || card.possible_swine()
          || card.possible_hyperswine()
         )
       && tpoints >= hi.value(Aiconfig::Type::limitdolle)
      )
    {
      c = card;
      if (!t.islastcard())
      {
        break;
      }
    }

    if(   card.isdolle()
       || card.possible_swine()
       || card.possible_hyperswine() )
      continue;

    c = card;
    break;
  } // for(i)

  // if can't win this trick and I'am not the last player: that's it
  if(   c.is_empty()
     && !t.islastcard() )
    return Card::empty;


  // or this is a color trick of a first color run with my card the first trump
  allmyteam =   allmyteam
    ||(   t.actcardno() > 1
       && hi.color_runs(t.startcard().color()) == 0
       && !t.startcard().istrump()
       && !t.contains_trump()
       && (   t.startcard().color()!= Card::heart
           || !HandCard(hi.hand(), Card::heart_ten ).isdolle()
          )
      );

  // find a better card
  for( i = 0; i < ha.cardsnumber(); i++ ) {
    HandCard const& card = ha.card(i);

    if (!t.isjabbed(card))
      continue;

    // non trump ace is always the best choice if color is not already jabbed
    if(   card.value() == Card::ace
       && !card.istrump()
       && (   !hi.colorjabbed( card.color() )
           || hi.jabbedbyownteam( card.color() )
           || allmyteam
          )
      ) {
      c = card;
      break;
    }

    if( card.istrump() ) {
      // find lowest trump
      if(     t.islastcard()
         && card.less(Card::diamond_jack)
         && card.less( c )
        )
      {
        c = card;
        continue;
      }

      if (!card.less(Card::diamond_jack)) {
        if(   !t.isstartcard()
           && card.less(Card::diamond_queen)
           && !t.startcard().istrump()
           && card.less( c )
           && (   !hi.colorjabbed( t.startcard().color() )
               || tpoints <= hi.value( Aiconfig::Type::limitqueen )
               || ha.count( Card::queen ) < 2
               || t.islastcard()
              )
          ) {
          c = card;
          continue;
        }
        if(    (card.value() == Card::queen)
           && (tpoints >= hi.value( Aiconfig::Type::limitqueen ) )
           && (  (    c.less(Card::diamond_queen)
                  && !t.islastcard()
                  && (    hi.color_runs( t.startcard().color() ) != 0
                      || (   hi.color_runs( t.startcard().color() ) == 0
                          && hi.game().type() == GameType::poverty
                         )
                      || t.startcard().istrump()
                      || tpoints > hi.value( Aiconfig::Type::limitdolle )
                     )
                 )
               || (    card.less( c )
                   &&  !(    c == Card::club_queen
                         && (hi.game().teaminfo().get( hi.player() ) == Team::unknown)
                         && (hi.game().tricks().current_no() > 3)
                        )

                  )
               || ( (card == Card::club_queen)
                   && (hi.game().teaminfo().get( hi.player() ) == Team::unknown)
                   && (hi.game().tricks().current_no() > 3)
                  )
               )
               )
               {

                 if ( (hi.hand().count_club_queens() > 0) // club queen is as good as dolle
                     && !hi.game().second_dolle_over_first()
                     && hi.color_runs( t.startcard().color() ) > 0)
                 {

                   c = Card::club_queen;
                 } else {
                   c = card;
                 }
                 continue;
               }

        if(
           (   card.isdolle()
            || card.possible_swine()
            || card.possible_hyperswine()
           )
           && (   card.less(c)
               || (   tpoints >= hi.value( Aiconfig::Type::limitdolle )
                   && c.less(Card::diamond_queen)
                   && (hi.color_runs( t.startcard().color() ) > 0)
                   && !t.islastcard()
                  )
              )
          )
        {

          if ( (hi.hand().count_club_queens() > 0) // club queen is as good as dolle
              && !hi.game().second_dolle_over_first()
              && card.isdolle() )
          {

            c = Card::club_queen;
          } else {
            c = card;
          }

          continue;
        }
      }//if card>jack
    }// if card istrump

  } // if winneris

  if (c.is_empty())
    return Card::empty;

  Trick tempTrick = t;
  tempTrick.add(c);

  //if opposite team can win trick no further optimization

  if (    oppositeTeamCanWinTrick( tempTrick, hi )
      && !t.islastcard()
      && (   (hi.color_runs(t.startcard().color()) != 0 )
          || (hi.game().cards().count(t.startcard().color())
              - hi.cards_information().played(t.startcard().color())
              <= t.remainingcardno()
             )
         ) ) {
    // make it as hard as possible with selected trump type
    if( c.value() ==Card::jack || c.value() == Card::queen )
      c = highest_card( hi.hand(), c.value(), hi );

    c = optimized_card_to_play(c, t, hi);
    if (   hi.same_team(t.winnerplayer())
        && c.less(valid_cards_behind_of_opposite_team(t, hi).next_higher_card(t.winnercard())))
      return Card::empty;

    return c;
  }



  bool onlyPovertyPlayer = true;
  for( unsigned i = t.actcardno() + 1; i < hi.game().playerno(); i++ )
    onlyPovertyPlayer = (   onlyPovertyPlayer
                         &&  (   hi.game().type() == GameType::poverty
                              && t.player_of_card( i ) == hi.game().players().soloplayer()
                             )); /// @todo poverty partner has no trumps

  // if last card or first_color_run of trick let's check for a better card
  // Ace, ten, king, nine
  if(   t.islastcard()
     || allmyteam
     || onlyPovertyPlayer
     ||(   !t.isstartcard()
        && !t.startcard().istrump()
        && hi.color_runs( t.startcard().color() ) < 1
        && (   t.startcard().tcolor() != Card::heart
            || !hi.game().rule()(Rule::Type::dollen)
            || (hi.hand().count_trump_aces() - hi.hand().count_swines()
                + hi.hand().count(hi.game().cards().trumpcolor(), Card::ten)
                >= 2 )
           )
       )
    ) {

    for( i = 0; i < ha.cardsnumber(); i++ ) {
      HandCard const& card = ha.card(i);


      if (!t.isjabbed(card))
        continue;

      // Maybe it's time to get the heart ten back home
      if(   card.isdolle()
         &&(   (  hi.hand().count_trumps() < 5
                &&  hi.game().tricks().remaining_no() < 7
                &&  hi.hand().count_dolle() == 1
                &&  hi.game().rule()(Rule::Type::extrapoint_dolle_jabs_dolle)
                &&  !hi.cards_information().played(Card::dolle)
               )
            || (   hi.hand().count_trumps()  < 6
                &&  hi.game().tricks().remaining_no() < 8
                && hi.game().swines().swines_announced()
                && hi.teamofplayer(hi.game().swines().swines_owner() ) != hi.team()
               )
            || (    hi.hand().count_trumps() < 6
                &&  hi.game().tricks().remaining_no()  < 8
                && hi.game().swines().hyperswines_announced()
                && hi.teamofplayer(hi.game().swines().hyperswines_owner() ) != hi.team()
               )
           )
        )
      {
        c = card;
        break;
      }

      // ace is always the best card
      if(   (card.value() == Card::ace)
         && !card.possible_swine()
         && (   t.islastcard()
             || allmyteam
             || (   hi.color_runs( t.startcard().color() ) == 0
                 && hi.game().type() != GameType::poverty
                 && !t.startcard().istrump()
                )
            )
        )
      {
        c = card;
        break;
      }

      if(    card.value() == Card::ten
         && !card.isdolle()
         && (   t.islastcard()
             || allmyteam
             || (   hi.color_runs( t.startcard().color() ) == 0
                 && hi.game().type() != GameType::poverty
                 && !t.startcard().istrump()
                )
            )
        )
      {
        c = card;
        continue;
      }

      if(   card.value() == Card::king
         && game.type() != GameType::solo_koehler
         && !card.possible_hyperswine()
        )
      {
        if( c.value()!=Card::ten || c.isdolle() )
        {
          c = card;
        }
        continue;
      }
      if (   card.value() == Card::nine
         && !card.possible_hyperswine() ) {
        if(  (   c.value() != Card::ten
              && c.value() != Card::king
              && c.value() != Card::ace
              && better_points_optimize( c, card, hi )
             )
           || c.isdolle()
           || c.possible_hyperswine()
          ) {
          c = card;
        }
        continue;
      }


      if(   c.istrump()
         && card.istrump()
         && card.less(c)
        ) {
        if(  (   c.value() != Card::ten
              && c.value() != Card::king
              && c.value() != Card::ace
              && better_points_optimize( c, card, hi )
              && !(   (c == Card::club_queen)
                   && (hi.game().teaminfo().get(hi.player()) == Team::unknown)
                   && !(hi.game().type() == GameType::marriage_silent))
             )
           || c.isdolle()
           || c.possible_hyperswine()
          ) {
          c = card;
        }
        continue;
      }
    } // for (i)
  } // if (last card or first color run)

  c = optimized_card_to_play(c, t, hi);
  if (   hi.same_team(t.winnerplayer())
      && c.less(valid_cards_behind_of_opposite_team(t, hi).next_higher_card(t.winnercard())))
    return Card::empty;

  return c;
} // Card Heuristics::best_winning_card_old(Trick t, HeuristicInformation hi, unsigned pointmodi)
