/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"
#include "heuristic.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

#include "ai.h"

#define DEBUG_HEURISTIC \
  if (   (   ::debug.filter("heuristics") \
          || ::debug.filter(name)) \
      && !hi.game().isvirtual()) \
cout << __FILE__ << '#' << __LINE__ << " (" << trick.actcardno() << ')' << hi.player().name() << ": " << name << " – "
#define DEBUG_HEURISTIC_REJECT \
  DEBUG_HEURISTIC << "reject: "
#define DEBUG_HEURISTIC_RETURN \
  DEBUG_HEURISTIC << "return: "

/*
 * Steps to create a new heuristic
 *
 *  1) src/player/aiconfig.h
 *     - enum Heuristic: add heuristic
 *  2) src/player/aiconfig.cpp
 *     - function name(): add heuristic
 *     - Aiconfig::Type::init_heuristic_states(): add heuristic to default ai
 *  3) src/player/ai/heuristic.h
 *     - add declaration of heuristic function
 *  4) src/player/ai/ai.cpp
 *     - Ai::nextcard(trick): add heuristic and function call in switch
 *  5) src/ui/gtkmm/players_db.cpp
 *     - PlayersDB::recreate_db(): add heuristic to switch (mind the group)
 *  6) src/player/ai/heuristic.cpp
 *     - add definition of heuristic (here is the work)
 *     - do not forget the documentation (see below)
 *  7) data/translations/de/aiconfig.heuristics.description
 *     - add german translation for heuristic name and description
 *  8) src/player/ai/aiconfigs.preset
 *     - create presets:
 *       a) save standard ai under 'aiconfig.defensive.saved'
 *       b) call 'make'
 *  9) test
 * 10) create references (remember: set the ai to standard types)
 *
 * Mostly the compiler will find an error.
 * To find the places in the files, just search for 'create_fehl' and 'create_fehl'.
 */

/* Documentation
 *
 * Each heuristic should contain the following documentation for automatic
 * generation of the description.
 * Note that the text shall not be split into multiple lines.
 * The name must be the first entry and the action the last.
 * For a real example look at 'Heuristics::serve_color_trick()'.
 */
// @heuristic::name        the name of the heuristic (as in '::name(Heuristic)')
// @heuristic::idea        the main idea for the heuristic
// @heuristic::comment     a simple comment
// @heuristic::condition   a condition for the heuristic
// @heuristic::condition   a second condition for the heuristic
// @heuristic::action      which card is taken



// missing

// @heuristic::name   choose pfund poverty
// @heuristic::idea   ???

// @heuristic::name   meatless: play highest color
// @heuristic::idea   ???

// @heuristic::name   picture: second best trump
// @heuristic::idea   ???

// @heuristic::name   color: choose ace
// @heuristic::idea   ???

// @heuristic::name   color: best winning
// @heuristic::idea   ???

// @heuristic::name   color: jab for ace
// @heuristic::idea   ???

// @heuristic::name   color: play low high
// @heuristic::idea   ???




// special heuristics

// @heuristic::name      no heuristic
// @heuristic::idea      internal use: no heuristic was used, yet

// @heuristic::name      manual
// @heuristic::idea      The card is played manual (from the human).
//
// @heuristic::name      bug report
// @heuristic::idea      Play the card from the bug report.
// @heuristic::comment   This heuristic is used so that in a bug report replay
// @heuristic::comment   the card suggestion is the card from the bug report.

// @heuristic::name      only valid card
// @heuristic::idea      If there is only one valid card, play it.


/*
   Ideen

   normales Spiel
 * Partner sitzt hinten:
 Farbe anspielen, die er nicht hat und von der noch mindestens 1 draußen ist
 * Startspieler aber kein Farb-As:
 Re: Karo König, Karo Neun, Karo Bube (damit Partner mit Kreuz-Dame rausrückt)
 Kontra: Lange Farbe anspielen (>= 4)
 * Partner sitzt hinter einem, beide Gegner vorne (Trumpfstich):
 - Pfunden (wenn Fuchs / Karo Zehn da)
 - rübergehen, damit Partner Sorgen los wird
 * Wenn Partner direkt nach mir sitzt und noch nicht dran war und kein As gespielt hat:
 Spiele Farbe an, damit er mit potentiellen As übernehmen kann

 Armut (z.B. seed 27350 (3) )
 * Erster Kontra Spieler nach Re:
 Wenn Partner noch übertrumpfen kann (trump limit?): Punkte/drunter bleiben.
 (damit er hinten sitzt)
 wenn Partner nicht übertrumpfen kann (-> cards information): selber übertrumpfen oder kleinen rein.
 * Armutler nach Partner:
 Gegner haben Stich: wenig Punkte geben
 */

/** For a first card in a trick play a color ace
 ** if there are still enough cards of this color in the game
 ** and the color has not been jabbed
 ** choose the ace with the most cards in that color still in the game
 ** If not the first player, take a color ace if the trick can get through.
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play (color ace or none)
 **/
Card
Heuristics::choose_ace(Trick const& trick, HeuristicInterface const& hi)
{
  // @heuristic::name   choose ace
  // @heuristic::idea   For the first card in a trick choose the best color Ace to play. For a following card in a color trick choose the ace, if the trick can get through.

  if (trick.isstartcard()) {
    // I am the start player

    // the number of remaining cards for the color
    map<Card::Color, unsigned> number_of_remaining_cards;

    // the card to play
    Card ace_to_play;
    // the maximal number of remaining cards found so far
    unsigned best_modified_remaining_cards = 0;
    for (auto
         c = hi.game().rule().card_colors().begin();
         c != hi.game().rule().card_colors().end();
         ++c) {
      Card::Color const& color = *c;
      Card const ace(color, Card::ace);

      // test that the color is no trumpcolor
      if ( color == hi.game().cards().trumpcolor())
        continue;
      // test that the player has an ace of the color
      if (!hi.hand().contains(ace))
        continue;

      { // each player behind not of the same team has the color
        unsigned i;
        for (i = trick.actcardno() + 1; i < hi.game().playerno(); ++i)
          if (   (hi.teamofplayer(trick.player_of_card(i)) != hi.team())
              && !hi.handofplayer(trick.player_of_card(i)).contains(color)) {
            break;
          }
        if (i < hi.game().playerno())
          continue;
      } // each player behind has the color

      // the number of remaining cards for the other players
      unsigned const remaining_cards
        = (hi.game().cards().count(color)
           - (hi.cards_information().played(*c) + hi.hand().count(*c)));
      // modified remaining cards
      // Modified by the number of aces of the color.
      // This is used to prefer single aces to double aces.
      unsigned const modified_remaining_cards
        = (remaining_cards
           +  (  (hi.hand().count(ace) + hi.cards_information().played(ace)
                  < hi.game().rule()(Rule::Type::number_of_same_cards))
               ? (hi.game().cards().count(color) - 1) // *Value*
               : 0) );

      // there are enough cards for all players
      if (remaining_cards < hi.game().playerno() - 1) {
        // check whether the opposite team must have the color
        bool opposite_team_must_serve = true;
        for (unsigned i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {
          auto const& player = trick.player_of_card(i);
          if (   (hi.teamofplayer(player) != hi.team())
              && !hi.cards_information(player).must_have(color)) {
            opposite_team_must_serve = false;
            break;
          }
        }
        if (!opposite_team_must_serve)
          continue;
      }

      // if the color has already been jabbed, it has to be from the own team
      if (   hi.colorjabbed(color)
          && (   !hi.jabbedbyownteam(color)
              || hi.jabbedbyotherteam(color)))
        continue;

      // no ace found so far
      if (   !ace_to_play.is_empty()
          // there must not be less remaining cards then the best
          && (   (modified_remaining_cards < best_modified_remaining_cards)
              // if both colors have the same cards out,
              // take the one the player has fewer aces
              || (  (modified_remaining_cards
                     == best_modified_remaining_cards)
                  ? (hi.hand().count(ace)
                     >= hi.hand().count(ace_to_play))
                  : false)
             ))
        continue;

      best_modified_remaining_cards = modified_remaining_cards;
      ace_to_play = ace;
    } // for (c)

    return ace_to_play;
  } // if (trick.isstartcard())

  if( !trick.isstartcard() ) {
    // I am not the start player

    // check for pure color trick
    if (trick.winnercard().istrump())
      return Card::empty;

    // the winnercard is no ace
    if (trick.winnercard().value() == Card::ace)
      return Card::empty;

    // the player has the ace of the color
    Card::Color const color = trick.startcard().color();
    auto const ace = Card(color, Card::ace);
    if (   !hi.hand().contains(ace)
        || ace.istrump(hi.game())
        || !trick.isvalid(ace, hi.hand()))
      return Card::empty;

    // there are enough remaining color cards, so the trick can go through
    if (hi.game().cards().count(color)
        - hi.cards_information().played(color)
        - hi.hand().count(color)
        < trick.remainingcardno() - 1)
      return Card::empty;

    // each player behind of the opposite team has the color
    for (unsigned c = trick.actcardno() + 1; c < hi.game().playerno(); ++c) {
      auto const& p = trick.player_of_card(c);
      if (   (hi.teamofplayer(p) != hi.team())
          && !hi.handofplayer(p).contains(color))
        return Card::empty;
    }

    // in solo games: check whether a ten is sufficient
    auto const ten = Card(color, Card::ten);
    if (   hi.game().is_solo()
        && !hi.is_soloplayer()
        && Heuristics::choose_ten(trick,hi) == ten
        && (   hi.cards_information().remaining_others(ace)
            || hi.cards_information().remaining_others(ten))
        ) {
      auto const& soloplayer = hi.game().players().soloplayer();
      if (hi.cards_information(soloplayer).can_have(color)
          + (trick.has_played(soloplayer) ? 1 : 0)
          > hi.hand().count(ace))
        return ten;
    }

    return ace;
  } // if( !trick.isstartcard() )

  return Card::empty;
} // Card Heuristics::choose_ace(Trick trick, HeuristicInterface hi)

/** For a first card in a trick play a color ten
 ** if there are still enough cards of this color in the game
 ** and the color has not been jabbed and the ten is the highest card
 ** choose the ten with the most cards in that color still in the game
 ** If not the first player, take a color ten if the trick can get through.
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play (color ten or none)
 **/
Card
Heuristics::choose_ten(Trick const& trick, HeuristicInterface const& hi)
{
  // @heuristic::name   choose ten
  // @heuristic::idea   For the first card in a trick choose the best color ten to play. For a following card in a color trick choose the ten, if the trick can get through.

  if (trick.isstartcard()) {

    // the number of remaining cards for the color
    map<Card::Color, unsigned> number_of_remaining_cards;

    // the card to play
    Card ten_to_play;
    // the maximal number of remaining cards found so far
    unsigned best_modified_remaining_cards = 0;
    for (auto const color :  hi.game().rule().card_colors()) {
      Card const ten(color, Card::ten);

      // there are no aces remaining of the color
      if (hi.cards_information().remaining(Card(color, Card::ace)))
        continue;

      // test that the color is no trumpcolor
      if (   (color != hi.game().cards().trumpcolor())
          // test that the player has an ten of the color
          && hi.hand().contains(ten)) {

        { // each player behind has the color
          unsigned i;
          for (i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {
            auto const& p = trick.player_of_card(i);
            if (   !hi.guessed_same_team(p)
                && !hi.handofplayer(p).contains(color)) {
              break;
            }
          }
          if (i < hi.game().playerno())
            continue;
        } // each player behind has the color

        // the number of remaining cards for the other players
        unsigned const remaining_cards
          = (hi.game().cards().count(color)
             - (hi.cards_information().played(color) + hi.hand().count(color)));
        // modified remaining cards
        // Modified by the number of tens of the color.
        // This is used to prefer single tens to double tens.
        unsigned const modified_remaining_cards
          = (remaining_cards
             +  (  (hi.hand().count(ten) + hi.cards_information().played(ten)
                    < hi.game().rule()(Rule::Type::number_of_same_cards))
                 ? (hi.game().cards().count(color) - 1) // *Value*
                 : 0) );

        // there are enough cards for all players
        if (   (remaining_cards >= hi.game().playerno() - 1)
            // if the color has already been jabbed, it has to be from the own team
            && (   !hi.colorjabbed(color)
                || (   hi.jabbedbyownteam(color))
                && !hi.jabbedbyotherteam(color))
            // no ten found so far
            && (   ten_to_play.is_empty()
                // there must not be less remaining cards then the best
                || (   (modified_remaining_cards >= best_modified_remaining_cards)
                    // if both colors have the same cards out,
                    // take the one the player has fewer tens
                    && (  (modified_remaining_cards
                           == best_modified_remaining_cards)
                        ? (hi.hand().count(ten)
                           < hi.hand().count(ten_to_play))
                        : true)
                   ) ) ) {
          best_modified_remaining_cards = modified_remaining_cards;
          ten_to_play = ten;
        } // if (take card)

      } // if (the color is no trumpcolor and the player has an ten of the color)
    } // for (i < ha.cardsnumber())

    return ten_to_play;
  } // if (trick.isstartcard())

  // I am not the start player
  if( !trick.isstartcard() ) {
    // check for pure color trick
    if (trick.winnercard().istrump())
      return Card::empty;

    // the winnercard is no ten or ace
    if (trick.winnercard().value() >= Card::ten)
      return Card::empty;

    // the player has the ten of the color
    Card::Color const color = trick.startcard().color();
    if (   !hi.hand().contains(Card(color, Card::ten))
        || Card(color, Card::ten).istrump(hi.game())
        || !trick.isvalid(Card(color, Card::ten), hi.hand()))
      return Card::empty;


    // there are enough remaining color cards, so the trick can go through
    if (hi.game().cards().count(color)
        - hi.cards_information().played(color)
        - hi.hand().count(color)
        < trick.remainingcardno() - 1)
      return Card::empty;

    // each player behind has the color
    for (unsigned c = trick.actcardno() + 1; c < hi.game().playerno(); ++c) {
      auto const& p = trick.player_of_card(c);
      if (   !hi.guessed_same_team(p)
          && (   !hi.handofplayer(p).contains(color)
              || hi.handofplayer(p).contains(Card(color, Card::ace))))
        return Card::empty;
    }

    return Card(color, Card::ten);
  } // if( !trick.isstartcard() )

  return Card::empty;
} // Card Heuristics::choose_ace(Trick trick, HeuristicInterface hi)

/** for a first card in a trick play a single color ace
 ** if there are still enough cards of this color in the game
 ** and the color has not been jabbed
 ** choose the single ace with the most cards in that color still in the game
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play (color ace or none)
 **/
Card
Heuristics::choose_single_ace(Trick const& trick, HeuristicInterface const& hi)
{
  // @heuristic::name   choose single ace
  // @heuristic::idea   for the first card in a trick choose the best single color Ace to play

  // I am not the start Player
  if( !trick.isstartcard() )
    return Card::empty;

  // the number of remaining cards for the color
  map<Card::Color, unsigned> number_of_remaining_cards;


  // the card to play
  Card ace_to_play;
  // the maximal number of remaining cards found so far
  unsigned best_modified_remaining_cards = 0;
  for (auto const color : hi.game().rule().card_colors()) {
    Card const ace(color, Card::ace);

    // test that the color is no trumpcolor
    if (   (color != hi.game().cards().trumpcolor())
        // test that the player has an ace of the color
        && hi.hand().contains( ace )
        && hi.hand().count( ace.color() ) == 1 ) {

      // the number of remaining cards for the other players
      unsigned const remaining_cards
        = (hi.game().cards().count(color)
           - (hi.cards_information().played(color) + hi.hand().count(color)));
      // modified remaining cards
      // Modified by the number of aces of the color.
      // This is used to prefer single aces to double aces.
      unsigned const modified_remaining_cards
        = (remaining_cards
           +  (  (hi.hand().count(ace) + hi.cards_information().played(ace)
                  < hi.game().rule()(Rule::Type::number_of_same_cards))
               ? (hi.game().cards().count(color) - 1) // *Value*
               : 0) );

      // there are enough cards for all players
      if (   (remaining_cards >= hi.game().playerno() - 1)
          // if the color has already been jabbed, it has to be from the own team
          && (   !hi.colorjabbed(color)
              || (   hi.jabbedbyownteam(color))
              && !hi.jabbedbyotherteam(color))
          // no ace found so far
          && (   ace_to_play.is_empty()
              // there must not be less remaining cards then the best
              || (   (modified_remaining_cards >= best_modified_remaining_cards)
                  // if both colors have the same cards out,
                  // take the one the player has fewer aces
                  && (  (modified_remaining_cards
                         == best_modified_remaining_cards)
                      ? (hi.hand().count(ace)
                         < hi.hand().count(ace_to_play))
                      : true)
                 ) ) ) {
        best_modified_remaining_cards = modified_remaining_cards;
        ace_to_play = ace;
      } // if (take card)

    } // if (the color is no trumpcolor and the player has an ace of the color)
  } // for (i < ha.cardsnumber())

  return ace_to_play;
} // Card Heuristics::choose_single_ace(Trick trick, HeuristicInterface hi)

/** play a good card for the last card in one Trick with color, which
 ** hi doesn't have anymore,
 **
 ** if this trick is won by own team
 **   play a pfund
 ** if there are only color cards in this trick
 **   choose best_winning_card or if there is no lowest_color_card
 ** if there is already a trump try jab this trump with lowest possible
 **   trump, if there are more then a limit of points already in this trick
 **
 ** @param    trick     trick to check
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::choose_for_color_trick(Trick const& trick,
                                   HeuristicInterface const& hi)
{
  constexpr auto name = "choose for color trick";
  // @heuristic::name   choose for color trick
  // @heuristic::idea   choose a card for a color trick either a pfund for the own team or a small card for the opponents
  // @heuristic::condition color trick
  // @heuristic::action play a pfund for the own team or a small card for the opponents or jab the trick

  if (trick.isempty()) {
    DEBUG_HEURISTIC_REJECT << "empty trick\n";
    return Card::empty;
  }

  if (trick.startcard().istrump()) {
    DEBUG_HEURISTIC_REJECT << "trump trick\n";
    return Card::empty;
  }

  if (   is_real_solo(hi.game().type())
      && soloPlayerBehind(trick, hi)) {
    DEBUG_HEURISTIC_REJECT << "soloplayer behind\n";
    return Card::empty;
  }

  auto const& game = hi.game();

  bool allmyteam = (   checkAllMyTeam( trick, hi )
                    && hi.jabbedbyownteam( trick.startcard().color() )
                    && !trick.islastcard());
  // Check whether the soloplayer in a poverty can jab the trick
  if (   (game.type() == GameType::poverty)
      && hi.same_player(game.players().poverty_partner())
      && !hi.handofplayer(game.players().soloplayer()).can_jab(trick))
    allmyteam = false;

  allmyteam = (   allmyteam
               && (   hi.teamofplayer(trick.winnerplayer())==hi.team()
                   || (   trick.winnercard().less( hi.trump_card_limit() )
                       && trick.winnercard().istrump() )
                  ) );

  // in a color trick assume, the partners can get over a nine or a king
  if (   !trick.winnercard().istrump()
      && checkAllMyTeam(trick, hi)
      && !trick.islastcard()
      && (   (trick.winnercard().value() == Card::nine)
          || (trick.winnercard().value() == Card::king)) )
    // ToDo: check whether a player behind can jab the trick
    allmyteam = true;

  if (!allmyteam) {
    // ToDo: if the partner has won the trick, check whether the opposite team can still get the trick
    for (unsigned i = trick.actcardno() + 2; i < game.playerno(); ++i) {
      Player const& player = trick.player_of_card(i);
      if (   (hi.teamofplayer(player) != hi.team())
          && (   (hi.handofplayer(player).contains(trick.startcard().tcolor())
                  && trick.winnercard().less(hi.handofplayer(player).highest_card(trick.startcard().tcolor())) )
              || (!hi.handofplayer(player).contains(trick.startcard().tcolor())
                  && hi.handofplayer(player).hastrump() ) ) ) {
        DEBUG_HEURISTIC_REJECT << "opposite team can jab the trick\n";
        return Card::empty;
      }
    } // for (i)
    // special case: poverty
    if (   (game.type() == GameType::poverty)
        && (hi.team() == Team::contra)
        && !trick.has_played(game.players().poverty_partner()) ) {
      DEBUG_HEURISTIC_REJECT << "poverty, re behind\n";
      return Card::empty;
    }
  } // if (!allmyteam)

  // lets see if this is the last card for this trick
  // or the first run of this color
  // and first card is no trump
  if ( (   trick.islastcard()
        || allmyteam
        || (   !trick.isstartcard()
            && hi.color_runs(trick.startcard().color()) == 0
           )
       )
      && !trick.startcard().istrump()
     )
  {
    HandCard c(hi.hand());
    DEBUG_HEURISTIC << "last card for this trick or first run of this color\n";
    auto const ha = hi.hand().validcards(trick);


    // first check if this trick goes to my team
    bool pred
      = (  (static_cast<int>(trick.points()) >= hi.value(Aiconfig::Type::limitdolle))
         ? (hi.teamofplayer(trick.winnerplayer()) == hi.team())
         : (hi.guessedteamofplayer(trick.winnerplayer()) == hi.team() ));

    if( game.type()==GameType::marriage ) {
      if (::is_selector(trick.startcard().tcolor(),
                        game.marriage().selector()) ) {
        pred = hi.is_soloplayer();
      } else { // if !(trick determines marriage)
        pred = hi.teamofplayer(trick.winnerplayer()) == maybe_to_team( hi.team() );
      } // if !(trick determines marriage)
    }  // if (game.type() == GameType::marriage)

    pred = (   pred
            && (    trick.islastcard()
                || (   trick.winnercard().value() == Card::ace
                    &&  (hi.color_runs(trick.startcard().color()) == 0)
                    &&  (hi.cards_information().remaining_others(trick.startcard().color()) >= trick.remainingcardno() - 1) )
                ||(   trick.winnercard().istrump()
                   && !trick.winnercard().less( hi.trump_card_limit() )
                  )
               ));

    pred = pred || allmyteam;

    if( game.type() == GameType::poverty ) {
      pred &=    (hi.same_player(game.players().poverty_partner()))
        && (  (   !game.rule()(Rule::Type::dollen)
               || trick.startcard().color() != Card::heart)
            || allmyteam);
    }

#if 0
    // -> PfundInFirstColorRun
    if (   pred
        || (   is_solo( game.type() )
            && trick.winnerplayer() != game.players().soloplayer()
            &&  hi.color_runs(trick.startcard().color()) == 0
            &&  trick.winnercard().less( hi.trump_card_limit() )
            &&  !trick.islastcard()
           )
       ) {
      DEBUG_HEURISTIC << "own team gets trick\n";
      auto const c = choose_best_fehl( trick, hi, false );
      if(   !c.is_empty()
         && hi.hand().count( c.color() ) == 1
         && hi.color_runs( c.color() ) == 0 ) {
        DEBUG_HEURISTIC_RETURN << "best fehl: " << c << '\n';
        return c;
      }
      {
        auto c = choose_pfund_card(trick, hi);
        DEBUG_HEURISTIC_RETURN << "pfund card: " << c << '\n';
        return c;
      }
    } // if own Team gets this trick
#endif

    if( !trick.contains_trump() ) {
      DEBUG_HEURISTIC << "no trump\n";
      c = best_winning_card(trick, hi);

      if( !ha.contains(trick.startcard().color() ) ) {
        if(!c.is_empty())
        {
          if (static_cast<int>(trick.points()) > hi.value(Aiconfig::Type::limitdolle)) {
            DEBUG_HEURISTIC_RETURN << "take card for points greater then limitdolle: " << c << '\n';
            return c;
          }
          if(   (static_cast<int>(trick.points()) > hi.value( Aiconfig::Type::limitqueen))
             && (  !Card::club_queen.less( c )
                 || trick.islastcard() )
            ) {
            DEBUG_HEURISTIC_RETURN << "take card for points greater then limitqueen: " << c << '\n';
            return c;
          }

          if( !Card::club_jack.less( c )
             || trick.islastcard() ) {
            DEBUG_HEURISTIC_RETURN << "take small trump: " << c << '\n';
            return c;
          }

        } // if(!c.is_empty())

        DEBUG_HEURISTIC_REJECT << "trump too high for this trick\n";
        return Card::empty;

      } else if (!c.is_empty()) { // if( !ha.contains(trick.startcard().color() ) )

        DEBUG_HEURISTIC_RETURN << "take color card: " << c << '\n';
        return c;
      } else { // if (c.is_empty())
        GameType gt = game.type();

        if( gt != GameType::solo_meatless ) {
          if(   hi.color_runs(trick.startcard().color()) == 0
             && trick.winnercard().value() != Card::ace
             && ha.count( trick.startcard().color(), Card::ace )  > 0 ) {
            auto const c = Card( trick.startcard().color(), Card::ace );
            DEBUG_HEURISTIC_RETURN << "take ace: " << c << '\n';
            return c;
          }

          if( hi.color_runs(trick.startcard().color()) > 0) {
            auto const c = Heuristics::lowest_color_card(trick.startcard().color(),
                                                         hi.hand());
            DEBUG_HEURISTIC_RETURN << "take lowest color card: " << c << '\n';
            return c;
          } else {
            auto const c = lowest_best_trump_card(trick, hi.hand(), hi );
            DEBUG_HEURISTIC_RETURN << "take lowest best trump card: " << c << '\n';
            return c;
          }
        } else { // if( gt != GameType::solo_meatless )
          auto const c = lowest_color_card_meatless( trick, hi.hand() );
          DEBUG_HEURISTIC_RETURN << "meatless: take lowest color card: " << c << '\n';
          return c;
        } // if( gt == GameType::solo_meatless )
      } // if (c.is_empty())


      DEBUG_HEURISTIC_REJECT << "no card found\n";
      return Card::empty;
    }// there was no other trumps in trick
    else // there is a trump in this trick
    {
      DEBUG_HEURISTIC << "trump in trick\n";
      // if there is already a trump
      // and my partner doesn't already win this trick
      // there are more then 10 points already in this trick
      // try jab this trump with lowest possible trump

      if( maybe_to_team( hi.teamofplayer( trick.winnerplayer() ) ) != hi.team() ) {
        c = best_winning_card(trick, hi);

        if (c) {
          if( static_cast<int>(trick.points()) > hi.value(Aiconfig::Type::limitdolle) ) {
            DEBUG_HEURISTIC_RETURN << "take card for points greater then limitdolle: " << c << '\n';
            return c;
          }
          if(   (static_cast<int>(trick.points()) > hi.value(Aiconfig::Type::limitqueen))
             && !Card::club_queen.less( c )
            ) {
            DEBUG_HEURISTIC_RETURN << "take card for points greater then limitqueen: " << c << '\n';
            return c;
          }
          if (static_cast<int>(trick.points())
              < hi.value(Aiconfig::Type::limit_throw_fehl)) {
            c = Heuristics::lowest_card(trick, hi.hand());
            if (   (c.value() == Card::nine)
                || (c.value() == Card::king))
              if (better_points_optimize(c, HandCard(hi.hand(), c), hi)) {
                DEBUG_HEURISTIC_RETURN << "take lowest card: " << c << '\n';
                return c;
              }
          } // if (trick.points() < 5)
          if( !Card::club_jack.less(c) ) {
            DEBUG_HEURISTIC_RETURN << "take low trump: " << c << '\n';
            return c;
          }

        } // if (!c.is_empty())
      } // if( maybe_to_team( hi.teamofplayer( trick.winnerplayer() ) ) != hi.team() )

      DEBUG_HEURISTIC_REJECT << "no reason found\n";
      return Card::empty;
    } // else
  } // check if trick is valid for this heuristic

  DEBUG_HEURISTIC_REJECT << "no reason found\n";
  return Card::empty;
} // Card Heuristics::choose_for_color_trick( Trick trick, HeuristicInterface hi )

/** Play a color again when not jabbed or the own team has jabbed it
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::retry_color(Trick const& trick, HeuristicInterface const& hi)
{
  constexpr auto name = "retry color";
  // @heuristic::name   retry color
  // @heuristic::idea   try a color again, which was not jabbed or jabbed by the own team
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  vector<Card::Color> colors;
  for (auto const color : hi.game().rule().card_colors()) {
    if (hi.color_runs(color) == 0)
      continue;
    if (hi.hand().count(color) == 0)
      continue;

    if (   hi.colorjabbed(color)
        && !hi.jabbedbyownteam(color))
      continue;

    // check that the trick could get through
    if (hi.cards_information().played(color)
        + (hi.game().playerno() - 1)
        + hi.hand().count(color)
        > hi.game().cards().count(color))
      continue;

    colors.push_back(color);
  }

  for (auto const color : colors) {
    if (   (hi.hand().count(Card(color, Card::ten)) > 0)
        && !Card(color, Card::ten).istrump(hi.game())
        && hi.jabbedbyownteam(color)
        && (hi.teamofplayer(trick.lastplayer()) == hi.team()) ) {
      auto const card = Card(color, Card::ten);
      DEBUG_HEURISTIC_RETURN << "own team at backhand – " << card << '\n';
      return card;
    }
  }

  for (auto const color : colors) {
    if (   hi.jabbedbyownteam(color)
        && hi.guessed_same_team(trick.lastplayer())) {
      auto const card = hi.hand().highest_card(color);
      DEBUG_HEURISTIC_RETURN << "color jabbed by own team – " << card << '\n';
      return card;
    }
  }
  for (auto const color : colors) {
    if (   hi.hand().contains(Card(color, Card::nine))
        && !Card(color, Card::nine).istrump(hi.game())) {
      auto const card = Card(color, Card::nine);
      DEBUG_HEURISTIC_RETURN << "take color nine – " << card << '\n';
      return card;
    }
  }
  for (auto const color : colors) {
    if (   hi.hand().contains(Card(color, Card::king))
        && !Card(color, Card::king).istrump(hi.game())) {
      auto const card = Card(color, Card::king);
      DEBUG_HEURISTIC_RETURN << "take color king – " << card << '\n';
      return card;
    }
  }

  DEBUG_HEURISTIC_REJECT << "no suitable color found\n";
  return Card::empty;
} // Card Heuristics::retry_color(Trick trick, HeuristicInterface hi)

/** Play a color the partner does not have
 ** prerequisites:
 **   1. it must be the first card in the trick
 **   2. the partner does not have the color
 **   3. the players behind the partner can have the color
 ** Then take the card with the greatest value in the shortest color
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::play_color_for_partner(Trick const& trick,
                                   HeuristicInterface const& hi)
{
  constexpr auto name = "play color for partner";
  // @heuristic::name   play color for partner
  // @heuristic::idea   play a color the partner does not have

  // @heuristic::condition   startplayer in the trick
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  // one partner known
  if (   (hi.game().type() == GameType::normal)
      && !hi.team_information().guessed_partner()) {
    DEBUG_HEURISTIC_REJECT << "no partner known\n";
    return Card::empty;
  }

  // best card so far to play
  Card best_card;
  // maximal remaining cards of the selected color
  unsigned max_remaining_cards_no = 0;

  for (auto const color : hi.game().rule().card_colors()) {
    // test that the color is no trumpcolor
    if (color == hi.game().cards().trumpcolor())
      continue;

    auto const remaining_cards_no
      = hi.cards_information().remaining_others(color);
    // @heuristic::condition   there are at least as many remaining cards as the best so far
    if (remaining_cards_no < max_remaining_cards_no) {
      DEBUG_HEURISTIC << color << ": less remaining cards than the best color so far\n";
      continue;
    }

    // @heuristic::condition   the partner does not have the color
    // @heuristic::condition   the partner has the highest trump
    // @heuristic::condition   or behind the partner all players have the color
    //                         (players - me - partner)
    Card highest_trump_of_opponents = Card::empty;
    Card highest_trump_of_partners = Card::empty;
    for (unsigned p = hi.game().playerno() - 1; p > trick.actcardno(); --p) {
      auto const& player = trick.player_of_card(p);
      // update the highest card of the opponents
      if (!hi.handofplayer(player).contains(color)) {
        if (hi.guessed_same_team(player)) {
          if (highest_trump_of_partners.less(hi.handofplayer(player).highest_trump()))
            highest_trump_of_partners = hi.handofplayer(player).highest_trump();
        } else { // if !(same team)
          if (highest_trump_of_opponents.less(hi.handofplayer(player).highest_trump()))
            highest_trump_of_opponents = hi.handofplayer(player).highest_trump();
        } // if !(same team)

      } // if (!hi.handofplayer(player).contains(color))
    } // for (p)
    if (highest_trump_of_partners.is_empty()) {
      DEBUG_HEURISTIC << color << ": the partner has the color\n";
      continue;
    }
    if (hi.game().cards().less(highest_trump_of_partners, highest_trump_of_opponents)) {
      highest_trump_of_opponents = Card::empty;
      bool bad_color = false; // whether this is a bad color to play
      for (unsigned p = hi.game().playerno() - 1; p > trick.actcardno(); --p) {
        auto const& player = trick.player_of_card(p);
        // update the highest card of the opponents
        if (   !hi.guessed_same_team(player)
            && !hi.handofplayer(player).contains(color))
          if (highest_trump_of_opponents.less(hi.handofplayer(player).highest_trump()))
            highest_trump_of_opponents = hi.handofplayer(player).highest_trump();
        // the last partner does not have the color and has a high trump
        if (hi.guessed_same_team(player)) {
          if (hi.handofplayer(player).contains(color))
            bad_color = true;
          if (   hi.handofplayer(player).highest_trump().less(hi.trump_card_limit())
              && hi.handofplayer(player).highest_trump().less(highest_trump_of_opponents))
            bad_color = true;
          break;
        }
      } // for (p : playerno)
      if (bad_color) {
        DEBUG_HEURISTIC << color << ": the partner does not have the highest trump\n";
        continue;
      }
    } // if (highest_trump_of_partners.less(highest_trump_of_opponents, hi.game()))

    // search the card with the highest value
    auto const& card = hi.hand().highest_card(color);

    if (   best_card.is_empty()
        || (remaining_cards_no > max_remaining_cards_no)
        || (card.value() > best_card.value()) ) {
      best_card = card;
      max_remaining_cards_no = remaining_cards_no;
    } // if (found better card)
  } // for (color : possible colors)

  // @heuristic::action   play the card with the highest value in the smallest color
  DEBUG_HEURISTIC_RETURN << "best card (highest value in the smallest color) – " << best_card << '\n';
  return best_card;
} // Card Heuristics::play_color_for_partner(Trick trick, HeuristicInterface hi)

/** Play a color the partner hopefully does not have
 ** prerequisites:
 **   1. it must be the first card in the trick
 **   2. the partner is the last player
 **   3. there is only one remaining card of the color
 **   4. another player could have the card
 ** Then take the card with the lowest value in the shortest color
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo     take card value into account
 **/
Card
Heuristics::try_color_for_partner(Trick const& trick,
                                  HeuristicInterface const& hi)
{
  constexpr auto name = "try color for partner";
  // @heuristic::name   try color for partner
  // @heuristic::idea   try a color the partner hopefully does not have

  // @heuristic::condition   startplayer in the trick
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  // @heuristic::condition   last player of the trick is partner
  if (!(::maybe_to_team(hi.teamofplayer(trick.lastplayer())) == hi.team())) {
    DEBUG_HEURISTIC_REJECT << "partner not the last player of the trick\n";
    return Card::empty;
  }

  // best card so far to play
  Card best_card;
  // maximal number of other players who can have the card
  // there must be at least one
  unsigned max_other_players = 1;

  for (auto const color : hi.game().rule().card_colors()) {
    // test that the color is no trumpcolor
    if (color == hi.game().cards().trumpcolor())
      continue;

    if (!hi.hand().contains(color)) {
      DEBUG_HEURISTIC << color << ": the player does not have the color\n";
      continue;
    }

    if (hi.cards_information().remaining_others(color) > 1) {
      DEBUG_HEURISTIC << color << ": more than one cards at the other players\n";
      continue;
    }
    if (hi.cards_information(trick.lastplayer()).must_have(color)) {
      DEBUG_HEURISTIC << color << ": the partner must have the color\n";
      continue;
    }

    // number of other players who can have the last card
    unsigned other_players = 0;

    for (unsigned p = hi.game().playerno() - 2; p > trick.actcardno(); --p) {
      auto const& player = trick.player_of_card(p);
      // ToDo: check that the card value for 'p' is greater than
      //       the card value for the partner
      if (hi.handofplayer(player).contains(color))
        other_players += 1;
    } // for (p : trick.actcardno())

    // search the card with the lowest value
    auto const& card = hi.hand().lowest_card(color);

    if (   (other_players > max_other_players)
        || (   (other_players == max_other_players)
            && (   best_card.is_empty()
                || (card.value() < best_card.value())) ) ) {
      best_card = card;
      max_other_players = other_players;
    } // if (found better card)
  } // for (c : possible colors)

  // @heuristic::action   play the card with the lowest value
  DEBUG_HEURISTIC_RETURN << "best card (lowest value in the smallest color) – " << best_card << '\n';
  return best_card;
} // Card Heuristics::try_color_for_partner(Trick trick, HeuristicInterface hi)

/** Start a new color hoping the partner has an ace
 ** prerequisites:
 **   1. it must be the first card in the trick
 **   2. first run of the color
 **   3. the partner is known and can have the ace
 ** Then take the card with the greatest value in the shortest color
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::play_color_for_partner_ace(Trick const& trick,
                                       HeuristicInterface const& hi)
{
  constexpr auto name = "play color for partner ace";
  // @heuristic::name   play color for partner ace
  // @heuristic::idea   play a color hoping the partner has the ace

  // @heuristic::condition   startplayer in the trick
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  // best card so far to play
  Card best_card;
  // maximal remaining cards of the selected color
  // @heuristic::condition   or there are enough remaining cards for all players
  //                         (players - me)
  unsigned max_remaining_cards_no = hi.game().playerno() - 1;

  for (auto const color : hi.game().cards().colors()) {
    auto const ace = Card(color, Card::ace);
    // @heuristic::condition   first run of the color
    if (hi.cards_information().played(color)
        > hi.cards_information(hi.player()).played(color))
      continue;

    // the player has the color
    if (hi.hand().contains(color) == 0)
      continue;

    // @heuristic::condition   the next in line who can have the ace is of the same team
    Player const* partner = nullptr;
    for (unsigned p = 1; p < hi.game().playerno(); ++p) {
      auto const& player = trick.player_of_card(p);
      auto const& hand = hi.handofplayer(player);
      if (hi.guessed_same_team(player)) {
        if (hand.contains(ace)) {
          partner = &player;
          break;
        }
      } else { // if !(same team)
        if (hand.contains(ace)) {
          break;
        }
        if (!hand.contains(color)) {
          break;
        }
      } // if !(same team)
    } // for (c < hi.game().playerno())

    // a partner is found
    if (!partner)
      continue;

    // @heuristic::condition   the partner has not started a trick with a non-color-ace
    for (unsigned t = 0; t < hi.game().tricks().current_no(); ++t) {
      auto const& trick = hi.game().tricks().trick(t);
      if (   (trick.startplayer() == *partner)
          && !trick.startcard().istrump()
          && (trick.startcard().value() == Card::ace) ) {
        partner = nullptr;
        break;
      }
    } // for (t < hi.game().tricks().current_no())
    if (!partner)
      continue;

    auto const remaining_cards_no
      = hi.cards_information().remaining_others(color);
    // there are at least as many remaining cards as the best so far
    if (remaining_cards_no < max_remaining_cards_no)
      continue;


    // search the card with the highest value
    auto const& card = hi.hand().highest_card(color);

    if (   (remaining_cards_no > max_remaining_cards_no)
        || (card.value() > best_card.value()) ) {
      best_card = card;
      max_remaining_cards_no = remaining_cards_no;
    } // if (found better card)
  } // for (color : possible colors)

  // @heuristic::action   play the card with the highest value in the smallest color
  DEBUG_HEURISTIC_RETURN << "best card – " << best_card << '\n';
  return best_card;
} // Card Heuristics::play_color_for_partner_ace(Trick trick, HeuristicInterface hi)

/** When a player cannot get a color trick take the card with the smallest value
 ** prerequisites:
 **   1. it must not be the first card in the trick
 **   2. the trick must be a color trick;
 **   3. one must have to serve the color
 **   4. the winnercard must be an ace or a trump or one does not have the ace;
 **   5. the winnerplayer must not be of the own team;
 ** The main case is that one is the last to play and has only got to
 ** take ones lowest card.
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::serve_color_trick(Trick const& trick, HeuristicInterface const& hi)
{
  constexpr auto name = "serve color trick";
  // @heuristic::name   serve color trick
  // @heuristic::idea   When a player cannot get a color trick take the card with the smallest value. The main case is that one is the last to play and has only got to take ones lowest card.

  // ToDo: Check, whether the trick can go to the own team

  // @heuristic::condition   it must not be the first card in the trick
  if (trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "empty trick\n";
    return Card::empty;
  }

  // @heuristic::condition   the trick must be a color trick
  if (trick.startcard().istrump()) {
    DEBUG_HEURISTIC_REJECT << "trump trick\n";
    return Card::empty;
  }

  Card::Color const color = trick.startcard().color();

  // @heuristic::condition   one cannot get the trick
  if (   hi.hand().contains(color)
      && hi.hand().contains(Card(color, Card::ace))
      && trick.winnercard().less(Card(color, Card::ace))
      && !HandCard(hi.hand(), color, Card::ace).is_special()) {
    DEBUG_HEURISTIC_REJECT << "can jab trick with ace\n";
    return Card::empty;
  }
  if (   hi.hand().contains(color)
      && (hi.cards_information().remaining(Card(color, Card::ace)) == 0)
      && hi.hand().contains(Card(color, Card::ten))
      && trick.winnercard().less(Card(color, Card::ten))
      && !HandCard(hi.hand(), color, Card::ten).is_special()) {
    DEBUG_HEURISTIC_REJECT << "can jab trick with ten\n";
    return Card::empty;
  }
  if (   !hi.hand().contains(color)
      && hi.hand().higher_card_exists(trick.winnercard())) {
    DEBUG_HEURISTIC_REJECT << "can jab trick with trump\n";
    return Card::empty;
  }

  // as last player: the trick should not go to the own team
  if (   (hi.guessedteamofplayer(trick.winnerplayer()) == hi.team())
      && !oppositeTeamCanWinTrick(trick, hi)) {
    DEBUG_HEURISTIC_REJECT << "own team has jabbed already\n";
    return Card::empty;
  }

  auto const color_cards = hi.hand().validcards(trick);

  // @heuristic::condition   the winnercard must be an ace or a trump or one does not have the ace in the non-trump case or there are not enough remaining cards to get through
  if (!(   (trick.winnercard().value() == Card::ace)
        || trick.winnercard().istrump() ) ) {
    if (   (color_cards.count(Card(color, Card::ace)) > 0)
        && (hi.cards_information().remaining_others(color)
            >= trick.remainingcardno() - 1) ) {
      DEBUG_HEURISTIC_REJECT << "can jab with ace\n";
      return Card::empty;
    }
    if ( !trick.winnercard().less(highest_card_behind_of_opposite_team(trick, hi))
        && (   color_cards.higher_card_exists(trick.winnercard())
            || trick.winnercard().less(highest_card_behind_of_same_team(trick,
                                                                        hi)) ) ) {
      DEBUG_HEURISTIC_REJECT << "own team can jab\n";
      return Card::empty;
    }
  } // if (trick.winnercard.value() != Card::ace) && (...)

  // @heuristic::condition   the winnerplayer must not be of the own team, or the trick does not get through
  if (   (hi.teamofplayer(trick.winnerplayer() ) == hi.team())
      && (   (hi.color_runs(color) != 0)
          || trick.winnercard().istrump()
          || (trick.winnercard() == Card(color, Card::ace))
         )
      && (hi.cards_information().remaining_others(color)
          >= trick.remainingcardno() - 1)
     ) {
    DEBUG_HEURISTIC_REJECT << "winnerplayer of own team and trick gets through\n";
    return Card::empty;
  }

  { // @heuristic::condition   no partner can jab

    Player const* p;

    for (p = &hi.game().players().following(trick.actplayer());
         p != &trick.startplayer();
         p = &hi.game().players().following(*p)) {
      auto const& hand = hi.handofplayer(*p);
      if (   (hi.guessedteamofplayer(*p) == hi.team())
          && hand.higher_card_exists(trick.winnercard()) ) {
        switch (trick.winnercard().value()) {
        case Card::ace:
          if (!Card(color, Card::ace).istrump(hi.game())
              && hand.contains(Card(color, Card::ace)))
            break;
        case Card::ten:
          if (!Card(color, Card::ten).istrump(hi.game())
              && hand.contains(Card(color, Card::ten)))
            break;
        case Card::king:
          if (!Card(color, Card::king).istrump(hi.game())
              && hand.contains(Card(color, Card::king)))
            break;
        case Card::queen:
          if (!Card(color, Card::queen).istrump(hi.game())
              && hand.contains(Card(color, Card::queen)))
            break;
        case Card::jack:
          if (!Card(color, Card::jack).istrump(hi.game())
              && hand.contains(Card(color, Card::jack)))
            break;
        case Card::nine:
          if (!Card(color, Card::nine).istrump(hi.game())
              && hand.contains(Card(color, Card::nine)))
            break;

        case Card::nocardvalue:
        case Card::unknowncardvalue:
          DEBUG_HEURISTIC_REJECT << "winnercard unknown\n";
          return Card::empty;
        } // switch (trick.winnercard().value())
      } // if (own team and player has higher card)
    } // for (p)
  } // no partner can jab

  // @heuristic::condition   it must not be an undetermined marriage
  if (    hi.game().marriage().is_undetermined()
      && !(   is_selector(color, hi.game().marriage().selector())
           && !hi.is_soloplayer())
     ) {
    DEBUG_HEURISTIC_REJECT << "undetermined marriage\n";
    return Card::empty;
  }

  // @heuristic::action   take last trump card
  if (   hi.hand().hastrump()
      && !hi.hand().contains(trick.startcard().color())
      && (hi.hand().count_trumps() <= 2)
      && (hi.hand().count_trumps() < hi.hand().cardsnumber() / 2) ) {
    auto const trumps = hi.hand().trumps();
    if (   (trumps[0].value() <= 3)
        && trumps[0].less(hi.lowest_trump_card_limit())) {
      if (trumps.size() == 1) {
        DEBUG_HEURISTIC_RETURN << "take trump: " << trumps[0] << '\n';
        return trumps[0];
      }
      if (   (trumps[1].value() <= 3)
          && trumps[1].less(hi.lowest_trump_card_limit())) {
        auto const card = trumps.lowest_trump();
        DEBUG_HEURISTIC_RETURN << "take lowest trump: " << card << '\n';
        return card;
      }
    } // if (first trump small)
  } // if (hi.hand().count_trumps() <= 2)

  // @heuristic::action   take lowest color card
  auto const card = Heuristics::lowest_card(trick, hi.hand());
  if (   hi.lowest_trump_card_limit().less(HandCard(hi.hand(), card))
      && hi.hand().lowest_trump().less(card)) {
    DEBUG_HEURISTIC_REJECT << "lowest trump is too high\n";
    return Card::empty;
  }
  // do not play a color ace blank in a solo
  if (   hi.game().is_solo()
      && (hi.hand().count(card.color()) == 2)
      && hi.hand().contains(card.color(), Card::ace)
      && hi.cards_information().remaining_others(card.color(), Card::ace)
     ) {
    DEBUG_HEURISTIC_REJECT << "no blank color ace in a solo\n";
    return Card::empty;
  }
  DEBUG_HEURISTIC_RETURN << "take lowest color card: " << card << '\n';
  return card;
} // Card Heuristics::serve_color_trick(Trick trick, HeuristicInterface hi)


/** for the last card in a trick, which wins your own team
 ** or where your teammate has already a played a high enough trump
 ** play a pfund
 **
 ** @param    t         current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::choose_pfund(Trick const& trick, HeuristicInterface const& hi)
{
  constexpr auto name = "choose pfund";
  // @heuristic::name   choose pfund
  // @heuristic::idea   choose for trump trick won by the own team a pfund

  if (trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "startcard\n";
    return Card::empty;
  }

  auto const pfund_card = HandCard(hi.hand(), choose_pfund_card(trick, hi));
  if (pfund_card == Card::empty) {
    DEBUG_HEURISTIC_REJECT << "no pfund card\n";
    return Card::empty;
  }
  if (   (pfund_card.value() <= 2)
      && !hi.guessed_same_team(trick.winnerplayer())
      && hi.hand().can_jab(trick)) {
    DEBUG_HEURISTIC_REJECT << "small pfund card and can jab the trick\n";
    return Card::empty;
  }

  // undetermined marriage => no pfund
  if (   hi.game().marriage().is_undetermined()
      && (  (  !is_selector(trick.startcard().color(), hi.game().marriage().selector())
             && hi.is_soloplayer())
          || !hi.is_soloplayer()
         )
     ) {
    DEBUG_HEURISTIC_REJECT << "undetermined marriage\n";
    return Card::empty;
  }

  // picture solo, trump trick, has to serve => no pfund
  if (   is_picture_solo(hi.game().type())
      && trick.startcard().istrump()
      && hi.hand().hastrump()) {
    DEBUG_HEURISTIC_REJECT << "picture solo, trump trick, have to serve\n";
    return Card::empty;
  }

  // last player
  if  (trick.islastcard()) {
    DEBUG_HEURISTIC << "last player\n";
    if (hi.same_team(trick.winnerplayer())) {
      DEBUG_HEURISTIC_RETURN << "  trick goes to my team: pfund " << pfund_card << '\n';
      return pfund_card;
    }
    if (!hi.guessed_same_team(trick.winnerplayer())) {
      DEBUG_HEURISTIC_REJECT << "  trick goes to opposite team\n";
      return Card::empty;
    }
    // guessed same team but not sure
    if (hi.value(Aiconfig::Type::trusting)
        || hi.team_information().team_value(trick.winnerplayer()) >= 50
        || hi.team_information().team_value(trick.winnerplayer()) <= -50
       ) {
      DEBUG_HEURISTIC_RETURN << "  guessed own team and trusting: pfund " << pfund_card << '\n';
      return pfund_card;
    }

    // has to get two foxes home
    if (   pfund_card.isfox()
        && !hi.game().rule()(Rule::Type::swines)
        && (hi.hand().count(pfund_card) > 1)) {
      DEBUG_HEURISTIC_RETURN << "  have to get two foxes home: pfund " << pfund_card << '\n';
      return pfund_card;
    }
    // have many pfund cards
    if (   (hi.hand().count_rich_cards() >= 3)
        || (hi.hand().count_rich_cards() * 2 >= hi.hand().cardsnumber())) {
      DEBUG_HEURISTIC_RETURN << "  have many pfund cards: pfund " << pfund_card << '\n';
      return pfund_card;
    }

    DEBUG_HEURISTIC_REJECT << "  no reason found to pfund\n";
    return Card::empty;
  } // if  (trick.islastcard())

  // special case: first color run
  auto color = trick.startcard().tcolor();
  if (   (color != Card::trump)
      && (hi.color_runs(color) == 0)
      && (hi.teamofplayer(trick.winnerplayer()) == hi.team())
      && (trick.winnercard().istrump()
          || (trick.winnercard() == Card(color, Card::ace)))
      && (hi.cards_information().remaining_others(color)
          >= trick.remainingcardno() - 1)
     ) {
    DEBUG_HEURISTIC_RETURN << "first color run, winnerplayer is my team and has ace/trump: pfund " << pfund_card << '\n';
    return pfund_card;
  }

  // special case: second color run
  if (   (color != Card::trump)
      && (hi.color_runs(color) == 1)
      && (hi.teamofplayer(trick.winnerplayer()) == hi.team())
      && (trick.winnercard().istrump()
          || (trick.winnercard() == Card(color, Card::ace)))
      && (hi.cards_information().remaining_others(color)
          >= trick.remainingcardno() - 1 + 1)
     ) {
    for (Player const*
         player = &(trick.game().players().following(trick.actplayer()));
         player != &trick.startplayer();
         player = &(trick.game().players().following(*player))) {
      if (   (hi.teamofplayer(*player) != hi.team())
          && (hi.handofplayer(*player).count(color) < 2)) {
        DEBUG_HEURISTIC_REJECT << "second color run, opponent have not enough color cards\n";
        return Card::empty;
      }
    } // for (player)
    DEBUG_HEURISTIC_RETURN << "second color run, opponent can have enough color cards: pfund " << pfund_card << '\n';
    return pfund_card;
  }

  // no opponent behind me can jab the trick
  if(   !trick.islastcard()
     && !checkAllMyTeam(trick, hi)
     && oppositeTeamCanWinTrick(trick, hi) ) {
    DEBUG_HEURISTIC_REJECT << "opponent behind me can jab the trick\n";
    return Card::empty;
  }

  // if winner is own team
  if (   hi.teamofplayer(trick.winnerplayer()) == hi.team()
      &&  !oppositeTeamCanWinTrick(trick, hi) ) {
    DEBUG_HEURISTIC << "winner is my team\n";
    auto const cards = hi.hand().validcards(trick);
    if (cards.lowest_value() == cards.highest_value()) {
      DEBUG_HEURISTIC_REJECT << "  no pfund to choose from\n";
      return Card::empty;
    }
    // small pfund and the last player in the next trick is of the opposite team
    if (   (pfund_card.value() <= 4)
        && cards.higher_card_exists(pfund_card)
        && trick.winnercard().less(hi.trump_card_limit())
        && !hi.same_team(hi.game().players().previous(trick.winnerplayer()))
       ) {
      if (!hi.same_team(hi.game().players().following(trick.actplayer()))) {
        DEBUG_HEURISTIC_REJECT << "  small pfund and the last player in the next trick is of the opposite team\n";
        return Card::empty;
      }
      if (best_winning_card(trick, hi).value() == pfund_card.value()) {
        DEBUG_HEURISTIC_REJECT << "  best winning card has same points as pfund card\n";
        return Card::empty;
      }
    }
    DEBUG_HEURISTIC_RETURN << "  small pfund: " << pfund_card << '\n';
    return pfund_card;
  }

  if (hi.guessed_same_team(trick.winnerplayer())) {
    DEBUG_HEURISTIC << "guessed winnerplayer is my team\n";
    if (!hi.cards_information().higher_card_exists(trick.winnercard())) {
      DEBUG_HEURISTIC_RETURN << "  no higher card then the winnercard exists: " << pfund_card << '\n';
      return pfund_card;
    }

    if (  !soloPlayerBehind(trick, hi)
        && (    trick.islastcard()
            || !trick.winnercard().less( hi.trump_card_limit() )
           )
       ) {
      DEBUG_HEURISTIC_RETURN << "  high winnecard or last card: " << pfund_card << '\n';
      return pfund_card;
    }
  } // if (   hi.guessed_same_team(trick.winnerplayer())

  if (   hi.game().is_solo()
      && trick.winnerplayer() == hi.game().players().soloplayer()
      && trick.winnercard().less( hi.trump_card_limit() )
      && !trick.islastcard()
      && (pfund_card.points() > 0)
      && trick.startcard().istrump()
     ) {
    DEBUG_HEURISTIC_RETURN << "solo, trump trick, soloplayer has winnercard but low, not last player: " << pfund_card << '\n';
    return pfund_card;
  }

  auto const* partner = hi.team_information().guessed_partner();
  if (   checkAllMyTeam(trick, hi)
      && ownTeamCanWinTrick(trick, hi)
      && trick.winnercard().less(hi.trump_card_limit())
      && (trick.winnercard() != Card(color, Card::ace))
      && !(   !trick.startcard().istrump()
           && trick.winnercard().istrump()
           && (partner
               ? hi.cards_information(*partner).can_have(trick.startcard().tcolor())
               : true))
      && !trick.islastcard()
     ) {
    if (partner) {
      auto hand = hi.handofplayer(*partner);
      if (   (pfund_card.points() <= 4)
          && (hi.handofplayer(*partner).highest_value() >= 10)) {
        DEBUG_HEURISTIC_REJECT << "partner behind can pfund better\n";
        return Card::empty;
      }
    }
    DEBUG_HEURISTIC_RETURN << "own team can jab trick behind: pfund " << pfund_card << '\n';
    return pfund_card;
  }

  DEBUG_HEURISTIC_REJECT << "no reason to pfund found\n";
  return Card::empty;
} // Card Heuristics::choose_pfund( Trick trick, HeuristicInterface hi)


/** if the partner has the highest card in the game, play a pfund (>= 10 points)
 **
 ** @param    trick   current trick
 ** @param    hi      heuristics interface
 **
 ** @return   card to play
 **/
Card
Heuristics::choose_pfund_before_partner(Trick const& trick,
                                        HeuristicInterface const& hi)
{
  // @heuristic::name   choose pfund before partner
  // @heuristic::idea   if the partner has the highest card in the game play a pfund

  // @heuristic::condition   the partner has the highest trump still in the game

  // whether the partner has the highest card
  bool partner_has_highest_card = false;

  // the highest trump the other players can have
  Card const highest_trump
    = hi.cards_information().highest_remaining_trump_of_others();
  // the trick can be jabbed with the highest trump
  if (   !trick.isempty()
      && !trick.isjabbed(highest_trump))
    return Card::empty;

  // the partner
  Player const* partner = hi.team_information().guessed_partner();

  // special case: first color run in a solo game, soloplayer behind
  if (   hi.game().is_solo()
      && (hi.team() == Team::contra)
      && !trick.isempty()
      && !trick.startcard().istrump()
      && !trick.has_played(hi.game().players().soloplayer())
      && (   (trick.winnercard().value() == Card::ace)
          || hi.same_team(trick.player_of_card(trick.actcardno() + 1)) )
      && (hi.color_runs(trick.startcard().tcolor()) == 0)
      && hi.handofplayer(hi.game().players().soloplayer()).contains(trick.startcard().tcolor())
     ) {
    return Heuristics::choose_pfund_card(trick, hi);
  }

  // special case: first trump run and card above 
  // @todo   Schweine/Hyperschweine beim Partner prüfen
  if (   is_color_solo(hi.game().type())
      && hi.team() == Team::contra
      && !trick.isempty()
      && trick.startcard().istrump()
      && hi.cards_information().trump_runs() < 2
      && hi.trump_card_limit().less(trick.winnercard())
      && !hi.game().announcements().last(hi.team()))
    return {};

  // color trick and the partner has to serve the color
  if (   !trick.isempty()
      && !trick.startcard().istrump()
      && hi.cards_information().remaining_others(trick.startcard().tcolor())
      && (   !partner
          || hi.handofplayer(*partner).contains(trick.startcard().tcolor())))
    return Card::empty;

  if (is_solo(hi.game().type())) {
    if (hi.team() == Team::contra) {
      if (   trick.has_played(hi.game().players().soloplayer())
          && trick.winnercard().less(highest_card_behind_of_same_team(trick, hi))
         ) {
        partner_has_highest_card = true;
      }
    } // if (hi.team() == Team::contra)
  } else if (partner) {
    // no partner behind
    if (trick.cardno_of_player(*partner) < trick.actcardno() )
      return Card::empty;

    // normal case
    if ( // the partner has the highest card
        hi.cards_information(*partner).must_have(highest_trump)
        // no other player then the partner has the highest card
        && (hi.cards_information(*partner).must_have(highest_trump)
            + hi.cards_information().played(highest_trump)
            + hi.hand().count(highest_trump)
            == hi.game().rule()(Rule::Type::number_of_same_cards) )
       )
      partner_has_highest_card = true;

    // special case: only the partner behind
    if (trick.actcardno() == hi.game().playerno() - 2) {
      // check, wether he has a card higher then the played one
      if (   (hi.cards_information(*partner).must_have_cards().higher_card_exists(trick.winnercard()))
          || (   trick.winnercard().less(hi.trump_card_limit())
              && hi.handofplayer(*partner).higher_card_exists(trick.winnercard())))
        partner_has_highest_card = true;
      // the player has more then one pfund
      unsigned const pfund_no
        = ((hi.hand().has_swines() ? 0 : hi.hand().count_trump_aces())
           + hi.hand().count(hi.game().cards().trumpcolor(), Card::ten));

      if (   !hi.game().is_solo()
          && (pfund_no >= 2)
          && hi.handofplayer(*partner).higher_card_exists(trick.winnercard())) {
        if (   is_solo(trick.game().type())
            || trick.winnercard().less(hi.value(Aiconfig::Type::limithigh))
            || (hi.hand().cardsnumber() < pfund_no * 3)
            || (   (hi.game().swines().swines_owner() == *partner)
                && hi.handofplayer(*partner).contains(hi.game().cards().swine()))
            || (   (hi.game().swines().hyperswines_owner() == *partner)
                && hi.handofplayer(*partner).contains(hi.game().cards().hyperswine()))
            || trick.points() >= 10)
          partner_has_highest_card = true;
      }

    } // if (only partner behind)
  } else { // if !(partner)
    // special case: partner re but I do not know who he is
    if (   (hi.game().type() == GameType::normal)
        // re team
        && (hi.team() == Team::re)
        // startcard or the player before seems to be contra
        && (   trick.isstartcard()
            || (   (trick.actcardno() == 1)
                && (hi.guessedteamofplayer(trick.startplayer()) == Team::contra))
           )
       ) {
      // the club queen is the highest card
      if (highest_trump == Card::club_queen)
        partner_has_highest_card = true;

    }
  } // if !(partner)

  // special case: partner unknown, but club queen not the highest card
  if (   !partner_has_highest_card
      && (hi.game().type() == GameType::normal)
      && (hi.team() == Team::re)
      && (hi.cards_information().played(Card::club_queen)
          == hi.cards_information(hi.player()).played(Card::club_queen) )
      && hi.hand().contains(hi.game().cards().trumpcolor(), Card::king)
      && trick.isvalid(Card(hi.game().cards().trumpcolor(), Card::king), hi.hand())
      && (   trick.isstartcard()
          || (   (trick.actcardno() == 1)
              && (hi.guessedteamofplayer(trick.startplayer()) == Team::contra)) )
      && !HandCard(hi.hand(),
                   hi.game().cards().trumpcolor(), Card::king).is_special()
      && !(   hi.hand().contains(Card(hi.game().cards().trumpcolor(), Card::ten))
           && !HandCard(hi.hand(),
                        hi.game().cards().trumpcolor(), Card::ten).is_special())
      && !(   hi.hand().contains(Card(hi.game().cards().trumpcolor(), Card::ace))
           && !HandCard(hi.hand(),
                        hi.game().cards().trumpcolor(), Card::ace).is_special())
     ) {
    return Card(hi.game().cards().trumpcolor(), Card::king);
  }

  if (!partner_has_highest_card)
    return Card::empty;

  // @heuristic::action   play a pfund of at least 10 points
  if (trick.isstartcard()) {
    // play a fox
    if (   (hi.hand().count_trump_aces() > 0)
        && !hi.hand().has_swines())
      return Card(hi.game().cards().trumpcolor(), Card::ace);

    // play a diamond ten
    Card const card(hi.game().cards().trumpcolor(), Card::ten);
    if (   hi.hand().contains(card)
        && !HandCard(hi.hand(), card).is_special())
      return card;

    // ToDo: play a color card of a color the partner does not have
    // ToDo: play a many points for the partner

  } else { // if !(trick.isstartcard())
    // the partner can play the card
    if (   trick.isempty()
        || trick.startcard().istrump()
        || (   partner
            && (trick.isvalid(hi.handofplayer(*partner).highest_trump(),
                              hi.handofplayer(*partner))))
        || (   hi.game().is_solo()
            && (hi.team() == Team::contra)
            && trick.has_played(hi.game().players().soloplayer())
            && !trick.startcard().istrump()
            && (   hi.handofplayer(trick.lastplayer()).can_jab(trick)
                || (   (trick.actcardno() == hi.game().playerno() - 2)
                    && hi.handofplayer(trick.secondlastplayer()).can_jab(trick))
               )
           )
       ) {
      // search a pfund card
      Card const card = Heuristics::choose_pfund_card(trick, hi);
      if (   !is_solo(trick.game().type())
          && !trick.winnercard().less(hi.value(Aiconfig::Type::limithigh))
          && HandCard(hi.hand(), card).isfox()
          && (hi.hand().count(card) == 1)
          && hi.hand().contains(Card::diamond_ten))
        return Card::diamond_ten;
      if (card.value() >= Card::ten)
        return card;
    }
  } // if !(trick.isstartcard())

  return Card::empty;
} // Card Heuristics::choose_pfund_before_partner(Trick trick, HeuristicInterface hi)


/** return lowest allowed Jack or Queen if they are trump
 ** and there is an ace to play, which is not already jabbed,
 ** else return empty card
 **
 ** @param    t    current trick
 ** @param    hi   heuristic interface
 **
 ** @result    -> description
 **/
Card
Heuristics::jab_for_ace( Trick const& t, HeuristicInterface const& hi )
{
  // @heuristic::name   jab for ace
  // @heuristic::idea   jab a trick to serve next trick a ace

  // must not be the first player
  if (t.isstartcard())
    return Card::empty;

  // must be a trump trick or a color trick without having to serve
  if (   !t.startcard().istrump()
      && hi.hand().contains(t.startcard().tcolor()))
    return Card::empty;

  // do not jab as marriage player in a decision trick
  if (   hi.game().marriage().is_undetermined()
      && hi.is_soloplayer()
      && (is_selector(t.startcard().tcolor(),
                      hi.game().marriage().selector()) ) )
    return Card::empty;

  // do not overjab the partner
  if (   hi.guessed_same_team(t.winnerplayer())
      && t.winnercard().istrump()
      && !t.winnercard().less(hi.lowest_trump_card_limit()) ) {
    // check that all players behind do have no ace to play
    // since the winnerplayer is of the own team, the remaining players must be of the opposite team
    if (t.actcardno() == 2) {
      // the last player must serve the trick
      if (   !t.startcard().istrump()
          && hi.handofplayer(t.lastplayer()).contains(t.startcard().tcolor()))
        return Card::empty;
      auto c = hi.game().rule().card_colors().begin();
      for (; c != hi.game().rule().card_colors().end(); ++c)
        if (   (hi.color_runs(*c) == 0)
            && hi.handofplayer(t.lastplayer()).contains(*c, Card::ace))
          break;
      if (c == hi.game().rule().card_colors().end())
        return Card::empty;
    } // if (t.actcardno() == 2)
  } // do not overjab the partner


  // the ace to jab for
  Card const ace = Heuristics::choose_ace( Trick(t.startplayer()), hi );

  if (ace.is_empty())
    return Card::empty;

  unsigned const remaining_cards
    = (hi.game().cards().count(ace.color())
       - (hi.cards_information().played(ace.color()) + hi.hand().count(ace.color())));
  if(   is_solo( hi.game().type() )
     && !hi.is_soloplayer()
     && remaining_cards < 5 )
    return Card::empty;


  // for a solo don't jab for ace, if there are players of my team behind me
  // and myself doesn't have that many trumps
  if (   hi.game().is_solo()
      && (t.winnerplayer() == hi.game().players().soloplayer())
      && t.winnercard().less( hi.trump_card_limit() )
      && !t.islastcard()
      && (hi.hand().count_trumps() < 4)
     )
    return Card::empty;
  // solo game, color trick, goes to own team
  if (   hi.game().is_solo()
      && (hi.team() == Team::contra)
      && !t.startcard().istrump()
      && (t.winnerplayer() != hi.game().players().soloplayer())
      && (   t.has_played(hi.game().players().soloplayer())
          || (   (   (t.winnercard().value() == Card::ace)
                  || hi.same_team(t.player_of_card(t.actcardno() + 1)) )
              && (hi.color_runs(t.startcard().tcolor()) == 0)
              && hi.handofplayer(hi.game().players().soloplayer()).contains(t.startcard().tcolor())))
      && (hi.hand().count_trumps() < (hi.game().tricks().remaining_no() * 2) / 3)
     )
    return Card::empty;

  if(   (hi.hand().count( ace )
         < hi.game().rule()(Rule::Type::number_of_same_cards))
     && (hi.color_runs(ace.color()) == 0)
     && (   (maybe_to_team(hi.teamofplayer(t.winnerplayer())) != hi.team())
         || (   (maybe_to_team(hi.teamofplayer(t.winnerplayer())) == hi.team())
             && t.winnercard().less( hi.trump_card_limit() )
            )
         || (hi.game().tricks().current_no() == 0)
        )
    )
  {
    // do not overjab the partner, if he has played a high card
    // normal case: do not overjab a club queen
    if (   (   is_normal(hi.game().type())
            || is_marriage(hi.game().type())
            || is_poverty(hi.game().type())
            || is_color_solo(hi.game().type()))
        && (maybe_to_team(hi.teamofplayer(t.winnerplayer())) == hi.team())
        && !t.winnercard().less(Card::club_queen) )
      return Card::empty;

    Card c;
    if(   (hi.game().tricks().current_no() == 0)
       && t.startcard().istrump()
       && !t.islastcard()
       && !checkAllMyTeam(t, hi )) {
      c = highest_trump_card( t, hi.hand() );
      if (!t.isjabbed(c))
        return Card::empty;

    } else {
#ifdef OUTDATED
      // 0.7.11 (2011-09-04)
      c =  best_winning_card_old( t, hi,
                                 hi.value( Aiconfig::Type::limitdolle ) - ( 4 - t.actcardno() ) + 10 );
#endif
      c = best_jabbing_card(t, hi);

    }

    if( !c.is_empty() ) {
      if (  !t.startcard().istrump()
          && (hi.color_runs(t.startcard().tcolor()) == 0))
        return c;

      if (   hi.game().marriage().is_undetermined()
          && (is_selector(t.startcard().color(), hi.game().marriage().selector()))
          && (hi.team() == Team::re)
         )
        return c;

      Trick tempTrick = t;
      tempTrick.add(HandCard( hi.hand(), c ));
      if (oppositeTeamCanWinTrick( tempTrick, hi ) ) {
        // as second last player or if both players behind are of the opposite team, try with the highest trump
        if (   (   (t.remainingcardno() == 2)
                && (hi.game().cards().trumpcolor() != Card::nocardcolor))
            || (   (t.remainingcardno() == 3)
                && (hi.guessedteamofplayer(t.startplayer()) == hi.team())
                && t.winnercard().less(hi.trump_card_limit()))
           ) {
          c = hi.hand().highest_trump();
          tempTrick = t;
          tempTrick.add(HandCard(hi.hand(), c));
          if (oppositeTeamCanWinTrick(tempTrick, hi))
            c = hi.hand().highest_trump_till(Card::club_queen);
          if (!t.winnercard().less(c))
            return Card::empty;
        } else {
          return Card::empty;
        }

        tempTrick = t;
        tempTrick.add(HandCard(hi.hand(), c));
      } // if (opposite team can win the trick)

      bool trickLost = tempTrick.winnercard().less(highest_card_behind_of_opposite_team(tempTrick, hi));
      if(   trickLost
         && (!c.less(HandCard( hi.hand(), Card::diamond_queen)))
         && (hi.hand().count(Card::queen) < 3)
        )
        return Card::empty;
    }

    { // only take a high card if there are enough single aces or many points
      // count the single aces
      unsigned single_aces = 0;
      for (auto color : hi.game().rule().card_colors()) {
        Card const ace = Card(color, Card::ace);
        if (   !ace.istrump(hi.game())
            && (hi.hand().count(ace) == 1)
            && (hi.cards_information().played(color) == 0))
          single_aces += 1;
      }
      if (   hi.game().cards().less(Card::club_queen, c)
          && (single_aces < 2)
          && (t.points() < 2 + 4 * t.actcardno()))
        return Card::empty;
      if (   hi.game().cards().less(hi.trump_card_limit(), c)
          && (single_aces == 0)
          && (t.points() < 2 + 4 * t.actcardno()))
        return Card::empty;
    } // only take a high card if there are enough single aces


    return c;
  } // if ( .... (hi.color_runs(ace.color()) == 0) .... )

  return Card::empty;
} // Card Heuristics::jab_for_ace( Trick t, HeuristicInterface hi )


/** plays high and low trumps changing
 **
 ** @param    t         current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::play_low_high(Trick const& t, HeuristicInterface const& hi)
{
  // @heuristic::name   play low high
  // @heuristic::idea   play changing low and high trumps to force opponent to jab or either draw his trumps

  if(    is_solo( hi.game().type() )
     && !hi.is_soloplayer() )
    return Card::empty;

  // @heuristic::condition do not use for undetermined marriage
  if (hi.game().marriage().is_undetermined())
    return Card::empty;

  // @heuristic::condition trump cards at the opposite players still remain
  if (!opposite_trump_behind(t, hi))
    return Card::empty;


  if(   t.isstartcard()
     && (hi.hand().count_trumps() > hi.game().tricks().remaining_no() / 2)
     && (hi.hand().higher_cards_no(Card::diamond_jack)
         + hi.hand().count(Card::diamond, Card::jack)
         > 3)
    )
  {
    if(  hi.next_low()
       || (hi.game().tricks().current_no() < 4) )
    {
      if( hi.hand().count( Card::queen ) < 2 * hi.hand().count( Card::jack ) )
      {
        Card c = lowest_jack( hi.hand(), hi );

        if(   !c.is_empty()
           && !t.isvalid(c,hi.hand())) // ToDo: Hier wird doch die erste Karte im Trumpf betrachtet, daher dürfte die Bedingung nie greifen (siehe auch unten)
          c = Card::empty;

        return c;
      }
    } else
    { // if !(hi.next_low())
      if( hi.hand().count( Card::queen ) > 1)
      {
        Card c = lowest_queen( hi.hand(), hi );

        if(   (hi.hand().count( Card::queen ) < 4)
           && (hi.hand().count_club_queens() == 1)
           && (hi.game().teaminfo().get( hi.player() ) == Team::unknown)
          )
          c = Card::club_queen;


        if(   !c.is_empty()
           && !t.isvalid( c, hi.hand() ) )
          c = Card::empty;


        return c;
      }
    } // if !(hi.next_low())
  } // if (t.isstartcard())

  return Card::empty;
} // Card Heuristics::play_low_high(Trick t,HeuristicInterface hi)

/** puts a nice card to a trick of a partner
 ** returns a king, a ten (no dolle), an ace (no swine) or a nine
 ** the order depends on the card the partner has played
 **
 ** @param    t         current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::play_for_team( Trick const& t, HeuristicInterface const& hi )
{
  // @heuristic::name   play for team
  // @heuristic::idea   put a good card in a trick of the team mate

  // I am not the start player
  if( t.isstartcard() )
    return Card::empty;

  bool const soloPlayer = soloPlayerBehind( t, hi );

  if (   soloPlayer
      && (t.game().type() != GameType::poverty) )
    return Card::empty;

  // @heuristic::condition   partner has jabbed the trick with a high card
  if (maybe_to_team(hi.teamofplayer(t.winnerplayer())) != hi.team())
    return Card::empty;

  if (t.winnercard().less(hi.value(Aiconfig::Type::limitthrowing)))
    return Card::empty;

  Card c;
  for( unsigned i = 0; i < hi.hand().cardsnumber(); i++ ) {
    Trick tr = t;
    if( tr.isvalid( hi.hand().card(i), hi.hand() ) ) {
      tr.add(hi.hand().card( i ));
      if (   !hi.same_player(tr.winnerplayer())
          && hi.hand().card(i).value()==Card::king
          && !hi.hand().card(i).istrump())
        c = hi.hand().card( i );
    } // if ( tr.isvalid( ... ) )
  } // for ( i < hi.hand().cardsnumber() )

  if( !c.is_empty() )
    return c;


  for(unsigned i = 0; i < hi.hand().cardsnumber(); i++ ) {
    Trick tr = t;
    if( tr.isvalid( hi.hand().card( i ), hi.hand() ) )
    {
      tr.add(hi.hand().card( i ));
      // a ten only, if the player has got enough points on his hand
      if(   !hi.same_player(tr.winnerplayer())
         && hi.hand().card(i).value() == Card::ten
         && !tr.winnercard().less(hi.value(Aiconfig::Type::limithigh))
         && (hi.hand().count(Card::ten) + hi.hand().count(Card::ace)
             - hi.hand().count_dolle() - hi.hand().count_swines()
             >= 3)
         && !hi.hand().card(i).is_special()
        )
        c=hi.hand().card(i);
    } // if ( tr.isvalid ( ... ) )
  } // for ( i < hi.hand().cardsnumber() )

  if( !c.is_empty() )
    return c;

  for(unsigned i = 0; i < hi.hand().cardsnumber(); i++ ) {
    Trick tr = t;
    if( tr.isvalid( hi.hand().card(i),hi.hand() ) )
    {
      tr.add(hi.hand().card( i ));

      // last player after me is player with poverty
      bool lastPoverty = (soloPlayer
                          && tr.actcardno() == 3
                          && hi.game().players().soloplayer() == tr.player_of_card(3) );

      if(   !hi.same_player(tr.winnerplayer())
         && hi.hand().card(i).value() == Card::ace
         && (   !tr.winnercard().less( hi.value( Aiconfig::Type::limithigh ) )
             || t.islastcard()
             || lastPoverty
            )
         && !hi.hand().card( i ).isswine()
         && (hi.hand().count(Card::ten) + hi.hand().count(Card::ace)
             >= 3)
         && (   !hi.hand().card( i ).isfox()
             || t.islastcard()
             || lastPoverty
            )
        )
        c = hi.hand().card( i );
    } // if ( tr.isvalid ( ... ) )
  } // for ( i < hi.hand().cardsnumber() )

  if( !c.is_empty())
    return c;


#ifdef OUTDATED
  // DK 2011-10-15
  // ToDo: Wieso ist eine Neun ein Spiel 'für' das eigene Team? Da gibt doch ein Bube mehr Punkte
  for(unsigned i = 0; i < hi.hand().cardsnumber(); i++ ) {
    Trick tr = t;
    if( tr.isvalid( hi.hand().card(i), hi.hand() ) )
    {
      tr += hi.hand().card( i );
      if(   !hi.same_player(tr.winnerplayer())
         && hi.hand().card(i).value() == Card::nine
         && !hi.hand().card(i).ishyperswine()
        )
        c = hi.hand().card( i );
    } // if ( tr.isvalid( ... ) )
  } // for ( i < hi.hand().cardsnumber() )
#endif

  // @heuristic::action   play a small card with some value
  return c;
} // Card Heuristics::play_for_team( Trick t, HeuristicInterface hi )


/** jabs a fox if in this in the trick
 **
 ** @param    t         current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::jab_fox( Trick const& t, HeuristicInterface const& hi )
{
  // @heuristic::name   jab fox
  // @heuristic::idea   jab a fox if played in this trick

  if (!t.contains_fox())
    return Card::empty;

  // @heuristic::condition   it must not be an undetermined marriage
  if (   hi.game().marriage().is_undetermined()
      && is_selector(t.startcard().color(), hi.game().marriage().selector()))
    return Card::empty;

  // player can play a trump
  if (   !t.startcard().istrump()
      && hi.hand().contains(t.startcard().tcolor()))
    return Card::empty;

  Card c;

  // lastcard and this is our trick nothing to do.
  if ( (    t.islastcard()
        || !oppositeTeamCanWinTrick( t, hi ) )
      && hi.same_team(t.winnerplayer())) {
    return Card::empty;
  }

  // partner has jabbed the trick with a high trump
  if (   hi.same_team(t.winnerplayer())
      && !t.winnercard().less(hi.value(Aiconfig::Type::limithigh) )  )
    return Card::empty;

  // if lastcard just take the best winning card
  if (   t.islastcard()
      && !hi.same_team(t.winnerplayer())) {
    return best_jabbing_card(t, hi);
#ifdef OUTDATED
    // DK, 0.7.12 (2012-05-05)
    return best_winning_card_old(t, hi, 10 ); ///@todo 10 replace with AiConfig
#endif
  }


  { // If I can jab the fox surely, jab it.
    // take the best cards of all players behind me that are not in my team
    HandCard const& highest_opposite_card
      = highest_card_behind_of_opposite_team(t, hi);
    if (!hi.hand().highest_trump().less(highest_opposite_card)) {
      Card const c = optimized_card_to_play(lowest_highest_trump(t, hi), t, hi);
      if (!c.is_empty())
        return c;
    }
  } // If I can jab the fox surely, jab it.
#ifdef OUTDATED
  // DK, 5. Mai 2012 
  { // if I can jab the fox surely, take the card
    // take the best cards of all players behind me that are not in my team
    Trick const& trick = t;
    HandCard highest_card_behind(hi.hand());
    for (Player const* player = &(trick.game().players().following(trick.actplayer()));
         player != &trick.startplayer();
         player = &(trick.game().players().following(*player))) {
      if( hi.teamofplayer(*player) != hi.team() )
      {
        HandCard const highest_card = player->hand().highest_card();
        if(  highest_card_behind.is_empty()
           || highest_card_behind.less( highest_card )
          )
          highest_card_behind = highest_card;
      } // if (hi.teamofplayer(*player) != hi.team())
    } // for (player)

    if (!highest_card_behind.is_empty()) {
      // the partner has the best possible card
      if(   hi.teamofplayer(trick.winnerplayer()) == hi.team()
         && !trick.winnercard().less( highest_card_behind )
         /// @todo next_higher_card_of_other_team
        )
        return Card::empty;

      // look whether I have the best card
      Card lowest_better_card;
      HandCards const valid_cards = hi.hand().validcards( t );
      for (auto c : valid_cards) {
        Trick tempTrick = trick;
        tempTrick += c;
        if (   trick.isjabbed(c)
            && !tempTrick.isjabbed(highest_card_behind)
            && (   lowest_better_card.is_empty()
                || lowest_better_card.less(c)
               )
           )
          lowest_better_card = c;
      } // for (c)

      if (!lowest_better_card.is_empty())
        return lowest_better_card;
    } // if (highest_card_behind)
  } // if I can jab the fox surely, take the card
#endif

  if (   (hi.game().tricks().current_no() == 0)
      && t.startcard().istrump()
      && !t.islastcard())
  {
    auto const card = optimized_card_to_play(highest_trump_card(t, hi.hand()), t, hi);
    if (t.isjabbed(card))
      return card;
  }
  {
    // for all other we take our highest queen
    HandCard const card(hi.hand(), highest_queen(hi.hand(), hi));
    if (!card)
      return Card::empty;

    if (   t.isvalid(card)
        && t.isjabbed(card))
      return card;
  }

  return Card::empty;
} // Card Heuristics::jab_fox( Trick t, HeuristicInterface hi )

/** play single color nine or king to get fehl in second run of a color
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::create_fehl(Trick const& trick, HeuristicInterface const& hi)
{
  // @heuristic::name   create fehl
  // @heuristic::idea   create a missing color (only allowed for first card depending on Aiconfig::Type::fehlcreationonfirstcard)

  if (hi.game().marriage().is_undetermined())
    return Card::empty;

  // @heuristic::condition   this trick is still worth fehl creation
  if (!(static_cast<int>(hi.game().tricks().current_no())
        < hi.value(Aiconfig::Type::lastfehlcreation)))
    return Card::empty;

  // @heuristic::condition   this is the firstcard
  if (hi.value(Aiconfig::Type::fehlcreationonfirstcard))
    if (trick.isstartcard())
      return choose_best_fehl(trick, hi);

  // @heuristic::condition   this is not the first colorrun and there are not more then 10 points
  if(   !trick.isstartcard()
     && (hi.color_runs(trick.startcard().color()))
     && (static_cast<int>(trick.points())
         < hi.value(Aiconfig::Type::limit_throw_fehl))
    ) {
    // @heuristic::condition   - the trick is jabbed high enough
    if(   trick.winnercard().istrump()
       && !trick.winnercard().less(hi.lowest_trump_card_limit())
      ) {
      return choose_best_fehl(trick, hi);
    }

    // @heuristic::condition   - the winnercard is already a no trump ace
    if(   !trick.winnercard().istrump()
       && (trick.winnercard().value() == Card::ace))
      return choose_best_fehl(trick, hi);
  }
  // @heuristic::action   play the best fehl card

  return Card::empty;
} // Card Heuristics::create_fehl(Trick trick, HeuristicInterface hi)

/** finds highest card of color c
 **
 ** @param    trick    trick to check
 ** @param    hi       heuristic interface
 ** @param    tcolor   color to choose
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **
 ** @todo      play the lowest highest card
 **/
Card
Heuristics::play_highest_color(Trick const& trick,
                               HeuristicInterface const& hi,
                               Card::TColor const& tcolor)
{
  // @heuristic::name   meatless: play highest color
  // @heuristic::idea   Heuristic for solo::meatless to play the longest color or follow previous color

  HandCard const& card = hi.hand().highest_card(tcolor);

  if (card.value() == Card::ace)
    return card;

  // @heuristic::condition   the player can jab the trick
  if (!(   trick.isempty()
        || trick.winnercard().less(card)))
    return Card::empty;

  // @heuristic::condition   the player has the highest card
  if (card.less(hi.cards_information().highest_remaining_card_of_others(tcolor)))
    return Card::empty;

  // @heuristic::action   play the highest card of the color
  return card;
} // Card Heuristics::play_highest_color( Trick t, HeuristicInterface hi, Card::Color co )

/** As first player and the partner in backhand, take a small trump.
 ** prerequisites:
 **   1. first player in the trick
 **   2. last player is partner
 ** take a small trump
 ** If the partner has the highest card still in the game, take a trump
 ** with a high value (fox or ten)
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo      check that the partner has no problems with his trumps
 **/
Card
Heuristics::partner_backhand_draw_trump(Trick const& trick,
                                        HeuristicInterface const& hi)
{
  // @heuristic::name   partner backhand draw trump
  // @heuristic::idea   as first player and the partner in backhand, draw trump

  // @heuristic::condition   first player in the trick
  if (!trick.isstartcard() )
    return Card::empty;

  // @heuristic::condition   last player is partner
  if (!(hi.teamofplayer(trick.lastplayer()) == hi.team()))
    return Card::empty;

  // @heuristic::condition   the opposite team has not made any better announcement then the own team
  if (  hi.game().announcements().last(opposite(hi.team()))
      > hi.game().announcements().last(hi.team()))
    return Card::empty;

  // the partner
  Player const& partner = trick.lastplayer();

  bool partner_has_highest_trump = false;
  { // check whether the partner has the highest trump in the game
    HandCard const highest_trump
      = HandCard(hi.hand(),
                 hi.cards_information().highest_remaining_trump_of_others());
    if (hi.handofplayer(partner).contains(highest_trump))
    {
      // check that no other player does have the card, also
      if (   highest_trump.isdolle()
          && hi.game().second_dolle_over_first()
          && hi.cards_information(partner).must_have(Card::dolle)) {
        partner_has_highest_trump = true;
      } else {

        partner_has_highest_trump = true;
        for (unsigned p = 0; p < hi.game().playerno(); ++p) {
          Player const& player = hi.game().player(p);
          if (   (hi.teamofplayer(player) != hi.team())
              && (hi.handofplayer(player).contains(highest_trump))) {
            partner_has_highest_trump = false;
            break;
          }
        } // for (p < hi.game().playerno())
      } // if !(second dolle over first)
    } // if (partner has highest trump)
  } // check whether the partner has the highest trump in the game

  vector<Card> cards_to_play;
  if (partner_has_highest_trump) {
    cards_to_play.push_back(Card::diamond_ace);
    cards_to_play.push_back(Card::diamond_ten);
  } else if (hi.hand().count_trumps() >= 3) { // *Value*
    cards_to_play.push_back(Card::diamond_king);
    cards_to_play.push_back(Card::diamond_jack);
    cards_to_play.push_back(Card::heart_jack);
    cards_to_play.push_back(Card::spade_jack);
    cards_to_play.push_back(Card::diamond_nine);
  }

  // @heuristic::condition   at least three trump on the hand if the partner does not have the highest card
  // ToDo: improve

  for (auto c : cards_to_play) {
    if (  hi.hand().contains(c)
        && !HandCard(hi.hand(), c).is_special()) {
      return c;
    }
  }

  return Card::empty;
} // static Card Heuristics::partner_backhand_draw_trump(Trick trick, HeuristicInterface hi)

/** as last player grab the trick
 ** prerequisites:
 **   1. the player must be the last player or the players behind cannot get the trick
 **   2. the player can get the trick
 ** take the best card to win the trick
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::grab_trick(Trick const& trick,
                       HeuristicInterface const& hi)
{
  // @heuristic::name   grab trick
  // @heuristic::idea   as last player jab the trick if possible

  // @heuristic::condition   last player in the trick
  if (!trick.islastcard()
      && oppositeTeamCanWinTrick(trick, hi))
    return Card::empty;

  // @heuristic::condition   the trick does not go to the own team
  if (hi.guessedteamofplayer(trick.winnerplayer()) == hi.team())
    return Card::empty;

  // @heuristic::action   Play the best winning card.
  return Heuristics::best_winning_card_old(trick, hi,
                                           Card::ace * hi.game().playerno());
} // Card Heuristics::grab_trick(Trick trick, HeuristicInterface hi)


/** I can throw fehl and the trick is not worth taking; I'am third or forth
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::let_trick_pass(Trick const& trick,
                           HeuristicInterface const& hi)
{
  // @heuristic::name   let trick pass
  // @heuristic::idea   not enough points to grab trick, throw fehl

  // @heuristic::condition   third or forth player in trick
  if (trick.remainingcardno() > 2 )
    return Card::empty;

  Card::TColor const tcolor = trick.startcard().tcolor();
  // @heuristic::condition  color trick
  if (tcolor == Card::trump)
    return Card::empty;

  // @heuristic::condition  can't serve color
  if ( hi.hand().contains(tcolor))
    return Card::empty;

  // @heuristic::condition  There are not many points in this trick
  if (static_cast<int>(trick.points()) > hi.value(Aiconfig::Type::limit_throw_fehl) )
    return Card::empty;

  int modi = 0;
  if (!trick.islastcard()) {
    // ToDo: analyse the cards (and team) the last player has (t.i. estimate, what he will play)
    Player const& player = trick.lastplayer();
    Hand const& hand = hi.handofplayer(player);
    if (hi.guessedteamofplayer(player) == hi.team()) {
      if (hand.contains(tcolor))
        modi += hand.lowest_value(tcolor);
      else
        modi += hand.lowest_value();
    } else { // if !(same team)
      // @heuristic::condition   Do not leave the trick to the opposite team
      if (!hand.contains(tcolor)
          || (hi.guessedteamofplayer(trick.winnerplayer()) != hi.team()) )
        return Card::empty;
      if (hand.contains(tcolor))
        modi += hand.highest_value(tcolor);
      else
        modi += hand.highest_value();
    } // if !(same team)
  } // if (!trick.islastcard())


  // There are not many points in this trick
  if (static_cast<int>(trick.points()) + modi > hi.value(Aiconfig::Type::limit_throw_fehl) )
    return Card::empty;

  // the card to play (or none)
  Card const card = choose_best_fehl(trick, hi, true);
  modi += card.value();

  // There are not many points in this trick
  if (static_cast<int>(trick.points()) + modi > hi.value(Aiconfig::Type::limit_throw_fehl) )
    return Card::empty;

  // @heuristic::action   Play the best fehl
  return card;
} // Card Heuristics::let_trick_pass(Trick trick, HeuristicInterface hi)


/** get a trick for announcement
 **
 ** @param    t         trick to check
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::get_trick_for_keeping_announcement(Trick const& t,
                                               HeuristicInterface const& hi)
{
  // @heuristic::name   get trick for announcement
  // @heuristic::idea   If this trick would open the possibility for the other team to win the announcement try to win this trick for the own team

  // @heuristic::condition   the trick goes to the opposite team
  if (hi.guessed_same_team(t.winnerplayer()))
    return Card::empty;

  unsigned const own_points = calcPointsOfOwnTeam(hi, false);
  unsigned const opp_points = calcPointsOfOppositeTeam(hi, false);

  // @heuristic::condition   the trick is critical for keeping the announcement or for deflecting the announcement of the opposite team
  Card const card = best_jabbing_card(t, hi);
#ifdef OUTDATED
  = best_winning_card_old(t, hi, 30);
#endif
  if (   (   hi.game().needed_points_to_win(hi.team())
          > own_points)
      && (   hi.game().needed_points_to_win(hi.team())
          <= (own_points + t.points() + card.value()) ) )
    return card;
  if (   (   hi.game().needed_points_to_win(opposite(hi.team()))
          > opp_points)
      && (   hi.game().needed_points_to_win(opposite(hi.team()))
          <= (opp_points + t.points()
              + hi.hand().validcards(t).lowest_value()
              + ((t.remainingcardno() - 1) * 4) )
         ) )
    // @heuristic::action   take the best winning card
    return card;

  return Card::empty;
} // Card Heuristics::get_trick_for_keeping_announcement(Trick t, HeuristicInterface hi)

/** Gives the partner a chance to drop his worries
 **
 ** @param    trick current trick
 ** @param    hi    heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::play_for_partner_worries(Trick const& trick,
                                     HeuristicInterface const& hi)
{
  constexpr auto name = "play for partner worries";
  // @heuristic::name   play for partner worries
  // @heuristic::idea   Plays a card that will win the trick so that the partner can drop a "worry-card"
  auto const* partner = hi.team_information().guessed_partner();
  if (!partner) {
    DEBUG_HEURISTIC_REJECT << "no guessed partner\n";
    return Card::empty;
  }

  if (trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "startcard\n";
    return Card::empty;
  }

  if (!trick.startcard().istrump()) {
    DEBUG_HEURISTIC_REJECT << "no trump trick\n";
    return Card::empty;
  }

  if (trick.cardno_of_player(*partner) < trick.actcardno()) {
    DEBUG_HEURISTIC_REJECT << "partner has played\n";
    return Card::empty;
  }

  // the player can win the trick
  if (!card_will_win_trick(trick, hi, hi.hand().highest_card())) {
    DEBUG_HEURISTIC_REJECT << "cannot win the trick\n";
    return Card::empty;
  }

  { // check for worries
    auto const validcards = partner->hand().validcards(trick);
    // the player has worries
    if (!std::contains_if(validcards, [](auto const& card) {
                          return (   (card.value() >= 10)
                                  && !card.isdolle()
                                  && !card.is_special());
                          })) {
      DEBUG_HEURISTIC_REJECT << "partner has no worries\n";
      return Card::empty;
    }
  } // check for worries

  // if the partner has almost only trump, we do have time to get the worrie
  if (   (hi.handofplayer(*partner).count_trumps() + 1
          >= hi.handofplayer(*partner).cardsnumber())
      || (hi.cards_information(*partner).must_have(Card::trump) + 1
          >= partner->cards_to_play()) ) {
    DEBUG_HEURISTIC_REJECT << "the partner has almost only trump\n";
    return Card::empty;
  }

  { // search the best winning card
    Card winning_card;

    HandCards const validcards = hi.hand().validcards(trick);
    for (auto const& c : validcards) {
      if (   card_will_win_trick(trick, hi, c)
          && (   winning_card.is_empty()
              || c.less(winning_card))) {
        winning_card = c;
      }
    } // search for lowest winning card

    DEBUG_HEURISTIC_RETURN << "return: best winning card: " << winning_card << '\n';
    return winning_card; // return the lowest winningcard
  } // search the best winning card
} // Card Heuristics::play_for_partner_worries(Trick trick, HeuristicInterface hi)

/** Start a new color hoping the partner has an ace
 ** prerequisites:
 **   1. it must be the first card in the trick
 **   2. first run of the color
 **   3. nobody has played the color
 ** Then take the card with the greatest value in the shortest color
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::start_with_color(Trick const& trick,
                             HeuristicInterface const& hi)
{
  constexpr auto name = "start with color";
  // @heuristic::name   start with color
  // @heuristic::idea   play a color hoping the partner has the ace

  // @heuristic::condition   startplayer in the trick
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  // best card so far to play
  HandCard const* best_card = &HandCard::empty;
  auto const& next_player = trick.player_of_card(1);

  if (hi.guessed_opposite_team(next_player)) {
    DEBUG_HEURISTIC_REJECT << "next player of opposite team\n";
    return {};
  }

  // maximal remaining cards of the selected color
  // @heuristic::condition   or there are enough remaining cards for all players
  //                         (players - me)
  auto max_remaining_cards_no = hi.game().playerno() - 1;

  for (auto const color : hi.game().rule().card_colors()) {
    // test that the color is no trumpcolor
    // @heuristic::condition   first run of the color
    if (   (color == hi.game().cards().trumpcolor())
        || (hi.cards_information().played(color)
            > hi.cards_information(hi.player()).played(color)) )
      continue;

    // the player has the color
    if (hi.hand().count(color) == 0)
      continue;

    // @heuristic::condition   following player is not of the opposite team
    if (   hi.handofplayer(next_player).contains(color)
        && !hi.handofplayer(next_player).contains(color, Card::ace))
      continue;

    auto const remaining_cards_no
      = hi.cards_information().remaining_others(color);
    // there are at least as many remaining cards as the best so far
    if (remaining_cards_no < max_remaining_cards_no)
      continue;

    // search the card with the highest value
    auto const& card = hi.hand().highest_card(color);

    if (   (remaining_cards_no > max_remaining_cards_no)
        || (card.value() > best_card->value()) ) {
      best_card = &card;
      max_remaining_cards_no = remaining_cards_no;
    } // if (found better card)
  } // for (c \in possible colors)

  // @heuristic::action   play the card with the highest value in the smallest color
  if (!*best_card) {
    DEBUG_HEURISTIC_REJECT << "no fitting color found\n";
    return Card::empty;
  }
  DEBUG_HEURISTIC_RETURN << "smallest color, highest value – " << *best_card << '\n';
  return *best_card;
} // Card Heuristics::start_with_color(Trick trick, HeuristicInterface hi)

/** Start with a low color to find the partner / in a solo game
 **
 ** @param    trick   current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo      split for normal game and for solo
 **/
Card
Heuristics::start_with_low_color(Trick const& trick,
                                 HeuristicInterface const& hi)
{
  constexpr auto name = "start with low color";
  // @heuristic::name   start with low color
  // @heuristic::idea   Start the trick with a low color (in a normal game to find the partner, in a solo game always)

  // @heuristic::condition   first card in trick
  if (!trick.isstartcard()) {
    DEBUG_HEURISTIC_REJECT << "no startcard\n";
    return Card::empty;
  }

  // no known partner in normal game or a solo game
  if (   !is_solo(hi.game().type())
      && hi.team_information().guessed_partner() ) {
    DEBUG_HEURISTIC_REJECT << "no solo game and partner is known\n";
    return Card::empty;
  }

  // do not play for the opposite team
  if (hi.guessed_opposite_team(trick.player_of_card(1))) {
    DEBUG_HEURISTIC_REJECT << "opposite team behind\n";
    return Card::empty;
  }

  if (!hi.hand().hascolor()) {
    DEBUG_HEURISTIC_REJECT << "has no color\n";
    return Card::empty;
  }
  // Everything below need color cards!

  Card c;
  { // @heuristic::action   when a singleton color card is found, take it
    for (auto const& card : hi.hand().cards()) {
      if (card.points() >= 10)
        continue;
      if (   !card.istrump()
          && (hi.hand().count(card.color()) == 1)) {
        if (c.is_empty()) {
          c = card;
        } else if (hi.cards_information().remaining_others(card.color())
                   > hi.cards_information().remaining_others(c.color()) ) {
          c = card;
        } else if (hi.color_runs(c.color()) > 0) {
          if (hi.color_runs(card.tcolor()) == 0)
            c = card;
          else if (c.value() > card.value())
            c = card;
        } // if ( color_runs > 0)
      } // if( ! card.istrump && count(card.color) == 1)
    } // for (card : hi.hand().cards())

    if (   !c.is_empty()
        && (c.points() < 10)) {
      DEBUG_HEURISTIC_RETURN << "singleton color card = " << c << '\n';
      return c;
    }
  } // take singleton color card

  { // @heuristic::action   take smallest card from the longest color
    auto longestcolor = Card::nocardcolor;
    // Determine the longest color
    for (auto const color : hi.game().rule().card_colors()) {
      if (   longestcolor == Card::nocardcolor
          || hi.hand().count(color) > hi.hand().count(longestcolor)) {
        if (hi.hand().lowest_card(color).points() < 10)
          longestcolor = color;
      }
    } // for (color : hi.game().rule().card_colors())

    c = hi.hand().lowest_card(longestcolor); // choose the lowest card
  } // take smallest card from the longest color

  if (c.points() >= 10) {
    DEBUG_HEURISTIC_REJECT << "lowest color card (" << c << ") has at least 10 points\n";
    return Card::empty;
  }
  DEBUG_HEURISTIC_RETURN << "lowest color card = " << c << '\n';
  return c;
} // Card Heuristics::start_with_low_color(Trick trick, HeuristicInterface hi)
