/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

#define DEBUG_HEURISTIC \
  if (   (   ::debug.filter("heuristics") \
          || ::debug.filter(name)) \
      && !hi.game().isvirtual()) \
cout << __FILE__ << '#' << __LINE__ << " (" << trick.actcardno() << ')' << hi.player().name() << ": " << name << " – "

/** returns a card to play for the bride in an open marriage
 **
 ** @param    trick   the current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play (or Card())
 **/
Card
Heuristics::play_to_marry( Trick const& trick,
                          HeuristicInterface const& hi )
{
  // @heuristic::name   play to marry
  // @heuristic::idea   play card to marry
  if( !(   hi.game().type() == GameType::marriage
        && !hi.is_soloplayer()
       )
    )
    return Card::empty;

  Card::Color color = Card::nocardcolor;

  if (trick.isempty()) {
    switch (hi.game().marriage().selector()) {
    case MarriageSelector::silent:
    case MarriageSelector::team_set:
      return Card::empty;

    case MarriageSelector::first_foreign:
    case MarriageSelector::first_trump: {
      // @heuristic::condition   trick empty and first foreign or first trump
      // @heuristic::action      play the highest trump, if only the bride can jab
      Card const trump = lowest_highest_trump(trick, hi);
      if (trump.is_empty())
        return Card::empty;

      if (!trump.less(highest_card_behind_of_opposite_team(trick, hi))) {
        if (   (trump == Card::dolle)
            && !hi.cards_information().remaining_others(Card::dolle)
            && hi.hand().contains(Card::spade, Card::queen))
          return Card::spade_queen;
        else
          return trump;
      }

      return Card::empty;
    }
    case MarriageSelector::first_club:
      color = Card::club;
    case MarriageSelector::first_spade:
      if( color == Card::nocardcolor )
        color = Card::spade;
    case MarriageSelector::first_heart:
      if( color == Card::nocardcolor )
        color = Card::heart;

      // @heuristic::condition   trick empty and first specific color
      // @heuristic::action      play the ace of this color, if it can get through
      if (   (hi.hand().count(Card(color, Card::ace)) > 0)
          && (hi.cards_information().remaining_others(Card(color, Card::ace))
              >= hi.game().playerno() - 1) )
        return Card(color, Card::ace);

      return Card::empty;
    case MarriageSelector::first_color:
      // this is caught by 'play ace'
      return Card::empty;

    } // switch (hi.game().marriage().selector())
  } else { // if !(trick.isempty())

    // @heuristic::condition   must be a selector trick
    if (!is_selector(trick.startcard().tcolor(),
                     hi.game().marriage().selector()) )
      return Card::empty;

    // @heuristic::condition   the player has the highest trump and can play it
    // @heuristic::action      play the highest trump, if only the bride can jab
    Card const card = choose_for_color_trick(trick, hi);
    if (!card.is_empty())
      return card;

    Card const trump = lowest_highest_trump(trick, hi);
    if (trump.is_empty())
      return Card::empty;
    if (   trick.isvalid(trump, hi.hand())
        && trick.winnercard().less(trump)
        && !trump.less(highest_card_behind_of_opposite_team(trick, hi)))
      return trump;
  }  // if !(trick.isempty())

  return Card::empty;
} // static Card Heuristics::play_to_marry(Trick trick, HeuristicInterface hi)

/** returns a card to play for the bride in an open marriage
 **
 ** @param    trick   the current trick
 ** @param    hi      heuristic interface
 **
 ** @return   card to play (or Card())
 **/
Card
Heuristics::play_to_get_married( Trick const& trick,
                                HeuristicInterface const& hi )
{
  // @heuristic::name   play to get married
  // @heuristic::idea   play card to get married

  // @heuristic::condition   marriage player
  if( !(   hi.game().type() == GameType::marriage
        && hi.is_soloplayer()
       )
    )
    return Card::empty;

  // @heuristic::action   As first player, play card of the marriage color to get married,
  // @heuristic::action   else in a marriage trick, play pfund.
  Card::Color color = Card::nocardcolor;

  switch( hi.game().marriage().selector() ) {
  case MarriageSelector::silent:
  case MarriageSelector::team_set:
    return Card::empty;

  case MarriageSelector::first_foreign:
    if (trick.isempty()) {
      vector<Card> cards;
      for (auto c : hi.game().rule().card_colors()) {
        if (   (hi.hand().count( c ) < 3)
            && !HandCard( hi.hand(), Card(c, Card::ace)).possible_swine()
            && hi.game().tricks().current_no() < hi.game().rule()(Rule::Type::marriage_determination) - 1)
          cards.push_back(Card(c, Card::ace));
      } // for ( c : hi.game().rule().card_colors() )
      if ( !HandCard( hi.hand(), Card(hi.game().cards().trumpcolor(), Card::ace)).possible_swine() )
        cards.push_back( Card(hi.game().cards().trumpcolor(), Card::ace ));
      for (auto c : hi.game().rule().card_colors()) {
        if ( !HandCard( hi.hand(), Card(c, Card::ten)).isdolle() ) {
          cards.push_back(Card(c, Card::ten));
        }
      } // foreach ( color in hi.game().rule().card_colors() )

      for (auto color : hi.game().rule().card_colors()) {
        cards.push_back(Card(color, Card::king));
      }

      for (auto c : hi.game().rule().card_colors()) {
        if ( !HandCard( hi.hand(), Card(c, Card::nine)).possible_hyperswine() )
          cards.push_back(Card(c, Card::nine));
      }

      for (auto card : cards) {
        if (hi.hand().contains(card))
          return card;
      }

      return Card::empty;
    }
    return Heuristics::choose_pfund_card( trick, hi );

  case MarriageSelector::first_trump:
    if(   !trick.isempty()
       && trick.startcard().istrump() )
      return Heuristics::choose_pfund_card( trick, hi );

    if( trick.isempty() )
    {
      // take a 'good' card
      if(   hi.hand().count_trump_aces() > 0
         && !hi.hand().has_swines()
        )
        return Card(hi.game().cards().trumpcolor(), Card::ace);

      if(  hi.hand().count( Card( hi.game().cards().trumpcolor(), Card::ten ) ) > 0 )
        return Card(hi.game().cards().trumpcolor(), Card::ten);

      return Heuristics::lowest_best_trump_card(trick, hi.hand(), hi);
    } // if (trick.isempty())
    break;

  case MarriageSelector::first_club:
    color = Card::club;
  case MarriageSelector::first_spade:
    if( color == Card::nocardcolor )
      color = Card::spade;
  case MarriageSelector::first_heart:
    if( color == Card::nocardcolor )
      color = Card::heart;

    if(   !trick.isempty()
       && trick.startcard().tcolor() == color
      )
      return Heuristics::choose_pfund_card( trick, hi );

    if( trick.isempty() )
    {
      vector<Card> cards;
      if (!HandCard(hi.hand(), Card(color, Card::ten)).is_special())
        cards.push_back(Card(color, Card::ten));
      if (!trick.isjabbed(Card(color, Card::ace)))
        cards.push_back(Card(color, Card::ace));
      cards.push_back(Card(color, Card::king));
      cards.push_back(Card(color, Card::nine));
      cards.push_back(Card(color, Card::ace)); // With Ace I will probably win the trick bad decision
      for( vector<Card>::const_iterator c = cards.begin();
          c != cards.end();
          ++c )
        if( hi.hand().count( *c ) > 0 )
          return *c;
    } // if( trick.isempty() )
    break;

  case MarriageSelector::first_color:
    if( trick.isempty() )
    {
      vector<Card> cards;
      for (auto
           c = hi.game().rule().card_colors().begin();
           c != hi.game().rule().card_colors().end();
           ++c)
      {
        Card::Color const& color = *c;
        if ( !HandCard( hi.hand(), Card(color, Card::ten)).isdolle() )
          cards.push_back(Card(color, Card::ten));
      } // foreach ( color in hi.game().rule().card_colors() )
      for (auto color : hi.game().rule().card_colors())
      {
        cards.push_back(Card(color, Card::king));

      } // foreach ( color in hi.game().rule().card_colors() )
      for (auto color : hi.game().rule().card_colors())
      {
        cards.push_back(Card(color, Card::nine));
      } // foreach ( color in hi.game().rule().card_colors() )

      for( vector<Card>::const_iterator card = cards.begin();
          card != cards.end();
          ++card )
      {
        if( ( hi.hand().count( *card ) > 0 ) && !card->istrump( hi.game() )  )
          return *card;
      } // foreach ( card in cards )

      return Card::empty;
    } // if( trick.isempty() )

    if(    hi.same_player(trick.startplayer())
       && !trick.startcard().istrump()
      )
      return Heuristics::choose_pfund_card( trick, hi );

    if( hi.same_player(trick.startplayer()) )
      return Heuristics::lowest_card(trick, hi.hand());

    break;

  } // switch(marriage_selector)

  return Card::empty;
} // static Card Heuristics::play_to_get_married(Trick trick, HeuristicInterface hi)


