/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

#define DEBUG_HEURISTIC \
  if (   (   ::debug.filter("heuristics") \
          || ::debug.filter(name)) \
      && !hi.game().isvirtual()) \
cout << __FILE__ << '#' << __LINE__ << " (" << trick.actcardno() << ')' << hi.player().name() << ": " << name << " – "

// ToDo: own trumplimit value for poverty

/** If the poverty partner is the last player play many points
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo       if the opposite team has played a high card (p.e. dolle)
 **             do _not_ pfund although the own teammate is last player.
 ** @todo       if an opposite player is behind, check, that the winnercard
 **             of the teammate is high enough
 **                (-> play_pfund?)
 **/
Card
Heuristics::poverty_special_play_pfund(Trick const& trick,
                                       HeuristicInterface const& hi)
{
  // @heuristic::name        poverty: special: play pfund
  // @heuristic::idea        If the trick goes or will go to the own team, play a pfund. This heuristic should only be used by the poverty player.

  if (trick.isstartcard())
    return Card::empty;

  // @heuristic::condition   poverty player
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (!hi.is_soloplayer())
    return Card::empty;

  // @heuristic::condition   the last player must be of the own team (oneself or the partner)
  if (hi.teamofplayer(trick.lastplayer()) != hi.team() )
    return Card::empty;

  auto const& teammate = hi.game().players().poverty_partner();

  // @heuristic::condition   the own team has already the trick or the partner is behind
  if (   (maybe_to_team(hi.teamofplayer(trick.winnerplayer())) != hi.team())
      && (trick.cardno_of_player(teammate) < trick.actcardno()) )
    return Card::empty;

  // @heuristic::condition   if the partner is behind, he can still jab the trick
  if (!trick.winnercard().less(hi.handofplayer(teammate).highest_trump()))
    return Card::empty;

  // @heuristic::action      play a pfund
  // ToDo: Borg: use 'play_pfund'
  Card c;
  HandCards const ha = hi.hand().validcards( trick );

  // find any card that's allowed for this trick
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if( !ha.card(i).istrump() ) {
      c = ha.card(i);
      break;
    }

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for( unsigned i = 0; i < ha.cardsnumber(); i++ ) {
    if(   !ha.card(i).istrump()
       && (ha.card(i).value() != Card::ace)
       && (ha.card(i).value() > c.value())
       && (   (ha.count( ha.card(i).color() ) >  ha.count(c.color()))
           || (ha.count( ha.card(i).color(), Card::ace ) == 0)
          )
      )
      c = ha.card(i);
  } // for ( i < ha.cardsnumber() )

  return c;
} // Card Heuristics::poverty_special_play_pfund(Trick trick, HeuristicInterface hi)


/** If the poverty partner has played high and still many points on hand play many points.
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_special_offer_pfund(Trick const& trick,
                                        HeuristicInterface const& hi)
{
  // @heuristic::name        poverty: special: offer pfund
  // @heuristic::idea        If the poverty partner has played high and still many points on hand play many points.

  if (trick.isstartcard())
    return Card::empty;

  // @heuristic::condition   poverty player
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (!hi.is_soloplayer())
    return Card::empty;

  // @heuristic::condition   the partner has jabbed with a high card
  if (   (hi.teamofplayer(trick.winnerplayer()) != hi.team())
      || trick.winnercard().less(hi.value(Aiconfig::Type::limithigh) )  )
    return Card::empty;

  // @heuristic::condition   the player has many high cards
  if (hi.hand().count(Card::ten) + hi.hand().count(Card::ace)
      < hi.hand().cardsnumber() / 2)
    return Card::empty;


  Card c;
  HandCards const ha = hi.hand().validcards( trick );

  // find any card that's allowed for this trick
  for (unsigned i = 0; i < ha.cardsnumber(); ++i) {
    if (!ha.card(i).istrump()) {
      c = ha.card(i);
      break;
    } // if (!ha.card(i).istrump())
  } // for ( i < ha.cardsnumber())

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for (unsigned i = 0; i < ha.cardsnumber(); ++i) {
    if (   !ha.card(i).istrump()
        && (ha.card(i).value() != Card::ace)
        && (ha.card(i).value() >= c.value())
        && (   (   ha.count( ha.card(i).color() ) >  ha.count(c.color()))
            || (ha.count( ha.card(i).color(), Card::ace ) == 0)
           )
       )
      c = ha.card(i);
  }   // for (i < ha.cardsnumber())

  if (c.value() < Card::ten) {
    for (unsigned i = 0; i < ha.cardsnumber(); ++i) {
      if (   !ha.card(i).istrump()
          && (ha.card(i).value() >= c.value())
          && ha.count( ha.card(i).color() ) > ha.count(c.color())
         )
        c = ha.card(i);
    } // for ( i < ha.cardsnumber() )
  } // if (c.value() < Card::ten)

  // @heuristic::action      play a ten or ace of the color with the most cards
  return c;
} // Card Heuristics::poverty_special_offer_pfund(Trick trick, HeuristicInterface hi)


/** 1) poverty player
 ** 2) contra players get trick
 ** give as few points as possible
 **
 ** reason: do not give points to the opposite team
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo      choose a 'good' color (to keep as much freedom as possible)
 ** @todo      'anfüttern' of a high re-card with a king
 ** @todo      not in the first color run, if the re has or can get the trick
 **/
Card
Heuristics::poverty_special_give_no_points(Trick const& trick,
                                           HeuristicInterface const& hi)
{

  // @heuristic::name        poverty: special: give no points
  // @heuristic::idea        If the contra team gets the trick, give as few points as possible.


  // @heuristic::condition   the player must be the poverty player
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (!hi.is_soloplayer())
    return Card::empty;

  // the accepter of the poverty
  Player const& re_player = hi.game().players().poverty_partner();

  // check whether the contra team gets the trick:
  // @heuristic::condition   the partner has played already
  // @heuristic::condition   contra already has the trick or the re player can be overjabbed
  if (trick.cardno_of_player(re_player) > trick.actcardno())
    return Card::empty;

  if (   trick.islastcard()
      && (trick.winnerplayer() == re_player) )
    return Card::empty;

  if (   trick.startcard().istrump()
      && hi.hand().hastrump())
    return Card::empty;

  /// @todo   Use 'cards information' for checking that the contra team can jab.
  if (   (trick.winnerplayer() == re_player)
      && (!trick.winnercard().less(hi.value(Aiconfig::Type::trumplimit_normal))
          || (   !trick.startcard().istrump()
              && (hi.color_runs(trick.startcard().tcolor()) == 0) ) ) ) {
    return Card::empty;
  }

  // @heuristic::action   play the lowest card
  // search the lowest card
  // ToDo: check for a 'good' color
  // ToDo: create and use 'best_loosing_card'
  HandCards const cards = hi.hand().cards_single();
  HandCard card; // the card to play
  for (const auto & c : cards) {
    if (!trick.isvalid(c))
      continue ;
    if (card.is_empty()) {
      card = c;
      continue;
    }
    if (c.value() < card.value())
      card = c;
  } // for (c \in cards)

  return card;
} // Card Heuristics::poverty_special_give_no_points(Trick trick, HeuristicInterface hi)


/** 1) re player in a poverty
 ** 2) color trick, cannot serve it
 ** play high trump (small queen or so)
 **
 ** reason: jab for the points
 **
 ** note: Should be called after 'serve color trick'
 **       because this is overreacted for the first color run.
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo      for more points take a higher card
 **/
Card
Heuristics::poverty_re_trump_color_trick_high(Trick const& trick,
                                              HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: re: trump color trick high
  // @heuristic::idea   Jab the trick with a small queen in order to get the points

  // @heuristic::condition   re player (not poverty player) in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (   (hi.team() != Team::re)
      || hi.is_soloplayer())
    return Card::empty;

  // @heuristic::condition only my team behind me, I don't need to jab high
  if (checkAllMyTeam( trick, hi ) )
    return Card::empty;

  // @heuristic::condition   neither first nor last card in the trick
  if (trick.isstartcard())
    return Card::empty;
  if (trick.islastcard())
    return Card::empty;
  // @heuristic::condition   color trick which must not be served
  if (   trick.startcard().istrump()
      || hi.hand().contains(trick.startcard().tcolor()))
    return Card::empty;

  // @heuristic::condition not first color run
  if (hi.color_runs(trick.startcard().color()) == 0)
    return Card::empty;

  // @heuristic::action   play high trump (small queen)
  // ToDo: Borg: use 'choose_best_card'
  // search in the order one of the following cards to play
  vector<Card> cards_to_play;

  if(   hi.color_runs( trick.startcard().color() ) > 0
     && !trick.startcard().istrump()
     && trick.points() >= 10 )
  {
    cards_to_play.push_back(Card::spade_queen);
    cards_to_play.push_back(Card::heart_queen);
    cards_to_play.push_back(Card::diamond_queen);
    cards_to_play.push_back(Card::club_jack);
    cards_to_play.push_back(Card::spade_jack);
    cards_to_play.push_back(Card::heart_jack);
    cards_to_play.push_back(Card::diamond_jack);
  } else {
    cards_to_play.push_back(Card::heart_queen);
    cards_to_play.push_back(Card::diamond_queen);
    cards_to_play.push_back(Card::club_jack);
    cards_to_play.push_back(Card::spade_jack);
    cards_to_play.push_back(Card::spade_queen);
    cards_to_play.push_back(Card::heart_jack);
    cards_to_play.push_back(Card::diamond_jack);
  }

  for (vector<Card>::const_iterator c = cards_to_play.begin();
       c != cards_to_play.end();
       ++c)
    if (   (hi.hand().count(*c) > 0)
        && trick.winnercard().less(*c))
      return *c;

  return Card::empty;
} // Card Heuristics::poverty_re_trump_color_trick_high(Trick trick, HeuristicInterface hi)


/** 1) re player in a poverty
 ** 2) first player
 ** play a trump (diamond jack, ..., club queen)
 **
 ** reason: draw trump from the contra players
 **
 ** note: this heuristic shall be called after 'draw trump'
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_re_play_trump(Trick const& trick,
                                  HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: re: play trump
  // @heuristic::idea   As first player draw trump fro the contra players.

  // @heuristic::condition   re player (not poverty player) in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;

  if (   (hi.team() != Team::re)
      || hi.is_soloplayer())
    return Card::empty;


  // @heuristic::condition   first player
  if (!trick.isstartcard())
    return Card::empty;


  vector<Card> cards_to_play;

  // @heuristic::condition   partner has no trump
  if (hi.cards_information().of_player(hi.game().players().soloplayer()).does_not_have(Card::trump))
  {
    // @heuristic::action   play a small jack
    bool jacksAdded = false;
    if(   (hi.game().tricks().remaining_no() > hi.game().tricks().max_size() / 2)
       || (hi.hand().count( Card::jack ) > hi.hand().count( Card::queen) )
      )
    {
      jacksAdded = true;
      cards_to_play.push_back(Card::club_jack);
      cards_to_play.push_back(Card::spade_jack);
      cards_to_play.push_back(Card::heart_jack);
      cards_to_play.push_back(Card::diamond_jack);
    }
    cards_to_play.push_back(Card::diamond_queen);
    cards_to_play.push_back(Card::heart_queen);
    cards_to_play.push_back(Card::spade_queen);
    cards_to_play.push_back(Card::club_queen);
    if( ! jacksAdded )
    {
      cards_to_play.push_back(Card::club_jack);
      cards_to_play.push_back(Card::spade_jack);
      cards_to_play.push_back(Card::heart_jack);
      cards_to_play.push_back(Card::diamond_jack);
    } // if ( ! jacksAdded )
  } else // partner has trump
  {
    // @heuristic::comment   \n\n
    // @heuristic::condition   partner has trump
    // @heuristic::action   play a high trump
    if( hi.hand().count_dolle() == 2 )
      cards_to_play.push_back(Card::dolle);

    cards_to_play.push_back(Card::club_queen);
    cards_to_play.push_back(Card::spade_queen);
    cards_to_play.push_back(Card::heart_queen);
    cards_to_play.push_back(Card::diamond_queen);
    cards_to_play.push_back(Card::club_jack);
    cards_to_play.push_back(Card::spade_jack);
    cards_to_play.push_back(Card::heart_jack);
    cards_to_play.push_back(Card::diamond_jack);
  } // end of: partner has trump

  for (vector<Card>::const_iterator c = cards_to_play.begin();
       c != cards_to_play.end();
       ++c)
    if (hi.hand().count(*c) > 0)
      return *c;


  return Card::empty;
} // Card Heuristics::poverty_re_play_trump(Trick trick, HeuristicInterface hi)


/** 1) contra player in a poverty
 ** 2) first player
 ** play a color
 **
 ** reason: do not draw trump
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_contra_play_color(Trick const& trick,
                                      HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: contra: play color
  // @heuristic::idea   Start with a color.

  // @heuristic::condition   contra player in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (hi.team() != Team::contra)
    return Card::empty;

  // @heuristic::condition   first player
  if (!trick.isstartcard())
    return Card::empty;


  // the player with the poverty
  Player const& poverty_player = hi.game().players().soloplayer();
  // the accepter of the poverty
  Player const& re_player = hi.game().players().poverty_partner();
  // the partner
  // ToDo: this formula does only work with four players
  Player const& partner = hi.game().player( (0 + 1 + 2 + 3)
                                           - poverty_player.no()
                                           - re_player.no()
                                           - hi.player().no());

  // @heuristic::action   Play a color the partner does not have; if the partner is behind the re player, take a high card, else take a low one. Else take the longest color.
  { // search a color to play
    vector<Card::Color> const& colors = hi.game().rule().card_colors();
    vector<Card::Color>::const_iterator c;
    // first search for a color the partner does not have

    for (c = colors.begin();
         c != colors.end();
         ++c)
      if (   hi.cards_information().of_player(partner).does_not_have(*c)
          && hi.hand().contains(*c))
        break;

    if (c != colors.end()) {
      // There is a color the partner does not have.
      HandCards const cards = hi.hand().cards_single();
      HandCard card; // the card to play
      for (const auto & c : cards) {
        if (c.istrump())
          continue;
        if (!hi.cards_information().of_player(partner).does_not_have(c.tcolor()))
          continue;
        if (card.is_empty()) {
          card = c;
          continue;
        }
        // If the partner is behind the re player, take the highest card
        // else take the lowest
        if (   (c.value() > card.value())
            == (trick.cardno_of_player(partner)
                > trick.cardno_of_player(re_player)) )
          card = c;
      } // for (c \in cards)

      return card;

    } else { // if !(c != colors.end())
      // There is no color the partner does not have.
      { // search an ace
        Card::TColor best_color = Card::nocardcolor;
        for (auto color : colors)
          if (   hi.hand().contains(color, Card::ace)
              && hi.cards_information().remaining_others(color)) {
            if (best_color == Card::nocardcolor)
              best_color = color;
            else if (hi.handofplayer(re_player).contains(color)
                     && !hi.handofplayer(re_player).contains(best_color))
              best_color = color;
            else if (   (   hi.handofplayer(re_player).contains(color)
                         || !hi.handofplayer(re_player).contains(best_color))
                     && (hi.cards_information().remaining_others(color)
                         < hi.cards_information().remaining_others(best_color)
                        ) )
              best_color = color;
          } // if (hi.hand().contains(*c, Card::ace))
        if (best_color != Card::nocardcolor)
          return Card(best_color, Card::ace);
      }

      // Play the color I have the most number of.
      Card::TColor color = Card::nocardcolor;
      unsigned max_num = 0;
      for (auto c : colors)
        if (hi.hand().count(c) > max_num)
          color = c, max_num = hi.hand().count(c);

      // play the smallest card of the color
      HandCards const cards = hi.hand().cards_single();
      HandCard card; // the card to play
      for (const auto & c : cards) {
        if (c.tcolor() != color)
          continue;
        if (c.value() < card.value())
          card = c;
      } // for (c \in cards)

      return card;
    } // if !(c != colors.end())

  } // search a color to play

  return Card::empty;
} // Card Heuristics::poverty_contra_play_color(Trick trick, HeuristicInterface hi)


/** 1) contra player in a poverty
 ** 2) color trick, cannot serve it
 ** 3) re player is behind
 ** 4) a) partner before:
 **       - partner has played low
 **       - there are some points in the trick
 **    b) partner behind:
 **       - partner can have the color
 ** play high trump (small queen or so)
 **
 ** reason: The re player does in most cases not have the color
 **         so make the trick expensive for him.
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo       use 'best_winning_card()'
 **/
Card
Heuristics::poverty_contra_trump_color_trick_high(Trick const& trick,
                                                  HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: contra: trump color trick high
  // @heuristic::idea   Play a high trump (small queen) in color tricks, so that the trick is expensive for the re player.

  // @heuristic::condition   contra player in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (hi.team() != Team::contra)
    return Card::empty;

  // @heuristic::condition   a color trck the player must not serve
  if (trick.isstartcard())
    return Card::empty;
  if (   trick.startcard().istrump()
      || hi.hand().contains(trick.startcard().tcolor()))
    return Card::empty;


  // the player with the poverty
  Player const& poverty_player = hi.game().players().soloplayer();
  // the accepter of the poverty
  Player const& re_player = hi.game().players().poverty_partner();
  // the partner
  // Bug: this formula does only work with four players
  Player const& partner = hi.game().player( (0 + 1 + 2 + 3)
                                           - poverty_player.no()
                                           - re_player.no()
                                           - hi.player().no());


  // @heuristic::condition   the re player is behind
  if (trick.cardno_of_player(re_player) < trick.actcardno())
    return Card::empty;

  if (trick.cardno_of_player(partner) < trick.actcardno()) {
    // @heuristic::condition   a) the partner has played low (trump limit)
    if (!(   !trick.card_of_player(partner).istrump()
          || trick.card_of_player(partner).less(hi.value(Aiconfig::Type::trumplimit_normal)))
       )
      return Card::empty;

    // @heuristic::condition   a) there are some points (>= 8) in the trick
    if (trick.points() < 8) // *Value*
      return Card::empty;
  } else { // if !(partner before)
    // @heuristic::condition   b) the partner is behind
    // @heuristic::condition   b) the partner can have the color
    if (hi.cards_information().of_player(partner).does_not_have(trick.startcard().tcolor()))
      return Card::empty;
  } // if !(partner before)
  // the partner was before and has played low

  // @heuristic::action   play high trump (small queen)
  // ToDo: Borg: use 'best_winning_card'
  // search in the order one of the following cards to play
  vector<Card> cards_to_play;
  cards_to_play.push_back(Card::heart_queen);
  cards_to_play.push_back(Card::diamond_queen);
  cards_to_play.push_back(Card::club_jack);
  cards_to_play.push_back(Card::spade_jack);
  cards_to_play.push_back(Card::spade_queen);
  cards_to_play.push_back(Card::heart_jack);
  cards_to_play.push_back(Card::diamond_jack);

  for (vector<Card>::const_iterator c = cards_to_play.begin();
       c != cards_to_play.end();
       ++c)
    if (hi.hand().count(*c) > 0)
      return *c;

  return Card::empty;
} // Card Heuristics::poverty_contra_trump_color_trick_high(Trick trick, HeuristicInterface hi)


/** 1) contra player in a poverty
 ** 2) re has played trump
 ** 3) partner is behind and can jab the trick
 ** play a small trump so that the partner can get the trick
 **
 ** reason: make myself be behind the re player
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_leave_to_partner(Trick const& trick,
                                     HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: contra: leave to partner
  // @heuristic::idea   Make myself be behind the re player.

  // @heuristic::condition   contra player in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (   (hi.team() != Team::contra)
      && !hi.is_soloplayer())
    return Card::empty;

  // the player with the poverty
  Player const& poverty_player = hi.game().players().soloplayer();
  // the accepter of the poverty
  Player const& re_player = hi.game().players().poverty_partner();
  // the partner
  // Bug: this formula does only work with four players
  Player const& partner = hi.game().player( (0 + 1 + 2 + 3)
                                           - poverty_player.no()
                                           - re_player.no()
                                           - hi.player().no());


  // @heuristic::condition   the re player has played a trump
  // more precise: it has been started with trump and re has already played
  // (contains the case that the poverty player has started with trump)
  if (trick.isstartcard())
    return Card::empty;
  if (!trick.startcard().istrump())
    return Card::empty;
  if (trick.cardno_of_player(re_player) > trick.actcardno())
    return Card::empty;

  // @heuristic::condition   the partner is behind and can jab the trick
  if (trick.cardno_of_player(partner) < trick.actcardno())
    return Card::empty;
  if (!trick.winnercard().less(hi.handofplayer(partner).highest_trump()))
    return Card::empty;
  // better: check over trumplimit_normal, whether I know, my player has a higher card (p.e. club queen, swine)
  if (!trick.winnercard().less(hi.value(Aiconfig::Type::trumplimit_normal)))
    return Card::empty;

  // @heuristic::action   play a low trump
  // ToDo: Borg: use 'play_pfund'?
  vector<Card> cards_to_play;

  // special case:
  //   poverty player behind has still trump and
  //   re player has played a small trump
  if (!(   (hi.game().type() == GameType::poverty)
        && (trick.startplayer().team() == Team::re)
        // ToDo: better card value
        && trick.startcard().less(hi.value(Aiconfig::Type::limithigh))
        && (trick.cardno_of_player(hi.game().players().soloplayer()) > trick.actcardno())
        && !hi.cards_information().of_player(hi.game().players().soloplayer()).does_not_have(Card::trump)) ) {
    if (!hi.hand().has_swines())
      cards_to_play.push_back(Card::diamond_ace);
    cards_to_play.push_back(Card::diamond_ten);
  } // if (special case poverty)
  cards_to_play.push_back(Card::diamond_king);
  cards_to_play.push_back(Card::diamond_nine);
  // ToDo: Hand function 'next greater card'
  cards_to_play.push_back(Card::diamond_jack);
  cards_to_play.push_back(Card::heart_jack);
  cards_to_play.push_back(Card::spade_jack);
  cards_to_play.push_back(Card::club_jack);
  cards_to_play.push_back(Card::diamond_queen);
  cards_to_play.push_back(Card::heart_queen);
  cards_to_play.push_back(Card::spade_queen);
  cards_to_play.push_back(Card::club_queen);

  for (auto c : cards_to_play) {
    if (hi.hand().count(c) > 0) {
      if (c == trick.winnercard()) {
        if (hi.hand().contains(hi.game().cards().next_higher_card(c)))
          return hi.game().cards().next_higher_card(c);
        else if (hi.hand().contains(hi.game().cards().next_higher_card(c, 2)))
          return hi.game().cards().next_higher_card(c, 2);
      }
      return c;
    }
  }

  return Card::empty;
} // Card Heuristics::poverty_leave_to_partner(Trick trick, HeuristicInterface hi)


/** 1) contra player in a poverty
 ** 2) re player has won with a trump
 ** 3) can play a better card than the re player
 ** play next higher card
 **
 ** reason: If you can, overjab the re player in order to get points.
 **
 ** note: Should be called after 'leave to partner'
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_overjab_re(Trick const& trick,
                               HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: contra: overjab re
  // @heuristic::idea   Overjab the re player in order to get the points.

  // @heuristic::condition   contra player in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (hi.team() != Team::contra)
    return Card::empty;

  // @heuristic::condition   the re player has won so far with a trump
  if (trick.winnerplayer() != hi.game().players().poverty_partner())
    return Card::empty;
  if (!trick.winnercard().istrump())
    return Card::empty;

  // @heuristic::condition   the player can play a better card then the re player
  if (!(   trick.startcard().istrump()
        || !hi.hand().contains(trick.startcard().tcolor()) ) )
    return Card::empty;

  // @heuristic::action   Take the best winning card.
  //return Heuristics::best_jabbing_card(trick, hi);
  return hi.hand().next_jabbing_card(trick.winnercard());
} // Card Heuristics::poverty_overjab_re(Trick trick, HeuristicInterface hi)


/** best winning card for the last card
 *
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::poverty_best_winning_card(Trick const& trick,
                                      HeuristicInterface const& hi)
{
  // @heuristic::name   poverty: best winning card
  // @heuristic::idea   As last player in a trick take the lowest card to win the trick.

  // @heuristic::condition   only for the not poverty player in a poverty
  if (hi.game().type() != GameType::poverty)
    return Card::empty;
  if (   (hi.team() != Team::re)
      || hi.is_soloplayer())
    return Card::empty;

  // @heuristic::condition   last player in the trick
  if (!trick.islastcard() )
    return Card::empty;

  // @heuristic::action   Play the best winning card.
  return Heuristics::best_winning_card(trick, hi);
} // Card Heuristics::poverty_best_winning_card(Trick trick, HeuristicInterface hi)

