/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "draw_trump.h"

#include "../ai.h"
#include "../team_information.h"
#include "../cards_information.h"
#include "../cards_information.of_player.h"
#include "../heuristics.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"


/** constructor
 **/
Heuristic::DrawTrump::DrawTrump(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::draw_trump)
{ }

/** destructor
 **/
Heuristic::DrawTrump::~DrawTrump() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::DrawTrump::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   this->condition_startcard(trick)
        && this->condition_opposite_team_has_trump()
        && !this->condition_opposite_team_announcement()
        && this->condition_has_many_trumps()
       ))
    return {};

  auto const& game = this->player().game();
  if (   game.type() == GameType::normal
      && this->team() == Team::re
      && this->hand().count_all(Card::club_queen) == 1) {
    auto const& team_information = this->ai_->team_information();
    auto const guessed_partner = team_information.guessed_partner();
    auto const& trick = game.tricks().current();
    auto partner_behind = (   guessed_partner
                           && !trick.has_played(*guessed_partner));
    partner_behind = (   partner_behind
                      || trick.isstartcard()
                      || team_information.guessed_team(trick.player_of_card(0)) == Team::contra);
    auto const& hand = this->hand();
    auto const& cards_information = this->ai_->cards_information();
    if (   partner_behind
        && cards_information.played(Card::club_queen) == hand.played(Card::club_queen)) {
      this->rationale_.add(_("Heuristic::condition::re partner behind with club queen on the hand"));
      auto const c = hand.lowest_trump();
      this->rationale_.add(_("Heuristic::return::low trump: %s", _(c)));
      return c;
    }
  }

  auto c = Heuristics::lowest_highest_trump(trick, *this->ai_);
  if (   (   c.is_special()
          || (c.points() >= 10))
      && c.less(Heuristics::highest_card_behind_of_opposite_team(trick, *this->ai_))) {
    this->rationale_.add(_("Heuristic::reject::%s can be jabbed by the opposite team", _(c)));
    return Card::empty;
  }

  this->rationale_.add(_("Heuristic::return::lowest highest trump: %s", _(c)));
  return c;
}

bool
Heuristic::DrawTrump::condition_has_many_trumps()
{
  auto const& game = this->player().game();
  auto const& hand = this->hand();
  auto const& cards_information = this->ai_->cards_information();
  auto const trumpno = hand.count_trumps();
  auto const other_trumpno = cards_information.remaining_trumps_others();

  if (other_trumpno >= trumpno + 6) {
    this->rationale_.add(_("Heuristic::condition::the others have it least 6 more trumps"));
    return false;
  }

  if (   is_solo(game.type())
      && this->team() == Team::re
      && other_trumpno >= trumpno * 1.5
     ) {
    this->rationale_.add(_("Heuristic::condition::soloplayer, the others have at least 1.5 times of trumps"));
    return false;
  }

  if (other_trumpno >= 2 * trumpno) {
    this->rationale_.add(_("Heuristic::condition::the others have at least two times of trumps"));
    return false;
  }

  this->rationale_.add(_("Heuristic::condition::many trumps"));
  return true;
}
