/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "play_trump.h"

#include "../ai.h"
#include "../team_information.h"
#include "../heuristics.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"


/** constructor
 **/
Heuristic::PlayTrump::PlayTrump(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::play_trump)
{ }

/** destructor
 **/
Heuristic::PlayTrump::~PlayTrump() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::PlayTrump::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   !this->condition_startcard(trick)
        && !this->condition_soloplayer_behind(trick)
       ))
    return {};

  auto const& game = trick.game();
  if (game.type() == GameType::poverty) {
    return this->get_card_in_poverty(trick);
  }

  auto const& hand = this->hand();
  auto const& team = this->player().team();
  // @heuristic::condition   only trumps
  if (hand.hascolor()) {
    this->rationale_.add(_("Heuristic::reject::still color on the hand"));
    return Card::empty;
  }

  // @heuristic::action      if the trick goes to the own team: play low card
  if (   (trick.winnerteam() == team)
      && (game.cards().less(this->trump_card_limit(), trick.winnercard()))) {
    auto const card = this->lowest_best_trump_card(trick);
    this->rationale_.add(_("Heuristic::return::own team has already jabbed high: take low trump %s", _(card)));
    return card;
  }

  auto const card = Heuristics::best_jabbing_card(trick, *this->ai_, true);

  // @heuristic::action      if the trick has enough points: jab it
  if (trick.points()
      > ((1.2 * game.remaining_points())
         / game.tricks().remaining_no()) ) {
    this->rationale_.add(_("Heuristic::return::many points in the trick: jab with %s", _(card)));
    return card;
  }
  // @heuristic::action      if the player has enough high cards: jab it
  if (   (trick.winnerteam() != team)
      && (hand.higher_cards_no(trick.winnercard())
          + hand.count(trick.winnercard())
          > hand.cardsnumber() / 2)) {
    this->rationale_.add(_("Heuristic::return::many high cards in the hand: jab with %s", _(card)));
    return card;
  }

  // special case: two fox on the hand, not many trumps and partner behind: try fox
  auto const guessed_partner = this->team_information().guessed_partner();
  if (   !game.is_solo()
      && (hand.count_trump_aces() == 2)
      && !game.rule()(Rule::Type::swines)
      && (hand.count_trumps() <= 4)
      && guessed_partner
      && !trick.has_played(*guessed_partner)
      && (   trick.startcard().istrump()
          || !this->guessed_hand(*guessed_partner).contains(trick.startcard().tcolor()) )
     ) {
    auto const card = Card(game.cards().trumpcolor(), Card::ace);
    this->rationale_.add(_("Heuristic::return::two fox on the hand, not many trumps, partner behind: try to save %s", _(card)));
    return card;
  }

  // @heuristic::action      else: play low card
  if (   this->trump_card_limit().less(trick.winnercard())
      && (trick.points() + 3 * trick.remainingcardno() < 16)) { // *Value*
    // ToDo: if there is no card greater then a fox in the game, play such
    auto const card = Heuristics::lowest_best_trump_card(trick, hand, *this->ai_);
    this->rationale_.add(_("Heuristic::return::few points, play low trump %s", _(card)));
    return card;
  }

  if (game.cards().less(this->trump_card_limit(), card)) { 
    this->rationale_.add(_("Heuristic::reject::low card to jab selected, just let it pass"));
    return Card::empty;
  }

  // ToDo: perhaps always return no card, because the player should have enough high trumps
  this->rationale_.add(_("Heuristic::return::just jab with %s", _(card)));
  return card;
}

/** @return  card to play or empty card for a poverty
 **/
Card
Heuristic::PlayTrump::get_card_in_poverty(Trick const& trick)
{
  if (this->team() == Team::re)
    return this->get_card_in_poverty_re(trick);
  else if (this->team() == Team::contra)
    return this->get_card_in_poverty_re(trick);
  else
    return {};
}

/** @return  card to play or empty card for a poverty (re player)
 **/
Card
Heuristic::PlayTrump::get_card_in_poverty_re(Trick const& trick)
{
  if (this->only_own_team_behind(trick)) {
    auto const card = Heuristics::best_jabbing_card(trick, *this->ai_, true);
    this->rationale_.add(_("Heuristic::return::poverty re and no opponent behind: take best jabbing card %s", _(card)));
    return card;
  }
  auto const color = trick.startcard().tcolor();
  if (this->color_runs(color) == 0) {
    auto const card = Heuristics::best_winning_card(trick, *this->ai_);
    this->rationale_.add(_("Heuristic::return::poverty re and first color run: take best winning card %s", _(card)));
    return card;
  }
  this->rationale_.add(_("Heuristic::reject::poverty, but opposite player behind and %s was already played", _(color)));
  return {};
}

/** @return  card to play or empty card for a poverty (contra player)
 **/
Card
Heuristic::PlayTrump::get_card_in_poverty_contra(Trick const& trick)
{
  auto const& game = trick.game();
  auto const& hand = this->hand();
  if (hand.contains(trick.startcard().tcolor())) {
    this->rationale_.add(_("Heuristic::reject::color in the hand"));
    return {};
  }
  Player const& re_player = game.players().poverty_partner();
  if (trick.has_played(re_player)) {
    this->rationale_.add(_("Heuristic::reject::re player has already played"));
    return {};
  }
  auto const trumpcolor = game.cards().trumpcolor();
  if (hand.contains(Card(trumpcolor, Card::ten)) > 0 ) {
    auto const card = Card(trumpcolor, Card::ten );
    this->rationale_.add(_("Heuristic::return::contra and first color run: take %s", _(card)));
    return card;
  }

  auto const card = this->lowest_best_trump_card(trick);
  this->rationale_.add(_("Heuristic::return::contra and first color run: take lowest best trump card %s", _(card)));
  return card;
}
