/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "trickweighting.h"

#include "virtual_games_interface.h"
#include "ai.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../game/game.h"
#include "../../game/exception.h"
#include "../../card/trick.h"
#include "../../misc/setting.h"
#include "../../ui/ui.h"

#ifdef BENDERS
#ifndef RELEASE
// whether to print some weighting informations
#define PLAYERCHECK 10

// Playernumber for which to produce output
unsigned playerCheck;
#endif
#endif


/** @return   whether the opposite team can win the trick
 **/
bool 
oppositeTeamCanWinTrick(Trick const& trick,
                        VirtualGamesInterface const& hi)
{
  if (trick.winnerplayer().team() != hi.team())
    return true;

  // take a look if all players coming in this trick of other team can still win the trick
  for (auto i = trick.actcardno(); i < hi.game().playerno(); ++i) {
    auto const& player = trick.player_of_card(i);
    if (hi.guessed_opposite_team(player)) {
      auto const& hand = hi.handofplayer(player);
      if (hand.higher_card_exists(trick.winnercard()))
        return true;
    }
  }
  return false;
} // bool oppositeTeamCanWinTrick ( Trick t, HeuristicInterface hi )

/** Modification for dolle and swines
 **/
int
TrickWeighting::modi_dolle_swines(VirtualGamesInterface const& vgi,
                                      Trick const& trick,
                                      Team const team,
                                      HandCard const& card,
                                      Team const winnerteam,
                                      Ai const& ai)
{
  int modi = 0;

  if( !card.istrump() )
    return modi;

  if (!(   card.isswine()
        || card.ishyperswine()
        || card.isdolle()))
    return modi;

  // have played a very high card

  auto const& hand = ai.hand();
  auto const& game = ai.game();
  auto const& swines = game.swines();
  auto const& cards = game.cards();

  // 2005-08-06: swines in the last three tricks
  // modi -= 29; // previous values: 23,14,21,24, 29, 33, 35,31
  modi -= (31 - std::max(static_cast<int>(cards.count_played()) - 9 , 0));
  // previous: 34, 31 (Reference 92672)
  // previous: no()-12

  if( trick.startcard().istrump() )
    modi -= 11; // reference

  if (  (static_cast<int>(trick.points()) - card.value()
         < ai.value(Aiconfig::Type::limitqueen)))
    modi-= 4;

  if (maybe_to_team(winnerteam) != team)
    modi -= 26;

  if (maybe_to_team(winnerteam) == team) {
    if (    // review Reference 260689
        trick.startplayer().no() == vgi.no() )  // serving with a high trump and something potential higher is still out there
    {
      if (    hand.count_swines() +
          hand.count_hyperswines() +
          hand.count( Card::dolle ) != hand.count_trumps() ) {
        modi -= 10; // reference
      }

      //high trump serving is generally not a good choice
      if (!(is_solo(vgi.game().type())
            && (vgi.game().players().soloplayer() == ai)))
        modi -= 19; // reference 274918
    }
  }

  if(   (vgi.no() != trick.winnerplayer().no())
     && !(  trick.cardno_of_player( card.player() ) == 3
          && trick.winnerplayer().team() == vgi.team() )
    ) {
    modi -= (36 - std::max(static_cast<int>(cards.count_played()) - 13, 0));
    modi += trick.points()/2;

    if( trick.cardno_of_player( trick.winnerplayer() )
       < trick.cardno_of_player( card.player() )
       || !trick.jabs_cards_before(  card.player() ))
      modi -= 16; // previously 9, 12
  }

  // not for all three cards

  if (  ! card.isdolle()) // so that a dolle is prefered to a swine
    modi -= 8;  //previous values: 2, 7,9

  if (  trick.startplayer().no() == vgi.no() )
  {
    // previous: 12, 18
    modi -= 30;  // 23: 002645, 30:
    if (    hand.count_swines() +
        hand.count_hyperswines() +
        hand.count(Card::dolle) == hand.count_trumps() )
    {

      if( card.isdolle()
         && (!swines.swines_announced()
             || ai.teamofplayer(swines.swines_owner()) == ai.team() )
         && (!swines.hyperswines_announced()
             || ai.teamofplayer(swines.hyperswines_owner()) == ai.team()) )
      {
        modi += 10;
      }
      if (  card.isswine()
          && (   !swines.hyperswines_announced()
              || ai.teamofplayer(swines.hyperswines_owner()) == ai.team())) {
        modi += 30;
      }
      if (card.ishyperswine()) {
        modi += 50;
      }

    } else {
      if( is_solo( vgi.game().type() )  // reference 246985
         && vgi.is_soloplayer() ) //2do no higher card still in game
        modi += 21;
    }
  }


  if (  trick.cardno_of_player(card.player()) == 1
      && vgi.game().cards().count_played(Card::dolle) == 0 //2do before trick
      && card.isdolle()
      && vgi.game().second_dolle_over_first() )
    modi -= 5;

  if (  trick.cardno_of_player(card.player()) == 0
      && vgi.game().cards().count_played(Card::dolle) == 0 //2do before trick
      && card.isdolle()
      && vgi.game().second_dolle_over_first() )
    modi -= 1; // reference 197840

  // depending on data in the trick

  Trick t(trick.startplayer());
  for (unsigned i = 0; i < trick.cardno_of_player(card.player()); ++i) {
    t.add(trick.card( i ));
  }

  if(  trick.cardno_of_player(card.player()) == 2
     && card.isdolle()
     && swines.swines_announced()  //2do extend to hyperswine, do same for swine hyperswine
     && trick.cardno_of_player(swines.swines_owner()) == 3 ) {
    // swine behind dolle
    modi -= 15;  // reference 176246
  }

  if(  trick.cardno_of_player(card.player()) == 1
     && card.isdolle()
     && swines.swines_announced()  //2do extend to hyperswine, do same for swine hyperswine
     && trick.cardno_of_player(swines.swines_owner()) >1 ) {
    // swine behind dolle
    modi -= 3;  // reference 010523
  }


  if(  card.isdolle()
     && swines.swines_announced()  //2do extend to hyperswine, do same for swine hyperswine
     && trick.cardno_of_player(swines.swines_owner()) < trick.cardno_of_player(card.player())) {
    // swines in game, but before player with dolle
    modi +=13; // reference 088597, 114803
  }

  if (trick.cardno_of_player(card.player()) == 3) {
    modi -= 21; // previous 18, 20, 24

    if (t.winnerplayer().team() == vgi.team()) { // reference 43530
      modi -= 9; // previous
    }

    if (t.winnerplayer().team() != vgi.team()) { // reference
      modi -= 19; // previous
    }
  }



  if ( ai.value(Aiconfig::Type::limithigh).less(t.winnercard()) // !oppositeTeamCanWinTrick( t, vgi ) ) 2do XXX
    && t.winnerplayer().team() == team ) {
      modi -= 26;
      if( is_solo( vgi.game().type() )
         && !vgi.is_soloplayer())
        modi -= 4;
    }


  if ( ( card.possible_swine() //2do for a swine possible_swine should return false?
        && !card.isswine() )
      || (  card.possible_hyperswine()
          && !card.ishyperswine()))
    modi += 9;


  // only for soli
  if( is_solo( vgi.game().type() )
     && vgi.is_soloplayer())
    modi += 3;

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MDS"<< modi << "\t";

  }
#endif
  return modi;
}


/** Modification for a queen
 **/
int
TrickWeighting::modi_queen(VirtualGamesInterface const& vgi,
                           Trick const& trick,
                           Team const team,
                           HandCard const& card, 
                           Team const winnerteam,
                           Ai const& ai)
{
  if (card.value() != Card::queen)
    return 0;
  if (!card.istrump())
    return 0;

  int modi = 0;
  Player const& ai_virt = trick.game().player(ai.no());

  modi -= 4; // previous 2, 3
  // 4: Reference 188872

  if (
      (static_cast<int>(trick.points()) - card.value()
       < ai.value(Aiconfig::Type::limitqueen)))
    modi-= 4;

  if ( trick.cardno_of_player(card.player()) == 3 )
  {
    modi -= 13; // previous 11, 13, 17, 14
    // 13 reference 043530

    Trick t( trick.startplayer());
    t.add(trick.card( 0 ));
    t.add(trick.card( 1 ));
    t.add(trick.card( 2 ));

    if (t.winnerplayer().team() == vgi.team()) {
      modi -= 7; // previous

    }
  }

  if (is_solo(vgi.game().type())
      && (vgi.game().players().soloplayer().no() == ai.no())
      && vgi.trump_card_limit().less(card)
      && card.less(ai.cards_information().highest_remaining_trump_of_others()))
    modi -= 8; // reference 246985



  if(   (  vgi.no() != trick.winnerplayer().no()
         && !(  trick.cardno_of_player( card.player() ) == 3
              && trick.winnerplayer().team() == vgi.team() )
        )

    )
  {
    modi -= (36 - std::max(static_cast<int>(ai.game().cards().count_played()) - 13, 0));
    modi += trick.points()/2;

    if( trick.cardno_of_player( trick.winnerplayer() )
       < trick.cardno_of_player( card.player() )
       || !trick.jabs_cards_before(  card.player() ))
      modi -= 18; // previously 9, 12, 16
    // 18 261053
  }

  if (   card == Card::spade_queen ) {
    if( vgi.no() != trick.winnerplayer().no() ) {
      modi -= 3; // reference
    }
  }

  if ( is_solo(vgi.game().type())
      && Card::diamond_queen == card )
    modi += 1; // reference 126305

  if (   card == Card::club_queen ) {
    if( vgi.no() != trick.winnerplayer().no() ) {
      modi -= 5; // reference 156330
    }

    if( vgi.game().teaminfo().get(ai) != ai.team() ) {
      modi += 3; // previous 1
    } else {
      modi -= 4; // previous 2, 3
    }

    // show your own team.
    if (    vgi.game().teaminfo().get(ai) != ai.team()
        && ai.hand().count_trumps() <= 3 + (ai.hand().has_swines()?0:ai.hand().count_trump_aces())
        && vgi.game().tricks().remaining_no() - 1 > ai.hand().count_trumps()
        && (trick.cardno_of_player(trick.winnerplayer())
            > trick.cardno_of_player(card.player())
            && trick.jabs_cards_before( card.player() ))
       )
    {
      modi += 4; // previous
      if (  vgi.game().cards().count_played(Card::club_queen) == 0
          && vgi.game().announcements().last(ai.team()).announcement == Announcement::noannouncement
         )
        modi += 12; // previous
    }

    if(    trick.cardno_of_player(ai_virt) == 3
       && trick.winnerplayer().team() != vgi.team() )
    {
      modi -= 3; // reference 234467
    }

    if(    trick.cardno_of_player(ai_virt) == 0 )
      modi -= 12; // reference 168423
  }


#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    COUT <<"MQ"<< modi << "\t";
  }
#endif
  return modi;
}

/** Modification for a jack
 **/
int
TrickWeighting::modi_jack(VirtualGamesInterface const& vgi,
                          Trick const& trick,
                          Team const team,
                          HandCard const& card,
                          Team const winnerteam,
                          Ai const& ai)
{
  int modi = 0;

  if (!card.istrump()) {
    return modi;
  }

  if( card.value() != Card::jack )
    return modi;

  if (  trick.cardno_of_player(card.player()) == 3 )
    modi -= 2;

  if (  trick.cardno_of_player(card.player()) == 1 )
    modi += 14; // reference 028089

  if (  trick.cardno_of_player(card.player()) == 0 )
    modi += 22; //reference 114558

  if( trick.startcard().istrump()
     && trick.cardno_of_player(card.player()) !=0 )
    modi += 5; //reference 099840


  // force enemy with a jack to play a queen
  if (   vgi.no() != trick.winnerplayer().no()
      && (trick.cardno_of_player(trick.winnerplayer())
          > trick.cardno_of_player(card.player())
          && trick.jabs_cards_before( card.player() ) )
      && (trick.winnercard().value() == Card::queen)
      && (ai.team() != winnerteam)
      && trick.winnercard().istrump() )
    modi += 1;

  Trick t( trick.startplayer());
  for(unsigned i = 0; i == trick.cardno_of_player(card.player()); i++ )
  {
    t.add(trick.card( i ));
  }

  if (   vgi.no() != trick.winnerplayer().no() ) // playing jack and not winning trick
  {
    modi -= 6; // reference 133381

  }

  // save charlie in endgame for extra point
  if(    card == Card::charlie
     && vgi.game().rule()(Rule::Type::extrapoint_charlie)
     && vgi.game().tricks().remaining_no() -1 == ai.hand().count_trumps())
    modi -= 23; // reference 71916

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MJ"<< modi << "\t";

  }
#endif
  return modi;
}

/** modification for a fox
 **/
int TrickWeighting::modi_fox(VirtualGamesInterface const& vgi,
                             Trick const& trick,
                             Team const team,
                             HandCard const& card,
                             Team const winnerteam,
                             Ai const& ai)
{
  int modi = 0;

  Player const& ai_virt = trick.game().player(ai.no());

  // calculations for fox in trick
  if (   vgi.no() != trick.winnerplayer().no()
      && (trick.cardno_of_player(trick.winnerplayer())
          > trick.cardno_of_player(card.player())
          && trick.jabs_cards_before( card.player() ) )
      && ai.value(Aiconfig::Type::limithigh).less(trick.winnercard())
      && ai.team() != winnerteam )
  {
    for (unsigned c = 0; c < trick.actcardno(); c++)
    {
      if (trick.card(c).isfox() )
        modi += 11;
    }

  }

  if ((  (vgi.no() != trick.startplayer().no())
       || card.istrump())
     ) {
    for (unsigned c = 0; c < trick.actcardno(); c++) {
      // test, whether the fox has been brought home

      if (trick.card(c).isfox() )
      {

        if( (vgi.game().type() != GameType::poverty)
           || !vgi.is_soloplayer())
        {
          unsigned fact = 25;
          if(   ai.value(Aiconfig::Type::hands_known)
             || ai.value(Aiconfig::Type::trusting) )
          {
            fact += 10;
            if( card.isfox() )
              fact += 5;
          }
          if( !card.isfox() )
            fact += 5;

          bool allmyteam = true;

          // take a look if all players coming in this trick are of mine team
          for( unsigned i = trick.cardno_of_player(ai_virt); i < vgi.game().playerno(); i++ )
            allmyteam = allmyteam && ( ai.team() == ai.teaminfo( trick.player_of_card( i ) ) );

          if (    card.isfox()
              && !trick.startcard().istrump()
              && ai.color_runs( trick.startcard().color() ) > 0
              && !allmyteam
              && oppositeTeamCanWinTrick( trick, vgi ))
          {
            fact -= 21;
          }



          modi += fact * ((team == winnerteam)
                          ? +1
                          : -1);

        } else { // fox from above neutralized.
          if ( card.isfox() )
            modi += 30 * ((team != winnerteam)
                          ? +1
                          : 0);
        }
      }
    } // for (c < trick.actcardno())
  } // if (  (vgi.no() != trick.startplayer().no()) || card.istrump())


  // own fox
  if( !card.isfox() )
  {
#ifdef PLAYERCHECK
    if (vgi.no() == playerCheck) {

      COUT <<"MFA"<< modi << "\t";

    }
#endif
    return modi;
  }

  // to play a fox it must be worth it
  {
    modi -= 12; // previous 7, 16 (ref 208166)
  }

  if (  (vgi.no() != trick.startplayer().no())) {

    if (maybe_to_team(winnerteam) == team)
    {
      if (  winnerteam == team
          && ai.value(Aiconfig::Type::limithigh).less(trick.winnercard())
         )
        modi += 30; // fox home

    }
  }

  // bad style to play fox as first card
  if( trick.startplayer().no() == vgi.no() )
  {
    modi -= 90; // previous
  }

  if( trick.cardno_of_player(card.player()) == 3
     && maybe_to_team(winnerteam) != team )
  {
    modi -= 6; // previous

    if( ai.hand().count_trumps() <= 3 )
      modi += 2; // reference
  }

  if( trick.cardno_of_player(card.player()) == 3
     && winnerteam != team )
  {// reference 93822
    modi -= 35; // previous

  }

  if( trick.cardno_of_player(card.player()) == 3
     && winnerteam == team )
  {
    modi += 5; // previous
  }

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MFB"<< modi << "\t";

  }
#endif

  return modi;
}

/** Modification for a meatless solo
 **/
int
TrickWeighting::modi_soli_meat_less(VirtualGamesInterface const& vgi,
                                    Trick const& trick,
                                    Team const team,
                                    HandCard const& card,
                                    Team const winnerteam,
                                    Ai const& ai)
{
  int modi = 0;

  if ( vgi.game().type() != GameType::solo_meatless)
    return modi;
  if (   ai.no() != vgi.game().players().soloplayer().no()
      && vgi.no() != trick.winnerplayer().no()
      && ai.hand().count(card.color(), vgi.game()) > 0
      && ai.hand().count(card.color(), vgi.game()) < 3
      && (   ai.hand().count(Card(card.color(),Card::ace)) > 0
          || ai.hand().count(Card(card.color(),Card::ten)) > 0
          || card.value() == Card::ace || card.value() == Card::ten )
     )
  { // prevent non soloplayer from blank playing of cards
    modi -= 12;
  }

  if (   ai.no() != vgi.game().players().soloplayer().no()
      && vgi.no() != trick.winnerplayer().no()
      && (card.value() == Card::ace || card.value() == Card::ten) )
    modi -=3;  // reference 004655


#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MSM"<< modi << "\t";

  }
#endif

  return modi;
}

/** Modification for a marriage
 **/
int
TrickWeighting::modi_marriage(VirtualGamesInterface const& vgi,
                              Trick const& trick,
                              Team const team,
                              HandCard const& card, 
                              Team const winnerteam,
                              Ai const& ai)
{
  if (vgi.game().type() != GameType::marriage)
    return 0;

  int modi = 0;

  if (   (vgi.game().marriage().selector() != MarriageSelector::team_set)
      && (card.value() >= 10) )
    modi -= 9;

  if (trick.cardno_of_player(card.player()) == 0) {
    switch (vgi.game().marriage().selector()) {
    case MarriageSelector::silent:
    case MarriageSelector::team_set:
      // the marriage is determined
      break;
    case MarriageSelector::first_color:
    case MarriageSelector::first_foreign:
      if (!card.istrump()) {
        modi -= 5;
      }
      break;
    case MarriageSelector::first_trump:
      if (card.istrump()) {
        modi -= 5;
      }
      break;
    case MarriageSelector::first_club:
      if (   !card.istrump()
          && card.color() == Card::club ) {
        modi -= 5;
      }
      break;
    case MarriageSelector::first_spade:
      if (   !card.istrump()
          && card.color() == Card::spade) {
        modi -= 5;
      }
      break;
    case MarriageSelector::first_heart:
      if ( !card.istrump()
          && card.color() == Card::heart ) {
        modi -= 5;
      }
      break;
    } // switch (this->game().marriage().selector())
  } // if !(marriage && !card is trickstartcard())

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    COUT <<"MM"<< modi << "\t";
  }
#endif
  return modi;
}

/** Modification for special points
 **/
int
TrickWeighting::modi_special_points(VirtualGamesInterface const& vgi,
                                    Trick const& trick,
                                    Team const team,
                                    HandCard const& card,
                                    Team const winnerteam,
                                    Ai const& ai)
{
  int modi = 0;

  Game const& virt_game = trick.game();

  // create the teams-vector
  vector<Team> teams;
  for (auto const& player : virt_game.players()) {
    // determine team of player from ai out of virtual game
    auto const t = ai.teamofplayer(player);
    if (t != Team::unknown)
      teams.push_back(maybe_to_team(t));
    else
      teams.push_back(opposite(maybe_to_team(ai.team())));
  } // for (player)

  if (  (vgi.no() != trick.startplayer().no())
      || card.istrump()) {

    if (maybe_to_team(winnerteam) == team) {
      auto const spv = trick.specialpoints(teams);
      modi += (30 * sum(spv, winnerteam, trick.startplayer().game()));
    } else {
      auto const spv = trick.specialpoints(teams);
      modi -= (30 * sum(spv, winnerteam, trick.startplayer().game()));
    }

  } // if (  (vgi.no() != trick.startplayer().no()) || card.istrump())
#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MSP"<< modi << "\t";

  }
#endif
  return modi;
}

/** Modification for no trump
 **/
int
TrickWeighting::modi_no_trump(VirtualGamesInterface const& vgi,
                              Trick const& trick,
                              Team const team,
                              HandCard const& card,
                              Team const winnerteam, 
                              Ai const& ai)
{
  int modi = 0;
  if (card.istrump () )
    return modi;

  Player const& ai_virt = trick.game().player(ai.no());


  if(  vgi.no() != trick.winnerplayer().no()
     && is_solo( vgi.game().type() )
     && ( trick.cardno_of_player( trick.winnerplayer() ) > trick.cardno_of_player( trick.game().players().soloplayer() ) )
     && !trick.card_of_player( trick.game().players().soloplayer() ).less( vgi.trump_card_limit() )
    )
  {
    modi -= card.value();
  }

  if( is_solo( vgi.game().type() )
     && trick.cardno_of_player( trick.game().players().soloplayer() ) == 0 )
  {
    modi += card.value();
    if ( card.value() >= 10 )
      modi += 3;
  }

  if( is_solo( vgi.game().type() )
     && trick.cardno_of_player(ai_virt) == 0 )
  {
    // last card with small value to serve a trick
    if(   ai.hand().count( card.color() ) == 0
       && card.value() < 10 )
      modi += 34; // reference 64029
  }

  // game nears it end, no real sense to keep non trump carts any longer, better to hold trumps
  if (    trick.cardno_of_player(ai_virt) == 0
      && vgi.game().tricks().remaining_no() < 4 )
  {
    modi += 23; // reference 058753

  }

  if (    trick.cardno_of_player(ai_virt) == 0
      && vgi.game().tricks().remaining_no() < 5 )
  {

    if ( vgi.team() == Team::re
        &&  vgi.game().type() != GameType::marriage
        && ai.cards_information().played(Card::club_queen)+ai.hand().count_club_queens()<2)
      modi += 3;  // reference 043205
  }

  if(  vgi.no() != trick.winnerplayer().no()
     && is_solo( vgi.game().type() )
     && ( trick.cardno_of_player( trick.winnerplayer() ) > trick.cardno_of_player( trick.game().players().soloplayer() ) )
     && trick.card_of_player( trick.game().players().soloplayer() ).less( vgi.trump_card_limit() )
    )
    modi += card.value();



  // prefer color cards a little bit
  if(    card.value() == Card::nine
     && trick.winnerplayer().team() != vgi.team() )
  {
    modi += 8; // previous: 6
    if( trick.cardno_of_player(ai_virt) == 0 )
      modi +=10; // reference 203953
  }
  if(    card.value() == Card::king
     && trick.winnerplayer().team() != vgi.team() )
    modi += 8;

  if( vgi.game().announcements().last( trick.winnerplayer().team() ).announcement != Announcement::noannouncement
     && trick.winnerplayer().team() != vgi.team() )
  {
    // preserve trumps against announcement
    modi += 6;  //reference 006877
  }

  if ( card.value() == Card::ace )
  {

    modi -= 2 - vgi.color_runs( card.color() );
  }

  if( card.value() == Card::jack )
    modi +=1; // good choice if not trump
  // reference 112599

  if( is_solo( vgi.game().type() )
     && vgi.team() == Team::contra )
    modi -=  2* ai.hand().count( card.color() ); // reference  137807


  if (  vgi.no() != trick.winnerplayer().no()
      && winnerteam != team
      && trick.startcard().color() != card.color()
      && !trick.startcard().istrump() )
  {
    modi -= 9; //previous 6, 16, 14, 8
    // 9 : Reference 001236

    if(    trick.cardno_of_player(ai_virt) == 3
       && card.value() >= 10 ) // reference 234467
      modi -= 29;

  }

  if(    trick.cardno_of_player(ai_virt) == 0
     && card.value() == Card::nine
     && vgi.color_runs( trick.startcard().color() ) > 0)
  { modi+=7; } //reference 272170

  bool allmyteam=true;

  for( unsigned i = trick.cardno_of_player(ai_virt);
      i < ai.game().playerno();
      i++ )
    allmyteam=allmyteam && ( ai.team() == ::maybe_to_team( vgi.teamofplayer( trick.player_of_card(i) ) ) );

  if(  !trick.startcard().istrump()
     && vgi.color_runs( trick.startcard().color() ) > 0
     && card.color() != trick.startcard().color()
     && !allmyteam)
    modi -= card.value()+2; //+2 reference 133381

  if( (ai.hand().count_trumps() - 1 <= (vgi.game().tricks().remaining_no() + 1) / 2 ) )
  {
    // reference 61146
    modi += 6;
    //if fox is still on hand, protect fox.
    if( ai.hand().count_trump_aces() > 0 && ai.hand().count_swines() == 0 )
    {
      modi += 6; //reference 052672
      if( ai.hand().count_trumps() == 2 )
        modi += 15; // reference 076221
    }
  }

  if (  vgi.color_runs( card.color() ) == 0
      && card.value() == Card::ace
      && ai.team() == winnerteam
     )
  {
    modi += 2;

  }

  if( is_solo( vgi.game().type() )
     && trick.cardno_of_player(trick.game().players().soloplayer()) > trick.cardno_of_player(card.player())
     && vgi.color_runs( card.color() ) == 0
     && card.value() == Card::ace
    )
  {
    modi += 22; // reference 9751
  }

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MNT"<< modi << "\t";

  }
#endif
  return modi;
}

/** Modification for a picture solo
 **/
int
TrickWeighting::modi_picture_soli(VirtualGamesInterface const& vgi,
                                  Trick const& trick,
                                  Team const team,
                                  HandCard const& card,
                                  Team const winnerteam,
                                  Ai const& ai)
{
  int modi = 0;

  if(  (   vgi.game().type() == GameType::solo_king
        || vgi.game().type() == GameType::solo_king_jack
        || vgi.game().type() == GameType::solo_king_queen
        || vgi.game().type() == GameType::solo_koehler)
     && card.value() == Card::king  )
  {
    if( vgi.no() != trick.winnerplayer().no() )
    {
      modi -= (36 - std::max(static_cast<int>(ai.game().cards().count_played()) - 13, 0));
      modi += trick.points()/2;

      if( trick.cardno_of_player( trick.winnerplayer() )
         < trick.cardno_of_player( card.player() )
         || !trick.jabs_cards_before( card.player() ) )
        modi -= 4;
    } else
      if(     trick.cardno_of_player( card.player() ) == 3
         && trick.points() < 16 )  // reference 131296
      {
        modi -= 25;
      }
  }

  if( card.istrump()
     && (   vgi.no() != trick.winnerplayer().no()
         || trick.winnercardno() != 3)
     && trick.cardno_of_player(card.player()) == 3 ) {
    if (   vgi.game().type() != GameType::solo_king
        && vgi.game().type() != GameType::solo_queen_jack
        && vgi.game().type() != GameType::solo_king_jack
        && vgi.game().type() != GameType::solo_king_queen
        && vgi.game().type() != GameType::solo_koehler
        && (   card.value() == Card::jack
            || card.value() == Card::queen)
       ) { // a king is better, but has two points more
      modi -= 3;
    }
  }

  if(  (   vgi.game().type()==GameType::solo_king
        || vgi.game().type()==GameType::solo_king_jack
        || vgi.game().type()==GameType::solo_king_queen
        || vgi.game().type()==GameType::solo_koehler )
     && card.value() == Card::king )
    modi -= 2;


  if(  card.istrump()
     && (    vgi.game().type()==GameType::solo_king
         ||  vgi.game().type()==GameType::solo_queen
         ||  vgi.game().type()==GameType::solo_jack
        )
    )
    modi -= 4;

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MPS"<< modi << "\t";

  }
#endif
  return modi;
}

/** Modification for other trump
 **/
int
TrickWeighting::modi_other_trumps(VirtualGamesInterface const& vgi,
                                  Trick const& trick,
                                  Team const team,
                                  HandCard const& card, 
                                  Team const winnerteam,
                                  Ai const& ai)
{
  int modi = 0;
  Game const& virt_game = trick.game();
  Player const& ai_virt = virt_game.player(ai.no());

  if( !card.istrump () )
    return modi;

  if (   trick.startcard().istrump()
      // ToDo: solo
      && !card.less(Card::diamond_queen)
      && (static_cast<int>(trick.points()) >= ai.value(Aiconfig::Type::limitqueen)) ) {
    modi += 2;
  }

  if (is_solo(vgi.game().type())) {
    modi -= 2;
    if  (   vgi.is_soloplayer()
         && card.value() == 0) {
      modi -= 12;
    }
  }

  if(  (   vgi.no() != trick.winnerplayer().no()
        || trick.winnercardno() != 3)
     && trick.cardno_of_player(card.player()) == 3 ) {
    modi -= 5; // previous -2
  }


  // have played a high card
  if (   (   ai.value(Aiconfig::Type::limithigh).less(card)
          || ai.value(Aiconfig::Type::limithigh) == card)
      && (static_cast<int>(trick.points()) - card.value()
          < ai.value(Aiconfig::Type::limitqueen) ) )
  {
    modi -= 3;
    if ((2 * ai.hand().count_ge(ai.value(Aiconfig::Type::limithigh))
         < ai.hand().count_trumps())
        && (ai.hand().count_ge(ai.value(Aiconfig::Type::limithigh))
            < 3))
      modi -= 13; // previous: 18
  }

  if (   !trick.startcard().istrump()
      && (trick.cardno_of_player(card.player()) < 3)
      && card.less(ai.value(Aiconfig::Type::limithigh))
      && (static_cast<int>(trick.points()) >= ai.value(Aiconfig::Type::limitqueen))
      && trick.jabs_cards_before(card.player()) ) {
    modi -= 3;
    if ( vgi.color_runs( trick.startcard().color() ) > 0 )
      modi -= 6;

    if (   is_solo(vgi.game().type())
        && trick.cardno_of_player(trick.game().players().soloplayer()) > trick.cardno_of_player(card.player()) )
      modi -= 7; // reference 197069
  }

  if(   (card.value() == Card::ten)
     && (!card.isdolle()) ) {
    modi -= 10; // reference 246985
  }

  if(   (card.value() == Card::ace)
     && !card.is_special() ) {
    modi -= 9; // reference 246985
  }

  if(   (card.value() == Card::king)
     &&  (winnerteam != team) ) {
    modi -= 14; // reference 256027
  }

  if(   (card.value() == Card::nine)
     && !card.is_special()
     && trick.winnercardno() == 3
     && trick.cardno_of_player(card.player()) == 3 )
    modi -= 3; // reference 064027_1


  if ( virt_game.cards().next_higher_card(card).value() != card.value() //2do nect_higher_card_still_in_game compar reference 260689
     && (trick.cardno_of_player(trick.winnerplayer())
         > trick.cardno_of_player(card.player()))
     && trick.jabs_cards_before( card.player() ))
  { // a higher card in value is needed to jab my card
    // for example a queen to jab my jack
    modi += 1;
  }


  bool allmyteam=true;

  for( unsigned i = trick.cardno_of_player(ai_virt);
      i < ai.game().playerno();
      i++ )
    allmyteam=allmyteam && ( ai.team() == ::maybe_to_team( vgi.teamofplayer( trick.player_of_card(i) ) ) );

  if(   card.less( ai.value(Aiconfig::Type::limithigh) )
     && !trick.startcard().istrump()
     && (trick.cardno_of_player(trick.winnerplayer())
         > trick.cardno_of_player(card.player())
         && trick.jabs_cards_before( card.player() ) )
     && !(    trick.islastcard()
          || allmyteam )
    ) {
    modi -= card.value() * std::min(vgi.color_runs( trick.startcard().color() ),
                                    2u);
  }

  if (  (vgi.no() == trick.startplayer().no()))
  {
    modi -= 2;

    if( ai.hand().count_trumps() == ai.hand().cardsnumber()
       && card.value() < 3 )
      modi +=5; // Reference 012664

    if( (card.value() == Card::king)
       && (winnerteam == team) )
      modi += 8;
  }

  { // the player has overjabbed his partner without need
    if (trick.winnerplayer() == ai) {
      Trick const t2 = trick.before_player(ai);
      if (  (::maybe_to_team(t2.winnerplayer().team()) == ai.team())
          && (vgi.trump_card_limit().less(t2.winnercard()))
         ) {
        modi -= 8;
      }
    }
  } // the player has overjabbed his partner without need

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    COUT <<"MOT"<< modi << "\t";
  }
#endif
  return modi;
}

/** Modification for a forced trump
 **/
int
TrickWeighting::modi_force_trump(VirtualGamesInterface const& vgi,
                                 Trick const& trick,
                                 Team const team,
                                 HandCard const& card, 
                                 Team const winnerteam,
                                 Ai const& ai)
{
  int modi = 0;

  // force enemy to play high card
  if (   vgi.no() != trick.winnerplayer().no()
      && (trick.cardno_of_player(trick.winnerplayer())
          > trick.cardno_of_player(card.player())
          && trick.jabs_cards_before( card.player() ) )
      && ai.value(Aiconfig::Type::limithigh).less(trick.winnercard())
      && ai.team() != winnerteam )
  {

    bool bIsForceTrump = true;
    // ToDo: start with
    for (unsigned c = trick.cardno_of_player(card.player()) + 1; c < trick.actcardno(); c++)
    {
      if (!trick.card(c).less(card))
        bIsForceTrump = false;
    } // for(c)
    if ( bIsForceTrump )
      modi += 27;
  }

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT <<"MFT"<< modi << "\t";

  }
#endif
  return modi;
}

/** Modification for a poverty
 **/
int
TrickWeighting::modi_poverty(VirtualGamesInterface const& vgi,
                             Trick const& trick,
                             Team const team,
                             HandCard const& card,
                             Team const winnerteam,
                             Ai const& ai)
{
  int modi = 0;
  if ( vgi.game().type() != GameType::poverty )
    return modi;
  Player const& ai_virt = trick.game().player(ai.no());

  if(    card.istrump()
     && trick.cardno_of_player(ai_virt) == 0)
    modi -= 3; // 1:reference 208016, 3: reference 194579

  //  get large cards of poverty player
  if((  //only for not poverty-player
      vgi.no() == vgi.game().players().poverty_partner().no() ))
  {

    if( (trick.cardno_of_player(trick.game().players().soloplayer()) > trick.cardno_of_player(card.player())) //
       && !card.less(ai.value(Aiconfig::Type::limithigh))
       &&  card.player().no() != vgi.game().players().previous(vgi.game().players().soloplayer()).no())
    {
      modi += 21;
      if( ai.value(Aiconfig::Type::limithigh).less(trick.winnercard()) )
        modi += 5;
      if ( trick.winnercard().isdolle()
          || trick.winnercard().isswine()
          || trick.winnercard().ishyperswine() )
        modi +=4;
    }
  }
#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    // 4. block
    COUT << "MPV" << modi << "\t";
  }
#endif

  return modi;
}

/** -> result
 **
 ** @param    vgi   the virtual games interface
 ** @param    trick   the trick to be examined
 ** @param    t   the team which should make points
 ** @param    card    the card played by the player
 **
 ** @return   the weighting of the trick for the team 'team'
 **
 ** @todo       perhaps points for getting the fox home
 ** @todo       perhaps when playing the highest card, subtract in modi
 **   the distance to the next lower card
 **/
int
TrickWeighting::modi(VirtualGamesInterface const& vgi,
                     Trick const& trick,
                     Team const t,
                     HandCard const& card)
{
  DEBUG_ASSERTION(trick.isfull(),
                  "TrickWeighting::modi()\n"
                  "  trick is not full");
  Game const& virt_game = trick.game();
  Ai const& ai = (dynamic_cast<Ai const*>(&virt_game.player(vgi.no()))
                  ? dynamic_cast<Ai const&>(virt_game.player( vgi.no() ))
                  : dynamic_cast<Ai const&>(vgi.game().player(vgi.no()))
                 );
  Player const& ai_virt = virt_game.player(ai.no());
  //Team const winnerteam = ai.teamofplayer(trick.winnerplayer());
  Team winnerteam
    = ::maybe_to_team(vgi.teamofplayer(trick.winnerplayer()));
  if (winnerteam == Team::unknown)
    winnerteam = ::opposite(t);

  if (virt_game.marriage().is_undetermined()) {
    // in an undetermined marriage all players shall play for themselves
    if (t == Team::re)
      winnerteam = ::maybe_to_team(ai.teaminfo(trick.winnerplayer()));
    else if (trick.winnerplayer().no() == vgi.no())
      winnerteam = ::maybe_to_team(t);
    else
      winnerteam = ::opposite(::maybe_to_team(t));
  }


  Team const team = ((virt_game.type() == GameType::marriage) 
                     ? maybe_to_team(t)
                     : t);

  int modi = 0;


  if (winnerteam == team)
    modi += trick.points();
  else
    modi -= trick.points();

#ifdef PLAYERCHECK
  // Playernumber for which to produce output
  playerCheck = PLAYERCHECK;

  // so in virtual games no informatin is shown
  if (vgi.game().isvirtual())
    playerCheck = 10;
#endif




#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    //COUT << trick.game().player(0).hand();
    COUT << '\n';
    COUT << trick << endl;  
    COUT << team << "\t" << trick.winnerplayer().no() << " " << winnerteam << std::endl;
    for( unsigned i = 0; i < ai.game().playerno(); i++ )
      COUT << virt_game.player( i ).team() << "\t";
    COUT << std::endl;
    // 1. block
    COUT << card << "\t";
  }
#endif

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {

    COUT << "START" << modi << "\t";

  }
#endif

  modi += modi_dolle_swines( vgi, trick, t, card, winnerteam, ai );

  modi += modi_queen( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_jack( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_fox( vgi, trick, t, card, winnerteam, ai );

  modi += modi_soli_meat_less( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_marriage( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_special_points( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_no_trump( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_picture_soli( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_other_trumps( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_force_trump( vgi, trick, t, card, winnerteam, ai  );

  modi += modi_poverty( vgi, trick, t, card, winnerteam, ai  );


  bool allmyteam=true;

  for (auto i = trick.cardno_of_player(ai_virt);
       i < ai.game().playerno();
       i++ )
    allmyteam = allmyteam && ( ai.team() == ::maybe_to_team( vgi.teamofplayer( trick.player_of_card(i) ) ) );

  /// @todo jabbed by player of other team behind me
  if(   !trick.startcard().istrump()
     && vgi.jabbedbyotherteam( trick.startcard().color() )
     && card.value() >= 10 
     && !allmyteam
     && oppositeTeamCanWinTrick( trick, vgi )
    ) {
    modi -= 4;
#ifdef PLAYERCHECK
    if (vgi.no() == playerCheck) {

      COUT << "XX " << modi << "\t";
    }
#endif
  }

#ifdef PLAYERCHECK
  if (vgi.no() == playerCheck) {
    COUT << "END" << modi << endl;
  }
#endif

  return modi;
} // int WVirtualGames::modi(Trick const& trick, Team const& team, Card const& card)


/** -> result
 **
 ** @param    vgi         virtual game interface
 ** @param    c           the card played by the player
 ** @param    virt_game   actual virtual game
 **
 ** @return   weighting modifier for getting own team in backhand
 **/
int
TrickWeighting::backhand(VirtualGamesInterface const& vgi,
                         HandCard const& c,
                         Game const& virt_game )                         
{
  int result = 0;

  // add some points, if the own team is in the back,
  if (virt_game.players().previous( virt_game.tricks().current().winnerplayer() ).team()
      == vgi.team() && !c.isfox() &&
      vgi.game().type() != GameType::solo_meatless )
  {
    result = 5;

    if( virt_game.tricks().current().points() > 10 )
      result -= 3;  // Bugreport reference 1236

    if (virt_game.type() == GameType::poverty )
      result -= 3;
  }
  else
    result = -6;


  return result;
}

