/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "ai_type.h"
#include "aiconfig_type.h"
#include "aiconfig_heuristic.h"
#include "../card/card.h"
#include "heuristicsmap.h"
#include "rating/rating.h"


#include <map>

class Aiconfig {
  public:
    // whether to write the standard type (all the whole configuration)
    static bool write_standard_type;

  public:
    using Difficulty = AiconfigDifficulty;
    using Type = AiconfigType;

    using Heuristic = AiconfigHeuristic;
    static Heuristic Heuristic_from_name(string const& name) 
    { return AiconfigHeuristic_from_name(name); }

    struct HeuristicState {
      HeuristicState() = delete;
      HeuristicState(Heuristic heuristic, bool active) :
        heuristic(heuristic), active(active) { }

      Heuristic heuristic;
      bool active;
    }; // struct HeuristicState

  public:
    // returns a preset ai
    static Aiconfig const& preset(Difficulty difficulty);
    // returns all (interesting) keys for the heuristic configuration
    static vector<HeuristicsMap::Key> const& keys();

  public:
    friend class Ai;

  public:
    Aiconfig();
    Aiconfig(Difficulty difficulty);
    Aiconfig(istream& istr);
    Aiconfig(istream& istr, Difficulty difficulty);
    Aiconfig(Aiconfig const& aiconfig);
    Aiconfig& operator=(Aiconfig const& aiconfig);
    virtual ~Aiconfig();

    // set the aiconfig to the default values
    void reset_to_hardcoded(unsigned no = UINT_MAX);

    // (re)sets the aiconfig to the given difficulty
    void set_to_difficulty(Difficulty difficulty);
    // the rule has changed
    void rule_changed(int type, void const* old_value);

    Difficulty difficulty() const;
  private:
    // update the difficulty
    void update_difficulty();
  public:
    // whether the aiconfig is in a (nested) update
    bool in_update() const;
    // enter a (nested) update
    void enter_update();
    // leave a (nested) update
    void leave_update();
    // leave a (nested) update and set the difficulty manual
    void leave_update(Difficulty difficulty);


    // whether the aiconfig is equal
    bool equal(Aiconfig const& aiconfig) const;

    AiType const& aitype(unsigned trickno) const;
    void set_aitype(unsigned trickno, AiType const& aitype);
    void set_aitype_for_all_tricks(AiType const& aitype);
    unsigned future_limit(unsigned trickno) const;
    void set_future_limit(unsigned trickno, unsigned future_limit);
    void set_future_limit_for_all_tricks(unsigned future_limit);
    Rating::Type const& rating(unsigned trickno) const;
    void set_rating(unsigned trickno, Rating::Type rating);
    void set_rating_for_all_tricks(Rating::Type rating);

    // returns the heuristic states
    vector<HeuristicState> const&
      heuristic_states(HeuristicsMap::Key key) const;
    vector<HeuristicState> const&
      heuristic_states(HeuristicsMap::GameTypeGroup gametype_group,
                       HeuristicsMap::PlayerTypeGroup playertype_group
                      ) const;
  private:
    vector<HeuristicState>&
      heuristic_states(HeuristicsMap::Key key);
    vector<HeuristicState>&
      heuristic_states(HeuristicsMap::GameTypeGroup gametype_group,
                       HeuristicsMap::PlayerTypeGroup playertype_group
                      );
  public:
    // returns the heuristics
    vector<Heuristic> heuristics(HeuristicsMap::GameTypeGroup const
                                 gametype_group,
                                 HeuristicsMap::PlayerTypeGroup const
                                 playertype_group
                                ) const;
    // returns the heuristics
    vector<Heuristic> heuristics(Player const& player) const;

    // returns the value
    bool value(HeuristicsMap::Key key,
               Heuristic heuristic) const;
    bool value(HeuristicsMap::GameTypeGroup gametype_group,
               HeuristicsMap::PlayerTypeGroup playertype_group,
               Heuristic heuristic) const;
    bool value(Type::Bool type) const;
    int value(Type::Int type) const;
    Card value(Type::Card type) const;

    // the minimum and maximum value
    int min(Type::Int const& type) const;
    int max(Type::Int const& type) const;

    // the valid cards for the type
    vector<Card> const& valid_cards(Type::Card type) const;

    // sets the value
    bool set(string const& type, string const& value);

    void set(HeuristicsMap::GameTypeGroup gametype_group,
             HeuristicsMap::PlayerTypeGroup playertype_group,
             Heuristic heuristic,
             string const& value);
    void move(HeuristicsMap::Key key,
              Heuristic heuristic,
              unsigned pos);
    void set(Heuristic heuristic, string const& value);
    void set(Type::Bool type, string const& value);
    void set(Type::Int type, string const& value);
    void set(Type::Card type, string const& value);

    void set(HeuristicsMap::Key key,
             Heuristic heuristic,
             bool value);
    void set(HeuristicsMap::GameTypeGroup gametype_group,
             HeuristicsMap::PlayerTypeGroup playertype_group,
             Heuristic heuristic,
             bool value);
    void set_to_default(HeuristicsMap::Key key);
    void set(Heuristic heuristic, bool value);
    void set(Type::Bool type, bool value);
    void set(Type::Int type, int value);
    void set(Type::Card type, Card const& value);

    bool load(string const& filename);
    bool read(istream& istr);
    bool read_heuristics(istream& istr);
    bool save(string const& filename) const;

    // output of the settings
    void write(ostream& ostr) const;

  private:
    // initializes the heuristic states
    void init_heuristic_states();
    // fills the heuristic states with all heuristics
    void fill_up_heuristic_states();

  private:
    Difficulty difficulty_ = Difficulty::standard_defensive;

    vector<AiType> aitype_p;
    vector<Rating::Type> rating_p;
    vector<unsigned> future_limit_p;

    map<HeuristicsMap::Key, vector<HeuristicState> > heuristic_states_;

    vector<bool> bool_;
    vector<int> int_;
    vector<Card> card_;

    // the level of the internal update
    unsigned update_level_;

}; // class Aiconfig

namespace HeuristicsMap {
typedef std::map<Key, vector<Aiconfig::Heuristic> > Map;
}; // namespace HeuristicsMap

// whether the aitype has support for rating
bool supports_rating(AiType type);

// whether the two aiconfigs are equal
bool operator==(Aiconfig const& lhs, Aiconfig const& rhs);
// whether the two aiconfigs are different
bool operator!=(Aiconfig const& lhs, Aiconfig const& rhs);

// whether the two heuristic states are equal
bool operator==(Aiconfig::HeuristicState lhs,
                Aiconfig::HeuristicState rhs);

// output of the Aiconfig
ostream& operator<<(ostream& ostr, Aiconfig const& aiconfig);
