/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"

#include "playersDb.h"
#include "../game/game.h"
#include "../game/game_summary.h"

#include <sstream>
using std::istringstream;

ostream& operator<<(ostream& ostr, TLWcount const& tlwcount)
{ return tlwcount.write(ostr); }
istream& operator>>(istream& istr, TLWcount& tlwcount)
{ return tlwcount.read(istr); }

ostream& operator<<(ostream& ostr, Ranking const& ranking)
{ return ranking.write(ostr); }
istream& operator>>(istream& istr, Ranking& ranking)
{ return ranking.read(istr); }


ostream& operator<<(ostream& ostr, PlayersDb const& players_db)
{ return players_db.write(ostr); }
istream& operator>>(istream& istr, PlayersDb& players_db)
{ return players_db.read(istr); }

/** constructor from a stream
 **
 ** @param    istr   stream with the database info
 **/
PlayersDb::PlayersDb(istream& istr)
{
  this->read(istr);
} // PlayersDb::PlayersDb(istream& istr)

/** writes the database to 'ostr'
 **
 ** @param    ostr   output stream
 **
 ** @return   output stream
 **/
ostream&
PlayersDb::write(ostream& ostr) const
{
  ostr << setw(24) << "ranking" << " =\t"
    << this->rank << '\n';
  ostr << setw(24) << "total games" << " =\t"
    << this->games_all() << '\n';
  ostr << setw(24) << "total special points" << " =\t"
    << this->specialpoints_all() << '\n';

  ostr << setw(24) << "average game points" << " =\t"
    << this->averageGamePoints() << '\n';
  ostr << setw(24) << "average game trick points" << " =\t"
    << this->averageGameTrickPoints() << '\n';

  ostr << '\n';

  ostr << "games\n"
    << "{\n";
  for (auto gt : game_type_list)
    ostr << setw(32) << to_string(gt) << " =\t"
      << this->games(gt) << '\n';
  ostr <<"}\n";

  ostr << '\n';

  ostr << "specialpoints\n"
    << "{\n";
  for (auto const sp : Specialpoint::types)
    ostr << setw(32) << to_string(sp) << " =\t"
      << this->specialpoints(sp) << '\n';
  ostr <<"}\n";

  return ostr;
} // ostream& PlayersDb::write(ostream& ostr) const

/** reads the database from 'istr'
 **
 ** @param    istr   stream with the database info
 **
 ** @return   input stream
 **/
istream&
PlayersDb::read(istream& istr)
{
  // first clear all
  this->averageGamePoints_ = 0;
  this->averageGameTrickPoints_ = 0;
  this->games_all_ = TLWcount();
  this->games_.clear();
  this->specialpoints_all_ = TLWcount();
  this->specialpoints_.clear();

  unsigned depth = 0;
  while (istr.good()) {
    Config config;
    istr >> config;
    if (istr.fail() || istr.eof())
      break;

    istringstream istr_value(config.value);
    if (!config.separator
        && (config.name == "{")) {
      depth += 1;
    } else if (!config.separator
               && (config.name == "}")) {
      if (depth == 0) {
        cerr << "players database: "
          << "found a '}' without a '{' before.\n"
          << "Finish reading the the file."
          << endl;
        break;
      } // if (depth == 0)
      depth -= 1;
      if (depth == 0)
        break;
    } else if (config.name == "ranking")
      this->rank.read(istr_value);
    else if (config.name == "total games")
      this->games_all_.read(istr_value);
    else if (config.name == "total special points")
      this->specialpoints_all_.read(istr_value);
    else if (config.name == "average game points")
      istr_value >> this->averageGamePoints_;
    else if (config.name == "average game trick points")
      istr_value >> this->averageGameTrickPoints_;
    else if (!config.separator
             && this->read_group(config.name, istr)) // NOLINT
      // this complex reading is necessary to easily add groups in child
      // classes
      ;
    else {
      cerr << "Reading the player database:\n"
        << "found following unknown line:\n"
        << (config.separator
            ? (config.name + " = " + config.value)
            : config.name + config.value)
        << '\n'
        << "ignoring it."
        << endl;
      SEGFAULT;
      exit(EXIT_FAILURE);
    }
  }; // while (istr.good())

  return istr;
} // istream& PlayersDb::read(istream& istr)

/** reads the group 'group' from 'istr'
 **
 ** @param    group   name of the group to read
 ** @param    istr   stream with the group info
 **
 ** @return   whether the group 'group' is known
 **/
bool
PlayersDb::read_group(string const& group, istream& istr)
{
  if (group == "games")
    this->read_group_games(istr);
  else if (group == "specialpoints")
    this->read_group_specialpoints(istr);
  else
    return false;

  return true;
} // bool PlayersDb::read_group(string const& group, istream& istr)

/** reads the games group from 'istr'
 **
 ** @param    istr   stream with the games info
 **/
void
PlayersDb::read_group_games(istream& istr)
{
  unsigned depth = 0;

  // load the configuration
  while (istr.good()) {
    Config config;
    istr >> config;

    if (config.separator) {
      istringstream istr_value(config.value);
      istr_value >> this->games_[game_type_from_string(config.name)];
    } else { // if (config.separator)
      if(config.name == "{") {
        depth += 1;
      } else if(config.name == "}") {
        if (depth == 0) {
          cerr << "players database: group 'games': "
            << "found a '}' without a '{' before.\n"
            << "Finish reading the the file."
            << endl;
          break;
        } // if (depth == 0)
        depth -= 1;
        if (depth == 0)
          break;
      } else if(config.name == "") {
        cerr << "players database: group 'games': "
          << "Ignoring line \'" << config.value << "\'.\n";
      } else {
        cerr << "players database: group 'games': "
          << "type '" << config.name << "' unknown.\n"
          << "Ignoring it.\n";
      } // if (config.name == .)
    } // config.separator
  } // while (istr.good())

  return ;
} // void PlayersDb::read_group_games(istream& istr)

/** reads the specialpoints group from 'istr'
 **
 ** @param    istr   stream with the specialpoints info
 **/
void
PlayersDb::read_group_specialpoints(istream& istr)
{
  unsigned depth = 0;

  // load the configuration
  while (istr.good()) {
    Config config;
    istr >> config;

    if (config.separator) {
      try {
        istringstream istr_value(config.value);
        istr_value >> this->specialpoints_[specialpoint_type_from_string(config.name)];
      } catch (...) {
        cerr << "players database: group 'specialpoints': "
          << "unknown type '" << config.name << "'."
          << endl;
      }
    } else { // if (config.separator)
      if(config.name == "{") {
        depth += 1;
      } else if(config.name == "}") {
        if (depth == 0) {
          cerr << "players database: group 'specialpoints': "
            << "found a '}' without a '{' before.\n"
            << "Finish reading the the file."
            << endl;
          break;
        } // if (depth == 0)
        depth -= 1;
        if (depth == 0)
          break;
      } else if(config.name == "") {
        cerr << "players database: group 'specialpoints': "
          << "Ignoring line \'" << config.value << "\'.\n";
      } else {
        cerr << "players database: group 'specialpoints': "
          << "type '" << config.name << "' unknown.\n"
          << "Ignoring it.\n";
      } // if (config.name == .)
    } // config.separator
  } // while (istr.good())
} // void PlayersDb::read_group_specialpoints(istream& istr)

/** adds the game to the db
 **
 ** @param    p   player
 ** @param    g   gamepoints
 **/
void
PlayersDb::evaluate(Player const& p, GameSummary const& g)
{
  if( g.winnerteam() == p.team() )
  {

    games_[g.game_type()].hasWon();
    rank.add( Ranking( 1.0 ) );
    averageGamePoints_ = (( averageGamePoints_ * games_all_.total() ) + g.points() )
      / ( games_all_.total() + 1 );
    games_all_.hasWon();
  } else
  {
    games_[g.game_type()].hasLost();
    rank.add( Ranking( -1.0 ) );
    games_all_.hasLost();
  }

  averageGameTrickPoints_ = 
    ( ( averageGameTrickPoints_ * games_all_.total() ) + g.points(p) ) 
    / (games_all_.total()+1);

  auto spv = g.specialpoints();  

  for(auto & i : spv) {
    if ((i.player_get_no == p.no() )
        || (i.player_get_no == UINT_MAX) )
      if( i.counting_team == p.team() ) {
        specialpoints_[i.type].hasWon();
        specialpoints_all_.hasWon();
      }
    if ((i.player_of_no == p.no() )
        || (i.player_of_no == UINT_MAX) )
      if( i.counting_team != p.team() ) {
        specialpoints_[i.type].hasLost();
        specialpoints_all_.hasLost();
      }
  }
} // virtual void PlayerDb::evaluate(Player p, GamePoints g )

/** clears the database
 **/
void
PlayersDb::clear()
{
  this->rank.clear();

  this->games_.clear();
  this->specialpoints_.clear();

  this->averageGamePoints_ = 0;
  this->averageGameTrickPoints_ = 0;

  this->games_all_.clear();
  this->specialpoints_all_.clear();
} // void PlayersDb::clear()

/** @return   a TLW for the games group 'marriage'
 **/
TLWcount
PlayersDb::games_group_marriage() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_marriage(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_marriage() const

/** @return   a TLW for the games group 'poverty'
 **/
TLWcount
PlayersDb::games_group_poverty() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_poverty(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_poverty() const

/** @return   a TLW for the games group 'solo'
 **/
TLWcount
PlayersDb::games_group_solo() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_solo(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_solo() const

/** @return   a TLW for the games group 'color solo'
 **/
TLWcount
PlayersDb::games_group_solo_color() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_color_solo(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_solo_color() const

/** @return   a TLW for the games group 'picture solo'
 **/
TLWcount
PlayersDb::games_group_solo_picture() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_picture_solo(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_solo_picture() const

/** @return   a TLW for the games group 'single picture solo'
 **/
TLWcount
PlayersDb::games_group_solo_picture_single() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_single_picture_solo(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_solo_picture_single() const

/** @return   a TLW for the games group 'double picture solo'
 **/
TLWcount
PlayersDb::games_group_solo_picture_double() const
{
  TLWcount tlwcount;

  for (auto gt : game_type_list)
    if (is_double_picture_solo(gt))
      tlwcount += this->games(gt);

  return tlwcount;
} // TLWcount PlayersDb::games_group_solo_picture_double() const

/** @return   a TLW for the specialpoints group 'winning'
 **/
TLWcount
PlayersDb::specialpoints_group_winning() const
{
  TLWcount tlwcount;

  for (auto const sp : Specialpoint::types)
    if (is_winning(sp))
      tlwcount += this->specialpoints(sp);

  return tlwcount;
} // TLWcount PlayersDb::specialpoints_group_winning() const

/** @return   a TLW for the specialpoints group 'announcement'
 **/
TLWcount
PlayersDb::specialpoints_group_announcement() const
{
  TLWcount tlwcount;

  for (auto const sp : Specialpoint::types)
    if (is_announcement(sp))
      tlwcount += this->specialpoints(sp);

  return tlwcount;
} // TLWcount PlayersDb::specialpoints_group_announcement() const
