/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"
#include "rating.h"
#include "ratings.h"

constexpr Rating::Type Rating::type_list[];

/** constructor
 **
 ** @param    type   rating type
 **/
Rating::Rating(Type const type) :
  type_(type)
{ }

/** -> result
 **
 ** @param    type   the rating type to return
 **
 ** @return   new rating of type 'type' (to be freed)
 **/
Rating*
Rating::new_(Type const type)
{
  switch(type) {
  case Type::max:
    return new Max();
  case Type::min:
    return new Min();
  case Type::second_max:
    return new SecondMax();
  case Type::second_min:
    return new SecondMin();
  case Type::median:
    return new Median();
  case Type::average:
    return new Average();
  case Type::average_physical:
    return new AveragePhysical();
  case Type::linear:
    return new Linear();
  case Type::linear_reverse:
    return new LinearReverse();
  } // switch(type)
  return {};
} // Rating* Rating::new_(Type const type)

/** -> result
 **
 ** @param    type   the rating type to return
 **
 ** @return   new rating of type 'type' (to be freed)
 **/
unique_ptr<Rating>
Rating::create(Type const type)
{
  switch(type) {
  case Type::max:
    return make_unique<Max>();
  case Type::min:
    return make_unique<Min>();
  case Type::second_max:
    return make_unique<SecondMax>();
  case Type::second_min:
    return make_unique<SecondMin>();
  case Type::median:
    return make_unique<Median>();
  case Type::average:
    return make_unique<Average>();
  case Type::average_physical:
    return make_unique<AveragePhysical>();
  case Type::linear:
    return make_unique<Linear>();
  case Type::linear_reverse:
    return make_unique<LinearReverse>();
  } // switch(type)
  return {};
} // unique_ptr<Rating> Rating::create(Type type)

/** write the rating in the stream
 **
 ** @param    ostr     stream to write into
 ** @param    rating   rating to write
 **/
ostream&
operator<<(ostream& ostr, Rating const& rating)
{
  rating.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, Rating rating)

/** write the rating in the stream
 **
 ** @param    ostr  stream to write into
 **/
void
Rating::write(ostream& ostr) const
{
  ostr << to_string(this->type_) << '\n';
  for (auto const& i : this->values)
    ostr << "  " << i << '\n';
} // void Rating::write(ostream& ostr) const

/** add the value
 **
 ** @param    value   value to add
 **/
void
Rating::add(int const value)
{
  this->values.push_back(value);
} // void Rating::add(int value)

/** deletes the worst 'n' arguments
 **
 ** @param    n   number of arguments to ignore
 **/
void
Rating::delete_worst(size_t const n)
{
  if (this->values.size() <= n) {
    this->values.clear();
    return ;
  }
  this->sort();
  this->values.erase(this->values.begin(), this->values.begin() + n);
} // void Rating::delete_worst(size_t n)

/** sorts the values
 **/
void
Rating::sort() const
{
  std::sort(this->values.begin(), this->values.end());
} // void Rating::sort() const

/** -> result
 **
 ** @param    type   rating type
 **
 ** @return   name of the rating type
 **/
string
to_string(Rating::Type const type)
{
  switch(type) {
  case Rating::Type::max:
    return "maximum";
    (void)_("Rating::maximum");
  case Rating::Type::min:
    return "minimum";
    (void)_("Rating::minimum");
  case Rating::Type::second_max:
    return "second maximum";
    (void)_("Rating::second maximum");
  case Rating::Type::second_min:
    return "second minimum";
    (void)_("Rating::second minimum");
  case Rating::Type::median:
    return "median";
    (void)_("Rating::median");
  case Rating::Type::average:
    return "average";
    (void)_("Rating::average");
  case Rating::Type::average_physical:
    return "physical average";
    (void)_("Rating::physical average");
  case Rating::Type::linear:
    return "linear";
    (void)_("Rating::linear");
  case Rating::Type::linear_reverse:
    return "linear reverse";
    (void)_("Rating::linear reverse");
  } // switch(type)
  return {};
} // string to_string(Rating::Type type)

/** -> result
 **
 ** @param    type  rating type
 **
 ** @return   translation of the rating type
 **/
string
gettext(Rating::Type const type)
{
  return gettext("Rating::" + to_string(type));
}

/** write the type in the output stream
 **
 ** @param    ostr   output stream
 ** @param    type   rating type
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Rating::Type const type)
{ return (ostr << to_string(type)); }

/** -> result
 **
 ** @param    name   name of the type
 **
 ** @return   the rating type
 **/
Rating::Type
Rating::type_from_string(string const& name)
{
  for (auto r : type_list)
    if (to_string(r) == name)
      return r;
  throw name;
  return Type::max;
} // Rating::Type type_from_name(string name)
