/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_SOUND

#include "sound.h"
#ifdef USE_SOUND_ALUT
#include "alut.h"
#endif
#ifdef USE_SOUND_COMMAND
#include "command.h"
#endif
#ifdef USE_SOUND_PLAYSOUND
#include "playsound.h"
#endif

#include "../party/party.h"
#include "../game/game.h"
#include "../player/player.h"
#include "../card/trick.h"

#include "../misc/setting.h"

#include "../utils/string.h"
#include "../utils/file.h"


/** -> result
 **
 ** @param    announcement   the announcement
 **
 ** @return   the type for the announcement
 **/
Sound::Type
Sound::type(GameType const gametype)
{
  switch (gametype) {
  case GameType::normal:
    return Type::gametype_normal;
  case GameType::thrown_nines:
    return Type::gametype_thrown_nines;
  case GameType::thrown_kings:
    return Type::gametype_thrown_kings;
  case GameType::thrown_nines_and_kings:
    return Type::gametype_thrown_nines_and_kings;
  case GameType::thrown_richness:
    return Type::gametype_thrown_richness;
  case GameType::redistribute:
    return Type::gametype_redistribute;
  case GameType::fox_highest_trump:
    return Type::gametype_fox_highest_trump;
  case GameType::poverty:
    return Type::gametype_poverty;
  case GameType::genscher:
    return Type::gametype_genscher;
  case GameType::marriage:
    return Type::gametype_marriage;
  case GameType::marriage_solo:
    return Type::gametype_marriage_solo;
  case GameType::marriage_silent:
    return Type::gametype_marriage_silent;
  case GameType::solo_meatless:
    return Type::gametype_solo_meatless;
  case GameType::solo_jack:
    return Type::gametype_solo_jack;
  case GameType::solo_queen:
    return Type::gametype_solo_queen;
  case GameType::solo_king:
    return Type::gametype_solo_king;
  case GameType::solo_queen_jack:
    return Type::gametype_solo_queen_jack;
  case GameType::solo_king_jack:
    return Type::gametype_solo_king_jack;
  case GameType::solo_king_queen:
    return Type::gametype_solo_king_queen;
  case GameType::solo_koehler:
    return Type::gametype_solo_koehler;
  case GameType::solo_club:
    return Type::gametype_solo_club;
  case GameType::solo_heart:
    return Type::gametype_solo_heart;
  case GameType::solo_spade:
    return Type::gametype_solo_spade;
  case GameType::solo_diamond:
    return Type::gametype_solo_diamond;
  } // switch (gametype)

  return Type::first;
} // static Sound::Type Sound::type(GameType gametype)

/** -> result
 **
 ** @param    marriage_selector   the marriage selector
 **
 ** @return   the type for the marriage selector
 **/
Sound::Type
Sound::type(MarriageSelector const marriage_selector)
{
  switch (marriage_selector) {
  case MarriageSelector::team_set:
    return Type::marriage_selector_team_set;
  case MarriageSelector::first_foreign:
    return Type::marriage_selector_first_foreign;
  case MarriageSelector::first_trump:
    return Type::marriage_selector_first_trump;
  case MarriageSelector::first_color:
    return Type::marriage_selector_first_color;
  case MarriageSelector::first_club:
    return Type::marriage_selector_first_club;
  case MarriageSelector::first_spade:
    return Type::marriage_selector_first_spade;
  case MarriageSelector::first_heart:
    return Type::marriage_selector_first_heart;
  case MarriageSelector::silent:
    return Type::marriage_selector_silent;
  } // switch (marriage_selector)

  return Type::first;
} // static Sound::Type Sound::type(MarriageSelector marriage_selector)

/** -> result
 **
 ** @param    announcement   the announcement
 **
 ** @return   the type for the announcement
 **/
Sound::Type
Sound::type(Announcement const announcement)
{
  switch (announcement) {
  case Announcement::noannouncement:
    return Type::announcement_noannouncement;
  case Announcement::no120:
    return Type::announcement_no120;
  case Announcement::no90:
    return Type::announcement_no90;
  case Announcement::no60:
    return Type::announcement_no60;
  case Announcement::no30:
    return Type::announcement_no30;
  case Announcement::no0:
    return Type::announcement_no0;
  case Announcement::reply:
    return Type::announcement_reply;
  case Announcement::no120_reply:
    return Type::announcement_no120_reply;
  case Announcement::no90_reply:
    return Type::announcement_no90_reply;
  case Announcement::no60_reply:
    return Type::announcement_no60_reply;
  case Announcement::no30_reply:
    return Type::announcement_no30_reply;
  case Announcement::no0_reply:
    return Type::announcement_no0_reply;
  } // switch (announcement)

  return Type::first;
} // static Sound::Type Sound::type(Announcement announcement)


/** -> result
 **
 ** @param    type   the sound type
 **
 ** @return   the name of the sound type
 **/
string
Sound::name(Sound::Type const type)
{
  switch (type) {
  case Type::gametype_normal:
    return "gametype: normal";
  case Type::gametype_thrown_nines:
    return "gametype: thrown nines";
  case Type::gametype_thrown_kings:
    return "gametype: thrown kings";
  case Type::gametype_thrown_nines_and_kings:
    return "gametype: thrown nines and kings";
  case Type::gametype_thrown_richness:
    return "gametype: thrown richness";
  case Type::gametype_redistribute:
    return "gametype: redistribute";
  case Type::gametype_fox_highest_trump:
    return "gametype: fox highest trump";
  case Type::gametype_poverty:
    return "gametype: poverty";
  case Type::gametype_genscher:
    return "gametype: genscher";
  case Type::gametype_marriage:
    return "gametype: marriage";
  case Type::gametype_marriage_solo:
    return "gametype: marriage solo";
  case Type::gametype_marriage_silent:
    return "gametype: marriage silent";
  case Type::gametype_solo_meatless:
    return "gametype: solo meatless";
  case Type::gametype_solo_jack:
    return "gametype: solo jack";
  case Type::gametype_solo_queen:
    return "gametype: solo queen";
  case Type::gametype_solo_king:
    return "gametype: solo king";
  case Type::gametype_solo_queen_jack:
    return "gametype: solo queen jack";
  case Type::gametype_solo_king_jack:
    return "gametype: solo king jack";
  case Type::gametype_solo_king_queen:
    return "gametype: solo king queen";
  case Type::gametype_solo_koehler:
    return "gametype: solo koehler";
  case Type::gametype_solo_club:
    return "gametype: solo club";
  case Type::gametype_solo_heart:
    return "gametype: solo heart";
  case Type::gametype_solo_spade:
    return "gametype: solo spade";
  case Type::gametype_solo_diamond:
    return "gametype: solo diamond";

  case Type::marriage_selector_team_set:
    return "marriage selector: team set";
  case Type::marriage_selector_first_foreign:
    return "marriage selector: first foreign";
  case Type::marriage_selector_first_trump:
    return "marriage selector: first trump";
  case Type::marriage_selector_first_color:
    return "marriage selector: first color";
  case Type::marriage_selector_first_club:
    return "marriage selector: first club";
  case Type::marriage_selector_first_spade:
    return "marriage selector: first spade";
  case Type::marriage_selector_first_heart:
    return "marriage selector: first heart";
  case Type::marriage_selector_silent:
    return "marriage selector: silent";

  case Type::cards_distributed:
    return "cards distributed";
  case Type::trick_move_in_pile:
    return "trick move in pile";
  case Type::card_played:
    return "card played";

  case Type::poverty:
    return "poverty";
  case Type::poverty_denied:
    return "poverty denied";
  case Type::poverty_accepted:
    return "poverty accepted";

  case Type::marriage:
    return "marriage";
  case Type::genscher:
    return "genscher";

  case Type::announcement_noannouncement:
    return "announcement: no announcement";
  case Type::announcement_re:
    return "announcement: re";
  case Type::announcement_contra:
    return "announcement: contra";
  case Type::announcement_no120:
    return "announcement: no 120";
  case Type::announcement_no90:
    return "announcement: no 90";
  case Type::announcement_no60:
    return "announcement: no 60";
  case Type::announcement_no30:
    return "announcement: no 30";
  case Type::announcement_no0:
    return "announcement: no 0";
  case Type::announcement_reply:
    return "announcement: reply";
  case Type::announcement_no120_reply:
    return "announcement: no 120 reply";
  case Type::announcement_no90_reply:
    return "announcement: no 90 reply";
  case Type::announcement_no60_reply:
    return "announcement: no 60 reply";
  case Type::announcement_no30_reply:
    return "announcement: no 30 reply";
  case Type::announcement_no0_reply:
    return "announcement: no 0 reply";

  case Type::swines:
    return "swines";
  case Type::hyperswines:
    return "hyperswines";

  case Type::no_winner:
    return "no winner";
  case Type::soloplayer_won:
    return "soloplayer won";
  case Type::soloplayer_lost:
    return "soloplayer lost";
  case Type::re_won:
    return "re won";
  case Type::contra_won:
    return "contra won";
  } // switch (type)

  return "";
} // static string Sound::name(Sound::Type type)

/** -> result
 **
 ** @param    type   the sound type
 **
 ** @return   the text for the sound type
 **/
string
Sound::text(Sound::Type const type)
{
  switch (type) {
  case Type::gametype_normal:
    return "normal";
  case Type::gametype_thrown_nines:
    return "thrown nines";
  case Type::gametype_thrown_kings:
    return "thrown kings";
  case Type::gametype_thrown_nines_and_kings:
    return "thrown nines and kings";
  case Type::gametype_thrown_richness:
    return "thrown richness";
  case Type::gametype_redistribute:
    return "redistribute";
  case Type::gametype_fox_highest_trump:
    return "fox highest trump";
  case Type::gametype_poverty:
    return "poverty";
  case Type::gametype_genscher:
    return "genscher";
  case Type::gametype_marriage:
    return "marriage";
  case Type::gametype_marriage_solo:
    return "marriage solo";
  case Type::gametype_marriage_silent:
    return "marriage silent";
  case Type::gametype_solo_meatless:
    return "solo meatless";
  case Type::gametype_solo_jack:
    return "solo jack";
  case Type::gametype_solo_queen:
    return "solo queen";
  case Type::gametype_solo_king:
    return "solo king";
  case Type::gametype_solo_queen_jack:
    return "solo queen jack";
  case Type::gametype_solo_king_jack:
    return "solo king jack";
  case Type::gametype_solo_king_queen:
    return "solo king queen";
  case Type::gametype_solo_koehler:
    return "solo koehler";
  case Type::gametype_solo_club:
    return "solo club";
  case Type::gametype_solo_heart:
    return "solo heart";
  case Type::gametype_solo_spade:
    return "solo spade";
  case Type::gametype_solo_diamond:
    return "solo diamond";

  case Type::marriage_selector_team_set:
    return ""; // "team set";
  case Type::marriage_selector_first_foreign:
    return "first foreign";
  case Type::marriage_selector_first_trump:
    return "first trump";
  case Type::marriage_selector_first_color:
    return "first color";
  case Type::marriage_selector_first_club:
    return "first club";
  case Type::marriage_selector_first_spade:
    return "first spade";
  case Type::marriage_selector_first_heart:
    return "first heart";
  case Type::marriage_selector_silent:
    return "silent";

  case Type::cards_distributed:
    return "cards distributed";
  case Type::trick_move_in_pile:
    return "trick move in pile";
  case Type::card_played:
    return "card played";

  case Type::poverty:
    return "poverty";
  case Type::poverty_denied:
    return "poverty denied";
  case Type::poverty_accepted:
    return "poverty accepted";

  case Type::marriage:
    return "marriage";
  case Type::genscher:
    return "genscher";

  case Type::announcement_noannouncement:
    return ""; // "no announcement";
  case Type::announcement_re:
    return "re";
  case Type::announcement_contra:
    return "contra";
  case Type::announcement_no120:
    return "no 120";
  case Type::announcement_no90:
    return "no 90";
  case Type::announcement_no60:
    return "no 60";
  case Type::announcement_no30:
    return "no 30";
  case Type::announcement_no0:
    return "no 0";
  case Type::announcement_reply:
    return "reply";
  case Type::announcement_no120_reply:
    return "no 120 reply";
  case Type::announcement_no90_reply:
    return "no 90 reply";
  case Type::announcement_no60_reply:
    return "no 60 reply";
  case Type::announcement_no30_reply:
    return "no 30 reply";
  case Type::announcement_no0_reply:
    return "no 0 reply";

  case Type::swines:
    return "swines";
  case Type::hyperswines:
    return "hyperswines";

  case Type::no_winner:
    return "no winner";
  case Type::soloplayer_won:
    return "soloplayer won";
  case Type::soloplayer_lost:
    return "soloplayer lost";
  case Type::re_won:
    return "re won";
  case Type::contra_won:
    return "contra won";
  } // switch (type)

  return "";
} // static string Sound::text(Sound::Type type)

/** -> result
 **
 ** @param    type   the sound type
 **
 ** @return   the file name for the sound type
 **/
string
Sound::filename(Sound::Type const type)
{
  string filename = Sound::text(type);
  String::replace_all(filename, " ",  "_");
  return filename;
} // static string Sound::filename(Sound::Type type)

/** -> result
 **
 ** @param    position   position of the sound
 ** @param    voice      voice (directory to take the sound from)
 ** @param    type       sound to play
 **
 ** @return   the filename of the sound, empty if not found
 **/
string
Sound::path(Position const position,
            string const& voice,
            Type const type)
{
  // search the sound file
  auto const basename = (Sound::filename(type)
                           + "." + ::setting(Setting::Type::sound_extension));
  for (auto const& dir : ::setting.data_directories()) {
    auto const path
      = (dir + "/"
         + ::setting(Setting::Type::sounds_directory) + "/"
         + voice + "/"
         + basename);
    if (File::isfile(path)) {
      return path;
    } // if (File::isfile(path_tmp))
  } // for (dir \in ::setting.data_directories)

  return {};
} // static string Sound::path(Position position, string voice, Type type)


/** create a sound class
 **/
unique_ptr<Sound>
Sound::create()
{
#ifdef USE_SOUND_ALUT
  return make_unique<Alut>();
#endif
#ifdef USE_SOUND_COMMAND
  return make_unique<Command>();
#endif
#ifdef USE_SOUND_PLAYSOUND
  return make_unique<WinPlaySound>();
#endif

  DEBUG_ASSERTION(false,
                  "Sound::create()"
                  "  no sound object created");

  return {};
} // static unique_ptr<Sound> Sound::create()

/** constructor
 **/
Sound::Sound()
{
  Game::signal_open().connect(*this, &Sound::game_open, this->disconnector_);
}

/** play the sound (global group)
 **
 ** @param    type   sound to play
 **/
void
Sound::play(Type const type)
{
  if (::setting(Setting::Type::sound))
    this->play(Position::center, "general", type);
} // virtual void Sound::play(Type type)

/** play the sound (group of the player)
 **
 ** @param    player   player who says it
 ** @param    type     sound to play
 **/
void
Sound::play(Player const& player, Type const type)
{
  if (::setting(Setting::Type::sound))
    this->play(::setting.position(player.no()), player.voice(), type);
} // virtual void Sound::play(Player player, Type type)

/** the game is opened
 **
 ** @param    game   game that is opened
 **/
void
Sound::game_open(Game const& game)
{
  {
    auto& game = ::party->game();
    game.signal_cards_distributed().connect(*this, &Sound::game_cards_distributed, this->disconnector_);
    game.tricks().signal_trick_move_in_pile().connect(*this, &Sound::trick_move_in_pile, this->disconnector_);
    game.signal_card_played().connect(*this, &Sound::card_played, this->disconnector_);
    game.signal_start().connect(*this, &Sound::game_start, this->disconnector_);
    game.signal_finish().connect(*this, &Sound::game_finish, this->disconnector_);
    game.poverty().signal_shift().connect(*this, &Sound::poverty_shift, this->disconnector_);
    game.poverty().signal_take_denied().connect(*this, &Sound::poverty_take_denied, this->disconnector_);
    game.poverty().signal_take_accepted().connect(*this, &Sound::poverty_take_accepted, this->disconnector_);
    game.marriage().signal_marriage().connect(*this, &Sound::marriage, this->disconnector_);
    game.marriage().signal_genscher().connect(*this, &Sound::genscher, this->disconnector_);
    game.announcements().signal_announcement_made().connect(*this, &Sound::announcement_made, this->disconnector_);
    game.swines().signal_swines_announced().connect(*this, &Sound::swines_announced, this->disconnector_);
    game.swines().signal_hyperswines_announced().connect(*this, &Sound::hyperswines_announced, this->disconnector_);
  }
} // void Sound::game_open(Game game)

/** the cards are distributed
 **/
void
Sound::game_cards_distributed()
{
  this->play(Type::cards_distributed);
} // void Sound::game_cards_distributed()

/** the trick is moved into a pile
 **
 ** @param    trick   trick
 **/
void
Sound::trick_move_in_pile(Trick const& trick)
{
  this->play(trick.winnerplayer(), Type::trick_move_in_pile);
} // void Sound::trick_move_in_pile(Trick trick)

/** the card is played
 **
 ** @param    card   card that is played
 **/
void
Sound::card_played(HandCard const& card)
{
  this->play(card.player(), Type::card_played);
} // void Sound::card_played(HandCard card)

/** the game is started
 **
 ** @param    game  game that is started
 **/
void
Sound::game_start()
{
  auto const& game = ::party->game();
  switch (game.type()) {
  case GameType::marriage:
    this->play(game.players().soloplayer(), Sound::type(game.type()));
    this->play(game.players().soloplayer(), Sound::type(game.marriage().selector()));
    break;
  case GameType::marriage_solo:
  case GameType::marriage_silent:
  case GameType::solo_meatless:
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
  case GameType::solo_koehler:
  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_diamond:
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::redistribute:
  case GameType::fox_highest_trump:
  case GameType::poverty:
    this->play(game.players().soloplayer(), Sound::type(game.type()));
    break;
  default:
    break;
  } // switch (game.type())
} // void Sound::game_start()

/** the game is finished
 **
 ** @param    game   game that is finished
 **/
void
Sound::game_finish()
{
  auto const& game = ::party->game();
  if (game.winnerteam() == Team::noteam) {
    this->play(Type::no_winner);
    return ;
  }

  if (is_solo(game.type())) {
    this->play(  (game.winnerteam() == Team::re)
               ? Type::soloplayer_won
               : Type::soloplayer_lost);
  } else { // if (GameType::is_solo(game.type()))
    this->play(  (game.winnerteam() == Team::re)
               ? Type::re_won
               : Type::contra_won);
  } // if !(GameType::is_solo(game.type()))
} // void Sound::game_finish()

/** the marriage partner has found a bride
 **
 ** @param    bridegroom   player with the marriage
 ** @param    bride        player who is married
 **/
void
Sound::marriage(Player const& bridegroom, Player const& bride)
{
  this->play(bridegroom, Type::marriage);
} // void Sound::marriage(Player bridegroom, Player bride)

/** a player has announced a genscher
 **
 ** @param    genscher   player with the genscher
 ** @param    partner    selected partner
 **/
void
Sound::genscher(Player const& genscher, Player const& partner)
{
  this->play(genscher, Type::genscher);
} // void Sound::genscher(Player genscher, Player partner)


// poverty

/** 'player' shifts 'cardno' cards
 **
 ** @param    player   player with the poverty
 ** @param    cardno   number of shifted cards
 **/
void
Sound::poverty_shift(Player const& player, unsigned const cardno)
{
#ifdef POSTPONED
  this->play(player, "poverty: "
             + String::to_string(cardno) + " cards");
#else
  this->play(player, Type::poverty);
#endif
} // void Sound::poverty_shift(Player player, unsigned cardno)

/** the player 'player' has denied the poverty trumps
 **
 ** @param    player   player who has denied the poverty
 **/
void
Sound::poverty_take_denied(Player const& player)
{
  this->play(player, Type::poverty_denied);
} // void Sound::poverty_take_denied(Player player)

/** and has returned 'cardno' cards with 'trumpno' trumps
 **
 ** @param    player    player who has accepted the poverty
 ** @param    cardno    number of reshifted cards
 ** @param    trumpno   number of reshifted trumps
 **/
void
Sound::poverty_take_accepted(Player const& player,
                             unsigned const cardno,
                             unsigned const trumpno)
{
#ifdef POSTPONED
  this->play(player, "poverty accepted: "
             + String::to_string(trumpno) + " trumps");
#else
  this->play(player, Type::poverty_accepted);
#endif
} // void Sound::poverty_take_accepted(Player player, unsigned cardno, unsigned trumpno)

/** an announcement is made
 **
 ** @param    announcement   announcement made
 ** @param    player         player who has announced
 **/
void
Sound::announcement_made(Announcement const announcement,
                         Player const& player)
{
  if (announcement == Announcement::no120)
    this->play(player, ((player.team() == Team::re)
                        ? Type::announcement_re
                        : Type::announcement_contra));
  else
    this->play(player, Sound::type(announcement));
} // void Sound::announcement_made(Announcement announcement, Player player)

/** the player has swines
 **
 ** @param    player    player with the swines
 **/
void Sound::swines_announced(Player const& player)
{
  this->play(player, Type::swines);
} // void Sound::swines_announced(Player const& player)

/** the player has hyperswines
 **
 ** @param    player   player with the hyperswines
 **/
void Sound::hyperswines_announced(Player const& player)
{
  this->play(player, Type::hyperswines);
} // void Sound::hyperswines_announced(Player player)

#endif // #ifdef USE_SOUND
