/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "card_suggestion.h"
#include "table.h"
#include "ui.h"
#include "cards.h"
#include "hand.h"

#include "../../party/party.h"
#include "../../game/game.h"
#include "../../player/player.h"
#include "../../player/ai/ai.h"
#include "../../os/bug_report_replay.h"

#include <gtkmm/box.h>
#include <gtkmm/eventbox.h>
#include <gtkmm/image.h>
#include <gtkmm/label.h>
#include <gtkmm/textview.h>
#include <gtkmm/main.h>
#include <gdk/gdkkeysyms.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table      the table
 **/
CardSuggestion::CardSuggestion(Table* const table) :
  Base(table),
  Gtk::StickyDialog("FreeDoko – " + _("Window::Card suggestion"))
{
  this->signal_realize().connect(sigc::mem_fun(*this, &CardSuggestion::init));
} // CardSuggestion::CardSuggestion(Table* table)

/** destructor
 **/
CardSuggestion::~CardSuggestion() = default;

/** init the dialog
 **/
void
CardSuggestion::init()
{
  { // action area
    add_close_button(*this);

    this->play_button = Gtk::manage(new Gtk::Button(_("Button::play card")));
    this->play_button->set_image_from_icon_name("dialog-ok-apply");
    this->play_button->set_always_show_image();
    this->add_action_widget(*this->play_button, Gtk::RESPONSE_NONE);
    this->play_button->signal_clicked().connect(sigc::mem_fun(*this, &CardSuggestion::play_card));

    this->play_button->set_can_default();
    this->play_button->grab_default();
    this->play_button->grab_focus();
    this->play_button->add_accelerator("activate",
                                       this->get_accel_group(),
                                       GDK_KEY_t,
                                       Gdk::ModifierType(0),
                                       Gtk::AccelFlags(0));
  } // action area

  { // vbox
    this->get_content_area()->set_spacing(20);

    this->card_box = Gtk::manage(new Gtk::EventBox());
    this->card_box->signal_button_press_event().connect(sigc::mem_fun(*this, &CardSuggestion::play_card_signal));

    this->card_image = this->ui->cards->new_managed_image();
    this->card_box->add(*this->card_image);
    this->card_image->set_padding(2 EM, 2 EX);
    this->heuristics_text = Gtk::manage(new Gtk::Label(_("Thinking, please wait.")));
    this->heuristics_description = Gtk::manage(new Gtk::TextView);
    this->heuristics_description->set_wrap_mode(Gtk::WRAP_WORD);
    this->heuristics_description->get_buffer()->set_text(_("Thinking, please wait."));

    this->get_content_area()->pack_start(*card_box, Gtk::PACK_SHRINK);
    this->get_content_area()->pack_start(*this->heuristics_text, Gtk::PACK_SHRINK);
    this->get_content_area()->pack_start(*this->heuristics_description);
  } // vbox

  this->show_all_children();
} // void CardSuggestion::init()

/** @return   the player for the suggestion,
 **       that is the current player of the game
 **/
::Player const&
CardSuggestion::player()
{
  return this->ui->game().players().current_player();
} // ::Player& CardSuggestion::player()

/** show the window
 **
 ** @param    show_window   whether to show the window
 **/
void
CardSuggestion::show_information(bool const show_window)
{
  if (!this->get_realized())
    this->realize();

  if (show_window)
    this->present();

  this->get_suggestion();
} // void CardSuggestion::show_information(bool show_window)

/** play the suggested card
 **/
void
CardSuggestion::play_card()
{
  if (this->suggested_card.is_empty())
    return ;

  const_cast<Player&>(this->player()).hand().request_card(this->suggested_card);
  this->ui->thrower(this->suggested_card, __FILE__, __LINE__);
} // void CardSuggestion::play_card()

/** a card was played
 **/
void
CardSuggestion::card_played()
{
  //this->hide();
  this->suggested_card = Card();
  this->ui->cards->change_managed(this->card_image, Card());
  this->heuristics_text->set_label("");
  this->heuristics_description->get_buffer()->set_text("");
  this->play_button->set_sensitive(false);
} // void CardSuggestion::card_played()

/** get the suggestion from the ai
 **/
void
CardSuggestion::get_suggestion()
{
  DEBUG_ASSERTION(dynamic_cast<Ai const*>(&this->player()),
                  "CardSuggestion::get_Suggestion():\n"
                  "  'this->player' cannot be cast to an ai.");

  if (this->thinking)
    return ;

  auto const& ai = dynamic_cast<::Ai const&>(this->player());
  if (this->suggested_card.is_empty()
      || (   ::bug_report_replay
          && (this->player().type() == Player::Type::human))) {
    this->thinking = true;

    this->heuristics_text->set_label(_("Thinking, please wait."));
    this->heuristics_description->get_buffer()->set_text("");
    this->ui->cards->change_managed(this->card_image, Card());

    this->play_button->set_sensitive(false);

    auto const thrower_depth = this->ui->thrower.depth();
    this->ui->thrower.set_depth(0);
    try {
      this->suggested_card = const_cast<Ai&>(ai).card_suggestion();
      if (ai.hand().contains(this->suggested_card))
        const_cast<Ai&>(ai).hand().request_card(this->suggested_card);
    } catch (Card const& card) {
      this->ui->thrower(card, __FILE__, __LINE__);
    } catch (GameStatus const game_status) {
      this->ui->thrower(game_status, __FILE__, __LINE__);
    } catch (...) {
      this->thinking = false;
      this->ui->thrower.set_depth(thrower_depth);
      throw;
    } // try
    this->ui->thrower.set_depth(thrower_depth);

    this->thinking = false;
  } // if (!this->suggested_card)

  dynamic_cast<Table*>(this->parent)->draw_all();

  this->ui->cards->change_managed(this->card_image, this->suggested_card);
  this->heuristics_text->set_label(_(ai.last_heuristic()));
  auto buffer = this->heuristics_description->get_buffer();
  buffer->set_text(_("AiConfig::Heuristic::Description::" + to_string(ai.last_heuristic())));
  auto const rationale_text = ai.last_heuristic_rationale().text();
  if (!rationale_text.empty()) {
    buffer->insert(buffer->end(),
                   "\n\n" + _("decision logic") + ":\n" + rationale_text);
  }

  this->play_button->set_sensitive(true);
} // void CardSuggestion::get_suggestion()

/** play the suggested card
 **
 ** @return   whether the card can be played
 **/
bool
CardSuggestion::play_card_signal(GdkEventButton* event)
{
  if (this->suggested_card.is_empty())
    return false;

  this->play_card();

  return true;
} // bool CardSuggestion::play_card_signal(GdkEventButton* event)

/** the mouse was clicked on the table
 **/
void
CardSuggestion::mouse_click_on_table()
{ }

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
