/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "cards.h"

#include "ui.h"
#include "splash_screen.h"
#include "help.h"

#include "../status_message.h"

#include "../../misc/setting.h"

#include "../../utils/file.h"

#include <gtkmm/image.h>
#include <gtkmm/messagedialog.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   the parent object
 **/
Cards::Cards(Base* const parent) :
  Base(parent)
{
  this->load_all();
}

/** destructor
 **/
Cards::~Cards() = default;

/** @return   the scaling
 **/
double
Cards::scaling() const
{ return this->scaling_; }

/** update the scaling
 **/
void
Cards::update_scaling()
{
  if (::setting(Setting::Type::original_cards_size))
    this->scaling_ = 1;
  else
    this->scaling_ = (static_cast<double>(::setting(Setting::Type::cards_height))
                      / this->height_original());

  // make that the width cannot get '0'
  if (static_cast<unsigned>(this->scaling_ * this->width_original())
      < ::setting.min(Setting::Type::cards_height))
    this->scaling_
      = (static_cast<double>(::setting.min(Setting::Type::cards_height))
         / this->width_original());


  // set the scaling of the 'ScaledPixbufRotations'
  this->back_.set_scaling(this->scaling());
  for (auto& c : this->card_)
    for (auto& v : c)
      v.set_scaling(this->scaling_);
} // void Cards::update_scaling()

/** -> result
 **
 ** @param    rotation   the rotation of the card
 **
 ** @return   the original width of the cards
 **/
int
Cards::width_original(Rotation const rotation) const
{
  DEBUG_ASSERTION(this->back_,
                  "Cards::width_original(rotation):\n"
                  "  'this->back_' is empty");

  return this->back_.get_orig_width(rotation);
}

/** -> result
 **
 ** @param    rotation   the rotation of the card
 **
 ** @return   the original height of the cards
 **/
int
Cards::height_original(Rotation const rotation) const
{
  DEBUG_ASSERTION(this->back_,
                  "Cards::height_original(rotation):\n"
                  "  'this->back_' is empty");

  return this->back_.get_orig_height(rotation);
}


/** -> result
 **
 ** @param    rotation   the rotation of the card
 **
 ** @return   the width of the cards
 **/
int
Cards::width(Rotation const rotation) const
{
  return this->back_.get_width(rotation);
}

/** -> result
 **
 ** @param    rotation   the rotation of the card
 **
 ** @return   the height of the cards
 **/
int
Cards::height(Rotation const rotation) const
{
  return this->back_.get_height(rotation);
}

/** -> result
 **
 ** @param    rotation   the rotation
 **
 ** @return   the back of the cards (with rotation 'rotation')
 **/
Gdk::ScaledPixbuf&
Cards::back(Rotation const rotation)
{
  return this->back_[rotation];
}

/** -> result
 **
 ** @param    card      the card
 ** @param    rotation   the rotation
 **
 ** @return   the pixbuf of 'card' (with rotation 'rotation')
 **/
Gdk::ScaledPixbuf&
Cards::card(Card const& card, Rotation const rotation)
{
  if (   (card == Card::empty)
      || (card == Card::unknown))
    return this->back(rotation);

  // a pointer to the vector with all cards with 'card.color'
  vector<Gdk::ScaledPixbufRotations>* card_color = nullptr;
  switch (card.color()) {
  case Card::club:
    card_color = &(this->card_[0]);
    break;
  case Card::spade:
    card_color = &(this->card_[1]);
    break;
  case Card::heart:
    card_color = &(this->card_[2]);
    break;
  case Card::diamond:
    card_color = &(this->card_[3]);
    break;
  default:
    DEBUG_ASSERTION(false,
                    "Cards::card(card, rotation):\n"
                    "  card.color '" << card.color() << "' not valid.");
    break;
  } // switch(card.color())

  // a pointer to the vector with all cards with 'card'
  Gdk::ScaledPixbufRotations* card_color_value = nullptr;
  switch (card.value()) {
  case Card::ace:
    card_color_value = &((*card_color)[0]);
    break;
  case Card::ten:
    card_color_value = &((*card_color)[1]);
    break;
  case Card::king:
    card_color_value = &((*card_color)[2]);
    break;
  case Card::queen:
    card_color_value = &((*card_color)[3]);
    break;
  case Card::jack:
    card_color_value = &((*card_color)[4]);
    break;
  case Card::nine:
    card_color_value = &((*card_color)[5]);
    break;
  default:
    DEBUG_ASSERTION(false,
                    "Cards::card(card, rotation):\n"
                    "  card.value '" << card.value() << "' not valid.");
    break;
  } // switch(card.color())

  return (*card_color_value)[rotation];
} // Gdk::ScaledPixbuf& Cards::card(Card card, Rotation rotation = up)

/** -> result
 **
 ** @param    rotation   the rotation
 **
 ** @return   the back of the cards (with rotation 'rotation') (scaled)
 **/
Glib::RefPtr<Gdk::Pixbuf> const&
Cards::back_original(Rotation const rotation)
{
  return this->back().orig_pixbuf();
}

/** -> result
 **
 ** @param    card      the card
 ** @param    rotation   the rotation
 **
 ** @return   the pixbuf of 'card' (with rotation 'rotation') (original)
 **/
Glib::RefPtr<Gdk::Pixbuf> const&
Cards::card_original(Card const& card, Rotation const rotation)
{
  DEBUG_ASSERTION(this->card(card, rotation),
                  "Cards::card_original(card, rotation):\n"
                  "  'card' is null");
  DEBUG_ASSERTION(this->card(card, rotation).orig_pixbuf(),
                  "Cards::card_original(card, rotation):\n"
                  "  'card.orig_pixbuf()' is null");

  return this->card(card, rotation).orig_pixbuf();
}

/** -> result
 **
 ** @param    card   card
 **           (default: Card())
 **
 ** @return   new image with the card 'card' that is managed
 **           (that is size and theme)
 **/
Gtk::Image*
Cards::new_managed_image(Card const& card)
{
  auto image = Gtk::manage(new Gtk::Image(this->card(card)));
  image->set_data("card", new Card(card));

  image->signal_unrealize().connect(sigc::bind<Gtk::Image const* const>(sigc::mem_fun(*this, &Cards::remove_managed), image));

  this->managed_image_.push_back(image);

  return image;
}

/** changes the type of 'image'
 **
 ** @param    image   image to change the type of
 ** @param    card   new card
 **/
void
Cards::change_managed(Gtk::Image* const image, Card const card)
{
  DEBUG_ASSERTION(image->get_data("card"),
                  "Cards::change_managed:\n"
                  "  image has no data 'type'");
  // ToDo: Test whether the image is in the vector

  delete static_cast<Card*>(image->steal_data("card"));
  image->set_data("card", new Card(card));

  image->set(this->card(card));
}

/** remove 'image' from the managed image list
 **
 ** @param    image   image to remove
 **/
void
Cards::remove_managed(Gtk::Image const* const image)
{
  for (auto i = this->managed_image_.begin();
       i != this->managed_image_.end();
      )
    if (*i == image) {
      delete static_cast<Card*>((*i)->steal_data("card"));
      i = this->managed_image_.erase(i);
      break;
    } else {
      ++i;
    }
} // void Cards::remove_managed(Gtk::Image const* image)

/** update the managed images
 **/
void
Cards::update_managed()
{
  for (auto i : this->managed_image_) {
    auto const card = *(static_cast<Card*>(i->get_data("card")));
    i->set(this->card(card));
  }
}

/** load the cards and the back
 **/
void
Cards::load_all()
{
  this->load_back();
  this->load_cards();
}

/** load the cards back
 **/
void
Cards::load_back()
{
  this->ui->add_status_message(_("loading the cards (cards back)"));
  auto back_new
    = Gdk::ScaledPixbufRotations(::setting.path(Setting::Type::cards_back));
  if (!back_new) {
    if (!this->back_) {
      // gettext: %s = back, %s = cardset
      this->ui->error(_("Error::loading the card back %s (%s)",
                        ::setting.value(Setting::Type::cards_back),
                        ::setting.value(Setting::Type::cardset)));
    }
    return ;
  } // if (!back_new)

  this->back_ = back_new;
  this->back_.set_scaling(this->scaling());

  if (::setting(Setting::Type::original_cards_size) == true)
    ::setting.set(Setting::Type::cards_height, this->height_original());

  this->update_scaling();

  this->ui->remove_status_message(_("loading the cards (cards back)"));
} // void Cards::load_back()

/** load the cards
 **/
void
Cards::load_cards()
{
  StatusMessages sm(*this->ui);
  if (   (::game_status != GameStatus::programstart)
      && (::game_status != GameStatus::party_initial_loaded) ) {
    sm.add(_("loading the cards"));
  } // if (::game_status != GameStatus::programstart)

  // load the cards
  vector<Card::Color> color = {Card::club, Card::spade, Card::heart, Card::diamond};
  vector<Card::Value> value = {Card::ace, Card::ten, Card::king, Card::queen, Card::jack, Card::nine};

  vector<vector<Gdk::ScaledPixbufRotations>> card_new;
  for (auto c : color) {
    card_new.push_back({});
    for (auto v : value) {
      sm.add(_("loading the cards (%t)", _(Card(c, v))));
      card_new.back().push_back(Gdk::ScaledPixbufRotations(::setting.path(Setting::Type::cardset) + "/"
                                                           + ::setting.value(Setting::Type::cards_directory) + "/"
                                                           + to_string(c)
                                                           + "/"
                                                           + to_string(v)
                                                           + "."
                                                           + ::setting(Setting::Type::graphic_extension)));
      card_new.back().back().set_scaling(this->scaling());
      if (!card_new.back().back())
        // gettext: %t = cardn, %s = cardset
        this->ui->error(_("Error::loading the card %t (%s)",
                          _(Card(c, v)),
                          ::setting.value(Setting::Type::cardset)));
      if (   (::game_status == GameStatus::programstart)
          || (::game_status == GameStatus::party_initial_loaded) )
        this->ui->add_progress(1.0 / (color.size() * value.size()));
    } // for (v \in value)
  } // for (c \in color)

  this->card_ = card_new;
} // void Cards::load_cards()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
