/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "game_debug.h"

#include "ui.h"
#include "cards.h"

#include "../../card/trick.h"
#include "../../player/ai/ai.h"
#include "../../utils/string.h"

#include <gtkmm/label.h>
#include <gtkmm/treeview.h>
#include <gtkmm/liststore.h>
#include <gtkmm/treestore.h>
#include <gtkmm/frame.h>
#include <gtkmm/box.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/textview.h>

namespace UI_GTKMM_NS {

/** constructor
 **/
GameDebug::PlayerInfoModel::PlayerInfoModel()
{
  this->add(this->name);
  this->add(this->team);
  this->add(this->swines);
  this->add(this->announcement);
  this->add(this->points);
  this->add(this->tricks);
} // GameDebug::PlayerInfoModel::PlayerInfoModel()

/** constructor
 **/
GameDebug::TeamInfoModel::TeamInfoModel()
{
  this->add(this->name);
  this->known_team.resize(::party->rule()(Rule::Type::number_of_players_in_game));
  for (auto& t : this->known_team)
    this->add(t);
} // GameDebug::TeamInfoModel::TeamInfoModel()

/** constructor
 **/
GameDebug::TricksModel::TricksModel()
{
  this->add(this->trickno);
  this->add(this->playerno);
  this->add(this->card);
  this->add(this->heuristic);
  this->add(this->heuristic_text);
  this->add(this->rationale);
  this->add(this->isvirtual);
} // GameDebug::TricksModel::TricksModel()

/** constructor
 **
 ** @param    parent   the parent widget
 **/
GameDebug::GameDebug(Base* const parent) :
  Base(parent),
  StickyDialog(_("Window::Game debug"), false)
{
  this->ui->add_window(*this);

  this->signal_realize().connect(sigc::mem_fun(*this, &GameDebug::init));

  Game::signal_open().connect_back(*this, &GameDebug::game_open, this->disconnector_);

#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_key_press_event().connect(sigc::mem_fun(*this, &GameDebug::on_key_press_event));
#endif
} // GameDebug::GameDebug(Base* parent)

/** destructor
 **/
GameDebug::~GameDebug() = default;

/** the game is opened
 **
 ** @param    game   game that is opened
 **/
void
GameDebug::game_open(Game& game)
{
  this->realize();

  this->tricks_rows.clear();
  this->tricks_store->clear();

  while (!this->ai_test_rows.empty())
    this->ai_test_rows.pop();

  for (auto const& p : game.players())
    this->name_changed(p);

  {
    game.signal_start().connect_back(*this, &GameDebug::update_info, this->disconnector_);
    game.tricks().signal_trick_open().connect_back(*this, &GameDebug::trick_open, this->disconnector_);
    game.signal_card_played().connect_back(*this, &GameDebug::card_played, this->disconnector_);
    game.announcements().signal_announcement_made().connect_back(*this, &GameDebug::announcement_made, this->disconnector_);
    game.swines().signal_swines_announced().connect_back(*this, &GameDebug::swines_announced, this->disconnector_);
    game.swines().signal_hyperswines_announced().connect_back(*this, &GameDebug::hyperswines_announced, this->disconnector_);
    game.marriage().signal_marriage().connect_back(*this, &GameDebug::marriage, this->disconnector_);
    game.marriage().signal_genscher().connect_back(*this, &GameDebug::genscher, this->disconnector_);
    game.signal_ai_test_card().connect_back(*this, &GameDebug::ai_test_card, this->disconnector_);
    game.signal_ai_card_weighting().connect_back(*this, &GameDebug::ai_card_weighting, this->disconnector_);
    game.signal_virtual_card_played().connect_back(*this, &GameDebug::virtual_card_played, this->disconnector_);
  }
} // void GameDebug::game_open(Game game)

/** a new trick has opened
 **
 ** @param    trick   new trick
 **/
void
GameDebug::trick_open(::Trick const& trick)
{ 
  if (!this->tricks_rows.empty())
    this->tricks_treeview->collapse_row(this->tricks_store->get_path(this->tricks_rows.back()));
  this->tricks_rows.push_back(*this->tricks_store->append());
  this->tricks_rows.back()[this->tricks_model.trickno]
    = std::to_string(trick.no());
} // void GameDebug::trick_open(Trick trick)

/** 'player' has played 'card'
 **
 ** @param    card   the played card
 **/
void
GameDebug::card_played(HandCard const& card)
{ 
  Player const& player = card.player();

  Gtk::TreeModel::Row row;
  if (this->ai_test_rows.empty())
    row = *this->tricks_store->append(this->tricks_rows.back().children());
  else {
    row = this->ai_test_rows.top();
    this->tricks_treeview->collapse_row(this->tricks_store->get_path(row));
    this->ai_test_rows.pop();
    while (!this->ai_test_rows.empty())
      this->ai_test_rows.pop();
  }
  this->tricks_treeview->expand_row(this->tricks_store->get_path(row.parent()), false);
  row[this->tricks_model.playerno]
    = std::to_string(player.no());
  row[this->tricks_model.card] = _(card);
  if (dynamic_cast<Ai const*>(&player)) {
    auto const& ai = dynamic_cast<Ai const&>(player);
    row[this->tricks_model.heuristic] = ai.last_heuristic();
    if (ai.last_heuristic() == Aiconfig::Heuristic::choose_best_card) {
      row[this->tricks_model.heuristic_text] = _(ai.aitype());
    } else {
      row[this->tricks_model.heuristic_text] = _(ai.last_heuristic());
    }
    row[this->tricks_model.rationale] = ai.last_heuristic_rationale();
  } else {
    row[this->tricks_model.rationale] = {};
  }

  this->update_info();
} // void GameDebug::card_played(HandCard card)

/** the ai tests the card 'card'
 **
 ** @param    card      the card that is tested
 ** @param    playerno   number of the player who plays the card
 **/
void
GameDebug::ai_test_card(Card const& card, unsigned const playerno)
{
#ifdef WINDOWS
  // Eventuell wird dadurch die Performance unter Windows wesentlich beschleunigt
  return;
#endif
  auto const& ai = dynamic_cast<Ai const&>(this->ui->game().players().current_player());
  if (this->ai_test_rows.empty()) {
    this->ai_test_rows.push(*this->tricks_store->append(this->tricks_rows.back().children()));
    auto& row = this->ai_test_rows.top();
    row[this->tricks_model.heuristic] = Aiconfig::Heuristic::no_heuristic;
    row[this->tricks_model.heuristic_text] = _(ai.aitype());
    row[this->tricks_model.rationale] = ai.last_heuristic_rationale();
  }
  this->ai_test_rows.push(*this->tricks_store->append(this->ai_test_rows.top().children()));
  if (this->ai_test_rows.size() == 2)
    this->tricks_treeview->expand_to_path(this->tricks_store->get_path(this->ai_test_rows.top().parent()));
  this->ai_test_rows.top()[this->tricks_model.card]
    = _(card);
  this->ai_test_rows.top()[this->tricks_model.playerno]
    = std::to_string(playerno);
} // void GameDebug::ai_test_card(Card card, unsigned playerno)

/** the ai has calculated a weighting for the tested card
 **
 ** @param    weighting   the weighting for the tested card
 **/
void
GameDebug::ai_card_weighting(int const weighting)
{
#ifdef WINDOWS
  // Eventuell wird dadurch die Performance unter Windows wesentlich beschleunigt
  return;
#endif
  if (this->ai_test_rows.empty())
    DEBUG_ASSERTION(false,
                    "GameDebug::ai_card_weighting():\n"
                    "  ai_test_rows is empty");
  while (this->ai_test_rows.top()[this->tricks_model.isvirtual])
    this->ai_test_rows.pop();

  if ( (weighting == INT_MIN)
      || (weighting == INT_MAX) )
    this->ai_test_rows.top()[this->tricks_model.heuristic_text]
      = "-";
  else
    this->ai_test_rows.top()[this->tricks_model.heuristic_text]
      = std::to_string(weighting);

  this->ai_test_rows.top()[this->tricks_model.isvirtual]
    = false;

  this->tricks_treeview->collapse_row(this->tricks_store->get_path(this->ai_test_rows.top()));
  this->ai_test_rows.pop();
} // void GameDebug::ai_card_weighting(int weighting)

/** virtual game: card is played
 **
 ** @param    card   the card that is played
 **/
void
GameDebug::virtual_card_played(HandCard const& card)
{
#ifdef WINDOWS
  // Eventuell wird dadurch die Performance unter Windows wesentlich beschleunigt
  return;
#endif
  auto const& player = card.player();

  this->ai_test_rows.push(*this->tricks_store->append(this->ai_test_rows.top().children()));
  auto row = this->ai_test_rows.top();

  row[this->tricks_model.card] = _(card);
  row[this->tricks_model.playerno] = std::to_string(player.no());
  if (dynamic_cast<Ai const*>(&player)) {
    auto const& ai = dynamic_cast<Ai const&>(player);
      row[this->tricks_model.heuristic] = ai.last_heuristic();
    if (ai.last_heuristic() == Aiconfig::Heuristic::choose_best_card) {
      row[this->tricks_model.heuristic_text] = _(ai.aitype());
      row[this->tricks_model.rationale] = {};
    } else {
      row[this->tricks_model.rationale] = ai.last_heuristic_rationale();
    }
  }
  row[this->tricks_model.isvirtual]
    = true;
} // void GameDebug::virtual_card_played(HandCard card)

/** the announcement 'announcement' has been made by player 'player'
 **
 ** @param    announcement   the announcement
 ** @param    player      the player, who has made the announcement
 **/
void
GameDebug::announcement_made(Announcement const announcement,
                             Player const& player)
{
  this->update_info();
} // void GameDebug::announcement_made(Announcement announcement, Player player)

/** the player has swines
 **
 ** @param    player      the player with the swines
 **/
void
GameDebug::swines_announced(Player const& player)
{
  this->update_info();
} // void GameDebug::swines_announced(Player player)

/** the player has hyperswines
 **
 ** @param    player      the player with the swines
 **/
void
GameDebug::hyperswines_announced(Player const& player)
{
  this->update_info();
} // void GameDebug::hyperswines_announced(Player player)

/** marriage: Information of the new team
 **
 ** @param    bridegroom   the player with the marriage
 ** @param    bride      the bride
 **/
void
GameDebug::marriage(Player const& bridegroom, Player const& bride)
{
  this->update_info();
} // void GameDebug::marriage(Player bridegroom, Player bride)

/** genscher: Information of the new team
 **
 ** @param    genscher   genscher player
 ** @param    partner    partner of the genscher player
 **/
void
GameDebug::genscher(Player const& genscher, Player const& partner)
{
  this->update_info();
} // void GameDebug::genscher(Player genscher, Player partner)

/** create all subelements
 **/
void
GameDebug::init()
{
  this->set_icon(this->ui->icon);

  this->set_default_size(static_cast<int>(6 * this->ui->cards->height()),
                         static_cast<int>(4 * this->ui->cards->height()));

  add_close_button(*this);

  auto const& party = this->ui->party();

  { // content
    auto content_box = Gtk::manage(new Gtk::Box());
    content_box->set_spacing(2 EM);
    content_box->set_border_width(1 EM);

    { // left box
      auto left_box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX));
      left_box->set_spacing(1 EX);
      { // game info
        auto box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX));
        box->set_homogeneous();
        box->set_border_width(1 EX);

        { // seed
          auto box2 = Gtk::manage(new Gtk::Box());
          auto label = Gtk::manage(new Gtk::Label(_("Label::seed") + ": "));
          box2->pack_start(*label, false, false);
          this->seed = Gtk::manage(new Gtk::Label(""));
          box2->pack_start(*this->seed, false, false);

          box->add(*box2);
        } // seed
        { // game type
          auto box2 = Gtk::manage(new Gtk::Box());
          auto label = Gtk::manage(new Gtk::Label(_("Label::gametype") + ": "));
          box2->pack_start(*label, false, false);
          this->gametype = Gtk::manage(new Gtk::Label(""));
          box2->pack_start(*this->gametype, false, false);

          box->add(*box2);
        } // game type
        { // soloplayer
          auto box2 = Gtk::manage(new Gtk::Box());
          auto label = Gtk::manage(new Gtk::Label(_("Label::soloplayer") + ": "));
          box2->pack_start(*label, false, false);
          this->soloplayer = Gtk::manage(new Gtk::Label(""));
          box2->pack_start(*this->soloplayer, false, false);

          box->add(*box2);
        } // seed

        auto frame = Gtk::manage(new Gtk::Frame(_("Label::game info")));
        frame->add(*box);
        left_box->pack_start(*frame, false, true);
      } // game info

      { // player_info
        this->player_info_list
          = Gtk::ListStore::create(this->player_info_model);
        this->player_info_treeview
          = Gtk::manage(new Gtk::TreeView(this->player_info_list));
        this->player_info_treeview->get_selection()->set_mode(Gtk::SELECTION_NONE);

        unsigned column = 0;
        this->player_info_treeview->append_column(_("name"),
                                                  this->player_info_model.name);
        column += 1;
        this->player_info_treeview->append_column(_("team"),
                                                  this->player_info_model.team);
        this->player_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->player_info_treeview->append_column(_("swines"),
                                                  this->player_info_model.swines);
        this->player_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->player_info_treeview->append_column(_("announcement"),
                                                  this->player_info_model.announcement);
        this->player_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->player_info_treeview->append_column(_("points"),
                                                  this->player_info_model.points);
        this->player_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->player_info_treeview->append_column(_("tricks"),
                                                  this->player_info_model.tricks);
        this->player_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;

        (void) column;

        for (unsigned p = 0;
             p < party.rule()(Rule::Type::number_of_players_in_game);
             ++p) {
          this->player_info_rows.push_back(*this->player_info_list->append());
          this->player_info_rows.back()[this->player_info_model.name]
            = party.players()[p].name();
        }

        auto frame = Gtk::manage(new Gtk::Frame(_("player info")));
        frame->add(*this->player_info_treeview);
        left_box->pack_start(*frame, false, true);
      } // player_info

      { // team_info
        this->team_info_list
          = Gtk::ListStore::create(this->team_info_model);
        this->team_info_treeview
          = Gtk::manage(new Gtk::TreeView(this->team_info_list));
        this->team_info_treeview->get_selection()->set_mode(Gtk::SELECTION_NONE);

        unsigned column = 0;
        this->team_info_treeview->append_column(_("name"),
                                                this->team_info_model.name);
        column += 1;
        for (unsigned p = 0;
             p < this->team_info_model.known_team.size();
             ++p) {
          this->team_info_treeview->append_column(party.players()[p].name(),
                                                  this->team_info_model.known_team[p]);
          this->team_info_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
          column += 1;
        }

        // real team
        this->team_info_rows.push_back(*this->team_info_list->append());
        // teaminfo of the game
        this->team_info_rows.push_back(*this->team_info_list->append());
        for (unsigned p = 0;
             p < this->team_info_model.known_team.size();
             ++p) {
          this->team_info_rows.push_back(*this->team_info_list->append());
          this->team_info_rows.back()[this->team_info_model.name]
            = party.players()[p].name();
        } // for (p)

        Gtk::Frame* frame = Gtk::manage(new Gtk::Frame(_("team info")));
        frame->add(*this->team_info_treeview);
        left_box->pack_start(*frame, false, true);
      } // team_info
      content_box->pack_start(*left_box, false, false);
    } // left box

    { // tricks
      auto right_box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 0));
      { // tricks
        this->tricks_store = Gtk::TreeStore::create(this->tricks_model);
        this->tricks_treeview
          = Gtk::manage(new Gtk::TreeView(this->tricks_store));
        this->tricks_treeview->get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
        this->tricks_treeview->set_rules_hint(true);

        unsigned column = 0;
        this->tricks_treeview->append_column(_("card"),
                                             this->tricks_model.card);
        column += 1;
        this->tricks_treeview->append_column(_("player"),
                                             this->tricks_model.playerno);
        this->tricks_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->tricks_treeview->append_column(_("trick"),
                                             this->tricks_model.trickno);
        this->tricks_treeview->get_column_cell_renderer(column)->set_property("xalign", 0.5);
        column += 1;
        this->tricks_treeview->append_column(_("heuristic"),
                                             this->tricks_model.heuristic_text);
        column += 1;
        this->tricks_treeview->get_selection()->signal_changed().connect(sigc::mem_fun(*this, &GameDebug::tricks_row_selection_changed));

        (void) column;
        this->tricks_treeview->get_selection()->set_mode(Gtk::SELECTION_SINGLE);

        auto scrolled_window = Gtk::manage(new Gtk::ScrolledWindow);
        scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC,
                                    Gtk::POLICY_AUTOMATIC);
        scrolled_window->set_hadjustment(this->tricks_treeview->get_hadjustment());
        scrolled_window->set_vadjustment(this->tricks_treeview->get_vadjustment());
        scrolled_window->add(*this->tricks_treeview);

        auto frame = Gtk::manage(new Gtk::Frame(_("tricks")));
        frame->add(*scrolled_window);

        right_box->pack_start(*frame, true, true);
      } // tricks
      { // rationale
        auto text = Gtk::manage(new Gtk::TextView());
        text->set_wrap_mode(Gtk::WRAP_WORD);
        text->set_editable(false);
        text->set_cursor_visible(false);
        this->heuristic_rationale = text->get_buffer();

        this->heuristic_rationale_frame = Gtk::manage(new Gtk::Frame(_("rationale")));
        this->heuristic_rationale_frame->set_margin_top(1 EX);
        this->heuristic_rationale_frame->add(*text);
        right_box->pack_start(*this->heuristic_rationale_frame, false, true);
      } // rationale
      content_box->pack_end(*right_box, true, true);
    }
    this->get_content_area()->pack_start(*content_box);
  } // content

  this->show_all_children();
  this->heuristic_rationale_frame->hide();
} // void GameDebug::init()

/** updates the information for the players
 **/
void
GameDebug::update_info()
{
  if (!(::game_status & GameStatus::game))
    return ;

  Game const& game = this->ui->game();

  this->seed->set_label(std::to_string(game.seed()));
  this->gametype->set_label(_(game.type()));
  if (game.players().exists_soloplayer())
    this->soloplayer->set_label(game.players().soloplayer().name());
  else
    this->soloplayer->set_label("-");

  for (unsigned p = 0; p < game.playerno(); ++p) {
    ::Player const& player = game.player(p);

    { // player info
      auto& row = this->player_info_rows[p];

      row[this->player_info_model.team]
        = _(player.team());
      row[this->player_info_model.swines]
        = (player.has_swines() ? "S" : "");
      if (player.has_hyperswines())
        row[this->player_info_model.swines]
          = row[this->player_info_model.swines] + "H";

      row[this->player_info_model.announcement]
        = ( (player.announcement() == Announcement::noannouncement)
           ? ""s
           : _(player.announcement())
          );
      row[this->player_info_model.points]
        = player.points_in_trickpile();
      row[this->player_info_model.tricks]
        = player.number_of_tricks();

    } // player info
    { // team info
      this->team_info_rows[0][this->team_info_model.known_team[p]]
        = _("Team::" + ::short_string(player.team()));
      this->team_info_rows[1][this->team_info_model.known_team[p]]
        = _("Team::" + ::short_string(player.game().teaminfo().get(player)));

      for (unsigned q = 0; q < game.playerno(); ++q) {
        Player const& player2 = player.game().player(q);
        if (dynamic_cast<Ai const*>(&player2)) {
          Ai const& ai2 = dynamic_cast<Ai const&>(player2);
          this->team_info_rows[2 + q][this->team_info_model.known_team[p]]
            = (_("Team::" + ::short_string(ai2.teaminfo(player)))
               + (is_real(ai2.teaminfo(player))
                  ? ""s
                  : " " + std::to_string(ai2.team_information().team_value(player))));
        } else {
          this->team_info_rows[2 + q][this->team_info_model.known_team[p]]
            = "";
        }
      } // for (q)
    } // team info
  } // for (p)
} // void GameDebug::update_info()

/** the name of 'player' has changed
 **
 ** @param    player   player whose name has changed
 **/
void
GameDebug::name_changed(Player const& player)
{
  if (!this->get_realized())
    return;

  if (::game_status < GameStatus::game_init)
    return ;
  if (!this->ui->game().players().contains(player))
    return ;

  this->team_info_treeview->get_column(1 + player.no())->set_title(player.name()); 

  this->player_info_rows[player.no()][this->player_info_model.name]
    = player.name();

  this->team_info_rows[2 + player.no()][this->team_info_model.name]
    = player.name();
} // void GameDebug::name_changed(Player player)

/** the row selection has changed:
 ** show the belonging game summary
 **/
void
GameDebug::tricks_row_selection_changed()
{
  auto selection = this->tricks_treeview->get_selection();

  if (!selection->get_selected()) {
    this->heuristic_rationale_frame->hide();
    this->heuristic_rationale->set_text({});
    return ;
  }
  auto row = *selection->get_selected();
  Aiconfig::Heuristic const heuristic = row[this->tricks_model.heuristic];
  if (heuristic != Aiconfig::Heuristic::no_heuristic) {
    Heuristic::Rationale const& rationale = row[this->tricks_model.rationale];
    if (!!rationale) {
      this->heuristic_rationale_frame->set_label(_("decision logic"));
      auto text = String::remove_trailing_newlines(rationale.text());
      this->heuristic_rationale->set_text(text);
    } else {
      this->heuristic_rationale_frame->set_label(_("description"));
      this->heuristic_rationale->set_text(_("AiConfig::Heuristic::Description::" + to_string(heuristic)));
    }
    this->heuristic_rationale_frame->show();
  } else {
    this->heuristic_rationale_frame->hide();
  }
}

/** a key has been pressed
 **
 ** @param    key   the key
 **
 ** @return   whether the key was managed
 **/
bool
GameDebug::on_key_press_event(GdkEventKey* const key)
{
  return this->ui->key_press(key);
} // bool GameDebug::on_key_press_event(GdkEventKey* key)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
