/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "game_finished.h"
#include "game_summary.h"
#include "game_review.h"

#include "ui.h"
#include "bug_report.h"
#include "main_window.h"
#include "table.h"
#include "party_points.h"

#include "../../party/party.h"
#include "../../game/game_summary.h"
#include "../../misc/setting.h"

#include <gtkmm/checkbutton.h>
#include <gtkmm/label.h>
#include <gtkmm/main.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   the parent widget
 **/
GameFinished::GameFinished(Base* const parent) :
  Base(parent),
  StickyDialog("FreeDoko – " + _("Window::Game finished"),
               *parent->ui->main_window, false)
{
  this->ui->add_window(*this);

  // the game summary is created here, because else I would have first
  // to show this window and afterwards set the gameno.
  this->game_summary = Gtk::manage(new GameSummary(this));

  this->signal_realize().connect(sigc::mem_fun(*this, &GameFinished::init));
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_hide().connect(sigc::mem_fun(*this, &GameFinished::on_hide));
#endif
} // GameFinished::GameFinished(Base* parent)

/** destructor
 **/
GameFinished::~GameFinished() = default;

/** create all subelements
 **/
void
GameFinished::init()
{
  this->set_icon(this->ui->icon);

  this->set_position(Gtk::WIN_POS_CENTER_ON_PARENT);

  this->get_content_area()->set_spacing(10);

  this->get_content_area()->pack_start(*this->game_summary);

  { // show party points
    this->show_party_points
      = Gtk::manage(new Gtk::Button(_("Button::show party points")));

    this->get_content_area()->pack_end(*this->show_party_points, Gtk::PACK_SHRINK);
    this->show_party_points->set_halign(Gtk::ALIGN_CENTER); 
    this->show_party_points->signal_clicked().connect(sigc::mem_fun0(this->ui->table->party_points(), &Gtk::Window::present));
  } // show party points

  { // show game review
    this->show_game_review
      = Gtk::manage(new Gtk::CheckButton(_("Button::show game review")));

    this->get_content_area()->pack_end(*this->show_game_review, Gtk::PACK_SHRINK);
    this->show_game_review->set_halign(Gtk::ALIGN_CENTER);
    this->show_game_review->signal_clicked().connect(sigc::mem_fun(*this, &GameFinished::toggle_game_review));
  } // show game review

  { // action area
    this->create_bug_report_button
      = Gtk::manage(new Gtk::Button(_("Button::create bug report")));
    //this->add_action_widget(*this->create_bug_report_button, Gtk::RESPONSE_NONE);
    this->get_action_area()->add(*this->create_bug_report_button);
    this->create_bug_report_button->signal_clicked().connect(sigc::mem_fun(*(this->ui->bug_report),
                                                                           &BugReport::create_report));

    this->replay_game_button = Gtk::manage(new Gtk::Button(_("Button::replay game")));
    this->replay_game_button->set_image_from_icon_name("edit-redo");
    this->replay_game_button->set_always_show_image();
    this->add_action_widget(*this->replay_game_button, Gtk::RESPONSE_CLOSE);
    this->replay_game_button->signal_clicked().connect(sigc::mem_fun(*this, &GameFinished::replay_game));

    this->next_game_button = Gtk::manage(new Gtk::Button(_("Button::next game")));
    this->next_game_button->set_image_from_icon_name("go-next");
    this->next_game_button->set_always_show_image();
    this->add_action_widget(*this->next_game_button, Gtk::RESPONSE_CLOSE);
    this->next_game_button->signal_clicked().connect(sigc::mem_fun(*this, &GameFinished::hide));

    this->next_game_button->set_can_default();
    this->next_game_button->grab_default();
    this->next_game_button->grab_focus();
  } // action area

  this->show_all_children();

  this->setting_update(Setting::Type::show_bug_report_button_in_game_finished_window);
} // void GameFinished::init()

/** the game has finished:
 ** show the summary and let the player review the game
 **/
void
GameFinished::game_finished()
{
  this->realize();

  auto const game_summary = ::GameSummary(this->ui->game());
  if (this->ui->party().is_replayed_game())
    this->game_summary->set_game_summary(game_summary);
  else
    this->game_summary->set_gameno(this->ui->party().gameno());
  if (this->ui->party().is_last_game()) {
    next_game_button->set_label(_("Button::end of party"));
    this->next_game_button->set_image_from_icon_name("go-last");
  } else {
    next_game_button->set_label(_("Button::next game"));
    this->next_game_button->set_image_from_icon_name("go-next");
  }

  this->present();

  while (   !this->ui->thrower
         && (::game_status == GameStatus::game_finished)
         && this->is_visible())
    ::ui->wait();

  this->show_game_review->set_active(false);
  if (this->ui->party().is_replayed_game())
    this->game_summary->remove_game_summary();

  this->hide();
} // void GameFinished::game_finished()

/** a setting has changed
 ** show_bug_report_button_in_game_finished_window:
 **   update whether to show the 'create bug report' button
 **
 ** @param    type   the type that has changed
 **/
void
GameFinished::setting_update(int const type)
{
  if (!this->get_realized())
    return ;

  switch (type) {
  case Setting::Type::show_bug_report_button_in_game_finished_window:
    if (::setting(Setting::Type::show_bug_report_button_in_game_finished_window))
      this->create_bug_report_button->show();
    else
      this->create_bug_report_button->hide();
    break;
  default:
    break;
  } // switch(type)
} // void GameFinished::setting_update(int type)

/** the name of 'player' has changed
 **
 ** @param    player   the player with the changed name
 **/
void
GameFinished::name_changed(::Player const& player)
{
  if (!this->get_realized())
    return ;

  if (this->game_summary)
    this->game_summary->name_changed(player);
} // void GameFinished::name_changed(Player const& player)

/** event: hide the window
 **/
void
GameFinished::on_hide()
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->Gtk::Dialog::on_hide();
#endif
  this->ui->main_quit();
  this->game_summary->set_gameno(UINT_MAX);
} // void GameFinished::on_hide()

/** toggles the game review
 **/
void
GameFinished::toggle_game_review()
{
  if (this->show_game_review->get_active()) {
    DEBUG_ASSERTION(!this->game_review,
                    "GameFinished::toggle_game_review:\n"
                    "  already in game review");
    this->game_review = std::make_unique<GameReview>(this);
  } else {
    DEBUG_ASSERTION(this->game_review,
                    "GameFinished::toggle_game_review:\n"
                    "  not in game review");
    this->game_review.reset();
  }
  this->ui->table->force_redraw_all();
} // void GameFinished::toggle_game_review()

/** replay the game
 **/
void
GameFinished::replay_game()
{
  this->ui->party().set_replay_game();
  this->hide();
} // void GameFinished::replay_game()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
