/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "gameinfo.gametype.h"

#include "ui.h"
#include "icons.h"

#include "../../game/game.h"
#include "../../player/player.h"

#include "../../utils/string.h"

#include <gtkmm/image.h>
#include <gtkmm/label.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   the parent object
 ** @param    game     the corresponding game
 **/
GameInfoDialog::Information::GameType::GameType(Base* const parent,
                                                Game const& game) :
  // gettext: %t = game.type()
  GameInfoDialog::Information(parent, _("Window::GameInfo::%t",
                                        _(game.type()))),
  game(&game)
{
  this->ui->icons->change_managed(this->icon, this->game->type());

  this->init();
} // GameInfoDialog::Information::GameType::GameType(Base parent, Game game)

/** destructor
 **/
GameInfoDialog::Information::GameType::~GameType() = default;

/** update all texts
 **/
void
GameInfoDialog::Information::GameType::update_texts()
{
  if (!(::game_status & GameStatus::game))
    return ;

  if (   (::game_status != GameStatus::game_reservation)
      && (::game_status != GameStatus::game_redistribute)
      && (::game_status != GameStatus::game_start))
    if (this->ui->game().type() != ::GameType::normal)
      return ;

  auto const& game = *this->game;

  switch (game.type()) {
  case ::GameType::normal:
    this->label->set_label(_(game.type()));
    break;
  case ::GameType::thrown_nines:
    this->label->set_label(_("%s has %u nines.",
                             game.players().soloplayer().name(),
                             game.players().soloplayer().hand().count_nines())
                          );
    break;
  case ::GameType::thrown_kings:
    this->label->set_label(_("%s has %u kings.",
                             game.players().soloplayer().name(),
                             game.players().soloplayer().hand().count_kings())
                          );
    break;
  case ::GameType::thrown_nines_and_kings:
    this->label->set_label(_("%s has %u nines and kings.",
                             game.players().soloplayer().name(),
                             game.players().soloplayer().hand().count_nines() + game.players().soloplayer().hand().count_kings())
                          );
    break;
  case ::GameType::thrown_richness:
    this->label->set_label(_("%s has %u richness.",
                             game.players().soloplayer().name(),
                             game.players().soloplayer().hand().points())
                          );
    break;
  case ::GameType::fox_highest_trump:
    // gettext: %t = gametype (fox highest trump)
    this->label->set_label(_("%s has %t.",
                             game.players().soloplayer().name(),
                             _(game.type()))
                          );
    break;
  case ::GameType::redistribute:
    this->label->set_label(_("%s forces a redistribution.",
                             game.players().soloplayer().name())
                          );
    break;
  case ::GameType::poverty: {
    // error
    // gettext: %t = gametype (poverty)
    string text = _("%s has a %t.",
                    game.players().soloplayer().name(),
                    _(game.type()));
    if (::game_status != GameStatus::game_redistribute) {
      for (auto const& player : game.players()) {
        if (   (player != game.players().soloplayer())
            && (player.team() == Team::re)) {
#ifdef POSTPONED
          text +=  ("        \n        "
                    + _("Partner is %s.",
                        player.name()));
#endif
          if (game.players().soloplayer().hand().count_trumps() == 1) {
            text +=  ("        \n        "
                      + _("%s has returned %u trump.",
                          player.name(),
                          game.players().soloplayer().hand().count_trumps()
                         ));
          } else {
            text +=  ("        \n        "
                      + _("%s has returned %u trumps.",
                          player.name(),
                          game.players().soloplayer().hand().count_trumps()
                         ));
          }

        } // if (player is the partner)
      } // for (player)
    } // if (::game_status != game_redistribute)
    this->label->set_label(text);
  } break;
  case ::GameType::genscher:
    // cannot happen
    break;
  case ::GameType::marriage:
  case ::GameType::marriage_solo:
    // gettext: %1s = marriage soloplayer, %t = type (marriage), %t = marriage selector
    this->label->set_label(_("%s has a %t: %t.",
                             game.players().soloplayer().name(),
                             _(game.type()),
                             _(game.marriage().selector()))
                          );
    break;
  case ::GameType::marriage_silent:
    // cannot happen
    break;
  case ::GameType::solo_meatless:
  case ::GameType::solo_jack:
  case ::GameType::solo_queen:
  case ::GameType::solo_king:
  case ::GameType::solo_queen_jack:
  case ::GameType::solo_king_jack:
  case ::GameType::solo_king_queen:
  case ::GameType::solo_koehler:
  case ::GameType::solo_club:
  case ::GameType::solo_spade:
  case ::GameType::solo_heart:
  case ::GameType::solo_diamond:
    // gettext: %s = player, %s = gametyp
    this->label->set_label(_("%s plays a %t.",
                             game.players().soloplayer().name(),
                             _(game.type()))
                          );
    break;
  } // switch (game.type())
} // void GameInfoDialog::Information::GameType::update_texts()

/** @return   whether the information is blocking the gameplay
 **/
bool
GameInfoDialog::Information::GameType::is_blocking() const
{
  return (this->Information::is_blocking() 
          || (::game_status == GameStatus::game_redistribute));
} // virtual bool GameInfoDialog::Information::GameType::is_blocking() const

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
