/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "gameinfo_dialog.h"
#include "gameinfo.h"
#include "gameinfo.announcement.h"
#include "gameinfo.swines.h"
#include "table.h"
#include "ui.h"
#include "main_window.h"

#include "../../party/party.h"
#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../player/human/human.h"

#include <gtkmm/separator.h>
#include <gtkmm/main.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table   the table
 **/
GameInfoDialog::GameInfoDialog(Table* const table) :
  Base(table),
  Gtk::StickyDialog("FreeDoko – " + _("Window::game information"), *table->ui->main_window, false)
{
  this->ui->add_window(*this);

  this->signal_realize().connect(sigc::mem_fun(*this,
                                               &GameInfoDialog::init));
} // GameInfoDialog::GameInfoDialog(Table* table)

/** destruktor
 **/
GameInfoDialog::~GameInfoDialog()
{
  this->clear();
} // GameInfoDialog::~GameInfoDialog()

/** init the dialog
 **/
void
GameInfoDialog::init()
{
  this->set_icon(this->ui->icon);

  this->set_skip_taskbar_hint();

  { // action area
    { // hyperswines announce
      this->hyperswines_announce_button
        = Gtk::manage(new Gtk::Button(_("Menu::Item::hyperswines announcement")));
      this->add_action_widget(*this->hyperswines_announce_button,
                              Gtk::RESPONSE_NONE);

      this->hyperswines_announce_button->signal_clicked().connect(sigc::mem_fun(*this, &GameInfoDialog::announce_hyperswines));
      this->hyperswines_announce_button->signal_clicked().connect(sigc::mem_fun(*this, &GameInfoDialog::close));

      // the button is only shown when the announcement is valid, see add_info().
      this->hyperswines_announce_button->show_all_children();
    } // hyperswines announce
    { // announcement reply button
      this->announcement_reply_button
        = Gtk::manage(new Gtk::Button(""));
      this->add_action_widget(*this->announcement_reply_button, Gtk::RESPONSE_CLOSE);

      this->announcement_reply_button->set_can_default();
      this->announcement_reply_button->grab_default();

      this->announcement_reply_button->signal_clicked().connect(sigc::mem_fun(*this, &GameInfoDialog::announcement_reply));
      this->announcement_reply_button->signal_clicked().connect(sigc::mem_fun(*this, &GameInfoDialog::close));

      // the button is only shown when the reply is valid, see add_info().
      this->announcement_reply_button->show_all_children();
    } // announcement reply button
    add_close_button(*this);
  } // action area

  { // move to the top left corner of the table
    int x, y;
    this->table().get_window()->get_origin(x, y);
    this->move(x, y);
  } // move to the top left corner of the table
} // void GameInfoDialog::init()

/** add the information and manage it
 **
 ** @param    info   game information to add
 **/
void
GameInfoDialog::add_info(Information* const info)
{
  this->infos.push_back(info);

  if (this->infos.empty()) {
    this->set_title(info->title);
  } else {
    this->get_content_area()->add(*Gtk::manage(new Gtk::Separator()));
    this->set_title(_("Window::game information"));
  }

  this->get_content_area()->add(*info);
  info->show();

  auto const& game = this->ui->game();

  { // check for announcement reply
    if (   (game.players().count_humans() == 1)
        && !this->ui->game().rule()(Rule::Type::knocking)
        && (!this->announcement_reply_button->is_visible())
        && (dynamic_cast<Information::Announcement*>(info)) ) {
      auto const& announcement_info
        = dynamic_cast<Information::Announcement const&>(*info);
      auto const& human = game.players().human();
      if (   (announcement_info.player->team()
              != human.team())
          && !is_reply(announcement_info.announcement) ) {
        if (game.announcements().is_valid(Announcement::reply,
                                          human)) {
          auto const opposite_team = ::opposite(human.team());
          auto const opposite_announcement = game.announcements().last(opposite_team).announcement;
          auto const reply = to_reply(opposite_announcement);
          this->announcement_reply_button->set_label((   (reply == Announcement::no120)
                                                      || (reply == Announcement::no120_reply))
                                                     ? _(human.team())
                                                     : _(reply));
          this->announcement_reply_button->show();
        } // if (reply)
      } // if (can reply)
    } // if (is announcement)
  } // check for announcement reply

  { // check for hyperswines announcement
    if (   (game.players().count_humans() == 1)
        && (!this->hyperswines_announce_button->is_visible())
        && (dynamic_cast<Information::Swines*>(info)) ) {
      if (game.swines().hyperswines_announcement_valid(game.players().human())) {
        this->hyperswines_announce_button->show();
      } // if (reply valid
    } // if (is announcement)
  } // check for hyperswines announcement
} // void GameInfoDialog::add_info(Information* info)

/** clear the window
 **/
void
GameInfoDialog::clear()
{
  for (auto& info : this->infos) {
    this->get_content_area()->remove(*info);
    delete info;
  }

  this->infos.clear();

  if (this->announcement_reply_button)
    this->announcement_reply_button->hide();

  if (this->hyperswines_announce_button)
    this->hyperswines_announce_button->hide();
} // void GameinfDialog::clear()

/** @return   the table
 **/
Table&
GameInfoDialog::table()
{
  return *dynamic_cast<Table*>(this->parent);
} // Table& GameinfDialog::table()

/** show the information and manage it
 **
 ** @param    info                  information to show
 ** @param    close_automatically   whether to close the window automatically
 ** @param    close_delay             the time (in tenth of seconds) to wait
 **                                  before closing the window
 **
 ** @todo      manage multiple delays
 **/
void
GameInfoDialog::show(Information* const info,
                     bool const close_automatically,
                     unsigned const close_delay)
{
  if (!this->get_realized())
    this->realize();

  this->add_info(info);

  this->present();

#ifdef WORKAROUND
  // @todo   manage multiple delays
  if (this->timeout_connection)
    this->timeout_connection.disconnect();

  if (close_automatically)
    this->timeout_connection
      = Glib::signal_timeout().connect(sigc::slot0<bool>(sigc::mem_fun(*this, &GameInfoDialog::on_timeout)),
                                       close_delay);
#endif

  if (this->is_blocking()) {
    while (   this->is_visible()
           && !this->ui->thrower) {
      ::ui->wait();
    }

    this->close();
  } // if (this->is_blocking)
} // void GameInfoDialog::show(Information* info, bool close_automatically, unsigned close_delay)

/** show the information and close it on a click
 **
 ** @param    info   information to show
 **/
void
GameInfoDialog::show(Information* const info)
{
  this->show(info, false, 0);
} // void GameInfoDialog::show(Information* info)

/** hide the window
 **/
void
GameInfoDialog::close()
{
  if (this->timeout_connection)
    this->timeout_connection.disconnect();

  this->hide();
  this->clear();
} // void GameInfoDialog::close()

/** update the announcement reply sensitivity
 **/
void
GameInfoDialog::update_announcement_reply()
{
  if (!this->announcement_reply_button)
    return ;

  auto const& game = this->ui->game();
  if (   this->announcement_reply_button->is_visible()
      && !game.announcements().is_valid(Announcement::reply,
                                        game.players().human()) ) {
    this->announcement_reply_button->hide();
  } // if (reply invalid)
} // void GameInfoDialog::update_announcement_reply()

/** the name of 'player' has changed
 **
 ** @param    player   player whose name has changed
 **/
void
GameInfoDialog::name_changed(Player const& player)
{
  for (auto& info : this->infos)
    info->name_changed(player);
} // void GameInfoDialog::name_changed(Player player)

/** the timeout event:
 ** close the window and exit the main loop
 **
 ** @return   false (to stop the timeout event)
 **/
bool
GameInfoDialog::on_timeout()
{
  this->close();

  return false;
} // bool GameInfoDialog::on_timeout()

/** reply to the last announcement
 **/
void
GameInfoDialog::announcement_reply()
{
  auto& game = this->ui->game();

  DEBUG_ASSERTION((game.players().count_humans() == 1),
                  "GameInfoDialog::announcement_reply()\n"
                  "  game has '" << game.players().count_humans() << "' humans, must be '1'");

  game.announcements().make_announcement(Announcement::reply, game.players().human());
} // void GameInfoDialog::announcement_reply()

/** announce the hyperswines
 **/
void
GameInfoDialog::announce_hyperswines()
{
  auto& game = this->ui->game();

  DEBUG_ASSERTION(game.players().count_humans() == 1,
                  "GameInfoDialog::announce_hyperswines()\n"
                  "  game has '" << game.players().count_humans() << "' humans, must be '1'");

  game.swines().hyperswines_announce(game.players().human());
} // void GameInfoDialog::announce_hyperswines()

/** @return   whether the dialog is blocking the gameplay
 **/
bool
GameInfoDialog::is_blocking() const
{
  for (auto& i : this->infos)
    if (i->is_blocking())
      return true;

  return false;
} // bool GameInfoDialog::is_blocking() const

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
