/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "hand.h"
#include "table.h"
#include "card_suggestion.h"
#include "ui.h"

#include "cards.h"
#include "icongroup.h"
#include "trickpile.h"
#include "name.h"

#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../player/player.h"
#include "../../player/human/human.h"
#include "../../player/ai/cards_information.of_player.h"
#include "../../card/trick.h"
#include "../../misc/setting.h"

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table   the table
 ** @param    position   the corresponding position
 **/
Hand::Hand(Table* const table, Position const position) :
  HTIN(table, position),
  last_status_{false, false, false, nullptr, {}}
{ }

/** destructor
 **/
Hand::~Hand() = default;

/**
 **/
bool
Hand::changed() const
{
  if (!this->surface_)
    return true;
  if (::game_status & GameStatus::game)
    return (this->last_status_ != this->current_status());
  return true;
} // bool Hand::changed() const

/** draws the hand
 **
 ** @param    cr       drawing context
 **/
void
Hand::draw(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(   (::game_status >= GameStatus::game_init)
        && (::game_status <= GameStatus::game_finished)) )
    return ;

  if (   (::game_status == GameStatus::game_finished)
      && !this->table().in_game_review()) {
    this->draw_tricks(cr);
    return ;
  }

  Game const& game = this->game();

  cr->save();
  cr->push_group();

  this->last_status_ = this->current_status();
  auto const& show = this->last_status_.show_hand;
  auto const& show_ai_information_hand = this->last_status_.show_ai_information_hand;
  auto const& ai = this->last_status_.ai;
  auto const sorted_hand(!show_ai_information_hand
                         ? this->player().sorted_hand()
                         : SortedHand(this->last_status_.hand) );
  auto const geometry = this->geometry();
  geometry.transform(cr);

  // for showing the 'must have' cards
  Card last_card;
  unsigned last_card_no = 0;

  // draw the cards
  for (unsigned c = 0; c < sorted_hand.cardsnumber_all(); c++) {
    if (this->table().in_game_review())
      if (sorted_hand.played_trick(c) < this->table().game_review_trickno())
        continue;

    if ( (show || show_ai_information_hand
          ? !sorted_hand.played(c)
          : !sorted_hand.hand().played(c) )
        || (::game_status == GameStatus::game_finished) ) {

      auto card = (show || show_ai_information_hand
                   ? this->ui->cards->card(sorted_hand.card_all(c))
                   : this->ui->cards->back()
                  ).pixbuf();
      int pos_x = geometry.cards_step_x() * c;
      int pos_y = 0;
      if (last_card != sorted_hand.card_all(c)) {
        last_card_no = 0;
        last_card = sorted_hand.card_all(c);
      }
      if (show_ai_information_hand
          && !(last_card_no < ai->cards_information().of_player(this->player()).must_have(sorted_hand.card_all(c))) ) {
        Glib::RefPtr<Gdk::Pixbuf> copy = card->copy();
        card->saturate_and_pixelate(copy, 5, true);
        card = copy;
      } else {
        last_card_no += 1;
      }

      // emphasize the valid cards
      if (   ::setting(Setting::Type::emphasize_valid_cards)
          && (::game_status == GameStatus::game_play)
          && (game.players().current_player() == this->player())
          && (this->player().type() == Player::Type::human)) {
        if (!game.tricks().current().isvalid(sorted_hand.card_all(c),
                                          sorted_hand.hand())) {
          Glib::RefPtr<Gdk::Pixbuf> copy = card->copy();
          card->saturate_and_pixelate(copy, 0, true);
          card = copy;
        } // if (!card valid)
      } // if (emphasize valid cards)
      if (   this->table().card_suggestion_
          && (this->player() == game.players().current_player())
          && (c == sorted_hand.requested_position_all()) ) {
        pos_y -= geometry.skip_y();
      } // if (suggested card)

      draw_pixbuf(cr, card, pos_x, pos_y - card->get_height());
    } // if (!sorted_hand.played(c))
  } // for (c < sorted_hand.cardsnumber_all()
  cr->pop_group_to_source();
  cr->paint();
  cr->restore();
} // void Hand::draw(Cairo::RefPtr<::Cairo::Context> cr)

/** draw the tricks the player has got
 **
 ** @param    cr       cairo context
 **/
void
Hand::draw_tricks(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(   (::game_status >= GameStatus::game_init)
        && (::game_status <= GameStatus::game_finished)) )
    return ;

  auto const geometry = this->geometry();
  auto width = geometry.width();
  auto const& game = this->game();
  auto const& trickpile = this->player().trickpile();

  // gap between two tricks
  int const trick_gap = 3;
  // number of gaps
  int const gapsno
    = (trickpile.size() * (game.playerno() - 1 + trick_gap) - trick_gap);

  cr->save();
  cr->push_group();
  geometry.transform(cr);

  int const card_step_x
    = (  (gapsno == 0)
       ? 0
       : min(this->ui->cards->width() / 4,
             (width - this->ui->cards->width()) / gapsno)
      );
  int const trick_step_x
    = (  (trickpile.size() <= 1)
       ? 0
       : max(trick_gap * card_step_x,
             static_cast<int>((width - this->ui->cards->width()
                               - trickpile.size() * (game.playerno() - 1) * card_step_x
                              ) / (trickpile.size() - 1)))
      );
  for (size_t t = 0; t < trickpile.size(); ++t) {
    auto const& trick = trickpile.trick(t);
    for (size_t c = 0; c < trick.actcardno(); ++c) {
      int pos_x = (t * trick_step_x
                   + (t * (game.playerno() - 1) + c) * card_step_x);
      int pos_y = 0;
      if (trick.is_specialpoints_card(c)) {
        pos_y -= geometry.skip_y();
      } // if (trick.is_specialpoints_card(c))

      Glib::RefPtr<Gdk::Pixbuf>
        card = this->ui->cards->card(trick.card(c));
      if (c != trick.winnercardno()) {
        Glib::RefPtr<Gdk::Pixbuf> copy = card->copy();
        card->saturate_and_pixelate(copy, 0, true);
        card = copy;
      } // if (c != trick.winnercardno())

      draw_pixbuf(cr, card, pos_x, pos_y - card->get_height());
    } // for (c < trick.actcardno())
  } // for (t < trickpile.size())
  cr->pop_group_to_source();
  cr->paint();
  cr->restore();
} // void Hand::draw_tricks(Cairo::RefPtr<::Cairo::Context> cr)

/** -> result
 **
 ** @param    position   Position of the hand
 **
 ** @return   geometry of the hand at 'position'
 **/
Hand::Geometry
Hand::geometry(Position const position) const
{
  Geometry geometry;

  geometry.margin_x = [=](){
    return (this->ui->cards->height() / 10);
  };
  geometry.margin_y = [=](){
    return (this->ui->cards->height() / 10);
  };

  geometry.skip_y = [=](){
    return (geometry.margin_y() * 2 / 3);
  };

  switch (this->table().layout()) {
  case Table::Layout::standard:
    switch(position) {
    case Position::north: {
      auto const rotation = Rotation::down;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return (this->table().width()
                - this->table().border_x()
                - this->geometry(Position::east).height()
                - this->geometry(Position::east).margin_y()
                - this->icongroup(Position::east).geometry().height());
      };
      geometry.pos_y = [=](){
        return this->table().border_y();
      };
      geometry.width = [=](){
        return max(2 * this->ui->cards->width(),
                   this->table().width()
                   - this->table().border_x()
                   - this->geometry(Position::west).height()
                   - this->geometry(Position::west).margin_y()
                   - this->icongroup(Position::west).geometry().height()
                   - this->table().border_x()
                   - this->geometry(Position::east).height()
                   - this->geometry(Position::east).margin_y()
                   - this->icongroup(Position::east).geometry().height());
      };

      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::south: {
      auto const rotation = Rotation::up;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return (this->table().border_x()
                + this->geometry(Position::west).height()
                + this->geometry(Position::west).margin_y()
                + this->icongroup(Position::west).geometry().height());
      };
      geometry.pos_y = [=](){
        return (this->table().height()
                - this->table().border_y());
      };
      geometry.width = [=](){
        return max(2 * this->ui->cards->width(),
                   this->table().width()
                   - this->table().border_x()
                   - this->geometry(Position::west).height()
                   - this->geometry(Position::west).margin_y()
                   - this->icongroup(Position::west).geometry().height()
                   - this->table().border_x()
                   - this->geometry(Position::east).height()
                   - this->geometry(Position::east).margin_y()
                   - this->icongroup(Position::east).geometry().height());
      };

      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::west: {
      auto const rotation = Rotation::right;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return this->table().border_x();
      };
      geometry.pos_y = [=](){
        return (this->table().border_y()
                + this->geometry(Position::north).height()
                + this->geometry(Position::north).margin_y()
                + this->icongroup(Position::north).geometry().height());
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };
      geometry.width = [=](){
        return max(2 * this->ui->cards->width(),
                   this->table().height()
                   - this->table().border_y()
                   - this->geometry(Position::north).height()
                   - this->geometry(Position::north).margin_y()
                   - this->icongroup(Position::north).geometry().height()
                   - this->table().border_y()
                   - this->geometry(Position::south).height()
                   - this->geometry(Position::south).margin_y()
                   - this->icongroup(Position::south).geometry().height());
      };

      break;
    }

    case Position::east: {
      auto const rotation = Rotation::left;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return (this->table().width()
                - this->table().border_x());
      };
      geometry.pos_y = [=](){
        return (this->table().height()
                - this->table().border_y()
                - this->geometry(Position::south).height()
                - this->geometry(Position::south).margin_y()
                - this->icongroup(Position::south).geometry().height());
      };
      geometry.width = [=](){
        return max(2 * this->ui->cards->width(),
                   this->table().height()
                   - this->table().border_y()
                   - this->geometry(Position::north).height()
                   - this->geometry(Position::north).margin_y()
                   - this->icongroup(Position::north).geometry().height()
                   - this->table().border_y()
                   - this->geometry(Position::south).height()
                   - this->geometry(Position::south).margin_y()
                   - this->icongroup(Position::south).geometry().height());
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::center:
      DEBUG_ASSERTION(false,
                      "Hand::geometry(position = " << position << ")\n"
                      "  wrong position 'center'");
      break;
    } // switch(position)
    break;

  case Table::Layout::widescreen:
    switch(position) {
    case Position::north: {
      auto const rotation = Rotation::down;
      geometry.rotation = rotation;
      geometry.pos_x = [=](){
        auto const name_geometry = this->name().geometry();
        auto const name_left = (name_geometry.pos_x()
                                < this->table().border_x() + this->ui->cards->height());

        auto pos = max(this->icongroup(Position::east).geometry().pos_x()
                       - this->icongroup(Position::east).geometry().height()
                       - 2 * this->ui->cards->height(),
                       this->table().border_x()
                       + 3 * this->ui->cards->width());
        pos = min(pos,
                  (  name_left
                   ? name_geometry.pos_x() + name_geometry.width()
                   : this->table().border_x() + this->ui->cards->height() / 2)
                  + 35 * this->ui->cards->height() / 10);
        pos = max(pos, this->table().width() * 1 / 3);
        return pos;
      };
      geometry.pos_y = [=](){
        return this->table().border_y();
      };
      geometry.width = [=](){
        auto const name_geometry = this->name().geometry();
        auto const name_left = (name_geometry.pos_x()
                                < this->table().border_x() + this->ui->cards->height());

        if (name_left) {
          return (geometry.pos_x()
                  - (  name_geometry.pos_x()
                     + name_geometry.width()
                     + geometry.margin_x()));
        } else {
          return (geometry.pos_x() - this->table().border_x() - this->ui->cards->height() / 2);
        }
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::south: {
      auto const rotation = Rotation::up;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return (  this->table().border_x()
                + this->geometry(Position::west).height()
                + this->geometry(Position::west).margin_y()
                + this->icongroup(Position::west).geometry().height()
                + this->ui->cards->height()
                //+ this->trickpile(Position::west).geometry().height());
          );
      };
      geometry.pos_y = [=](){
        return (this->table().height()
                - this->table().border_y());
      };
      geometry.width = [=](){
        return max(2 * this->ui->cards->width(),
                   this->icongroup(Position::east).geometry().pos_x()
                   - this->icongroup(Position::east).geometry().height()
                   - geometry.pos_x()
                   - this->geometry(Position::east).margin_y());
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::west: {
      auto const rotation = Rotation::right;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return this->table().border_x();
      };
      geometry.pos_y = [=](){
        return (  this->icongroup(Position::north).geometry().pos_y()
                + this->icongroup(Position::north).geometry().height()
                + this->name().geometry().height()
                + 3 * geometry.margin_y());
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };
      geometry.width = [=](){
        return (this->table().height()
                - this->table().border_y()
                - geometry.pos_y());
      };

      break;
    }

    case Position::east: {
      auto const rotation = Rotation::left;
      geometry.rotation = rotation;

      geometry.pos_x = [=](){
        return (this->table().width()
                - this->table().border_x());
      };
      geometry.pos_y = [=](){
        return (this->table().height()
                - this->table().border_y()
                - this->geometry(Position::south).height()
                - this->geometry(Position::south).margin_y()
                - this->icongroup(Position::south).geometry().height());
      };
      geometry.width = [=](){
        return (geometry.pos_y()
                - this->table().border_y()
                - this->name().geometry().height()
                - geometry.margin_x());
      };
      geometry.height = [=](){
        return this->ui->cards->height();
      };

      break;
    }

    case Position::center:
      DEBUG_ASSERTION(false,
                      "Hand::geometry(position = " << position << ")\n"
                      "  wrong position 'center'");
      break;
    } // switch(position)
    break;
  } // switch (this->table().layout())

  geometry.cards_step_x = [=](){
    return (geometry.width() - this->ui->cards->width()) / (this->player().sorted_hand().cardsnumber_all() - 1.0);
  };

  return geometry;
} // Hand::Geometry Hand::geometry(Position const position) const

/** @return   geometry of the hand
 **/
Hand::Geometry
Hand::geometry() const
{
  return this->geometry(this->position());
} // Hand::Geometry Hand::geometry() const

/** @return   the outline of the hand
 **/
Hand::Outline
Hand::outline() const
{
  return this->geometry().outline();
} // Outline Hand::outline() const

/** -> result
 **
 ** @param    x   x position
 ** @param    y   y position
 **
 ** @return   the number of the card at the given position,
 **      UINT_MAX, if there is no card
 **
 ** @todo   check for transparency
 **/
optional<unsigned>
Hand::cardno_at_position(int x, int y) const
{
  // note: the position of the hand is complex to calculate
  // (see Hand::draw()), so since it does not take much time to get
  // through all cards I just check all cards.
  auto const geometry = this->geometry();
  geometry.retransform(x, y);
  for (unsigned cb = this->player().sorted_hand().cardsnumber_all();
       cb > 0;
       cb--) {
    unsigned const c = cb - 1;
    if (this->player().sorted_hand().played(c))
      continue;

    int pos_x = geometry.cards_step_x() * c;
    int pos_y = 0;

    // emphasize suggested card
    if (   this->table().card_suggestion_
        && (this->player() == this->game().players().current_player())
        && (c == this->player().sorted_hand().requested_position()) ) {
      pos_y -= geometry.skip_y();
    } // if (suggested card

    if ((x >= pos_x)
        && (x < (pos_x + this->ui->cards->width()))
        && (y < pos_y)
        && (y >= (pos_y - this->ui->cards->height()))
       ) {
      // the position is over the card
      // ToDo: check for transparency
      return c;
    }
  } // for (cb)

  return {};
} // optional<unsigned> Hand::cardno_at_position(int x, int y) const

/** -> result
 **
 ** @param    x   x position
 ** @param    y   y position
 **
 ** @return   the card at the given position
 **      Card(), if there is no card
 **/
Card
Hand::card_at_position(int const x, int const y) const
{
  auto const cardnox = this->cardno_at_position(x, y);

  if (!cardnox)
    return Card();

  auto const& cardno = *cardnox;

  return this->player().sorted_hand().card_all(cardno);
} // Card Hand::card_at_position(int x, int y) const

/** @return   the current status
 **/
Hand::Status
Hand::current_status() const
{
  bool const show_tricks = (  (::game_status == GameStatus::game_finished)
                            && !this->table().in_game_review());
  bool show_hand = ::setting(Setting::Type::show_all_hands);
  if (::game_status == GameStatus::game_finished) {
    show_hand = true;
  }

  auto const& game = this->game();
  auto const& player = this->player();
  if (   ::game_status == GameStatus::game_redistribute
      && player == game.players().soloplayer() ) {
    show_hand = true;
  }
  if (player.type() == Player::Type::human) {
    if (game.players().count_humans() == 1) {
      show_hand = true;
    } else {
      show_hand = (player == game.players().current_player());
    }
  }
#if !defined(RELEASE) && defined(DKNOF)
  if (  ::game_status != GameStatus::game_poverty_shift
      && this->position() == Position::south) {
    show_hand = true;
  }
#endif

  // the ai to show the cards information of
  Ai const* const ai
    = (  (game.players().count_humans() == 1)
#if !defined(RELEASE) && defined(DKNOF)
       && false
#endif
       ? dynamic_cast<Ai const*>(&game.players().human())
       : dynamic_cast<Ai const*>(&this->hand(Position::south).player()));
  bool const show_ai_information_hand
    = (   !show_hand
       && (::game_status >= GameStatus::game_play)
       && ::setting(Setting::Type::show_ai_information_hands)
       && (ai != nullptr));

  // the hand to draw
  auto const hand(!show_ai_information_hand
                  ? player.hand()
                  : ai->handofplayer(player));

  return {show_tricks, show_hand, show_ai_information_hand, ai, hand};
} // Hand::Status Hand::current_status() const

/** comparison operator
 **/
bool
operator!=(Hand::Status const& lhs, Hand::Status const& rhs)
{
  if (lhs.show_tricks != rhs.show_tricks)
    return true;
  if (lhs.show_tricks)
    return false;
  if (lhs.show_hand != rhs.show_hand)
    return true;
  if (lhs.show_hand == true) {
    if (lhs.show_ai_information_hand != rhs.show_ai_information_hand)
      return true;
  }
  return (lhs.hand != rhs.hand);
} // bool operator!=(Hand::Status lhs, Hand::Status rhs)

/** write the status in ostr
 **
 ** @param    ostr     output stream
 ** @param    status   status to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Hand::Status const& status)
{
  ostr << "show_tricks = " << status.show_tricks << ", "
    << "show_hand = " << status.show_hand << ", "
    << "show_ai_information_hand = " << status.show_ai_information_hand << ", "
    << "ai = " << (status.ai ? status.ai->name() : "--"s) << ", "
    << "hand = " << status.hand.cardsnumber();
  return ostr;
} // ostream& operator<<(ostream& ostr, Hand::Status const& status)

/** @return   the outline of the hand
 **/
Hand::Outline
Hand::Geometry::outline() const
{
  return this->transform({0, -this->skip_y() - this->height(),
                         this->width(), this->height() + this->skip_y()});
} // Outline Hand::Geometry::outline() const

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
