/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#ifdef USE_UI_GTKMM

#pragma once

#include "base.h"

#include "../../basetypes.h"

#include "../../misc/setting.theme.h"

class Party;

#include "widgets/file_chooser_dialog_wrapper.h"
#include <gtkmm/menubar.h>
#include <sigc++/connection.h>
namespace Gtk {
class Menu;
class MenuItem;
class CheckMenuItem;
class Label;
} // namespace Gtk

namespace UI_GTKMM_NS {
class MainWindow;
class Preferences;

/**
 ** the menu
 **
 ** party
 **   +- new party		C-n
 **   +- load party		C-l
 **   +- save party		C-s
 **   +- end party
 **   `- quit			C-q
 ** information
 **   +- card suggestion	h
 **   +- last trick		l
 **   +- rules		r
 **   +- players
 **   `- party points		p
 ** announcement
 **   `-  ...			n	m
 ** preferences
 **   +- show all hands
 **   +- show valid cards
 **   +- cards order
 **   `- preferences		F2
 ** bug report
 ** help
 **   +- help index		F1
 **   +- license
 **   +- cardset license
 **   +- changelog
 **   `- about
 ** 
 ** Note: the translation of the announcements are managed internally
 ** 
 **
 ** @todo   replace 'party_points_*_connection' (see sigc-doc)
 ** @todo   update the player names in the announcement
 **/
class Menu : public Base, public Gtk::MenuBar {
  friend class Preferences; // for changing the preferences menu items
  friend class UI_GTKMM; // for shortcuts

  public:
  explicit Menu(MainWindow* main_window);
  ~Menu() override;

  Menu() = delete;
  Menu(Menu const&) = delete;
  Menu& operator=(Menu const&) = delete;

  // set all signal
  void set_signals();
  // sets the setting 'show all hands'
  void show_all_hands_set();
  // sets the setting 'emphasize vaild cards'
  void emphasize_valid_cards_set();

  // a party is opened
  void party_open(Party& party);
  // a party is started
  void party_start();
  // a game is started
  void game_start();
  // a trick is opened
  void trick_open();
  // a card is requested
  void card_get();
  // a card has been played
  void card_got();
  // the game is finished
  void game_finish();
  // the game is closed
  void game_close();
  // the party is finished
  void party_finish();
  // the party is closed
  void party_close();

  // update whether the announcements are valid
  void announcements_update();
  private:
  // initialize
  void init();

  public:
  // an announcement shall be made
  void announcement_event(unsigned playerno, Announcement announcement);
  // a swines announcement shall be made
  void swines_announcement_event(unsigned playerno);
  // a hyperswines announcement shall be made
  void hyperswines_announcement_event(unsigned playerno);

  private:
  // the party shall be saved
  bool save_party_event(string filename);
  // a party shall be loaded
  bool load_party_event(string filename);

  // the theme list shall be updated
  void update_themes_sensitivity();
  // a theme as selected
  void apply_theme_event(Setting::Theme const* theme);

  private:
  AutoDisconnector disconnector_;

  MainWindow* main_window = nullptr;

  Gtk::Menu* party_menu = nullptr;
  Gtk::MenuItem* party = nullptr;
  Gtk::MenuItem* new_party = nullptr;
  Gtk::MenuItem* load_party = nullptr;
  Gtk::MenuItem* save_party = nullptr;
  Gtk::MenuItem* end_party = nullptr;
  Gtk::MenuItem* quit = nullptr;

  Gtk::Menu* information_menu = nullptr;
  Gtk::MenuItem* information = nullptr;
  Gtk::MenuItem* card_suggestion = nullptr;
  Gtk::MenuItem* last_trick = nullptr;
  Gtk::MenuItem* game_debug = nullptr;
  Gtk::MenuItem* bug_report_replay = nullptr;
  Gtk::MenuItem* rules = nullptr;
  Gtk::MenuItem* party_points = nullptr;
  Gtk::MenuItem* players = nullptr;
  Gtk::MenuItem* players_db = nullptr;

  Gtk::MenuItem* announcements = nullptr;
  vector<Gtk::MenuItem*> announcements_player;
  vector<Gtk::Menu*> announcements_player_menu;
  vector<vector<Gtk::MenuItem*> > announcements_announcement;
  vector<Gtk::MenuItem*> swines_separator;
  vector<Gtk::MenuItem*> swines_announcement;
  vector<Gtk::MenuItem*> hyperswines_announcement;

  Gtk::Menu* preferences_menu = nullptr;
  Gtk::MenuItem* preferences = nullptr;
  protected:
  Gtk::CheckMenuItem* emphasize_valid_cards = nullptr;
  Gtk::CheckMenuItem* show_all_hands = nullptr;
  private:
  Gtk::MenuItem* cards_order = nullptr;
  Gtk::Menu* themes_menu = nullptr;
  Gtk::MenuItem* themes_item = nullptr;
  vector<Gtk::MenuItem*> themes;
  Gtk::MenuItem* preferences_preferences = nullptr;

  Gtk::MenuItem* bug_report = nullptr;

  Gtk::Menu* help_menu = nullptr;
  Gtk::MenuItem* help = nullptr;
  Gtk::MenuItem* help_index = nullptr;
  Gtk::MenuItem* license = nullptr;
  Gtk::MenuItem* changelog = nullptr;
  Gtk::MenuItem* support = nullptr;
  Gtk::MenuItem* about = nullptr;
  Gtk::MenuItem* debug = nullptr;
  Gtk::MenuItem* generate_error = nullptr;

  sigc::connection party_points_present_connection;

  unique_ptr<Gtk::FileChooserDialogWrapper> load_party_file_chooser;
  unique_ptr<Gtk::FileChooserDialogWrapper> save_party_file_chooser;
}; // class Menu : public Base, public Gtk::MenuBar

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
