/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "name.h"
#include "table.h"
#include "hand.h"
#include "icongroup.h"

#include "ui.h"
#include "cards.h"

#include "../../player/player.h"
#include "../../party/rule.h"
#include "../../game/game.h"
#include "../../card/trick.h"

namespace UI_GTKMM_NS {

#define name_TRICKS_STEP_BIG_PART \
static_cast<int>(this->ui->cards->height() / 15.0)
#define name_TRICKS_STEP_SMALL_PART \
static_cast<int>(this->ui->cards->width() / 30.0)
#define name_CARDS_STEP_BIG_PART \
static_cast<int>(this->ui->cards->height() * 0)
#define name_CARDS_STEP_SMALL_PART \
static_cast<int>(this->ui->cards->width() / 20.0)

/** constructor
 **
 ** @param    table      the table
 ** @param    position   the position of the name
 **/
Name::Name(Table* const table, Position const position) :
  HTIN(table, position),
  last_status_{"", false, "", 0}
{ }

/** destructor
 **/
Name::~Name() = default;

/** returns the surface
 ** If the surface is outdated, redraw it.
 ** Update width and height of the name
 **
 ** @return   surface with the drawing in it
 **/
Cairo::RefPtr<Cairo::ImageSurface> 
Name::surface()
{
  this->HTIN::surface();

  auto cr = Cairo::Context::create(this->surface_);
  auto layout = Pango::Layout::create(cr);
  layout->set_font_description(Pango::FontDescription(::setting(Setting::Type::name_font)));
  layout->set_text(this->player().name());
#ifdef WORKAROUND
  // Bei kursiven Schriften wird bei den Grenzen von layout->get_pixel_size() rechts etwas abgeschnitten
  auto r = layout->get_pixel_ink_extents();
  this->width_ = max(0, r.get_x()) + r.get_width();
  this->height_ = max(0, r.get_y()) + r.get_height();
#else
  layout->get_pixel_size(this->width_, this->height_);
#endif
  if (   (this->width_ != this->surface_->get_width())
      || (this->height_ != this->surface_->get_height()) )
    this->surface_.clear();

  return this->HTIN::surface();
} // Cairo::RefPtr<Cairo::ImageSurface> Name::surface()


/** @return   whether the status has changed;
 **/
bool
Name::changed() const
{
  if (!this->surface_)
    return true;
  if (::game_status & GameStatus::game) {
    if (this->last_status_ != this->current_status())
      return (this->last_status_ != this->current_status());
  }
  return true;
} // bool Name::changed() const

/** @return   the geometry of the name
 **/
Name::Geometry
Name::geometry() const
{
  Geometry geometry;
  auto const hand_geometry = this->hand().geometry();
  geometry.width = [=](){
    return this->width_;
  };
  geometry.height = [=](){
    return this->height_;
  };
  switch (this->table().layout()) {
  case Table::Layout::standard:
    switch (this->position()) {
    case Position::south:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x(); 
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.height()
                - hand_geometry.margin_y());
      };
      break;

    case Position::north:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                - hand_geometry.width());
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                + hand_geometry.height()
                + hand_geometry.margin_y()
                + this->height_);
      };
      break;

    case Position::west:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x();
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.margin_y());
      };
      break;

    case Position::east:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                - hand_geometry.height());
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.width()
                - hand_geometry.margin_y());
      };
      break;

    case Position::center:
      DEBUG_ASSERTION(false,
                      "Name::geometry()\n"
                      "  wrong position 'center' "
                      << "(" << this->position() << ")");
    } // switch (this->position())
    break;
  case Table::Layout::widescreen:
    switch (this->position()) {
    case Position::south:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x(); 
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.height()
                - hand_geometry.margin_y());
      };
      break;

    case Position::north:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        // Wenn genug Platz ist, den Namen links vom Blatt setzen, ansonsten rechts
        if (this->width_ + hand_geometry.margin_x()
            < this->ui->cards->height()) {
          return max(this->table().border_x(),
                     this->ui->cards->height() / 2 - hand_geometry.margin_x() - this->width_);
        } else {
          return (hand_geometry.pos_x()
                  + hand_geometry.margin_x());
        }
      };
      geometry.pos_y = [=](){
        return (this->table().border_y()
                + this->height_);
        return (hand_geometry.pos_y()
                + hand_geometry.height()
                + hand_geometry.margin_y()
                + this->height_);
      };
      break;

    case Position::west:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x();
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.margin_y());
      };
      break;

    case Position::east:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                - hand_geometry.height());
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.width()
                - hand_geometry.margin_y());
      };
      break;

    case Position::center:
      DEBUG_ASSERTION(false,
                      "Name::geometry()\n"
                      "  wrong position 'center' "
                      << "(" << this->position() << ")");
    } // switch (this->position())
    break;
  } // switch (this->table().layout())
  return geometry;
} // Name::Geometry Name::geometry() const

/** @return   the outline of the name
 **/
Name::Outline
Name::outline() const
{
  return this->geometry().outline();
}

/** @return   the outline of the name
 **/
Name::Outline
Name::Geometry::outline() const
{
  return this->transform({0, -this->height(), this->width(), this->height()});
} // Outline Name::Geometry::outline() const

/** draws the name
 **
 ** @param    cr       cairo context
 **/
void
Name::draw(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(::game_status & GameStatus::game))
    return ;

  this->last_status_ = this->current_status();
  auto const is_active = this->last_status_.is_active;
  auto const color = Gdk::RGBA(this->last_status_.color);
  auto const progress = this->last_status_.progress;

  cr->save();
  cr->push_group();
  {
    cr->set_source_rgb(color.get_red(), color.get_green(), color.get_blue());
    auto layout = Pango::Layout::create(cr);
    layout->set_font_description(Pango::FontDescription(::setting(Setting::Type::name_font)));
    layout->set_text(this->player().name());
    layout->get_pixel_size(this->width_, this->height_);
    auto const geometry = this->geometry();
    geometry.transform(cr);
    cr->translate(0, -geometry.height());
    layout->show_in_cairo_context(cr);
    cr->save();
    if (   is_active 
        && (progress != 1) ) {
      {
        cr->rectangle(0, 0,
                      geometry.width() * this->ui->progress(),
                      geometry.height());
        cr->clip();
        auto color = Gdk::RGBA(::setting(Setting::Type::name_reservation_font_color));
        cr->set_source_rgb(color.get_red(), color.get_green(), color.get_blue());
        layout->show_in_cairo_context(cr);
      }
    } // if !(progress and active player)
    cr->restore();
  }

  cr->pop_group_to_source();
  cr->paint();
  cr->restore();
} // void Name::draw(Cairo::RefPtr<::Cairo::Context> cr)

/** @return   the current status
 **/
Name::Status
Name::current_status() const
{
  // test whether it is the active player
  bool is_active = false;
  switch (::game_status) {
  case GameStatus::game_reservation:
    // in the reservation show the current player
    if (this->game().players().current_player().type() == Player::Type::human) {
      // for the human show the startplayer
      if (   is_solo(this->game().type())
          && (this->game().rule()(Rule::Type::lustsolo_player_leads)))
        is_active = (this->player()
                     == this->game().players().current_player());
      else
        is_active = (this->player()
                     == this->game().startplayer());
    } else {
      // for all other players show the player to wait for
      is_active = (this->player()
                   == this->game().players().current_player());
    }
    break;
  case GameStatus::game_start:
  case GameStatus::game_init:
    is_active = (this->player()
                 == this->game().players().current_player());
    break;
  case GameStatus::game_poverty_shift:
    if (this->game().players().current_player().type() == Player::Type::human)
      is_active = (this->player()
                   == this->game().startplayer());
    else
      is_active = (this->player()
                   == this->game().players().current_player());

    break;
  case GameStatus::game_play:
  case GameStatus::game_full_trick:
  case GameStatus::game_trick_taken:
    is_active = (this->player() == this->game().players().current_player());
    break;
  case GameStatus::game_finished:
    if (this->table().in_game_review())
      is_active = (   this->table().game_review_trick_visible()
                   && (this->player()
                       == this->table().game_review_trick().winnerplayer()));
    else
      is_active = (   (this->player().team()
                       == this->game().winnerteam())
                   || (   (this->game().winnerteam()
                           == Team::noteam)
                       && (this->player().team() == Team::re)) );
#ifdef POSTPONED
    if (this->gui().playfield().table().last_trick().trickno()
        == UINT_MAX) {
      is_active = (this->player().team()
                   == this->player().game().winnerteam());
    } else { // if (show a specific trick)
      is_active = (this->player()
                   == this->player().game().tricks().trick(this->gui().playfield().table().last_trick().trickno()).winnerplayer());
    } // if (show a specific trick)
#endif
    break;

  default:
    break;
  } // switch (::game_status)

  auto color = ::setting(is_active
                         ? Setting::Type::name_active_font_color
                         : (   (::game_status == GameStatus::game_reservation)
                            && this->game().has_made_reservation(this->player())
                            && (this->game().reservation(this->player()).game_type
                                != GameType::normal) )
                         ? Setting::Type::name_reservation_font_color
                         : Setting::Type::name_font_color);

  return {this->player().name(),
    is_active,
    color,
    this->ui->progress()};
} // Name::Status Name::current_status() const

/** comparison operator
 **/
bool
operator!=(Name::Status const& lhs, Name::Status const& rhs)
{
  return (   (lhs.name != rhs.name)
          || (lhs.is_active != rhs.is_active)
          || (lhs.color != rhs.color)
          || (lhs.progress != rhs.progress) );
} // bool operator!=(Name::Status lhs, Name::Status rhs)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
