/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "players_db.h"
#include "ui.h"
#include "cards.h"

#include "../../party/party.h"
#include "../../player/player.h"
#include "../../player/playersDb.h"
#include "../../player/ai/aiDb.h"

#include "../../utils/string.h"


#include <gtkmm/treeview.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/scrolledwindow.h>
#include <gdk/gdkkeysyms.h>
namespace UI_GTKMM_NS {

constexpr PlayersDB::Statistic PlayersDB::statistic_all[];

/** -> result
 **
 ** @param    statistic   statistic
 **
 ** @return   string of 'statistic'
 **/
string
  PlayersDB::to_string(Statistic const statistic)
  {
    switch(statistic) {
    case Statistic::total:
      return "total";
      (void)_("Statistic::total");
    case Statistic::won:
      return "won";
      (void)_("Statistic::won");
    case Statistic::lost:
      return "lost";
      (void)_("Statistic::lost");
    case Statistic::percent_won:
      return "percent won";
      (void)_("Statistic::percent won");
    } // switch(statistic)

    return "";
  } // static string PlayersDB::to_string(Statistic statistic)

/** translation
 **
 ** @param    statistic   statistic
 **
 ** @return   translation
 **/
string
  PlayersDB::gettext(Statistic const statistic)
  {
    return ::gettext("Statistic::" + to_string(statistic));
  }

/** Constructor for the model
 **
 ** @param    playerno   number of players in the party
 **/
PlayersDB::PlayersDBModel::PlayersDBModel(unsigned const playerno) :
  Gtk::TreeModel::ColumnRecord(),
  type(),
  statistic(playerno)
{
  this->add(this->type);
  for (unsigned p = 0; p < playerno; p++)
    this->add(this->statistic[p]);
} // PlayersDB::PlayersDBModel::PlayersDBModel(unsigned playerno)

/** constructor
 **
 ** @param    parent   the parent object
 **/
PlayersDB::PlayersDB(Base* const parent) :
  Base(parent),
  StickyDialog("FreeDoko – " + _("Window::players database"), false)
{
  this->ui->add_window(*this);

  this->set_default_size(0, 3 * this->ui->cards->height());
  this->signal_realize().connect(sigc::mem_fun(*this, &PlayersDB::init));

#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_key_press_event().connect(sigc::mem_fun(*this, &PlayersDB::on_key_press_event));
#endif
} // PlayersDB::PlayersDB(Base* parent)

/** destructor
 **/
PlayersDB::~PlayersDB() = default;

/** create all subelements
 **/
void
PlayersDB::init()
{
  this->set_icon(this->ui->icon);

  { // hbox
    auto hbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_HORIZONTAL, 1 EM));
    { // statistic type
      auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));

      Gtk::RadioButton::Group statistic_group;
      for (auto statistic : statistic_all) {
        auto const radio_button
          = Gtk::manage(new Gtk::RadioButton(statistic_group,
                                             gettext(statistic)));
        vbox->add(*radio_button);
        radio_button->signal_clicked().connect(sigc::bind<Statistic const>(sigc::mem_fun(*this, &PlayersDB::set_statistic), statistic));
        if (statistic == Statistic::percent_won)
          radio_button->set_active(true);

      } // for (statistic)
      vbox->set_valign(Gtk::ALIGN_START);
      hbox->add(*vbox);
    } // statistic type

    { // the treeview

      // the points table
      this->players_db_model = std::make_unique<PlayersDBModel>(this->ui->party().players().size());

      this->players_db_list
        = Gtk::TreeStore::create(*this->players_db_model);
      this->players_db_treeview
        = Gtk::manage(new Gtk::TreeView(this->players_db_list));

      this->players_db_treeview->append_column(_("players database data type"),
                                               this->players_db_model->type);
      this->players_db_treeview->get_column_cell_renderer(0)->set_property("xalign", 1);
      for (unsigned p = 0; p < this->ui->party().players().size(); p++) {
        this->players_db_treeview->append_column(this->ui->party().players()[p].name(),
                                                 this->players_db_model->statistic[p]);
        this->players_db_treeview->get_column_cell_renderer(p + 1)->set_property("xalign", 0.5);
      } // for (p < this->ui->party().players().size())

      { // the scrolled window
        auto scrolled_window = Gtk::manage(new Gtk::ScrolledWindow);
        scrolled_window->set_policy(Gtk::POLICY_NEVER,
                                    Gtk::POLICY_AUTOMATIC);
        scrolled_window->add(*(this->players_db_treeview));

        hbox->pack_start(*scrolled_window);
      } // the scrolled window
    } // the treeview
    this->get_content_area()->pack_start(*hbox);
  } // hbox

  { // action area
    { // clear button
      auto clear_button = Gtk::manage(new Gtk::Button(_("Button::clear database")));
      clear_button->set_image_from_icon_name("edit-clear-all");
      clear_button->set_always_show_image();
      this->add_action_widget(*clear_button, Gtk::RESPONSE_NONE);
      clear_button->signal_clicked().connect(sigc::mem_fun(*this, &PlayersDB::clear_db));
    } // clear button

    add_close_button(*this);
  } // action area

  this->show_all_children();

  this->signal_show().connect(sigc::mem_fun(*this,
                                            &PlayersDB::recreate_db));
  this->recreate_db();

  Party::signal_open().connect_back([this](Party&) { this->recreate_db(); }, this->disconnector_);
  party->signal_close().connect_back(*this, &PlayersDB::hide, this->disconnector_);
} // void PlayersDB::init()

/** the name of 'player' has changed
 **
 ** @param    player   the player with the changed name
 **/
void
PlayersDB::name_changed(Player const& player)
{
  if (!this->get_realized())
    return ;

  this->players_db_treeview->get_column(this->ui->party().players().no(player) + 1
                                       )->set_title(" " + player.name() + " ");
} // void PlayersDB::name_changed(Player player)

/** recreates the table
 **/
void
PlayersDB::recreate_db()
{
  if (!this->get_realized())
    return ;

  if (!this->players_db_list)
    return ;

  this->players_db_list->clear();

  { // ranking
    this->ranking = *this->players_db_list->append();
    this->ranking[this->players_db_model->type]
      = _("ranking");
  } // ranking

  { // average game points
    this->average_game_points = *this->players_db_list->append();
    this->average_game_points[this->players_db_model->type]
      = _("average game points");
  } // average game points

  { // average game trick points
    this->average_game_trick_points = *this->players_db_list->append();
    this->average_game_trick_points[this->players_db_model->type]
      = _("average game trick points");
  } // average game trick points

  {
    for (unsigned p = 0; p < 4; ++p) {
      this->ranking[this->players_db_model->statistic[p]]
        = this->statistic_data(TLWcount());
    }
  }

  { // game type
    this->games = *this->players_db_list->append();
    this->games[this->players_db_model->type]
      = _("games");
    this->games_marriage
      = *(this->players_db_list->append(this->games.children()));
    this->games_marriage[this->players_db_model->type]
      = _("GameType::Group::marriage");
    this->games_poverty
      = *(this->players_db_list->append(this->games.children()));
    this->games_poverty[this->players_db_model->type]
      = _("GameType::Group::poverty");
    this->games_solo
      = *(this->players_db_list->append(this->games.children()));
    this->games_solo[this->players_db_model->type]
      = _("GameType::Group::solo");
    this->games_solo_color
      = *(this->players_db_list->append(this->games_solo.children()));
    this->games_solo_color[this->players_db_model->type]
      = _("GameType::Group::Solo::color");
    this->games_solo_picture
      = *(this->players_db_list->append(this->games_solo.children()));
    this->games_solo_picture[this->players_db_model->type]
      = _("GameType::Group::Solo::picture");
    this->games_solo_picture_single
      = *(this->players_db_list->append(this->games_solo_picture.children()));
    this->games_solo_picture_single[this->players_db_model->type]
      = _("GameType::Group::Solo::Picture::single");
    this->games_solo_picture_double
      = *(this->players_db_list->append(this->games_solo_picture.children()));
    this->games_solo_picture_double[this->players_db_model->type]
      = _("GameType::Group::Solo::Picture::double");

    this->game.clear();
    for (auto gt : game_type_list) {
      switch (gt) {
      case GameType::normal:
        this->game.push_back(*this->players_db_list->prepend(this->games.children()));
        break;
      case GameType::thrown_nines:
      case GameType::thrown_kings:
      case GameType::thrown_nines_and_kings:
      case GameType::thrown_richness:
      case GameType::fox_highest_trump:
      case GameType::redistribute:
      case GameType::poverty:
        this->game.push_back(*this->players_db_list->append(this->games_poverty.children()));
        break;
      case GameType::genscher:
      case GameType::marriage:
      case GameType::marriage_solo:
      case GameType::marriage_silent:
        this->game.push_back(*this->players_db_list->append(this->games_marriage.children()));
        break;
      case GameType::solo_jack:
      case GameType::solo_queen:
        this->game.push_back(*this->players_db_list->append(this->games_solo_picture_single.children()));
        break;
      case GameType::solo_king:
      case GameType::solo_queen_jack:
      case GameType::solo_king_jack:
      case GameType::solo_king_queen:
        this->game.push_back(*this->players_db_list->append(this->games_solo_picture_double.children()));
        break;
      case GameType::solo_koehler:
        this->game.push_back(*this->players_db_list->append(this->games_solo_picture.children()));
        break;
      case GameType::solo_club:
      case GameType::solo_heart:
      case GameType::solo_spade:
      case GameType::solo_diamond:
        this->game.push_back(*this->players_db_list->append(this->games_solo_color.children()));
        break;
      case GameType::solo_meatless:
        this->game.push_back(*this->players_db_list->append(this->games_solo.children()));
        break;
#if 0
      default:
        DEBUG_ASSERTION(false,
                        "UI_GTKMM::PlayersDB::recreate_db()\n"
                        "  game type '" << static_cast<GameType>(gt) << "' is not considered");
        break;
#endif
      } // switch(gt)
      this->game.back()[this->players_db_model->type] = _(gt);
    } // for (gt \in GameType)
  } // game type

  { // special points
    this->specialpoints = *this->players_db_list->append();
    this->specialpoints[this->players_db_model->type]
      = _("Specialpoint::specialpoints");
    this->specialpoints_winning
      = *this->players_db_list->append(this->specialpoints.children());
    this->specialpoints_winning[this->players_db_model->type]
      = _("Specialpoint::Group::winning");
    this->specialpoints_announcement
      = *this->players_db_list->append(this->specialpoints.children());
    this->specialpoints_announcement[this->players_db_model->type]
      = _("Specialpoint::Group::announcement");

    this->specialpoint.clear();
    for (auto sp : Specialpoint::types) {
      switch (sp) {
      case Specialpoint::Type::nospecialpoint:
      case Specialpoint::Type::caught_fox:
      case Specialpoint::Type::caught_fox_last_trick:
      case Specialpoint::Type::fox_last_trick:
      case Specialpoint::Type::charlie:
      case Specialpoint::Type::caught_charlie:
      case Specialpoint::Type::dolle_caught_dolle:
      case Specialpoint::Type::heart_trick:
      case Specialpoint::Type::doppelkopf:
        this->specialpoint[sp] = *this->players_db_list->prepend(this->specialpoints.children());
        break;
      case Specialpoint::Type::won:
      case Specialpoint::Type::no90:
      case Specialpoint::Type::no60:
      case Specialpoint::Type::no30:
      case Specialpoint::Type::no0:
      case Specialpoint::Type::contra_won:
      case Specialpoint::Type::solo:
      case Specialpoint::Type::bock:
        this->specialpoint[sp] = *this->players_db_list->append(this->specialpoints_winning.children());
        break;
      case Specialpoint::Type::no120_said:
      case Specialpoint::Type::no90_said:
      case Specialpoint::Type::no60_said:
      case Specialpoint::Type::no30_said:
      case Specialpoint::Type::no0_said:
      case Specialpoint::Type::no90_said_120_got:
      case Specialpoint::Type::no60_said_90_got:
      case Specialpoint::Type::no30_said_60_got:
      case Specialpoint::Type::no0_said_30_got:
      case Specialpoint::Type::no120_reply:
      case Specialpoint::Type::no90_reply:
      case Specialpoint::Type::no60_reply:
      case Specialpoint::Type::no30_reply:
      case Specialpoint::Type::no0_reply:
        this->specialpoint[sp] = *this->players_db_list->append(this->specialpoints_announcement.children());
        break;
      } // switch(sp)
      this->specialpoint[sp][this->players_db_model->type] = _(sp);
    } // for (sp \in Specialpoint)
  } // special points

  { // heuristics
    this->heuristics = *this->players_db_list->append();
    this->heuristics[this->players_db_model->type]
      = _("AiConfig::Group::heuristics");
    this->heuristics_general
      = *this->players_db_list->append(this->heuristics.children());
    this->heuristics_general[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::normal game");
    this->heuristics_poverty
      = *this->players_db_list->append(this->heuristics.children());
    this->heuristics_poverty[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::poverty");
    this->heuristics_solo
      = *this->players_db_list->append(this->heuristics.children());
    this->heuristics_solo[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::solo");
    this->heuristics_solo_color
      = *this->players_db_list->append(this->heuristics_solo.children());
    this->heuristics_solo_color[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::Solo::color");
    this->heuristics_solo_picture
      = *this->players_db_list->append(this->heuristics_solo.children());
    this->heuristics_solo_picture[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::Solo::picture");
    this->heuristics_solo_meatless
      = *this->players_db_list->append(this->heuristics_solo.children());
    this->heuristics_solo_meatless[this->players_db_model->type]
      = _("AiConfig::Heuristic::Group::Solo::meatless");

    this->heuristic.clear();
    for (auto const heuristic : aiconfig_heuristic_list) {
      switch (heuristic) {
      case Aiconfig::Heuristic::error:
        break;
      case Aiconfig::Heuristic::no_heuristic:
      case Aiconfig::Heuristic::manual:
      case Aiconfig::Heuristic::bug_report:
        this->heuristic.push_back(this->heuristics);
        break;
      case Aiconfig::Heuristic::only_valid_card:
        this->heuristic.push_back(*this->players_db_list->prepend(this->heuristics.children()));
        break;
      case Aiconfig::Heuristic::choose_best_card:
        this->heuristic.push_back(*this->players_db_list->prepend(this->heuristics.children()));
        std::swap(this->heuristic[static_cast<int>(Aiconfig::Heuristic::only_valid_card)],
                  this->heuristic.back());
        break;
      case Aiconfig::Heuristic::choose_ace:
      case Aiconfig::Heuristic::choose_ten:
      case Aiconfig::Heuristic::choose_single_ace:
      case Aiconfig::Heuristic::play_to_marry:
      case Aiconfig::Heuristic::play_to_get_married:
      case Aiconfig::Heuristic::choose_for_color_trick:
      case Aiconfig::Heuristic::jab_color_over_fox:
      case Aiconfig::Heuristic::start_with_color:
      case Aiconfig::Heuristic::start_with_low_color:
      case Aiconfig::Heuristic::start_with_low_trump:
      case Aiconfig::Heuristic::retry_color:
      case Aiconfig::Heuristic::play_color_for_partner:
      case Aiconfig::Heuristic::try_color_for_partner:
      case Aiconfig::Heuristic::play_color_for_partner_ace:
      case Aiconfig::Heuristic::play_bad_color:
      case Aiconfig::Heuristic::jab_color_trick_with_dolle:
      case Aiconfig::Heuristic::save_dolle:
      case Aiconfig::Heuristic::pfund_in_first_color_run:
      case Aiconfig::Heuristic::serve_color_trick:
      case Aiconfig::Heuristic::serve_trump_trick:
      case Aiconfig::Heuristic::choose_pfund:
      case Aiconfig::Heuristic::choose_pfund_before_partner:
      case Aiconfig::Heuristic::jab_for_ace:
      case Aiconfig::Heuristic::create_fehl:
      case Aiconfig::Heuristic::best_winning:
      case Aiconfig::Heuristic::low_high:
      case Aiconfig::Heuristic::play_trump:
      case Aiconfig::Heuristic::play_for_team:
      case Aiconfig::Heuristic::jab_fox:
      case Aiconfig::Heuristic::jab_for_doppelkopf:
      case Aiconfig::Heuristic::try_for_doppelkopf:
      case Aiconfig::Heuristic::play_for_partner_worries:
      case Aiconfig::Heuristic::draw_trump:
      case Aiconfig::Heuristic::play_to_jab_later:
      case Aiconfig::Heuristic::partner_backhand_draw_trump:
      case Aiconfig::Heuristic::play_highest_color_card_in_game:
      case Aiconfig::Heuristic::get_trick_for_announcement:
      case Aiconfig::Heuristic::grab_trick:
      case Aiconfig::Heuristic::let_trick_pass:
        this->heuristic.push_back(*this->players_db_list->append(this->heuristics_general.children()));
        break;
      case Aiconfig::Heuristic::color_jab_for_ace:
      case Aiconfig::Heuristic::color_low_high:
      case Aiconfig::Heuristic::play_color_in_solo:
        this->heuristic.push_back(*this->players_db_list->append(this->heuristics_solo_color.children()));
        break;
      case Aiconfig::Heuristic::picture_secondbesttrump:
        this->heuristic.push_back(*this->players_db_list->append(this->heuristics_solo_picture.children()));
        break;
      case Aiconfig::Heuristic::choose_pfund_poverty:
      case Aiconfig::Heuristic::poverty_special_play_pfund:
      case Aiconfig::Heuristic::poverty_special_give_no_points:
      case Aiconfig::Heuristic::poverty_special_offer_pfund:
      case Aiconfig::Heuristic::poverty_re_trump_color_trick_high:
      case Aiconfig::Heuristic::poverty_re_play_trump:
      case Aiconfig::Heuristic::poverty_contra_play_color:
      case Aiconfig::Heuristic::poverty_contra_trump_color_trick_high:
      case Aiconfig::Heuristic::poverty_leave_to_partner:
      case Aiconfig::Heuristic::poverty_overjab_re:
      case Aiconfig::Heuristic::poverty_best_winning_card:
        this->heuristic.push_back(*this->players_db_list->append(this->heuristics_poverty.children()));
        break;
      case Aiconfig::Heuristic::meatless_playhighestcolor:
        this->heuristic.push_back(*this->players_db_list->append(this->heuristics_solo_meatless.children()));
        break;
      } // switch(h)
      this->heuristic.back()[this->players_db_model->type]
        = _(heuristic);
    } // for (h \in Aiconfig::Heuristic)
  } // heuristics

  this->update_db();

  this->players_db_treeview->show_all();
} // void PlayersDB::recreate_db()

/** updates the statistics
 **/
void
PlayersDB::update_db()
{
  if (!this->get_realized())
    return ;

  if (!this->players_db_list)
    return ;

  Party const& party = this->ui->party();

  for (unsigned p = 0; p < party.players().size(); ++p) {
    auto const& player = this->ui->party().players()[p];
    auto const& db = player.db();

    this->ranking[this->players_db_model->statistic[p]]
      = String::to_string(static_cast<int>(db.rank.value() * 10) / 10.0);

    this->average_game_points[this->players_db_model->statistic[p]]
      = String::to_string(static_cast<int>(db.averageGamePoints() * 10) / 10.0);

    this->average_game_trick_points[this->players_db_model->statistic[p]]
      = String::to_string(static_cast<int>(db.averageGameTrickPoints() * 10) / 10.0);

    { // game type
      this->games[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_all());
      this->games_marriage[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_marriage());
      this->games_poverty[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_poverty());
      this->games_solo[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_solo());
      this->games_solo_color[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_solo_color());
      this->games_solo_picture[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_solo_picture());
      this->games_solo_picture_single[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_solo_picture_single());
      this->games_solo_picture_double[this->players_db_model->statistic[p]]
        = this->statistic_data(db.games_group_solo_picture_double());
      for (auto gt : game_type_list) {
        this->game[static_cast<int>(gt)][this->players_db_model->statistic[p]]
          = this->statistic_data(db.games(gt));
      } // for (gt \in gameTYPE)
    } // game type

    { // special points
      this->specialpoints[this->players_db_model->statistic[p]]
        = this->statistic_data(db.specialpoints_all());
      this->specialpoints_winning[this->players_db_model->statistic[p]]
        = this->statistic_data(db.specialpoints_group_winning());
      this->specialpoints_announcement[this->players_db_model->statistic[p]]
        = this->statistic_data(db.specialpoints_group_announcement());
      for (auto sp : Specialpoint::types) {
        this->specialpoint[sp][this->players_db_model->statistic[p]]
          = this->statistic_data(db.specialpoints(sp));
      } // for (sp : Specialpoint::types)
    } // special points
    { // heuristics
      if (dynamic_cast<AiDb const*>(&db) == nullptr)
        continue;
      auto const& aidb = dynamic_cast<AiDb const&>(db);
      auto const total = aidb.heuristic(Aiconfig::Heuristic::no_heuristic);

      this->heuristics[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, total);
      this->heuristics_general[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_general());
      this->heuristics_poverty[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_poverty());
      this->heuristics_solo[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_solo());
      this->heuristics_solo_color[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_solo_color());
      this->heuristics_solo_picture[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_solo_picture());
      this->heuristics_solo_meatless[this->players_db_model->statistic[p]]
        = this->heuristic_statistic_data(total, aidb.heuristics_group_solo_meatless());
      for (auto const heuristic : aiconfig_heuristic_list) {
        if (this->heuristic[static_cast<int>(heuristic)] == this->heuristics)
          continue;
        this->heuristic[static_cast<int>(heuristic)][this->players_db_model->statistic[p]]
          = this->heuristic_statistic_data(total, aidb.heuristic(heuristic));
      } // for (h \in Heuristic)
    } // heuristics
  } // for (p < party.players().size())
} // void PlayersDB::update_db()

/** sets the statistic
 **
 ** @param    statistic   new statistic
 **/
void
PlayersDB::set_statistic(Statistic const statistic)
{
  this->statistic = statistic;
  this->update_db();
} // void PlayersDB::set_statistic(Statistic const statistic)

/** clears the database
 **
 ** @todo    remove 'const cast'
 **/
void
PlayersDB::clear_db()
{
  for (auto& p : this->ui->party().players())
    p.db().clear();

  this->update_db();
} // void PlayersDB::clear_db()

/** -> result
 **
 ** @param    tlwcount   counter
 **
 ** @return   string of the data analysed by 'this->statistic'
 **/
Glib::ustring
PlayersDB::statistic_data(TLWcount const& tlwcount) const
{
  switch(this->statistic) {
  case Statistic::total:
    return String::to_string(tlwcount.total());
  case Statistic::won:
    if (tlwcount.total() > 0)
      return String::to_string(tlwcount.won());
    else
      return "-";
  case Statistic::lost:
    if (tlwcount.total() > 0)
      return String::to_string(tlwcount.lost());
    else
      return "-";
  case Statistic::percent_won:
    if (tlwcount.total() > 0)
      return (String::to_string(tlwcount.won() * 100
                                / tlwcount.total())
              + "%");
    else
      return "-";
  } // switch(this->statistic)

  return "-";
} // void PlayersDB::statistic_data(TLWcout tlwcount) const

/** -> result
 **
 ** @param    total   total number of heuristics
 ** @param    num     number of the active heuristics
 **
 ** @return   string of the data analysed by 'this->statistic'
 **/
Glib::ustring
PlayersDB::heuristic_statistic_data(unsigned const total,
                                    unsigned const num) const
{
  if (num == 0)
    return "-";

  switch(this->statistic) {
  case Statistic::total:
  case Statistic::won:
  case Statistic::lost:
    return String::to_string(num);
  case Statistic::percent_won:
    return (String::to_string((num * 100 / total))
            + "."
            + String::to_string((num * 1000 / total) % 10)
            + "%");
  } // switch(this->statistic)

  return "-";
} // void PlayersDB::heuristic_statistic_data(unsigned total, unsigned num) const

/** a key has been pressed
 ** C-o: output of the statistics on 'stdout'
 **
 ** @param    key   the key
 **
 ** @return   whether the key was managed
 **/
bool
PlayersDB::on_key_press_event(GdkEventKey* const key)
{
  bool managed = false;

  if ((key->state & ~GDK_SHIFT_MASK) == GDK_CONTROL_MASK) {
    switch (key->keyval) {
    case GDK_KEY_o: // ouput of the preferences
      for (auto& p : this->ui->party().players()) {
        cout << p.name() << '\n'
          << "{\n"
          <<   p.db()
          << "}\n";
      }
      managed = true;
      break;
    } // switch (key->keyval)
  } // if (key->state == GDK_CONTROL_MASK)

  return (managed
          || this->StickyDialog::on_key_press_event(key)
          || this->ui->key_press(key));
} // bool PlayersDB::on_key_press_event(GdkEventKey* key)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
