/********************************************************************** *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "thrower.h"

#include "../../card/card.h"

#ifdef DKNOF
//#define THROWER_INFO
#endif

namespace UI_GTKMM_NS {

/** constructor
 **/
Thrower::Thrower() = default;

/** @return   the exception type
 **/
Thrower::Exception
Thrower::exception() const
{
  return this->exception_;
} // Thrower::Exception Thrower::exception() const

/** @return   the exception type
 **/
Thrower::Type
Thrower::exception_type() const
{
  return this->exception_type_;
} // Thrower::Type Thrower::exception_type() const

/** @return   whether an exception will be thrown
 **/
Thrower::operator bool() const
{
  return this->will_throw();
} // Thrower::operator bool() const

/** @return   the depth
 **/
unsigned
Thrower::depth() const
{
  return this->depth_;
} // unsigned Thrower::depth() const

/** @return   whether an exception will be thrown
 **/
bool
Thrower::will_throw() const
{
  return (this->exception_type() != Type::none);
} // bool Thrower::will_throw() const

/** throws the delayed exception
 **/
void
Thrower::throw_it()
{
  if (this->exception_type() == Type::none)
    return ;

#ifdef THROWER_INFO
  cout << "thrower: throw it" << endl;
#endif

  // Since this function is left by throwing an exception,
  // I have to delete the type before the 'switch' (or in each 'case')
  // and I cannot use 'this->clear()' because then I do not have the
  // information afterwards.
  auto const type = this->exception_type();

  this->exception_type_ = Type::none;

  switch (type) {
  case Type::none:
    break;
  case Type::gamestatus:
    throw (this->exception_.game_status);
    break;
  case Type::card:
    {
      auto const card = *this->exception_.card;
      delete this->exception_.card;
      this->exception_.card = nullptr;

      throw (card);
    }
    break;
  }; // switch(type)
} // void Thrower::throw_it()

/** removes the exception to throw
 **/
void
Thrower::clear()
{
#ifdef THROWER_INFO
  cout << "thrower: clear" << endl;
#endif
  switch (this->exception_type()) {
  case Type::none:
  case Type::gamestatus:
    break;
  case Type::card:
    delete this->exception_.card;
    break;
  } // switch (this->exception_type())

  this->exception_type_ = Type::none;
} // void Thrower::clear()

/** increment the depth
 **/
void
Thrower::inc_depth()
{
  this->depth_ += 1;
} // void Thrower::inc_depth()

/** decrement the depth
 ** if depth '0' is reached, throw
 **/
void
Thrower::dec_depth()
{
  DEBUG_ASSERTION((this->depth() > 0),
                  "Thrower::dec_depth()\n"
                  "  depth is '0'");
  DEBUG_ASSERTION((this->depth() < 10000),
                  "Thrower::dec_depth()\n"
                  "  depth is too big: " << this->depth());

  this->depth_ -= 1;

  if (this->depth() == 0) {
    this->throw_it();
  }
} // void Thrower::dec_depth()

/** set the depth
 ** only use if you know exactly why you need it!
 **
 ** @param    depth   new depth
 **/
void
Thrower::set_depth(unsigned const depth)
{
  this->depth_ = depth;
} // void Thrower::set_depth(unsigned depth)

/** stores 'game_status' for throwing
 **
 ** @param    game_status   object to throw
 ** @param    file          the file of the throw
 ** @param    line          the line number of the throw
 **/
void
Thrower::operator()(GameStatus const game_status,
                    string const& file,
                    unsigned const line)
{
#ifdef THROWER_INFO
  cout << "thrower: throw game status '" << game_status << "'" << endl;
#endif
#ifdef DKNOF
  if (this->will_throw()
      && (   (this->exception_type() != Type::gamestatus)
          || (this->exception_.game_status != game_status)) )
    DEBUG_ASSERTION(!this->will_throw(),
                    "Thrower::operator()(GameStatus):\n"
                    "  will already throw '"
                    << static_cast<int>(this->exception_type()) << "'\n"
                    "  first throw: " << this->file << "::" << this->line
                    << '\n'
                    << "  new throw: " << file << "::" << line << '\n');
#endif

  this->exception_type_ = Type::gamestatus;
  this->exception_.game_status = game_status;

  this->file = file;
  this->line = line;
} // bool Thrower::operator()(GameStatus game_status, string file, unsigned line)

/** stores 'card' for throwing
 **
 ** @param    card   object to throw
 ** @param    file   the file of the throw
 ** @param    line   the line number of the throw
 **/
void
Thrower::operator()(Card const& card,
                    string const& file,
                    unsigned const line)
{
  // double throw (by clicking to fast)
  if (this->exception_type_ == Type::card)
    return ;

#ifdef THROWER_INFO
  cout << "thrower: throw card '" << card << "'" << endl;
#endif
  DEBUG_ASSERTION(!this->will_throw(),
                  "Thrower::operator()(Card):\n"
                  "  will already throw '"
                  << static_cast<int>(this->exception_type()) << "'\n"
                  "  first throw: " << this->file << "::" << this->line
                  << '\n'
                  << "  new throw: " << file << "::" << line << '\n');

  this->exception_type_ = Type::card;
  this->exception_.card = new Card(card);

  this->file = file;
  this->line = line;
} // bool Thrower::operator()(Card card, string file, unsigned line)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
