/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "trickpile.h"
#include "table.h"
#include "hand.h"
#include "icongroup.h"
#include "name.h"

#include "ui.h"
#include "cards.h"

#include "../../player/player.h"
#include "../../party/party.h"
#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../card/trick.h"
#include "../../misc/setting.h"

#include "../../utils/string.h"

namespace UI_GTKMM_NS {

/** Constructor
 **
 ** @param    table   the table
 ** @param    position   the position of the trickpile
 **/
TrickPile::TrickPile(Table* const table, Position const position) :
  HTIN(table, position)
{ }

/** destructor
 **/
TrickPile::~TrickPile() = default;

/**
 **/
bool
TrickPile::changed() const
{
  if (!this->surface_)
    return true;
  if (::game_status & GameStatus::game)
    return (this->last_status_ != this->current_status());
  return true;
} // bool TrickPile::changed() const

/** draws the trickpile
 **
 ** @param    cr       cairo context
 **/
void
TrickPile::draw(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(::game_status & GameStatus::game))
    return ;

  if (!(   (::game_status >= GameStatus::game_init)
        && (::game_status <= GameStatus::game_finished)))
    return ;

  this->last_status_ = this->current_status();

  this->draw_tricks(cr);
  this->draw_points(cr);
} // void TrickPile::draw(Cairo::RefPtr<::Cairo::Context> cr)

/** draws the tricks in the trickpile
 **
 ** @param    cr    cairo context
 **
 ** @todo   skip the cards, the human player(s) do know not to make a special points (because of further team informations)
 **/
void
TrickPile::draw_tricks(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(::game_status & GameStatus::game))
    return ;

  if (   (::game_status == GameStatus::game_finished)
      && !this->table().in_game_review())
    return ;

  cr->save();
  cr->push_group();
  auto const geometry = this->geometry();
  geometry.transform(cr);

  unsigned trickno = geometry.number_of_tricks();

  if (this->table().in_game_review()) {
    trickno = 0;
    for (auto const& t : this->player().trickpile()) {
      if (t.no() + 1 < this->table().game_review_trickno())
        trickno += 1;
      else
        break;
    } // for (auto t : this->player.trickpile())
  } // if (this->table().in_game_review())

  for (unsigned t = 0; t < trickno; ++t) {
#ifdef TEST_LAYOUT
    vector<HandCard> cards;
#else
    auto cards = this->player().trickpile().trick(t).specialpoints_cards();
#endif
    auto const specialpoints_no = cards.size();
    // ToDo: skip the cards, the human player(s) do know not to make a special points (because of further team informations)

    while (cards.size() < this->game().playerno())
      cards.push_back(HandCard());
    for (unsigned c = 0; c < cards.size(); ++c) {
      auto const& card = cards[c];
      int x = (  geometry.tricks_step_x() * t
               + geometry.cards_step_x() * c);
      int y = (  geometry.tricks_step_y() * t
               + geometry.cards_step_y() * c);
      if (card.is_empty()) {
        x -= (specialpoints_no * geometry.cards_step_x());
        y -= (specialpoints_no * geometry.cards_step_y());
      } // if !(card)
      if (card.is_empty()) {
        draw_pixbuf(cr, this->ui->cards->card(card, Rotation::left),
                    x, y - this->ui->cards->width());
      } else {
        draw_pixbuf(cr, this->ui->cards->card(card),
                    x, y - this->ui->cards->height());
      }
    } // for (c < cards.size())
  } // for (t < this->player().number_of_tricks())
  cr->pop_group_to_source();
  cr->paint();
  cr->restore();
} // void TrickPile::draw_tricks(Cairo::RefPtr<::Cairo::Context> cr))

/** draws the points in the trickpile
 **
 ** @param    cr       cairo context
 **/
void
TrickPile::draw_points(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(::game_status & GameStatus::game))
    return ;
  if (!(   ::setting(Setting::Type::show_trickpiles_points)
        || (::game_status == GameStatus::game_finished)))
    return ;

  auto const geometry = this->geometry();

  auto number_of_tricks = this->player().number_of_tricks();
  if (number_of_tricks == 0)
    return ;

  auto points_layout = Pango::Layout::create(cr);

  if (::game_status == GameStatus::game_finished) {
    auto points = this->player().points_in_trickpile();
    if (this->table().in_game_review()) {
      number_of_tricks = 0;
      points = 0;

      for (auto const& t : this->player().trickpile()) {
        if (t.no() + 1 < this->table().game_review_trickno()) {
          points += t.points();
          number_of_tricks += 1;
        } else {
          break;
        }
      } // for (auto t : this->player().trickpile())
      if (number_of_tricks == 0)
        return ;
    } // if (this->table().in_game_review())

    points_layout->set_text(String::to_string(points));
  } else { // if !(::game_status == GameStatus::game_finished)
    points_layout->set_text(String::to_string(this->player().points_in_trickpile()));
  } // if !(::game_status == GameStatus::game_finished)
  points_layout->set_font_description(Pango::FontDescription(::setting(Setting::Type::trickpile_points_font)));

  auto number_of_cards
    = max(1,
          static_cast<int>(this->game().playerno())
          - static_cast<int>(this->player().trickpile().last_trick().specialpoints_cards().size()));

  int layout_width, layout_height;
  points_layout->get_pixel_size(layout_width, layout_height);
  int x = (  (geometry.tricks_step_x()
              * (number_of_tricks - 1))
           + (geometry.cards_step_x()
              * (number_of_cards - 1))
           + this->ui->cards->height() / 2
          );
  int y = (  (geometry.tricks_step_y()
              * (number_of_tricks - 1))
           + (geometry.cards_step_y()
              * (number_of_cards - 1))
           - this->ui->cards->width() / 2
          );

  Gdk::RGBA color;
  if (   (::game_status == GameStatus::game_finished)
      && !this->table().in_game_review()
     ) {
    if (   (this->player().team()
            == this->game().winnerteam())
        || (   (this->game().winnerteam() == Team::noteam)
            && (this->player().team() == Team::re)) )
      color = Gdk::RGBA(::setting(Setting::Type::name_font_color));
    else
      color = Gdk::RGBA(::setting(Setting::Type::name_active_font_color));
#if 0
    Gdk::Rectangle r(x - layout_width / 2,
                     y - layout_height / 2,
                     2 * layout_width, 2 * layout_height);
    gc->set_clip_rectangle(r);
#endif
  } else {
    color = Gdk::RGBA(::setting(Setting::Type::trickpile_points_font_color));
  } // if ()

  cr->save();
  cr->push_group();
  geometry.transform(cr);
  cr->move_to(x, y);
  cr->set_identity_matrix();
  cr->rel_move_to(-layout_width / 2, -layout_height / 2);
  cr->set_source_rgb(color.get_red(), color.get_green(), color.get_blue());
  points_layout->show_in_cairo_context(cr);
  cr->pop_group_to_source();
  cr->paint();
  cr->restore();
} // void TrickPile::draw_points(Cairo::RefPtr<::Cairo::Context> cr)

/** @return   the outline of the trick
 **/
TrickPile::Outline
TrickPile::outline() const
{
  return this->geometry().outline();
} // Outline TrickPile::outline() const

/** @return   the geometry of the name
 **/
TrickPile::Geometry
TrickPile::geometry() const
{
  Geometry geometry;
  auto const hand_geometry = this->hand().geometry();
  auto const icongroup_geometry = this->icongroup().geometry();
  geometry.number_of_tricks = [=](){
#ifdef TEST_LAYOUT
    return 12;
#else
    return this->player().number_of_tricks();
#endif
  };
  geometry.tricks_step_x = [=](){
    return (this->ui->cards->height() / 15);
  };
  geometry.tricks_step_y = [=](){
    return -(this->ui->cards->height() / 30);
  };
  geometry.cards_step_x = [=](){
    return (this->ui->cards->height() * 0);
  };
  geometry.cards_step_y = [=](){
    return -(this->ui->cards->height() / 20);
  };
  geometry.width = [=](){
    return (max(this->ui->cards->height(), this->ui->cards->width())
            + ((geometry.number_of_tricks() - 1)
               * std::abs(geometry.tricks_step_x()))
            + ((this->player().game().playerno() - 1)
               * std::abs(geometry.cards_step_x()))
           );
  };
  geometry.height = [=](){
    return (max(this->ui->cards->height(), this->ui->cards->width())
            + ((geometry.number_of_tricks() - 1)
               * std::abs(geometry.tricks_step_y()))
            + ((this->player().game().playerno() - 1)
               * std::abs(geometry.cards_step_y()))
           );
  };

  switch (this->table().layout()) {
  case Table::Layout::standard:
    switch (this->position()) {
    case Position::south:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                + hand_geometry.width()
                + hand_geometry.margin_x()); 
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y());
      };
      break;

    case Position::north:
      geometry.rotation = Rotation::down;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                - hand_geometry.width()
                - hand_geometry.margin_x());
      };
      geometry.pos_y = [=](){
        return hand_geometry.pos_y();
      };
      break;

    case Position::west:
      geometry.rotation = Rotation::right;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x();
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                + hand_geometry.width()
                + hand_geometry.margin_x());
      };
      break;

    case Position::east:
      geometry.rotation = Rotation::left;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x();
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                - hand_geometry.width()
                - hand_geometry.margin_x()
                - this->name().geometry().height()
                - hand_geometry.margin_x());
      };
      break;

    case Position::center:
      DEBUG_ASSERTION(false,
                      "TrickPile::geometry()\n"
                      "  wrong position 'center' "
                      << "(" << this->position() << ")");
    } // switch (this->position())
    break;

  case Table::Layout::widescreen:
    switch (this->position()) {
    case Position::south:
      geometry.rotation = Rotation::up;
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                + hand_geometry.width()
                + hand_geometry.margin_x()
                ); 
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y());
      };
      break;

    case Position::north:
      geometry.rotation = Rotation::down;
      geometry.pos_x = [=](){
        return hand_geometry.pos_x();
        return (this->table().border_x()
                + hand_geometry.width() / 2
                + this->ui->cards->height() / 2);
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                + hand_geometry.height()
                + hand_geometry.margin_y()
                + (  (   this->icongroup().current_status().swines
                      || this->icongroup().current_status().hyperswines)
                   ? icongroup_geometry.height()
                   : 0)
               );
      };
      break;

    case Position::west:
      geometry.rotation = Rotation::right;
      geometry.tricks_step_x = [=](){
        return -(this->ui->cards->height() / 15);
      };
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                + hand_geometry.height()
                + hand_geometry.margin_y()
                + (  (   this->icongroup().current_status().swines
                      || this->icongroup().current_status().hyperswines)
                   ? icongroup_geometry.height()
                   : 0));
      };
      geometry.pos_y = [=](){
        return (hand_geometry.pos_y()
                + hand_geometry.width()
                - this->ui->cards->height());
      };
      break;

    case Position::east:
      geometry.rotation = Rotation::left;
      geometry.tricks_step_x = [=](){
        return -(this->ui->cards->height() / 15);
      };
      geometry.pos_x = [=](){
        return (hand_geometry.pos_x()
                - hand_geometry.height()
                - hand_geometry.margin_y()
                - (  !(   this->icongroup().current_status().swines
                      || this->icongroup().current_status().hyperswines)
                   ? 0
                   : icongroup_geometry.height() + hand_geometry.margin_y()));
      };
      geometry.pos_y = [=](){
        return (this->table().border_y() + this->ui->cards->height());
      };
      break;

    case Position::center:
      DEBUG_ASSERTION(false,
                      "TrickPile::geometry()\n"
                      "  wrong position 'center' "
                      << "(" << this->position() << ")");
    } // switch (this->position())
    break;
  } // switch (this->table().layout())
  return geometry;
} // TrickPile::Geometry Name::geometry() const

/** @return   the outline of the name
 **/
TrickPile::Outline
TrickPile::Geometry::outline() const
{
  return this->transform({0 + min(0, this->tricks_step_x()
                                  * (this->number_of_tricks() - 1)),
                         -this->height(),
                         this->width(), this->height()});
} // Outline TrickPile::Geometry::outline() const

/** -> result
 **
 ** @param    x   x position
 ** @param    y   y position
 **
 ** @return   the pointer to the trick under the mouse curser
 **           nullptr if the pointer is over no trick
 **           or the trick cannot be shown (-> rules)
 **
 ** @todo     transparency
 **/
::Trick const*
TrickPile::over_trick(int x, int y) const
{
  if (!::in_running_game())
    return {};

  auto const& trickpile = this->player().trickpile();

  int const c = this->game().playerno() - 1;
  auto geometry = this->geometry();
  geometry.retransform(x, y);
  if (   (x < 0)
      || (y > 0)
      || (x >= geometry.width())
      || (y <= -geometry.height()) )
    return {};

  auto const cards_step_x = geometry.cards_step_x();
  auto const cards_step_y = geometry.cards_step_y();
  auto const tricks_step_x = geometry.tricks_step_x();
  auto const tricks_step_y =  geometry.tricks_step_y();
  for (int t = trickpile.size() - 1; t >= 0; --t) {
    int const pos_x = (  t * tricks_step_x
                       + c * cards_step_x);

    int const pos_y = (  t * tricks_step_y
                       + c * cards_step_y);

    // ToDo: check special points (open cards)
    if (   (x >= pos_x)
        && (x < (pos_x + this->ui->cards->height()))
        && (y <= pos_y)
        && (y > (pos_y - this->ui->cards->width()))
       ) {
      return &trickpile.trick(t);
    }
  } // for (t)

  return {};
} // ::Trick const* TrickPile::over_trick(int x, int y)

/** @return   the current status
 **/
TrickPile::Status
TrickPile::current_status() const
{
  return {this->player().number_of_tricks(),
    this->game().teaminfo().get_all()};
} // TrickPile::Status TrickPile::current_status() const

/** comparison operator
 **/
bool
operator!=(TrickPile::Status const& lhs, TrickPile::Status const& rhs)
{
  if (lhs.size != rhs.size)
    return true;
  if (lhs.teaminfo.size() != rhs.teaminfo.size())
    return true;
  for (size_t i = 0; i < lhs.teaminfo.size(); ++i)
    if (lhs.teaminfo[i] != rhs.teaminfo[i])
      return true;
  return false;
} // bool operator!=(TrickPile::Status lhs, TrickPile::Status rhs)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
