/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "help.h"

#include "ui.h"
#include "../utils/windows.h"
#include "../utils/string.h"
#include "../utils/file.h"

/** show the help page of the given location
 **
 ** @param    location   location of the page to show
 **/
void
Help::show(string const& location)
{
#ifdef WINDOWS
  STARTUPINFO si = {sizeof(si)};
  PROCESS_INFORMATION pi;
  CreateProcess(NULL,
                const_cast<CHAR*>((::setting(Setting::Type::browser_command) + " " + location).c_str()),
                NULL, NULL, FALSE, 0, NULL, NULL,
                &si, &pi);
#else
  int const a =
    ::system((::setting(Setting::Type::browser_command) + " " + location
              + " &").c_str());
  (void) a;
#endif

  return ;
} // void Help::show(string const location)

/** @return   the location of the homepage
 **/
string
Help::homepage_location()
{
  vector<string> const languages = {"en", "de"};

  if (!std::contains(languages, ::setting(Setting::Type::language)))
    return ("http://free-doko.sourceforge.net/en/");

  return ("http://free-doko.sourceforge.net/"
          + ::setting(Setting::Type::language) + "/");
} // string Help::homepage_location()

/** @return   the location of the sourceforge page to download cardsets
 **/
string
Help::cardsets_download_location()
{
  return "http://sourceforge.net/project/showfiles.php?group_id=38051&package_id=30877";
} // string Help::cardsets_download_location()

/** -> result
 **
 ** @param    page   help page
 **                   default: index
 **
 ** @return   the location with the manual help in page 'page'
 **
 ** @bug       using the function 'realpath'
 **/
string
Help::manual_location(string const& page)
{
  // Search through a list of local directories, if that does not help,
  // return the internet address of the manual.

  // the directories to search
  vector<string> directories;

  { // fill the directories list
    if (getenv("FREEDOKO_MANUAL_DIRECTORY")) {
      directories.push_back(getenv("FREEDOKO_MANUAL_DIRECTORY"));
      directories.push_back(getenv("FREEDOKO_MANUAL_DIRECTORY")
                            + ("/" + ::setting(Setting::Type::language)));
    }
#ifdef WINDOWS
    {
      string dir = Reg_read(HKEY_LOCAL_MACHINE,
                            "Software\\FreeDoko",
                            "Help Directory");
      if (!dir.empty())
        directories.push_back(dir + "/" + ::setting(Setting::Type::language));
    }
#endif
    vector<string> const datadir = ::setting.data_directories();
    for (auto const& d : datadir)
      directories.push_back((::setting(Setting::Type::manual_directory)[0] == '/'
                            ? "" : d + "/")
                            + ::setting(Setting::Type::manual_directory) + "/"
                            + ::setting(Setting::Type::language));
  } // fill the directories list


  { // search the manual page
    for (auto const& d : directories)
      if (File::isfile(d + "/" + page))
        return ("file://" + File::absolute_path(d + "/" + page));
  } // search the manual page

  return ("http://free-doko.sourceforge.net/doc/manual/"
          + ::setting(Setting::Type::language) + "/" + page);
} // string Help::manual_location(string page)

/** -> result
 **
 ** @param    page   help page
 **
 ** @return   the location with the operation help in page 'page'
 **/
string
Help::operation_location(string const& page)
{
  return manual_location("operation/" + page);
} // string Help::operation_location(string page)

/** -> result
 **
 ** @param    type   rule type to show
 **
 ** @return   the location of the help for the rule 'type'
 **/
string
Help::location(Rule::Type::Bool const type)
{
  return operation_location("rules_config.html"
                            + String::replaced_all(to_string(type),
                                                             " ", "_"));
} // string Help::location(Rule::Type::Bool type)

/** -> result
 **
 ** @param    type   rule type to show
 **
 ** @return   the location of the help for the rule 'type'
 **/
string
Help::location(Rule::Type::Unsigned const type)
{
  return operation_location("rules_config.html"
                            + String::replaced_all(to_string(type), " ", "_"));
} // string Help::location(Rule::Type::Unsigned type)

/** -> result
 **
 ** @param    type   setting type to show
 **
 ** @return   the location of the help for the setting 'type'
 **/
string
Help::location(Setting::Type::Bool const type)
{
  return operation_location("settings.html"
                            + String::replaced_all(to_string(type), " ", "_"));
} // string Help::location(Setting::Type::Bool type)

/** -> result
 **
 ** @param    type   setting type to show
 **
 ** @return   the location of the help for the setting 'type'
 **/
string
Help::location(Setting::Type::Unsigned const type)
{
  return operation_location("settings.html"
                            + String::replaced_all(to_string(type), " ", "_"));
} // string Help::location(Setting::Type::Unsigned type)

/** -> result
 **
 ** @param    type   setting type to show
 **
 ** @return   the location of the help for the setting 'type'
 **/
string
Help::location(Setting::Type::String const type)
{
  return operation_location("settings.html"
                            + String::replaced_all(to_string(type), " ", "_"));
} // string Help::location(Setting::Type::String type)

/** -> result
 **
 ** @param    type   setting type to show
 **
 ** @return   the location of the cards order help
 **/
string
Help::location(Setting::Type::CardsOrder const type)
{
  return operation_location("settings.html"
                            + String::replaced_all(to_string(type), " ", "_"));
} // string Help::location(Setting::Type::CardsOrder type)
