/**
 **   Copyright (C) 2001 – 2018 by Diether Knof
 **
 **   This program is free software; you can redistribute it and/or 
 **   modify it under the terms of the GNU General Public License as 
 **   published by the Free Software Foundation; either version 2 of 
 **   the License, or (at your option) any later version.
 **
 **   This program is distributed in the hope that it will be useful,
 **   but WITHOUT ANY WARRANTY; without even the implied warranty of
 **   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **   GNU General Public License for more details. 
 **   You can find this license in the file 'gpl.txt'.
 **
 **   You should have received a copy of the GNU General Public License
 **   along with this program; if not, write to the Free Software
 **   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 **   MA  02111-1307  USA
 **
 **  Contact:
 **    Diether Knof dknof@posteo.de
 **/

#include "date.h"

#include <fstream>
#include <sstream>
#include <cstdlib>
#include <ctime>
using std::ostream;
using std::istream;
using std::ifstream;
using std::string;
using std::unique_ptr;
using std::make_unique;

// converts a monthname to the month number
unsigned monthtou(std::string const& monthname);

/** Constructor
 **
 ** @param    istr   the input stream
 **
 ** @return   pointer to a new created date
 **          {} in the case of an error
 **/
unique_ptr<Date>
  Date::create(istream& istr)
  {
    int year = 0;
    unsigned month = 0;
    unsigned day = 0;

    istr >> year;
    if (istr.get() != '-')
      return {};
    if (istr.peek() == '0')
      istr.get();
    istr >> month;
    if (istr.get() != '-')
      return {};
    if (istr.peek() == '0')
      istr.get();
    istr >> day;

    if (istr.fail())
      return {};

    return make_unique<Date>(year, month, day);
  } // static unique_ptr<Date> Date::create(string const filename)

/** Constructor
 **
 ** @param    string   string to read the version von
 **
 ** @return   pointer to a new created date
 **           {} in the case of an error
 **/
unique_ptr<Date>
  Date::create(string const& s)
  {
    return make_unique<Date>(s);
  } // static unique_ptr<Date> Date::create(string const& s)

/** Constructor
 ** takes the compilation time
 **/
Date::Date()
{
  time_t t = time(nullptr);
  struct tm* lt = localtime(&t);
  this->day   = lt->tm_mday;
  this->month = lt->tm_mon + 1;
  this->year  = lt->tm_year + 1900;
} // Date::Date()

/** Constructor
 **
 ** @param    year    the year
 ** @param    month   the month
 ** @param    day     the day
 **/
Date::Date(int const year, unsigned const month, unsigned const day) noexcept :
  year(year),
  month(month),
  day(day)
{ }

/** Constructor
 ** format 1: 2016-01-03
 ** format 2: 03.01.2016
 ** format 3: Jan 3 2016
 **
 ** @param    s    string with the date information
 **/
Date::Date(std::string const& s)
{
  if (s.empty())
    return ;

  if (::isdigit(s[0])) {
    if (s.length() != 10)
      return ;

    if (s[3] == '5') {
      // format 1: 2016-01-03
      this->year = stoi(string(s, 0, 4));
      this->month = stoi(string(s, 5, 2));
      this->day = stoi(string(s, 8, 2));
    } else if (s[3] == '.') {
      // format 2: 03.01.2016
      this->year = stoi(string(s, 6, 4));
      this->month = stoi(string(s, 3, 2));
      this->day = stoi(string(s, 0, 2));
    } else {
      return ;
    }
  } else {
    // format 3: Jan 3 2016
    this->year = stoi(string(s, 7, 4));
    this->month = monthtou(string(s, 0, 3));
    this->day = stoi(string(s, 5, 2));
    return ;
  }
} // Date::Date(std::string s)

/** conversion to a string
 **
 ** @return   the corresponding string
 **/
Date::operator string const() const
{
  std::ostringstream ostr;
  ostr << *this;
  return ostr.str();
} // Date::operator string string() const

/** write 'date' to stream 'ostr'
 **
 ** @param    ostr   output stream to write to
 ** @param    date   the date
 **
 ** @return   the output stream
 **/
ostream&
operator<<(ostream& ostr, Date const& date)
{
  ostr << date.year << '-'
    << (date.month < 10 ? "0" : "") << date.month << '-'
    << (date.day < 10 ? "0" : "") << date.day;
  return ostr;
} // ostream& operator<<(ostream& ostr, Date const& date)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether the dates are equal
 **/
bool
operator==(Date const& lhs, Date const& rhs)
{
  return (   (lhs.day == rhs.day)
          && (lhs.month == rhs.month)
          && (lhs.year == rhs.year));
} // bool operator==(Date lhs, Date rhs)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether the dates differ
 **/
bool
operator!=(Date const& lhs, Date const& rhs)
{
  return (   (lhs.day != rhs.day)
          || (lhs.month != rhs.month)
          || (lhs.year != rhs.year));
} // bool operator!=(Date lhs, Date rhs)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether 'lhs' is older than 'rhs'
 **/
bool
operator<(Date const& lhs, Date const& rhs)
{
  if (lhs.year < rhs.year)
    return true;
  if (lhs.year > rhs.year)
    return false;

  // now we have 'lhs.year == rhs.year'
  if (lhs.month < rhs.month)
    return true;
  if (lhs.month > rhs.month)
    return false;

  // now we have 'lhs.month == rhs.month'
  if (lhs.day < rhs.day)
    return true;
  if (lhs.day > rhs.day)
    return false;

  return false;
} // bool operator<(Date lhs, Date rhs)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether 'lhs' is newer than 'rhs'
 **/
bool
operator>(Date const& lhs, Date const& rhs)
{
  if (lhs.year > rhs.year)
    return true;
  if (lhs.year < rhs.year)
    return false;

  // now we have 'lhs.year == rhs.year'
  if (lhs.month > rhs.month)
    return true;
  if (lhs.month < rhs.month)
    return false;

  // now we have 'lhs.month == rhs.month'
  if (lhs.day > rhs.day)
    return true;
  if (lhs.day < rhs.day)
    return false;

  return false;
} // bool operator>(Date lhs, Date rhs)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether 'lhs' is not newer than 'rhs'
 **/
bool
operator<=(Date const& lhs, Date const& rhs)
{
  if (lhs.year < rhs.year)
    return true;
  if (lhs.year > rhs.year)
    return false;

  // now we have 'lhs.year == rhs.year'
  if (lhs.month < rhs.month)
    return true;
  if (lhs.month > rhs.month)
    return false;

  // now we have 'lhs.month == rhs.month'
  if (lhs.day < rhs.day)
    return true;
  if (lhs.day > rhs.day)
    return false;

  return true;
} // bool operator<=(Date lhs, Date rhs)

/** -> result
 **
 ** @param    lhs   first date
 ** @param    rhs   second date
 **
 ** @return   whether 'lhs' is not older as 'rhs'
 **/
bool
operator>=(Date const& lhs, Date const& rhs)
{
  if (lhs.year > rhs.year)
    return true;
  if (lhs.year < rhs.year)
    return false;

  // now we have 'lhs.year == rhs.year'
  if (lhs.month > rhs.month)
    return true;
  if (lhs.month < rhs.month)
    return false;

  // now we have 'lhs.month == rhs.month'
  if (lhs.day > rhs.day)
    return true;
  if (lhs.day < rhs.day)
    return false;

  return true;
} // bool operator>=(Date lhs, Date rhs)

/** -> result
 **
 ** @param    monthname   name of the month (english, short form, ex: Dec)
 **
 ** @return   number of the month 'monthname',
 **          0 on failure
 **/
unsigned
monthtou(std::string const& monthname)
{
  if (monthname == "Jan")
    return 1;
  if (monthname == "Feb")
    return 2;
  if (monthname == "Mar")
    return 3;
  if (monthname == "Apr")
    return 4;
  if (monthname == "May")
    return 5;
  if (monthname == "Jun")
    return 6;
  if (monthname == "Jul")
    return 7;
  if (monthname == "Aug")
    return 8;
  if (monthname == "Sep")
    return 9;
  if (monthname == "Oct")
    return 10;
  if (monthname == "Nov")
    return 11;
  if (monthname == "Dec")
    return 12;

  // Error
  return 0;
} // unsigned monthtou(std::string monthname)
