/**
 **   Copyright (C) 2001 – 2018 by Diether Knof
 **
 **   This program is free software; you can redistribute it and/or 
 **   modify it under the terms of the GNU General Public License as 
 **   published by the Free Software Foundation; either version 2 of 
 **   the License, or (at your option) any later version.
 **
 **   This program is distributed in the hope that it will be useful,
 **   but WITHOUT ANY WARRANTY; without even the implied warranty of
 **   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **   GNU General Public License for more details. 
 **   You can find this license in the file 'gpl.txt'.
 **
 **   You should have received a copy of the GNU General Public License
 **   along with this program; if not, write to the Free Software
 **   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 **   MA  02111-1307  USA
 **
 **  Contact:
 **    Diether Knof dknof@posteo.de
 **/

#pragma once

#include <memory>
using std::unique_ptr;
#include <string>
using std::string;
using namespace std::literals::string_literals;

// Optionales Element
template<typename T>
class optional final {
  public:
    optional() = default;
    optional(T const& value);
    optional(optional const& value);
    optional& operator=(optional const& value);
    optional& operator=(T const& value);

    ~optional() = default;

    explicit operator bool() const;
    T& value();
    T const& value() const;

    T const& operator*() const;
    T& operator*();
    T const* operator->() const;
    T* operator->();

    void reset();

    bool operator==(optional const& rhs) const;
    bool operator!=(optional const& rhs) const;
    bool operator==(T const& rhs) const;
    bool operator!=(T const& rhs) const;

  private:
    unique_ptr<T> value_;
}; // class optional


// constructor
template<typename T>
optional<T>::optional(T const& value) :
  value_(make_unique<T>(value))
{ }

// constructor
template<typename T>
optional<T>::optional(optional<T> const& value) :
  value_(value ? make_unique<T>(*value) : nullptr)
{ }

// copy operator
template<typename T>
optional<T>&
optional<T>::operator=(optional<T> const& value)
{
  if (!value)
    this->value_.reset();
  else
    this->value_ = make_unique<T>(*value);
  return *this;
}

// copy operator
template<typename T>
optional<T>&
optional<T>::operator=(T const& value)
{
  this->value_ = make_unique<T>(value);
  return *this;
}

// boolean converter
template<typename T>
optional<T>::operator bool() const
{
  return this->value_.operator bool();
}

// Wert zurückgeben
template<typename T>
T const&
optional<T>::value() const
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return *this->value_;
}

template<typename T>
T&
optional<T>::value()
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return *this->value_;
}

// Wert zurückgeben
template<typename T>
T const&
optional<T>::operator*() const
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return *this->value_;
}

template<typename T>
T&
optional<T>::operator*()
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return *this->value_;
}

template<typename T>
T const*
optional<T>::operator->() const
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return this->value_.operator->();
}

template<typename T>
T*
optional<T>::operator->()
{
  if (!*this)
    throw "optional: kein Wert enthalten"s;
  return this->value_.operator->();
}

template<typename T>
void
optional<T>::reset()
{
  this->value_.reset();
}

// comparison
template<typename T>
bool
optional<T>::operator==(optional const& rhs) const
{
  if (!this->value_)
    return !rhs.value_;
  if (!rhs.value_)
    return false;
  return (**this == *rhs);
}

// comparison
template<typename T>
bool
optional<T>::operator!=(optional const& rhs) const
{
  if (!this->value_)
    return rhs.value_;
  if (!rhs.value_)
    return true;
  return (**this != *rhs);
}

// comparison
template<typename T>
bool
optional<T>::operator==(T const& rhs) const
{
  if (!this->value_)
    return false;
  return (**this == rhs);
}

// comparison
template<typename T>
bool
optional<T>::operator!=(T const& rhs) const
{
  if (!this->value_)
    return true;
  return (**this != rhs);
}
