/**
 **   Copyright (C) 2006 – 2018 by Diether Knof
 **
 **   This program is free software; you can redistribute it and/or 
 **   modify it under the terms of the GNU General Public License as 
 **   published by the Free Software Foundation; either version 2 of 
 **   the License, or (at your option) any later version.
 **
 **   This program is distributed in the hope that it will be useful,
 **   but WITHOUT ANY WARRANTY; without even the implied warranty of
 **   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **   GNU General Public License for more details. 
 **   You can find this license in the file 'gpl.txt'.
 **
 **   You should have received a copy of the GNU General Public License
 **   along with this program; if not, write to the Free Software
 **   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 **   MA  02111-1307  USA
 **
 **  Contact:
 **    Diether Knof dknof@posteo.de
 **/

#pragma once

#include <time.h>
#include <memory>

using MSecCounter = long unsigned;

/**
 ** a class to check the runtime
 ** uses 'clock()'
 **
 ** @todo     thread savety (start/stop: either save the thread id or do not use)
 **/
class StopWatch {
  public:
    /** A proxy for simplier measuring
     ** Starts at construct, stop at destruct
     **/
    class StartStopProxy {
      public:
        StartStopProxy() = delete;
        StartStopProxy(StopWatch& stop_watch);
        ~StartStopProxy();
      private:
        StopWatch* const stop_watch;
    }; // class StartStopProxy

  private:
    // returns the time betwenn 'start_time' and 'now'
    static MSecCounter diff_to_now(clock_t start_time);

  public:
    StopWatch();
    StopWatch(StopWatch const& stop_watch) = default;
    StopWatch& operator=(StopWatch const&) = delete;
    ~StopWatch();

    // create a start stop proxy
    std::unique_ptr<StartStopProxy> start_stop_proxy();

    // start the time counter (the time is added)
    void start();
    // stop the time counter
    void stop();
    // reset the time counter
    void reset();

    // whether the stop watch is running
    bool running() const;

    // elapsed time in mili seconds
    unsigned long msec() const;
    // how many calls have been made
    int calls() const;

  private:
    // the time when 'add' was called
    clock_t start_time;
    // elapsed time in mili seconds
    MSecCounter msec_ = 0;
    // how many 'start's still remain
    int depth = 0;

    // how many calls (of start) have been made
    int calls_ = 0;
}; // class StopWatch
