/**
 **   Copyright (C) 2001 – 2018 by Diether Knof
 **
 **   This program is free software; you can redistribute it and/or 
 **   modify it under the terms of the GNU General Public License as 
 **   published by the Free Software Foundation; either version 2 of 
 **   the License, or (at your option) any later version.
 **
 **   This program is distributed in the hope that it will be useful,
 **   but WITHOUT ANY WARRANTY; without even the implied warranty of
 **   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **   GNU General Public License for more details. 
 **   You can find this license in the file 'gpl.txt'.
 **
 **   You should have received a copy of the GNU General Public License
 **   along with this program; if not, write to the Free Software
 **   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 **   MA  02111-1307  USA
 **
 **  Contact:
 **    Diether Knof dknof@posteo.de
 **/

#include "version.h"

#include <fstream>
#include <sstream>
using std::ostream;
using std::istream;
using std::ifstream;
using std::string;
using std::istringstream;
using std::ostringstream;
using std::unique_ptr;
using std::make_unique;

/** Constructor
 **
 ** @param    filename   the filename
 **
 ** @return   pointer to a new created version
 **   NULL in the case of an error
 **/
unique_ptr<Version>
  Version::create_from_file(string const filename)
  {
    ifstream istr(filename.c_str());
    if (istr.fail())
      return {};

    return Version::create(istr);
  } // static unique_ptr<Version> create_from_file(string filename)

/** Constructor
 **
 ** @param    istr   the input stream
 **
 ** @return   pointer to a new created version
 **   NULL in the case of an error
 **
 ** @todo   read whether it is a release
 **/
unique_ptr<Version>
Version::create(istream& istr)
{
  unsigned major = 0;
  unsigned minor = 0;
  unsigned micro = 0;
  char patch = '\0';
  auto date = Date{__DATE__};
  bool release = false;
  unsigned rc = 0;


  // read the date from the stream.
  // Format: 'major.minor.micro  (date)'
  // Example: '0.6.1  (2003-12-21)'
  istr >> major;
  if (istr.get() != '.')
    return {};
  istr >> minor;
  if (istr.get() != '.')
    return {};
  istr >> micro;
  if (istr.peek() != ' ')
    patch = istr.get();

  while (istr.peek() == ' ') {
    istr.get();
  }


  { // Datum in Klammern
    if (istr.peek() != '(') {
      return make_unique<Version>(major, minor, micro, patch, date, release);
    }
    istr.get();

    auto date2 = Date::create(istr);
    if (!date2) {
      return make_unique<Version>(major, minor, micro, patch, date, release);
    }
    date = *date2;

    if (istr.get() != ')') {
      return {};
    }
  } // Datum in Klammern

  while (istr.peek() == ' ') {
    istr.get();
  }

  { // Versionsbeschreibung in Klammern
    if (istr.peek() != '(') {
      return make_unique<Version>(major, minor, micro, patch, date, release);
    }
    istr.get();

    string release_str;
    while (   (istr.peek() != ')')
           && istr.good())
      release_str += istr.get();

    if (   (release_str != "release")
        && (release_str != "developer version")) {
      return make_unique<Version>(major, minor, micro, patch, date, release);
    }

    if (release_str == "release") {
      release = true;
    } else if (string(release_str, 0, 2) == "rc") {
      release = true;
      rc = stoul(string(release_str, 2, string::npos));
    }
  } // Versionsbeschreibung in Klammern

  while (istr.peek() == ' ') {
    istr.get();
  }

  { // Beschreibung
    if (istr.peek() != '(') {
      if (rc)
        return make_unique<Version>(major, minor, micro, date, release, rc);
      else
        return make_unique<Version>(major, minor, micro, patch, date, release);
    }
    istr.get();

    string description;
    while (   (istr.peek() != ')')
           && istr.good()) {
      description += istr.get();
    }
    if (rc)
      return make_unique<Version>(major, minor, micro, date, release, rc, description);
    else
      return make_unique<Version>(major, minor, micro, patch, date, release, description);
  } // Beschreibung
} // static unique_ptr<Version> Version::create(string filename)

/** Constructor
 **
 ** @param    string   string to read the version von
 **
 ** @return   pointer to a new created version
 **   nullptr in the case of an error
 **/
unique_ptr<Version>
Version::create(string const& s)
{
  istringstream istr(s.c_str());

  return Version::create(istr);
} // static unique_ptr<Version> Version::create(string s)


/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'
 **
 ** @param    major   major version number
 ** @param    minor   minor version number
 ** @param    micro   micro version number
 ** @param    patch   character to denote the patch version
 **                    default: none
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 char const patch) :
  major_(major),
  minor_(minor),
  micro_(micro),
  patch_(patch)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'-'release'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    date          the date
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 Date const& date) :
  major_(major),
  minor_(minor),
  micro_(micro),
  date_(date)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'-'release'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    patch         character to denote the patch version
 ** @param    date          the date
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 char const patch,
                 Date const& date) :
  major_(major),
  minor_(minor),
  micro_(micro),
  patch_(patch),
  date_(date)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'-'release'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    date          the date
 ** @param    release         whether this is a release
 ** @param    description   description
 **                       default: ""
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 Date const& date,
                 bool const release,
                 std::string const description) :
  major_(major),
  minor_(minor),
  micro_(micro),
  date_(date),
  release_(release),
  description_(description)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'-'release'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    date          the date
 ** @param    release       whether this is a release (must be true)
 ** @param    rc            release candidate
 ** @param    description   description
 **                       default: ""
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 Date const& date,
                 bool const release,
                 unsigned const rc,
                 std::string const description) :
  major_(major),
  minor_(minor),
  micro_(micro),
  date_(date),
  release_(release),
  rc_(rc),
  description_(description)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro''patch'-'release'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    patch         character to denote the patch version
 ** @param    date          the date
 ** @param    release       whether this is a release
 ** @param    description   description
 **                       default: ""
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 char const patch,
                 Date const& date,
                 bool const release,
                 std::string const description) :
  major_(major),
  minor_(minor),
  micro_(micro),
  patch_(patch),
  date_(date),
  release_(release),
  description_(description)
{ }

/** Constructor
 ** the version number will be 'major'.'minor'.'micro'
 **
 ** @param    major         major version number
 ** @param    minor         minor version number
 ** @param    micro         micro version number
 ** @param    description   the description
 **/
Version::Version(unsigned const major,
                 unsigned const minor, unsigned const micro,
                 std::string const description) :
  major_(major),
  minor_(minor),
  micro_(micro),
  description_(description)
{ }

/** conversion to a string
 **
 ** @return   the corresponding string
 **/
Version::operator string() const
{
  ostringstream ostr;

  ostr << *this;

  return ostr.str();
} // Version::operator string string() const

/** conversion of the number to a string
 **
 ** @return   the corresponding string
 **/
string
Version::number_to_string() const
{
  ostringstream ostr;

  ostr << this->major_
    << "."
    << this->minor_
    << "."
    << this->micro_;

  if (this->patch_ != '\0')
    ostr << this->patch_;

  return ostr.str();
} // string Version::number_to_string() const

/** write the version to the stream
 **
 ** @param    ostr   output stream to write to
 **
 ** @return   the output stream
 **/
ostream&
Version::write(ostream& ostr) const
{
  ostr << this->number_to_string();

  ostr << "  (" << this->date_ << ')';

  if (this->release_) {
    if (this->rc_)
      ostr << " (rc" << this->rc_ << ")";
    else
      ostr << " (release)";
  } else {
    ostr << " (developer version)";
  }

  if (!this->description_.empty())
    ostr << " (" << this->description_ << ")";

  return ostr;
} // ostream& Version::write(ostream& ostr) const

/** write 'version' to stream 'ostr'
 **
 ** @param    ostr   output stream to write to
 ** @param    version   the version
 **
 ** @return   the output stream
 **/
ostream&
operator<<(ostream& ostr, Version const& version)
{
  version.write(ostr);

  return ostr;
} // ostream& operator<<(ostream& ostr, Version const& version)

/** -> result
 **
 ** @param    v1   first version
 ** @param    v2   second version
 **
 ** @return   whether the versions are equal
 **/
bool
operator==(Version const& v1, Version const& v2)
{
  return (   (v1.micro_ == v2.micro_)
          && (v1.minor_ == v2.minor_)
          && (v1.major_ == v2.major_)
          && (v1.patch_ == v2.patch_)
          && (v1.date_ == v2.date_)
          && (v1.release_ == v2.release_)
          && (v1.rc_ == v2.rc_)
          && (v1.description_ == v2.description_));
} // bool operator==(Version const& v1, Version const& v2)

/** -> result
 **
 ** @param    v1   first version
 ** @param    v2   second version
 **
 ** @return   whether 'v1' is older than 'v2'
 **/
bool
operator<(Version const& v1, Version const& v2)
{

  if (v1.major_ < v2.major_)
    return true;
  if (v1.major_ > v2.major_)
    return false;

  // now we have 'v1.major == v2.major'
  if (v1.minor_ < v2.minor_)
    return true;
  if (v1.minor_ > v2.minor_)
    return false;

  // now we have 'v1.minor == v2.minor'
  if (v1.micro_ < v2.micro_)
    return true;
  if (v1.micro_ > v2.micro_)
    return false;

  // now we have 'v1.micro == v2.micro'
  if (v1.patch_ < v2.patch_)
    return true;
  if (v1.patch_ > v2.patch_)
    return false;

  // now we have 'v1.patch == v2.patch'
  if (v1.release_ && v2.release_) {
    if (!v1.rc_)
      return false;
    if (!v2.rc_)
      return true;
    if (v1.rc_ < v2.rc_)
      return true;
    if (v1.rc_ > v2.rc_)
      return false;
  }

  // now the version numbers are equal
  return (v1.date_ < v2.date_);
} // bool operator<(Version const& v1, Version const& v2)

/** -> result
 **
 ** @param    v1   first version
 ** @param    v2   second version
 **
 ** @return   whether 'v1' is not newer than 'v2'
 **/
bool
operator<=(Version const& v1, Version const& v2)
{
  return ((v1 < v2) || (v1 == v2));
} // bool operator<=(Version const& v1, Version const& v2)

/** -> result
 **
 ** @param    v1   first version
 ** @param    v2   second version
 **
 ** @return   whether 'v1' is newer than 'v2'
 **/
bool
operator>(Version const& v1, Version const& v2)
{
  return (v2 < v1);
} // bool operator>(Version const& v1, Version const& v2)

/** -> result
 **
 ** @param    v1   first version
 ** @param    v2   second version
 **
 ** @return   whether 'v1' is not older than 'v2'
 **/
bool
operator>=(Version const& v1, Version const& v2)
{
  return (v2 <= v1);
} // bool operator>=(Version v1, Version v2)
